/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef SM_CAMERA_H_
#define SM_CAMERA_H_

#include <nw/gfx.h>
#include "../include/SmMessage.h"
#include "../include/SmPadStatus.h"
#include "../include/SmRectCollision.h"
#include "../include/SmTouchPanelStatus.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
//------------------------------------------------------------------------------
// Camera data structure.
// Borrowed from Viewer.
struct CameraInfo
{
/*
    float           AspectRatio;
    float           Height;
    float           Rotate;
    float           Twist;*/

    float           Near;
    float           Far;
    float           Fovy;
    nw::math::VEC3  Position;
    nw::math::VEC3  TargetPos;
    nw::math::VEC3  UpVector;


    float           RotationInit;
    float           RotationStep;
    float           RotationMax;
    float           TranslationInit;
    float           TranslationStep;
    float           TranslationMax;
    float           TiltInit;
    float           DollyInit;

    bool            TumbleXInverse;
    bool            TumbleYInverse;
    bool            TrackXInverse;
    bool            TrackYInverse;
    bool            DollyInverse;
    bool            DollyTargetInverse;
    bool            TiltXInverse;
    bool            TiltYInverse;
    CameraInfo();
};



//------------------------------------------------------------------------------
// Camera class.
class SmCamera : public SmMessage
{
    friend class GfxCtrl;

private:
    // Constructor
    SmCamera( nw::gfx::SceneNode* parent, nw::gfx::Camera* camera );

    // Destructor
    ~SmCamera();

public:
    // Camera animation array.
    #define ANIM_COUNT (32)
    typedef nw::ut::FixedSizeArray<nw::gfx::AnimEvaluator*, ANIM_COUNT> CameraAnimEvaluatorArray;

    // Configure the view.
    void SetPosition( nw::math::VEC3* position )
    {
        NW_NULL_ASSERT( position );
        NW_NULL_ASSERT( m_Camera );

        // The view.
        m_Position = *position;
        m_Camera->Transform().SetTranslate( m_Position );
    }

    void SetPadCameraPosition( nw::math::VEC3* position )
    {
        NW_NULL_ASSERT( position );
        NW_NULL_ASSERT( m_PadCamera );
        m_PadCamera->Transform().SetTranslate( *position );
    }



    // Get the view.
    void GetPosition( nw::math::VEC3* position )
    {
        NW_NULL_ASSERT( position );
        *position = m_Position;
    }

    // Set the viewpoint.
    void SetTarget( nw::math::VEC3* target )
    {
        NW_NULL_ASSERT( target );
        NW_NULL_ASSERT( m_Camera );

        m_Target = *target;

        nw::gfx::ResCameraViewUpdater resViewUpdater = m_Camera->GetViewUpdater()->GetResource();
        nw::gfx::ResAimTargetViewUpdater resAimTargetViewUpdater =
            nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>(resViewUpdater);
        nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater =
            nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>(resViewUpdater);
        nw::gfx::ResRotateViewUpdater resRotateViewUpdater =
            nw::gfx::ResDynamicCast<nw::gfx::ResRotateViewUpdater>(resViewUpdater);

        // Store the viewpoint.
        if (resAimTargetViewUpdater.IsValid())
        {
            resAimTargetViewUpdater.SetTargetPosition( m_Target );
        }
        else if (resLookAtTargetViewUpdater.IsValid())
        {
            resLookAtTargetViewUpdater.SetTargetPosition( m_Target );
        }
        else if (resRotateViewUpdater.IsValid())
        {
            //
            NW_FATAL_ERROR("Unxxxxx type of resource view updater");
        }
        else
        {
            NW_FATAL_ERROR("Invalid type of resource view updater");
        }
    }

    void SetPadCameraTarget( nw::math::VEC3* target )
    {
        NW_NULL_ASSERT( m_PadCamera );

        nw::gfx::LookAtTargetViewUpdater* viewUpdater =
            nw::ut::DynamicCast<nw::gfx::LookAtTargetViewUpdater*>(m_PadCamera->GetViewUpdater());
        nw::gfx::ResLookAtTargetViewUpdater resViewUpdater =
            nw::gfx::ResStaticCast<nw::gfx::ResLookAtTargetViewUpdater>(viewUpdater->GetResource());

       resViewUpdater.SetTargetPosition(*target);
    }

    // Get the viewpoint.
    void GetTarget( nw::math::VEC3* target )
    {
        NW_NULL_ASSERT( target );
        *target = m_Target;
    }

    // Configure the camera.
    void SetGxCamera( nw::gfx::Camera* camera );

    // Get the camera.
    nw::gfx::Camera* GetGxCamera();

    nw::gfx::Camera* GetBaseCamera()
    {
        return m_Camera;
    }

    nw::gfx::Camera* GetPadCamera()
    {
        return m_PadCamera;
    }

    bool IsPadCamera()
    {
        return m_PadCameraAttached;
    }

    /*
        Animation features.
    */
    // Add animations.
    void AddAnimEvaluator( nw::gfx::AnimEvaluator* animEvaluator );

    // Get the number of animations.
    uint GetAnimEvaluatorNum() const;

    // Cancel animations.
    bool DetachAnimEvaluator();

    // Configure animations.
    bool SetAnimEvaluatorNo( uint animEvaluatorNo );

    // Configure the animation frame.
    void SetAnimationFrame( f32 setFrame );

    // Advance the animation frame.
    bool AddAnimationFrame( f32 addFrame, bool loop = true );

    // Get the animation frame.
    f32 GetAnimationFrame() const;

    // Receive message
    virtual bool ReceveMessage( SmMessageType type, void* object, uint targetId );

    // Operate the camera the same as Viewer.
    void MoveCamera3D( SmPadStatus* padStaus, SmTouchPanelStatus* panelStatus );

    // Update camera matrix.
    void UpdateMatrix();

private:
    nw::gfx::SceneNode*         m_Parent;
    nw::gfx::Camera*            m_Camera;
    nw::gfx::Camera*            m_PadCamera;
    bool                        m_PadCameraAttached;

    nw::math::VEC3              m_Position;
    nw::math::VEC3              m_Target;

    bool                        m_IsAnimPlaying;
    f32                         m_AnimationFrame;
    s32                         m_CurrentAnimEvNo;
    CameraAnimEvaluatorArray    m_CameraAnimEvaluatorArray;

    f32                         mRotation;
    f32                         mTranslation;
    CameraInfo                  mCameraInfo;
    s32                         mTouchPanelX;
    s32                         mTouchPanelY;

    bool                        m_IsGrab;
    SmRectCollision             m_Collision;
};
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif


#endif  // SM_CAMERA_H_
