/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef SM_INITIAL_FILE_PARSER_H_
#define SM_INITIAL_FILE_PARSER_H_

#include <ctype.h>
#include <stdlib.h>

#include <nn/gx.h>
#include "../include/SmBase.h"
#include "../include/SmFile.h"



#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
//------------------------------------------------------------------------------
//
class SmParam
{
public:
    enum
    {
        SM_PARAM_TYPE_U32       = 0,
        SM_PARAM_TYPE_S32       = 1,
        SM_PARAM_TYPE_F32       = 2,
        SM_PARAM_TYPE_STRING    = 3
    };
    typedef u32 SmParamType;

    //
    SmParam()
    : m_Type( SM_PARAM_TYPE_U32 ),
      m_u32Param( 0 )
    {
        m_String = "";
    }

    //
    SmParamType GetType() const { return m_Type; }

    //
    u32 GetU32() const
    {
        switch( m_Type )
        {
        case SM_PARAM_TYPE_U32:
            {
                return m_u32Param;
            }
        case SM_PARAM_TYPE_S32:
            {
                return static_cast<u32>(m_s32Param);
            }
        case SM_PARAM_TYPE_F32:
            {
                return static_cast<u32>(m_f32Param);
            }
        case SM_PARAM_TYPE_STRING:
            {
                return static_cast<u32>(atoi(m_String.c_str()));
            }

        }
        return 0;
    }

    s32 GetS32() const
    {
        switch( m_Type )
        {
        case SM_PARAM_TYPE_U32:
            {
                return static_cast<s32>(m_u32Param);
            }
        case SM_PARAM_TYPE_S32:
            {
                return m_s32Param;
            }
        case SM_PARAM_TYPE_F32:
            {
                return static_cast<s32>(m_f32Param);
            }
        case SM_PARAM_TYPE_STRING:
            {
                return static_cast<s32>(atoi(m_String.c_str()));
            }
        }
        return 0;
    }

    f32 GetF32() const
    {
        switch( m_Type )
        {
        case SM_PARAM_TYPE_U32:
            {
                return static_cast<f32>(m_u32Param);
            }
        case SM_PARAM_TYPE_S32:
            {
                return static_cast<f32>(m_s32Param);
            }
        case SM_PARAM_TYPE_F32:
            {
                return m_f32Param;
            }
        case SM_PARAM_TYPE_STRING:
            {
                return static_cast<f32>(atof(m_String.c_str()));
            }
        }
        return 0;
    }

    const char* GetString() const
    {
        switch( m_Type )
        {
        case SM_PARAM_TYPE_U32:
        case SM_PARAM_TYPE_S32:
        case SM_PARAM_TYPE_F32:
            {
//                m_String.Format( "%d", m_u32Param );
//                return m_String.c_str();
                NW_FATAL_ERROR( "no support!" );
            }

        case SM_PARAM_TYPE_STRING:
            {
                return m_String.c_str();
            }
        }

        return NULL;
    }

    //
    void SetParam( u32 param )
    {
        m_Type      = SM_PARAM_TYPE_U32;
        m_u32Param  = param;
    }

    void SetParam( s32 param )
    {
        m_Type      = SM_PARAM_TYPE_S32;
        m_s32Param  = param;
    }

    void SetParam( f32 param )
    {
        m_Type      = SM_PARAM_TYPE_F32;
        m_f32Param  = param;
    }

    void SetParam( const char* param )
    {
        m_Type      = SM_PARAM_TYPE_STRING;
        m_String    = param;
    }

private:
    SmParamType m_Type;
    union
    {
        u32 m_u32Param;
        s32 m_s32Param;
        f32 m_f32Param;
    };
    std::string m_String;
};


//---------------------------------------------------------------------------
//
class SmCmd : public SmBase
{
friend class SmInitialFileParser;

private:
    //
    struct SmData
    {
        std::string name;       // Name
        SmParam     param;      // Value
    };

public:
    //
    SmCmd()
    : m_Parent( NULL ),
      m_Child( NULL ),
      m_Next( NULL )
    {
        Init();
    }

    //
    void Init()
    {
        m_Command = "NULL";

        for ( int i=0; i<16; ++i )
        {
            m_Data[i].name = "null";
            m_Data[i].param.SetParam( 0 );
        }
    }

    //
    const char* GetCommand() const
    {
        return m_Command.c_str();
    }

    //
    f32 GetF32( const char* strName ) const
    {
        for ( int i=0; i<16; ++i )
        {
            const char* name = m_Data[i].name.c_str();
            if ( strcmp( name, strName ) == 0 )
            {
                return m_Data[i].param.GetF32();
            }
        }

        return 0.f;
    }

private:
    SmCmd*          m_Parent;
    SmCmd*          m_Child;
    SmCmd*          m_Next;

    std::string     m_Command;     // <XXXX
    SmData          m_Data[16];    // tx = 10  (fixed to 16)
};


//---------------------------------------------------------------------------
//
class SmInitialFileParser : public  SmBase
{
    typedef nw::ut::FixedSizeArray<SmCmd*, 64> SmCmdArray;

public:
    SmInitialFileParser();
    ~SmInitialFileParser();

    //
    bool    IsGettable();

    //
    SmCmd   GetCmd();

    //
    bool Load( const wchar_t* configFileName );

    //
    u32 GetCmdNum() const { return m_CmdArray.size(); }

    //
    const SmCmd* GetCmd( u32 no ) { return m_CmdArray[no]; }

private:
    //
    void _createCmdTree();

private:
    SmFile          m_File;

    char*           m_Top;
    char*           m_Current;

    SmCmdArray      m_CmdArray;
};
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif



#endif // SM_INITIAL_FILE_PARSER_H_
