/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef SM_LAYOUT_H_
#define SM_LAYOUT_H_

#include <nw/lyt.h>

#include "../include/SmDef.h"
#include "../include/SmBase.h"
#include "../include/SmRectCollision.h"


#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
//------------------------------------------------------------------------------
//
class SmLayoutResource : public SmBase
{
public:
    //
    #define ANIM_COUNT (32)
    typedef nw::ut::FixedSizeArray<nw::lyt::AnimTransform*, ANIM_COUNT> AnimationArray;

public:
    //
    SmLayoutResource()
      : m_Layout( NULL )
    {
        m_AnimationArray.clear();
    }

    //
    ~SmLayoutResource()
    {
        SM_SAFE_DELETE( m_Layout );
    }

    //
    bool InitLayout( const void* bclyt, nw::lyt::ArcResourceAccessor* pResAccessor )
    {
        NW_NULL_ASSERT(!m_Layout);
        NW_NULL_ASSERT(bclyt);
        NW_NULL_ASSERT(pResAccessor);

        m_Layout = new nw::lyt::Layout();
        NW_NULL_ASSERT( m_Layout );

        if ( !m_Layout->Build( bclyt, pResAccessor ) )
        {
            NW_FATAL_ERROR("can not build layout resource.\n");
        }

        return true;
    }

    //
    bool AddAnimation( const void* bclan, nw::lyt::ArcResourceAccessor* pResAccessor )
    {
        NW_NULL_ASSERT(m_Layout);
        NW_NULL_ASSERT(bclan);
        NW_NULL_ASSERT(pResAccessor);

        nw::lyt::AnimTransform* pAnimTrans = m_Layout->CreateAnimTransform( bclan, pResAccessor );
        if ( !pAnimTrans ) return false;
        m_AnimationArray.push_back( pAnimTrans );

        return true;
    }

    //
    uint GetAnimationNum() const
    {
        return m_AnimationArray.size();
    }

    //
    f32 GetAnimationFrameSize( uint animNo ) const
    {
        if ( animNo > GetAnimationNum() ) return 0.f;

        return m_AnimationArray[animNo]->GetFrameSize();
    }

    //
    nw::ut::Rect GetOriginalRect( const char* name ) const
    {
        nw::lyt::Pane* rootPane = m_Layout->GetRootPane();
        NW_NULL_ASSERT( rootPane );
        nw::lyt::Pane* pane = rootPane->FindPaneByName( name );
        NW_NULL_ASSERT( pane );
        return pane->GetPaneRect();
    }

    //
    nw::ut::Rect GetRect( const char* name );

    //
    nw::lyt::Pane* GetPane( const char* name ) const
    {
        nw::lyt::Pane* rootPane = m_Layout->GetRootPane();
        NW_NULL_ASSERT( rootPane );
        nw::lyt::Pane* pane = rootPane->FindPaneByName( name );
        return pane;
    }

    //
    nw::lyt::Layout* GetLayout() { return m_Layout; }

    //
    AnimationArray& GetAnimationArray() { return m_AnimationArray; }

private:
    nw::lyt::Layout*    m_Layout;
    AnimationArray      m_AnimationArray;
};


//------------------------------------------------------------------------------
//
class SmLayout : public SmBase
{
public:
    //
    SmLayout( SmLayoutResource* layoutResource );

    //
    ~SmLayout();

    //
    void SetPosition( f32 x, f32 y );

    //
    f32 GetPosX() const { return m_Position.x; }
    f32 GetPosY() const { return m_Position.y; }

    //
    //
    void SetPanePosition( nw::lyt::Pane* pane, f32 x, f32 y )
    {
        m_OffsetPane        = pane;
        m_OffsetPosition.x  = x;
        m_OffsetPosition.y  = y;
        m_OffsetPosition.z  = 0.f;
    }

    //
    bool SetAnimationNo( uint animTransformNo );

    //
    void SetAnimationFrame( f32 setFrame );

    //
    bool AddAnimationFrame( f32 addFrame, bool loop = true );

    //
    f32 GetAnimationFrame() const
    {
        return m_AnimationFrame;
    }

    //
    void SetVisible( bool visible )
    {
        m_Visible = visible;
    }

    //
    bool GetVisible() const
    {
        return m_Visible;
    }

    //
    void SetVisiblePane( const char* paneName, bool visible )
    {
        nw::lyt::Pane* pane = m_LayoutResource->GetPane( paneName );
        if( pane )
        {
            pane->SetVisible( visible );
        }
    }

    //
    SmLayoutResource* GetLayoutResource() const
    {
        return m_LayoutResource;
    }

    //
    bool SetupLayout();

    //
    void ResetLayout();

    //
    void SetString( const wchar_t* string )
    {
        // todo
        m_String = string;
    }

private:
    SmLayoutResource*           m_LayoutResource;
    nw::lyt::Pane*              m_OffsetPane;

    nw::math::VEC3              m_Position;
    nw::math::VEC3              m_OffsetPosition;

    f32                         m_AnimationFrame;
    u32                         m_AnimationNo;
    bool                        m_Visible;
    const wchar_t*              m_String;
};
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif


#endif  // SM_LAYOUT_H_
