/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#ifndef SM_MESSAGE_H_
#define SM_MESSAGE_H_


#include "../include/SmBase.h"
#include "../include/SmPrimitive.h"


#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
// The message type.
enum
{
    SM_MESSAGE_NONE = 0,                    //
    SM_MESSAGE_TUCHPANEL_PRESS = 1,         //
    SM_MESSAGE_TUCHPANEL_RELEASE = 2,       //
    SM_MESSAGE_TUCHPANEL_MOTION = 3,        //
    SM_MESSAGE_PAD_UPDATE = 4,              //
    SM_MESSAGE_SCENE_UPDATE = 5,            //
    SM_MESSAGE_UPDATE_PARAM = 6             //
};
typedef u32 SmMessageType;

//------------------------------------------------------------------------------
// The message class.
class SmMessage : public SmBase
{
public:

    // Constructor
    SmMessage( SmMessage* parent )
               : m_Parent( parent ),
                 m_Child( NULL ),
                 m_Next( NULL ),
                 m_Enable( true ),
                 m_Visible( true ),
                 m_Target( NULL ),
                 m_Id( 0 ){}

    // Send message.
    virtual bool SendMessage(   SmMessageType   type,
                                void*           object = NULL,
                                uint            targetId = 0 )
    {
        bool ret = false;

        // Execute self.
        if ( m_Enable )
        {
            if ( this->ReceveMessage( type, object, targetId ) )
            {
                return true;
            }
        }

        // Go deeper if there are children.
        if ( m_Enable && m_Child )
        {
            ret = m_Child->sendMessageSub( type, object, targetId );
            if ( ret ) return true;
        }

        return ret;
    }

    // Receive message.
    // Do not receive subsequent messages if true.
    virtual bool ReceveMessage( SmMessageType   type,
                                void*           object = NULL,
                                uint            targetId = 0 )
    {
        NW_UNUSED_VARIABLE(type);
        NW_UNUSED_VARIABLE(object);
        NW_UNUSED_VARIABLE(targetId);
        return false;
    }

    // Set the target.
    void SetTarget( SmMessage* target, uint id )
    {
        m_Target    = target;
        m_Id        = id;
    }

    // Set the parent.
    void SetParent( SmMessage* parent )
    {
        m_Parent = parent;
    }

    // Get the parent.
    SmMessage* GetParent()
    {
        return m_Parent;
    }

    // Add child.
    void AddChild( SmMessage* child )
    {
        child->SetParent( this );

        if ( !m_Child )
        {
            m_Child = child;
        }
        else
        {
            // Get last Next of children and add child using that Next.
            //
            SmMessage* tmp = m_Child->getNextTail();
            tmp->setNext( child );
        }
    }

    // Execute drawing routines.
    virtual void Render(){ return; }

    // Draw the entire tree.
    void DrawTree()
    {
        // Execute self.
        if ( m_Visible )
        {
            this->Render();
        }

        // Process siblings, too.
        SmMessage* next = this->getNext();
        while( next )
        {
            if ( next->m_Visible )
            {
                next->Render();

                if ( next->m_Child )
                {
                    next->m_Child->DrawTree();
                }
            }
            next = next->getNext();
        }

        // Go deeper if there are children.
        if ( m_Visible && m_Child )
        {
            m_Child->DrawTree();
        }
    }

    // Valid/invalid
    void SetEnable( bool enable )
    {
        // Execute self.
        m_Enable = enable;
        {
            SetVisible( m_Enable );
        }

        // Go deeper if there are children.
        if ( m_Child )
        {
            m_Child->setEnableFunc( enable );
        }

        return;
    }

    bool GetEnable()
    {
        return m_Enable;
    }

    virtual void SetVisible( bool visible )
    {
        m_Visible = visible;
    }

private:
    bool sendMessageSub( SmMessageType   type,
                          void*           object,
                          uint            targetId )
    {
        bool ret = false;

        // Execute self.
        if ( m_Enable )
        {
            if ( this->ReceveMessage( type, object, targetId ) )
            {
                return true;
            }
        }

        // Process siblings, too.
        SmMessage* next = this->getNext();
        while( next )
        {
            if ( next->m_Enable )
            {
                ret = next->ReceveMessage( type, object, targetId );
                if ( ret ) return true;

                if ( next->m_Child )
                {
                    ret = next->m_Child->sendMessageSub( type, object, targetId );
                    if ( ret ) return true;
                }
            }
            next = next->getNext();
        }

        // Go deeper if there are children.
        if ( m_Enable && m_Child )
        {
            ret = m_Child->sendMessageSub( type, object, targetId );
            if ( ret ) return true;
        }

        return ret;
    }

    void setEnableFunc( bool enable )
    {
        // Execute self.
        m_Enable = enable;
        {
            SetVisible( m_Enable );
        }

        // Process siblings, too.
        SmMessage* next = this->getNext();
        while( next )
        {
            next->setEnableFunc( enable );
            next = next->getNext();
        }

        // Go deeper if there are children.
        if ( m_Child )
        {
            m_Child->setEnableFunc( enable );
        }

        return;
    }

    void setNext( SmMessage* next )
    {
        m_Next = next;
    }

    SmMessage* getNext()
    {
        return m_Next;
    }

    SmMessage* getNextTail()
    {
        if ( !m_Next )
        {
            return this;
        }
        else
        {
            SmMessage* pret = this->getNext();
            while( pret )
            {
                if ( !pret->getNext() ) break;
                pret = pret->getNext();
            }
            return pret;
        }
    }


private:
    SmMessage*          m_Parent;
    SmMessage*          m_Child;
    SmMessage*          m_Next;

    bool                m_Enable;
    bool                m_Visible;

protected:
    SmMessage*          m_Target;
    uint                m_Id;

};
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif  // SM_MESSAGE_H_
