/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <nn/os.h>
#include <nn/fs.h>

#include <nw/types.h>
#include <nw/demo/demo_Utility.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

#include "../include/SmDef.h"
#include "../include/GfxCtrl.h"

namespace
{

/*
  
 */
bool
LoadResources( GfxCtrl::GfxResourceSet* resourceSet, const wchar_t* resourcePath,
              nw::os::IAllocator* allocator, nw::os::IAllocator* deviceAllocator, GfxCtrl::GfxVramType vramType = GfxCtrl::GFX_CTRL_VRAM_NONE )
{
    static const int RESOURCE_ALIGNMENT = 128;

    resourceSet->buffer = nw::demo::Utility::LoadFile( deviceAllocator , resourcePath, RESOURCE_ALIGNMENT );
    if (!resourceSet->buffer)
    {
        NW_FATAL_ERROR("can not open gfx archive.\n");
    }
    resourceSet->resource = nw::gfx::ResGraphicsFile(&(resourceSet->buffer.front()));

    if (vramType != GfxCtrl::GFX_CTRL_VRAM_NONE)
    {
        // Use the checkboxes to switch out the vertex information that is being displayed in the list.
        resourceSet->resource.ForeachTexture(nw::gfx::TextureLocationFlagSetter(vramType | GL_NO_COPY_FCRAM_DMP));
        resourceSet->resource.ForeachIndexStream(nw::gfx::IndexStreamLocationFlagSetter(vramType | GL_NO_COPY_FCRAM_DMP));
        resourceSet->resource.ForeachVertexStream(nw::gfx::VertexStreamLocationFlagSetter(vramType | GL_NO_COPY_FCRAM_DMP));
    }

    // Set up loaded resources and output any errors.
    nw::gfx::Result result = resourceSet->resource.Setup(allocator);
    if (result.IsFailure())
    {
        NW_WARNING(false ,"Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    return true;
}



/*
  
 */
GfxCtrl::GfxCtrl( SmSceneCtrl* sceneCtrl )
: m_GfxSceneCtrl(sceneCtrl),
  m_SceneRoot(NULL),
  m_Camera(NULL),
  m_ResCamera(NULL),
  m_SmCamera(NULL),
  m_SmParticle(NULL),
  m_ParticleContext(NULL),
  m_ParticleSceneUpdater(NULL),
  m_SceneEnvSetting(NULL),
  m_LightAnimation(NULL)
{
    m_ResourcesArray.clear();
    m_AnimationArray.clear();
    m_SmModelInstArray.clear();

    // Create root node.
    m_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(m_Allocator);
    NW_NULL_ASSERT(m_SceneRoot);

    // Create SmCamera instance.
    m_SmCamera = new SmCamera( m_SceneRoot, NULL );

    // Create particle context.
    m_ParticleContext = nw::gfx::ParticleContext::Builder()
        .MaxEmission(1000)
        .Create( m_Allocator );

    // Create particle scene updater.
    m_ParticleSceneUpdater = nw::gfx::ParticleSceneUpdater::Builder()
        .Create( m_Allocator );
}


/*
  
 */
GfxCtrl::~GfxCtrl()
{
    // Destroy animations.
    for (int animIdx = 0; animIdx < m_AnimationArray.Size(); ++animIdx)
    {
        nw::gfx::SafeDestroy(m_AnimationArray[animIdx]);
    }
    m_AnimationArray.clear();

    // Destroy SmModel instances.
    for (int modelIdx = 0; modelIdx < m_SmModelInstArray.Size(); ++modelIdx)
    {
        SM_SAFE_DELETE(m_SmModelInstArray[modelIdx]);
    }
    m_SmModelInstArray.clear();

    // Destroy particle-related objects.
    nw::gfx::SafeDestroy(m_ParticleSceneUpdater);
    nw::gfx::SafeDestroy(m_ParticleContext);

    // Destroy scene node starting from top node.
    nw::gfx::SafeDestroyBranch(m_SceneRoot);

    // Destroy resources.
    nw::demo::SafeCleanupResources(m_ResourcesArray);
    nw::ut::SafeDestroy(m_SceneEnvSetting);

    SM_SAFE_DELETE(m_SmCamera);
    SM_SAFE_DELETE(m_SmParticle);

    m_ResourcesArray.clear();
}


/*
  
 */
void
GfxCtrl::InitializeScene()
{

}


/*
  
 */
void
GfxCtrl::TraverseScene()
{
    m_GfxSceneCtrl->TraverseScene(m_SceneRoot);
}


/*
  
 */
void
GfxCtrl::Update()
{
    uint i = 0;

    // Update particles.
    m_ParticleSceneUpdater->UpdateNode( m_GfxSceneCtrl->GetSceneContext(), m_ParticleContext );

    // Update SmModel.
    // Resolve constraints.
    for ( i = 0; i < m_SmModelInstArray.size(); i++ )
    {
        m_SmModelInstArray[i]->Update();
    }

    // Update camera.
    m_SmCamera->UpdateMatrix();
}


/*
  
 */
SmModel*
GfxCtrl::CreateModel( SmModel* parentModel, const wchar_t* resourcePath, GfxVramType vramType )
{
    NW_NULL_ASSERT( resourcePath );

    bool isPushed;
    NW_UNUSED_VARIABLE(isPushed);
    GfxCtrl::GfxResourceSet resSet;

    // Load and set up resources.
    LoadResources( &resSet, resourcePath, m_Allocator, m_DeviceAllocator, vramType );

    // Create SmModel instance and set required data.
    SmModel* pModel = new SmModel();
    NW_NULL_ASSERT( pModel );

    //------------------------------------------------------------------
    // Set up models. (Only supports one model per resource.)
    nw::gfx::ResModelArray models = resSet.resource.GetModels();
     if ( models.size() != 1 ) { NW_FATAL_ERROR("1 Resource is 1 Model"); }

    // Generate node.
    nw::gfx::SceneNode* pAddNode = nw::demo::Utility::CreateSceneNode(
            m_DeviceAllocator,
            nw::gfx::ResSceneObject((*models.begin())));

    // Resolve parent-child references.
    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(m_Allocator);
    sceneNodeArray.push_back( pAddNode );
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    // Add model to scene.
     if ( parentModel )
    {
        // Add to specified parent node.
        parentModel->GetSceneNode()->AttachChild( pAddNode );
        pModel->SetSceneNode( parentModel->GetSceneNode(), pAddNode );
    }
    else
    {
        // Add to root node.
        m_SceneRoot->AttachChild( pAddNode );
        pModel->SetSceneNode( m_SceneRoot, pAddNode );
    }

    // Set model/skeletal model.
    nw::gfx::Model*         pGfxModel  = nw::ut::DynamicCast<nw::gfx::Model*>(pAddNode);
    nw::gfx::SkeletalModel* pSkelModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(pAddNode);
    pModel->SetModel( pGfxModel, pSkelModel );
     if ( !pSkelModel ) goto _anim_end;


    //------------------------------------------------------------------
    // Set up the animation.
    {
        nw::gfx::ResSkeletalModel resModel = pSkelModel->GetResSkeletalModel();
        nw::gfx::ResSkeleton resSkeleton = resModel.GetSkeleton();
        int maxBones = resSkeleton.GetBonesCount();

        uint resAnimNum = resSet.resource.GetSkeletalAnimsCount();

        for ( uint i = 0; i < resAnimNum; i++ )
        {
            nw::anim::ResAnim resAnim = resSet.resource.GetSkeletalAnims(i);
            if (!resAnim.IsValid()){ NW_FATAL_ERROR("Animation is Valid"); }

            // Creates a transform animation evaluator.
            nw::gfx::TransformAnimEvaluator* evaluator = nw::gfx::TransformAnimEvaluator::Builder()
                .AnimData(resAnim)
                .MaxMembers(maxBones)
                .AllocCache(true)
                .Create(m_Allocator);

            if (!evaluator){ NW_FATAL_ERROR("Animation Evaluator is Valid"); }

            // Set object to SmModel.
            pModel->AddAnimEvaluator( evaluator );

            // Store animation.
            isPushed = m_AnimationArray.push_back( evaluator );
            NW_ASSERT(isPushed);
        }
    }

_anim_end:

    //------------------------------------------------------------------
    // Material Animations
    {
        int materialAnimCount = resSet.resource.GetMaterialAnimsCount();
        if ( materialAnimCount == 1 )
        {
            nw::anim::ResAnim resAnim = resSet.resource.GetMaterialAnims(0);

            nw::gfx::AnimGroup* animGroup = pGfxModel->GetMaterialAnimGroup();
            if ( !animGroup ) goto _setup_end;
            int maxMembers = animGroup->GetMemberCount();

            nw::gfx::AnimEvaluator* matAnimEvaluator = nw::gfx::AnimEvaluator::Builder()
                .AnimData(resAnim)
                .MaxMembers(maxMembers)
                .AllocCache(true)
                .Create(m_Allocator);

            // Bind the animation.
            bool bindResult = matAnimEvaluator->Bind( animGroup );

            // Set animation for model.
            if ( bindResult )
            {
                pGfxModel->SetMaterialAnimObject( matAnimEvaluator );
            }

            // Store animation.
            isPushed = m_AnimationArray.push_back( matAnimEvaluator );
            NW_ASSERT(isPushed);
        }
    }

_setup_end:

    // Store loaded resource.
    isPushed = m_ResourcesArray.push_back( resSet );
    NW_ASSERT(isPushed);

    // Store SmModel instance.
    isPushed = m_SmModelInstArray.push_back( pModel );
    NW_ASSERT(isPushed);

    return pModel;
}


/*
  
 */
bool GfxCtrl::AddModel( SmModel* model )
{
    // Add to root node.
    m_SceneRoot->AttachChild( model->GetSceneNode() );
    model->SetSceneNode( m_SceneRoot, model->GetSceneNode() );

    return true;
}



/*
  
 */
SmParticle*
GfxCtrl::CreateParticle( const wchar_t* resourcePath, GfxVramType vramType  )
{
    NW_ASSERT( !m_SmParticle );

    NW_NULL_ASSERT( resourcePath );

    bool isPushed;
    NW_UNUSED_VARIABLE(isPushed);

    GfxCtrl::GfxResourceSet resSet;
    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(m_Allocator);

    // Load resources.
    LoadResources( &resSet, resourcePath, m_Allocator, m_DeviceAllocator, vramType );

    // Create SmParticle instance and set required data.
    m_SmParticle = new SmParticle( m_SceneRoot );
    NW_NULL_ASSERT( m_SmParticle );

    // Generate node.
    nw::gfx::ResModelArray models = resSet.resource.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* pAddNode = nw::demo::Utility::CreateSceneNode(
                m_DeviceAllocator,
                nw::gfx::ResSceneObject((*modelResource).ptr()));


        // todo: Update layer ID here.
        {
            nw::gfx::ParticleModel* pParticleModel  = nw::ut::DynamicCast<nw::gfx::ParticleModel*>(pAddNode);
            pParticleModel->SetLayerId( 6 );
        }

        sceneNodeArray.push_back( pAddNode );

        // Add ParticleModel to SmParticle.
        nw::gfx::ParticleModel* pParticleModel  = nw::ut::DynamicCast<nw::gfx::ParticleModel*>(pAddNode);
        NW_NULL_ASSERT( pParticleModel );
        m_SmParticle->AddParticleModel( pParticleModel );
    }

    // Create emitter.
    nw::gfx::ResEmitterArray emitters = resSet.resource.GetEmitters();
    nw::gfx::ResEmitterArray::iterator emittersEnd = emitters.end();
    for (nw::gfx::ResEmitterArray::iterator emitterResource = emitters.begin();
         emitterResource != emittersEnd; ++emitterResource)
    {
        nw::gfx::SceneNode* pAddEmitNode = nw::demo::Utility::CreateSceneNode(
                m_DeviceAllocator,
                nw::gfx::ResSceneObject((*emitterResource).ptr()));

        sceneNodeArray.push_back( pAddEmitNode );

        // Add ParticleModel to SmParticle.
        nw::gfx::ParticleEmitter* pParticleEmitter  = nw::ut::DynamicCast<nw::gfx::ParticleEmitter*>(pAddEmitNode);
        NW_NULL_ASSERT( pParticleEmitter );
        m_SmParticle->AddParticleEmitter( pParticleEmitter );
    }
/*
    // Material Animations
    {
        int materialAnimCount = resSet.resource.GetMaterialAnimsCount();
        nw::anim::ResAnim resAnim = resSet.resource.GetMaterialAnims(0);
        nw::gfx::AnimGroup* animGroup = pGfxModel->GetMaterialAnimGroup();
        if ( !animGroup ) goto _setup_end;
        int maxMembers = animGroup->GetMemberCount();
        nw::gfx::AnimEvaluator* matAnimEvaluator = nw::gfx::AnimEvaluator::Builder()
            .AnimData(resAnim)
            .MaxMembers(maxMembers)
            .AllocCache(true)
            .Create(m_Allocator);
        // Bind the animation.
        bool bindResult = matAnimEvaluator->Bind( animGroup );
        // Add animation to model.
        if ( bindResult )
        {
            pGfxModel->SetMaterialAnimObject( matAnimEvaluator );
        }
        // Store animation.
        isPushed = m_AnimationArray.push_back( matAnimEvaluator );
        NW_ASSERT(isPushed);
    }






*/
    // Initialize settings and resolve particle object references.
    nw::gfx::ParticleUtil::SetupParticleObject( &sceneNodeArray, m_ParticleContext );

    // Resolve parent-child references.
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    // Add models and emitters to scene.
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode( m_SceneRoot )
    );

    // Store loaded resource.
    isPushed = m_ResourcesArray.push_back( resSet );
    NW_ASSERT(isPushed);

    return m_SmParticle;
}


/*
  
 */
void
GfxCtrl::CreateSceneEnvironment( const wchar_t* resourcePath )
{
    bool isPushed = false;
    NW_UNUSED_VARIABLE(isPushed);
    GfxCtrl::GfxResourceSet resSet;

    nw::gfx::SceneNode* firstLightNode = NULL;

    // Load resources.
    LoadResources( &resSet, resourcePath, m_Allocator, m_DeviceAllocator );

    // Create light objects.
    nw::gfx::ResLightArray lights = resSet.resource.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource( nw::gfx::ResSceneObject((*lightResource).ptr()) )
            .CreateObject(m_Allocator, m_DeviceAllocator);

        nw::gfx::SceneNode* node = nw::ut::DynamicCast<nw::gfx::SceneNode*>(sceneObject);

        // Store first light and bind animation.
        if ( !firstLightNode )
        {
            firstLightNode = node;
        }

        m_SceneRoot->AttachChild(node);
    }

    // Set up light animations.
    int lightAnimCount = resSet.resource.GetLightAnimsCount();
    if ( lightAnimCount == 1 && firstLightNode )
    {
        nw::anim::ResAnim resAnim = resSet.resource.GetLightAnims(0);

        nw::gfx::Light* light =
                        nw::ut::DynamicCast<nw::gfx::Light*>(firstLightNode);

        nw::gfx::AnimGroup* animGroup = light->GetAnimGroup();
        if (animGroup)
        {
            int maxMembers = animGroup->GetMemberCount();

            nw::gfx::AnimEvaluator* lightAnimEvaluator = nw::gfx::AnimEvaluator::Builder()
                .AnimData(resAnim)
                .MaxMembers(maxMembers)
                .AllocCache(true)
                .Create(m_Allocator);

            // Bind the animation.
            bool bindResult = lightAnimEvaluator->Bind( animGroup );

            // Set animation for model.
            if ( bindResult )
            {
                light->SetAnimObject( lightAnimEvaluator );
            }

            // Store animation.
            isPushed = m_AnimationArray.push_back( lightAnimEvaluator );
            NW_ASSERT(isPushed);

            if( !m_LightAnimation )
            {
                m_LightAnimation = lightAnimEvaluator;
            }
        }
    }


    // Set up fogs.
    {
        nw::gfx::ResFogArray fogs = resSet.resource.GetFogs();
        nw::gfx::ResFogArray::iterator fogsEnd = fogs.end();
        for (nw::gfx::ResFogArray::iterator fogResource = fogs.begin();
             fogResource != fogsEnd; ++fogResource)
        {
            nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
                m_DeviceAllocator,
                nw::gfx::ResSceneObject(*fogResource)
            );
            NW_NULL_ASSERT(node);

            m_SceneRoot->AttachChild(node);
        }
    }

    // Create a scene environment instance.
    if (!m_SceneEnvSetting)
    {
        nw::gfx::ResSceneEnvironmentSettingArray settings = resSet.resource.GetSceneEnvironmentSettings();
        nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
        for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
            settingResource != settingsEnd; ++settingResource)
        {
            nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
                .Resource(*settingResource)
                .CreateObject(m_Allocator, m_DeviceAllocator);

            nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
                nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

            NW_NULL_ASSERT(sceneEnvironmentSetting);
            m_SceneEnvSetting = sceneEnvironmentSetting;;
        }
    }


    // Store loaded resource.
    isPushed = m_ResourcesArray.push_back( resSet );
    NW_ASSERT(isPushed);
}


/*
  
 */
void
GfxCtrl::CreateCamera( const wchar_t* resourcePath )
{
    NW_ASSERT( !m_ResCamera );

    GfxCtrl::GfxResourceSet resSet;

    // Load resources.
    LoadResources( &resSet, resourcePath, m_Allocator, m_DeviceAllocator );

    // Only supports one camera per resource.
    nw::gfx::ResCameraArray cameras = resSet.resource.GetCameras();
    if ( cameras.size() != 1 ) { NW_FATAL_ERROR("1 Resource is 1 Camera"); }

    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource( nw::gfx::ResSceneObject( *cameras.begin() ) )
            .CreateObject(m_Allocator, m_DeviceAllocator);

        nw::gfx::SceneNode* node = nw::ut::DynamicCast<nw::gfx::SceneNode*>(sceneObject);

        // Cast to gfxCamera.
        m_ResCamera = nw::ut::DynamicCast<nw::gfx::Camera*>(node);
        NW_ASSERT(m_ResCamera);

        // Set camera to SmCamera instance.
        m_SmCamera->SetGxCamera( m_ResCamera );
    }

    // Store loaded resource.
    bool isPushed = m_ResourcesArray.push_back( resSet );
    NW_UNUSED_VARIABLE(isPushed);
    NW_ASSERT(isPushed);

    // Create camera animation.
    {
        nw::gfx::AnimGroup* animGroup = m_ResCamera->GetAnimGroup();
        if (animGroup == NULL) // There is no animation group for animations.
        {
            return;
        }

        int maxMembers = animGroup->GetMemberCount();

        int cameraAnimCount = resSet.resource.GetCameraAnimsCount();
        for (int i = 0; i < cameraAnimCount; ++i)
        {
            nw::anim::ResAnim resAnim = resSet.resource.GetCameraAnims(i);

            // Creates a transform animation evaluator.
            nw::gfx::AnimEvaluator* evaluator = nw::gfx::AnimEvaluator::Builder()
                .AnimData(resAnim)
                .MaxMembers(maxMembers)
                .AllocCache(true)
                .Create(m_Allocator);

            if (!evaluator){ NW_FATAL_ERROR("Animation Evaluator is Valid"); }

            // Add to SmCamera.
            m_SmCamera->AddAnimEvaluator( evaluator );

            // Store animation.
            isPushed = m_AnimationArray.push_back( evaluator );
            NW_ASSERT(isPushed);
        }
    }
}


} // Namespace


