/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <nn/os.h>

#include <nw/types.h>
#include <nw/ut.h>

#include "../include/SmLayout.h"

namespace
{

/*
  
 */
nw::ut::Rect SmLayoutResource::GetRect( const char* name )
{
    nw::ut::Rect rect;
    const nw::lyt::Pane* pane = GetPane( name );
     if ( !pane ) return rect;

    rect = pane->GetPaneRect();
    nw::math::VEC3 pos = pane->GetTranslate();

    // Correct to screen coordinates.
    rect.top    = -rect.top;
    rect.bottom = -rect.bottom;

    // Account for the rectangle's center.
    f32 posx = 0.f;
    f32 posy = 0.f;

    switch( pane->GetBasePositionH() )
    {
    case nw::lyt::HORIZONTALPOSITION_LEFT:
        posx = SM_LOWER_SCREEN_HALF_SIZE_W + pos.x;
        break;

    case nw::lyt::HORIZONTALPOSITION_CENTER:
        posx = SM_LOWER_SCREEN_HALF_SIZE_W + pos.x - rect.GetWidth()/2.f;
        break;

    case nw::lyt::HORIZONTALPOSITION_RIGHT:
        posx = SM_LOWER_SCREEN_HALF_SIZE_W + pos.x;
        break;
    }

    switch( pane->GetBasePositionV() )
    {
    case nw::lyt::VERTICALPOSITION_TOP:
        posy = SM_LOWER_SCREEN_HALF_SIZE_H - pos.y;
        break;

    case nw::lyt::VERTICALPOSITION_CENTER:
        posy = SM_LOWER_SCREEN_HALF_SIZE_H - pos.y - rect.GetHeight()/2.f;
        break;

    case nw::lyt::VERTICALPOSITION_BOTTOM:
        posy = SM_LOWER_SCREEN_HALF_SIZE_H - pos.y;
        break;
    }

    rect.MoveTo( posx, posy );

    return rect;
}




/*
  
 */
SmLayout::SmLayout( SmLayoutResource* layoutResource )
: m_LayoutResource( layoutResource ),
  m_OffsetPane( NULL ),
  m_AnimationFrame( 0.0f ),
  m_AnimationNo( 0 ),
  m_Visible( true ),
  m_String( NULL )
{
    SetPosition( 0.f, 0.f );
}


/*
  
 */
SmLayout::~SmLayout()
{

}


/*
  
 */
void
SmLayout::SetPosition( f32 x, f32 y )
{
    m_Position.x = x;
    m_Position.y = - y;
    m_Position.z = 0.0f;
}


/*
  
 */
bool
SmLayout::SetAnimationNo( uint animTransformNo )
{
    if ( animTransformNo > m_LayoutResource->GetAnimationNum() ) return false;
    m_AnimationNo = animTransformNo;
    return true;
}


/*
  
 */
void
SmLayout::SetAnimationFrame( f32 setFrame )
{
    m_AnimationFrame = setFrame;
}


/*
  
 */
bool
SmLayout::AddAnimationFrame( f32 addFrame, bool loop )
{
    bool ret = false;
    if ( m_AnimationNo >= m_LayoutResource->GetAnimationNum() ) return ret;
    f32 frameSize = m_LayoutResource->GetAnimationFrameSize( m_AnimationNo );

    // Add to frame.
    m_AnimationFrame += addFrame;

    // Last frame.
    if ( m_AnimationFrame > frameSize )
    {
        if ( loop )
        {
            m_AnimationFrame = 0.f;
        }
        else
        {
            m_AnimationFrame -= frameSize;
        }

        ret = true;
    }
    return ret;
}


/*
  
 */
bool
SmLayout::SetupLayout()
{
    nw::lyt::Layout* layout = m_LayoutResource->GetLayout();
     if ( !layout ) return false;

    // Set the display position of the entire layout.
    nw::lyt::Pane* pane = layout->GetRootPane();
    if ( pane )
    {
        pane->SetTranslate( m_Position );
    }

    // If an offset has already been applied to the Pane instance, set its position.
    if ( m_OffsetPane )
    {
        m_OffsetPane->SetTranslate( m_OffsetPosition );
    }

    // Configure animations.
    if ( m_AnimationNo < m_LayoutResource->GetAnimationNum() )
    {
        SmLayoutResource::AnimationArray& animArray = m_LayoutResource->GetAnimationArray();

        // Configure animations.
        if ( !animArray.empty() )
        {
            layout->UnbindAnimation( NULL );
            layout->BindAnimation( animArray[m_AnimationNo] );

            animArray[m_AnimationNo]->SetFrame( m_AnimationFrame );
        }
    }

    return true;
}


/*
  
 */
void
SmLayout::ResetLayout()
{
    // Restore the Pane instance with offset applied.
    if ( m_OffsetPane )
    {
        m_OffsetPane->SetTranslate( nw::math::VEC3::Zero() );
    }
}



} // Namespace


