/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <nn/os.h>

#include <nw/types.h>
#include <nw/ut.h>

#include "../include/SmLytButton.h"
#include "../include/SmTouchPanelStatus.h"

namespace
{

//----------------------------------------
// Constructor
SmLytButton::SmLytButton( SmLayout* smLayout, const wchar_t* string, const char* bounding )
: SmMessage( NULL ),
  m_Layout( smLayout ),
  m_LastX( 0 ),
  m_LastY( 0 ),
  m_IsPressed( false ),
  m_IsGrab( false ),
  m_IsSwitch( true )
{
    NW_NULL_ASSERT( m_Allocator );
    NW_NULL_ASSERT( m_Layout );

    // Register the collision shape.
    SmLayoutResource* layoutRes = m_Layout->GetLayoutResource();
    nw::ut::Rect rectBounding = layoutRes->GetRect( bounding );
    m_Collision.SetRect( rectBounding );

     if ( string )
    {
        m_Layout->SetString( string );
    }

#if 0
    // Register text.
    // Is it TextBox pane?
    const nw::lyt::Pane* pane = layoutRes->GetPane("TextBox_00");
    if ( nw::lyt::TextBox* pTextBox = nw::ut::DynamicCast<nw::lyt::TextBox*>(pane) )
    {
        pTextBox->SetString( string );
    }
#endif

    // Initial state
    SetSwitch( m_IsSwitch );
}


//----------------------------------------
// Destructor
SmLytButton::~SmLytButton()
{

}


//----------------------------------------
// position
void SmLytButton::SetPosition( f32 x, f32 y )
{
    m_Layout->SetPosition( x, y );
//    m_Collision.SetXY( x, y );

    // Determine the collision rectangle.
    m_Collision.SetXY(  SM_LOWER_SCREEN_HALF_SIZE_W + x - m_Collision.GetWidth()/2.f,
                        SM_LOWER_SCREEN_HALF_SIZE_H + y - m_Collision.GetHeight()/2.f );
}



//----------------------------------------
// Rendering
void SmLytButton::Render()
{
    return;
}


//----------------------------------------
// Set button status.
void SmLytButton::SetSwitch( bool bswitch )
{
    m_IsSwitch = bswitch;

     if ( m_IsSwitch )
    {
        m_Layout->SetAnimationNo( 0 );
    }
    else
    {
        m_Layout->SetAnimationNo( 1 );
    }
}



//----------------------------------------
// Swap button status.
void SmLytButton::SwapSwitch()
{
     if ( m_IsSwitch )
    {
        m_IsSwitch = false;
        m_Layout->SetAnimationNo( 1 );
    }
    else
    {
        m_IsSwitch = true;
        m_Layout->SetAnimationNo( 0 );
    }
}


//----------------------------------------
// Receive message
bool SmLytButton::ReceveMessage( SmMessageType type, void* object, uint targetId )
{
    NW_UNUSED_VARIABLE(targetId);

    SmTouchPanelStatus*  touchPanelStaus = NULL;

    switch( type )
    {
    case SM_MESSAGE_TUCHPANEL_PRESS:
        {
            touchPanelStaus = static_cast<SmTouchPanelStatus*>(object);
            NW_NULL_ASSERT( touchPanelStaus );

            // If not being used by another process...
             if ( !touchPanelStaus->IsGrab() )
            {
                // If inside the collision zone,
                 if ( m_Collision.CheckInner( touchPanelStaus->GetX(), touchPanelStaus->GetY() ) )
                {
                    // Grab.
                    touchPanelStaus->Grab();
                    m_IsGrab    = true;
                    m_IsPressed = true;

                    m_Layout->SetAnimationNo( 2 );
                }
            }
        }
        break;

    case SM_MESSAGE_TUCHPANEL_MOTION:
        {
            touchPanelStaus = static_cast<SmTouchPanelStatus*>(object);
            NW_NULL_ASSERT( touchPanelStaus );
             if ( m_IsGrab )
            {
                m_LastX = touchPanelStaus->GetX();
                m_LastY = touchPanelStaus->GetY();

                // Play the animation being selected.
                m_Layout->AddAnimationFrame( 1.f );
            }
        }
        break;

    case SM_MESSAGE_TUCHPANEL_RELEASE:
        {
            touchPanelStaus = static_cast<SmTouchPanelStatus*>(object);
            NW_NULL_ASSERT( touchPanelStaus );

            // Release.
             if ( m_IsGrab )
            {
                touchPanelStaus->UnGrab();
                m_IsGrab = false;
                m_IsPressed = false;

                // Return the selected animation to 0.f frame.
                m_Layout->SetAnimationFrame( 0.f );

                // Send a message to the specified target if inside the collision zone.
                 if ( m_Collision.CheckInner( m_LastX, m_LastY ) )
                {
                    // Swap the state.
                    SwapSwitch();

                    // Receive update.
                     if ( m_Target )
                    {
                        m_Target->SendMessage( SM_MESSAGE_UPDATE_PARAM, NULL, m_Id );
                    }
                }
            }
        }
        break;

    default:
        break;
    }
    return 0;
}

} // Namespace


