/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nn/os.h>

#include <nw/types.h>
#include <nw/ut.h>

#include "../include/SmTouchPanelCamera.h"
#include "../include/SmTouchPanelStatus.h"

namespace
{


//----------------------------------------
// Constructor
SmTouchPanelCamera::SmTouchPanelCamera( SmCamera* smcamera )
: SmMessage( NULL ),
  m_SmCamera( smcamera ),
  m_IsGrab( false ),
  m_TouchPanelPosLastX( 0 ),
  m_TouchPanelPosLastY( 0 ),
  m_TouchPanelPosX( 0 ),
  m_TouchPanelPosY( 0 ),
  m_DegX( 0.f ),
  m_DegY( 0.f )
{
    // Camera viewpoint
    m_SmCamera->GetPosition( &m_CameraPos );

    // Camera viewpoint
    m_SmCamera->GetTarget( &m_CameraTarget );

    // Base camera orientation.
    m_CameraToward.x = m_CameraTarget.x - m_CameraPos.x;
    m_CameraToward.y = m_CameraTarget.y - m_CameraPos.y;
    m_CameraToward.z = m_CameraTarget.z - m_CameraPos.z;
    m_CameraToward.Normalize();

    // Set touch panel collision.
    m_Collision.SetRect( 40, 70, 160, 160 );
}


//----------------------------------------
// Receive message
bool SmTouchPanelCamera::ReceveMessage( SmMessageType type, void* object, uint targetId )
{
    NW_UNUSED_VARIABLE(targetId);

    SmTouchPanelStatus*  touchPanelStaus = NULL;

    bool ret = false;

    switch( type )
    {
    case SM_MESSAGE_TUCHPANEL_PRESS:
        {
            touchPanelStaus = static_cast<SmTouchPanelStatus*>(object);
            NW_NULL_ASSERT( touchPanelStaus );

            if ( m_Collision.CheckInner( touchPanelStaus->GetX(), touchPanelStaus->GetY() ) )
            {
                // If not being used by another process...
                if ( !touchPanelStaus->IsGrab() )
                {
                    // Grab.
                    touchPanelStaus->Grab();
                    m_IsGrab    = true;
                }
            }
        }
        break;

    case SM_MESSAGE_TUCHPANEL_MOTION:
        {
            touchPanelStaus = static_cast<SmTouchPanelStatus*>(object);
            NW_NULL_ASSERT( touchPanelStaus );

            // Update camera
            if ( m_IsGrab )
            {
                // Update coordinates.
                m_TouchPanelPosLastX    = static_cast<s32>( m_TouchPanelPosX );
                m_TouchPanelPosLastY    = static_cast<s32>( m_TouchPanelPosY );
                m_TouchPanelPosX        = static_cast<s32>( touchPanelStaus->GetX() );
                m_TouchPanelPosY        = static_cast<s32>( touchPanelStaus->GetY() );

                // Calculate camera.
                calcCamera();
            }
        }
        break;

    case SM_MESSAGE_TUCHPANEL_RELEASE:
        {
            touchPanelStaus = static_cast<SmTouchPanelStatus*>(object);
            NW_NULL_ASSERT( touchPanelStaus );

            // Release.
            if ( m_IsGrab )
            {
                touchPanelStaus->UnGrab();
                m_IsGrab = false;
            }
        }
        break;

    default:
        break;
    }

    return ret;
}


//----------------------------------------
// Rendering
void SmTouchPanelCamera::Render()
{
    Sm2DDynamicPrimPC prim;
    if ( prim.Begin( 4, GL_TRIANGLE_STRIP, 0.f, 0.f ) )
    {
        prim.SetVertex( m_Collision.GetX(),                             m_Collision.GetY(),                             nw::ut::Color8::YELLOW );
        prim.SetVertex( m_Collision.GetX(),                             m_Collision.GetY() + m_Collision.GetHeight(),   nw::ut::Color8::YELLOW );
        prim.SetVertex( m_Collision.GetX() + m_Collision.GetWidth(),    m_Collision.GetY(),                             nw::ut::Color8::YELLOW );
        prim.SetVertex( m_Collision.GetX() + m_Collision.GetWidth(),    m_Collision.GetY() + m_Collision.GetHeight(),   nw::ut::Color8::YELLOW );
        prim.End();
    }

    if ( m_IsGrab )
    {
        if ( prim.Begin( 4, GL_TRIANGLE_STRIP, 0.f, 0.f ) )
        {
            prim.SetVertex( m_TouchPanelPosX,           m_TouchPanelPosY,       nw::ut::Color8::RED );
            prim.SetVertex( m_TouchPanelPosX,           m_TouchPanelPosY + 4.f, nw::ut::Color8::RED );
            prim.SetVertex( m_TouchPanelPosX + 4.f,     m_TouchPanelPosY,       nw::ut::Color8::RED );
            prim.SetVertex( m_TouchPanelPosX + 4.f,     m_TouchPanelPosY + 4.f, nw::ut::Color8::RED );
            prim.End();
        }
    }
}


//----------------------------------------
// Calculate walk-through camera.
void SmTouchPanelCamera::calcCamera()
{
    nw::math::MTX34 matrixRaotY;
    matrixRaotY.SetupIdentity();

    nw::math::VEC3 toward;

    f32 degX = 0.f;

    // Left-right pan at screen center.
    {
        #define _DEGREE_SPEED   (0.3f)
        #define _DEGREE_MAX     (15.f)

        s32 centerX = m_Collision.GetX() + m_Collision.GetWidth()/2.f;
        s32 centerY = m_Collision.GetY() + m_Collision.GetHeight()/2.f;

        if (  m_TouchPanelPosX < centerX )
        {
            if ( m_DegY < _DEGREE_MAX )
            {
                m_DegY += _DEGREE_SPEED;
            }
        }
        else
        {
            if ( m_DegY > -_DEGREE_MAX )
            {
                m_DegY -= _DEGREE_SPEED;
            }
        }

        if (  m_TouchPanelPosY < centerY )
        {
            if ( m_DegX < _DEGREE_MAX )
            {
                m_DegX += _DEGREE_SPEED;
            }
        }
        else
        {
            if ( m_DegX > -_DEGREE_MAX )
            {
                m_DegX -= _DEGREE_SPEED;
            }
        }
    }

    // Update orientation.
    MTX34RotXYZDeg( &matrixRaotY, m_DegX, m_DegY, 0.f );
    VEC3TransformNormal( &toward, &matrixRaotY, &m_CameraToward );

    // Viewpoint based on position and orientation.
    m_CameraTarget.x = m_CameraPos.x + toward.x * 10.f;
    m_CameraTarget.y = m_CameraPos.y + toward.y * 10.f;
    m_CameraTarget.z = m_CameraPos.z + toward.z * 10.f;

    // Update camera.
    updateCamera();
}


//----------------------------------------
// Update camera.
void SmTouchPanelCamera::updateCamera()
{
    // todo:
    // Disabled
     if ( !GetEnable() ) return;

    m_SmCamera->SetPosition( &m_CameraPos );
    m_SmCamera->SetTarget( &m_CameraTarget );
}


} // Namespace


