﻿/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


#define NW_DEBUG_CHECK_MEMORY_LEAK


#include <nn/os.h>
#include <nn/fs.h>
#include <nn/hid/CTR/hid_TouchPanelReader.h>

#include <nw/types.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>
#include <nw/lyt.h>

#include "../include/SmRenderSystem.h"
#include "../include/SmSceneCtrl.h"
#include "../include/SmCommandUtility.h"
#include "../include/demo.h"
#include "../include/framework.h"
#include "../include/GfxCtrl.h"
#include "../include/LytCtrl.h"
#include "../include/SmDef.h"
#include "../include/SmPerf.h"
#include "../include/SmFile.h"
#include "../include/SmPrimitive.h"
#include "../include/SmTouchPanelStatus.h"
#include "../include/SmPadStatus.h"
#include "../include/SmSliderBar.h"
#include "../include/SmButton.h"
#include "../include/SmCommandUtility.h"


namespace
{


//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 2;
#define DEMO_UPPER_SCREEN    (0)
#define DEMO_LOWER_SCREEN    (1)
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;
RenderTargetArray           s_RenderTargets;

SmRenderSystem*             s_RenderSystem = NULL;
SmSceneCtrl*                s_GfxSceneCtrl = NULL;

//----------------------------------------
// Memory-related
nw::demo::DemoAllocator     s_DeviceAllocator;
//nw::demo::DemoDeviceMemoryAllocator         s_DeviceAllocator;


//----------------------------------------
// Management of each object group

// Management of the Gfx object
GfxCtrl*    s_GfxCtrl       = NULL;

// Management of the Lyt object
LytCtrl*    s_LytCtrl       = NULL;

// Demo:
Demo*       s_Demo          = NULL;

// Framework status
FrameWorkStatus s_FrameStatus;

const s32 s_BaseCameraIndex = 0;
const s32 s_PadCameraIndex = 1;


//----------------------------------------
// OpenGL rendering
SmOgl*      s_SmOgl     = NULL;


//----------------------------------------
// Performance measurement
SmCpuPerf*      s_CpuPerf = NULL;
SmGpuPerf*      s_GpuPerf = NULL;
SmPerfDrawer*   s_PerfDrawer = NULL;


//----------------------------------------
// Parallax camera
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;


//----------------------------------------
// Static command cache
SmStaticCommandCache* s_StaticCommandCache = NULL;



/*
  
 */
void
InitializeGraphics()
{
    // Set the allocator for SmBase.
    SmBase::SetAllocator( &s_DeviceAllocator, &s_DeviceAllocator );

    // The constructor configures the standard settings for renderDescription.
    SmRenderSystem::Description renderDescription;
    {
        // Creates a render system.
        s_RenderSystem = new SmRenderSystem( renderDescription );

        // Generates the render target.
        s_RenderTargets.push_back(
            nw::gfx::IRenderTarget::Builder()
                    .BufferSize(
                        renderDescription.upperScreenDescription.height,
                        renderDescription.upperScreenDescription.width
                    )
                    .ColorFormat(renderDescription.renderColorFormat)
                    .Create(&s_DeviceAllocator)
        );
        NW_ASSERT(!s_RenderTargets.empty());
//        s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

        s_RenderTargets.push_back(
            nw::gfx::IRenderTarget::Builder()
                    .BufferSize(
                        renderDescription.lowerScreenDescription.height,
                        renderDescription.lowerScreenDescription.width
                    )
                    .ColorFormat(renderDescription.renderColorFormat)
                    .Create(&s_DeviceAllocator)
        );
        NW_ASSERT(!s_RenderTargets.empty());
    }

    // Initializes SmOgl (OpenGL wrapper).
    s_SmOgl = SmOgl::InitializeSmOgl( NW_DEMO_FILE_PATH(L"shader.shbin") );

    // Generates static command cache.
    s_StaticCommandCache = new SmStaticCommandCache();

    NW_GL_ASSERT();
}


/*
  
 */
void
TerminateGraphics()
{
    // Destroys static command cache.
    SM_SAFE_DELETE( s_StaticCommandCache );

    // Terminates SmOgl.
    SmOgl::TerminateSmOgl();

    // Destroys render target.
    nw::gfx::SafeDestroyAll(s_RenderTargets);

    // Destroys render system.
    SM_SAFE_DELETE( s_RenderSystem );

    NW_GL_ASSERT();
}


/*
  
 */
void
InitializeGfx()
{
    // Generates SmSceneCtrl.
    s_GfxSceneCtrl = new SmSceneCtrl( s_RenderSystem->GetRenderContext() );

    // Generates GfxCtrl.
    s_GfxCtrl = new GfxCtrl( s_GfxSceneCtrl );
    NW_NULL_ASSERT( s_GfxCtrl );

    // Processing bar
    s_CpuPerf       = new SmCpuPerf();
    NW_NULL_ASSERT( s_CpuPerf );
    s_GpuPerf       = new SmGpuPerf();
    NW_NULL_ASSERT( s_GpuPerf );

    s_PerfDrawer    = new SmPerfDrawer();
    NW_NULL_ASSERT( s_PerfDrawer );
}


/*
  
 */
void
TerminateGfx()
{
    SM_SAFE_DELETE(s_GfxCtrl);
    SM_SAFE_DELETE(s_CpuPerf);
    SM_SAFE_DELETE(s_GpuPerf);
    SM_SAFE_DELETE(s_PerfDrawer);

    SM_SAFE_DELETE(s_GfxSceneCtrl);
}


/*
  
 */
void
InitializeLyt()
{
    // Generates LytCtrl.
    s_LytCtrl = new LytCtrl();
    NW_NULL_ASSERT( s_LytCtrl );
}


/*
  
 */
void
TerminateLyt()
{
    SM_SAFE_DELETE(s_LytCtrl);
}


/*
  
 */
void
OutputLog()
{
    f32 cpuCost = 0.f;
    f32 gpuCost = 0.f;

    NW_UNUSED_VARIABLE(cpuCost);
    NW_UNUSED_VARIABLE(gpuCost);
/*
    NW_DEV_LOG( "\n[COST LOG]\n" );
    NW_DEV_LOG( "RED   :GFX\n" );
    NW_DEV_LOG( "BLUE  :SYS\n" );
    NW_DEV_LOG( "GREEN :LYT\n" );
    NW_DEV_LOG( "YELLOW:PERFORMANCE METER\n" );
*/
    NW_DEV_LOG( "\n[CPU COST LOG]\n" );
//    for ( int i = 0; i < s_CpuPerf->GetCostCount(); i++ )
//    {
//        NW_DEV_LOG( "%s:%8.4f[%8.4f%%]\n", s_CpuPerf->GetCostName(i), s_CpuPerf->GetCost(i), s_CpuPerf->GetCost(i)/16.666f*100.f );
//        cpuCost += s_CpuPerf->GetCostEnd(i) - s_CpuPerf->GetCostStart(i);
//    }
    if (s_CpuPerf->GetCostCount())
    {
        cpuCost = s_CpuPerf->GetCostEnd(s_CpuPerf->GetCostCount() - 1) - s_CpuPerf->GetFrameStart();
    }
    NW_DEV_LOG( "CPU COST:%8.4f[%8.4f%%]\n", cpuCost, cpuCost/16.666f*100.f );
    NW_DEV_LOG( "[GPU COST LOG]\n" );
//    for ( int i = 0; i < s_GpuPerf->GetCostCount(); i++ )
//    {
//        NW_DEV_LOG( "%s:%8.4f[%8.4f%%]\n", s_CpuPerf->GetCostName(i), s_CpuPerf->GetCost(i), s_CpuPerf->GetCost(i)/16.666f*100.f );
//        cpuCost += s_CpuPerf->GetCostEnd(i) - s_CpuPerf->GetCostStart(i);
//    }
    if (s_GpuPerf->GetCostCount())
    {
        gpuCost = s_GpuPerf->GetCostEnd(s_GpuPerf->GetCostCount() - 1) - s_GpuPerf->GetCostStart(0);
    }
    NW_DEV_LOG( "GPU COST:%8.4f[%8.4f%%]\n\n", gpuCost, gpuCost/16.666f*100.f );
}


/*
  
 */
SmMessageType
SendTouchPanelMassage( SmTouchPanelStatus& touchPanelStatus )
{
    SmMessageType messageType = SM_MESSAGE_NONE;

    // Touch panel messages
    static u16 touchPanelX = 0;
    static u16 touchPanelY = 0;

    // Divides the current status into press, release, motion and sends a message.
    messageType = SM_MESSAGE_NONE;

    // Press ?
    if ( ( touchPanelX == 0 && touchPanelStatus.GetX() != 0 ) || ( touchPanelY == 0 && touchPanelStatus.GetY() != 0 ) )
    {
        messageType = SM_MESSAGE_TUCHPANEL_PRESS;
    }
    else
    // Release ?
    if ( ( touchPanelX != 0 && touchPanelStatus.GetX() == 0 ) || ( touchPanelY != 0 && touchPanelStatus.GetY() == 0 ) )
    {
        messageType = SM_MESSAGE_TUCHPANEL_RELEASE;
    }
    else
    // Motion ?
     if ( ( touchPanelX != touchPanelStatus.GetX() ) || ( touchPanelY != touchPanelStatus.GetY() ) )
    {
        messageType = SM_MESSAGE_TUCHPANEL_MOTION;
    }
    else
    if ( touchPanelStatus.IsGrab() )
    {
        messageType = SM_MESSAGE_TUCHPANEL_MOTION;      // If grabbing, go to motion.
    }

    touchPanelX = touchPanelStatus.GetX();
    touchPanelY = touchPanelStatus.GetY();

    // Sends a message concerning the demo.
    if ( messageType != SM_MESSAGE_NONE )
    {
        s_Demo->SendMessage( messageType, &touchPanelStatus );
    }

    return messageType;
}


/*
  
 */
void
SendPadMassage( SmPadStatus& padStatus )
{
    // Framework pad operation
    if ( padStatus.IsHold( SmPadStatus::SM_BUTTON_R ) )
    {
        // Log output
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_A ) )
        {
            OutputLog();
        }

        // Enables/disables command double buffering.
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_B ) )
        {
            s_FrameStatus.SwapCommandBufferMode();
        }

        // Enables/disables layout.
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_Y ) )
        {
            s_FrameStatus.SwapLayoutVisible();
        }

        // Enables/disables CpuMeter.
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_X ) )
        {
            s_FrameStatus.SwapCpuMeterVisible();
        }
    }
    else
    {
        // Starts/stops play.
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_A ) )
        {
            s_FrameStatus.SwapPlay();
        }

        // Frame advance
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_X ) )
        {
            s_FrameStatus.SetAddFrame();
        }

        // Rewind
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_B ) )
        {
            s_FrameStatus.SetFrame( 0.f );
            s_Demo->Reset();
            NW_DEV_LOG( "[DEMO] Demo Reset\n\n" );
        }

        // Outputs hidden log.
        if ( padStatus.IsTrigger( SmPadStatus::SM_BUTTON_L ) )
        {
            NW_DEV_LOG( "[DEMO] FRAME_END - START:%8.4f[%8.4f%%]\n", s_FrameStatus.GetAllTime(), s_FrameStatus.GetAllTime()/16.666f );
            NW_DEV_LOG( "[DEMO] WAIT GPU         :%8.4f[%8.4f%%]\n", s_FrameStatus.GetGpuWaitTime(), s_FrameStatus.GetGpuWaitTime()/16.666f );
            NW_DEV_LOG( "[DEMO] COMMNAD REQ COUNT      :%d\n", s_FrameStatus.GetUsedReqCount() );
            NW_DEV_LOG( "[DEMO] STEREO PARAM LEVEL     :%f\n", s_Demo->GetStereoLevel() );
            NW_DEV_LOG( "[DEMO] STEREO PARAM RANGE     :%f\n", s_Demo->GetStereoRange() );
            s_GfxCtrl->OutputLog();
            NW_DEV_LOG( "\n" );
        }
    }

    // Sends a message concerning the demo.
    s_Demo->SendMessage( SM_MESSAGE_PAD_UPDATE, &padStatus );
}


/*
  
 */

#define _SET_PERF_CPU_BEGIN( costName, costColor ) \
    s_CpuPerf->Begin( costName, costColor ); \

#define _SET_PERF_CPU_END() \
    s_CpuPerf->End(); \

#define _SET_PERF_GPU_BEGIN( costName, costColor ) \
    s_GpuPerf->Begin( costName, costColor ); \

#define _SET_PERF_GPU_END() \
    s_GpuPerf->End(); \


/*
  
 */
void
CreateGfxCache( SmCommandReuser& CmdReuser, s32 CameraIndex, nw::gfx::Camera* CurrentCamera, SmRenderSystem::CommandBufferSide side )
{
    // Updates the scene.
    _SET_PERF_CPU_BEGIN( "GFX :update scene", nw::ut::Color8::MAGENTA );
    bool bUpdate = true;

    // Resets the render context.
    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();
    renderContext->ResetState();

    // Processes the demo content.
    // If less than 60 fps, changes the number of frames to be added to 2.f～3.f.
     if ( s_FrameStatus.IsPlaying() )
    {
        int vSyncCount = s_FrameStatus.UpdateVSyncCounter();

        vSyncCount = nw::ut::Clamp(vSyncCount, FrameWorkStatus::STEPFRAME_MIN, FrameWorkStatus::STEPFRAME_MAX);

        s_FrameStatus.AddFrame( vSyncCount );
    }
    else
    {
        if ( s_FrameStatus.GetAddFrame() )
        {
            s_FrameStatus.AddFrame( 1.f );
        }
        else
        {
            bUpdate = false;
        }
    }

    // Sets the current frame in the demo.
    s_Demo->Exec( s_FrameStatus.GetFrame() );

    // Processes updates if the frames have not been updated.
    if ( bUpdate )
    {
        // Updates the scene.
        s_GfxSceneCtrl->UpdateScene();

        // Updates the Sm scene.
        s_GfxCtrl->Update();
    }

    // Sets the render environment for lighting.
//        s_GfxSceneCtrl->SetEnvironment(CameraIndex);
    renderContext->SetActiveCamera(CameraIndex);

    // Updates camera/fog and processes the rendering software.
    s_GfxSceneCtrl->SubmitView(CurrentCamera);
    _SET_PERF_CPU_END();

     // Generates the command cache.
    _SET_PERF_CPU_BEGIN( "GFX :create draw command", nw::ut::Color8::RED );

    CmdReuser.StartCaching( s_RenderSystem->GetCacheCmdListUpper(side) );

    s_GfxSceneCtrl->Render();
    nngxFlush3DCommand();

    CmdReuser.EndCaching();

    _SET_PERF_CPU_END();
/*
    u32 uframe = s_FrameStatus.GetFrame();
    if (uframe % 200 == 0)
    {
        NW_DEV_LOG( "CMD BUFF SIZE:%d\n", CmdReuser.GetCacheSize() );
    }
*/
}

/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::Camera* currentCamera = NULL;

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeGfx();
    InitializeLyt();

    // Sets the GPU measurement callback.
    nngxBindCmdlist( s_RenderSystem->GetCmdList(SmRenderSystem::COMMAND_BUFFER_SIDE_0) );
    s_GpuPerf->SetCallback();
    nngxBindCmdlist( s_RenderSystem->GetCmdList(SmRenderSystem::COMMAND_BUFFER_SIDE_1) );
    s_GpuPerf->SetCallback();


    // Generates the demo class and playback class.
    // In the demo class, manages the Gfx object and layout.
    {
        s_Demo  = new Demo( s_GfxCtrl, s_LytCtrl );
        s_Demo->InitializeScene();
    }

    // Initializes and updates the scenes.
    s_GfxSceneCtrl->InitializeScene( s_GfxCtrl->GetSceneRoot());
    s_GfxSceneCtrl->UpdateScene();

    // Resolves and sets references for the scene environment.
    s_GfxSceneCtrl->SetSceneEnvironmentSetting(s_GfxCtrl->GetSceneEnvSetting());

    // Set the camera.
    SmCamera* pSmCamera = s_GfxCtrl->GetSmCamera();
    nw::gfx::SceneEnvironment& sceneEnvironment = renderContext->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, pSmCamera->GetBaseCamera());
    sceneEnvironment.SetCamera(s_PadCameraIndex, pSmCamera->GetPadCamera());

    // Sets the camera aspect ratio.
    nw::demo::Utility::SetCameraAspectRatio(s_GfxCtrl->GetSmCamera()->GetBaseCamera(), s_RenderTargets[0]);
    nw::demo::Utility::SetCameraAspectRatio(s_GfxCtrl->GetSmCamera()->GetPadCamera(), s_RenderTargets[0]);

    // Generates camera for stereo.
    {
        s_LeftCamera =
            nw::gfx::Camera::DynamicBuilder()
            .MaxChildren(0)
            .MaxCallbacks(0)
            .Create(&s_DeviceAllocator);

        s_RightCamera =
            nw::gfx::Camera::DynamicBuilder()
            .MaxChildren(0)
            .MaxCallbacks(0)
            .Create(&s_DeviceAllocator);
    }

    // Commands
    SmCommandReuser                 CmdReuser;

    SmRenderSystem::CommandBufferSide currentCmdListSide = SmRenderSystem::COMMAND_BUFFER_SIDE_0;
    SmRenderSystem::CommandBufferSide runningCmdListSide = SmRenderSystem::COMMAND_BUFFER_SIDE_1;

    SmTouchPanelStatus              touchPanelStatus;
    SmPadStatus                     padStatus;

    nn::hid::CTR::PadReader         padReader;
    nn::hid::CTR::TouchPanelReader  touchPanelReader;


    #define TRANSPARENCY_COLOR  (0)

    // Logs and outputs operational description.
    {
        NW_DEV_LOG( "[DEMO] PAD_A:PLAY/STOP\n" );
        NW_DEV_LOG( "[DEMO] PAD_X:ADD FRAME\n" );
        NW_DEV_LOG( "[DEMO] PAD_R+A:OUTPUT LOG\n" );
        NW_DEV_LOG( "[DEMO] PAD_R+B:COMMAND BUFFER DOUBLE/SINGLE SWAP\n" );
        NW_DEV_LOG( "[DEMO] PAD_R+Y:LAYOUT DRAW ON/OFF\n" );
        NW_DEV_LOG( "[DEMO] PAD_R+X:CPU METER DRAW ON/OFF\n" );
        NW_DEV_LOG( "\n" );
    }

    /*
        [Main loop]

        ・ Processing takes place in the following sequence.

          - Processes games.
          - gfx Renders scenes.
          - Renders layout.
          - Renders processing meters (GL).
          - Presents rendered screen.

        ・ Caution is required because the handling of the command list changes significantly depending on whether the mode is the double buffering mode.

        ・ Processing like buffer clearing takes place at stage rendering.

        ・Classes generated within demos derive from the SmMessage class. These classes handle the processing of messages from the touch panel, pad, and UI operations.
    
*/

    s_FrameStatus.UpdateVSyncCounter();

    while ( ! nw::demo::Utility::IsTerminating() )
    {
        // ------------------------------------------------------------------------
        // Prepares process measurements.

        // Records the frame start time.
        s_FrameStatus.SetTimerStart();

        // Resets process measurement.
        s_CpuPerf->Reset( s_FrameStatus.GetStartTime() );
        s_GpuPerf->Reset( s_FrameStatus.GetStartTime() );

        // ------------------------------------------------------------------------
        // Swaps the command list.

        if ( s_FrameStatus.IsCommandBufferDouble() )
        {
            // In the double buffering mode, commands generated in the preceding frame start executing here.
            nngxBindCmdlist( s_RenderSystem->GetCmdList(currentCmdListSide) );
            // Calls the nngxUpdateBuffer function on all device memory regions in accordance with the CTR-SDK performance tips. 
            // 
            nngxUpdateBuffer(
                reinterpret_cast<void*>(nn::os::GetDeviceMemoryAddress()),
                nn::os::GetDeviceMemorySize() );
            nngxRunCmdlist();
            s_GpuPerf->SetCmdlistRun();

            // Swaps the command lists.
            runningCmdListSide = currentCmdListSide;
            if ( runningCmdListSide == SmRenderSystem::COMMAND_BUFFER_SIDE_0 )
            {
                currentCmdListSide = SmRenderSystem::COMMAND_BUFFER_SIDE_1;
            }
            else
            {
                currentCmdListSide = SmRenderSystem::COMMAND_BUFFER_SIDE_0;
            }

            // Binds the command list that executes in the next frame.
            nngxBindCmdlist( s_RenderSystem->GetCmdList(currentCmdListSide) );
        }

        // ------------------------------------------------------------------------
        // Game processing

        _SET_PERF_CPU_BEGIN( "GAME:Update", nw::ut::Color8::BLACK );

        // Processes the touch panel messages.
        touchPanelReader.ReadLatest( touchPanelStatus.GetHidTouchPanelStatus() );
        SmMessageType touchPanelMessageType = SendTouchPanelMassage( touchPanelStatus );

        // Processes the pad messages.
        padReader.ReadLatest( padStatus.GetHidPadStatus() );
        SendPadMassage( padStatus );

        _SET_PERF_CPU_END();

        // ------------------------------------------------------------------------
        // Renders the upper screen (gfx).

        // Switches the render target.
        s_RenderSystem->SetRenderTarget(s_RenderTargets[DEMO_UPPER_SCREEN]);

        // Gets the current camera.
        currentCamera = s_GfxCtrl->GetCamera();
        NW_NULL_ASSERT( currentCamera );

        // Puts the commands that render the scene in cache.
        if( pSmCamera->IsPadCamera() )
        {
            CreateGfxCache( CmdReuser, s_PadCameraIndex, currentCamera, currentCmdListSide );
        }
        else
        {
            CreateGfxCache( CmdReuser, s_BaseCameraIndex, currentCamera, currentCmdListSide );
        }
        NW_ASSERT( CmdReuser.IsCache() );

        // Uses the commands in the cache to render the left and right screens.
        _SET_PERF_CPU_BEGIN( "GFX :draw", nw::ut::Color8::WHITE );

        // Calculates parallax.
        nw::gfx::res::ResCameraProjectionUpdater resProjectionUpdater =
            currentCamera->GetProjectionUpdater()->GetResource();
        NW_ASSERT(resProjectionUpdater.IsValid());

        // Calculates the parallax parameters.
        f32 flevel = s_Demo->GetStereoLevel() * (resProjectionUpdater.GetFar() - resProjectionUpdater.GetNear()) + resProjectionUpdater.GetNear();
        f32 frange = s_Demo->GetStereoRange();
        s_RenderSystem->CalcStereoCamera( s_LeftCamera, s_RightCamera, currentCamera, flevel, frange);

        // Renders and transfers the image for the left eye.
        _SET_PERF_GPU_BEGIN( "GPU :draw GFX (left)", nw::ut::Color8::RED );
        renderContext->SetCameraMatrix( s_LeftCamera );
        CmdReuser.UseCommand(false);
        _SET_PERF_GPU_END();
        _SET_PERF_GPU_BEGIN( "GPU :transfer upper screen", nw::ut::Color8::BLUE );
        s_RenderSystem->MakeTransferBufferCommand(nw::demo::UPPER_SCREEN, s_FrameStatus.IsCommandBufferDouble() );
        _SET_PERF_GPU_END();

        // Renders the right eye in the stereo render mode.
        if ( s_Demo->IsStereoEnable() )
        {
            _SET_PERF_GPU_BEGIN( "GPU :draw GFX (right)", nw::ut::Color8::RED );
            renderContext->SetCameraMatrix( s_RightCamera );
            CmdReuser.UseCommand(false);
            _SET_PERF_GPU_END();
            _SET_PERF_GPU_BEGIN( "GPU :transfer ext screen", nw::ut::Color8::BLUE );
             s_RenderSystem->MakeTransferBufferCommand(nw::demo::EXTENSION_SCREEN, s_FrameStatus.IsCommandBufferDouble() );
            _SET_PERF_GPU_END();
        }
        _SET_PERF_CPU_END();

        // ------------------------------------------------------------------------
        // Renders the lower screen (lyt and gl).

        // Switches the render target.
        s_RenderSystem->SetRenderTarget(s_RenderTargets[DEMO_LOWER_SCREEN]);

        // Skips processing if the layout is not rendered.
        if ( s_FrameStatus.IsLayoutVisible() )
        {
            _SET_PERF_CPU_BEGIN( "LYT :draw", nw::ut::Color8::GREEN );
            _SET_PERF_GPU_BEGIN( "GPU :draw LYT", nw::ut::Color8::GREEN );

            s_LytCtrl->Render();
            nngxFlush3DCommand();
            _SET_PERF_GPU_END();
            _SET_PERF_CPU_END();
        }

        // Skips processing if the load meter (GL render section) is not processed.
        if ( s_FrameStatus.IsCpuMeterVisible() )
        {
            // Updates the GL state.
            glBindFramebuffer(GL_FRAMEBUFFER, s_RenderSystem->GetRenderContext()->GetRenderTarget()->GetBufferObject());
            nngxUpdateState(NN_GX_STATE_ALL);

            // Renders the load meter.
            _SET_PERF_CPU_BEGIN( "PERF:draw", nw::ut::Color8::YELLOW );
            _SET_PERF_GPU_BEGIN( "GPU :draw PERF", nw::ut::Color8::YELLOW );

            s_SmOgl->SetBegin2D();
            s_PerfDrawer->Render( s_CpuPerf, s_GpuPerf );

            _SET_PERF_GPU_END();
            _SET_PERF_CPU_END();
        }

        // Transfers the lower screen.
        _SET_PERF_GPU_BEGIN( "GPU :transfer lower screen", nw::ut::Color8::BLUE );
        s_RenderSystem->MakeTransferBufferCommand(nw::demo::LOWER_SCREEN, s_FrameStatus.IsCommandBufferDouble() );
        _SET_PERF_GPU_END();

        // ------------------------------------------------------------------------
        // Presents the rendered screen.

        // Gets the total command list count.
        s_FrameStatus.UpdateUsedReqCount();

        // Processes the command list. The processing sequence changes depending on whether or not there is double buffering.
        if ( s_FrameStatus.IsCommandBufferDouble() )
        {
            // Waits for the completion of commands generated in the previous frame.
            s_FrameStatus.SetGpuWaitStart();
            nngxWaitCmdlistDone();
            s_FrameStatus.SetGpuWaitEnd();

            // Stops and clears the command list so that completed command lists can be used to generate commands in the next frame.
            // 
            nngxBindCmdlist( s_RenderSystem->GetCmdList(runningCmdListSide) );
            nngxStopCmdlist();
            nngxClearCmdlist();
        }
        else
        {
            // Calls the nngxUpdateBuffer function on all device memory regions in accordance with the CTR-SDK performance tips. 
            // 
            nngxUpdateBuffer(
                reinterpret_cast<void*>(nn::os::GetDeviceMemoryAddress()),
                nn::os::GetDeviceMemorySize() );
            // Starts executing the commands generated in this frame.
            nngxRunCmdlist();
            s_GpuPerf->SetCmdlistRun();

            // Waits for completion.
            s_FrameStatus.SetGpuWaitStart();
            nngxWaitCmdlistDone();
            s_FrameStatus.SetGpuWaitEnd();

            // Stops and clears the command list so that completed command lists can be used to generate/execute commands in the next frame.
            // 
            nngxStopCmdlist();
            nngxClearCmdlist();
        }

        // Swaps the display buffer and presets the display.
        s_RenderSystem->SwapBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        // Calculates all of the time required for a frame.
        s_FrameStatus.SetTimerEnd();

        // Waits for VSync.
        s_RenderSystem->WaitVSync(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);
    }

    // Destroy the cameras used for stereo.
    nw::gfx::SafeDestroy(s_LeftCamera);
    nw::gfx::SafeDestroy(s_RightCamera);

    SM_SAFE_DELETE( s_Demo );

    TerminateGfx();
    TerminateLyt();
}

} // Namespace


/*
  
 */
void
nnMain()
{
    nn::applet::Enable();
    
    // Initializes the graphics system.
    SmInitializeGraphicsSystem( &s_DeviceAllocator );

    // Initializes the layout library.
    nw::lyt::Initialize(&s_DeviceAllocator, &s_DeviceAllocator);

    // initialized the touch  panel.
    nn::hid::CTR::Initialize();

    {
        InitializeGraphics();
        nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

        {
            DemoScene();
        }

        nw::demo::PadFactory::Finalize();
        TerminateGraphics();
    }

    // Terminates the graphics system.
    SmFinalizeGraphicsSystem( &s_DeviceAllocator );
}


