/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_DEMO_GRAPHICSSYSTEM_H_
#define NW_DEMO_GRAPHICSSYSTEM_H_

#include <nw/demo/demo_DisplayBufferSwapper.h>
#include <nw/demo/demo_CameraController.h>
#include <nw/demo/demo_CommandListSwapper.h>

#include <nw/gfx/gfx_IRenderTarget.h>
#include <nw/gfx/gfx_RenderQueue.h>
#include <nw/gfx/gfx_SceneUpdater.h>

#include <nw/ut/ut_MoveArray.h>
#include <nw/ut/ut_ResUtil.h>
#include <nw/math.h>

#include <nn/gx.h>
#include <nn/ulcd/CTR/ulcd_StereoCamera.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace nw {
namespace gfx {

class RenderContext;
class MeshRenderer;
enum RenderColorFormat;
class Shader;

class SceneNode;
class SceneContext;
class SceneTraverser;
class SceneInitializer;
class Camera;
} // namespace gfx

namespace os {

class IAllocator;

} // namespace os

namespace ut {

struct FloatColor;

} // namespace ut

namespace demo {

class RenderSystem;
class SceneSystem;
class GraphicsDrawing;
class DemoAllocator;

//---------------------------------------------------------------------------
//
//
//---------------------------------------------------------------------------
enum UpperScreenMode
{
    UPPER_SCREEN_MODE_NORMAL = NN_GX_DISPLAYMODE_NORMAL,       //
    UPPER_SCREEN_MODE_STEREO = NN_GX_DISPLAYMODE_STEREO        //
};

//
struct ResourceSet
{
    nw::ut::MoveArray<u8> buffer;
    nw::gfx::ResGraphicsFile resource;
};

const int MAX_RESOURCES_COUNT = 128;
typedef nw::ut::FixedSizeArray<ResourceSet, MAX_RESOURCES_COUNT> ResourceArray;

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class RenderSystem
{
public:

    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    struct Description
    {
        //
        struct ScreenSize
        {
            s32 width;  //
            s32 height; //
            //
            ScreenSize(s32 w, s32 h) : width(w), height(h) {}
        };

        //
        struct ScreenOffset
        {
            s32 x;  //
            s32 y;  //
            //
            ScreenOffset(s32 x, s32 y) : x(x), y(y) {}
        };
        //
        DisplayBufferSwapper::Description upperScreenDescription;

        //
        DisplayBufferSwapper::Description lowerScreenDescription;

        gfx::RenderColorFormat renderColorFormat;   //
        size_t commandBufferSize;                   //
        size_t commandRequestCount;                 //
        size_t reusableCommandBufferSize;           //
        size_t reusableCommandRequestCount;         //
        UpperScreenMode upperScreenMode;            //

        size_t commandListCount;                    //
        size_t displayBufferCount;                  //

        size_t maxGpuProfilingEntryCount;           //

        //
        //
        Description()
        : renderColorFormat(nw::gfx::RENDER_COLOR_FORMAT_RGBA8),
          commandBufferSize(0x100000),
          commandRequestCount(512),
          reusableCommandBufferSize(0),
          reusableCommandRequestCount(0),
          upperScreenMode(UPPER_SCREEN_MODE_NORMAL),
          commandListCount(2),
          displayBufferCount(2),
          maxGpuProfilingEntryCount(8)
        {
            upperScreenDescription.screenKind = nw::demo::UPPER_SCREEN;
            upperScreenDescription.width  = NN_GX_DISPLAY0_HEIGHT;
            upperScreenDescription.height = NN_GX_DISPLAY0_WIDTH;

            lowerScreenDescription.screenKind = nw::demo::LOWER_SCREEN;
            lowerScreenDescription.width  = NN_GX_DISPLAY1_HEIGHT;
            lowerScreenDescription.height = NN_GX_DISPLAY1_WIDTH;
        }
    };

    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    struct LoadMeterDescription
    {
        //
        LoadMeterDescription()
            : startTick(0),
              loadTime(0.0f),
              loadGpuTime(0.0f),
              waitTime(0.0f),
              transferTime(0.0f),
              gpuOtherTime(0.0f),
              cumulativeTime(0.0f),
              cumulativeGpuTime(0.0f),
              cumulativeWaitTime(0.0f),
              drawCommandBufferSize(0),
              callCount(0)
        {}

        s64 startTick; //
        float loadTime; //
        float loadGpuTime; //
        float waitTime; //
        float transferTime; //
        float gpuOtherTime; //
        float cumulativeTime; //
        float cumulativeGpuTime; //
        float cumulativeWaitTime; //

        int drawCommandBufferSize; //

        s32 callCount; //

        static const float NANO_TO_MILLI;
        static const float MILLI_TO_RATE;
    };

    //----------------------------------------
    //
    //


    //
    //
    //
    //
    //
    //
    //
    static RenderSystem* Create(
        os::IAllocator* allocator,
        const Description& description
    );

    //
    //
    //
    //
    void LoadSkyModel(const wchar_t* fileName);

    //
    void Destroy();

    //

    //----------------------------------------
    //
    //

    //
    //
    //
    //
    void SetRenderTarget(gfx::IRenderTarget* renderTarget);

    //
    //
    //
    //
    gfx::RenderContext*  GetRenderContext() { return this->m_RenderContext; }

    //
    //
    //
    //
    gfx::SceneEnvironment&  GetSceneEnvironment()
    {
        NW_NULL_ASSERT(this->m_RenderContext);
        return this->m_RenderContext->GetSceneEnvironment();
    }

    //
    //
    //
    //
    void SetRenderSortMode(gfx::ISceneUpdater::RenderSortMode renderSortMode);

    //
    //
    //
    //
    gfx::ISceneUpdater::RenderSortMode GetRenderSortMode() const
    {
        return this->m_RenderSortMode;
    }

    //
    //
    //
    //
    gfx::RenderQueue* GetRenderQueue() const
    {
        return this->m_RenderQueue;
    }

    //
    //
    //
    //
    gfx::MeshRenderer* GetMeshRenderer() const
    {
        return this->m_MeshRenderer;
    }

    //
    CommandListSwapper* GetCommandListSwapper() const
    {
        return this->m_CommandListSwapper;
    }

    //
    //
    //
    //
    //
    //
    //
    void EnqueueRenderCommand(
        gfx::RenderCommand* renderCommand,
        gfx::ResMaterial::TranslucencyKind translucencyKind,
        u8 priority,
        u8 layerId)
    {
        this->m_RenderQueue->EnqueueCommand(
            renderCommand,
            translucencyKind,
            priority,
            layerId);
    }

    //

    //----------------------------------------
    //
    //

    //
    //
    //
    //
    //
    void SetSceneEnvironmentSettings(
        SceneSystem* sceneSystem,
        ut::MoveArray<gfx::SceneEnvironmentSetting*>* sceneEnvironmentSettings);

    //
    //
    //
    //
    //
    void SetEnvironment(
        SceneSystem* sceneSystem);

    //
    //
    //
    //
    void SubmitView(
        SceneSystem* sceneSystem);

    //
    //
    //
    //
    //
    //
    void RenderScene(
        gfx::Camera* camera,
        s32 screenKind,
        bool isCommandCacheDump = false);

    //
    //
    //
    //
    //
    //
    //
    //
    void RenderStereoScene(
        gfx::Camera* leftCamera,
        gfx::Camera* rightCamera,
        bool isCommandCacheDump = false);

    //
    //
    //
    //
    void ClearBuffer(ut::FloatColor clearColor);

    //
    //
    //
    //
    void ClearBySkyModel(const gfx::Camera* camera);

    //
    //
    //
    //
    void TransferBuffer(s32 screenKind);

    //
    void WaitCommandList();

    //
    void RunCommandList();

    //
    void SwapCommandList();

    //
    //
    //
    //
    void WaitVSync(s32 screenKind);

    //
    //
    //
    //
    void SwapBuffer(s32 screenKind);

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    void PresentBuffer(s32 screenKind);

    //
    //
    //
    //
    //
    //
    //
    //
    void CalcStereoCamera(
        gfx::Camera* leftCamera,
        gfx::Camera* rightCamera,
        gfx::Camera* baseCamera,
        f32 depthLevel = 1.0f,
        f32 depthRange = 1.0f);

    //

    //----------------------------------------
    //
    //

    //
    void BeginLoadMeter();

    //
    void EndLoadMeter();

    //
    void CalcLoadMeter();

    //
    void SuspendLoadMeter();

    //
    void ResumeLoadMeter();

    //
    void ResetCumulativeLoadMeter();

    //
    const LoadMeterDescription& GetLoadMeter() { return m_LoadMeterDescription; }

    //
    const LoadMeterDescription& GetCumulativeLoadMeter() { return m_CumulativeLoadMeterDescription; }

    //
    void AddLoadMeterCommandSize(s32 commandSize) { m_LoadMeterDescription.drawCommandBufferSize += commandSize; }

    //
    //
    //
    //
    //
    ut::Signal1<void, const gfx::Camera*>* GetPostRenderCallback() const
    {
        return this->m_PostRenderCallback;
    }

    //
    //
    //
    //
    ut::Signal1<void, const gfx::Camera*>* GetPostRightRenderCallback() const
    {
        return this->m_PostRightRenderCallback;
    }
    //

private:
    //
    RenderSystem()
    : m_Allocator(NULL),
      m_RenderContext(NULL),
      m_PriorMaterialRenderKeyFactory(NULL),
      m_RenderQueue(NULL),
      m_MeshRenderer(NULL),
      m_RenderSortMode(gfx::ISceneUpdater::ALL_MESH_BASE_SORT),
      m_UpperSwapper(NULL),
      m_LowerSwapper(NULL),
      m_ExtensionSwapper(NULL),
      m_CommandListSwapper(NULL),
      m_StereoCamera(NULL),
      m_SkyModel(NULL),
      m_PostRenderCallback(NULL),
      m_PostRightRenderCallback(NULL)
    {
    }

    os::IAllocator* m_Allocator;
    gfx::RenderContext* m_RenderContext;
    gfx::RenderKeyFactory* m_PriorMaterialRenderKeyFactory;
    gfx::RenderKeyFactory* m_PriorDepthRenderKeyFactory;
    gfx::RenderQueue* m_RenderQueue;
    gfx::MeshRenderer* m_MeshRenderer;
    gfx::ISceneUpdater::RenderSortMode m_RenderSortMode;

    DisplayBufferSwapper* m_UpperSwapper;
    DisplayBufferSwapper* m_LowerSwapper;
    DisplayBufferSwapper* m_ExtensionSwapper;

    CommandListSwapper* m_CommandListSwapper;

    LoadMeterDescription m_LoadMeterDescription;
    LoadMeterDescription m_CumulativeLoadMeterDescription;

    nn::ulcd::CTR::StereoCamera* m_StereoCamera;
    ResourceArray m_Resources;
    gfx::Model* m_SkyModel;

    ut::Signal1<void, const gfx::Camera*>* m_PostRenderCallback;
    ut::Signal1<void, const gfx::Camera*>* m_PostRightRenderCallback;
};


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class SceneSystem
{
public:
    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    struct Description
    {
        s32 maxSceneNodes; //
        s32 maxModels; //
        s32 maxSkeletalModels; //
        s32 maxCameras; //
        s32 maxLights; //
        s32 maxFragmentLights; //
        s32 maxVertexLights; //
        s32 maxHemiSphereLights; //
        s32 maxAmbientLights; //
        s32 maxFogs; //
        s32 maxParticleSets; //
        s32 maxParticleEmitters; //
        s32 maxParticleModels; //
        s32 maxBones; //
        s32 maxMaterials; //
        s32 maxUserRenderNodes; //
        bool isFixedSizeMemory; //

        //
        Description()
        : maxSceneNodes(gfx::SceneContext::DEFAULT_MAX_SCENE_NODES),
          maxModels(gfx::SceneContext::DEFAULT_MAX_MODELS),
          maxSkeletalModels(gfx::SceneContext::DEFAULT_MAX_SKELETAL_MODELS),
          maxCameras(gfx::SceneContext::DEFAULT_MAX_CAMERAS),
          maxLights(gfx::SceneContext::DEFAULT_MAX_LIGHTS),
          maxFragmentLights(gfx::SceneContext::DEFAULT_MAX_FRAGMENT_LIGHTS),
          maxVertexLights(gfx::SceneContext::DEFAULT_MAX_VERTEX_LIGHTS),
          maxHemiSphereLights(gfx::SceneContext::DEFAULT_MAX_HEMISPHERE_LIGHTS),
          maxAmbientLights(gfx::SceneContext::DEFAULT_MAX_AMBIENT_LIGHTS),
          maxFogs(gfx::SceneContext::DEFAULT_MAX_FOGS),
          maxParticleSets(gfx::SceneContext::DEFAULT_MAX_PARTICLESETS),
          maxParticleEmitters(gfx::SceneContext::DEFAULT_MAX_PARTICLEEMITTERS),
          maxParticleModels(gfx::SceneContext::DEFAULT_MAX_PARTICLEMODELS),
          maxBones(128),
          maxMaterials(32),
          maxUserRenderNodes(gfx::SceneContext::DEFAULT_MAX_USER_RENDER_NODES),
          isFixedSizeMemory(true)
        {}
    };

    //----------------------------------------
    //
    //

    //---------------------------------------------------------------------------
    //
    //
    //
    //
    //
    //
    //---------------------------------------------------------------------------
    static SceneSystem* Create(
        os::IAllocator* allocator,
        const Description& description
    );

    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    void Destroy();
    //

    //---------------------------------------------------------------------------
    //
    //
    //
    //---------------------------------------------------------------------------
    void InitializeScene(gfx::SceneNode* sceneRoot);

    //---------------------------------------------------------------------------
    //
    //
    //
    //---------------------------------------------------------------------------
    void TraverseScene(gfx::SceneNode* sceneRoot);

    //---------------------------------------------------------------------------
    //
    //---------------------------------------------------------------------------
    void UpdateScene();

    //
    gfx::SceneContext*   GetSceneContext() { return m_SceneContext; }

    //
    gfx::ISceneUpdater* GetSceneUpdater() { return m_SceneUpdater; }

    //
    nw::demo::CameraController* GetCameraController() { return m_CameraController; }

private:
    SceneSystem()
    : m_Allocator(NULL),
      m_SceneContext(NULL),
      m_SceneTraverser(NULL),
      m_SceneUpdater(NULL),
      m_SceneInitializer(NULL),
      m_CameraController(NULL)
    {}

    os::IAllocator*      m_Allocator;
    gfx::SceneContext*   m_SceneContext;
    gfx::SceneTraverser* m_SceneTraverser;
    gfx::SceneUpdater*   m_SceneUpdater;
    gfx::SceneInitializer* m_SceneInitializer;
    CameraController*    m_CameraController;
};

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void InitializeGraphicsSystem(nw::demo::DemoAllocator* deviceAllocator);

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void FinalizeGraphicsSystem();

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
template <typename TArray>
NW_INLINE void
SafeCleanupResources(TArray& resources)
{
    typename TArray::iterator end = resources.end();
    for (typename TArray::iterator it = resources.begin();
        it != end;
        ++it)
    {
        nw::ut::SafeCleanup(it->resource);
    }
}


//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void* AllocateGraphicsMemory(GLenum area, GLenum aim, GLuint id, GLsizei size);


//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void DeallocateGraphicsMemory(GLenum area, GLenum aim, GLuint id, void* addr);

} // namespace demo
} // namespace nw
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif

#endif // NW_DEMO_GRAPHICSSYSTEM_H_
