/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_DEMO_MEMORY_H_
#define NW_DEMO_MEMORY_H_

#include <nw/types.h>
#include <nw/os/os_Memory.h>
#include <nn/dbg.h>
#include <nn/fnd.h>

namespace nw {
namespace demo {

class DemoAllocator;

// In the demo, all memory is to be allocated from the device memory.

// The memory size used by the demo.
const size_t DEMO_MEMORY_SIZE = 0x1000000;

// The memory size for particles.
const size_t DEMO_PARTICLE_MEMORY_SIZE = 0x100000;

//=====================
//
//

//
//
//
//
//
//
void InitializeDemoAllocator(DemoAllocator* allocator, size_t size, bit32 option = 0 );

//
//
//
//
void FinalizeDemoAllocator(DemoAllocator* allocator);

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void InitializeDemoMemory();

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
void* UnAlignMemory( void* memory );

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void* Alloc(size_t size, u8 alignment = 4);

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void  Free(void* memory);

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void Dump();

//

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
class DemoAllocator : public nw::os::IAllocator
{
public:
    //
    DemoAllocator()
    : m_StartAddress(0),
      m_BreakAllocAddress(NULL),
      m_BreakAllocSize(0),
      m_BreakFreeAddress(NULL),
      m_BreakFreeSize(0)
    {}

    //=====================
    //
    //

    //
    //
    //
    //
    //
    //
    //
    //
    void Initialize(uptr startAddress, size_t size, bit32 option = 0);

    //
    //
    void Finalize();

    //


    //
    //
    //
    uptr GetStartAddress() const { return m_StartAddress; }

    //=====================
    //
    //

    //
    //
    //
    //
    //
    //
    virtual void* Alloc(size_t size, u8 alignment)
    {
        NW_ASSERT(size != 0);

        void* memory = m_Heap.Allocate(size, alignment);

#if defined(NW_DEBUG_TRAP_ALLOCATOR)
        if (memory != NULL)
        {
            void* unalignMemory = UnAlignMemory(memory);
            if (unalignMemory == m_BreakAllocAddress &&
                (m_BreakAllocSize == 0 || GetMemoryBlockSize(unalignMemory) == m_BreakAllocSize))
            {
                NW_DEV_LOG("Alloc(): Specified memory block allocated\n");
    #if defined(NW_DEBUG_BREAK_ALLOCATOR)
                nn::dbg::Break(nn::dbg::BREAK_REASON_USER);
    #endif
            }
        }
#endif

        return memory;
    }

    using nw::os::IAllocator::Alloc;

    //
    //
    //
    virtual void Free(void* memory)
    {
#if defined(NW_DEBUG_TRAP_ALLOCATOR)
        void* unalignMemory = UnAlignMemory(memory);
        if (unalignMemory == m_BreakFreeAddress &&
            (m_BreakFreeSize == 0 || GetMemoryBlockSize(unalignMemory) == m_BreakFreeSize))
        {
            NW_DEV_LOG("Free(): Specified memory block freed\n");
#if defined(NW_DEBUG_BREAK_ALLOCATOR)
            nn::dbg::Break(nn::dbg::BREAK_REASON_USER);
#endif
        }
#endif

        m_Heap.Free(memory);
    }

    //
    //=====================
    //
    //



    //---------------------------------------------------------------------------
    //
    //
    //
    //---------------------------------------------------------------------------
    size_t GetFreeSize();

    //---------------------------------------------------------------------------
    //
    //
    //
    //---------------------------------------------------------------------------
    size_t GetTotalSize();

    //---------------------------------------------------------------------------
    //
    //
    //
    //
    //
    //---------------------------------------------------------------------------
    size_t GetMemoryBlockSize(void* address);


    //
    void Dump();

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    void SetBreakAlloc(int address, size_t size)
    {
#if defined(NW_DEBUG_TRAP_ALLOCATOR)
        m_BreakAllocAddress = reinterpret_cast<void *>(address);
        m_BreakAllocSize = size;
#else
        NW_UNUSED_VARIABLE(address);
        NW_UNUSED_VARIABLE(size);
#endif
    }

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    void SetBreakFree(int address, size_t size)
    {
#if defined(NW_DEBUG_TRAP_ALLOCATOR)
        m_BreakFreeAddress = reinterpret_cast<void *>(address);
        m_BreakFreeSize = size;
#else
        NW_UNUSED_VARIABLE(address);
        NW_UNUSED_VARIABLE(size);
#endif
    }

    //



private:
    uptr m_StartAddress;

    void* m_BreakAllocAddress;
    size_t m_BreakAllocSize;

    void* m_BreakFreeAddress;
    size_t m_BreakFreeSize;

    nn::fnd::ExpHeap m_Heap;
};


//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
class FrameHeapAllocator : public nw::os::IAllocator
{
public:
    //
    FrameHeapAllocator()
    : m_StartAddress(0)
    {}

    //
    //
    //
    //
    //
    //
    //
    //
    FrameHeapAllocator(uptr startAddress, size_t size, bit32 option = 0)
    : m_StartAddress(startAddress)
    {
        Initialize(startAddress, size, option);
    }

    //=====================
    //
    //

    //
    //
    //
    //
    //
    //
    //
    //
    void Initialize(uptr startAddress, size_t size, bit32 option = 0);

    //
    //
    void Finalize();

    //

    //
    //
    //
    uptr GetStartAddress() const { return m_StartAddress; }

    //=====================
    //
    //

    //
    //
    //
    //
    //
    //
    virtual void* Alloc(size_t size, u8 alignment)
    {
        NW_ASSERT(size != 0);

        return m_FrameHeap.Allocate(size, alignment);
    }

    //
    virtual void Free(void*)
    {
        // Since the frame heap can only deallocate everything, nothing is done.
    }

    //
    void FreeAll()
    {
        m_FrameHeap.Free();
    }

    //

private:
    uptr m_StartAddress;

    nn::fnd::FrameHeap m_FrameHeap;
};


} // namespace demo
} // namespace nw

/* NW_DEMO_MEMORY_H_ */
#endif

