/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


//------------------------------------------------------------------
// Demo: ResFont
//
// Overview
//   This is a sample of building and destroying nw::font::ResFont.
//
// Controls
//   None.
//
//------------------------------------------------------------------

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <nn.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/applet.h>
#include <nw/font/font_TextWriter.h>
#include <nw/font/font_ResFont.h>
#include <nw/demo.h>

namespace
{

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
MemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
#if defined(NW_PLATFORM_CTR)
    return nw::demo::SimpleApp::Allocate(size, alignment);
#else
    return nw::demo::Alloc(size, alignment);
#endif
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void*
DevMemAlloc(
    size_t  size,
    u8      alignment = 4
)
{
#if defined(NW_PLATFORM_CTR)
    return nw::demo::SimpleApp::AllocateDeviceMemory(size, alignment);
#else
    return nw::demo::Alloc(size, alignment);
#endif
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
MemFree(void* memory)
{
#if defined(NW_PLATFORM_CTR)
    nw::demo::SimpleApp::Free(memory);
#else
    nw::demo::Free(memory);
#endif
}

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
InitShaders(nw::font::TextWriterResource* pResource)
{
    const wchar_t* shaders = NW_DEMO_FILE_PATH( NW_FONT_SHADERBINARY );
    nn::fs::FileReader shaderReader(shaders);

    const u32 fileSize = (u32)shaderReader.GetSize();

    void* shaderBinary = MemAlloc(fileSize);

    s32 read = shaderReader.Read(shaderBinary, fileSize);
    NN_ASSERT(read == fileSize);

    pResource->InitResource(shaderBinary, fileSize);

    MemFree(shaderBinary);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
InitDraw()
{
    glClearColor(0.3f, 0.3f, 0.3f, 1.0f);

    glDisable(GL_CULL_FACE);            // Disable culling
    glDisable(GL_SCISSOR_TEST);         // Disable scissoring
    glDisable(GL_POLYGON_OFFSET_FILL);  // Disable polygon offset
#if defined(NW_PLATFORM_CTR)
    glDisable(GL_EARLY_DEPTH_TEST_DMP); // Disable early depth test
#endif
    glDisable(GL_DEPTH_TEST);           // Disable depth test
    glDisable(GL_STENCIL_TEST);         // Disable stencil test

    // Make all elements of the color buffer writable
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
bool
InitFont(
    nw::font::ResFont*  pFont,
    const wchar_t*      filePath
)
{
    // Loads font resources
    nn::fs::FileReader fontReader(filePath);

    s32 fileSize = (s32)fontReader.GetSize();
    if ( fileSize <= 0 )
    {
        return false;
    }

    void* buffer = DevMemAlloc(fileSize, nw::font::GlyphDataAlignment);
    if (buffer == NULL)
    {
        return false;
    }

    s32 readSize = fontReader.Read(buffer, fileSize);
    if (readSize != fileSize)
    {
        MemFree(buffer);
        return false;
    }

    // Sets the font resource.
    bool bSuccess = pFont->SetResource(buffer);
    NN_ASSERT(bSuccess);

    //--- Fails if the resource is already set, already loaded, or invalid.
    if (! bSuccess)
    {
        MemFree(buffer);
    }

    // Sets the render buffer.
    // Rendering is possible without making this setting, but the processing speed gets slower.
    // 
    const u32 drawBufferSize = nw::font::ResFont::GetDrawBufferSize(buffer);
    void* drawBuffer = MemAlloc(drawBufferSize, 4);
    pFont->SetDrawBuffer(drawBuffer);

    return bSuccess;
}


//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
void
CleanupFont(nw::font::ResFont* pFont)
{
    // Disable render buffer
    // When the render buffer is set, the pointer to the buffer passed to SetDrawBuffer at build time is returned.
    // 
    void *const drawBuffer = pFont->SetDrawBuffer(NULL);
    if (drawBuffer != NULL)
    {
        MemFree(drawBuffer);
    }

    // Returns a pointer to the resource passed when SetResource was called if a font is set.
    // 
    void *const resource = pFont->RemoveResource();
    if (resource != NULL)
    {
        MemFree(resource);
    }

    // After RemoveResource is executed, data cannot be used as a font until SetResource is executed again.
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
void
SetupTextCamera(
    const nw::font::TextWriterResource&
                textWriterResource,
    int         width,
    int         height
)
{
#ifdef NW_PLATFORM_CTRWIN
    const GLsizei lcdWidth  = width;
    const GLsizei lcdHeight = height;
#else
    const GLsizei lcdWidth  = height;
    const GLsizei lcdHeight = width;
#endif
    glViewport(0, 0, lcdWidth, lcdHeight);

    // Set the projection matrix for an orthogonal projection
    {
        // Sets the origin at the upper left, and the Y and Z axes in opposite directions.
        nn::math::MTX44 proj;
        f32 znear   =  0.0f;
        f32 zfar    = -1.0f;
        f32 t       =  0;
        f32 b       =  static_cast<f32>(height);
        f32 l       =  0;
        f32 r       =  static_cast<f32>(width);
        nn::math::MTX44OrthoPivot(&proj, l, r, b, t, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP);
        textWriterResource.SetProjectionMtx(proj);
    }

    // Set the model view matrix as an identity matrix
    {
        nn::math::MTX34 mv;
        nn::math::MTX34Identity(&mv);
        textWriterResource.SetViewMtx(mv);
    }
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void
DrawAscii(
    nw::font::ResFont*              pFont,
    nw::font::TextWriterResource*   pTextWriterResource
)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    nw::font::TextWriter writer;
    writer.SetFont(pFont);
    writer.SetTextWriterResource(pTextWriterResource);
    writer.SetupGX();
    writer.SetCursor(0, 0);

    (void)writer.Print("DEMO: ResFont\n");
    (void)writer.Print("\n");

    // Display a sample of ASCII characters
    (void)writer.Print("All ASCII Character listing:\n");
    (void)writer.Print("\n");
    (void)writer.Print(" !\"#$%&'()*+,-./\n");
    (void)writer.Print("0123456789:;<=>?\n");
    (void)writer.Print("@ABCDEFGHIJKLMNO\n");
    (void)writer.Print("PQRSTUVWXYZ[\\]^_\n");
    (void)writer.Print("`abcdefghijklmno\n");
    (void)writer.Print("pqrstuvwxyz{|}~\n");

    glFlush();
}

}   // Namespace

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
nnMain()
{
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();

    demoApp.Initialize();

    nw::font::ResFont font;

    // Build font
    {
        bool bSuccess = InitFont(&font, NW_DEMO_FILE_PATH(L"tahoma.bcfnt"));
        NN_ASSERTMSG(bSuccess, "Fail to load ResFont.");
    }

    // Build the render resource
    nw::font::TextWriterResource textWriterResource;
    InitShaders(&textWriterResource);

    textWriterResource.ActiveGlProgram();

    InitDraw();

    // Display font data
    volatile bool loop = true;
    while (loop)
    {
        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            glBindFramebuffer(GL_FRAMEBUFFER, demoApp.GetFrameBufferObject());

            SetupTextCamera(
                textWriterResource,
                demoApp.DISPLAY0_WIDTH,
                demoApp.DISPLAY0_HEIGHT);
            DrawAscii(&font, &textWriterResource);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            glBindFramebuffer(GL_FRAMEBUFFER, demoApp.GetFrameBufferObject());

            SetupTextCamera(
                textWriterResource,
                demoApp.DISPLAY1_WIDTH,
                demoApp.DISPLAY1_HEIGHT);
            DrawAscii(&font, &textWriterResource);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    // Destroy render resource
    textWriterResource.DeleteResource();

    // Destroy font
    CleanupFont(&font);
}
