/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_FONT_CHARWRITER_H_
#define NW_FONT_CHARWRITER_H_

#if defined(_MSC_VER) && _MSC_VER >= 1500
#pragma once
#endif

// Upper left origin point
#define COORDINATE_LT 1

#include <cstddef>
#include <cstdio>
#include <GLES2/gl2.h>
#include <nn/types.h>
#include <nw/math.h>
#include <nw/ut/ut_Color.h>
#include <nw/font/font_Font.h>
#include <nw/font/font_TextWriterResource.h>
#include <nw/font/font_Stopwatch.h>
#include <nw/math/math_Types.h>

#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
#endif
// Shader binary filename for character rendering
#define NW_FONT_SHADERBINARY    L"nwfont_TextWriterShader.shbin"

#define NW_FONT_RECTDRAWER_SHADERBINARY L"nwfont_RectDrawerShader.shbin"

namespace nw {

namespace font {

struct DispStringBuffer;
class RectDrawer;

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
class CharWriter
{
public:
    /* ------------------------------------------------------------------------
            Types
       ------------------------------------------------------------------------ */

    //
    enum GradationMode
    {
        GRADMODE_NONE,                          //
        GRADMODE_H,                             //
        GRADMODE_V,                             //

        NUM_OF_GRADMODE
    };



    /* ------------------------------------------------------------------------
            Constants
       ------------------------------------------------------------------------ */
    static const u32 DEFAULT_COLOR_MAPPING_MIN = 0x00000000UL;
    static const u32 DEFAULT_COLOR_MAPPING_MAX = 0xFFFFFFFFUL;


    /* ------------------------------------------------------------------------
            Functions
       ------------------------------------------------------------------------ */

    //
    //

    //
                        CharWriter();

    //
                        ~CharWriter();

    //

    //
    //

    //
    //
    //
    //
    void                SetFont(const Font* pFont)  { m_pFont = pFont; }

    //
    //
    //
    //
    //
    const Font*         GetFont() const             { return m_pFont; }

    //

    //
    //

    //
    void                SetupGX();

    //
    static void         FinalizeGX()
    {
        TextWriterResource::FinalizeGX();
    }

    //

    //
    //

    //
    //
    //
    //
    //
    void                SetColorMapping(
                            ut::Color8  min,
                            ut::Color8  max
    )
    {
        m_ColorMapping.min = min;
        m_ColorMapping.max = max;
    }

    //
    //
    //
    //
    const ut::Color8    GetColorMappingMin() const { return m_ColorMapping.min; }

    //
    //
    //
    //
    const ut::Color8    GetColorMappingMax() const { return m_ColorMapping.max; }

    //
    void                ResetColorMapping()
    {
        SetColorMapping(DEFAULT_COLOR_MAPPING_MIN, DEFAULT_COLOR_MAPPING_MAX);
    }

    //

    //
    //

    //
    //
    //
    //
    void                SetAlpha(u8 alpha)
    {
        m_Alpha = alpha;
    }

    //Gets additional alpha value when rendering characters.

    //
    //
    //
    //
    u8                  GetAlpha() const    { return m_Alpha; }

    //
    //
    //
    //
    void                SetGradationMode(GradationMode mode)
    {
        NW_FONT_MINMAX_ASSERT(mode, GRADMODE_NONE, NUM_OF_GRADMODE - 1);
        m_GradationMode = mode;
        UpdateVertexColor();
    }

    //
    //
    //
    //
    GradationMode       GetGradationMode() const    { return m_GradationMode; }

    //
    //
    //
    //
    void                SetTextColor(ut::Color8 color)
    {
        m_TextColors[internal::TEXTCOLOR_START] = color;
        UpdateVertexColor();
    }

    //
    //
    //
    //
    //
    void                SetTextColor(
                            ut::Color8  start,
                            ut::Color8  end
    )
    {
        m_TextColors[internal::TEXTCOLOR_START] = start;
        m_TextColors[internal::TEXTCOLOR_END  ] = end;
        UpdateVertexColor();
    }

    //
    //
    //
    //
    const ut::Color8     GetTextColor() const           { return m_TextColors[internal::TEXTCOLOR_START]; }

    //
    //
    //
    //
    const ut::Color8     GetGradationStartColor() const { return m_TextColors[internal::TEXTCOLOR_START]; }

    //
    //
    //
    //
    const ut::Color8     GetGradationEndColor() const   { return m_TextColors[internal::TEXTCOLOR_END]; }

    //

    //
    //

    //
    //
    //
    //
    //
    void                SetScale(
                            f32 hScale,
                            f32 vScale
    )
    {
        m_Scale.x = hScale;
        m_Scale.y = vScale;
    }

    //
    //
    //
    //
    void                SetScale(f32 hvScale)
    {
        m_Scale.x = hvScale;
        m_Scale.y = hvScale;
    }

    //
    //
    //
    //
    f32                 GetScaleH() const   { return m_Scale.x; }

    //
    //
    //
    //
    f32                 GetScaleV() const   { return m_Scale.y; }

    //
    //
    //
    //
    //
    void                SetFontSize(
                            f32 width,
                            f32 height);

    //
    //
    //
    //
    void                SetFontSize(f32 height);

    //
    //
    //
    //
    //
    f32                 GetFontWidth() const;

    //
    //
    //
    //
    //
    f32                 GetFontHeight() const;

    //
    //
    //
    //
    f32                 GetFontAscent() const;

    //
    //
    //
    //
    f32                 GetFontDescent() const;

    //

    //
    //

    //
    //
    //
    //
    void                EnableFixedWidth(bool isFixed)  { m_IsWidthFixed = isFixed; }

    //
    //
    //
    //
    bool                IsWidthFixed() const            { return m_IsWidthFixed; }

    //
    //
    //
    //
    void                SetFixedWidth(f32 width)        { m_FixedWidth = width; }

    //
    //
    //
    //
    f32                 GetFixedWidth() const           { return m_FixedWidth; }

    //

    //
    //

    //
    //
    //
    //
    //
    //
    f32                 Print(CharCode code);

    //
    //
    //
    //
    void                DrawGlyph(const Glyph& glyph);

    //

    //
    //

    //
    //
    //
    //
    //
    void                SetCursor(
                            f32 x,
                            f32 y
    )
    {
        m_CursorPos.x = x;
        m_CursorPos.y = y;
    }

    //
    //
    //
    //
    //
    //
    void                SetCursor(
                            f32 x,
                            f32 y,
                            f32 z
    )
    {
        m_CursorPos.x = x;
        m_CursorPos.y = y;
        m_CursorPos.z = z;
    }

    //
    //
    //
    //
    //
    void                MoveCursor(
                            f32 dx,
                            f32 dy
    )
    {
        m_CursorPos.x += dx;
        m_CursorPos.y += dy;
    }

    //
    //
    //
    //
    //
    //
    void                MoveCursor(
                            f32 dx,
                            f32 dy,
                            f32 dz
    )
    {
        m_CursorPos.x += dx;
        m_CursorPos.y += dy;
        m_CursorPos.z += dz;
    }

    //
    //
    //
    //
    void                SetCursorX(f32 x)           { m_CursorPos.x = x; }

    //
    //
    //
    //
    void                SetCursorY(f32 y)           { m_CursorPos.y = y; }

    //
    //
    //
    //
    void                SetCursorZ(f32 z)           { m_CursorPos.z = z; }

    //
    //
    //
    //
    void                MoveCursorX(f32 dx)         { m_CursorPos.x += dx; }

    //
    //
    //
    //
    void                MoveCursorY(f32 dy)         { m_CursorPos.y += dy; }

    //
    //
    //
    //
    void                MoveCursorZ(f32 dz)         { m_CursorPos.z += dz; }

    //
    //
    //
    //
    f32                 GetCursorX() const          { return m_CursorPos.x; }

    //
    //
    //
    //
    f32                 GetCursorY() const          { return m_CursorPos.y; }

    //
    //
    //
    //
    f32                 GetCursorZ() const          { return m_CursorPos.z; }

    //

    //
    //

    //
    //
    //
    //
    TextWriterResource* GetTextWriterResource() const
    {
        return m_pTextWriterResource;
    }

    //
    //
    //
    //
    void                SetTextWriterResource(TextWriterResource* pTextWriterResource)
    {
        m_pTextWriterResource = pTextWriterResource;
    }

    //

    //
    //

    //
    //
    //
    //
    //
    //
    static u32          GetDispStringBufferSize(u32 charNum);

    //
    //
    //
    //
    //
    //
    //
    static DispStringBuffer*
                        InitDispStringBuffer(
                            void*   drawBuffer,
                            u32     charNum);

    //
    //
    //
    //
    DispStringBuffer*   GetDispStringBuffer() const
    {
        return m_pDispStringBuffer;
    }

    //
    //
    //
    //
    //
    void                SetDispStringBuffer(DispStringBuffer* buffer)
    {
        m_pDispStringBuffer = buffer;
    }

    //
    //
    //
    void                StartPrint();

    //
    //
    //
    void                EndPrint();

    //
    //
    //
    //
    //
    //
    //
    //
#if !defined(NW_FONT_ENABLE_EDGE_FONT)
    void                UseCommandBuffer( RectDrawer* pRectDrawer = NULL );
#else
    void                UseCommandBuffer( FontCombinerType combiner = FONT_COMBINER_TYPE_DEFAULT, RectDrawer* pRectDrawer = NULL );
#endif

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
#if !defined(NW_FONT_ENABLE_EDGE_FONT)
    u32*                UseCommandBuffer( u32* pCmdStream, RectDrawer* pRectDrawer = NULL );
#else
    u32*                UseCommandBuffer( u32* pCmdStream, FontCombinerType combiner = FONT_COMBINER_TYPE_DEFAULT, RectDrawer* pRectDrawer = NULL );
#endif

    //

    #if defined(NW_FONT_PROFILE)
        void                PrintProfile();
    #endif

protected:
    #if defined(NW_FONT_PROFILE)
        void                CopyProfileData(const CharWriter& other) const;
    #endif

private:

    /* ------------------------------------------------------------------------
            Types
       ------------------------------------------------------------------------ */
    typedef math::VEC2      CharScale;
    typedef math::VEC3      CursorPos;

    //

    struct ColorMapping
    {
        ut::Color8          min;
        ut::Color8          max;
    };

    //

    /* ------------------------------------------------------------------------
            Functions
       ------------------------------------------------------------------------ */
    //---- Prepare to draw

    //
    //
    //
    //
    void                SetupGXDefault(bool bAlphaTex = false);

    //
    //
    //
    //
    //
    void                SetupGXWithColorMapping(bool bAlphaTex = false);

    //
    void                SetupVertexFormat();

    //
    void                SetupGXCommon();

    //
    //
    //
    //
    //
    //
    void                PrintGlyph(
                            f32             x,
                            const Glyph&    glyph);

    //
    void                UpdateVertexColor();

    //
    //
    //
    //
    void                LoadTexture(const Glyph& glyph);

    /* ------------------------------------------------------------------------
            Variables
       ------------------------------------------------------------------------ */
    static const GLushort       s_VertexIndexs[];

    ColorMapping                m_ColorMapping;         //
    ut::Color8                  m_VertexColors[internal::TRIFAN_VTX_MAX];
                                                        //
    ut::Color8                  m_TextColors[internal::TEXTCOLOR_MAX];  //
    GradationMode               m_GradationMode;
    CharScale                   m_Scale;                //
    CursorPos                   m_CursorPos;            //
    f32                         m_FixedWidth;           //
    const Font*                 m_pFont;                //

    //
    TextWriterResource*         m_pTextWriterResource;

    //
    DispStringBuffer*           m_pDispStringBuffer;

    bool                        m_IsWidthFixed;         //
    u8                          m_Alpha;                //

    #if defined(NW_FONT_PROFILE)
        mutable Stopwatch           m_LoadTextureSw;
        mutable Stopwatch           m_PrintCharSw;
        mutable u32                 m_LoadTextureCount;
        mutable u32                 m_PrintCharCount;
        mutable Stopwatch           m_DispCharSw;
        mutable u32                 m_DispCharCount;
    #endif
};

}   // namespace font
}   // namespace nw
#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif
#endif
#endif //  NW_FONT_CHARWRITER_H_
