/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_FONT_DISPSTRINGBUFFER_H_
#define NW_FONT_DISPSTRINGBUFFER_H_

#if defined(_MSC_VER) && _MSC_VER >= 1500
#pragma once
#endif

#include <nw/font/font_Types.h>

namespace nw {
namespace font {

namespace internal {

struct CharAttribute
{
    math::VEC4          pos;
    ut::Color8          color[TEXTCOLOR_MAX];
    math::VEC4          tex;
    const internal::TextureObject*
                        pTexObj;        //
};

}   // namespace internal

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
struct DispStringBuffer
{
    //
    //
    enum CommandBufferFlag
    {
        // Generates command sequence statically.
        CBF_STATIC = 1 << 0,
        // Uses command jump.
        CBF_COMMAND_JUMP = 1 << 1,
        // Uses the VRAM command buffer.
        CBF_VRAM = 1 << 2,
        // Conducts double buffering.
        CBF_DOUBLE_BUFFER = 1 << 3,
        //
        CBF_DEFAULT = 0
    };

    //
    enum CommandBufferMode
    {
        //
        CBM_DYNAMIC = CBF_DEFAULT,
        //
        CBM_VRAM = CBF_STATIC | CBF_COMMAND_JUMP | CBF_VRAM,
        //
        CBM_DOUBLE_BUFFER = CBF_COMMAND_JUMP | CBF_DOUBLE_BUFFER,

        //
        CBM_NUM
    };

    const u32           charCountMax;           //

    u16                 charCount;              //
    u16                 drawCharCount;          //

    u8                  generatedCommand;       //
    u8                  commandBufferMode;      //
    u8                  padding[2];             //

    u8*                 drawFlags;              //
    u32*                commandBuffer;          //
    u32*                initialCommandBuffer;   //
    u32*                subCommandBuffer;       //

    u32                 commandBufferSize;      //
    u32                 commandBufferCapacity;  //
    u32                 textColorCommandOffset; //

    //
    //
    internal::CharAttribute*
                        GetCharAttrs() const
    {
        return reinterpret_cast<internal::CharAttribute*>(
            reinterpret_cast<uptr>(this) + sizeof(*this) );
    }

    static u32          CalcCommandBufferCapacity(u32 charNum);

    //
                        DispStringBuffer(u32 charNum);

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    void                SetCommandBufferMode(CommandBufferMode mode)
    {
        u8 newValue = static_cast<u8>(mode);

        // If the status changes, disables the command and regenerates it.
        if (commandBufferMode != newValue)
        {
            this->ClearCommand();
            commandBuffer = initialCommandBuffer;
        }

        commandBufferMode = newValue;
    }

    //
    //
    //
    //
    //
    //
    //
    CommandBufferMode GetCommandBufferMode() const
    {
        return static_cast<CommandBufferMode>(commandBufferMode);
    }

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    void*               SetVramCommandBuffer(void* pCmdBuffer)
    {
        void* prev = subCommandBuffer;
        subCommandBuffer = static_cast<u32*>(pCmdBuffer);
        return prev;
    }

    //
    //
    //
    //
    //
    //
    //
    void*               GetVramCommandBuffer() const
    {
        NW_ASSERT(this->GetCommandBufferMode() == CBM_VRAM);
        return subCommandBuffer;
    }

    //
    //
    //
    //
    //
    //
    //
    //
    void                SendToVram() const;

    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    //
    void                SetSecondCommandBuffer(void* pCmdBuffer)
    {
        subCommandBuffer =  static_cast<u32*>(pCmdBuffer);
    }

    //
    //
    //
    //
    //
    //
    //
    void*               GetSecondCommandBuffer() const
    {
        NW_ASSERT(this->GetCommandBufferMode() == CBM_DOUBLE_BUFFER);
        return subCommandBuffer;
    }

    //
    //
    //
    //
    //
    void                SwapCurrentCommandBuffer()
    {
        NW_ASSERT(this->GetCommandBufferMode() & CBF_DOUBLE_BUFFER);

        if (subCommandBuffer == NULL)
        {
            return;
        }

        if (commandBuffer == initialCommandBuffer)
        {
            commandBuffer = subCommandBuffer;
        }
        else
        {
            commandBuffer = initialCommandBuffer;
        }
    }

    //
    void                ClearCommand()
    {
        generatedCommand = false;
        commandBufferSize = 0;
    }

    //
    bool                IsGeneratedCommand()
    {
        return 0 != generatedCommand;
    }

    //
    bool                IsCommandEmpty()
    {
        return 0 == commandBufferSize;
    }

    //
    u16                 GetDrawCharCount() const    { return drawCharCount; }

    //
    void                SetDrawCharCount(u16 count)
    {
        drawCharCount = count;
    }

    //
    void                ResetDrawCharCount()
    {
        SetDrawCharCount(0xFFFF);
    }
};

}   // namespace font
}   // namespace nw

#endif //  NW_FONT_DISPSTRINGBUFFER_H_
