/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW_FONT_TAGPROCESSORBASE_H_
#define NW_FONT_TAGPROCESSORBASE_H_

#include <nn/types.h>
#include <nw/ut/ut_Rect.h>
#include <nw/font/font_Types.h>

namespace nw {
namespace font {



template <typename CharType>
class TextWriterBase;


template <typename CharType>
struct PrintContext
{
    TextWriterBase<CharType>*   writer;     //
    const CharType*             str;        //
    const f32                   xOrigin;    //
    const f32                   yOrigin;    //
    u32                         flags;      //

    PrintContext(
        TextWriterBase<CharType>*   aWriter,
        const CharType*             aStr,
        const f32                   aXOrigin,
        const f32                   aYOrigin,
        u32                         aFlags
    )
    :   writer(aWriter),
        str(aStr),
        xOrigin(aXOrigin),
        yOrigin(aYOrigin),
        flags(aFlags)
    {}
};

//---------------------------------------------------------------------------
//
//
//
//---------------------------------------------------------------------------
template <typename CharType>
class TagProcessorBase
{
public:
    /* ------------------------------------------------------------------------
            Types
       ------------------------------------------------------------------------ */

    //
    enum Operation
    {
        //
        //
        //
        OPERATION_DEFAULT,

        //
        //
        //
        //
        OPERATION_NO_CHAR_SPACE,

        //
        //
        //
        //
        OPERATION_CHAR_SPACE,

        //
        //
        //
        //
        OPERATION_NEXT_LINE,

        //
        //
        OPERATION_END_DRAW,

        //
        NUM_OF_OPERATION
    };

    /* ------------------------------------------------------------------------
            Functions
       ------------------------------------------------------------------------ */

    //
    //

    //
                        TagProcessorBase() {}


    //
    virtual             ~TagProcessorBase() {}


    //


    //
    //

    //
    //
    //
    //
    //
    //
    //
    virtual Operation   Process(
                            u16                     code,
                            PrintContext<CharType>* pContext
                        )
    {
        NN_ASSERT(code < ' ');
        NN_POINTER_ASSERT(pContext);

        switch (code)
        {
        case '\n':
            ProcessLinefeed(pContext);
            return OPERATION_NEXT_LINE;

        case '\t':
            ProcessTab(pContext);
            return OPERATION_NO_CHAR_SPACE;

        default:
            // nothing to do
            break;
        }

        return OPERATION_DEFAULT;
    }

    //
    //
    //
    //
    //
    //
    //
    //
    //
    virtual Operation   CalcRect(
                            ut::Rect*               pRect,
                            u16                     code,
                            PrintContext<CharType>* pContext
                        )
    {
#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
#endif
        NN_POINTER_ASSERT(pRect);
        NN_ASSERT(code < ' ');
        NN_POINTER_ASSERT(pContext);

        switch (code)
        {
        case '\n':
            {
                TextWriterBase<CharType>& writer = *pContext->writer;

                pRect->right    = writer.GetCursorX();
                pRect->top      = writer.GetCursorY();

                ProcessLinefeed(pContext);

                pRect->left     = writer.GetCursorX();

                // ProcessLinefeed is just the amount the cursour moves, and does not include the height of the next line. To include the height of the next line, add FontHeight.
                // 
                pRect->bottom   = writer.GetCursorY() + pContext->writer->GetFontHeight();

                pRect->Normalize();
            }
            return OPERATION_NEXT_LINE;

        case '\t':
            {
                TextWriterBase<CharType>& writer = *pContext->writer;

                pRect->left     = writer.GetCursorX();

                ProcessTab(pContext);

                pRect->right    = writer.GetCursorX();
                pRect->top      = writer.GetCursorY();
                pRect->bottom   = pRect->top + writer.GetFontHeight();

                pRect->Normalize();
            }
            return OPERATION_NO_CHAR_SPACE;

        default:
            // nothing to do
            break;
        }

        return OPERATION_DEFAULT;
#ifdef NW_PLATFORM_CTR
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif
#endif
    }

    //

private:
    /* ------------------------------------------------------------------------
            Types
       ------------------------------------------------------------------------ */
    typedef PrintContext<CharType>  ContextType;

    void                ProcessLinefeed(ContextType* pContext);
    void                ProcessTab(ContextType* pContext);
};

/* ------------------------------------------------------------------------
        private tag processing
   ------------------------------------------------------------------------ */
template <typename CharType>
void
TagProcessorBase<CharType>::ProcessLinefeed(ContextType* pContext)
{
    NN_POINTER_ASSERT(pContext);

    TextWriterBase<CharType>& writer = *pContext->writer;

    const f32 x   = pContext->xOrigin;
    const f32 y   = writer.GetCursorY() + writer.GetLineHeight();

    writer.SetCursor(x, y);
}

template <typename CharType>
void
TagProcessorBase<CharType>::ProcessTab(ContextType* pContext)
{
    NN_POINTER_ASSERT(pContext);

    TextWriterBase<CharType>& writer = *pContext->writer;
    const int tabWidth = writer.GetTabWidth();

    if (tabWidth > 0)
    {
        const f32 aCharWidth  = writer.IsWidthFixed() ? writer.GetFixedWidth(): writer.GetFontWidth();
        const f32 dx          = writer.GetCursorX() - pContext->xOrigin;
        const f32 tabPixel    = tabWidth * aCharWidth;
        const int numTab      = static_cast<int>(dx / tabPixel) + 1;
        const f32 x           = pContext->xOrigin + tabPixel * numTab;

        writer.SetCursorX(x);
    }
}

}   // namespace font
}   // namespace nw

#endif //  NW_FONT_TAGPROCESSORBASE_H_
