/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <nn/types.h>
#include <nn/assert.h>
#if defined(NW_PLATFORM_CTR)
    #include <nn/version.h>
    #include <nn/gx/CTR/gx_CommandAccess.h>
    #include <nn/gr/CTR/gr_Utility.h>
#endif
#include <nw/font/font_RectDrawer.h>
#include <nw/font/font_DispStringBuffer.h>

#define NW_FONT_ADD_COMMANDS( pStrBuf, command, size )                                                  \
    NW_ASSERT((pStrBuf)->commandBufferSize + (size) / sizeof(u32) <= (pStrBuf)->commandBufferCapacity); \
    std::memcpy(&(pStrBuf)->commandBuffer[(pStrBuf)->commandBufferSize], command, size );               \
    (pStrBuf)->commandBufferSize += (size) / sizeof(u32)

#define NW_FONT_ADD_SINGLE_COMMAND( pStrBuf, header, data )                             \
    NW_ASSERT((pStrBuf)->commandBufferSize + 2 <= (pStrBuf)->commandBufferCapacity);    \
    (pStrBuf)->commandBuffer[(pStrBuf)->commandBufferSize + 0] = (data);                \
    (pStrBuf)->commandBuffer[(pStrBuf)->commandBufferSize + 1] = (header);              \
    (pStrBuf)->commandBufferSize += 2

// Set the draw function.
// 0x253 [0:0]
#define NW_FONT_CMD_SET_DRAW_FUNC( func )                   \
    PICA_CMD_DATA_DRAW_MODE1( func, 0 /* mode */ ),         \
    PICA_CMD_HEADER_SINGLE_BE( PICA_REG_DRAW_MODE1, 0x1 )

// Draw mode
// 0x253 [0:8]
#define NW_FONT_CMD_SET_DRAW_MODE1( mode )                  \
    PICA_CMD_DATA_DRAW_MODE1( 0 /* func */, mode ),         \
    PICA_CMD_HEADER_SINGLE_BE( PICA_REG_DRAW_MODE1, 0x2 )

// Load array N elements 8-11: number of bytes, number of elements
#define NW_FONT_CMD_DATA_LOAD_ARRAY_ELEMENT8( elm8, elm9, elm10, elm11, bytes, num )    \
      ((elm8)  & 0x0F) <<  0 | ((elm9)  & 0x0F) <<  4                                   \
    | ((elm10) & 0x0F) <<  8 | ((elm11) & 0x0F) << 12                                   \
    | ((bytes) & 0xFF) << 16 | ((num)   & 0x0F) << 28

// Load array N address offset: element, number of bytes, number of elements
#define NW_FONT_CMD_SET_LOAD_ARRAY( offset, elm0, elm1, elm2, elm3, elm4, elm5, elm6,       \
    elm7, elm8, elm9, elm10, elm11, bytes, num )                                            \
    (offset) & 0x0FFFFFFF,  /* Address offset */                                        \
    PICA_CMD_DATA_LOAD_ARRAY0_ELEMENT( elm0, elm1, elm2, elm3, elm4, elm5, elm6, elm7 ),    \
    NW_FONT_CMD_DATA_LOAD_ARRAY_ELEMENT8( elm8, elm9, elm10, elm11, bytes, num )

#define NW_FONT_CMD_SET_LOAD_ARRAY_ZERO()                                           \
    NW_FONT_CMD_SET_LOAD_ARRAY(                                                     \
        0, /* offset */                                                             \
        PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR,  \
        PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR,  \
        PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR,  \
        PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR, PICA_DATA_VERTEX_0_ATTR,  \
        0,  /* bytes */                                                             \
        0)  /* num */

#define NW_FONT_CMD_SET_TEXTURE_LOD_DISABLE( addr )                                               \
    PICA_CMD_DATA_TEXTURE_LOD_LEVEL( 0 /* lodBias */, 0 /* maxLodLevel */, 0 /* minLodLevel */ ), \
    PICA_CMD_HEADER_SINGLE( addr )

// Clears the internal state of the vertex index.
#define NW_FONT_CMD_SET_VERTEX_INDEX_CLEAR()                \
    /* 0x25f */                                             \
    PICA_CMD_DATA_START_DRAW_FUNC1( true ),                 \
    PICA_CMD_HEADER_SINGLE( PICA_REG_START_DRAW_FUNC1 )

#if defined(NW_PLATFORM_CTR)
extern unsigned* __cb_current_command_buffer;
#endif

namespace nw {
namespace font {
namespace {

#if defined(NW_PLATFORM_CTR)

const u8    MATRIX_UNIFORM_NUM      = 3;    // Number of uniforms the array uses.
const u8    COLOR_USE_COUNT         = 4;    // Number of uniforms used for each character color.

const u32   TEX_FUNC_DATA =
    PICA_CMD_DATA_TEXTURE_FUNC(
        PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_2D /* texture0SamplerType */,
        0 /* texture1SamplerType */,
        0 /* texture2SamplerType */,
        0 /* texture3Texcoord */,
        0 /* texture3SamplerType */,
        0 /* texture2Texcoord */,
        1 /* clearTextureCache */);

const u32   TEX_ENV_SRC3 =
    PICA_CMD_DATA_TEX_ENV_SRC(
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT );

const u32   TEX_ENV_SRC4 =
    PICA_CMD_DATA_TEX_ENV_SRC(
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS );

const u32   TEX_ENV_OPERAND3 =
    PICA_CMD_DATA_TEX_ENV_OPERAND(
        PICA_DATA_OPE_RGB_SRC_COLOR,            PICA_DATA_OPE_RGB_SRC_COLOR,    PICA_DATA_OPE_RGB_SRC_COLOR,
        PICA_DATA_OPE_ALPHA_SRC_ALPHA,          PICA_DATA_OPE_ALPHA_SRC_ALPHA,  PICA_DATA_OPE_ALPHA_SRC_ALPHA );

const u32   TEX_ENV_OPERAND4 =
    PICA_CMD_DATA_TEX_ENV_OPERAND(
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_COLOR,  PICA_DATA_OPE_RGB_SRC_COLOR,    PICA_DATA_OPE_RGB_SRC_COLOR,
        PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_ALPHA,PICA_DATA_OPE_ALPHA_SRC_ALPHA,  PICA_DATA_OPE_ALPHA_SRC_ALPHA );

const u32   TEX_ENV_OPERAND3_ALPHA =
    PICA_CMD_DATA_TEX_ENV_OPERAND(
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_COLOR,  PICA_DATA_OPE_RGB_SRC_COLOR,    PICA_DATA_OPE_RGB_SRC_COLOR,
        PICA_DATA_OPE_ALPHA_SRC_ALPHA,          PICA_DATA_OPE_ALPHA_SRC_ALPHA,  PICA_DATA_OPE_ALPHA_SRC_ALPHA );

const u32   TEX_ENV_OPERAND4_ALPHA =
    PICA_CMD_DATA_TEX_ENV_OPERAND(
        PICA_DATA_OPE_RGB_SRC_COLOR,            PICA_DATA_OPE_RGB_SRC_COLOR,    PICA_DATA_OPE_RGB_SRC_COLOR,
        PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_ALPHA,PICA_DATA_OPE_ALPHA_SRC_ALPHA,  PICA_DATA_OPE_ALPHA_SRC_ALPHA );

const u32   TEX_ENV_COMBINE_MODULATE =
    PICA_CMD_DATA_TEX_ENV_COMBINE(
        PICA_DATA_TEX_ENV_COMBINE_MODULATE,
        PICA_DATA_TEX_ENV_COMBINE_MODULATE );

const u32   TEX_ENV_COMBINE_MULT_ADD =
    PICA_CMD_DATA_TEX_ENV_COMBINE(
        PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP,
        PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP );

const u32   TEX_ENV_CONST_COLOR_WHITE   = 0xffffffff;
const u32   TEX_ENV_CONST_COLOR_BLACK   = 0x00000000;

const u32   TEX_ENV_SCALE_1 =
    PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 );


// Maximum size during burst access.
const u32   BURST_MAX = 0x80;

const u32   DRAW_MODE0_DEFAULT  = 0;    // When glDrawElements and the mode is GL_TRIANGLES, set to 1, in all other cases, set to 0
                                        // 

#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
const u32   DRAW_MODE2_DEFAULT  = 0;
const u32   DRAW_FUNC_DEFAULT   = 0;
#else
const u32   DRAW_MODE0_ELEMENT  = 1;
const u32   DRAW_MODE2_ELEMENT  = PICA_DATA_DRAW_TRIANGLES;
#endif

struct ShaderBinaryFileHeader
{
    u8      signature[4];
    u32     exeCount;
    u32     exeOffsetTop;
};

struct ShaderPackageHeader
{
    u8      signature[4];
    u8      version[2];
    u16     reserved0;
    u32     instOffset;    // Program code information offset
    u32     instCount;     // Number of program code data (*4 to get number of bytes)
    u32     swizzleOffset;
    u32     swizzleCount;
    u32     lineOffset;
    u32     lineCount;
    u32     stringOffset;
    u32     stringSize;
};

struct SwizzleInfo
{
    u32     value;
    u16     usedInfo;
    u16     reserve;
};

// Execution image information header
struct ExeImageHeader
{
    u8      signature[4];

    u8      version[2];
    u8      shaderType;
    u8      mergeOutputMapsDebug;

    u32     mainAddr;
    u32     endAddr;

    u16     inputMask;
    u16     outputMask;

    u8      geometryDataMode;
    u8      startIndex;
    u8      subdPatchSize;
    u8      constVertexNumber;

    u32     setupOffset;
    u32     setupCount;
    u32     labelOffset;
    u32     labelCount;
    u32     outmapOffset;
    u32     outmapCount;
    u32     bsymOffset;
    u32     bsymCount;
    u32     stringOffset;
    u32     stringSize;
};

struct SetupInfo
{
    u16     type;
    u16     index;
    u32     value[4];
};

struct OutmapInfo
{
    u16     type;
    u16     index;
    u16     mask;
    u16     reserve;
};

enum
{
    SHADEROUTMAPTYPE_POSITION,
    SHADEROUTMAPTYPE_QUATERNION,
    SHADEROUTMAPTYPE_COLOR,
    SHADEROUTMAPTYPE_TEXCOORD0,
    SHADEROUTMAPTYPE_TEXCOORD0W,
    SHADEROUTMAPTYPE_TEXCOORD1,
    SHADEROUTMAPTYPE_TEXCOORD2,
    SHADEROUTMAPTYPE_VIEW,

    SHADEROUTMAPTYPE_MAX
};

// Dummy command required once before setting 0x080 to 0x0bf.
const u32   TextureDummyCommands[] =
{
    0, PICA_CMD_HEADER_BURST_BE(PICA_REG_TEXTURE_FUNC, 3, 0x0), 0, 0,
};

const u32   TextInitCommands[] =
{
    // Sets the number of texture coordinates in the vertex shader.
    // 0x2b1
    PICA_CMD_DATA_VS_INT( 0x01, 0, 0 ),
    PICA_CMD_HEADER_SINGLE( PICA_REG_VS_INT0 ),

    // Clear the color depth buffer (required once before setting 0x100 to 0x130).
    // 0x111
    PICA_CMD_DATA_COLOR_DEPTH_BUFFER_CLEAR(true),
    PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),

    // Default blend setting (disables logical operations)
    // 0x100, 0x101
    NW_FONT_COMMAND_SET_BLEND_DEFAULT,

    // Enables alpha testing (GREATER 0)
    // 0x104
    PICA_CMD_DATA_FRAGOP_ALPHA_TEST( true, PICA_DATA_ALPHA_TEST_GREATER, 0 ),
    PICA_CMD_HEADER_SINGLE( PICA_REG_FRAGOP_ALPHA_TEST ),

    // Final stage texture combiner settings
    PICA_CMD_DATA_TEX_ENV_SRC(
        PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
        PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
    PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV5, PICA_REG_TEX_ENV5_SCALE - PICA_REG_TEX_ENV5 + 1 ),
    PICA_CMD_DATA_TEX_ENV_OPERAND(
        PICA_DATA_OPE_RGB_SRC_COLOR,   PICA_DATA_OPE_RGB_SRC_COLOR,   PICA_DATA_OPE_RGB_SRC_COLOR,
        PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
    TEX_ENV_COMBINE_MODULATE,
    TEX_ENV_CONST_COLOR_BLACK,
    TEX_ENV_SCALE_1,

    // Shader output settings
    // Outputs texture coordinate 0 only.
    // 0x06f
    PICA_CMD_DATA_VS_GS_OUT_ATTR_CLK(
        false,  // posZ
        false,  // col
        true,   // tex0
        false,  // tex1
        false,  // tex2
        false,  // tex0_w
        false), // view_quart
    PICA_CMD_HEADER_SINGLE_BE( PICA_REG_VS_OUT_ATTR_CLK, 0x2 ),

    // Dummy command required once before setting 0x080 to 0x0bf.
    // 0x080
    0, PICA_CMD_HEADER_BURST_BE(PICA_REG_TEXTURE_FUNC, 3, 0x0), 0, 0,

    // Texture unit settings
    // Select sampler type, coordinates
    // 0x080
    TEX_FUNC_DATA,
    PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEXTURE_FUNC, 0xB ),
};


#if !defined(NW_FONT_ENABLE_EDGE_FONT)

const u32   TexEnvOperand34Commands[] =
{
    //      First stage operand
    TEX_ENV_OPERAND3,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

    //      Second stage operand
    TEX_ENV_OPERAND4,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
};

const u32   TexEnvOperand34AlphaCommands[] =
{
    //      First stage operand
    TEX_ENV_OPERAND3_ALPHA,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

    //      Second stage operand
    TEX_ENV_OPERAND4_ALPHA,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
};

const u32   TexEnv34Commands[] =
{
    TEX_ENV_SRC3,
    PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_COMBINE - PICA_REG_TEX_ENV3 + 1 ),
    TEX_ENV_OPERAND3,
    TEX_ENV_COMBINE_MODULATE,

    TEX_ENV_SCALE_1,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_SCALE ),

    TEX_ENV_SRC4,
    PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
    TEX_ENV_OPERAND4,
    TEX_ENV_COMBINE_MULT_ADD,

    TEX_ENV_SCALE_1,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
};

const u32   TexEnv34AlphaCommands[] =
{
    TEX_ENV_SRC3,
    PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_COMBINE - PICA_REG_TEX_ENV3 + 1 ),
    TEX_ENV_OPERAND3_ALPHA,
    TEX_ENV_COMBINE_MODULATE,

    TEX_ENV_SCALE_1,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_SCALE ),

    TEX_ENV_SRC4,
    PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
    TEX_ENV_OPERAND4_ALPHA,
    TEX_ENV_COMBINE_MULT_ADD,

    TEX_ENV_SCALE_1,
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
};

#else

    //-----------------------------------
    // Various commands
    //-----------------------------------

    const u32   TEX_ENV_EDGE_SRC_3 =
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,        // Texture alpha
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,        // 0
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,        // Unused
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,        // Texture alpha
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,        // Unused
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT );      // Unused

    const u32   TEX_ENV_NO_EDGE_SRC_3 =
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,        // Txture color (basic fixed a 1)
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,        // Unused
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,        // Unused
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,        // Texture alpha
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,        // 0
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT );      // Unused

    const u32   TEX_ENV_COMMON_SRC_4 =
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,               // White interpolated color
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,    // Black interpolated color
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,               // Prior stage output color
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,               // White interpolated alpha
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,    // Black interpolated alpha
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS );             // Prior stage output color

    const u32   TEX_ENV_EDGE_OPERAND_3 =
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_ALPHA,      PICA_DATA_OPE_RGB_SRC_ALPHA,    0,        // Sets unused section to 0.
            PICA_DATA_OPE_ALPHA_SRC_ALPHA,    0,                              0 );

    // TEX_ENV_OPERAND3_ALPHA inverts texture color only.  =>  Does not use color, so TEX_ENV_EDGE_OPERAND_3_ALPHA should be unnecessary.

    const u32   TEX_ENV_NO_EDGE_OPERAND_3 =
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR,      0,                                0,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA,    PICA_DATA_OPE_ALPHA_SRC_ALPHA,    0 );

    const u32   TEX_ENV_NO_EDGE_OPERAND_3_ALPHA =
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_ONE_MINUS_SRC_COLOR,    0,                                0,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA,            PICA_DATA_OPE_ALPHA_SRC_ALPHA,    0 );

    const u32   TEX_ENV_COMMON_OPERAND_4 =
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR,      PICA_DATA_OPE_RGB_SRC_COLOR,      PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA,    PICA_DATA_OPE_ALPHA_SRC_ALPHA,    PICA_DATA_OPE_ALPHA_SRC_ALPHA );

    const u32   TEX_ENV_COMBINE_EDGE_3 =
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_ADD_SIGNED,      // C3 = 2 * ( Ta + 0.0 - 0.5 )
            PICA_DATA_TEX_ENV_COMBINE_REPLACE );       // A3 = 2 * Ta

    const u32   TEX_ENV_COMBINE_NO_EDGE_3 =
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_REPLACE,         // C3 = Tc
            PICA_DATA_TEX_ENV_COMBINE_ADD_SIGNED );    // A3 = 2 * ( Ta + 0.0 - 0.5 )

    const u32   TEX_ENV_COMBINE_INTERPOLATE =
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE,     // C4 = C3 * Wc + ( 1.0 - C3 ) * Bc
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE );   // A4 = C3 * Wa + ( 1.0 - A3 ) * Ba

    const u32   TEX_ENV_SCALE_2_2 =
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_2, PICA_DATA_TEX_ENV_SCALE_2 );

    const u32   TEX_ENV_SCALE_1_2 =
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_2 );


    //-----------------------------------
    // Operand
    //-----------------------------------

    // Regular font
    const u32   TexEnvOperand34Commands_Default[] =
    {
        // First stage
        TEX_ENV_OPERAND3,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

        // Second stage
        TEX_ENV_OPERAND4,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
    };

    // Border fonts with borders
    const u32   TexEnvOperand34Commands_Edge[] =
    {
        TEX_ENV_EDGE_OPERAND_3,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

        TEX_ENV_COMMON_OPERAND_4,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
    };

    // Border fonts without borders
    const u32   TexEnvOperand34Commands_NoEdge[] =
    {
        TEX_ENV_NO_EDGE_OPERAND_3,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

        TEX_ENV_COMMON_OPERAND_4,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
    };

    const u32* TexEnvOperand34Commands[ FONT_COMBINER_TYPE_MAX ] =
    {
        TexEnvOperand34Commands_Default,
        TexEnvOperand34Commands_Edge,
        TexEnvOperand34Commands_NoEdge
    };

    const u32 TexEnvOperand34CommandsSize[ FONT_COMBINER_TYPE_MAX ] =
    {
        sizeof( TexEnvOperand34Commands_Default ),
        sizeof( TexEnvOperand34Commands_Edge ),
        sizeof( TexEnvOperand34Commands_NoEdge )
    };


    //-----------------------------------
    // Operand (for A4 and A8 textures)
    //-----------------------------------
    const u32   TexEnvOperand34AlphaCommands_Default[] =
    {
        TEX_ENV_OPERAND3_ALPHA,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

        TEX_ENV_OPERAND4_ALPHA,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
    };

    const u32   TexEnvOperand34AlphaCommands_Edge[] =
    {
        TEX_ENV_EDGE_OPERAND_3,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

        TEX_ENV_COMMON_OPERAND_4,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
    };

    const u32   TexEnvOperand34AlphaCommands_NoEdge[] =
    {
        TEX_ENV_NO_EDGE_OPERAND_3_ALPHA,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_OPERAND ),

        TEX_ENV_COMMON_OPERAND_4,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_OPERAND ),
    };

    const u32*   TexEnvOperand34AlphaCommands[ FONT_COMBINER_TYPE_MAX ] =
    {
        TexEnvOperand34AlphaCommands_Default,
        TexEnvOperand34AlphaCommands_Edge,
        TexEnvOperand34AlphaCommands_NoEdge
    };

    const u32   TexEnvOperand34AlphaCommandsSize[ FONT_COMBINER_TYPE_MAX ] =
    {
        sizeof( TexEnvOperand34AlphaCommands_Default ),
        sizeof( TexEnvOperand34AlphaCommands_Edge ),
        sizeof( TexEnvOperand34AlphaCommands_NoEdge )
    };


    //-----------------------------------
    // Combiner style (when changing textures)
    //-----------------------------------
    const u32   TexEnv34Commands_Default[] =
    {
        TEX_ENV_SRC3,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_COMBINE - PICA_REG_TEX_ENV3 + 1 ),
        TEX_ENV_OPERAND3,
        TEX_ENV_COMBINE_MODULATE,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_SCALE ),

        TEX_ENV_SRC4,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
        TEX_ENV_OPERAND4,
        TEX_ENV_COMBINE_MULT_ADD,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
    };

    const u32   TexEnv34Commands_Edge[] =
    {
        TEX_ENV_EDGE_SRC_3,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_SCALE - PICA_REG_TEX_ENV3 + 1 ),
        TEX_ENV_EDGE_OPERAND_3,
        TEX_ENV_COMBINE_EDGE_3,
        TEX_ENV_CONST_COLOR_BLACK,
        TEX_ENV_SCALE_2_2,

        TEX_ENV_COMMON_SRC_4,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
        TEX_ENV_COMMON_OPERAND_4,
        TEX_ENV_COMBINE_INTERPOLATE,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
    };

    const u32   TexEnv34Commands_NoEdge[] =
    {
        TEX_ENV_NO_EDGE_SRC_3,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_SCALE - PICA_REG_TEX_ENV3 + 1 ),
        TEX_ENV_NO_EDGE_OPERAND_3,
        TEX_ENV_COMBINE_NO_EDGE_3,
        TEX_ENV_CONST_COLOR_BLACK,
        TEX_ENV_SCALE_1_2,

        TEX_ENV_COMMON_SRC_4,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
        TEX_ENV_COMMON_OPERAND_4,
        TEX_ENV_COMBINE_INTERPOLATE,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
    };

    const u32*   TexEnv34Commands[ FONT_COMBINER_TYPE_MAX ] =
    {
        TexEnv34Commands_Default,
        TexEnv34Commands_Edge,
        TexEnv34Commands_NoEdge
    };

    const u32   TexEnv34CommandsSize[ FONT_COMBINER_TYPE_MAX ] =
    {
        sizeof( TexEnv34Commands_Default ),
        sizeof( TexEnv34Commands_Edge ),
        sizeof( TexEnv34Commands_NoEdge )
    };


    //-----------------------------------
    // Combiner style for A4 and A8 textures (when changing textures)
    //-----------------------------------
    const u32   TexEnv34AlphaCommands_Default[] =
    {
        TEX_ENV_SRC3,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_COMBINE - PICA_REG_TEX_ENV3 + 1 ),
        TEX_ENV_OPERAND3_ALPHA,
        TEX_ENV_COMBINE_MODULATE,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3_SCALE ),

        TEX_ENV_SRC4,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
        TEX_ENV_OPERAND4_ALPHA,
        TEX_ENV_COMBINE_MULT_ADD,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
    };

    const u32   TexEnv34AlphaCommands_Edge[] =
    {
        TEX_ENV_EDGE_SRC_3,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_SCALE - PICA_REG_TEX_ENV3 + 1 ),
        TEX_ENV_EDGE_OPERAND_3,
        TEX_ENV_COMBINE_EDGE_3,
        TEX_ENV_CONST_COLOR_BLACK,
        TEX_ENV_SCALE_2_2,

        TEX_ENV_COMMON_SRC_4,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
        TEX_ENV_COMMON_OPERAND_4,
        TEX_ENV_COMBINE_INTERPOLATE,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
    };

    const u32   TexEnv34AlphaCommands_NoEdge[] =
    {
        TEX_ENV_NO_EDGE_SRC_3,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, PICA_REG_TEX_ENV3_SCALE - PICA_REG_TEX_ENV3 + 1 ),
        TEX_ENV_NO_EDGE_OPERAND_3_ALPHA,
        TEX_ENV_COMBINE_NO_EDGE_3,
        TEX_ENV_CONST_COLOR_BLACK,
        TEX_ENV_SCALE_1_2,

        TEX_ENV_COMMON_SRC_4,
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, PICA_REG_TEX_ENV4_COMBINE - PICA_REG_TEX_ENV4 + 1 ),
        TEX_ENV_COMMON_OPERAND_4,
        TEX_ENV_COMBINE_INTERPOLATE,

        TEX_ENV_SCALE_1,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4_SCALE ),
    };

    const u32*   TexEnv34AlphaCommands[ FONT_COMBINER_TYPE_MAX ] =
    {
        TexEnv34AlphaCommands_Default,
        TexEnv34AlphaCommands_Edge,
        TexEnv34AlphaCommands_NoEdge
    };

    const u32   TexEnv34AlphaCommandsSize[ FONT_COMBINER_TYPE_MAX ] =
    {
        sizeof( TexEnv34AlphaCommands_Default ),
        sizeof( TexEnv34AlphaCommands_Edge ),
        sizeof( TexEnv34AlphaCommands_NoEdge )
    };


#endif


const u32   s_DrawInitCommands[] =
{
    // 0x111, 0x110
    PICA_CMD_SET_COLOR_DEPTH_BUFFER_CLEAR( 1 /* data1 */, 1 /* data2 */ ),

    // Issued by the DisableTextures function called at initialization so here the 0x080 x 3 dummy commands are not needed.
    // 
    // ...

    // Default blend settings
    // 0x100, 0x101
    NW_FONT_COMMAND_SET_BLEND_DEFAULT,

    // Enables alpha testing (GREATER 0)
    // 0x104
    PICA_CMD_DATA_FRAGOP_ALPHA_TEST( true, PICA_DATA_ALPHA_TEST_GREATER, 0 ),
    PICA_CMD_HEADER_SINGLE( PICA_REG_FRAGOP_ALPHA_TEST ),

    // Disables fog.
    // 0x0e0 [2:0] [3:3] [16:16] [25:24]
    // Sets the TexEnv buffer input in PREV_BUFFER.
    // 0x0e0 [15:8]
    PICA_CMD_DATA_GAS_FOG_MODE(
        PICA_DATA_FOG_FALSE,
        0 /* dmp_Gas.shadingDensitySrc */,
        0 /* dmp_Fog.zFlip */ ),
    PICA_CMD_HEADER_SINGLE( PICA_REG_GAS_FOG_MODE ),

    // Disables fragment writing.
    // 0x08f, 0x1c6
    PICA_CMD_SET_FRAG_LIGHT_ENABLE( false ),

    // Color Buffer Read
    // 0x112
    0x0000000f,
    PICA_CMD_HEADER_SINGLE( PICA_REG_COLOR_BUFFER_READ ),

    // Disables texture LOD.
    // 0x084
    NW_FONT_CMD_SET_TEXTURE_LOD_DISABLE( PICA_REG_TEXTURE0_LOD ),
    // 0x094
    NW_FONT_CMD_SET_TEXTURE_LOD_DISABLE( PICA_REG_TEXTURE1_LOD ),
    // 0x09c
    NW_FONT_CMD_SET_TEXTURE_LOD_DISABLE( PICA_REG_TEXTURE2_LOD ),

    // Disables shadow textures.
    // 0x08b
    PICA_CMD_DATA_TEXTURE_SHADOW(
        false,  /* perspectiveShadow */
        0,      /* shadowZBias */
        0),     /* shadowZScale */
    PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE0_SHADOW ),

#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
// Settings for glDrawArrays.
// Set the draw mode and other settings to the defaults, just in case.

    // Vertex start offset (for glDrawArrays, the argument first value)
    // 0x22a
    PICA_CMD_DATA_DRAW_VERTEX_OFFSET(0),
    PICA_CMD_HEADER_SINGLE( PICA_REG_DRAW_VERTEX_OFFSET ),

    // Draw mode
    // 0x229 [8:8]
    PICA_CMD_SET_DRAW_MODE0( DRAW_MODE0_DEFAULT ),

    // Specify draw function, draw mode.
    // 0x253 [0:0],[8:8]
    PICA_CMD_SET_DRAW_MODE1( DRAW_FUNC_DEFAULT, DRAW_MODE0_DEFAULT ),

    // Draw mode
    // 0x25e [9:8]
    PICA_CMD_SET_DRAW_MODE2( DRAW_MODE2_DEFAULT ),

    // Clears the internal state of the vertex index.
    // Only call the first time when glDrawArrays is set to GL_TRIANGLES mode and the number of vertices is a multiple of three.
    // 
    // 0x25f
    NW_FONT_CMD_SET_VERTEX_INDEX_CLEAR(),
#else
// Settings for glDrawElements GL_TRIANGLES.

    // Draw mode
    // 0x229 [8:8]
    PICA_CMD_SET_DRAW_MODE0( DRAW_MODE0_ELEMENT ),

    // Specify draw function, draw mode.
    // 0x253 [0:0],[8:8]
    PICA_CMD_SET_DRAW_MODE1( DRAW_FUNC_DEFAULT, DRAW_MODE0_ELEMENT ),

    // Draw mode
    // 0x25e [9:8]
    PICA_CMD_SET_DRAW_MODE2( DRAW_MODE2_ELEMENT ),
#endif
};

const u32   s_DrawFinalizeCommands[] =
{
    // 0x232-0x235 fixed vertex attribute values
    PICA_CMD_SET_VS_FIXED_ATTR(
        0x0,            /* order */
        0x3f000000,     /* c0 */
        0x00000000,     /* c1 */
        0x00000000),    /* c2 */

#if ! defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
// Returns to default for the rendering mode, etc.
    // Draw mode
    // 0x229 [8:8]
    PICA_CMD_SET_DRAW_MODE0( DRAW_MODE0_DEFAULT ),

    // Draw mode
    // 0x253 [0:8]
    NW_FONT_CMD_SET_DRAW_MODE1( DRAW_MODE0_DEFAULT ),

    // Draw mode
    // 0x25e [9:8]
    PICA_CMD_SET_DRAW_MODE2( DRAW_MODE2_DEFAULT ),
#endif

    // 0x111, 0x110
    PICA_CMD_SET_COLOR_DEPTH_BUFFER_CLEAR( 1 /* data1 */, 1 /* data2 */ ),
};

const u32   DrawCommands[] =
{
#if !defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    // Clears the internal state of the vertex index.
    // Required when in the GL_TRIANGLES mode of glDrawElements.
    // 0x25f
    NW_FONT_CMD_SET_VERTEX_INDEX_CLEAR(),
#endif

#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    // 0x253
    NW_FONT_CMD_SET_DRAW_FUNC( 1 /* func */ ),

    PICA_CMD_SET_START_DRAW_ARRAY( 0x00000001 /* start */ ),

    // 0x253
    NW_FONT_CMD_SET_DRAW_FUNC( 0 /* func */ ),
#else
    PICA_CMD_SET_START_DRAW_ELEMENT( 0x00000001 /* start */ ),
#endif
};

const u32   DRAWBEGIN_STATICCOMMAND_BYTES   =
    158 * 4 + sizeof(s_DrawInitCommands)
#if !defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    + sizeof(u32) * 2 * 1
#endif
    ;

const u32   DRAWEND_STATICCOMMAND_BYTES     = 176 + sizeof(s_DrawFinalizeCommands);

#endif

u32
GetColorU32Raw(ut::Color8 color)
{
    return *reinterpret_cast<const u32*>(&color);
}

void
NormalizeF32Colors(
    math::VEC4* __restrict          dst,
    const ut::Color8* __restrict    src
)
{
    register const f32 d = 1.f / 255.f;
    register f32 colors[4];

    for (int i = 0; i < internal::TEXTCOLOR_MAX; ++i)
    {
        colors[0] = src[i].r * d;
        colors[1] = src[i].g * d;
        colors[2] = src[i].b * d;
        colors[3] = src[i].a * d;

        dst[i].x = colors[0];
        dst[i].y = colors[1];
        dst[i].z = colors[2];
        dst[i].w = colors[3];
    }
}

void
SetVertexColor(
    math::VEC4* __restrict          dst,
    const ut::Color8* __restrict    src,
    int                             mode
)
{
    math::VEC4 wkCols[internal::TEXTCOLOR_MAX];
    NormalizeF32Colors(wkCols, src);
    dst[VERTEX_LT] = wkCols[internal::TEXTCOLOR_START];
    dst[VERTEX_RT] = wkCols[
        mode != CharWriter::GRADMODE_H ? internal::TEXTCOLOR_START: internal::TEXTCOLOR_END];
    dst[VERTEX_LB] = wkCols[
        mode != CharWriter::GRADMODE_V ? internal::TEXTCOLOR_START: internal::TEXTCOLOR_END];
    dst[VERTEX_RB] = wkCols[
        mode == CharWriter::GRADMODE_NONE ? internal::TEXTCOLOR_START: internal::TEXTCOLOR_END];
}

inline
bool
IsDrawChar(
    const u8*   drawFlags,
    int         index
)
{
    return 0 != ut::internal::ReadBits(drawFlags[index / 8], 1, index % 8);
}

inline
void
SetDrawChar(
    u8*         drawFlags,
    int         index
)
{
    drawFlags[index / 8] = ut::internal::WriteBits(drawFlags[index / 8], 1, 1, index % 8);
}

#if defined(NW_PLATFORM_CTR)
const ShaderBinaryFileHeader&
GetShaderBinaryFileHeader(const void* shaderBinary)
{
    return *static_cast<const ShaderBinaryFileHeader*>(shaderBinary);
}
void
EnsureShaderBinaryFileHeader(const void* shaderBinary)
{
    const ShaderBinaryFileHeader& header =
        *static_cast<const ShaderBinaryFileHeader*>(shaderBinary);
    NN_ASSERT(  header.signature[0] == 'D' && header.signature[1] == 'V'
             && header.signature[2] == 'L' && header.signature[3] == 'B' );
}

const ShaderPackageHeader&
GetShaderPackageHeader(const void* shaderBinary)
{
    const ShaderBinaryFileHeader& fileHead = GetShaderBinaryFileHeader(shaderBinary);
    return *static_cast<const ShaderPackageHeader*>(
            ut::AddOffsetToPtr(
                shaderBinary,
                sizeof(fileHead) + (fileHead.exeCount - 1) * sizeof(u32) ) );
}

void
EnsureShaderPackageHeader(const ShaderPackageHeader& header)
{
    (void)header;
    NN_ASSERT(  header.signature[0] == 'D' && header.signature[1] == 'V'
             && header.signature[2] == 'L' && header.signature[3] == 'P' );
}

const ExeImageHeader*
GetExeImageHeader(const void* shaderBinary)
{
    const ShaderBinaryFileHeader& fileHead = GetShaderBinaryFileHeader(shaderBinary);
    return static_cast<const ExeImageHeader*>(
            ut::AddOffsetToPtr(
                shaderBinary,
                fileHead.exeOffsetTop ) );
}

void
EnsureExeImageHeader(const ExeImageHeader& header)
{
    (void)header;
    NN_ASSERT(  header.signature[0] == 'D' && header.signature[1] == 'V'
             && header.signature[2] == 'L' && header.signature[3] == 'E' );
}

u32
CalcInstCommandCount(u32 count)
{
    u32 size = (count / BURST_MAX) * math::RoundUp(2 + 1 + BURST_MAX, 2);
    if (0 != count % BURST_MAX)
    {
        size += math::RoundUp(2 + 1 + (count % BURST_MAX), 2);
    }
    return size;
}

u32
CalcDrawBeginCommandBytes(const ShaderPackageHeader& pkgHead, const ExeImageHeader& exeHead)
{
    const u32 instBytes =
          sizeof(u32)
        * ( CalcInstCommandCount(pkgHead.instCount) +
            CalcInstCommandCount(pkgHead.swizzleCount) +
            6 * exeHead.setupCount );
    // NN_LOG("inst swizzle bytes %d\n", instBytes);
    return DRAWBEGIN_STATICCOMMAND_BYTES + instBytes;
}

// Adds commands for the vertex attribute array.
void
AddVertexAttributeArrayCommand(
    ut::internal::CmdCache& cmdCache,
    u32                     vboBaseAddr,
    uptr                    phyArrayBuffer,
    u32                     vtxAttrNum
)
{
    const u32 fixedVtxAttrMask = 0;
    const u32 loadAryBytes = 16;        // Number of bytes near one vertex of the load array.
    const u32 loadAryElementNum = 1;

    const u32 command200[] =
    {
        // 0x200  base address
        PICA_CMD_DATA_VERTEX_ATTR_ARRAYS_BASE_ADDR( vboBaseAddr ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_VTX_ATTR_ARRAYS_BASE_ADDR, (0x202 - 0x200 + 1) + (0x205 - 0x203 + 1) * 12 ),

        // 0x201 The types for internal vertex attributes 0-7
        PICA_CMD_DATA_VERTEX_ATTR_ARRAYS(
            PICA_DATA_SIZE_4_FLOAT, PICA_DATA_SIZE_1_BYTE, PICA_DATA_SIZE_1_BYTE,
            PICA_DATA_SIZE_1_BYTE,  PICA_DATA_SIZE_1_BYTE, PICA_DATA_SIZE_1_BYTE,
            PICA_DATA_SIZE_1_BYTE,  PICA_DATA_SIZE_1_BYTE),

        // 0x202 The types for internal vertex attributes 8-11. The number of fixed vertex attribute masks and vertex attributes.
        PICA_DATA_SIZE_1_BYTE <<  0 | PICA_DATA_SIZE_1_BYTE <<  4
            | PICA_DATA_SIZE_1_BYTE <<  8 | PICA_DATA_SIZE_1_BYTE << 12
            | fixedVtxAttrMask << 16
            | (vtxAttrNum - 1) << 28,

        // 0x203-0x205; load array 0 address offset and elements; number of bytes and number of elements for 1-12.
        NW_FONT_CMD_SET_LOAD_ARRAY(
            phyArrayBuffer - vboBaseAddr,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            PICA_DATA_VERTEX_0_ATTR,
            loadAryBytes,
            loadAryElementNum),
    };
    cmdCache.Add(command200, sizeof(command200));

    static const u32 command206[] =
    {
        // 0x206-0x226 load arrays 1-11
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 1
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 2
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 3
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 4
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 5
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 6
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 7
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 8
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 9
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 10
        NW_FONT_CMD_SET_LOAD_ARRAY_ZERO(),  // Load array 11

        // 0x2bb  Input register map settings registers 1-8 of vertex attributes
        PICA_CMD_DATA_VS_ATTR_IN_REG_MAP0( 0, 0, 0, 0, 0, 0, 0, 0 ),
        PICA_CMD_HEADER_SINGLE( PICA_REG_VS_ATTR_IN_REG_MAP0 ),

        // 0x2bc  Input register map settings registers 9-12 of vertex attributes
        PICA_CMD_DATA_VS_ATTR_IN_REG_MAP1( 0, 0, 0, 0 ),
        PICA_CMD_HEADER_SINGLE( PICA_REG_VS_ATTR_IN_REG_MAP1 ),
    };
    cmdCache.Add(command206, sizeof(command206));
}
#endif  // #if defined(NW_PLATFORM_CTR)

}   // Namespace

bool        RectDrawer::s_StaticInitialized     = false;

RectDrawer::VertexBuffer
            RectDrawer::s_VertexBuffer;

void*
RectDrawer::GetVertexBufferData()
{
    if (! s_StaticInitialized)
    {
        InitializeStatic();

        s_StaticInitialized = true;
    }

    return &s_VertexBuffer;
}

u32
RectDrawer::GetCommandBufferSize(
    const void* shaderBinary,
    u32         /* size */
)
{
#if defined(NW_PLATFORM_CTR)
    EnsureShaderBinaryFileHeader(shaderBinary);
    const ShaderBinaryFileHeader& fileHead = GetShaderBinaryFileHeader(shaderBinary);
    NN_ASSERT(fileHead.exeCount == 1);  // Provides support only when there is one execute image.

    const ShaderPackageHeader& pkgHead = GetShaderPackageHeader(shaderBinary);
    EnsureShaderPackageHeader(pkgHead);

    const ExeImageHeader& exeHead = GetExeImageHeader(shaderBinary)[0];
    EnsureExeImageHeader(exeHead);

    return  CalcDrawBeginCommandBytes(pkgHead, exeHead)
        +   DRAWEND_STATICCOMMAND_BYTES;
#else
    NN_ASSERT(false);
    return 0;
#endif
}

u32
RectDrawer::GetVertexBufferCommandBufferSize(
    const void* shaderBinary,
    u32         size
)
{
    return  GetVertexBufferSize()
        +   GetCommandBufferSize(shaderBinary, size);
}

RectDrawer::RectDrawer()
:   m_UniformAddrIndex(0),
    m_UniformMtxIndex(0),
    m_UniformDataIndex(0),
    m_UniformTextColorIndex(0),
    m_UniformAddr(reinterpret_cast<UniformAddress*>(&m_UniformAddrBuffer[3])),
    m_UniformMtx(reinterpret_cast<math::VEC4*>(&m_UniformMtxBuffer[3])),
    m_UniformData(reinterpret_cast<math::VEC4*>(&m_UniformDataBuffer[3])),
    m_UniformTextColor(reinterpret_cast<math::VEC4*>(&m_UniformTextColorBuffer[3])),
    m_CommandBuffer(NULL),
    m_VertexBufferArray(NULL),
#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    m_VertexBufferBaseAddr(0),
#endif
    m_CommandJumpFirstReturnAddr(NULL),
    m_CommandJumpReturnAddr(NULL),
    m_CommandJumpSizePtr(NULL),
    m_IsColorDirty(false)
{
}

RectDrawer::~RectDrawer()
{
    this->Finalize();
}

void
RectDrawer::Initialize(
    void*       vtxBufCmdBuf,
    const void* shaderBinary,
    u32         size
)
{
    const u32 cmdBufSize = GetCommandBufferSize(shaderBinary, size);

    void *const vertexBuffer = ut::AddOffsetToPtr(vtxBufCmdBuf, cmdBufSize);

    InitializeCMD(vertexBuffer, vtxBufCmdBuf, shaderBinary, size, true);
}

void
RectDrawer::Initialize(
    void*       vertexBuffer,
    void*       commandBuffer,
    const void* shaderBinary,
    u32         size
)
{
    InitializeCMD(vertexBuffer, commandBuffer, shaderBinary, size, false);
}

void
RectDrawer::Finalize()
{
    m_CommandBuffer = NULL;
    m_VertexBufferArray = NULL;
}

void
RectDrawer::DrawBegin(u32 flag)
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->DrawBegin(pCmdStream, flag);
    SetCurrentCommandBuffer(pCmdStream);
}

/* virtual */
u32*
RectDrawer::DrawBegin(u32* pCmdStream, u32 flag)
{
    NN_ASSERT(  m_UniformAddrIndex      == 0
            &&  m_UniformDataIndex      == 0
            &&  m_UniformMtxIndex       == 0
            &&  m_UniformTextColorIndex == 0 );
    NW_ASSERT(pCmdStream);

    if (0 == (flag & DONT_USE_SETUP_COMMAND))
    {
        pCmdStream = UseBeginCommand(pCmdStream);
    }

    m_CommandJumpFirstReturnAddr = 0;

    return pCmdStream;
}

void
RectDrawer::DrawEnd(u32 flag)
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->DrawEnd(pCmdStream, flag);
    SetCurrentCommandBuffer(pCmdStream);
}

/* virtual */
u32*
RectDrawer::DrawEnd(u32* pCmdStream, u32 flag)
{
    if (0 == (flag & DONT_USE_SETUP_COMMAND))
    {
        pCmdStream = UseEndCommand(pCmdStream);
    }

    return pCmdStream;
}

void
RectDrawer::SetProjectionMtx(const nn::math::MTX44& mtx)
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->SetProjectionMtx(pCmdStream, mtx);
    SetCurrentCommandBuffer(pCmdStream);
}

u32*
RectDrawer::SetProjectionMtx(u32* pCmdStream, const nn::math::MTX44& mtx)
{
#if defined(NW_PLATFORM_CTR)
    nn::math::MTX44 *const pDst = reinterpret_cast<nn::math::MTX44*>(&m_UniformProjBuffer[3]);
    nn::math::MTX44Copy(pDst, &mtx);

    const u32 size = SetUniformCommand(m_UniformProjBuffer, UNIFORM_PROJ_NUM + 1);
    NW_FONT_ADD_COMMANDS_PTR(pCmdStream, m_UniformProjBuffer, size);
#else
    NN_ASSERT(false);
#endif
    return pCmdStream;
}

void
RectDrawer::SetViewMtxForText(const nn::math::MTX34& mtx)
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->SetViewMtxForText(pCmdStream, mtx);
    SetCurrentCommandBuffer(pCmdStream);
}

u32*
RectDrawer::SetViewMtxForText(u32* pCmdStream, const nn::math::MTX34& mtx)
{
#if defined(NW_PLATFORM_CTR)
    nn::math::MTX34 *const pDst = reinterpret_cast<nn::math::MTX34*>(&m_UniformMtxBuffer[3]);
    nn::math::MTX34Copy(pDst, &mtx);

    m_UniformMtxIndex = sizeof(mtx) / sizeof(nn::math::VEC4);
    const u32 size = SetUniformCommand(m_UniformMtxBuffer, m_UniformMtxIndex);
    NW_FONT_ADD_COMMANDS_PTR(pCmdStream, m_UniformMtxBuffer, size);
    m_UniformMtxIndex = 0;
#else
    NN_ASSERT(false);
#endif
    return pCmdStream;
}

u32*
RectDrawer::UseBeginCommand(u32* pCmdStream)
{
    pCmdStream = DisableTextures(pCmdStream);

    NW_FONT_ADD_COMMANDS_PTR(
        pCmdStream,
        m_CmdCacheDrawBegin.GetAddr(),
        m_CmdCacheDrawBegin.GetSize());

    return pCmdStream;
}

u32*
RectDrawer::UseEndCommand(u32* pCmdStream)
{
    NW_FONT_ADD_COMMANDS_PTR(
        pCmdStream,
        m_CmdCacheDrawEnd.GetAddr(),
        m_CmdCacheDrawEnd.GetSize());
    return pCmdStream;
}

u32*
RectDrawer::DisableTextures(u32* pCmdStream)
{
#if defined(NW_PLATFORM_CTR)
    // 0x06f
    NW_FONT_ADD_SINGLE_COMMAND_PTR(
        pCmdStream,
        PICA_CMD_HEADER_SINGLE_BE( PICA_REG_VS_OUT_ATTR_CLK, 0x2 ),
        0 );

    NW_FONT_ADD_COMMANDS_PTR(pCmdStream, TextureDummyCommands, sizeof(TextureDummyCommands));

    // 0x080
    // Select sampler type, coordinates
    NW_FONT_ADD_SINGLE_COMMAND_PTR(
        pCmdStream,
        PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEXTURE_FUNC, 0xB ),
        PICA_CMD_DATA_TEXTURE_FUNC(
            PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_FALSE /* texture0SamplerType */,
            0 /* texture1SamplerType */,
            0 /* texture2SamplerType */,
            0 /* texture3Texcoord */,
            0 /* texture3SamplerType */,
            0 /* texture2Texcoord */,
            0 /* clearTextureCache */));
#else
    NN_ASSERT(false);
#endif
    return pCmdStream;
}

u32*
RectDrawer::AddUniformMtx(u32* pCmdStream)
{
#if defined(NW_PLATFORM_CTR)
    const u32 size = SetUniformCommand(m_UniformMtxBuffer, m_UniformMtxIndex);
    NW_FONT_ADD_COMMANDS_PTR(pCmdStream, m_UniformMtxBuffer, size);
#else
    NN_ASSERT(false);
#endif

    m_UniformMtxIndex = 0;
    return pCmdStream;
}

void
RectDrawer::SetParallax(
    f32     parallax,
    f32     dLevel,
    f32     addDist
)
{
#if defined(NW_PLATFORM_CTR)
    nn::math::VEC4 *const pParallaxData =
        reinterpret_cast<nn::math::VEC4*>(&m_UniformProjBuffer[3 + 4 * 4]);
    pParallaxData->Set(parallax, dLevel, addDist, 0.0f);
#else
    NN_ASSERT(false);
#endif
}

void
RectDrawer::InitializeStatic()
{
    static const int vtx[DRAW_VTX_NUM] =
    {
        VERTEX_RT, VERTEX_LT, VERTEX_RB,
        VERTEX_RB, VERTEX_LT, VERTEX_LB,
    };

#if defined(NW_PLATFORM_CTR)
#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    for (int addr = 0; addr < UNIFORM_ADDR_NUM; ++addr)
    {
        for (int vi = 0; vi < DRAW_VTX_NUM; ++vi)
        {
            s_VertexBuffer.position[addr][vi][0] = addr;
            s_VertexBuffer.position[addr][vi][1] = vtx[vi];
            s_VertexBuffer.position[addr][vi][2] = vtx[vi] % 2 ?  1.f : 0.f; // Right and left
            s_VertexBuffer.position[addr][vi][3] = vtx[vi] / 2 ? -1.f : 0.f; // Top and bottom
        }
    }
#else
    for (int addr = 0; addr < UNIFORM_ADDR_NUM; ++addr)
    {
        for (int vi = 0; vi < DRAW_VTX_NUM; ++vi)
        {
            s_VertexBuffer.indexes[addr][vi] = addr * VERTEX_MAX + vtx[vi];
        }
    }

    for (int addr = 0; addr < UNIFORM_ADDR_NUM; ++addr)
    {
        for (int vi = 0; vi < VERTEX_MAX; ++vi)
        {
            s_VertexBuffer.position[addr][vi][0] = addr;
            s_VertexBuffer.position[addr][vi][1] = vi;
            s_VertexBuffer.position[addr][vi][2] = vi % 2 ?  1.f : 0.f; // Right and left
            s_VertexBuffer.position[addr][vi][3] = vi / 2 ? -1.f : 0.f; // Top and bottom
        }
    }
#endif
#else
    NN_ASSERT(false);
#endif
}

void
RectDrawer::InitializeCMD(
    void*       vertexBuffer,
    void*       commandBuffer,
    const void* shaderBinary,
    u32         /* size */,
    bool        initVertexBuffer
)
{
    NN_NULL_ASSERT(vertexBuffer);

#if defined(NW_PLATFORM_CTR)
    EnsureShaderBinaryFileHeader(shaderBinary);

    const ShaderBinaryFileHeader& fileHead = GetShaderBinaryFileHeader(shaderBinary);

    NN_ASSERT(fileHead.exeCount == 1);  // Provides support only when there is one execute image.

    const ShaderPackageHeader& pkgHead = GetShaderPackageHeader(shaderBinary);

    EnsureShaderPackageHeader(pkgHead);

    const ExeImageHeader& exeHead = GetExeImageHeader(shaderBinary)[0];

    EnsureExeImageHeader(exeHead);

    NN_ASSERT(exeHead.shaderType == 0); // Supports only the vertex shader.

    m_VertexBufferArray = vertexBuffer;
    if (initVertexBuffer)
    {
        std::memcpy(vertexBuffer, GetVertexBufferData(), GetVertexBufferSize());
        nngxUpdateBuffer(vertexBuffer, GetVertexBufferSize());
    }

    m_CommandBuffer = commandBuffer;
    const u32 DrawBeginCommandBytes = CalcDrawBeginCommandBytes(pkgHead, exeHead);
    m_CmdCacheDrawBegin.Init(m_CommandBuffer, DrawBeginCommandBytes);
    m_CmdCacheDrawEnd.Init(
        ut::AddOffsetToPtr(m_CommandBuffer, DrawBeginCommandBytes),
        DRAWEND_STATICCOMMAND_BYTES);

    const u32 vtxAttrNum = 1;           // Number of vertex attributes

    // Whether to use the geometry shader
    const bool useGeometryShader = false;

    // Whether to use the sub-divisions of the reserved geometry shader
    const bool useGeometryShaderSubdivision = false;

    // Generates the command cache of DrawBegin.
    {
        static const u32 commands[] =
        {
            // 0x251~, 0x200~ (44 words)
            PICA_CMD_SET_DRAW_MODE0_DUMMY_BEGIN(),

            // 0x229 [1:0] Not using the geometry shader.
            PICA_CMD_DATA_DRAW_MODE0(
                useGeometryShader,
                0                   /* drawMode */,
                false),             /* useGeometryShaderSubdivision */
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_DRAW_MODE0, 0x1 ),

            // 0x200~ (32 words)
            PICA_CMD_SET_DRAW_MODE0_DUMMY_END(),

            // 0x244 [0:0] copy of shared processor
            PICA_CMD_SET_VS_COM_MODE( useGeometryShader ),
        };
        m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
    }

    {   // Vertex shader program code
        const u32* progCodes =
            static_cast<const u32*>(ut::AddOffsetToPtr(&pkgHead, pkgHead.instOffset));

        for (u32 totalCnt = 0; totalCnt < pkgHead.instCount; )
        {
            const u32 writeCnt = math::Min(BURST_MAX, pkgHead.instCount - totalCnt);
            const u32 commands[] =
            {
                // 0x2cb  program shader load address
                PICA_CMD_DATA_VS_PROG_ADDR( totalCnt ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_PROG_ADDR ),

                // 0x2cc  program code data
                progCodes[totalCnt + 0],
                PICA_CMD_HEADER_BURST( PICA_REG_VS_PROG_DATA0, writeCnt ),
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
            m_CmdCacheDrawBegin.Add(&progCodes[totalCnt + 1], sizeof(u32) * (writeCnt - 1));
            m_CmdCacheDrawBegin.RoundUp(8);
            totalCnt += writeCnt;
        }

        {
            static const u32 commands[] =
            {
                // 0x2bf  Notification that program updating has completed.
                PICA_CMD_DATA_VS_PROG_END( 1 ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_PROG_RENEWAL_END ),
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
        }
    }

    {   // Vertex shader swizzle pattern
        const SwizzleInfo *const swDatas =
            static_cast<const SwizzleInfo*>(ut::AddOffsetToPtr(&pkgHead, pkgHead.swizzleOffset));

        for (u32 totalCnt = 0; totalCnt < pkgHead.swizzleCount; )
        {
            const u32 writeCnt = math::Min(BURST_MAX, pkgHead.swizzleCount - totalCnt);
            const u32 commands[] =
            {
                // 0x2d5  Swizzle pattern load address
                PICA_CMD_DATA_VS_PROG_SWIZZLE_ADDR( totalCnt ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_PROG_SWIZZLE_ADDR ),

                // 0x2d6  Swizzle pattern data
                swDatas[totalCnt + 0].value,
                PICA_CMD_HEADER_BURST( PICA_REG_VS_PROG_SWIZZLE_DATA0, writeCnt ),
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));

            for (int i = 1; i < writeCnt; ++i)
            {
                m_CmdCacheDrawBegin.Add(&swDatas[totalCnt + i].value, sizeof(u32));
            }

            m_CmdCacheDrawBegin.RoundUp(8);
            totalCnt += writeCnt;
        }
    }

    {   // Floating-point value constants register
        const SetupInfo* setupInfos =
            static_cast<const SetupInfo*>(
                ut::AddOffsetToPtr(
                    &exeHead,
                    exeHead.setupOffset ) );

        for (int i = 0; i < exeHead.setupCount; ++i)
        {
            const SetupInfo& setupInfo = setupInfos[i];

            // Implements only when there are floating-point values.
            NN_ASSERT(setupInfo.type == 2);

            const int dataNum = 3;  // 24bit * 4 components
            const int writeCount = 1 + dataNum; // Number of writes (addresses and data)
            const u32 commands[] =
            {
                // 0x2c0
                PICA_CMD_DATA_VS_FLOAT_ADDR( PICA_DATA_VS_F24, setupInfo.index ),
                PICA_CMD_HEADER_BURSTSEQ( PICA_REG_VS_FLOAT_ADDR, writeCount ),

                setupInfo.value[3] <<  8 | ((setupInfo.value[2] >> 16) & 0x0000FF),
                setupInfo.value[2] << 16 | ((setupInfo.value[1] >>  8) & 0x00FFFF),
                setupInfo.value[1] << 24 | ((setupInfo.value[0] >>  0) & 0xFFFFFF),
                0,  // Alignment adjustments
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
        }
    }

    const uptr phyArrayBuffer = nngxGetPhysicalAddr(reinterpret_cast<uptr>(m_VertexBufferArray));

    // Vertex array base address
    // Configures the 16-byte aligned values that are smaller than all vertex array and index array addresses.
    const u32 vboBaseAddr = math::RoundDown(phyArrayBuffer, 16);

    AddVertexAttributeArrayCommand(m_CmdCacheDrawBegin, vboBaseAddr, phyArrayBuffer, vtxAttrNum);

    {
        const u32 commands[] =
        {
            // 0x229 [31:31]
            PICA_CMD_DATA_DRAW_MODE0(
                false,                          /* useGeometryShader */
                0,                              /* drawMode */
                useGeometryShaderSubdivision),
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_DRAW_MODE0, 0x8 ),

            // 0x2b9  Number of vertex attributes input to the vertex shader.
            PICA_CMD_DATA_VS_ATTR_NUM0( vtxAttrNum ),
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_VS_ATTR_NUM0, 0xb ),

            // 0x2ba  Vertex shader start address
            PICA_CMD_DATA_VS_START_ADDR( exeHead.mainAddr ),
            PICA_CMD_HEADER_SINGLE( PICA_REG_VS_START_ADDR ),
        };
        m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
    }

    {
        const u32 OutRegNum = 7;
        const int CompNum = 4;
        static const u8 outAttrs[SHADEROUTMAPTYPE_MAX] =
        {
            PICA_DATA_VS_OUT_ATTR_X,
            PICA_DATA_VS_OUT_ATTR_QUART_X,
            PICA_DATA_VS_OUT_ATTR_R,
            PICA_DATA_VS_OUT_ATTR_TEX0_U,
            PICA_DATA_VS_OUT_ATTR_TEX0_W,
            PICA_DATA_VS_OUT_ATTR_TEX1_U,
            PICA_DATA_VS_OUT_ATTR_TEX2_U,
            PICA_DATA_VS_OUT_ATTR_VIEW_X
        };
        u8 outmaps[OutRegNum][CompNum];
        bool useRegs[OutRegNum];
        u8 useOutRegTypes[SHADEROUTMAPTYPE_MAX];

        for (u32 i = 0; i < SHADEROUTMAPTYPE_MAX; ++i)
        {
            useOutRegTypes[i] = 0;
        }

        for (u32 j = 0; j < OutRegNum; ++j)
        {
            useRegs[j] = false;
            for (int i = 0; i < CompNum; ++i)
            {
                outmaps[j][i] = PICA_DATA_VS_OUT_ATTR_INVALID;
            }
        }

        const OutmapInfo *const outmapInfos =
            static_cast<const OutmapInfo*>(
                ut::AddOffsetToPtr(
                    &exeHead,
                    exeHead.outmapOffset ) );
        for (int j = 0; j < math::Min(exeHead.outmapCount, OutRegNum); ++j)
        {
            if (outmapInfos[j].type == 9)   // generic
            {
                continue;
            }

            useRegs[outmapInfos[j].index] = true;
            u8* outmap = outmaps[outmapInfos[j].index];
            useOutRegTypes[outmapInfos[j].type] = static_cast<u8>(outmapInfos[j].mask);
            u8 ci = outAttrs[outmapInfos[j].type];
            for (int i = 0; i < CompNum; ++i)
            {
                if (0 != (outmapInfos[j].mask & (1 << i)))
                {
                    outmap[i] = ci;
                    ++ci;
                }
            }
        }

        int outRegUseNum = 0;
        for (int j = 0; j < OutRegNum; ++j)
        {
            if (useRegs[j])
            {
                ++outRegUseNum;
            }
        }

        {
            const u32 commands[] =
            {
                // 0x2bd  Vertex shader output register mask
                PICA_CMD_DATA_VS_OUT_MASK( exeHead.outputMask ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_OUT_REG_MASK ),

                // 0x251  Number of vertex shader output registers used.
                outRegUseNum - 1,
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_OUT_REG_NUM2 ),

                // 0x252  Geometry shader mode
                PICA_CMD_DATA_GS_MISC_REG0( PICA_DATA_GS_OTHER_MODE ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_GS_MISC_REG0 ),

                // 0x242  Number of vertex attributes input to the vertex shader.
                PICA_CMD_DATA_VS_ATTR_NUM1( vtxAttrNum ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_ATTR_NUM1 ),

                // 0x24a, 0x25e[3:0], 0x04f  Number of vertex shader output registers used.
                outRegUseNum - 1,
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_OUT_REG_NUM1 ),
                PICA_CMD_DATA_VS_GS_OUT_REG_NUM3( outRegUseNum, 0 /* mode */ ),
                PICA_CMD_HEADER_SINGLE_BE( PICA_REG_VS_OUT_REG_NUM3, 0x1 ),
                outRegUseNum,
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_OUT_REG_NUM0 ),
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
        }

    // 0x050-0x056 Vertex shader output attributes

        {
            const u32 commands[] =
            {
                *reinterpret_cast<u32*>(outmaps[0]),
                PICA_CMD_HEADER_BURSTSEQ( PICA_REG_VS_OUT_ATTR0, OutRegNum ),
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
            m_CmdCacheDrawBegin.Add(reinterpret_cast<u32*>(outmaps[1]), sizeof(u32) * (OutRegNum - 1));
        }

        {
            bool usePosZ = false;
            {
                int outCompNum = 0;
                for (int i = 0; i < 4; ++i)
                {
                    if (useOutRegTypes[SHADEROUTMAPTYPE_POSITION] & (1 << i))
                    {
                        ++outCompNum;
                    }
                }

                usePosZ = outCompNum >= 3;
            }

            const bool useViewQuaternion =
                    0 != useOutRegTypes[SHADEROUTMAPTYPE_VIEW]
                ||  0 != useOutRegTypes[SHADEROUTMAPTYPE_QUATERNION];
            const u32 outAttrClk =
                PICA_CMD_DATA_VS_GS_OUT_ATTR_CLK(
                    usePosZ,                                            /* posZ */
                    0 != useOutRegTypes[SHADEROUTMAPTYPE_COLOR],        /* col */
                    0 != useOutRegTypes[SHADEROUTMAPTYPE_TEXCOORD0],    /* tex0 */
                    0 != useOutRegTypes[SHADEROUTMAPTYPE_TEXCOORD1],    /* tex1 */
                    0 != useOutRegTypes[SHADEROUTMAPTYPE_TEXCOORD2],    /* tex2 */
                    0 != useOutRegTypes[SHADEROUTMAPTYPE_TEXCOORD0W],   /* tex0_w */
                    useViewQuaternion);                                 /* view_quart*/
            const u32 TexEnableBits =
                PICA_CMD_DATA_VS_GS_OUT_ATTR_CLK(
                    false,      /* posZ */
                    false,      /* col */
                    true,       /* tex0 */
                    true,       /* tex1 */
                    true,       /* tex2 */
                    true,       /* tex0_w */
                    false);     /* view_quart*/

            const bool useTexCoords = 0 != (outAttrClk & TexEnableBits);
            const u32 outAttrMode = useTexCoords ? 1 : 0;
            const u32 commands[] =
            {
                // 0x064  Case 1: texture coordinates being output from the vertex shader.
                PICA_CMD_DATA_VS_GS_OUT_ATTR_MODE( outAttrMode ),
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_OUT_ATTR_MODE ),

                // 0x06f  Output attribute clock controls from the vertex shader.
                outAttrClk,
                PICA_CMD_HEADER_SINGLE( PICA_REG_VS_OUT_ATTR_CLK ),
            };
            m_CmdCacheDrawBegin.Add(commands, sizeof(commands));
        }
    }

    m_CmdCacheDrawBegin.Add(s_DrawInitCommands, sizeof(s_DrawInitCommands));

#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    m_VertexBufferBaseAddr = vboBaseAddr;
#else
    {
        void *const eleArrayBuf = ut::AddOffsetToPtr(m_VertexBufferArray, sizeof(s_VertexBuffer.position));
        const uptr phyEleArrayBuf = nngxGetPhysicalAddr(reinterpret_cast<uptr>(eleArrayBuf));
        const u32 indexOffset = phyEleArrayBuf - vboBaseAddr;
        const u32 commands[] =
        {
            // 0x227
            PICA_CMD_DATA_INDEX_ARRAY_ADDR_OFFSET(
                indexOffset,
                1),             /* type */
            PICA_CMD_HEADER_SINGLE( PICA_REG_INDEX_ARRAY_ADDR_OFFSET ),
        };
        m_CmdCacheDrawBegin.Add(commands, sizeof(commands) );
    }
#endif

    // m_CmdCacheDrawBegin.Dump();

    // Generates the DrawEnd command cache.
    AddVertexAttributeArrayCommand(m_CmdCacheDrawEnd, vboBaseAddr, phyArrayBuffer, vtxAttrNum);

    m_CmdCacheDrawEnd.Add(s_DrawFinalizeCommands, sizeof(s_DrawFinalizeCommands));

    // m_CmdCacheDrawEnd.Dump();


    { // Initializing the uniform command

        // Initializing the projection matrix command
        InitUniformBuffer(m_UniformProjBuffer, UNIFORM_PROJ_START);

        // Initializing the parallax information command
        SetParallax(0.0f, 0.0f, 0.0f);

        // Initializing buffers where addresses are added
        InitUniformBuffer(m_UniformAddrBuffer, UNIFORM_ADDR_START);

        // Initializing buffers where matrix and vertex color are added.
        InitUniformBuffer(m_UniformMtxBuffer, UNIFORM_MTX_START);

        // Initializing buffers where texture coordinates and rectangular information are added.
        InitUniformBuffer(m_UniformDataBuffer, UNIFORM_DATA_START);
    }
#else
    NN_ASSERT(false);
#endif
}

void
RectDrawer::InitUniformBuffer(
    u32* __restrict buf,
    u32             addr
)
{
#if defined(NW_PLATFORM_CTR)
    const u32 data =
        PICA_CMD_DATA_VS_FLOAT_ADDR( PICA_DATA_VS_F32, addr );

    // 0x2c0
    buf[0] = data;
    buf[1] = PICA_CMD_HEADER_SINGLE( PICA_REG_VS_FLOAT_ADDR );
#endif
}

#if !defined(NW_FONT_ENABLE_EDGE_FONT)
void
RectDrawer::BuildTextCommand(CharWriter* pCharWriter)
#else
void
RectDrawer::BuildTextCommand(CharWriter* pCharWriter, FontCombinerType combiner )
#endif
{
#if defined(NW_PLATFORM_CTR)
    NN_ASSERT(  m_UniformAddrIndex      == 0
            &&  m_UniformDataIndex      == 0
            &&  m_UniformMtxIndex       == 0
            &&  m_UniformTextColorIndex == 0 );

    DispStringBuffer* __restrict pStringBuffer = pCharWriter->GetDispStringBuffer();
    pStringBuffer->commandBufferSize = 0;
    pStringBuffer->textColorCommandOffset = 0;
    pStringBuffer->generatedCommand = true;     // Completes the generation of commands.

    const u16 charCount = math::Min(pStringBuffer->charCount, pStringBuffer->GetDrawCharCount());

    // Exit when the number of characters to draw is 0.
    if (charCount == 0)
    {
        return;
    }

    m_UniformAddress.addrMtx = 0;
    m_IsColorDirty = false;

    // Initializing buffers where texture colors are added
    InitUniformBuffer(m_UniformTextColorBuffer, UNIFORM_TEXTCOLOR_START);

    NW_FONT_ADD_COMMANDS(pStringBuffer, TextInitCommands, sizeof(TextInitCommands));

    // Set vertex color.
    ut::Color8 vtxColors[internal::TEXTCOLOR_MAX] =
    {
        pCharWriter->GetGradationStartColor(),
        pCharWriter->GetGradationEndColor(),
    };

    AddTextColor(
        &m_UniformTextColor[m_UniformTextColorIndex],
        vtxColors,
        pCharWriter->GetGradationMode());

    // Flag indicating whether drawing has taken place.
    u8 *const drawFlags = pStringBuffer->drawFlags;
    for (int i = math::RoundUp(charCount, 8) / 8; i > 0; --i)
    {
        drawFlags[i - 1] = 0;

        /*
         * std::memset(drawFlags, 0, charCount);
         Loop was faster than * .
         */
    }

    u32 addr = 0;
    u32 size = 0;
    u32 format = 0xFFFFFFFF;
    u32 texFilter = 0xFFFFFFFF;
    u32 startCharIndex = 0;
    internal::CharAttribute* __restrict charAttrs = pStringBuffer->GetCharAttrs();

    for (u32 drawCharCount = 0; drawCharCount < charCount; )
    {
        u32 i = startCharIndex;
        for ( ; IsDrawChar(drawFlags, i); ++i)
        {
            ;
        }
        startCharIndex = i + 1;     // Next search start position

        u32 cmdTexAddr = charAttrs[i].pTexObj->GetImage();
        if (addr != cmdTexAddr)
        {
            addr = cmdTexAddr;

            // Switches textures, so writes data from buffer.
            if (drawCharCount > 0 && m_UniformAddrIndex > 0)
            {
                UniformAndDrawText(pStringBuffer);
            }

            if (format != charAttrs[i].pTexObj->GetFormat())
            {
                const u32 newFormat = charAttrs[i].pTexObj->GetFormat();

                const bool isNewAlpha  = newFormat == FONT_SHEET_FORMAT_A4 || newFormat == FONT_SHEET_FORMAT_A8;
                const u32* commands = NULL;
                u32 commandSize = 0;
                if (format == 0xFFFFFFFF)   // The first time
                {
                #if !defined(NW_FONT_ENABLE_EDGE_FONT)

                    // Adds all settings for texture combiner stages 3 and 4.
                    // 
                    // The following settings' colors' values will be overwritted directly before issuing (the UseCommandBuffer function), so take care not to change the position in the command list.
                    // 
                    // 
                    // 
                    if (isNewAlpha)
                    {
                        commands = TexEnv34AlphaCommands;
                        commandSize = sizeof(TexEnv34AlphaCommands);
                    }
                    else
                    {
                        commands = TexEnv34Commands;
                        commandSize = sizeof(TexEnv34Commands);
                    }

                #else

                    // Adds all settings for texture combiner stages 3 and 4.
                    if (isNewAlpha)
                    {
                        commands = TexEnv34AlphaCommands[ combiner ];
                        commandSize = TexEnv34AlphaCommandsSize[ combiner ];
                    }
                    else
                    {
                        commands = TexEnv34Commands[ combiner ];
                        commandSize = TexEnv34CommandsSize[ combiner ];
                    }

                #endif

                    NW_FONT_ADD_COMMANDS(pStringBuffer, commands, commandSize);
                }
                else
                {
                    // Adds an amount equivalent to the change in the texture combiner.
                    const bool isCrntAlpha = format == FONT_SHEET_FORMAT_A4 || format == FONT_SHEET_FORMAT_A8;

                    if (isCrntAlpha != isNewAlpha)
                    {
                    #if !defined(NW_FONT_ENABLE_EDGE_FONT)

                        if (isNewAlpha)
                        {
                            commands = TexEnvOperand34AlphaCommands;
                            commandSize = sizeof(TexEnvOperand34AlphaCommands);
                        }
                        else
                        {
                            commands = TexEnvOperand34Commands;
                            commandSize = sizeof(TexEnvOperand34Commands);
                        }

                    #else

                        if (isNewAlpha)
                        {
                            commands = TexEnvOperand34AlphaCommands[ combiner ];
                            commandSize = TexEnvOperand34AlphaCommandsSize[ combiner ];
                        }
                        else
                        {
                            commands = TexEnvOperand34Commands[ combiner ];
                            commandSize = TexEnvOperand34CommandsSize[ combiner ];
                        }

                    #endif

                        NW_FONT_ADD_COMMANDS(pStringBuffer, commands, commandSize);
                    }

                    NW_FONT_ADD_COMMANDS(pStringBuffer, TextureDummyCommands, sizeof(TextureDummyCommands));
                }

                format = newFormat;

                // 0x08e
                NW_FONT_ADD_SINGLE_COMMAND(
                    pStringBuffer,
                    PICA_CMD_HEADER_SINGLE(PICA_REG_TEXTURE0_FORMAT),
                    format);
            }
            else
            {
                NW_FONT_ADD_COMMANDS(pStringBuffer, TextureDummyCommands, sizeof(TextureDummyCommands));
            }

            // Clears the texture cache.
            // 0x080
            NW_FONT_ADD_SINGLE_COMMAND(
                pStringBuffer,
                PICA_CMD_HEADER_SINGLE_BE(PICA_REG_TEXTURE_FUNC, 0x4),
                TEX_FUNC_DATA);

            // Texture address
            // 0x085
            NW_FONT_ADD_SINGLE_COMMAND(
                pStringBuffer,
                PICA_CMD_HEADER_SINGLE(PICA_REG_TEXTURE0_ADDR1),
                addr);

            // Texture size
            if (size != charAttrs[i].pTexObj->GetSize().size)
            {
                size = charAttrs[i].pTexObj->GetSize().size;
                // 0x082
                NW_FONT_ADD_SINGLE_COMMAND(
                    pStringBuffer,
                    PICA_CMD_HEADER_SINGLE(PICA_REG_TEXTURE0_SIZE),
                    size);
            }

            const u32 crntTexFilter = charAttrs[i].pTexObj->GetWrapFilter();
            if (texFilter != crntTexFilter)
            {
                texFilter = crntTexFilter;
                // 0x083
                NW_FONT_ADD_SINGLE_COMMAND(
                    pStringBuffer,
                    PICA_CMD_HEADER_SINGLE(PICA_REG_TEXTURE0_WRAP_FILTER),
                    texFilter);
            }
        }

        for ( ; i < charCount; ++i)
        {
            if (IsDrawChar(drawFlags, i) || addr != charAttrs[i].pTexObj->GetImage())
            {
                continue;
            }

            const internal::CharAttribute& charAttr= pStringBuffer->GetCharAttrs()[i];
            if ( vtxColors[internal::TEXTCOLOR_START] != charAttr.color[internal::TEXTCOLOR_START]
              || vtxColors[internal::TEXTCOLOR_END  ] != charAttr.color[internal::TEXTCOLOR_END  ]
            )
            { // Set vertex color.
                vtxColors[internal::TEXTCOLOR_START] = charAttr.color[internal::TEXTCOLOR_START];
                vtxColors[internal::TEXTCOLOR_END  ] = charAttr.color[internal::TEXTCOLOR_END  ];
                AddTextColor(
                    &m_UniformTextColor[m_UniformTextColorIndex],
                    vtxColors,
                    pCharWriter->GetGradationMode());
            }

            { // Texture coordinate settings
                m_UniformAddress.addrTexCoord = m_UniformDataIndex;
                math::VEC4* __restrict pTexCoord = &m_UniformData[m_UniformDataIndex];
                *pTexCoord = charAttr.tex;

                ++m_UniformDataIndex;
            }

            { // Rectangle settings

                m_UniformAddress.addrSizeAndVtx = m_UniformDataIndex;
                math::VEC4* __restrict pQuadInfo = &m_UniformData[m_UniformDataIndex];
                *pQuadInfo = charAttr.pos;

                ++m_UniformDataIndex;
            }

            SetDrawChar(drawFlags, i);
            ++drawCharCount;

            m_UniformAddr[m_UniformAddrIndex] = m_UniformAddress;
            ++m_UniformAddrIndex;

#ifdef NW_FONT_RECTDRAWER_USE_DRAW_BUFFER
            if ( m_UniformAddrIndex + 1                  > UNIFORM_ADDR_NUM
              || m_UniformDataIndex + 1 + 1              > UNIFORM_DATA_NUM
              || m_UniformTextColorIndex + COLOR_USE_COUNT > UNIFORM_TEXTCOLOR_NUM
            )
            {
                UniformAndDrawText(pStringBuffer);
            }
#else
            UniformAndDrawText(pStringBuffer);
#endif
        }
    }

    if (m_UniformAddrIndex > 0)
    {
        UniformAndDrawText(pStringBuffer);
    }
    m_UniformTextColorIndex = 0;

    // Adds an end command to the sub-routine command list if command jump is enabled.
    if (pStringBuffer->GetCommandBufferMode() & DispStringBuffer::CBF_COMMAND_JUMP)
    {
        RectDrawer::BuildReturnCommand(pStringBuffer);
    }
#else
    NN_ASSERT(false);
#endif
}

void
RectDrawer::UniformAndDrawText(DispStringBuffer* pStringBuffer)
{
#if defined(NW_PLATFORM_CTR)
    NN_ASSERT(m_UniformAddrIndex > 0);

    const int vtxNum = DRAW_VTX_NUM * m_UniformAddrIndex;   // Number of vertices to draw

    // Uniform that goes to uUniformAddr.
    {
        u32 size = SetUniformCommand(m_UniformAddrBuffer, m_UniformAddrIndex);
        NW_FONT_ADD_COMMANDS(pStringBuffer, m_UniformAddrBuffer, size);
        m_UniformAddrIndex = 0;
    }

    if (m_IsColorDirty)
    { // The vertex color uniform that goes to uUniformMtx.
        u32 size = SetUniformCommand(m_UniformTextColorBuffer, m_UniformTextColorIndex);

        // Saves the text color command position.
        if (pStringBuffer->textColorCommandOffset == 0)
        {
            pStringBuffer->textColorCommandOffset = pStringBuffer->commandBufferSize;

            // Since the first colors are for animations, they are kept from being included in the command buffer.
            // Configures only when additional colors are registered.
            if (m_UniformTextColorIndex > VERTEX_MAX)
            {
                // Location of the second color.
                const u32 index = 3 + VERTEX_MAX * 4;

                NW_FONT_ADD_SINGLE_COMMAND(
                    pStringBuffer,
                    PICA_CMD_HEADER_SINGLE(PICA_REG_VS_FLOAT_ADDR),
                    PICA_CMD_DATA_VS_FLOAT_ADDR(PICA_DATA_VS_F32, UNIFORM_TEXTCOLOR_START + VERTEX_MAX));
                NW_FONT_ADD_SINGLE_COMMAND(
                    pStringBuffer,
                    PICA_CMD_HEADER_VS_F32(m_UniformTextColorIndex - VERTEX_MAX),
                    m_UniformTextColorBuffer[index]);
                NW_FONT_ADD_COMMANDS(pStringBuffer, m_UniformTextColorBuffer + index + 1, size - sizeof(u32) * (index + 1));
            }
        }
        else
        {
            NW_FONT_ADD_COMMANDS(pStringBuffer, m_UniformTextColorBuffer, size);
        }

        const u8 lastColorIdx = m_UniformTextColorIndex - VERTEX_MAX;
        if (lastColorIdx > 0)
        {
            // If there are two or more colors, brings the last color to the front.
            for (int i = 0; i < VERTEX_MAX; ++i)
            {
                m_UniformTextColor[i] = m_UniformTextColor[lastColorIdx + i];
            }
            m_IsColorDirty = true;
        }
        else
        {
            // If only one, then nothing happens. However, returns the value that was swapped with the header.
            m_UniformTextColorBuffer[3] = m_UniformTextColorBuffer[2];
            m_IsColorDirty = false;
        }
        m_UniformAddress.addrColor = MATRIX_UNIFORM_NUM;
        m_UniformTextColorIndex = VERTEX_MAX;
    }

    if (m_UniformDataIndex > 0)
    { // Uniform that goes to uUniformData

        u32 size = SetUniformCommand(m_UniformDataBuffer, m_UniformDataIndex);
        NW_FONT_ADD_COMMANDS(pStringBuffer, m_UniformDataBuffer, size);
        m_UniformDataIndex = 0;
    }

    // Adds draw commands to the command list.
#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    const u32 addrOffset = GetVertexIndexAddressOffset(vtxNum);
#else
    const u32 addrOffset = 0;   // Dummy
#endif

    u32* cmdPtr = &pStringBuffer->commandBuffer[pStringBuffer->commandBufferSize];
    cmdPtr = internal::SetVertexNumCmd(cmdPtr, addrOffset, vtxNum);
    pStringBuffer->commandBufferSize = cmdPtr - &pStringBuffer->commandBuffer[0];

    NW_FONT_ADD_COMMANDS( pStringBuffer, GetDrawCommands(), GetDrawCommandSize() );
#else
    NN_ASSERT(false);
#endif
}

#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
u32
RectDrawer::GetVertexIndexAddressOffset(u32 vtxNum) const
{
    return (((vtxNum > 0x10 ? (vtxNum - 0x10) * 2: 0) + m_VertexBufferBaseAddr) & 0xFFF) >= 0xFE0 ? 0x20: 0;
}
#endif

const u32*
RectDrawer::GetDrawCommands() const
{
#if defined(NW_PLATFORM_CTR)
    return DrawCommands;
#else
    return NULL;
#endif
}

u32
RectDrawer::GetDrawCommandSize() const
{
#if defined(NW_PLATFORM_CTR)
    return sizeof(DrawCommands);
#else
    return 0;
#endif
}

void
RectDrawer::AddTextColor(
    math::VEC4* __restrict          dst,
    const ut::Color8* __restrict    src,
    int                             mode
)
{
#if defined(NW_PLATFORM_CTR)
    SetVertexColor(dst, src, mode);
    m_UniformAddress.addrColor = static_cast<f32>(MATRIX_UNIFORM_NUM + m_UniformTextColorIndex);
    m_UniformTextColorIndex += COLOR_USE_COUNT;
    m_IsColorDirty = true;
#else
    NN_ASSERT(false);
#endif
}

// ---- Below, a code other than that for the RectDrawer class  ----

u32
DispStringBuffer::CalcCommandBufferCapacity(u32 charNum)
{
#if defined(NW_PLATFORM_CTR)
#if !defined(NW_FONT_ENABLE_EDGE_FONT)
    NN_ASSERT(sizeof(TexEnv34Commands) == sizeof(TexEnv34AlphaCommands));
    NN_ASSERT(sizeof(TexEnvOperand34Commands) == sizeof(TexEnvOperand34AlphaCommands));
#else
    // Because FONT_COMBINER_TYPE cannot be gotten here, return the maximum size.
    // 
    u32 TexEnvOperandCommandsSize = 0;
    u32 TexEnvCommandsSize        = 0;

    {
        const u32* cmd_size   = TexEnv34CommandsSize;
        const u32* ope_size   = TexEnvOperand34CommandsSize;
        const u32* cmd_size_a = TexEnv34AlphaCommandsSize;
        const u32* ope_size_a = TexEnvOperand34AlphaCommandsSize;

        for( int i = 1;  i < FONT_COMBINER_TYPE_MAX;  ++i )
        {
            NN_ASSERT(cmd_size[i] == cmd_size_a[i]);
            NN_ASSERT(ope_size[i] == ope_size_a[i]);
        }

        TexEnvOperandCommandsSize = ope_size[0];
        TexEnvCommandsSize        = cmd_size[0];

        for( int i = 1;  i < FONT_COMBINER_TYPE_MAX;  ++i )
        {
            if( ope_size[i] >= TexEnvOperandCommandsSize )
                TexEnvOperandCommandsSize = ope_size[i];

            if( cmd_size[i] >= TexEnvCommandsSize )
                TexEnvCommandsSize = cmd_size[i];
        }
    }
#endif

    const int UniformComponents = 4;    // The number of uniform components (xyzw).
    const int SingleCommand = 2;        // Single command
    // The header of the command for uniform transfer
    const int UniformHeader = math::RoundUp(SingleCommand + 1, 2);

    u32 cap =
                  sizeof(TextInitCommands) / sizeof(u32)      // Blend combiner shader settings
                + UniformComponents * COLOR_USE_COUNT       // Initial character colors
              #if !defined(NW_FONT_ENABLE_EDGE_FONT)
                + sizeof(TexEnv34Commands) / sizeof(u32)    // Combiner settings
              #else
                + TexEnvCommandsSize / sizeof(u32)
              #endif
                + charNum *
                    (
                        #if !defined(NW_FONT_ENABLE_EDGE_FONT)
                          sizeof(TexEnvOperand34Commands) / sizeof(u32)     // Amount of combiner change
                        #else
                          TexEnvOperandCommandsSize / sizeof(u32)
                        #endif
                        + sizeof(TextureDummyCommands) / sizeof(u32)     // 0x080 dummy command
                        + SingleCommand * 5     // Four single commands for texture
                                                // (address, size, format, filter, cache clear)
                        + UniformHeader + UniformComponents * 1                 // Address
                        + UniformHeader + UniformComponents * COLOR_USE_COUNT   // Character color
                        + UniformHeader + UniformComponents * (1 + 1)   // Texture coordinates and rectangle information
#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
                        + SingleCommand * 1     // Vertex index offset
#endif
                        + SingleCommand * 1     // Number of vertices
                        + sizeof(DrawCommands) / sizeof(u32)    // Draw commands
                    )
            ;
    // Keeps at a multiple of 4 words, including the return kick command (2 words) from the sub-routine.
    cap = math::RoundUp(cap + 2, 4);
    return cap;
#else
    return 0;
#endif
}

#if defined(NW_PLATFORM_CTR)
u32*
RectDrawer::AddTextVertexColorCommand(u32* pCmdStream, const u32* cmdBuf)
{
    NW_FONT_ADD_SINGLE_COMMAND_PTR(
        pCmdStream,
        PICA_CMD_HEADER_SINGLE(PICA_REG_VS_FLOAT_ADDR),
        PICA_CMD_DATA_VS_FLOAT_ADDR(PICA_DATA_VS_F32, UNIFORM_TEXTCOLOR_START));
    NW_FONT_ADD_SINGLE_COMMAND_PTR(
        pCmdStream,
        PICA_CMD_HEADER_VS_F32(VERTEX_MAX),
        cmdBuf[0]);
    NW_FONT_ADD_COMMANDS_PTR(
        pCmdStream,
        cmdBuf + 1, sizeof(u32) * (VERTEX_MAX * 4));
    return pCmdStream;
}

u32*
#if defined(NW_FONT_ENABLE_EDGE_FONT)
RectDrawer::AddTextMappingColorCommand(u32* pCmdStream, u32 white, u32 black, FontCombinerType combiner)
#else
RectDrawer::AddTextMappingColorCommand(u32* pCmdStream, u32 white, u32 black)
#endif
{
#if defined(NW_FONT_ENABLE_EDGE_FONT)
    if( combiner != FONT_COMBINER_TYPE_DEFAULT )
    {
        // Black Buffer color
        // 0x0fd
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE(PICA_REG_TEX_ENV_BUFFER_COLOR),
            black);
        // White Constant colors of the 4th stage of the combiner.
        // 0x0f3
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE(PICA_REG_TEX_ENV4_COLOR),
            white);
    }
    else
#endif
    {
        // White Constant colors of the 3rd stage of the combiner.
        // 0x0db
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE(PICA_REG_TEX_ENV3_COLOR),
            white);
        // Black Constant colors of the 4th stage of the combiner.
        // 0x0f3
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE(PICA_REG_TEX_ENV4_COLOR),
            black);
    }

    return pCmdStream;
}
#endif

#if !defined(NW_FONT_ENABLE_EDGE_FONT)
void
CharWriter::UseCommandBuffer( RectDrawer* pRectDrawer )
{
    u32* pCmdStream = RectDrawer::GetCurrentCommandBuffer();
    pCmdStream = this->UseCommandBuffer(pCmdStream, pRectDrawer);
    RectDrawer::SetCurrentCommandBuffer(pCmdStream);
}
#else
void
CharWriter::UseCommandBuffer( FontCombinerType combiner, RectDrawer* pRectDrawer )
{
    u32* pCmdStream = RectDrawer::GetCurrentCommandBuffer();
    pCmdStream = this->UseCommandBuffer(pCmdStream, combiner, pRectDrawer);
    RectDrawer::SetCurrentCommandBuffer(pCmdStream);
}
#endif

#if !defined(NW_FONT_ENABLE_EDGE_FONT)
u32*
CharWriter::UseCommandBuffer( u32* pCmdStream, RectDrawer* pRectDrawer )
#else
u32*
CharWriter::UseCommandBuffer( u32* pCmdStream, FontCombinerType combiner, RectDrawer* pRectDrawer )
#endif
{
#if defined(NW_PLATFORM_CTR)
    NN_NULL_ASSERT(m_pDispStringBuffer);
    NN_NULL_ASSERT(m_pDispStringBuffer->commandBuffer);
    NN_ASSERT(m_pDispStringBuffer->IsGeneratedCommand());

    if (m_pDispStringBuffer->IsCommandEmpty())
    {
        return pCmdStream;
    }

    // Set vertex color.
    {
        u32 cmdBuf[VERTEX_MAX * 4 + 1];
        math::VEC4* dstCol = reinterpret_cast<math::VEC4*>(cmdBuf);
        SetVertexColor(dstCol, m_TextColors, GetGradationMode());
        cmdBuf[VERTEX_MAX * 4] = 0;
        pCmdStream = RectDrawer::AddTextVertexColorCommand(pCmdStream, cmdBuf);
    }

    // Sets the black and white interpolation colors.
    {
        const u32 alpha = m_Alpha;
        ut::Color8 white = m_ColorMapping.max;
        ut::Color8 black = m_ColorMapping.min;
        white.a = static_cast<u8>(white.a * alpha / 255);
        black.a = static_cast<u8>(black.a * alpha / 255);

#if defined(NW_FONT_ENABLE_EDGE_FONT)
        pCmdStream = RectDrawer::AddTextMappingColorCommand(pCmdStream, GetColorU32Raw(white), GetColorU32Raw(black), combiner);
#else
        pCmdStream = RectDrawer::AddTextMappingColorCommand(pCmdStream, GetColorU32Raw(white), GetColorU32Raw(black));
#endif
    }

    // Applies the content of the internal buffer to the command buffer the hardware references.
    if (m_pDispStringBuffer->GetCommandBufferMode() & DispStringBuffer::CBF_COMMAND_JUMP)
    {
        NN_NULL_ASSERT(pRectDrawer);

        pCmdStream = RectDrawer::BuildPaddingCommand(
            pCmdStream,
            (u32)pCmdStream % 16 + 24);

        uptr sizePtr;
        if (m_pDispStringBuffer->GetCommandBufferMode() & DispStringBuffer::CBF_VRAM)
        {
            NW_NULL_ASSERT(m_pDispStringBuffer->subCommandBuffer);
            pCmdStream = nn::gr::CTR::MakeChannel0SubroutineCommand(
                pCmdStream,
                &sizePtr,
                reinterpret_cast<uptr>(m_pDispStringBuffer->subCommandBuffer),
                m_pDispStringBuffer->commandBufferSize * sizeof(u32));
        }
        else
        {
            pCmdStream = nn::gr::CTR::MakeChannel0SubroutineCommand(
                pCmdStream,
                &sizePtr,
                reinterpret_cast<uptr>(m_pDispStringBuffer->commandBuffer),
                m_pDispStringBuffer->commandBufferSize * sizeof(u32));
        }

        NW_ASSERT((u32)pCmdStream % 16 == 0);

        if (pRectDrawer)
        {
            pRectDrawer->FixCommandJump(pCmdStream);
            pRectDrawer->SetCommandJumpInfo(pCmdStream, (u32*)sizePtr);
        }
    }
    else
    {
        NW_FONT_ADD_COMMANDS_PTR(
            pCmdStream,
            m_pDispStringBuffer->commandBuffer,
            m_pDispStringBuffer->commandBufferSize * sizeof(u32));
    }
#else
    NN_ASSERT(false);
#endif

    return pCmdStream;
}

u32*
RectDrawer::BuildPaddingCommand(u32* command, size_t size)
{
#if defined(NW_PLATFORM_CTR)
    if (size % 16)
    {
        int num = (size >> 2) % 4;
        for (int i = 4; i > num; --i)
        {
            *command++ = 0;
        }
    }

    return command;
#else
    NN_ASSERT(false);
    return NULL;
#endif
}

void
RectDrawer::BuildReturnCommand(DispStringBuffer* pDispStringBuffer)
{
#if defined(NW_PLATFORM_CTR)
    u32 size = pDispStringBuffer->commandBufferSize;
    u32* command = &pDispStringBuffer->commandBuffer[size];

    // Uses dummy commands for corrections so that the size of the command buffer is a multiple of 16.
    // (So that the kick command, which is added later, comes to 8 bytes.)
    command = RectDrawer::BuildPaddingCommand(command, size * sizeof(u32) + 8);

    // Inserts the kick command that returns from the sub-routine.
    command = nn::gr::CTR::MakeChannelKickCommand(command, nn::gr::CTR::CMD_BUF_CH1);
    size = command - pDispStringBuffer->commandBuffer;
    NW_ASSERT(size %4 == 0);
    NW_ASSERT(size <= pDispStringBuffer->commandBufferCapacity);
    pDispStringBuffer->commandBufferSize = size;

    // Flushes the cache.
    nngxUpdateBuffer(pDispStringBuffer->commandBuffer, size * sizeof(u32));
#else
    NN_ASSERT(false);
#endif
}

void
RectDrawer::SetCommandJumpInfo(void* command, void* sizePtr)
{
#if defined(NW_PLATFORM_CTR)
    // m_CommandJumpSizePtr will be reset by FixCommandJump.
    NW_ASSERTMSG(m_CommandJumpSizePtr == NULL, "FixCommandJump() required.");

    m_CommandJumpReturnAddr = command;
    m_CommandJumpSizePtr = sizePtr;

    if (!m_CommandJumpFirstReturnAddr)
    {
        m_CommandJumpFirstReturnAddr = m_CommandJumpReturnAddr;
    }
#else
    NN_ASSERT(false);
#endif
}

void
RectDrawer::FixCommandJump(void* command)
{
#if defined(NW_PLATFORM_CTR)
    if (m_CommandJumpSizePtr)
    {
        // Sets the command size to /8.
        *(u32*)m_CommandJumpSizePtr = ut::GetOffsetFromPtr(m_CommandJumpReturnAddr, command) >> 3;

        // Make this null so that "already fixed" is indicated.
        m_CommandJumpSizePtr = NULL;
    }
#else
    NN_ASSERT(false);
#endif
}

/* static */
u32*
RectDrawer::GetCurrentCommandBuffer()
{
#if defined(NW_PLATFORM_CTR)
    return static_cast<u32*>(__cb_current_command_buffer);
#else
    return NULL;
#endif
}

/* static */
void
RectDrawer::SetCurrentCommandBuffer(u32 *pCmdStream)
{
#if defined(NW_PLATFORM_CTR)
    __cb_current_command_buffer = static_cast<unsigned*>(pCmdStream);
#endif
}

void
DispStringBuffer::SendToVram() const
{
#if defined(NW_PLATFORM_CTR)
    NW_ASSERT(commandBufferMode == CBM_VRAM);
    NW_NULL_ASSERT(subCommandBuffer);
    NW_ASSERT(generatedCommand);

    if (subCommandBuffer)
    {
        nngxAddVramDmaCommand(commandBuffer, subCommandBuffer, commandBufferSize * sizeof(u32));
    }
#else
    NN_ASSERT(false);
#endif
}

}   // namespace font
}   // namespace nw

