/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nn/assert.h>
#include <nw/ut/ut_Inlines.h>
#include <nw/font/font_ResFont.h>

namespace nw {
namespace font {

namespace {

/*
  

  
  
 */
template <typename T>
inline void
ResolveOffset(T*& ptr, void* base)
{
    ptr = reinterpret_cast<T*>(
            reinterpret_cast<u8*>(base) + reinterpret_cast<u32>(ptr)
    );
}

template <typename T>
inline void
DeResolveOffset(T*& ptr, void* base)
{
    ptr = reinterpret_cast<T*>(reinterpret_cast<u32>(ptr) - reinterpret_cast<u32>(base));
}

}   // Namespace


void
ResFont::RevertResource(void* resource)
{
    ut::BinaryFileHeader* fileHeader =
        reinterpret_cast<ut::BinaryFileHeader*>(resource);
    NW_POINTER_ASSERT(fileHeader);
    NW_ALIGN_ASSERT(fileHeader, GlyphDataAlignment);

    // Immediately return if already initialized.
    if (fileHeader->signature != BINFILE_SIG_FONT_RESOLEVED) { return; }

    //----------------------------------------
    // Restore the offset value.
    ut::BinaryBlockHeader* blockHeader;
    FontInformation* info = NULL;
    int nBlocks = 0;

    blockHeader = reinterpret_cast<ut::BinaryBlockHeader*>(
        reinterpret_cast<u8*>(fileHeader) + fileHeader->headerSize
        );

    while (nBlocks < fileHeader->dataBlocks)
    {
        NW_POINTER_ASSERT(blockHeader);
        switch (blockHeader->kind)
        {
        //--------------------------------------------------
        // INFO block
        case BINBLOCK_SIG_FINF:
            // There must be 1 INFO block
            NW_ASSERT(info == NULL);
            {
                info = reinterpret_cast<FontInformation*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NW_ASSERT(info->fontType == FONT_TYPE_TEXTURE);
                NW_ASSERT(info->alterCharIndex != GLYPH_INDEX_NOT_FOUND);

                // pGlyph is required
                NW_NULL_ASSERT(info->pGlyph);
                DeResolveOffset(info->pGlyph, fileHeader);

                // OK if there is no pWidth and pMap
                if (info->pWidth != NULL)
                {
                    DeResolveOffset(info->pWidth, fileHeader);
                }
                if (info->pMap != NULL)
                {
                    DeResolveOffset(info->pMap, fileHeader);
                }
            }
            break;

        //--------------------------------------------------
        // TGLP block
        case BINBLOCK_SIG_TGLP:
            // There must also be one TGLP block, but having multiple blocks is not fatal
            {
                FontTextureGlyph* glyph = reinterpret_cast<FontTextureGlyph*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NW_NULL_ASSERT(glyph->sheetImage);
                DeResolveOffset(glyph->sheetImage, fileHeader);
            }
            break;

        //--------------------------------------------------
        // CWDH block
        case BINBLOCK_SIG_CWDH:
            {
                FontWidth* width = reinterpret_cast<FontWidth*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NW_ASSERT( width->indexBegin <= width->indexEnd );

                if (width->pNext != NULL)
                {
                    DeResolveOffset(width->pNext, fileHeader);
                }
            }
            break;

        //--------------------------------------------------
        // CMAP block
        case BINBLOCK_SIG_CMAP:
            {
                FontCodeMap* map = reinterpret_cast<FontCodeMap*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NW_ASSERT( map->ccodeBegin <= map->ccodeEnd );
                NW_ASSERT( (map->mappingMethod == FONT_MAPMETHOD_DIRECT)
                                || (map->mappingMethod == FONT_MAPMETHOD_TABLE)
                                || (map->mappingMethod == FONT_MAPMETHOD_SCAN) );

                if (map->pNext != NULL)
                {
                    DeResolveOffset(map->pNext, fileHeader);
                }
            }
            break;

        //--------------------------------------------------
        // GLGR Block
        case BINBLOCK_SIG_GLGR:
            {
                // ResFont does not reference the GLGR block, so nothing needs to be done.
            }
            break;

        //--------------------------------------------------
        // unknown
        default:
            NW_ASSERTMSG(false, "The font has unknown block('%c%c%c%c').",
                (blockHeader->kind >> 24) & 0xFF,
                (blockHeader->kind >> 16) & 0xFF,
                (blockHeader->kind >>  8) & 0xFF,
                (blockHeader->kind >>  0) & 0xFF
                );
            return;
        }

        blockHeader = reinterpret_cast<ut::BinaryBlockHeader*>(
            reinterpret_cast<u8*>(blockHeader) + blockHeader->size
        );
        nBlocks++;
    }

    //---- Change the signature and restore to original state.
    fileHeader->signature = BINFILE_SIG_FONT;
}

/* ------------------------------------------------------------------------
        Constructors/destructors
   ------------------------------------------------------------------------ */

ResFont::ResFont()
{
}

ResFont::~ResFont()
{
    if (!IsManaging(NULL))
    {
        RemoveResource();
    }
}



/* ------------------------------------------------------------------------
        Build
   ------------------------------------------------------------------------ */

u32
ResFont::GetDrawBufferSize(const void* bfnt)
{
    NN_POINTER_ASSERT(bfnt);
    NN_ALIGN_ASSERT(bfnt, GlyphDataAlignment);

    const FontTextureGlyph* pGlyph = NULL;
    const ut::BinaryFileHeader* fileHeader =
        reinterpret_cast<const ut::BinaryFileHeader*>(bfnt);

    //---- Is this font offset resolved?
    if (fileHeader->signature == BINFILE_SIG_FONT_RESOLEVED)
    {
    }
    else
    {
        //---- Has the correct file header?

        // Check if this is a supported version.
        if (! IsValidBinaryFile(fileHeader, BINFILE_SIG_FONT, FONT_FILE_VERSION, 2))
        {
            // Unsupported format.
            NN_LOG("Invalid font resource.");
            return 0;
        }
    }

    // Search FontInformation

    const ut::BinaryBlockHeader* blockHeader = reinterpret_cast<const ut::BinaryBlockHeader*>(
        reinterpret_cast<const u8*>(fileHeader) + fileHeader->headerSize
    );

    int nBlocks = 0;
    while (nBlocks < fileHeader->dataBlocks)
    {
        NN_POINTER_ASSERT( blockHeader );
        if (blockHeader->kind == BINBLOCK_SIG_TGLP)
        {
            pGlyph = reinterpret_cast<const FontTextureGlyph*>(
                reinterpret_cast<const u8*>(blockHeader) + sizeof(*blockHeader)
            );
            break;
        }

        blockHeader = reinterpret_cast<const ut::BinaryBlockHeader*>(
            reinterpret_cast<const u8*>(blockHeader) + blockHeader->size
        );
        nBlocks++;
    }

    if (pGlyph == NULL)
    {
        return 0;
    }

    return sizeof(internal::TextureObject) * pGlyph->sheetNum;
}

void*
ResFont::SetDrawBuffer(void* buffer)
{
    void *const prevBuffer = GetTextureObjectsBufferPtr();
    if (prevBuffer == buffer)
    {
        return buffer;
    }

    if (NULL != prevBuffer)
    {
        DeleteTextureNames();
    }

    SetTextureObjectsBufferPtr(buffer);

    if (NULL != buffer)
    {
        GenTextureNames();
    }

    return prevBuffer;
}

bool
ResFont::SetResource(void* bfnt)
{
    NN_POINTER_ASSERT(bfnt);
    NN_ALIGN_ASSERT(bfnt, GlyphDataAlignment);

    FontInformation* pFontInfo = NULL;
    ut::BinaryFileHeader* fileHeader =
        reinterpret_cast<ut::BinaryFileHeader*>(bfnt);

    //----Resource wasn't already set?
    if (! IsManaging(NULL))
    {
        NN_LOG("Font resource already atached.");
        return false;
    }

    //---- Is this font offset resolved?
    if (fileHeader->signature == BINFILE_SIG_FONT_RESOLEVED)
    {
        // Search FontInformation
        pFontInfo = static_cast<nw::font::FontInformation *>(FindBlock(fileHeader, BINBLOCK_SIG_FINF));
    }
    else
    {
        //---- Has the correct file header?

        // Check if this is a supported version.
        if (! IsValidBinaryFile(fileHeader, BINFILE_SIG_FONT, FONT_FILE_VERSION, 2))
        {
            // Unsupported format.
            NN_LOG("Invalid font resource.");
            return false;
        }

        // Rebuild
        pFontInfo = Rebuild(fileHeader);
    }

    if (pFontInfo == NULL)
    {
        return false;
    }

    SetResourceBuffer(bfnt, pFontInfo);

    return true;
}

void* ResFont::FindBlock(ut::BinaryFileHeader* fileHeader, ut::SigWord sigword)
{
    ut::BinaryBlockHeader* blockHeader;
    int nBlocks = 0;

    blockHeader = reinterpret_cast<ut::BinaryBlockHeader*>(
        reinterpret_cast<u8*>(fileHeader) + fileHeader->headerSize
    );

    while (nBlocks < fileHeader->dataBlocks)
    {
        NW_ASSERT_VALID_POINTER( blockHeader );
        if (blockHeader->kind == sigword)
        {
            return reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader);
        }

        blockHeader = reinterpret_cast<ut::BinaryBlockHeader*>(
            reinterpret_cast<u8*>(blockHeader) + blockHeader->size
        );
        nBlocks++;
    }

    return NULL;
}

void*
ResFont::RemoveResource()
{
    if (IsManaging(NULL))
    {
        NN_LOG("ResFont::RemoveResource(): Res font is not loaded.\n");
        return 0;
    }

    return RemoveResourceBuffer();
}

FontInformation*
ResFont::Rebuild(ut::BinaryFileHeader* fileHeader)
{
    NN_POINTER_ASSERT(fileHeader);
    NN_ALIGN_ASSERT(fileHeader, GlyphDataAlignment);

    ut::BinaryBlockHeader* blockHeader;
    FontInformation* info = NULL;
    int nBlocks = 0;

    blockHeader = reinterpret_cast<ut::BinaryBlockHeader*>(
        reinterpret_cast<u8*>(fileHeader) + fileHeader->headerSize
    );

    while (nBlocks < fileHeader->dataBlocks)
    {
        NN_POINTER_ASSERT(blockHeader);
        switch (blockHeader->kind)
        {
        //--------------------------------------------------
        // INFO block
        case BINBLOCK_SIG_FINF:
            // There must be 1 INFO block
            NN_ASSERT(info == NULL);
            {
                info = reinterpret_cast<FontInformation*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NN_ASSERT(info->fontType == FONT_TYPE_TEXTURE);
                NN_ASSERT(info->alterCharIndex != GLYPH_INDEX_NOT_FOUND);

                // pGlyph is required
                NN_NULL_ASSERT(info->pGlyph);
                ResolveOffset(info->pGlyph, fileHeader);
                NN_POINTER_ASSERT(info->pGlyph);

                // OK if there is no pWidth and pMap
                if (info->pWidth != NULL)
                {
                    ResolveOffset(info->pWidth, fileHeader);
                    NN_POINTER_ASSERT(info->pWidth);
                }
                if (info->pMap != NULL)
                {
                    ResolveOffset(info->pMap, fileHeader);
                    NN_POINTER_ASSERT(info->pMap);
                }
            }
            break;

        //--------------------------------------------------
        // TGLP block
        case BINBLOCK_SIG_TGLP:
            // There must also be one TGLP block, but having multiple blocks is not fatal
            {
                FontTextureGlyph* glyph = reinterpret_cast<FontTextureGlyph*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NN_NULL_ASSERT(glyph->sheetImage);
                ResolveOffset(glyph->sheetImage, fileHeader);
                NN_POINTER_ASSERT(glyph->sheetImage);

                //---- 32x32 I4 is the smallest, 1024x1024 RGBA8 is the largest
                NW_FONT_MIN_ASSERT       ( glyph->cellWidth,     1 );
                NW_FONT_MIN_ASSERT       ( glyph->cellHeight,    1 );
                NW_FONT_MINMAX_ASSERT    ( glyph->sheetSize,     32 * 32 / 2,    1024 * 1024 * 4 );
                NW_FONT_MIN_ASSERT       ( glyph->sheetNum,      1 );
                NW_FONT_MIN_ASSERT       ( glyph->sheetRow,      1 );
                NW_FONT_MIN_ASSERT       ( glyph->sheetLine,     1 );
                NW_FONT_MINMAX_ASSERT    ( glyph->sheetWidth,    32,             1024 );
                NW_FONT_MINMAX_ASSERT    ( glyph->sheetHeight,   32,             1024 );
            }
            break;

        //--------------------------------------------------
        // CWDH block
        case BINBLOCK_SIG_CWDH:
            {
                FontWidth* width = reinterpret_cast<FontWidth*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NN_ASSERT( width->indexBegin <= width->indexEnd );

                if (width->pNext != NULL)
                {
                    ResolveOffset(width->pNext, fileHeader);
                    NN_POINTER_ASSERT( width->pNext );
                }
            }
            break;

        //--------------------------------------------------
        // CMAP block
        case BINBLOCK_SIG_CMAP:
            {
                FontCodeMap* map = reinterpret_cast<FontCodeMap*>(
                    reinterpret_cast<u8*>(blockHeader) + sizeof(*blockHeader)
                );

                NN_ASSERT( map->ccodeBegin <= map->ccodeEnd );
                NN_ASSERT( (map->mappingMethod == FONT_MAPMETHOD_DIRECT)
                                || (map->mappingMethod == FONT_MAPMETHOD_TABLE)
                                || (map->mappingMethod == FONT_MAPMETHOD_SCAN) );

                if (map->pNext != NULL)
                {
                    ResolveOffset(map->pNext, fileHeader);
                    NN_POINTER_ASSERT(map->pNext);
                }
            }
            break;

        //--------------------------------------------------
        // GLGR Block
        case BINBLOCK_SIG_GLGR:
            {
                // ResFont does not reference the GLGR block, so nothing needs to be done.
            }
            break;

        //--------------------------------------------------
        // unknown
        default:
            NN_ASSERTMSG(false, "The font has unknown block('%c%c%c%c').",
                (blockHeader->kind >> 24) & 0xFF,
                (blockHeader->kind >> 16) & 0xFF,
                (blockHeader->kind >>  8) & 0xFF,
                (blockHeader->kind >>  0) & 0xFF
                );
            return NULL;
        }

        blockHeader = reinterpret_cast<ut::BinaryBlockHeader*>(
            reinterpret_cast<u8*>(blockHeader) + blockHeader->size
        );
        nBlocks++;
    }


    //---- Signature is changed and marked as offset-resolved.
    fileHeader->signature = BINFILE_SIG_FONT_RESOLEVED;

    return info;
}

}   // namespace font
}   // namespace nw
