﻿/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <nn.h>
#include <nw/font/font_TextWriterResource.h>
#include <nw/font/font_DispStringBuffer.h>
#include <nw/font/font_RectDrawer.h>

#include <nw/assert.h>  // NW_GL_ASSERT用

#define NW_FONT_TEXENV3   "dmp_TexEnv[3]"
#define NW_FONT_TEXENV4   "dmp_TexEnv[4]"
#define NW_FONT_TEXENV5   "dmp_TexEnv[5]"

namespace nw {
namespace font {

namespace
{

int
GetUniformLocation(
    GLuint      programId,
    const char* name
)
{
    int result = glGetUniformLocation(programId, name);
    NN_ASSERT(result != -1);
    return result;
}

}   // namespace

TextWriterResource::TextWriterResource()
:   m_PosZ(0.f),
    m_IsInitialized(false)
{
    ResetLoadingTexture();          // m_LoadingTexture
}

TextWriterResource::~TextWriterResource()
{
    Finalize();
}

void
TextWriterResource::InitResource(
    const void* shaderBinary,
    u32         size
)
{
    NN_ASSERT(! m_IsInitialized);

    m_ProgramId = glCreateProgram();
    NN_ASSERT(m_ProgramId != 0);

    GLuint shaderHandle = glCreateShader(GL_VERTEX_SHADER);
    NN_ASSERT(shaderHandle != 0);
    glShaderBinary(1, &shaderHandle, GL_PLATFORM_BINARY_DMP, shaderBinary, size);

    glAttachShader(m_ProgramId, shaderHandle);

    glDeleteShader(shaderHandle);
    glAttachShader(m_ProgramId, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(m_ProgramId, internal::VERTEX_ATTR_POS,      "aPosition");
    glBindAttribLocation(m_ProgramId, internal::VERTEX_ATTR_POS_Z,    "aPositionZ");
    glBindAttribLocation(m_ProgramId, internal::VERTEX_ATTR_COLOR,    "aColor");
    glBindAttribLocation(m_ProgramId, internal::VERTEX_ATTR_TEXCOORD, "aTexCoord");

    glLinkProgram(m_ProgramId);

    glGenTextures(1, &m_TextureId);

    NW_GL_ASSERT();

    m_UniformLocations[internal::LOC_PROJECTION] =
        GetUniformLocation(m_ProgramId, "uProjection");
    m_UniformLocations[internal::LOC_MODELVIEW] =
        GetUniformLocation(m_ProgramId, "uModelView");
    m_UniformLocations[internal::LOC_FRAGMENTLIGHTING_ENABLED] =
        GetUniformLocation(m_ProgramId, "dmp_FragmentLighting.enabled"),
    m_UniformLocations[internal::LOC_TEXTURE0_SAMPLERTYPE] =
        GetUniformLocation(m_ProgramId, "dmp_Texture[0].samplerType"),
    m_UniformLocations[internal::LOC_FOG_MODE] =
        GetUniformLocation(m_ProgramId, "dmp_Fog.mode");
    m_UniformLocations[internal::LOC_FRAGOPERATION_ENABLEALPHATEST] =
        GetUniformLocation(m_ProgramId, "dmp_FragOperation.enableAlphaTest");
    m_UniformLocations[internal::LOC_FRAGOPERATION_ALPHAREFVALUE] =
        GetUniformLocation(m_ProgramId, "dmp_FragOperation.alphaRefValue");
    m_UniformLocations[internal::LOC_FRAGOPERATION_ALPHATESTFUNC] =
        GetUniformLocation(m_ProgramId, "dmp_FragOperation.alphaTestFunc");

    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_SRCRGB      ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".srcRgb"      );
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_SRCALPHA    ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".srcAlpha"    );
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_OPERANDRGB  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".operandRgb"  );
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_OPERANDALPHA] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".operandAlpha");
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_COMBINERGB  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".combineRgb"  );
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_COMBINEALPHA] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".combineAlpha");
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_SCALERGB    ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".scaleRgb"    );
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_SCALEALPHA  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".scaleAlpha"  );
    m_TexEnvUniformLocations[internal::TEXENV_3][internal::TCLOC_CONSTRGBA   ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV3 ".constRgba"   );

    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_SRCRGB      ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".srcRgb"      );
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_SRCALPHA    ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".srcAlpha"    );
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_OPERANDRGB  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".operandRgb"  );
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_OPERANDALPHA] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".operandAlpha");
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_COMBINERGB  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".combineRgb"  );
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_COMBINEALPHA] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".combineAlpha");
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_SCALERGB    ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".scaleRgb"    );
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_SCALEALPHA  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".scaleAlpha"  );
    m_TexEnvUniformLocations[internal::TEXENV_4][internal::TCLOC_CONSTRGBA   ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV4 ".constRgba"   );

    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_SRCRGB      ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".srcRgb"      );
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_SRCALPHA    ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".srcAlpha"    );
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_OPERANDRGB  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".operandRgb"  );
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_OPERANDALPHA] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".operandAlpha");
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_COMBINERGB  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".combineRgb"  );
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_COMBINEALPHA] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".combineAlpha");
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_SCALERGB    ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".scaleRgb"    );
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_SCALEALPHA  ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".scaleAlpha"  );
    m_TexEnvUniformLocations[internal::TEXENV_5][internal::TCLOC_CONSTRGBA   ] =
        GetUniformLocation(m_ProgramId, NW_FONT_TEXENV5 ".constRgba"   );

    m_IsInitialized = true;
}

void
TextWriterResource::FinalizeGX()
{
    for (int i = 0; i < internal::VERTEX_ATTR_NUM; ++i)
    {
        glDisableVertexAttribArray(i);
    }
}

void
TextWriterResource::Finalize()
{
    if (m_IsInitialized)
    {
        FinalizeGX();
        glDeleteTextures(1, &m_TextureId);
        m_IsInitialized = false;

        glUseProgram(0);
        glDeleteProgram(m_ProgramId);

        NN_LOG("nw::font::TextWriterResource: delete buffer\n");
    }
}

void
TextWriterResource::SetProjectionMtx(const nn::math::MTX44& mtx) const
{
    NN_ASSERT(m_IsInitialized);

    glUniform4fv(m_UniformLocations[internal::LOC_PROJECTION], 4, mtx.a);
}

void
TextWriterResource::SetViewMtx(const nn::math::MTX34& mtx) const
{
    NN_ASSERT(m_IsInitialized);

    glUniform4fv(m_UniformLocations[internal::LOC_MODELVIEW], 3, mtx.a);
}

void
TextWriterResource::SetPosZ(f32 posZ)
{
    m_PosZ = posZ;
    glVertexAttrib1f(internal::VERTEX_ATTR_POS_Z, m_PosZ);
}

}   // namespace font
}   // namespace nw
