/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>

#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>
#include <nw/anim.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;
nw::demo::DemoAllocator s_ParticleAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

const wchar_t* RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"Mobile.bcres"),
    NW_DEMO_FILE_PATH(L"Mobile.bcskla"),
    NW_DEMO_FILE_PATH(L"SceneEnvironmentSetting.bcenv"),
    NW_DEMO_FILE_PATH(L"Light.bclgt"),
};

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;
nw::gfx::ParticleContext* s_ParticleContext = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
const int SCENE_NODE_COUNT = 4;
nw::gfx::SceneNode* s_SceneRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;

const s32 s_BaseCameraIndex = 0;

bool s_SceneDone = false;

//----------------------------------------
// Scene environment-related
const s32 ENVIRONMENT_SETTINGS_COUNT = 1;

typedef nw::ut::FixedSizeArray<nw::gfx::SceneEnvironmentSetting*, ENVIRONMENT_SETTINGS_COUNT> SceneEnvironmentSettingArray;
SceneEnvironmentSettingArray s_SceneEnvironmentSettings;

//----------------------------------------
// Animation-related
const int MAX_ANIM_OBJECTS = 2;
nw::ut::FixedSizeArray<nw::gfx::AnimObject*, MAX_ANIM_OBJECTS> s_AnimObjects;

const int ANIMATION_RESOURCES_COUNT = 4;
const int ANIMATION_TARGETS_COUNT = 8;

nw::ut::FixedSizeArray<nw::gfx::ResGraphicsFile, ANIMATION_RESOURCES_COUNT> s_ResAnims;
nw::ut::FixedSizeArray<nw::gfx::SceneObject*, ANIMATION_TARGETS_COUNT> s_AnimTargets;

nw::gfx::BaseAnimEvaluator* s_MainEvaluator = NULL;

struct AnimBindingTable
{
    const char* targetName;
    const char* animName;
    nw::demo::Utility::AnimationType animType;
    u8 padding[3];
};

const AnimBindingTable ANIM_BINDING_TABLE[] =
{
    { "Toy_Mobile", "Toy_Mobile", nw::demo::Utility::SKELETAL_ANIMATION },
    { "pointLight1", "pointLight1", nw::demo::Utility::LIGHT_ANIMATION }
};

//----------------------------------------
// Particle-related
nw::gfx::ParticleSceneUpdater* s_ParticleSceneUpdater = NULL;
const int EMITTER_COUNT = 3;

typedef nw::ut::FixedSizeArray<nw::gfx::ParticleModel*, EMITTER_COUNT> ParticleMdlArray;
ParticleMdlArray s_ParticleModels;


//----------------------------------------
// Constrain-related
nw::gfx::SkeletalModel* s_SkeletalModel = NULL;
int s_BoneIdx = -1;

typedef nw::ut::FixedSizeArray<nw::gfx::ParticleEmitter*, EMITTER_COUNT> EmitterArray;
EmitterArray s_Emitters;

nw::gfx::Light* s_Light = NULL;

/*
  

  
 */
class CallbackConstraint : public nw::gfx::Skeleton::CalculateMatrixSlot
{
public:
    //
    //
    //
    //
    //
    //
    //
    CallbackConstraint(nw::os::IAllocator* allocator)
    : nw::gfx::Skeleton::CalculateMatrixSlot(allocator),
      m_ConstrainedObjects(allocator),
      m_Skeleton(NULL),
      m_BoneIdx(0)
    {}

    //
    //
    //
    //
    //
    //
    //
    //
    CallbackConstraint(size_t maxConstrainedCount, nw::os::IAllocator* allocator)
    : nw::gfx::Skeleton::CalculateMatrixSlot(allocator),
      m_ConstrainedObjects(maxConstrainedCount, allocator),
      m_Skeleton(NULL),
      m_BoneIdx(0)
    {}

    //
    //
    //
    //
    //
    bool AddConstrainedObject(nw::gfx::TransformNode* constrainedObject)
    {
        constrainedObject->Transform().DisableFlags(
            nw::gfx::CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED);
        return m_ConstrainedObjects.push_back(constrainedObject);
    }

    //
    //
    //
    //
    //
    //
    bool SetTargetObject(
        nw::gfx::SkeletalModel* skeletalModel,
        const char* boneName
    )
    {
        nw::gfx::Skeleton* skeleton = skeletalModel->GetSkeleton();
        int boneIdx = skeleton->GetResSkeleton().GetBonesIndex(boneName);
        if (boneIdx < 0)
        {
            return false;
        }

        m_Skeleton = skeleton;
        m_BoneIdx = boneIdx;

        return true;
    }

    //
    virtual void Invoke(nw::gfx::Skeleton* skeleton, int boneIdx)
    {
        if (skeleton != m_Skeleton || boneIdx != m_BoneIdx)
        {
            return;
        }

        nw::gfx::Skeleton::MatrixPose& pose = skeleton->WorldMatrixPose();

        TransformNodeArray::iterator end = m_ConstrainedObjects.end();
        for (TransformNodeArray::iterator iter = m_ConstrainedObjects.begin() ; iter != end ; ++iter)
        {
            (*iter)->WorldMatrix() = *pose.GetMatrix(boneIdx);

            // Since the world matrix was changed, invalidate the reverse matrix's cache.
            (*iter)->InvalidateInverseWorldMatrix();
        }
    }

protected:
    typedef nw::ut::MoveArray<nw::gfx::TransformNode*> TransformNodeArray;
    TransformNodeArray m_ConstrainedObjects;
    nw::gfx::Skeleton* m_Skeleton;
    int m_BoneIdx;
};


/*
  
                
 */
class CallbackResourceBasedConstraint : public CallbackConstraint
{
public:
    //
    //
    //
    //
    //
    //
    //
    CallbackResourceBasedConstraint(nw::os::IAllocator* allocator)
    : CallbackConstraint(allocator)
    {}

    //
    //
    //
    //
    //
    //
    //
    //
    CallbackResourceBasedConstraint(size_t maxConstrainedCount, nw::os::IAllocator* allocator)
    : CallbackConstraint(maxConstrainedCount, allocator)
    {}

    //
    //
    //
    //
    //
    //
    //
    bool AddConstrainedObject(nw::gfx::TransformNode* constrainedObject)
    {
        return m_ConstrainedObjects.push_back(constrainedObject);
    }

    //
    virtual void Invoke(nw::gfx::Skeleton* skeleton, int boneIdx)
    {
        if (skeleton != m_Skeleton || boneIdx != m_BoneIdx)
        {
            return;
        }

        nw::gfx::Skeleton::MatrixPose& pose = skeleton->WorldMatrixPose();

        TransformNodeArray::iterator end = m_ConstrainedObjects.end();
        for (TransformNodeArray::iterator iter = m_ConstrainedObjects.begin() ; iter != end ; ++iter)
        {
            (*iter)->SetResourceBasedTransform(*pose.GetMatrix(boneIdx));
        }
    }
};

CallbackConstraint* s_ConstraintEmitter;
CallbackResourceBasedConstraint* s_ResourceBasedConstraintEmitter;

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    s_ParticleContext = nw::gfx::ParticleContext::Builder()
        .Create(&s_DeviceAllocator);

    s_ParticleSceneUpdater = nw::gfx::ParticleSceneUpdater::Builder()
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_ParticleSceneUpdater);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroy(s_ParticleSceneUpdater);

    nw::gfx::SafeDestroy(s_ParticleContext);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateAnim(void)
{
    for (int animIdx = 0; animIdx < s_AnimObjects.Size(); ++animIdx)
    {
        nw::gfx::SafeDestroy(s_AnimObjects[animIdx]);
    }
    s_AnimObjects.clear();
}


/*
  

  
 */
void
SetStepFrame(f32 stepFrame)
{
    nw::gfx::SceneContext* sceneContext = s_SceneSystem->GetSceneContext();

    nw::gfx::SceneNodeArray::iterator end = sceneContext->GetSceneNodesEnd();
    for (nw::gfx::SceneNodeArray::iterator i = sceneContext->GetSceneNodesBegin(); i != end; ++i)
    {
        nw::gfx::ParticleEmitter* emitter = nw::ut::DynamicCast<nw::gfx::ParticleEmitter*>(*i);
        if (emitter != NULL)
        {
            emitter->ParticleAnimFrameController().SetStepFrame(stepFrame);
        }
        else
        {
            nw::gfx::ParticleModel* model = nw::ut::DynamicCast<nw::gfx::ParticleModel*>(*i);
            if (model != NULL)
            {
                model->ParticleAnimFrameController().SetStepFrame(stepFrame);
            }
        }
    }
}

/*
  
 */
void
ResetParticle()
{
    // Reset the particles.
    if ( s_SceneDone )
    {
        nw::gfx::SceneContext* sceneContext = s_SceneSystem->GetSceneContext();

        // Performs the particle model's reset.
        // You cannot call ParticleSetsClear during a GPU process.
        {
            ParticleMdlArray::iterator end = s_ParticleModels.end();
            for (ParticleMdlArray::iterator iter = s_ParticleModels.begin() ; iter != end ; ++iter)
            {
                s_SceneRoot->AttachChild(*iter);

                (*iter)->ForeachParticleSet(nw::gfx::ParticleSetsClear());
                (*iter)->ParticleAnimFrameController().SetFrame(0);
            }

            s_ParticleModels.clear();
        }

        // Performs the particle emitter's reset.
        {
            EmitterArray::iterator end = s_Emitters.end();
            for (EmitterArray::iterator iter = s_Emitters.begin() ; iter != end ; ++iter)
            {
                (*iter)->Reset();
            }
        }

        // Reset the animation.
        for (int animIdx = 0; animIdx < s_AnimObjects.Size(); ++animIdx)
        {
            nw::gfx::BaseAnimEvaluator* evaluator = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(s_AnimObjects[animIdx]);
            if (evaluator)
            {
                evaluator->SetFrame(0);
            }
        }

        s_SceneSystem->InitializeScene(s_SceneRoot);
        s_SceneDone = false;
    }
}


/*
  
 */
void
LoopAnimation()
{
    // When the animation has arrived at the end frame, detach ParticleModel.
    if (s_MainEvaluator->GetFrame() >= s_MainEvaluator->GetEndFrame())
    {
        nw::gfx::SceneContext* sceneContext = s_SceneSystem->GetSceneContext();

        // Detach ParticleModel.
        // Since ParticleSetsClear cannot be called during a GPU process, after the GPU process ends, attach ParticleModel and perform the ParticleSetsClear call process.
        // 
        nw::gfx::ParticleModelArray::iterator end = sceneContext->GetParticleModelEnd();
        for (nw::gfx::ParticleModelArray::iterator iter = sceneContext->GetParticleModelBegin();
            iter != end; ++iter)
        {
            bool isPushed = s_ParticleModels.push_back(*iter);
            NW_ASSERT(isPushed);
            s_SceneRoot->DetachChild(*iter);
        }

        s_SceneSystem->InitializeScene(s_SceneRoot);
        s_SceneDone = true;
    }
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);

}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(15.0f, 10.0f, 15.0f),
        nw::math::VEC3(0.0f, 0.0f, 0.0f)
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}

/*
  
 */
void
BuildResources(nw::demo::ResourceSet* resourceSet)
{
    resourceSet->resource.ForeachTexture(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachIndexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachVertexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));

    nw::gfx::Result result = resourceSet->resource.Setup(&s_DeviceAllocator);
    if (result.IsFailure())
    {
        NW_FATAL_ERROR("Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(&s_DeviceAllocator);

    nw::gfx::ResModelArray models = resourceSet->resource.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*modelResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);

        s_AnimTargets.push_back(node);

        nw::gfx::SkeletalModel* skeletalModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);
        if (skeletalModel)
        {
            s_SkeletalModel = skeletalModel;
        }
    }

    nw::gfx::ResLightArray lights = resourceSet->resource.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*lightResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);

        s_AnimTargets.push_back(node);

        s_Light = nw::ut::DynamicCast<nw::gfx::Light*>(node);
    }

    nw::gfx::ResEmitterArray emitters = resourceSet->resource.GetEmitters();
    for (nw::gfx::ResEmitterArray::iterator emitterResource = emitters.begin();
        emitterResource != emitters.end(); ++emitterResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*emitterResource)
            );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);

        s_Emitters.push_back(nw::ut::DynamicCast<nw::gfx::ParticleEmitter*>(node));
    }

    nw::gfx::ResCameraArray cameras = resourceSet->resource.GetCameras();
    nw::gfx::ResCameraArray::iterator camerasEnd = cameras.end();
    for (nw::gfx::ResCameraArray::iterator cameraResource = cameras.begin();
         cameraResource != camerasEnd; ++cameraResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*cameraResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);
    }

    // Resolve parent-child reference relationships
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    nw::gfx::ParticleUtil::SetupParticleObject(&sceneNodeArray, s_ParticleContext);

    // Add models and emitters to the scene.
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(s_SceneRoot)
    );

    nw::gfx::ResSceneEnvironmentSettingArray settings = resourceSet->resource.GetSceneEnvironmentSettings();
    nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
    for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
        settingResource != settingsEnd; ++settingResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource(*settingResource)
            .CreateObject(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
            nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

        NW_NULL_ASSERT(sceneEnvironmentSetting);
        s_SceneEnvironmentSettings.push_back(sceneEnvironmentSetting);
    }
}

/*
  
 */
void
BuildAnimations()
{
    NW_FOREACH(AnimBindingTable animBindingTable, ANIM_BINDING_TABLE)
    {
        s_AnimObjects.PushBack(
            nw::demo::Utility::BindAnimationByName(
                &s_DeviceAllocator,
                s_AnimTargets,
                s_ResAnims,
                animBindingTable.targetName,
                animBindingTable.animName,
                animBindingTable.animType,
                false)
            );
    }

    // Set the first registered animation as a base for loop playback.
    NW_ASSERT(!s_AnimObjects.empty());
    s_MainEvaluator = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(s_AnimObjects[0]);
    NW_NULL_ASSERT(s_MainEvaluator);
}

/*
  
                
 */
void
SetupCallbackConstraint()
{
    void* slotMemory = s_DeviceAllocator.Alloc(sizeof(CallbackConstraint));
    s_ConstraintEmitter =
        new(slotMemory) CallbackConstraint(s_Emitters.size(), &s_DeviceAllocator);

    EmitterArray::iterator end = s_Emitters.end();
    for (EmitterArray::iterator iter = s_Emitters.begin() ; iter != end ; ++iter)
    {
        if ( ::std::strcmp( (*iter)->GetName(), "LineEffectEmitter") == 0 )
        {
            // LineEffectEmitter sets the constrain in the SetupCallbackResourceBasedConstraint function so it is excluded here.
            // 
            continue;
        }

        // Add restricted emitters to the function object.
        s_ConstraintEmitter->AddConstrainedObject(*iter);
    }

    // Set the RocketNozzle bone to be restricted.
    bool result = s_ConstraintEmitter->SetTargetObject(s_SkeletalModel, "RocketNozzle");
    NW_ASSERT(result);

    // Connect the function object to the signal.
    s_SkeletalModel->GetSkeleton()->PostCalculateMatrixSignal().Connect(s_ConstraintEmitter);
}

/*
  
                

                
 */
void
SetupCallbackResourceBasedConstraint()
{
    void* slotMemory = s_DeviceAllocator.Alloc(sizeof(CallbackResourceBasedConstraint));
    s_ResourceBasedConstraintEmitter =
        new(slotMemory) CallbackResourceBasedConstraint(1, &s_DeviceAllocator);

    EmitterArray::iterator end = s_Emitters.end();
    for (EmitterArray::iterator iter = s_Emitters.begin() ; iter != end ; ++iter)
    {
        if ( ::std::strcmp( (*iter)->GetName(), "LineEffectEmitter") == 0 )
        {
            // Add restricted emitters to the function object.
            s_ResourceBasedConstraintEmitter->AddConstrainedObject(*iter);
        }
    }

    // Set the Rocket bone to be restricted.
    bool result = s_ResourceBasedConstraintEmitter->SetTargetObject(s_SkeletalModel, "Rocket");
    NW_ASSERT(result);

    // Connect the function object to the signal.
    s_SkeletalModel->GetSkeleton()->PostCalculateMatrixSignal().Connect(s_ResourceBasedConstraintEmitter);
}

/*
  
                
 */
void
SetupConstraint()
{
    // Get the index of the restricted bone.
    s_BoneIdx = s_SkeletalModel->GetSkeleton()->GetResSkeleton().GetBonesIndex("RocketNozzle");

    // Invalidate the calculation of the light's world matrix.
    s_Light->Transform().DisableFlags(nw::gfx::CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED);
}

/*
  
                
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    NW_FOREACH(const wchar_t* name, RESOURCE_FILES)
    {
        nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, name, &s_DeviceAllocator);
        BuildResources(resourceSet);
        s_ResAnims.push_back(resourceSet->resource);
    }

    NW_NULL_ASSERT(s_SkeletalModel);
    NW_NULL_ASSERT(s_Light);

    BuildAnimations();

    // Set the constrain that used the callback.
    SetupCallbackConstraint();

    // Set the constrain the used the callback and consider the resource's transform.
    SetupCallbackResourceBasedConstraint();

    // Prepare the constrain that occurs without using the callback.
    SetupConstraint();

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Resolves and sets references for the scene environment.
    s_RenderSystem->SetSceneEnvironmentSettings(s_SceneSystem, &s_SceneEnvironmentSettings);
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);
    nw::ut::SafeDestroyAll(s_SceneEnvironmentSettings);

    TerminateAnim();

    NW_GL_ASSERT();

    s_ResAnims.clear();
    s_AnimTargets.clear();

    s_Resources.clear();
    s_Emitters.clear();
    s_SceneEnvironmentSettings.clear();

    s_SceneRoot = NULL;
    s_SkeletalModel = NULL;
    s_Light = NULL;
}

/*
  
 */
void
ConstrainLight()
{
    nw::gfx::Skeleton::MatrixPose& pose = s_SkeletalModel->GetSkeleton()->WorldMatrixPose();
    s_Light->WorldMatrix() = *pose.GetMatrix(s_BoneIdx);

    // Since the world matrix was changed, invalidate the reverse matrix's cache.
    s_Light->InvalidateInverseWorldMatrix();
}

/*
  
 */
void
UpdateScene()
{
    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    ConstrainLight();

    s_ParticleSceneUpdater->UpdateNode(
        s_SceneSystem->GetSceneContext(),
        s_ParticleContext);

    nw::gfx::ResCameraProjectionUpdater resProjectionUpdater =
        s_BaseCamera->GetProjectionUpdater()->GetResource();

    int near = resProjectionUpdater.GetNear();

    s_RenderSystem->CalcStereoCamera(
        s_LeftCamera,
        s_RightCamera,
        s_BaseCamera,
        near + 5.0f);

    ++s_FrameCount;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}

/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();

        LoopAnimation();

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();
        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        // After the GPU process ends, perform the particle reset process.
        ResetParticle();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);
    nw::demo::InitializeDemoAllocator(&s_ParticleAllocator, nw::demo::DEMO_PARTICLE_MEMORY_SIZE);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, &s_ParticleAllocator, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeDemoAllocator(&s_ParticleAllocator);

    nw::demo::FinalizeGraphicsSystem();
}
