/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>

#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>
#include <nw/anim.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

const wchar_t* MODEL_RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"Male.bcmdl"),
    NW_DEMO_FILE_PATH(L"SceneEnvironmentSetting.bcenv"),
    NW_DEMO_FILE_PATH(L"FragmentLight.bcenv"),
};

const wchar_t* SKELETAL_ANIM_RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"Walk.bcskla"),
};

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
const int SCENE_NODE_COUNT = 4;
nw::gfx::SceneNode* s_SceneRoot = NULL;
nw::gfx::SceneNode* s_ModelRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
const f32 s_fNearPlane = 0.1f;

//----------------------------------------
// Scene environment-related
const s32 ENVIRONMENT_SETTINGS_COUNT = 1;

typedef nw::ut::FixedSizeArray<nw::gfx::SceneEnvironmentSetting*, ENVIRONMENT_SETTINGS_COUNT> SceneEnvironmentSettingArray;
SceneEnvironmentSettingArray s_SceneEnvironmentSettings;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Animation-related
float s_AnimWeight0 = 0.5f;
float s_AnimWeight1 = 0.5f;

const int MAX_ANIM_MODELS = 80;
struct AnimModelSet {
    nw::gfx::SkeletalModel* skeletalModel;
    nw::anim::ResAnim resAnim;
    nw::gfx::TransformAnimEvaluator* evaluator;
    bool isSkeletonShared;
    u8 padding[3];
};
typedef nw::ut::FixedSizeArray<AnimModelSet, MAX_ANIM_MODELS> AnimModelSetArray;
AnimModelSetArray s_AnimModelSets;
AnimModelSet s_AnimModelSetTemplate;

u32 s_NumCopy = 0;
u32 s_NumSkeletonSharedCopy = 0;

//----------------------------------------
// Operations-related
s32 s_WhirlDepth;
s32 s_WhirlPos;
/*
  
 */
enum DuplicationMode
{
    DUPLICATION,            //
    SKELETON_SHARED         //
};

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    nw::demo::SceneSystem::Description sceneDescription;
    sceneDescription.maxSceneNodes = MAX_ANIM_MODELS + 10;
    sceneDescription.maxSkeletalModels = MAX_ANIM_MODELS + 1;
    sceneDescription.maxModels = MAX_ANIM_MODELS + 1;
    sceneDescription.isFixedSizeMemory = false;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  

  
  
  

  
 */
nw::gfx::TransformAnimEvaluator*
CreateTransformAnimEvaluator(
    const int maxMembers,
    const bool translateAnimEnabled,
    nw::anim::ResAnim resAnim
)
{
    if (!resAnim.IsValid())
    {
        return NULL;
    }

    //----------------------------------------
    // Creates a transform animation evaluator.
    //
    // When an animation is application to only one model, and for frame-format data, setting AllocCache to FALSE will lower the processing load.
    // 
    nw::gfx::TransformAnimEvaluator* evaluator = nw::gfx::TransformAnimEvaluator::Builder()
        .AnimData(resAnim)
        .MaxMembers(maxMembers)
        .MaxAnimMembers(resAnim.GetMemberAnimSetCount())
        .AllocCache(false)
        .Create(&s_DeviceAllocator);

    // Set the invalidation flag for movement animation.
    evaluator->SetIsTranslateDisabled(!translateAnimEnabled);

    return evaluator;
}

/*
  
 */
nw::gfx::TransformAnimEvaluator*
InitializeSkeletalAnim(
    nw::gfx::SkeletalModel* model,
    nw::anim::ResAnim resAnim
)
{
    nw::gfx::AnimGroup* animGroup = model->GetSkeletalAnimGroup();
    NW_NULL_ASSERT(animGroup);

    nw::gfx::ResSkeletalModel resModel = model->GetResSkeletalModel();
    nw::gfx::ResSkeleton resSkeleton = resModel.GetSkeleton();
    const int maxBones = resSkeleton.GetBonesCount();
    const bool translateAnimEnabled =
        nw::ut::CheckFlag(resSkeleton.GetFlags(), nw::gfx::ResSkeleton::FLAG_TRANSLATE_ANIMATION_ENABLED);

    nw::gfx::TransformAnimEvaluator* evaluator = CreateTransformAnimEvaluator(
        maxBones, translateAnimEnabled, resAnim);

    NW_NULL_ASSERT(evaluator);

    // Bind the animation.
    bool bindResult = evaluator->Bind(animGroup);

    // Set animation for model.
    model->SetSkeletalAnimObject(evaluator);

    return evaluator;
}

/*
  
 */
void
TerminateAnim(void)
{
    for (int animIdx = 0; animIdx < s_AnimModelSets.Size(); ++animIdx)
    {
        nw::gfx::SafeDestroy(s_AnimModelSets[animIdx].evaluator);
    }
    s_AnimModelSets.clear();

    nw::gfx::SafeDestroy(s_AnimModelSetTemplate.evaluator);

}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);

    NW_ASSERT(s_ModelRoot == NULL);
    s_ModelRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    s_SceneRoot->AttachChild(s_ModelRoot);
    NW_NULL_ASSERT(s_ModelRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(50.0f, 40.0f, 50.0f),
        nw::math::VEC3(0.0f, 10.0f, 0.0f),
        s_fNearPlane
    );
    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}

/*
  
 */
void
BuildResources(nw::demo::ResourceSet* resourceSet)
{
    resourceSet->resource.ForeachTexture(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachIndexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachVertexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));

    nw::gfx::Result result = resourceSet->resource.Setup(&s_DeviceAllocator);
    if (result.IsFailure())
    {
        NW_FATAL_ERROR("Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(&s_DeviceAllocator);

    nw::gfx::ResModelArray models = resourceSet->resource.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*modelResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);

        s_AnimModelSetTemplate.skeletalModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);
    }

    nw::gfx::ResLightArray lights = resourceSet->resource.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*lightResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);
    }

    // Resolve parent-child reference relationships
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    // Add a model to the scene
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(s_ModelRoot)
    );

    nw::gfx::ResSceneEnvironmentSettingArray settings = resourceSet->resource.GetSceneEnvironmentSettings();
    nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
    for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
        settingResource != settingsEnd; ++settingResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource(*settingResource)
            .CreateObject(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
            nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

        NW_NULL_ASSERT(sceneEnvironmentSetting);
        s_SceneEnvironmentSettings.push_back(sceneEnvironmentSetting);
    }
}

/*
  
 */
void
BuildAnimModelTemplate()
{
    NW_NULL_ASSERT(s_AnimModelSetTemplate.skeletalModel);

    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, SKELETAL_ANIM_RESOURCE_FILES[0], &s_DeviceAllocator);
    NW_ASSERT (resourceSet->resource.GetSkeletalAnimsCount() != 0);

    s_AnimModelSetTemplate.skeletalModel->GetSkeleton()->GetResSkeleton().EnableFlags(nw::gfx::ResSkeleton::FLAG_MODEL_COORDINATE);

    s_AnimModelSetTemplate.resAnim = resourceSet->resource.GetSkeletalAnims(0);
    s_AnimModelSetTemplate.evaluator = InitializeSkeletalAnim(s_AnimModelSetTemplate.skeletalModel, s_AnimModelSetTemplate.resAnim);
}

/*
  
 */
void
GetNewDisplayPosition(f32* pPosX, f32* pPosZ)
{
    // Lays out model in a spiral configuration.
    int whirlSideLength = (s_WhirlDepth * 2 + 2);
    int posX;
    int posZ;

    // Finds where the new position will be on the rectangle.
    switch (s_WhirlPos / whirlSideLength)
    {
    case 0:
        // Right
        posX = s_WhirlDepth + 1;
        posZ = s_WhirlPos % whirlSideLength - s_WhirlDepth;
        break;
    case 1:
        // Top
        posX = -(s_WhirlPos % whirlSideLength) + s_WhirlDepth;
        posZ = s_WhirlDepth + 1;
        break;
    case 2:
        // Left
        posX = -s_WhirlDepth - 1;
        posZ = -(s_WhirlPos % whirlSideLength) + s_WhirlDepth;
        break;
    case 3:
        // Bottom
        posX = s_WhirlPos % whirlSideLength - s_WhirlDepth;
        posZ = -s_WhirlDepth - 1;
        break;
    default:
        NW_ASSERT(false);
    }

    *pPosX = posX * 10.0f;
    *pPosZ = posZ * 10.0f;

    s_WhirlPos++;
    if (s_WhirlPos >= whirlSideLength * 4)
    {
        s_WhirlDepth++;
        s_WhirlPos = 0;
    }
    return;
}

/*
  
 */
void
ReleaseLastDisplayPosition()
{
    s_WhirlPos--;
    if (s_WhirlPos < 0)
    {
        s_WhirlDepth--;
        s_WhirlPos = (s_WhirlDepth * 2 + 2) * 4 - 1;
    }

    NW_ASSERT(s_WhirlDepth >= 0);
}

/*
  
 */
void
DuplicateModel(DuplicationMode duplicationMode)
{
    if (s_AnimModelSets.size() >= MAX_ANIM_MODELS)
    {
        return;
    }

    nw::gfx::SkeletalModel* skeletalModel;
    AnimModelSet animModelSet;

    switch (duplicationMode)
    {
    case SKELETON_SHARED:
        // Generate an instance of a skeletal model that shares a skeleton.
        // When sharing a skeleton, you need to set the FLAG_MODEL_COORDINATE flag to ResSkeleton.
        // In addition, since the first generated skeleton has ownership rights, caution is needed when destroying skeletons.
        // In this demo, s_AnimModelSetTemplate.skeletalModel has the skeleton's ownership rights.
        skeletalModel = nw::gfx::SkeletalModel::Builder()
            .SharedSkeleton( s_AnimModelSetTemplate.skeletalModel->GetSkeleton() )
            .SharedMaterialModel( s_AnimModelSetTemplate.skeletalModel )
            .Create(NULL, nw::gfx::ResSceneObject(s_AnimModelSetTemplate.skeletalModel->GetResModel()), &s_DeviceAllocator);

        NW_NULL_ASSERT(skeletalModel);

        animModelSet.isSkeletonShared = true;
        s_NumSkeletonSharedCopy++;
        break;

    case DUPLICATION:
        // Generate an instance of the skeletal model.
        skeletalModel = nw::gfx::SkeletalModel::Builder()
            .SharedMaterialModel( s_AnimModelSetTemplate.skeletalModel )
            .Create(NULL, nw::gfx::ResSceneObject(s_AnimModelSetTemplate.skeletalModel->GetResModel()), &s_DeviceAllocator);

        NW_NULL_ASSERT(skeletalModel);

        animModelSet.isSkeletonShared = false;
        s_NumCopy++;
        break;

    default:
        NW_ASSERT(false);
    }

    if (s_ModelRoot->AttachChild(skeletalModel))
    {
        f32 posX;
        f32 posZ;

        animModelSet.skeletalModel = skeletalModel;
        animModelSet.resAnim = s_AnimModelSetTemplate.resAnim;
        switch (duplicationMode)
        {
        case SKELETON_SHARED:
            // When sharing a skeleton, it is okay to evaluate an animation for one shared skeleton, so the creation of animation evaluations and binding is not performed for duplicate skeletons.
            // 
            animModelSet.evaluator = NULL;
            break;

        case DUPLICATION:
            animModelSet.evaluator = InitializeSkeletalAnim(animModelSet.skeletalModel, animModelSet.resAnim);
            NW_NULL_ASSERT(animModelSet.evaluator);
            break;
        }
        s_AnimModelSets.push_back( animModelSet );

        GetNewDisplayPosition(&posX, &posZ);
        skeletalModel->Transform().SetTranslate(posX, 0.0f, posZ);

        // Since the scene structure changed, perform initialization.
        s_SceneSystem->InitializeScene(s_SceneRoot);
    }
    else
    {
        nw::ut::SafeDestroy(skeletalModel);
    }
}

/*
  
 */
void
DestroyLastDuplicatedModel()
{
    if (s_AnimModelSets.empty())
    {
        return;
    }

    AnimModelSet destroyModelSet = s_AnimModelSets.back();
    s_AnimModelSets.pop_back();

    s_ModelRoot->DetachChild(destroyModelSet.skeletalModel);

    nw::ut::SafeDestroy(destroyModelSet.skeletalModel);
    nw::ut::SafeDestroy(destroyModelSet.evaluator);

    s_SceneSystem->InitializeScene(s_SceneRoot);

    ReleaseLastDisplayPosition();

    if (destroyModelSet.isSkeletonShared)
    {
        s_NumSkeletonSharedCopy--;
    }
    else
    {
        s_NumCopy--;
    }
}

/*
  
 */
void
InitializeScenes()
{
    s_AnimModelSetTemplate.skeletalModel = NULL;

    BuildRootNodes();

    BuildCameras();

    NW_FOREACH(const wchar_t* name, MODEL_RESOURCE_FILES)
    {
        BuildResources(nw::demo::Utility::LoadResources(s_Resources, name, &s_DeviceAllocator));
    }

    BuildAnimModelTemplate();

    s_NumCopy = 0;
    s_NumSkeletonSharedCopy = 0;

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Resolves and sets references for the scene environment.
    s_RenderSystem->SetSceneEnvironmentSettings(s_SceneSystem, &s_SceneEnvironmentSettings);

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;

    s_WhirlDepth = 0;
    s_WhirlPos = 0;

    // Begin by placing some objects in advance.
    DuplicateModel(SKELETON_SHARED);
    DuplicateModel(SKELETON_SHARED);
    DuplicateModel(DUPLICATION);
    DuplicateModel(DUPLICATION);
    DuplicateModel(DUPLICATION);
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);
    nw::ut::SafeDestroyAll(s_SceneEnvironmentSettings);

    TerminateAnim();

    NW_GL_ASSERT();

    s_Resources.clear();
    s_SceneEnvironmentSettings.clear();
    s_ModelRoot = NULL;
}

/*
  
 */
void
ControlModels()
{
    if (nw::demo::PadFactory::GetPad()->IsButtonRepeatFast(nw::demo::Pad::BUTTON_L))
    {
        // Generate a model with the L Button.
        // Press the L Button while holding the +Control Pad upwards to generate a model that shares a skeleton.
        if (nw::demo::PadFactory::GetPad()->IsButtonPress(nw::demo::Pad::BUTTON_UP))
        {
            DuplicateModel(SKELETON_SHARED);
        }
        else
        {
            DuplicateModel(DUPLICATION);
        }
    }
    else if (nw::demo::PadFactory::GetPad()->IsButtonRepeatFast(nw::demo::Pad::BUTTON_R))
    {
        // Destroy the last generated model.
        DestroyLastDuplicatedModel();
    }
}

/*
  
 */
void
UpdateScene()
{
    ControlModels();

    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    ++s_FrameCount;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    const int dumpPositionX = 10;
    const int dumpPositionY = 10;
    s_GraphicsDrawing.DrawString(
        dumpPositionX,
        dumpPositionY,
        "       original: 1\n"
        "           copy: %d\n"
        "skeleton shared: %d",
        s_NumCopy,
        s_NumSkeletonSharedCopy
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}

/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    nw::demo::DebugUtility::PostInitializeScenes();

    NW_DEV_LOG("[Multi Instance Demo usage]\n");
    NW_DEV_LOG("r         : delete\n");
    NW_DEV_LOG("l         : add copy\n");
    NW_DEV_LOG("up + l    : add skeleton shared copy\n");

    bool isContinuing = true;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();
        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, NULL, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeGraphicsSystem();
}
