/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

#include <nn/hid/CTR/hid_TouchPanelReader.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;
nw::demo::DemoAllocator s_ParticleAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;
nw::gfx::ParticleContext* s_ParticleContext = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
nw::gfx::SceneNode* s_SceneRoot = NULL;
nw::gfx::SceneNode* s_SceneRootInvisible = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
const f32 s_fNearPlane = 0.1f;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Particle-related
nw::gfx::ParticleSceneUpdater* s_ParticleSceneUpdater = NULL;


/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    sceneDescription.maxSceneNodes = 256;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    s_ParticleContext = nw::gfx::ParticleContext::Builder()
        .Create(&s_DeviceAllocator);

    s_ParticleSceneUpdater = nw::gfx::ParticleSceneUpdater::Builder()
        .Create(&s_DeviceAllocator);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_ParticleSceneUpdater);

    nw::gfx::SafeDestroy(s_ParticleContext);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .MaxChildren(16)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);

    s_SceneRootInvisible = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .MaxChildren(16)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRootInvisible);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(0.0f, 20.0f, 50.0f),
        nw::math::VEC3(0.0f, 0.0f, 0.0f),
        s_fNearPlane
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}


/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::gfx::SafeDestroyBranch(s_SceneRootInvisible);
    nw::demo::SafeCleanupResources(s_Resources);

    NW_GL_ASSERT();

    s_Resources.clear();
}

/*
  
 */
void
UpdateScene()
{
    NW_ASSERT(0 < s_RenderTargets.size());

    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    NW_NULL_ASSERT(s_ParticleSceneUpdater);
    s_ParticleSceneUpdater->UpdateNode(
        s_SceneSystem->GetSceneContext(),
        s_ParticleContext);

    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    s_FrameCount++;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}




//----------------------------------------
// Variable unique to the demo

// Particle effect class
nw::demo::ParticleEffect* s_ParticleEffect = NULL;

// Particle node class
nw::demo::ParticleNode* s_ParticleNode = NULL;

// Counter
u32 s_Count = 0;

// Number of generations
#define PARTICLE_EFFECT_NUM (10)

// Effect filename to be loaded
const wchar_t* EFFECT_RESOURCE_FILE = NW_DEMO_FILE_PATH(L"light_particle.bcptl");

// Effect shader file to be loaded
const wchar_t* SHADER_RESOURCE_FILE_NAME = NW_DEMO_FILE_PATH(L"nwgfx_ParticleDefaultShader.bcsdr");

/*
  
 */
void
InitializeParticleApplicationDemo()
{
    // This demo prepares 10 effect instances, and performs playback and re-use based on input.
    // 
    // By preparing the needed instances in advance, the cost of generation and destruction can be saved, and the application can be executed at a stabler cost.
    // 
    // 

    // Loads shader binaries.
    nw::demo::ParticleEffect::InitializeShaderBinary(SHADER_RESOURCE_FILE_NAME, &s_DeviceAllocator);

    // Generate the particle effect class.
    s_ParticleEffect = nw::demo::ParticleEffect::Create(&s_DeviceAllocator, &s_DeviceAllocator, false, s_ParticleContext);

    // Load the effect data and setup.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, EFFECT_RESOURCE_FILE, &s_DeviceAllocator);

    // Sets up the resources.
    s_ParticleEffect->Setup(resourceSet->resource, true);

    // Set resources to ParticleEffect.
    s_ParticleEffect->Register(resourceSet->resource);

    // Prepare 10 instances.
    s_ParticleEffect->AddPool(PARTICLE_EFFECT_NUM);

    // Lease the particle node class' instance.
    s_ParticleNode = s_ParticleEffect->LeaseInstance();

    // Add to the scene.
    s_SceneRoot->AttachChild(s_ParticleNode);
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    // Logs and outputs operational description.
    NW_DEV_LOG("[Particle Application Demo usage]\n");
    NW_DEV_LOG("touch panel: create\n");
}

/*
  
 */
void
UpdateParticleApplicationDemo( nn::hid::CTR::TouchPanelReader&  touchPanelReader)
{
    // Processes the touch panel messages.
    nn::hid::CTR::TouchPanelStatus touchPanelStatus;
    static u16 touchPanelX = 0;
    static u16 touchPanelY = 0;
    touchPanelReader.ReadLatest( &touchPanelStatus );

    bool isSceneUpdate = false;

    // Generate an effect when touched.
    if ( ( touchPanelX == 0 && touchPanelStatus.x != 0 ) || ( touchPanelY == 0 && touchPanelStatus.y != 0 ) )
    {
        // Borrow an available instance from s_FireEffect.
        nw::demo::ParticleNode* particleNode = s_ParticleEffect->LeaseInstance();
        if (particleNode != NULL)
        {
            // Sets the display position.
            f32 fposx = static_cast<f32>(touchPanelStatus.x)/8.f - 20.f;
            f32 fposy = static_cast<f32>(touchPanelStatus.y)/8.f - 15.f;
            particleNode->SetTranslate(fposx, -fposy, 0);

            // Add to the scene.
            s_SceneRoot->AttachChild(particleNode);

            isSceneUpdate = true;
        }
    }

    // If there are ended effects, detach them from the scene.
    for ( u32 i = 0; i < PARTICLE_EFFECT_NUM; ++i)
    {
        nw::demo::ParticleNode* particleNode = s_ParticleEffect->GetActiveEffect(i);
        if (particleNode && particleNode->IsDone())
        {
            s_SceneRoot->DetachChild(particleNode);
            s_ParticleEffect->ReleaseInstance(particleNode);
            isSceneUpdate = true;
        }
    }

    // If there are changes to the scene, perform updating.
    if (isSceneUpdate)
    {
        s_SceneSystem->InitializeScene(s_SceneRoot);
    }

    touchPanelX = touchPanelStatus.x;
    touchPanelY = touchPanelStatus.y;

    ++s_Count;
}

/*
  
 */
void
ReportParticleApplicationDemo()
{
    if(s_Count%20 != 0)
    {
        s_GraphicsDrawing.DrawString(120, 60, "TOUCH ME!!");
    }

    u32 alive[PARTICLE_EFFECT_NUM];
    for ( u32 i = 0; i < PARTICLE_EFFECT_NUM; ++i)
    {
        nw::demo::ParticleNode* particleNode = s_ParticleEffect->GetActiveEffect(i);
        if (!particleNode)
        {
            alive[i] = false;
        }
        else
        {
            if (particleNode->IsDone())
            {
                alive[i] = false;
            }
            else
            {
                alive[i] = true;
            }
        }
    }

    s_GraphicsDrawing.DrawString(0, 10, "[%d][%d][%d][%d][%d][%d][%d][%d][%d][%d]",
                    alive[0],alive[1],alive[2],alive[3],alive[4],alive[5],alive[6],alive[7],alive[8],alive[9]);
}


/*
  
 */
void
TerminateParticleApplicationDemo()
{
    nw::demo::ParticleNode* prevTarget = s_ParticleEffect->GetActiveEffect(0);
    while (prevTarget != NULL)
    {
        s_ParticleEffect->ReleaseInstance(prevTarget);
        prevTarget = s_ParticleEffect->GetActiveEffect(0);
    }

    s_ParticleEffect->FreePool();

    nw::gfx::SafeDestroy(s_ParticleEffect);

    // Destroy the loaded particle shader.
    nw::demo::ParticleEffect::FinalizeShaderBinary();
}




/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    // The initialization process for ParticleApplicationDemo.
    InitializeParticleApplicationDemo();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    // Touch panel reader
    nn::hid::CTR::TouchPanelReader  s_TouchPanelReader;

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();

        // The fixed term process for ParticleApplicationDemo.
        UpdateParticleApplicationDemo( s_TouchPanelReader );

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();

        // Display the report for ParticleApplicationDemo.
        ReportParticleApplicationDemo();

        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    // The finalization process for ParticleApplicationDemo.
    TerminateParticleApplicationDemo();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);
    nw::demo::InitializeDemoAllocator(&s_ParticleAllocator, nw::demo::DEMO_PARTICLE_MEMORY_SIZE);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, &s_ParticleAllocator, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeDemoAllocator(&s_ParticleAllocator);

    nw::demo::FinalizeGraphicsSystem();
}
