/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

static const wchar_t* MODEL_RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"Cube.bcmdl"),
    NW_DEMO_FILE_PATH(L"SceneEnvironmentSetting.bcenv"),
    NW_DEMO_FILE_PATH(L"FragmentLight.bcenv"),
};

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
const int SCENE_NODE_COUNT = 4;
nw::gfx::SceneNode* s_SceneRoot = NULL;
nw::gfx::Camera* s_Camera = NULL;
s32 s_FrameCount = 0;

//----------------------------------------
// Scene environment-related
const s32 ENVIRONMENT_SETTINGS_COUNT = 1;

typedef nw::ut::FixedSizeArray<nw::gfx::SceneEnvironmentSetting*, ENVIRONMENT_SETTINGS_COUNT> SceneEnvironmentSettingArray;
SceneEnvironmentSettingArray s_SceneEnvironmentSettings;

const s32 s_BaseCameraIndex = 0;

/*
  
 */
void
InitializeGraphics()
{
    // Generate an instance of the demo rendering instance.
    // The setting of the standard values to the renderDescription structure for the generation settings is performed by the structure's constructor.
    // 
    nw::demo::RenderSystem::Description renderDescription;
    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    // Initialize the GraphicsDrawing for debug display.
    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());

    // Generate an instance of the demo scene system.
    // The setting of the standard values to the sceneDescription structure for the generation settings is performed by the structure's constructor.
    // 
    nw::demo::SceneSystem::Description sceneDescription;
    sceneDescription.isFixedSizeMemory = true;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    NW_GL_ASSERT();
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);
}

/*
  
 */
void
BuildCameras()
{
    s_Camera = nw::demo::Utility::CreateCamera(
        &s_DeviceAllocator,
        nw::math::VEC3(7.0f, 3.5f, -5.0f)
    );
    s_SceneRoot->AttachChild(s_Camera);
    s_SceneSystem->GetCameraController()->Register(s_Camera);
}

/*
  
 */
void
BuildResources(nw::demo::ResourceSet* resourceSet)
{
    // A function object is set to ResTexture, ResIndexStream, and ResVertexStream within ResGraphicsFile with ForeachTexture, ForeachIndexStream, and ForeachVertexStream.
    // 
    // Function objects specify a memory placement location with SetLocationFlag.
    // Here, textures and vertex data are placed in VRAMA and VRAMB.
    //
    resourceSet->resource.ForeachTexture(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachIndexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachVertexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));

    // Sets up the resource set. The following processes occur in the set up of resources.
    // - Reference resolution for shaders and textures
    // - VRAM transfer for textures and vertices
    // - Material command generation
    //
    // The return value Result for Setup is a ResourceResult, and returns reference resolution errors, etc.
    // When an error code is returned, provide the necessary ResGraphicsFile to the argument, and re-do Setup to eliminate the error.
    // Resources that undergo Setup must call Cleanup to destroy the internally allocated memory, etc.
    //
    nw::gfx::Result result = resourceSet->resource.Setup(&s_DeviceAllocator);
    if (result.IsFailure())
    {
        NW_FATAL_ERROR("Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(SCENE_NODE_COUNT, &s_DeviceAllocator);

    // Generate an instance of the model.
    nw::gfx::ResModelArray models = resourceSet->resource.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*modelResource),
            false
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);
    }

    // Generate an instance of the light.
    nw::gfx::ResLightArray lights = resourceSet->resource.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*lightResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);
    }

    // By passing all of the SceneNode inheritance classes included within the file, we construct ResSceneNode's parent-child hierarchy in SceneNode.
    // 
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    // Add the SceneNode inheritance class to the scene root.
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(s_SceneRoot)
    );

    // Create a scene environment instance.
    // Scene environments are needed to switch the light sets, cameras and fogs by material.
    // Only light sets are used in SimpleDemo.
    nw::gfx::ResSceneEnvironmentSettingArray settings = resourceSet->resource.GetSceneEnvironmentSettings();
    nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
    for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
        settingResource != settingsEnd; ++settingResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource(*settingResource)
            .CreateObject(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
            nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

        NW_NULL_ASSERT(sceneEnvironmentSetting);
        s_SceneEnvironmentSettings.push_back(sceneEnvironmentSetting);
    }
}

/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildCameras();

    // Create the model, light, and scene environment instances from the file and add them to the scene root.
    NW_FOREACH(const wchar_t* name, MODEL_RESOURCE_FILES)
    {
        BuildResources(nw::demo::Utility::LoadResources(s_Resources, name, &s_DeviceAllocator));
    }

    // With SceneInitializer, you set the MaterialId that determines the order for material sorts.
    // With SceneTraverser, you collect and serialize scene nodes, and add them to SceneContext.
    s_SceneSystem->InitializeScene(s_SceneRoot);

    // To set the camera position in world coordinates, you must update once.
    s_SceneSystem->UpdateScene();

    // Resolves and sets references for the scene environment.
    s_RenderSystem->SetSceneEnvironmentSettings(s_SceneSystem, &s_SceneEnvironmentSettings);

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_Camera);
    nw::demo::Utility::SetCameraAspectRatio(s_Camera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);
    nw::gfx::SafeDestroyAll(s_SceneEnvironmentSettings);

    NW_GL_ASSERT();

    s_Resources.clear();
    s_SceneEnvironmentSettings.clear();
}

/*
  
 */
void
UpdateScene()
{
    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_Camera->UpdateCameraMatrix();

    ++s_FrameCount;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}

/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    while ( isContinuing )
    {
        // Update the test framework.
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        // Perform pad updates.
        nw::demo::PadFactory::GetPad()->Update();

        // For the processing order here, refer to the Graphics Rendering Sequence in the Processing Flow documentation.
        // 

        // Update the animations, nodes, skeleton and camera.
        UpdateScene();

        // Set the camera for lights, environment maps, and WScale.
        // Since ActiveCamera is reset by RenderContext::ResetState, settings must be performed before rendering each frame.
        // 
        renderContext->SetActiveCamera(s_BaseCameraIndex);

        // Perform the update dependent on viewpoint and the RenderQueue construction.
        s_RenderSystem->SubmitView(s_SceneSystem);

        // Render to the upper screen.
        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderScene(s_Camera, nw::demo::UPPER_SCREEN);

        // Clear the lower screen.
        s_RenderSystem->ClearBySkyModel(s_Camera);

        // Render font to the lower screen.
        ReportDemo();
        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        // Display the rendered buffer.
        // The PresentBuffer function waits for V-Sync internally and processes the command list.
        // For more information see the PresentBuffer function.
        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN);

        // After rendering in the gfx library, ResetState is called, so the state used for rendering is reset.
        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    // Initialize the SDK and device memory.
    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    // NW_DEMO_TEST_LOOP is a macro that initializes the profiler, etc.
    // If the NW_DEBUG_CHECK_MEMORY_LEAK macro is enabled, a check for memory leaks is enabled, and the following block will loop.
    // At this time, pressing START on the pad will result in a TRUE for the nw::demo::Utility::IsTerminating function, and will exit from the loop within the DemoScene function.
    // 
    // When the following loop is repeating, the memory leak check occurs.
    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, NULL, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    // Perform end processing for device memory.
    nw::demo::FinalizeGraphicsSystem();
}
