/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nn/gr.h>

#include <nw/types.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>

#include "GrRenderer.h"

namespace
{

class GrUserRenderCommand;

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

const wchar_t* MODEL_RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"Male.bcmdl"),
    NW_DEMO_FILE_PATH(L"SceneEnvironmentSetting.bcenv"),
    NW_DEMO_FILE_PATH(L"FragmentLight.bcenv")
};

const wchar_t* GR_SHADER_FILE = NW_DEMO_FILE_PATH(L"UserRenderCommandShader.shbin");

//----------------------------------------
// Rendering-related
const int RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;

static nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
const int SCENE_NODE_COUNT = 8;
nw::gfx::SceneNode* s_SceneRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Scene environment-related
const s32 ENVIRONMENT_SETTINGS_COUNT = 1;

typedef nw::ut::FixedSizeArray<nw::gfx::SceneEnvironmentSetting*, ENVIRONMENT_SETTINGS_COUNT> SceneEnvironmentSettingArray;
SceneEnvironmentSettingArray s_SceneEnvironmentSettings;

//----------------------------------------
// Related to rendering with user-defined render nodes
nw::gfx::UserRenderNode* s_UserRenderNode = NULL;
GrUserRenderCommand*     s_GrUserRenderCommand = NULL;
GrPrimitiveRenderer s_GrPrimitiveRenderer;

/*
  
 */
class GrUserRenderCommand : public nw::gfx::UserRenderCommand
{
public:
    //
    virtual void Invoke(nw::gfx::RenderContext* renderContext)
    {
        // Get the UserRenderNode that is set by UserRenderCommand
        nw::gfx::UserRenderNode* userRenderNode = GetUserRenderNode();
        NW_NULL_ASSERT( userRenderNode );

        nw::math::MTX34& worldMatrix = userRenderNode->WorldMatrix();

        // Get a pointer to the render command buffer
        u32* command = NULL;
        nngxGetCmdlistParameteri(
            NN_GX_CMDLIST_CURRENT_BUFADDR,
            reinterpret_cast<GLint*>(&command));

        // Save the starting address for the current command list to advance the pointer at the end
        const u32* start_addr = command;

        command = s_GrPrimitiveRenderer.MakeSceneBeginCommand(command);

        command = s_GrPrimitiveRenderer.MakeDrawBoxCommand(command, worldMatrix);

        command = s_GrPrimitiveRenderer.MakeSceneEndCommand(command);

        // Advance the command list pointer
        nngxMoveCommandbufferPointer((command - start_addr) * sizeof(u32));

        // Since the shaders are switched to render in GR, reset the render context.
        // 
        renderContext->ResetState(
            nw::gfx::RenderContext::RESETSTATEMODE_SHADER_PROGRAM |
            nw::gfx::RenderContext::RESETSTATEMODE_MATERIAL_CACHE
            ,0);

    }

public:
    //
    GrUserRenderCommand(){}

    //
    virtual ~GrUserRenderCommand() {}
};

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount      = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;

    // Because the default value of maxUserRenderNodes is 0, values are set here.
    sceneDescription.maxUserRenderNodes = 1;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    s_GrPrimitiveRenderer.Initialize( &s_DeviceAllocator, GR_SHADER_FILE );

    NW_GL_ASSERT();
}


/*
  
 */
void
TerminateGraphics()
{
    s_GrPrimitiveRenderer.Finalize();

    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);


    NW_GL_ASSERT();
}

/*
  
 */
void
BuildUserRenderNode()
{
    // Generate a user-defined render command
    void* userCommandMemory = s_DeviceAllocator.Alloc(sizeof(GrUserRenderCommand));
    NW_NULL_ASSERT(userCommandMemory);
    s_GrUserRenderCommand = new(userCommandMemory) GrUserRenderCommand();

    // Generate a user-defined render node
    NW_ASSERT(s_UserRenderNode == NULL);
    s_UserRenderNode = nw::gfx::UserRenderNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_UserRenderNode);

    // Since UserRenderNode is not Z-sorted, render order can be control only by render priority and layer ID.
    // 
    s_UserRenderNode->SetPriority( 0 );
    s_UserRenderNode->SetLayerId( 0 );

    // Set the render command in the user-defined render node
    s_UserRenderNode->SetUserRenderCommand(s_GrUserRenderCommand);

    // Attach the user-defined render node
    s_SceneRoot->AttachChild(s_UserRenderNode);

    s_UserRenderNode->Transform().SetTranslate(nw::math::VEC3(-5.0f, 5.0f, 0.0f));
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .IsFixedSizeMemory(false)
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(23.0f, 20.0f, 21.0f),
        nw::math::VEC3(0.0f, 8.0f, 0.0f)
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}

/*
  
 */
void
BuildResources(nw::demo::ResourceSet* resourceSet)
{
    resourceSet->resource.ForeachTexture(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachIndexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));
    resourceSet->resource.ForeachVertexStream(nw::gfx::LocationFlagSetter(NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP));

    nw::gfx::Result result = resourceSet->resource.Setup(&s_DeviceAllocator);
    if (result.IsFailure())
    {
        NW_FATAL_ERROR("Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(SCENE_NODE_COUNT, &s_DeviceAllocator);

    nw::gfx::ResModelArray models = resourceSet->resource.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*modelResource),
            false
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);

        nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>(node);
        model->Transform().SetTranslate(nw::math::VEC3(5.0f, 0.0f, 0.0f));
    }

    nw::gfx::ResLightArray lights = resourceSet->resource.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*lightResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);
    }

    // Resolve parent-child reference relationships
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(s_SceneRoot)
    );

    nw::gfx::ResSceneEnvironmentSettingArray settings = resourceSet->resource.GetSceneEnvironmentSettings();
    nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
    for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
        settingResource != settingsEnd; ++settingResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource(*settingResource)
            .CreateObject(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
            nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

        NW_NULL_ASSERT(sceneEnvironmentSetting);
        s_SceneEnvironmentSettings.push_back(sceneEnvironmentSetting);
    }
}

/*
  
 */
void
InitializeScenes()
{
    BuildRootNodes();

    BuildUserRenderNode();

    BuildCameras();

    NW_FOREACH(const wchar_t* name, MODEL_RESOURCE_FILES)
    {
        BuildResources(nw::demo::Utility::LoadResources(s_Resources, name, &s_DeviceAllocator));
    }

    // Traverse and initialize the scene tree.
    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    s_RenderSystem->SetSceneEnvironmentSettings(s_SceneSystem, &s_SceneEnvironmentSettings);

    // Set the camera.
    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    s_DeviceAllocator.Free(s_GrUserRenderCommand);
    nw::gfx::SafeDestroyBranch(s_UserRenderNode);

    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);
    nw::gfx::SafeDestroyAll(s_SceneEnvironmentSettings);

    NW_GL_ASSERT();

    s_Resources.clear();
}

/*
  
 */
void
UpdateScene()
{
    s_SceneSystem->GetCameraController()->Update();

    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();

    nw::gfx::ResCameraProjectionUpdater resProjectionUpdater =
        s_BaseCamera->GetProjectionUpdater()->GetResource();

    int near = resProjectionUpdater.GetNear();

    s_RenderSystem->CalcStereoCamera(
        s_LeftCamera,
        s_RightCamera,
        s_BaseCamera,
        near + 5.0f);

    ++s_FrameCount;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}

/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    InitializeScenes();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();

        UpdateScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);

        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);
        s_RenderSystem->RenderStereoScene(s_LeftCamera, s_RightCamera);

        s_RenderSystem->ClearBySkyModel(s_BaseCamera);
        ReportDemo();
        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, NULL, &s_RenderSystem)
    {
        InitializeGraphics();

        DemoScene();

        TerminateGraphics();
    }

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeGraphicsSystem();
}
