/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_ParticleSceneUpdater.h>

#include <nw/gfx/gfx_SceneContext.h>
#include <nw/gfx/gfx_ParticleContext.h>
#include <nw/gfx/gfx_SkeletalModel.h>
#include <nw/gfx/gfx_ParticleModel.h>
#include <nw/gfx/gfx_ParticleSet.h>
#include <nw/gfx/gfx_ParticleEmitter.h>

#include <nw/ut/ut_Foreach.h>


namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_ROOT_DEFINITION(ParticleSceneUpdater);

//----------------------------------------
ParticleSceneUpdater*
ParticleSceneUpdater::Builder::Create(
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    void* memory = allocator->Alloc(sizeof(ParticleSceneUpdater));
    if (memory == NULL)
    {
        return NULL;
    }

    return new(memory) ParticleSceneUpdater(
        allocator);
}

//----------------------------------------
void
ParticleSceneUpdater::SetStepFrame(
    SceneContext* sceneContext,
    f32 stepFrame
)
{
    NW_PROFILE("ParticleSceneUpdater::SetStepFrame");

    {
        ParticleModelArray::iterator end = sceneContext->GetParticleModelEnd();
        for (ParticleModelArray::iterator i = sceneContext->GetParticleModelBegin(); i != end;)
        {
            ParticleModel* model = *i++;
            NW_NULL_ASSERT(model);

            model->ParticleAnimFrameController().SetStepFrame(stepFrame);
        }
    }

    {
        ParticleEmitterArray::iterator end = sceneContext->GetParticleEmitterEnd();
        for (ParticleEmitterArray::iterator i = sceneContext->GetParticleEmitterBegin(); i != end;)
        {
            ParticleEmitter* emitter = *i++;
            NW_NULL_ASSERT(emitter);

            emitter->ParticleAnimFrameController().SetStepFrame(stepFrame);
        }
    }
}

//----------------------------------------
void
ParticleSceneUpdater::UpdateNode(
    SceneContext* sceneContext,
    ParticleContext* particleContext,
    bool enableSwapBuffer
)
{
    NW_PROFILE("ParticleSceneUpdater::UpdateNode");

    // Process in order from Frame update to Emission to UpdateParticles
    // Since there is no dependency relationship among emitters, UpdateFrame for emitters and Emission occurs together (previously Emission happened after UpdateFrame completed)
    // 

    {
        ParticleModelArray::iterator end = sceneContext->GetParticleModelEnd();
        for (ParticleModelArray::iterator i = sceneContext->GetParticleModelBegin(); i != end;)
        {
            ParticleModel* model = *i++;
            NW_NULL_ASSERT(model);

            model->UpdateParticleFrame();
        }
    }

    {
        ParticleEmitterArray::iterator end = sceneContext->GetParticleEmitterEnd();
        for (ParticleEmitterArray::iterator i = sceneContext->GetParticleEmitterBegin(); i != end;)
        {
            ParticleEmitter* emitter = *i++;
            NW_NULL_ASSERT(emitter);

            emitter->UpdateParticleFrame();
            emitter->Emission(particleContext);
        }
    }

    {
        ParticleSetArray::iterator end = sceneContext->GetParticleSetEnd();
        for (ParticleSetArray::iterator i = sceneContext->GetParticleSetBegin(); i != end;)
        {
            ParticleSet* particleSet = *i++;
            particleSet->UpdateParticles(particleContext, enableSwapBuffer);
        }
    }
}

} // namespace gfx
} // namespace nw
