/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_Config.h>
#include <nw/gfx/gfx_SimpleMaterialActivator.h>
#include <nw/gfx/gfx_MaterialState.h>
#include <nw/gfx/gfx_Model.h>
#include <nn/gx.h>
#include <nw/dev.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(SimpleMaterialActivator, IMaterialActivator);

//----------------------------------------
SimpleMaterialActivator*
SimpleMaterialActivator::Create(os::IAllocator* allocator)
{
    void* memory = allocator->Alloc(sizeof(SimpleMaterialActivator));

    if (memory == NULL)
    {
        return NULL;
    }
    else
    {
        return new(memory) SimpleMaterialActivator(allocator);
    }
}

//----------------------------------------
void
SimpleMaterialActivator::Activate(RenderContext* renderContext, const Material* material)
{
    NW_NULL_ASSERT(renderContext);

    const SceneEnvironment& sceneEnvironment = renderContext->GetSceneEnvironment();
    bool areFragmentLightsDirty = sceneEnvironment.IsFragmentLightsDirty() || sceneEnvironment.IsAmbientLightDirty();
    bool areVertexLightsDirty = sceneEnvironment.IsVertexLightsDirty() || sceneEnvironment.IsHemiSphereLightDirty();

    // With SimpleMaterialActivator a resouce comparison is performed.
    // When the directly previously set material is using the buffer, this cannot be skipped.
    const Material* cacheMaterial = renderContext->GetMaterialCache();
    if ( ( cacheMaterial != NULL) &&
         ( cacheMaterial->GetBufferCount() == 0) &&
         ( cacheMaterial->GetOriginal() == material->GetOriginal() ) &&
         ( !areFragmentLightsDirty ) &&
         ( !areVertexLightsDirty ) )
    {
        return;
    }

    const ShaderProgram* shaderProgram = renderContext->GetShaderProgram();
    NW_NULL_ASSERT(shaderProgram);

    ResMaterial original = material->GetOriginal();
    const Model* owner = material->GetOwnerModel();
    NW_NULL_ASSERT(owner);

    RenderContext::MaterialHash materialHash = renderContext->GetMaterialHash();

    ResMaterial resShaderParameterMaterial = original;
    u32 shaderParametersHash = resShaderParameterMaterial.GetShaderParametersHash();
    bool isShaderParametersEnabled =
        (shaderParametersHash == 0x0) ||
        (shaderParametersHash != materialHash.shaderParameter) ||
        renderContext->IsShaderProgramDirty();
    if (isShaderParametersEnabled)
    {
        i::MaterialState::ActivateShaderParameter(shaderProgram, resShaderParameterMaterial);
        materialHash.shaderParameter = shaderParametersHash;
    }

    ResMaterial resShadingParametersMaterial = original;

    // FragmentLightEnabled must be set.
    i::MaterialState::ActivateFragmentLightParameters(sceneEnvironment, shaderProgram, resShadingParametersMaterial);

    u32 shadingParametersHash = resShadingParametersMaterial.GetShadingParameterHash();
    bool isShadingParametersEnabled =
        (shadingParametersHash == 0x0) || (shadingParametersHash != materialHash.shadingParameter) || areVertexLightsDirty;
    if (isShadingParametersEnabled)
    {
        i::MaterialState::ActivateShadingParameters(sceneEnvironment, shaderProgram, resShadingParametersMaterial);
        materialHash.shadingParameter = shadingParametersHash;
    }

    ResRasterization resRasterization = original.GetRasterization();
    u32 rasterizationHash = original.GetRasterizationHash();
    bool isRasterizationEnabled = (rasterizationHash == 0x0) || (rasterizationHash != materialHash.rasterization);
    if (isRasterizationEnabled)
    {
        i::MaterialState::ActivateRasterization(resRasterization);
        materialHash.rasterization = rasterizationHash;
    }

    ResMaterial resTextureCoordinatorsMaterial = original;
    u32 textureCoordinatorHash = resTextureCoordinatorsMaterial.GetTextureCoordinatorsHash();
    bool isTextureCoordinatorEnabled = (textureCoordinatorHash == 0x0) || (textureCoordinatorHash != materialHash.textureCoordinator);
    if (isTextureCoordinatorEnabled)
    {
        i::MaterialState::ActivateTextureCoordinators(renderContext, shaderProgram, resTextureCoordinatorsMaterial);
        materialHash.textureCoordinator = textureCoordinatorHash;
    }

    ResMaterial resTextureMappersMaterial = original;
    u32 textureMappersHash = resTextureMappersMaterial.GetTextureMappersHash();
    bool isTextureMappersEnabled = (textureMappersHash == 0x0) || (textureMappersHash != materialHash.textureMapper);
    if (isTextureMappersEnabled)
    {
        i::MaterialState::ActivateTextureMappers(resTextureMappersMaterial);
        materialHash.textureMapper = textureMappersHash;
    }

#if defined(NW_GFX_PROCTEX) // Procedural texture settings
    ResMaterial resProceduralTextureMapperMaterial = original;
    i::MaterialState::ActivateProceduralTextureMapper(resProceduralTextureMapperMaterial);
#endif

    bool reflectionEnabled = false;
    bool lightEnabled = ut::CheckFlag(original.GetFlags(), ResMaterialData::FLAG_FRAGMENTLIGHT_ENABLED);
    u32 fragmentLightingHash = original.GetFragmentLightingHash();
    u32 fragmentLightingTableHash = original.GetFragmentLightingTableHash();
    if (lightEnabled)
    {
        ResFragmentLighting resFragmentLighting =original.GetFragmentShader().GetFragmentLighting();
        reflectionEnabled = ut::CheckFlag(resFragmentLighting.GetFlags(), ResFragmentLightingData::FLAG_REFLECTION_ENABLED);
        bool isFragmentLightingEnabled = (fragmentLightingHash == 0x0) ||
                                         (fragmentLightingHash != materialHash.fragmentLighting) ||
                                         areFragmentLightsDirty;
        if (isFragmentLightingEnabled)
        {
            i::MaterialState::ActivateFragmentLighting(sceneEnvironment, resFragmentLighting);
        }

        ResFragmentLightingTable resFragmentLightingTable =original.GetFragmentShader().GetFragmentLightingTable();
        bool isFragmentLightingTableEnabled = (fragmentLightingTableHash == 0x0) || (fragmentLightingTableHash != materialHash.fragmentLightingTable);
        if (isFragmentLightingTableEnabled)
        {
            i::MaterialState::ActivateFragmentLightingTable(resFragmentLighting, resFragmentLightingTable);
        }
    }
    // Must update because lightEnabled is included in the hash.
    materialHash.fragmentLighting = fragmentLightingHash;
    materialHash.fragmentLightingTable = fragmentLightingTableHash;

    ResMaterialColor resMaterialColor =  original.GetMaterialColor();
    u32 materialColorHash = original.GetMaterialColorHash();
    bool isMaterialColorEnabled = (materialColorHash == 0x0) || (materialColorHash != materialHash.materialColor) || areFragmentLightsDirty;
    if (!isMaterialColorEnabled && cacheMaterial != NULL)
    {
        // When a material color is changed to reflectionEnabled, since it must be reset, the previously set material is referenced.
        bool isPreReflectionEnabled = ut::CheckFlag(cacheMaterial->GetOriginal().GetFragmentShader().GetFragmentLighting().GetFlags(), ResFragmentLightingData::FLAG_REFLECTION_ENABLED);
        isMaterialColorEnabled = (reflectionEnabled != isPreReflectionEnabled);
    }
    if (isMaterialColorEnabled)
    {
        i::MaterialState::ActivateMaterialColor(sceneEnvironment, shaderProgram, resMaterialColor, reflectionEnabled);
        materialHash.materialColor = materialColorHash;
    }

    ResFragmentShader resTextureCombinerFragmentShader = original.GetFragmentShader();
    u32 textureCombinersHash = original.GetTextureCombinersHash();
    bool isTextureCombinersEnabled = (textureCombinersHash == 0x0) || (textureCombinersHash != materialHash.textureCombiner);
    if (isTextureCombinersEnabled || isMaterialColorEnabled)
    {
        i::MaterialState::ActivateTextureCombiners(resTextureCombinerFragmentShader, resMaterialColor);
        materialHash.textureCombiner = textureCombinersHash;
    }

    ResAlphaTest resAlphaTest = original.GetFragmentShader().GetAlphaTest();
    u32 alphaTestHash = original.GetAlphaTestHash();
    bool isAlphaTestEnabled = (alphaTestHash == 0x0) || (alphaTestHash != materialHash.alphaTest);
    if (isAlphaTestEnabled)
    {
        i::MaterialState::ActivateAlphaTest(resAlphaTest);
        materialHash.alphaTest = alphaTestHash;
    }

    ResFragmentOperation resFragmentOperation = original.GetFragmentOperation();
    u32 fragmentOperationHash = original.GetFragmentOperationHash();
    bool isFragmentOperationEnabled = (fragmentOperationHash == 0x0) || (fragmentOperationHash != materialHash.fragmentOperation);
    if (isFragmentOperationEnabled)
    {
        i::MaterialState::ActivateFragmentOperation(resFragmentOperation);
        materialHash.fragmentOperation = fragmentOperationHash;
    }

    renderContext->SetMaterialHash(materialHash);\
}

//----------------------------------------
SimpleMaterialActivator::SimpleMaterialActivator(os::IAllocator* allocator)
    : IMaterialActivator(allocator)
{
}

//----------------------------------------
SimpleMaterialActivator::~SimpleMaterialActivator()
{
}

} // namespace gfx
} // namespace nw
