/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_AnimGroup.h>
#include <nw/gfx/gfx_CalculatedTransform.h>
#include <nw/gfx/gfx_TransformAnimBlendOp.h>
#include <nw/gfx/gfx_TransformAnimEvaluator.h>
#include <nw/gfx/gfx_TransformAnimInterpolator.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(TransformAnimInterpolator, AnimInterpolator);

//----------------------------------------------------------
const anim::AnimResult*
TransformAnimInterpolator::GetResult(
    void* target,
    int memberIdx
) const
{
    //----------------------------------------------------------
    // Get the blend operation
    const anim::AnimBlendOp* blendOp = m_AnimGroup->GetBlendOperation(memberIdx);

    const bit32 TRANSFORM_FLAG_MASK =
        CalculatedTransform::FLAG_IS_IDENTITY |
        CalculatedTransform::FLAG_IS_ROTATE_TRANSLATE_ZERO |
        CalculatedTransform::FLAG_IS_ROTATE_ZERO |
        CalculatedTransform::FLAG_IS_TRANSLATE_ZERO |
        CalculatedTransform::FLAG_IS_SCALE_ONE |
        CalculatedTransform::FLAG_IS_UNIFORM_SCALE;

    if (m_IsOldMethod)
    {
        // TODO: Clean up the blocks here.

        //----------------------------------------------------------
        // Find the total weight of valid child animations for each SRT
        float weightSums[3] = { 0.0f, 0.0f, 0.0f };
        for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        {
            if (m_AnimObjects[animIdx] == NULL)
            {
                continue;
            }
            const float weight = m_Weights[animIdx];
            if (!AnimWeightNearlyEqualZero(weight))
            {
                const AnimObject* animObj = m_AnimObjects[animIdx];
                if (animObj->HasMemberAnim(memberIdx))
                {
                    const TransformAnimEvaluator* evaluator =
                        ut::DynamicCast<const TransformAnimEvaluator*>(animObj);
                    if (evaluator != NULL)
                    {
                        if (!evaluator->GetIsScaleDisabled())
                        {
                            weightSums[0] += weight;
                        }
                        if (!evaluator->GetIsRotateDisabled())
                        {
                            weightSums[1] += weight;
                        }
                        if (!evaluator->GetIsTranslateDisabled())
                        {
                            weightSums[2] += weight;
                        }
                    }
                    else
                    {
                        weightSums[0] += weight;
                        weightSums[1] += weight;
                        weightSums[2] += weight;
                    }
                }
            }
        }

        // When there are no valid child animations, NULL is returned without changing target
        // 
        if (TransformAnimEvaluator::CheckWeightsNearlyZero(weightSums))
        {
            return NULL;
        }

        // When the weight total is not 1.0, calculate a normalized scale
        float weightNormalizeScale[3] =
        {
            GetAnimWeightNormalizeScale(weightSums[0]),
            GetAnimWeightNormalizeScale(weightSums[1]),
            GetAnimWeightNormalizeScale(weightSums[2])
        };

        //----------------------------------------------------------
        // If there is operational space for interpolation it should occur there, and after evacuating target's CONVERTED flag, it should be turned on
        // 
        CalculatedTransform* transform =
            reinterpret_cast<CalculatedTransform*>(target);
        const bit32 flagsBak = transform->GetFlags();
        const bool convertedBak = transform->IsEnabledFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);
        transform->EnableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);

        //----------------------------------------------------------
        // Raise the target's IGNORE flag, and set to unwritten state
        transform->EnableFlags(CalculatedTransform::FLAG_IS_IGNORE_ALL);

        //----------------------------------------------------------
        // Loop for all child animations
        CalculatedTransform workResult;
        bool written = false;
        bool firstRotateFlag = true;
        math::MTX34 firstRotateMtx;
        bit32 transformFlag = TRANSFORM_FLAG_MASK;
        //for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
        {
            const AnimObject* animObj = m_AnimObjects[animIdx];
            if (animObj == NULL)
            {
                continue;
            }

            const float childWeight = m_Weights[animIdx];
            NW_ASSERT(childWeight >= 0.0f);
            float srcWeights[3] =
            {
                childWeight * weightNormalizeScale[0],
                childWeight * weightNormalizeScale[1],
                childWeight * weightNormalizeScale[2]
            };
            TransformAnimEvaluator::DisableSRTWeightsIfNeeded(srcWeights, animObj);

            if (!TransformAnimEvaluator::CheckWeightsNearlyZero(srcWeights))
            {
                // Reflect the evacuated CONVERTED flag to work
                // If the child is a Blender that handles CONVERTED, expect the CONVERTED results to be returned as-is
                // When the child is an Evaluator, the flag is returned as off but there is no problem for conversion to occur within blendOp
                workResult.EnableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND, convertedBak);

                // Evaluations
                const anim::AnimResult* childResult =
                    animObj->GetResult(&workResult, memberIdx);

                if (childResult != NULL)
                {
                    written = true;
                    const bool evaluatorFlag =
                        ut::DynamicCast<const TransformAnimEvaluator*>(animObj) != NULL;

                    // Save the first animation's rotation matrix for when the row of the blend result's rotation matrix is a zero vector
                    // 
                    if (evaluatorFlag && firstRotateFlag &&
                        !workResult.IsEnabledFlags(CalculatedTransform::FLAG_IS_IGNORE_ROTATE))
                    {
                        firstRotateFlag = false;
                        firstRotateMtx = workResult.TransformMatrix();
                    }

                    // Blend
                    if (!blendOp->Blend(reinterpret_cast<anim::AnimResult*>(transform),
                        NULL, childResult, srcWeights))
                    {
                        break;
                    }

                    transformFlag &= workResult.GetFlags();
                }
            }
        }

        //----------------------------------------------------------
        // Restore target's CONVERTED flag
        if (!convertedBak)
        {
            transform->DisableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);
        }

        if (!written) // No valid child animation
        {
            transform->RestoreFlags(CalculatedTransform::FLAG_IS_IGNORE_ALL, flagsBak);
            return NULL;
        }

        //----------------------------------------------------------
        // Execute if there are processes after blending
        if (!convertedBak && blendOp->HasPostBlend())
        {
            if (!blendOp->PostBlend(reinterpret_cast<anim::AnimResult*>(transform), NULL))
            {
                // If the row of the blend result's rotation matrix is a zero vector, use the first animation's rotation matrix.
                // 
                // When there is a problem with the quality of the results, you must avoid this by slightly shifting the original data.
                // 
                transform->AdjustZeroRotateMatrix(!firstRotateFlag, firstRotateMtx);
            }
        }

        transform->RestoreFlags(TRANSFORM_FLAG_MASK, transformFlag);
    }
    else
    {
        bool isValidAnim = false;
        for (int animIdx = 0; animIdx < m_AnimObjects.Size(); ++animIdx)
        {
            if (m_AnimObjects[animIdx] && m_AnimObjects[animIdx]->HasMemberAnim(memberIdx))
            {
                isValidAnim = true;
                break;
            }
        }

        // When there are no valid child animations, NULL is returned without changing target
        // 
        if (!isValidAnim)
        {
            return NULL;
        }

        // When normalization of the animation weight is needed, perform it.
        if (m_IsWeightDirty && m_IsWeightNormalizationEnabled)
        {
            NormalizeWeight();
        }

        //----------------------------------------------------------
        // If there is operational space for interpolation it should occur there, and after evacuating target's CONVERTED flag, it should be turned on
        // 
        CalculatedTransform* transform =
            reinterpret_cast<CalculatedTransform*>(target);
        const bit32 flagsBak = transform->GetFlags();
        const bool convertedBak = transform->IsEnabledFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);
        transform->EnableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);

        //----------------------------------------------------------
        // Raise the target's IGNORE flag, and set to unwritten state
        transform->EnableFlags(CalculatedTransform::FLAG_IS_IGNORE_ALL);

        //----------------------------------------------------------
        // Loop for all child animations
        CalculatedTransform workResult;
        bool firstRotateFlag = true;
        math::MTX34 firstRotateMtx;
        bit32 transformFlag = TRANSFORM_FLAG_MASK;
        for (int animIdx = m_AnimObjects.Size() - 1; animIdx >= 0; --animIdx)
        {
            const AnimObject* animObj = m_AnimObjects[animIdx];

            const float childWeight = m_NormalizedWeights[animIdx];
            NW_ASSERT(childWeight >= 0.0f);
            const float srcWeights[3] =
            {
                childWeight,
                childWeight,
                childWeight
            };

            if (!TransformAnimEvaluator::CheckWeightsNearlyZero(srcWeights))
            {
                // Reflect the evacuated CONVERTED flag to work
                // If the child is a Blender that handles CONVERTED, expect the CONVERTED results to be returned as-is
                // When the child is an Evaluator, the flag is returned as off but there is no problem for conversion to occur within blendOp
                workResult.EnableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND, convertedBak);

                // Evaluations
                const anim::AnimResult* childResult = NULL;
                if (animObj != NULL)
                {
                    childResult = animObj->GetResult(&workResult, memberIdx);
                }

                if (childResult == NULL)
                {
                    const math::Transform3* originalValue =
                        static_cast<const math::Transform3*>(GetAnimGroup()->GetOriginalValue(memberIdx));
                    workResult.SetTransform(*originalValue);
                    workResult.UpdateScaleFlags();
                    workResult.UpdateRotateFlagsStrictly();
                    workResult.UpdateTranslateFlags();
                    workResult.UpdateCompositeFlags();
                    childResult = reinterpret_cast<anim::AnimResult*>(&workResult);
                }

                const bool evaluatorFlag =
                    ut::DynamicCast<const TransformAnimEvaluator*>(animObj) != NULL;

                // Save the first animation's rotation matrix for when the row of the blend result's rotation matrix is a zero vector
                // 
                if (evaluatorFlag && firstRotateFlag &&
                    !workResult.IsEnabledFlags(CalculatedTransform::FLAG_IS_IGNORE_ROTATE))
                {
                    firstRotateFlag = false;
                    firstRotateMtx = workResult.TransformMatrix();
                }

                // Blend
                if (!blendOp->Blend(reinterpret_cast<anim::AnimResult*>(transform),
                    NULL, childResult, srcWeights))
                {
                    break;
                }

                transformFlag &= workResult.GetFlags();
            }
        }

        //----------------------------------------------------------
        // Restore target's CONVERTED flag
        if (!convertedBak)
        {
            transform->DisableFlags(CalculatedTransform::FLAG_CONVERTED_FOR_BLEND);
        }

        //----------------------------------------------------------
        // Execute if there are processes after blending
        if (!convertedBak && blendOp->HasPostBlend())
        {
            if (!blendOp->PostBlend(reinterpret_cast<anim::AnimResult*>(transform), NULL))
            {
                // If the row of the blend result's rotation matrix is a zero vector, use the first animation's rotation matrix.
                // 
                // When there is a problem with the quality of the results, you must avoid this by slightly shifting the original data.
                // 
                transform->AdjustZeroRotateMatrix(!firstRotateFlag, firstRotateMtx);
            }
        }

        transform->RestoreFlags(TRANSFORM_FLAG_MASK, transformFlag);
    }

    // Cast AnimResult to the target and return (use with the parent's blender)
    return reinterpret_cast<anim::AnimResult*>(target);
}

} // namespace gfx
} // namespace nw
