/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_AnimGroup.h>
#include <nw/gfx/gfx_CalculatedTransform.h>
#include <nw/gfx/gfx_TransformAnimBlendOp.h>
#include <nw/gfx/gfx_TransformAnimEvaluator.h>
#include <nw/gfx/gfx_TransformAnimInterpolatorLite.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(TransformAnimInterpolatorLite, AnimInterpolator);

namespace
{
    NW_INLINE bool TransformBlend(
        CalculatedTransform* dstX,
        const CalculatedTransform* srcX,
        float weight,
        bool isFirst)
    {
        const math::MTX34& srcMtx = srcX->TransformMatrix();
        math::MTX34& dstMtx = dstX->DirectTransformMatrix();

        f32 temp00 = srcMtx.f._00;
        f32 temp01 = srcMtx.f._01;
        f32 temp02 = srcMtx.f._02;
        f32 temp03 = srcMtx.f._03;
        f32 temp10 = srcMtx.f._10;
        f32 temp11 = srcMtx.f._11;
        f32 temp12 = srcMtx.f._12;
        f32 temp13 = srcMtx.f._13;
        f32 temp23 = srcMtx.f._23;
        f32 temps0 = srcX->Scale().x;
        f32 temps1 = srcX->Scale().y;
        f32 temps2 = srcX->Scale().z;

        // In the process after blending, the rotation matrix is ortho-normalized, so the second line of calculation is unnecessary

        if (isFirst)
        {
            temp00 = temp00 * weight;
            temp01 = temp01 * weight;
            temp02 = temp02 * weight;
            temp10 = temp10 * weight;
            temp11 = temp11 * weight;
            temp12 = temp12 * weight;

            temp03 = temp03 * weight;
            temp13 = temp13 * weight;
            temp23 = temp23 * weight;

            temps0 = temps0 * weight;
            temps1 = temps1 * weight;
            temps2 = temps2 * weight;
        }
        else
        {
            temp00 = dstMtx.f._00 + temp00 * weight;
            temp01 = dstMtx.f._01 + temp01 * weight;
            temp02 = dstMtx.f._02 + temp02 * weight;
            temp10 = dstMtx.f._10 + temp10 * weight;
            temp11 = dstMtx.f._11 + temp11 * weight;
            temp12 = dstMtx.f._12 + temp12 * weight;

            temp03 = dstMtx.f._03 + temp03 * weight;
            temp13 = dstMtx.f._13 + temp13 * weight;
            temp23 = dstMtx.f._23 + temp23 * weight;

            temps0 = dstX->Scale().x + temps0 * weight;
            temps1 = dstX->Scale().y + temps1 * weight;
            temps2 = dstX->Scale().z + temps2 * weight;
        }

        dstMtx.f._00 = temp00;
        dstMtx.f._01 = temp01;
        dstMtx.f._02 = temp02;
        dstMtx.f._03 = temp03;
        dstMtx.f._10 = temp10;
        dstMtx.f._11 = temp11;
        dstMtx.f._12 = temp12;
        dstMtx.f._13 = temp13;
        dstMtx.f._23 = temp23;
        dstX->DirectScale().Set(temps0, temps1, temps2);

        return true;
    }

    NW_INLINE bool PostTransformBlend(CalculatedTransform* dstX)
    {
        // Ortho-normalizes the rotation matrix.

        // Calculate the third row of the rotation matrix with a vector product
        math::VEC3* v0 = reinterpret_cast<math::VEC3*>(&dstX->DirectTransformMatrix().m[0]);
        math::VEC3* v1 = reinterpret_cast<math::VEC3*>(&dstX->DirectTransformMatrix().m[1]);
        math::VEC3* v2 = reinterpret_cast<math::VEC3*>(&dstX->DirectTransformMatrix().m[2]);

        math::VEC3Cross(v2, v0, v1);

        float lengthSquareV0 = v0->LengthSquare();
        float lengthSquareV2 = v2->LengthSquare();

        if (lengthSquareV0 < CalculatedTransform::s_VecSquareLenTol ||
            lengthSquareV2 < CalculatedTransform::s_VecSquareLenTol)
        {
            // The rotation axis is the zero vector.
            return false;
        }
        else
        {
            f32 temp00 = v0->x;
            f32 temp01 = v0->y;
            f32 temp02 = v0->z;
            f32 temp10;
            f32 temp11;
            f32 temp12;
            f32 temp20 = v2->x;
            f32 temp21 = v2->y;
            f32 temp22 = v2->z;
            f32 tempr0;
            f32 tempr2;

            tempr0 = 1.0f / nn::math::FSqrt(lengthSquareV0);
            tempr2 = 1.0f / nn::math::FSqrt(lengthSquareV2);
            temp00 = temp00 * tempr0;
            temp01 = temp01 * tempr0;
            temp02 = temp02 * tempr0;
            temp20 = temp20 * tempr2;
            temp21 = temp21 * tempr2;
            temp22 = temp22 * tempr2;

            // Calculate the cross product
            temp10 = temp21 * temp02 - temp22 * temp01;
            temp11 = temp22 * temp00 - temp20 * temp02;
            temp12 = temp20 * temp01 - temp21 * temp00;

            v0->x = temp00;
            v0->y = temp01;
            v0->z = temp02;
            v1->x = temp10;
            v1->y = temp11;
            v1->z = temp12;
            v2->x = temp20;
            v2->y = temp21;
            v2->z = temp22;
        }

        return true;
    }
}

//----------------------------------------------------------
const anim::AnimResult*
TransformAnimInterpolatorLite::GetResult(
    void* target,
    int memberIdx
) const
{
    const bit32 TRANSFORM_FLAG_MASK =
        CalculatedTransform::FLAG_IS_IDENTITY |
        CalculatedTransform::FLAG_IS_ROTATE_TRANSLATE_ZERO |
        CalculatedTransform::FLAG_IS_ROTATE_ZERO |
        CalculatedTransform::FLAG_IS_TRANSLATE_ZERO |
        CalculatedTransform::FLAG_IS_SCALE_ONE |
        CalculatedTransform::FLAG_IS_UNIFORM_SCALE;

    bool isValidAnim = false;
    const int animObjectCount = m_AnimObjects.Size();
    for (int animIdx = 0; animIdx < animObjectCount; ++animIdx)
    {
        NW_ASSERT(m_AnimObjects[animIdx] == NULL || nw::ut::IsTypeOf<const TransformAnimEvaluator>(m_AnimObjects[animIdx]));
        const TransformAnimEvaluator* transformAnimObj =
            reinterpret_cast<const TransformAnimEvaluator*>(m_AnimObjects[animIdx]);

        if (transformAnimObj && transformAnimObj->TransformAnimEvaluator::HasMemberAnim(memberIdx))
        {
            isValidAnim = true;
            break;
        }
    }

    // When there are no valid child animations, NULL is returned without changing target
    // 
    if (!isValidAnim)
    {
        return NULL;
    }

    //----------------------------------------------------------
    // Since all child animations are TransformAnimEvaluator, there is no need to consider the target CONVERTED flag.
    // 

    CalculatedTransform* transform =
        reinterpret_cast<CalculatedTransform*>(target);
    transform->EnableFlags(
        CalculatedTransform::FLAG_IS_IGNORE_ALL |
        CalculatedTransform::FLAG_IS_DIRTY);

    //----------------------------------------------------------
    // Loop for all child animations
    CalculatedTransform workResult;
    bit32 transformFlag = TRANSFORM_FLAG_MASK;
    bool isFirst = true;

    for (int animIdx = animObjectCount - 1; animIdx >= 0; --animIdx)
    {
        NW_ASSERTMSG(m_AnimObjects[animIdx] == NULL || nw::ut::IsTypeOf<const TransformAnimEvaluator>(m_AnimObjects[animIdx]),
            "Except TransformAnimEvaluator can't set to TransformAnimInterpolatorLite");
        const TransformAnimEvaluator* transformAnimObj =
            reinterpret_cast<const TransformAnimEvaluator*>(m_AnimObjects[animIdx]);

        const float childWeight = m_NormalizedWeights[animIdx];
        NW_ASSERT(childWeight >= 0.0f);

        // Evaluations
        const anim::AnimResult* childResult = NULL;
        if (transformAnimObj != NULL)
        {
            childResult = transformAnimObj->TransformAnimEvaluator::GetResult(&workResult, memberIdx);
        }

        if (childResult == NULL)
        {
            const math::Transform3* originalValue =
                static_cast<const math::Transform3*>(GetAnimGroup()->GetOriginalValue(memberIdx));
            workResult.SetTransform(*originalValue);
            workResult.UpdateScaleFlags();
            workResult.UpdateRotateFlagsStrictly();
            workResult.UpdateTranslateFlags();
            workResult.UpdateCompositeFlags();
        }

        TransformBlend(transform, &workResult, childWeight, isFirst);
        isFirst = false;

        transformFlag &= workResult.GetFlags();
    }

    //----------------------------------------------------------
    // Since all child animations are TransformAnimEvaluator, turn the target CONVERTED flag off.
    // 
    // Also, turn the IGNORE flag off since SRT is always written.
    transform->DisableFlags(
        CalculatedTransform::FLAG_CONVERTED_FOR_BLEND |
        CalculatedTransform::FLAG_IS_IGNORE_SCALE |
        CalculatedTransform::FLAG_IS_IGNORE_ROTATE |
        CalculatedTransform::FLAG_IS_IGNORE_TRANSLATE);

    if (!PostTransformBlend(transform))
    {
        // If a row in the rotation matrix for the blend result is a zero vector, sets a unit matrix.
        // 
        // When there is a problem with the quality of the results, you must avoid this by slightly shifting the original data.
        // 
        *transform = CalculatedTransform::Identity();
    }

    transform->RestoreFlags(TRANSFORM_FLAG_MASK, transformFlag);

    // Cast AnimResult to the target and return (use with the parent's blender)
    return reinterpret_cast<anim::AnimResult*>(target);
}

} // namespace gfx
} // namespace nw
