/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/gfx/gfx_TransformNode.h>
#include <nw/gfx/gfx_ISceneVisitor.h>

#include <nw/ut/ut_ResUtil.h>
#include <nw/ut/ut_ResDictionary.h>

namespace nw
{
namespace gfx
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(TransformNode, SceneNode);

//----------------------------------------
TransformNode::TransformNode(
    os::IAllocator* allocator,
    ResTransformNode resObj,
    const TransformNode::Description& description)
: SceneNode(allocator, resObj, description),
    m_InverseWorldMatrix(math::MTX34::Identity()),
    m_IsInverseWorldMatrixValid(false),
    m_PostCalculateWorldMatrixSignal(NULL),
    m_IsBranchWorldMatrixCalculationEnabled(true),
    m_Description(description)
{
    if (resObj.IsValid())
    {
        this->m_Transform.SetTransform(resObj.GetTransform());
        this->m_WorldMatrix = resObj.GetWorldMatrix();
    }
    else
    {
        math::MTX34Identity(&this->m_WorldMatrix);
    }
}

//----------------------------------------
TransformNode*
TransformNode::DynamicBuilder::Create(
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    void* memory = allocator->Alloc(sizeof(TransformNode));
    NW_NULL_ASSERT(memory);

    TransformNode* node = new(memory) TransformNode(
        allocator,
        ResTransformNode(),
        m_Description);

    Result result = node->Initialize(allocator);

    NW_ASSERT(result.IsSuccess());

    return node;
}

//----------------------------------------
TransformNode*
TransformNode::Create(
    SceneNode* parent,
    ResSceneObject resource,
    const TransformNode::Description& description,
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    ResTransformNode resNode = ResStaticCast<ResTransformNode>(resource);

    void* memory = allocator->Alloc(sizeof(TransformNode));
    NW_NULL_ASSERT(memory);

    TransformNode* node = new(memory) TransformNode(
        allocator,
        resNode,
        description);

    Result result = node->Initialize(allocator);

    NW_ASSERT(result.IsSuccess());

    if (parent)
    {
        bool isAttached = parent->AttachChild(node);
        NW_ASSERT(isAttached);
    }
    return node;
}


//----------------------------------------
Result
TransformNode::CreateCallbacks(os::IAllocator* allocator)
{
    Result result = INITIALIZE_RESULT_OK;

    if (m_Description.isFixedSizeMemory)
    {
        if (m_Description.maxCallbacks == 0)
        {
            m_PostCalculateWorldMatrixSignal =
                CalculateMatrixSignal::CreateInvalidateSignal(allocator);
        }
        else
        {
            m_PostCalculateWorldMatrixSignal =
                CalculateMatrixSignal::CreateFixedSizedSignal(m_Description.maxCallbacks, allocator);
        }
    }
    else
    {
        m_PostCalculateWorldMatrixSignal =
            CalculateMatrixSignal::CreateVariableSizeSignal(allocator);
    }

    // When there was a failure in memory allocation for the dynamic array
    if (m_PostCalculateWorldMatrixSignal == NULL)
    {
        result |= Result::MASK_FAIL_BIT;
    }

    return result;
}

//----------------------------------------
void
TransformNode::Accept(
    ISceneVisitor* visitor
)
{
    visitor->VisitTransformNode(this);
    AcceptChildren(visitor);
}

//----------------------------------------
const math::MTX34&
TransformNode::InverseWorldMatrix() const
{
    if (!m_IsInverseWorldMatrixValid)
    {
        NW_FAILSAFE_IF(math::MTX34Inverse(&m_InverseWorldMatrix, m_WorldMatrix) == 0)
        {
            m_InverseWorldMatrix = math::MTX34::Identity();
        }

        m_IsInverseWorldMatrixValid = true;
    }

    return m_InverseWorldMatrix;
}

//----------------------------------------
void
TransformNode::InvalidateInverseWorldMatrix()
{
    m_IsInverseWorldMatrixValid = false;
}

//----------------------------------------
void
TransformNode::UpdateTransform(
    WorldMatrixUpdater* worldMatrixUpdater,
    SceneContext* sceneContext)
{
    // Calls callback even when CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED is set to 0.
    if (this->Transform().IsEnabledFlags(CalculatedTransform::FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED))
    {
        // Skip calculations referencing the Dirty flag set by Transform
        if (this->IsEnabledResults(SceneNode::FLAG_IS_DIRTY))
        {
            const CalculatedTransform* parentWorldTransform;
            const CalculatedTransform* parentLocalTransform;

            if (this->GetParent() == NULL)
            {
                parentWorldTransform = &CalculatedTransform::Identity();
                parentLocalTransform = &CalculatedTransform::Identity();
            }
            else
            {
                parentWorldTransform = &this->GetParent()->TrackbackWorldTransform();
                parentLocalTransform = &this->GetParent()->TrackbackLocalTransform();
            }

            worldMatrixUpdater->UpdateBasic(
                &this->WorldMatrix(),
                &this->WorldTransform(),
                this->Transform(),
                *parentWorldTransform,
                *parentLocalTransform);

            this->InvalidateInverseWorldMatrix();
        }
    }
    else
    {
        // Since FLAG_IS_WORLDMATRIX_CALCULATION_ENABLED is handled as calculated, the Dirty flag for inheritance is set to 0.
        // 
        this->DisableTraversalResults(SceneNode::FLAG_IS_DIRTY);
    }

    this->UpdateDirection();

    // Make the Dirty flag 0.
    this->Transform().DisableFlags(CalculatedTransform::FLAG_IS_DIRTY);

    this->PostCalculateWorldMatrixSignal()(this, sceneContext);
}


//----------------------------------------
Result
TransformNode::Initialize(os::IAllocator* allocator)
{
    Result result = INITIALIZE_RESULT_OK;

    result |= SceneNode::Initialize(allocator);
    NW_ENSURE_AND_RETURN(result);

    result |= CreateCallbacks(allocator);
    NW_ENSURE_AND_RETURN(result);

    return result;
}

} // namespace gfx
} // namespace nw
