/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../precompiled.h"

#include <nw/ut/ut_ResUtil.h>
#include <nw/ut/ut_ResDictionary.h>
#include <nw/gfx/gfx_Common.h>
#include <nw/gfx/gfx_SceneObject.h>
#include <nw/gfx/res/gfx_ResTexture.h>
#include <nw/gfx/gfx_ShaderProgram.h>
#include <nw/gfx/res/gfx_ResGraphicsFile.h>

#include <nn/gx.h>

namespace nw {
namespace gfx {
namespace res {

typedef void (*SetupFunc)(ResTexture resTex);

static void ResImageTexture_Setup(ResTexture resTex);
static void ResCubeTexture_Setup(ResTexture resTex);
static void ResReferenceTexture_Setup(ResTexture resTex);
static void ResProceduralTexture_Setup(ResTexture resTex);
static void ResShadowTexture_Setup(ResTexture resTex);

static void ResImageTexture_Cleanup(ResTexture resTex);
static void ResCubeTexture_Cleanup(ResTexture resTex);
static void ResReferenceTexture_Cleanup(ResTexture resTex);
static void ResProceduralTexture_Cleanup(ResTexture resTex);
static void ResShadowTexture_Cleanup(ResTexture resTex);

static SetupFunc s_TextureSetupTable[] =
{
    ResImageTexture_Setup,
    ResCubeTexture_Setup,
    ResReferenceTexture_Setup,
    ResProceduralTexture_Setup,
    ResShadowTexture_Setup
};

static SetupFunc s_TextureCleanupTable[] =
{
    ResImageTexture_Cleanup,
    ResCubeTexture_Cleanup,
    ResReferenceTexture_Cleanup,
    ResProceduralTexture_Cleanup,
    ResShadowTexture_Cleanup
};

namespace {
    // The table for obtaining pixel formats.
    static const uint GL_FORMAT_TABLE[] =
    {
        GL_RGBA_NATIVE_DMP,                 // Each pixel is 4 bytes; R, G, B, and A are 8 bits each.
        GL_RGB_NATIVE_DMP,                  // Each pixel is 3 bytes; R, G, and B are 8 bits each.
        GL_RGBA_NATIVE_DMP,                 // Each pixel is 2 bytes;  R, G, and B are 5 bits each and A is 1 bit.
        GL_RGB_NATIVE_DMP,                  // Each pixel is 2 bytes;  R, G, and  B are 5, 6, and 5 bits respectively.
        GL_RGBA_NATIVE_DMP,                 // Each pixel is 2 bytes; R, G, B, and A are 4 bits each.
        GL_LUMINANCE_ALPHA_NATIVE_DMP,      // Format where 1 pixel is 2 bytes, and Alpha and Luminance are 8 bits each.
        GL_HILO8_DMP_NATIVE_DMP,            // Each pixel is 2 bytes; X and Y are 8 bits each.
        GL_LUMINANCE_NATIVE_DMP,            // Each pixel is 1 byte; Luminance is 8 bits.
        GL_ALPHA_NATIVE_DMP,                // Each pixel is 1 byte; Alpha is 8 bits.
        GL_LUMINANCE_ALPHA_NATIVE_DMP,      // Format where 1 pixel is 1 bytes, and Alpha and Luminance are 4 bits each.
        GL_LUMINANCE_NATIVE_DMP,            // Format  where 1 pixel is 4 bits and Luminance is 4 bits.
        GL_ALPHA_NATIVE_DMP,                // Format  where 1 pixel is 4 bits and Alpha is 4 bits.
        GL_ETC1_RGB8_NATIVE_DMP,            // Compressed format. Each pixel is equivalent to 4 bytes.
        GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP    // Compressed format. Each pixel is equivalent to 8 bytes.
    };

    static const uint GL_FORMAT_TYPE_TABLE[] =
    {
        GL_UNSIGNED_BYTE,                   // Each pixel is 4 bytes; R, G, B, and A are 8 bits each.
        GL_UNSIGNED_BYTE,                   // Each pixel is 3 bytes; R, G, and B are 8 bits each.
        GL_UNSIGNED_SHORT_5_5_5_1,          // Each pixel is 2 bytes;  R, G, and B are 5 bits each and A is 1 bit.
        GL_UNSIGNED_SHORT_5_6_5,            // Each pixel is 2 bytes;  R, G, and  B are 5, 6, and 5 bits respectively.
        GL_UNSIGNED_SHORT_4_4_4_4,          // Each pixel is 2 bytes; R, G, B, and A are 4 bits each.
        GL_UNSIGNED_BYTE,                   // Format where 1 pixel is 2 bytes, and Alpha and Luminance are 8 bits each.
        GL_UNSIGNED_BYTE,                   // Each pixel is 2 bytes; X and Y are 8 bits each.
        GL_UNSIGNED_BYTE,                   // Each pixel is 1 byte; Luminance is 8 bits.
        GL_UNSIGNED_BYTE,                   // Each pixel is 1 byte; Alpha is 8 bits.
        GL_UNSIGNED_BYTE_4_4_DMP,           // 1 pixel is 1 byte, this is a format where Alpha and Luminance each are 4 bits
        GL_UNSIGNED_4BITS_DMP,              // Format  where 1 pixel is 4 bits and Luminance is 4 bits.
        GL_UNSIGNED_4BITS_DMP,              // Format  where 1 pixel is 4 bits and Alpha is 4 bits.
        0,                                  // Compressed format. Each pixel is equivalent to 4 bytes.
        0                                   // Compressed format. Each pixel is equivalent to 8 bytes.
    };

    const f32 BYTESIZE_4 = 4.0f;
    const f32 BYTESIZE_3 = 3.0f;
    const f32 BYTESIZE_2 = 2.0f;
    const f32 BYTESIZE_1 = 1.0f;
    const f32 BYTESIZE_HALF = 0.5f;
    static const f32 BYTESIZE_PER_PIXEL_TABLE[] =
    {
        BYTESIZE_4,     // FORMAT_HW_RGBA8
        BYTESIZE_3,     // FORMAT_HW_RGB8
        BYTESIZE_2,     // FORMAT_HW_RGBA5551
        BYTESIZE_2,     // FORMAT_HW_RGB565
        BYTESIZE_2,     // FORMAT_HW_RGBA4
        BYTESIZE_2,     // FORMAT_HW_LA8
        BYTESIZE_2,     // FORMAT_HW_HILO8
        BYTESIZE_1,     // FORMAT_HW_L8
        BYTESIZE_1,     // FORMAT_HW_A8
        BYTESIZE_1,     // FORMAT_HW_LA4
        BYTESIZE_HALF,  // FORMAT_HW_L4
        BYTESIZE_HALF,  // FORMAT_HW_A4
        BYTESIZE_HALF,  // FORMAT_HW_ETC1
        BYTESIZE_1      // FORMAT_HW_ETC1A4
    };
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
static void
TexImage2D( u32 texID, ResPixelBasedImage resImage, u32 loadFlag )
{
    u32 size = resImage.GetImageDataCount();
    GLenum transtype = loadFlag & 0xFFFF0000;
    const u32 NN_GX_MEM_MASK = 0x00030000;
    void* address = NULL;

    if (size == 0)
    {
        size = resImage.GetWidth() * resImage.GetHeight() * resImage.GetBitsPerPixel() / 8;
    }

    switch (transtype)
    {
    case (NN_GX_MEM_FCRAM | GL_NO_COPY_FCRAM_DMP):
        nngxUpdateBuffer( resImage.GetImageData(), size );
        break;

    case (NN_GX_MEM_VRAMA | GL_NO_COPY_FCRAM_DMP):
    case (NN_GX_MEM_VRAMB | GL_NO_COPY_FCRAM_DMP):
        {
            GLuint area = (transtype & NN_GX_MEM_MASK);
            address = __dmpgl_allocator(area, NN_GX_MEM_TEXTURE, texID, size);
            if ( resImage.GetImageData() )
            {
                nngxAddVramDmaCommand( resImage.GetImageData(), address, size );
            }

            resImage.SetLocationAddress( address );
            resImage.ref().m_MemoryArea = area;
        }
        break;

    case (NN_GX_MEM_FCRAM | GL_COPY_FCRAM_DMP):
        {
            address = __dmpgl_allocator(NN_GX_MEM_FCRAM, NN_GX_MEM_TEXTURE, texID, size);
            if ( resImage.GetImageData() )
            {
                nw::os::MemCpy( address, resImage.GetImageData(), size );
                nngxUpdateBuffer( address, size );
            }

            resImage.SetLocationAddress( address );
            resImage.ref().m_MemoryArea = NN_GX_MEM_FCRAM;
        }
        break;

    case (NN_GX_MEM_VRAMA | GL_COPY_FCRAM_DMP):
    case (NN_GX_MEM_VRAMB | GL_COPY_FCRAM_DMP):
        NW_FATAL_ERROR("(GL_COPY_FCRAM_DMP | NN_GX_MEM_VRAM) is not supported!");
        break;
    default:
        break;
    }
}


//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
static void
DeleteImage2D( u32 texID, ResPixelBasedImage resImage )
{
    GLuint area = resImage.ref().m_MemoryArea;

    if (area != ResPixelBasedImageData::AREA_NO_MALLOC)
    {
        void* address = reinterpret_cast<void*>( resImage.GetLocationAddress() );
        __dmpgl_deallocator( area, NN_GX_MEM_TEXTURE, texID, address );
        resImage.SetLocationAddress( static_cast<u32>(NULL) );
    }
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
static inline void
SetTexImage2D( u32 texID, ResPixelBasedImage resImage, u32 loadFlag )
{
    NW_ASSERT( resImage.IsValid() );

    // When the address is already set, the data copy is not performed.
    if (resImage.GetLocationAddress() != NULL)
    {
        resImage.ref().m_MemoryArea = ResPixelBasedImageData::AREA_NO_MALLOC;
        return;
    }

    // Specifies loading to the texture memory.
    // The default is NO_COPY from FCRAM.
    if (loadFlag == 0)
    {
        loadFlag = NN_GX_MEM_FCRAM | GL_NO_COPY_FCRAM_DMP;
    }

    // When FCRAM, NO_COPY, alignment check. Device memory and 128-byte alignment.
    NW_ASSERT( (loadFlag != (NN_GX_MEM_FCRAM | GL_NO_COPY_FCRAM_DMP)) || resImage.CheckMemoryLocation() );

    TexImage2D( texID, resImage, loadFlag );
}


/*
  

  
 */
static void
ResImageTexture_Setup(ResTexture resTex)
{
    ResImageTexture resImgTex = ResDynamicCast<ResImageTexture>( resTex );

    NW_ASSERT( resImgTex.IsValid() );

    if ( resImgTex.GetTextureObject() != 0 )
    {
        return;
    }

    u32 texID = reinterpret_cast<u32>(resTex.ptr());
    resImgTex.SetTextureObject( texID );

    ResPixelBasedImage resImage = resImgTex.GetImage();

    u32 loadFlag = resImgTex.GetLocationFlag();

    SetTexImage2D( texID, resImage, loadFlag );
}

/*
  

  
 */
static void
ResShadowTexture_Setup(ResTexture resTex)
{
    ResShadowTexture resSdwTex = ResDynamicCast<ResShadowTexture>( resTex );

    NW_ASSERT( resSdwTex.IsValid() );

    if ( resSdwTex.GetTextureObject() != 0 )
    {
        return;
    }

    u32 texID = reinterpret_cast<u32>(resTex.ptr());
    resSdwTex.SetTextureObject( texID );

    ResPixelBasedImage resImage = resSdwTex.GetImage();

    u32 loadFlag = resSdwTex.GetLocationFlag();

    SetTexImage2D( texID, resImage, loadFlag );
}

/*
  

  
 */
static void
ResCubeTexture_Setup(ResTexture resTex)
{
    ResCubeTexture resCubeTex = ResDynamicCast<ResCubeTexture>( resTex );

    NW_ASSERT( resCubeTex.IsValid() );

    if ( resCubeTex.GetTextureObject() != 0 )
    {
        return;
    }

    u32 texID = reinterpret_cast<u32>(resTex.ptr());
    resCubeTex.SetTextureObject( texID );

    u32 loadFlag = resCubeTex.GetLocationFlag();

    // Set up the texture for each face.
    // Note that the same texture might be applied on separate faces.
    for ( int face = 0; face < ResCubeTexture::MAX_CUBE_FACE; ++face )
    {
        ResPixelBasedImage resImage = resCubeTex.GetImage( static_cast<ResCubeTexture::CubeFace>( face ) );

        SetTexImage2D( texID, resImage, loadFlag );
    }
}

/*
  

  
 */
static void
ResReferenceTexture_Setup(ResTexture resTex)
{
    ResReferenceTexture resRefTex = ResStaticCast<ResReferenceTexture>( resTex );

    NW_ASSERT( resRefTex.GetTargetTexture().IsValid() );

    switch(resRefTex.GetTargetTexture().ref().typeInfo)
    {
    case ResImageTexture::TYPE_INFO:
        {
            ResImageTexture_Setup(resRefTex.GetTargetTexture());
            break;
        }
    case ResCubeTexture::TYPE_INFO:
        {
            ResCubeTexture_Setup(resRefTex.GetTargetTexture());
            break;
        }
    case ResReferenceTexture::TYPE_INFO:
        {
            ResReferenceTexture_Setup(resRefTex.GetTargetTexture());
            break;
        }
    case ResProceduralTexture::TYPE_INFO:
        {
            ResProceduralTexture_Setup(resRefTex.GetTargetTexture());
            break;
        }
    case ResShadowTexture::TYPE_INFO:
        {
            ResShadowTexture_Setup(resRefTex.GetTargetTexture());
            break;
        }
    default:
        {
            NW_FATAL_ERROR("Unsupported texture type.\n");
        }
    }
}

/*
  

  
 */
static void
ResProceduralTexture_Setup(ResTexture resTex)
{
    NW_UNUSED_VARIABLE( resTex )
}

/*
  

  
 */
static void
ResImageTexture_Cleanup(ResTexture resTex)
{
    ResImageTexture resImgTex = ResDynamicCast<ResImageTexture>( resTex );
    NW_ASSERT( resImgTex.IsValid() );

    GLuint texID = resImgTex.GetTextureObject();
    if ( texID == 0 )
    {
        return;
    }

    ResPixelBasedImage resImage = resImgTex.GetImage();
    DeleteImage2D( texID, resImage );
    resImgTex.SetTextureObject( 0 );
}

/*
  

  
 */
static void
ResShadowTexture_Cleanup(ResTexture resTex)
{
    ResShadowTexture resSdwTex = ResDynamicCast<ResShadowTexture>( resTex );
    NW_ASSERT( resSdwTex.IsValid() );

    GLuint texID = resSdwTex.GetTextureObject();
    if ( texID == 0 )
    {
        return;
    }

    ResPixelBasedImage resImage = resSdwTex.GetImage();
    DeleteImage2D( texID, resImage );
    resSdwTex.SetTextureObject( 0 );
}

/*
  

  
 */
static void
ResCubeTexture_Cleanup(ResTexture resTex)
{
    ResCubeTexture resCubeTex = ResDynamicCast<ResCubeTexture>( resTex );
    NW_ASSERT( resCubeTex.IsValid() );

    GLuint texID = resCubeTex.GetTextureObject();
    if ( texID == 0 )
    {
        return;
    }

    for ( int face = 0; face < ResCubeTexture::MAX_CUBE_FACE; ++face )
    {
        ResPixelBasedImage resImage = resCubeTex.GetImage( static_cast<ResCubeTexture::CubeFace>( face ) );
        DeleteImage2D( texID, resImage );
    }

    resCubeTex.SetTextureObject( 0 );
}

/*
  

  
 */
static void
ResReferenceTexture_Cleanup(ResTexture resTex)
{
    ResReferenceTexture resRefTex = ResStaticCast<ResReferenceTexture>( resTex );

    resRefTex.ref().toTargetTexture.set_ptr(NULL);
}

/*
  

  
 */
static void
ResProceduralTexture_Cleanup(ResTexture resTex)
{
    NW_UNUSED_VARIABLE( resTex )
}

//-------------------------------------------------------------------------
Result
ResTexture::Setup(os::IAllocator* allocator, ResGraphicsFile graphicsFile)
{
    NW_UNUSED_VARIABLE(allocator);
    NW_UNUSED_VARIABLE(graphicsFile);

    NW_ASSERT( internal::ResCheckRevision( *this ) );

    Result result = RESOURCE_RESULT_OK;
    switch ( this->ref().typeInfo )
    {
    case ResImageTexture::TYPE_INFO:
        {
            s_TextureSetupTable[0]( *this );
        }
        break;
    case ResCubeTexture::TYPE_INFO:
        {
            s_TextureSetupTable[1]( *this );
        }
        break;
    case ResReferenceTexture::TYPE_INFO:
        {
            s_TextureSetupTable[2]( *this );
        }
        break;
    case ResProceduralTexture::TYPE_INFO:
        {
            s_TextureSetupTable[3]( *this );
        }
        break;
    case ResShadowTexture::TYPE_INFO:
        {
            s_TextureSetupTable[4]( *this );
        }
        break;
    default:
        {
            NW_FATAL_ERROR("Unsupported texture type.\n");
        }
    }

    return result;
}

//---------------------------------------------------------------------------
Result
ResTexture::Setup(os::IAllocator* allocator)
{
    return this->Setup(allocator, ResGraphicsFile(NULL));
}

/*
  
 */
void
ResTexture::Cleanup()
{
    switch ( this->ref().typeInfo )
    {
    case ResImageTexture::TYPE_INFO:
        {
            s_TextureCleanupTable[0]( *this );
        }
        break;
    case ResCubeTexture::TYPE_INFO:
        {
            s_TextureCleanupTable[1]( *this );
        }
        break;
    case ResReferenceTexture::TYPE_INFO:
        {
            s_TextureCleanupTable[2]( *this );
        }
        break;
    case ResProceduralTexture::TYPE_INFO:
        {
            s_TextureCleanupTable[3]( *this );
        }
        break;
    case ResShadowTexture::TYPE_INFO:
        {
            s_TextureCleanupTable[4]( *this );
        }
        break;
    default:
        {
            NW_FATAL_ERROR("Unsupported texture type.\n");
        }
    }
}

//----------------------------------------
ResImageTexture
ResImageTexture::DynamicBuilder::Create(
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    // Check for power of 2
    NW_ASSERT( (m_Description.height & (m_Description.height - 1)) == 0 );
    NW_ASSERT( (m_Description.width & (m_Description.width - 1)) == 0 );

    // Texture size check
    const s32 TEXTURE_SIZE_MAX = 1024;
    const s32 TEXTURE_SIZE_MIN =
        ( m_Description.formatHW == ResPixelBasedTexture::FORMAT_HW_ETC1 ? 16 : 8 );

    NW_MINMAX_ASSERT( m_Description.height, TEXTURE_SIZE_MIN, TEXTURE_SIZE_MAX );
    NW_MINMAX_ASSERT( m_Description.width, TEXTURE_SIZE_MIN, TEXTURE_SIZE_MAX );

    // Mipmap sizes must be specified at 1 or higher.
    NW_ASSERT(m_Description.mipmapSize > 0);

    // Memory allocation for ResImageTextureData
    void* textureDataMemory = allocator->Alloc(sizeof(ResImageTextureData));
    NW_NULL_ASSERT(textureDataMemory);
    ResImageTextureData* textureData = new(textureDataMemory) ResImageTextureData();

    //--------------------------------
    // Member initialization for ResSceneObjectData
    textureData->typeInfo = ResImageTexture::TYPE_INFO;
    textureData->m_Header.revision = ResImageTexture::BINARY_REVISION;
    textureData->m_Header.signature = ResImageTexture::SIGNATURE;
    textureData->toName.set_ptr(NULL);
    textureData->m_UserDataDicCount = 0;
    textureData->toUserDataDic.set_ptr(NULL);

    //--------------------------------
    // Member initialization for ResPixelBasedTextureData
    textureData->m_Height = m_Description.height;
    textureData->m_Width = m_Description.width;
    textureData->m_Format = GL_FORMAT_TABLE[m_Description.formatHW];
    textureData->m_FormatType = GL_FORMAT_TYPE_TABLE[m_Description.formatHW];
    textureData->m_MipmapSize = m_Description.mipmapSize;
    textureData->m_TextureObject = 0;
    textureData->m_LocationFlag = m_Description.locationFlag;
    textureData->m_FormatHW = m_Description.formatHW;

    // Memory allocation for ResPixelBasedImageData
    void* imageDataMemory = allocator->Alloc(sizeof(ResPixelBasedImageData));
    NW_NULL_ASSERT(imageDataMemory);
    ResPixelBasedImageData* imageData = new(imageDataMemory) ResPixelBasedImageData();

    imageData->m_DynamicAllocator = allocator;
    imageData->m_Height = m_Description.height;
    imageData->m_Width = m_Description.width;
    imageData->m_LocationAddress = m_Description.locationAddress;

    f32 bytesizePerPixel = BYTESIZE_PER_PIXEL_TABLE[m_Description.formatHW];
    imageData->m_BitsPerPixel = static_cast<u32>(bytesizePerPixel * 8);

    u32 byteCount = 0;
    s32 h = m_Description.height;
    s32 w = m_Description.width;

    // Calculate the memory size to allocate for textures.
    // When using mipmaps, calculate the memory allocation amount corresponding to the number of mipmaps.
    for (s32 mipmapSize = m_Description.mipmapSize; mipmapSize > 0; --mipmapSize)
    {
        byteCount += static_cast<u32>(bytesizePerPixel * h * w);

        if (h == TEXTURE_SIZE_MIN && w == TEXTURE_SIZE_MIN)
        {
            textureData->m_MipmapSize = m_Description.mipmapSize - mipmapSize + 1;
            break;
        }

        if (h > TEXTURE_SIZE_MIN)
        {
            h /= 2;
        }
        if (w > TEXTURE_SIZE_MIN)
        {
            w /= 2;
        }
    }

    // Memory allocation for pixel data
    // When isDynamicAllocation is specified, perform memory allocation
    if (m_Description.isDynamicAllocation)
    {
        const u8 ALIGNMENT = 128;
        NW_ASSERT(byteCount != 0);
        void* bufferMemory = allocator->Alloc(byteCount, ALIGNMENT);
        NW_NULL_ASSERT(bufferMemory);
        u8* buffer = static_cast<u8*>(bufferMemory);
        if (m_Description.executingMemoryFill)
        {
            std::fill_n(NW_CHECKED_ARRAY_ITERATOR(buffer, byteCount), byteCount, 0);
        }

        imageData->m_ImageDataTableCount = byteCount;
        imageData->toImageDataTable.set_ptr(buffer);
    }
    else
    {
        imageData->m_ImageDataTableCount = 0;
        imageData->toImageDataTable.set_ptr(NULL);
    }

    //--------------------------------
    // Member initialization for ResImageTextureData
    textureData->toImage.set_ptr(imageData);

    return ResImageTexture(textureData);
}

//----------------------------------------
ResCubeTexture
ResCubeTexture::DynamicBuilder::Create(
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    // Check for power of 2
    NW_ASSERT( (m_Description.width & (m_Description.width - 1)) == 0 );

    // Texture size check
    const s32 TEXTURE_SIZE_MAX = 1024;
    const s32 TEXTURE_SIZE_MIN =
        ( m_Description.formatHW == ResPixelBasedTexture::FORMAT_HW_ETC1 ? 16 : 8 );

    // For cube textures, FORMAT_HW_RGB8 cannot be used
    NW_ASSERT(m_Description.formatHW != ResPixelBasedTexture::FORMAT_HW_RGB8);

    NW_MINMAX_ASSERT( m_Description.width, TEXTURE_SIZE_MIN, TEXTURE_SIZE_MAX );

    // Mipmap sizes must be specified at 1 or higher.
    NW_ASSERT(m_Description.mipmapSize > 0);

    // Memory allocation for ResImageTextureData
    void* textureDataMemory = allocator->Alloc(sizeof(ResCubeTextureData));
    NW_NULL_ASSERT(textureDataMemory);
    ResCubeTextureData* textureData = new(textureDataMemory) ResCubeTextureData();

    //--------------------------------
    // Member initialization for ResSceneObjectData
    textureData->typeInfo = ResCubeTexture::TYPE_INFO;
    textureData->m_Header.revision = ResCubeTexture::BINARY_REVISION;
    textureData->m_Header.signature = ResCubeTexture::SIGNATURE;
    textureData->toName.set_ptr(NULL);
    textureData->m_UserDataDicCount = 0;
    textureData->toUserDataDic.set_ptr(NULL);

    //--------------------------------
    // Member initialization for ResPixelBasedTextureData
    textureData->m_Height = m_Description.width;
    textureData->m_Width = m_Description.width;
    textureData->m_Format = GL_FORMAT_TABLE[m_Description.formatHW];
    textureData->m_FormatType = GL_FORMAT_TYPE_TABLE[m_Description.formatHW];
    textureData->m_MipmapSize = m_Description.mipmapSize;
    textureData->m_TextureObject = 0;
    textureData->m_LocationFlag = m_Description.locationFlag;
    textureData->m_FormatHW = m_Description.formatHW;

    // Memory allocation for ResPixelBasedImageData
    void* imageDataMemory = allocator->Alloc(sizeof(ResPixelBasedImageData) * MAX_CUBE_FACE);
    NW_NULL_ASSERT(imageDataMemory);
    ResPixelBasedImageData* imageData = new(imageDataMemory) ResPixelBasedImageData[MAX_CUBE_FACE];

    f32 bytesizePerPixel = BYTESIZE_PER_PIXEL_TABLE[m_Description.formatHW];

    u32 byteCount = 0;
    s32 h = m_Description.width;
    s32 w = m_Description.width;

    // Calculate the memory size to allocate for textures.
    // When using mipmaps, calculate the memory allocation amount corresponding to the number of mipmaps.
    for (s32 mipmapSize = m_Description.mipmapSize; mipmapSize > 0; --mipmapSize)
    {
        byteCount += static_cast<u32>(bytesizePerPixel * h * w);

        if (h == TEXTURE_SIZE_MIN && w == TEXTURE_SIZE_MIN)
        {
            textureData->m_MipmapSize = m_Description.mipmapSize - mipmapSize + 1;
            break;
        }

        if (h > TEXTURE_SIZE_MIN)
        {
            h /= 2;
        }
        if (w > TEXTURE_SIZE_MIN)
        {
            w /= 2;
        }
    }

    uptr bufferMemory = 0;

    // Memory allocation for pixel data
    // When isDynamicAllocation is specified, perform memory allocation
    if (m_Description.isDynamicAllocation)
    {
        u32 totalByteCount = byteCount * MAX_CUBE_FACE;
        const u8 ALIGNMENT = 128;
        NW_ASSERT(totalByteCount != 0);
        bufferMemory = reinterpret_cast<uptr>(allocator->Alloc(totalByteCount, ALIGNMENT));
        NW_NULL_ASSERT(bufferMemory);
        u8* buffer = reinterpret_cast<u8*>(bufferMemory);
        if (m_Description.executingMemoryFill)
        {
            std::fill_n(NW_CHECKED_ARRAY_ITERATOR(buffer, totalByteCount), totalByteCount, 0);
        }
    }

    for (int i = 0 ; i < MAX_CUBE_FACE ; ++i)
    {
        // For safety's sake, set only the first layer of DynamicAllocator to be able to deallocate ImageDataTable.
        imageData[i].m_DynamicAllocator = (i == 0) ? allocator : NULL;
        imageData[i].m_Height = m_Description.width;
        imageData[i].m_Width = m_Description.width;
        imageData[i].m_LocationAddress = m_Description.locationAddress[i];

        imageData[i].m_BitsPerPixel = static_cast<u32>(bytesizePerPixel * 8);

        if (m_Description.isDynamicAllocation)
        {
            imageData[i].m_ImageDataTableCount = byteCount;
            imageData[i].toImageDataTable.set_ptr(reinterpret_cast<void*>(bufferMemory + byteCount * i));
        }
        else
        {
            imageData[i].m_ImageDataTableCount = 0;
            imageData[i].toImageDataTable.set_ptr(NULL);
        }

        //--------------------------------
        // Member initialization for ResImageTextureData
        textureData->toImagesTables[i].set_ptr(&imageData[i]);
    }

    return ResCubeTexture(textureData);
}

//----------------------------------------
ResShadowTexture
ResShadowTexture::DynamicBuilder::Create(
    os::IAllocator* allocator
)
{
    NW_NULL_ASSERT(allocator);

    // Check for power of 2
    NW_ASSERT( (m_Description.height & (m_Description.height - 1)) == 0 );
    NW_ASSERT( (m_Description.width & (m_Description.width - 1)) == 0 );

    // Texture size check
    const s32 TEXTURE_SIZE_MAX = 1024;
    const s32 TEXTURE_SIZE_MIN = 8;
    NW_MINMAX_ASSERT( m_Description.height, TEXTURE_SIZE_MIN, TEXTURE_SIZE_MAX );
    NW_MINMAX_ASSERT( m_Description.width, TEXTURE_SIZE_MIN, TEXTURE_SIZE_MAX );

    // Memory allocation for ResShadowTextureData
    void* textureDataMemory = allocator->Alloc(sizeof(ResShadowTextureData));
    NW_NULL_ASSERT(textureDataMemory);
    ResShadowTextureData* textureData = new(textureDataMemory) ResShadowTextureData();

    //--------------------------------
    // Member initialization for ResSceneObjectData
    textureData->typeInfo = ResShadowTexture::TYPE_INFO;
    textureData->m_Header.revision = ResShadowTexture::BINARY_REVISION;
    textureData->m_Header.signature = ResShadowTexture::SIGNATURE;
    textureData->toName.set_ptr(NULL);
    textureData->m_UserDataDicCount = 0;
    textureData->toUserDataDic.set_ptr(NULL);

    //--------------------------------
    // Member initialization for ResPixelBasedTextureData
    textureData->m_Height = m_Description.height;
    textureData->m_Width = m_Description.width;
    textureData->m_Format = GL_SHADOW_DMP;
    textureData->m_FormatType = GL_UNSIGNED_INT;
    textureData->m_MipmapSize = 0;
    textureData->m_TextureObject = 0;
    textureData->m_LocationFlag = m_Description.locationFlag;
    textureData->m_FormatHW = ResPixelBasedTexture::FORMAT_HW_RGBA8;

    // Memory allocation for ResPixelBasedImageData
    void* imageDataMemory = allocator->Alloc(sizeof(ResPixelBasedImageData));
    NW_NULL_ASSERT(imageDataMemory);
    ResPixelBasedImageData* imageData = new(imageDataMemory) ResPixelBasedImageData();

    imageData->m_DynamicAllocator = allocator;
    imageData->m_Height = m_Description.height;
    imageData->m_Width = m_Description.width;
    imageData->m_LocationAddress = m_Description.locationAddress;

    const f32 BYTESIZE_4 = 4.0f;
    f32 bytesizePerPixel = BYTESIZE_4;

    imageData->m_BitsPerPixel = static_cast<u32>(bytesizePerPixel * 8);

    u32 byteCount = 0;
    s32 h = m_Description.height;
    s32 w = m_Description.width;

    // Calculate the memory size to allocate for textures.
    // When using mipmaps, calculate the memory allocation amount corresponding to the number of mipmaps.
    byteCount += static_cast<u32>(bytesizePerPixel * h * w);

    // Memory allocation for pixel data
    // When isDynamicAllocation is specified, perform memory allocation
    if (m_Description.isDynamicAllocation)
    {
        const u8 ALIGNMENT = 128;
        NW_ASSERT(byteCount);
        void* bufferMemory = allocator->Alloc(byteCount, ALIGNMENT);
        NW_NULL_ASSERT(bufferMemory);
        u8* buffer = static_cast<u8*>(bufferMemory);
        if (m_Description.executingMemoryFill)
        {
            std::fill_n(NW_CHECKED_ARRAY_ITERATOR(buffer, byteCount), byteCount, 0);
        }

        imageData->m_ImageDataTableCount = byteCount;
        imageData->toImageDataTable.set_ptr(buffer);
    }
    else
    {
        imageData->m_ImageDataTableCount = 0;
        imageData->toImageDataTable.set_ptr(NULL);
    }

    //--------------------------------
    // Member initialization for ResImageTextureData
    textureData->toImage.set_ptr(imageData);

    //--------------------------------
    // Member initialization for ResShadowTextureData
    textureData->m_IsPerspectiveShadow = m_Description.isPerspectiveShadow;
    textureData->m_ShadowZBias = m_Description.shadowZBias;

    return ResShadowTexture(textureData);
}

//-----------------------------------------
void
ResImageTexture::DynamicDestroy()
{
    os::IAllocator* allocator = this->GetImage().ref().m_DynamicAllocator;

    NW_NULL_ASSERT( allocator );

    if (this->IsValid())
    {
        ResPixelBasedImage resPixelBasedImage = this->GetImage();

        if (resPixelBasedImage.IsValid())
        {
            // Memory deallocation for pixel data
            if (resPixelBasedImage.ptr()->toImageDataTable.to_ptr() != NULL)
            {
                allocator->Free(resPixelBasedImage.ptr()->toImageDataTable.to_ptr());
            }

            // Memory deallocation for ResPixelBasedImageData
            allocator->Free(resPixelBasedImage.ptr());
        }

        // Memory deallocation for ResImageTextureData
        allocator->Free(this->ptr());
        this->invalidate();
    }
}

//-----------------------------------------
void
ResCubeTexture::DynamicDestroy()
{
    ResPixelBasedImage firstFace = this->GetImage(static_cast<CubeFace>(0));

    os::IAllocator* allocator = firstFace.ref().m_DynamicAllocator;

    NW_NULL_ASSERT( allocator );

    if (this->IsValid())
    {
        // Since memory was allocated for all 6 faces at once, when the address for the first face is deallocated, all 6 faces are deallocated.
        // 
        if (firstFace.IsValid())
        {
            // Memory deallocation for pixel data
            if (firstFace.ptr()->toImageDataTable.to_ptr() != NULL)
            {
                allocator->Free(firstFace.ptr()->toImageDataTable.to_ptr());
            }

            // Memory deallocation for ResPixelBasedImageData
            allocator->Free(firstFace.ptr());
        }

        // Memory deallocation for ResImageTextureData
        allocator->Free(this->ptr());
        this->invalidate();
    }
}

//-----------------------------------------
void
ResShadowTexture::DynamicDestroy()
{
    os::IAllocator* allocator = this->GetImage().ref().m_DynamicAllocator;

    NW_NULL_ASSERT( allocator );

    ResPixelBasedImage resPixelBasedImage = this->GetImage();

    if(resPixelBasedImage.IsValid())
    {
        // Memory deallocation for pixel data
        if (resPixelBasedImage.ptr()->toImageDataTable.to_ptr() != NULL)
        {
            allocator->Free(resPixelBasedImage.ptr()->toImageDataTable.to_ptr());
        }

        // Memory deallocation for ResPixelBasedImageData
        allocator->Free(resPixelBasedImage.ptr());
    }

    // Memory deallocation for ResImageTextureData
    allocator->Free(this->ptr());
    this->invalidate();
}

//---------------------------------------------------------------------------
bool
ResPixelBasedImage::CheckMemoryLocation() const
{
    if (! nw::os::IsDeviceMemory(this->GetImageData()))
    {
        return false;
    }

    u32 startAddress = reinterpret_cast<u32>( this->GetImageData() );
    u32 endAddress = reinterpret_cast<u32>( this->GetImageData() ) + static_cast<u32>(this->GetImageDataCount()) - 1;

    // 128-byte alignment check.
    if ((startAddress & 0x7F) != 0)
    {
        return false;
    }

    // We realized that there were no hardware issues with a 32-MB alignment check, so we removed it.

    return true;
}



} /* namespace res */
} /* namespace gfx */
} /* namespace nw */

