/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: extUserData
//
// Overview
//   This demo obtains extended user data.
//
//   Extended user data is set for text boxes placed in the layout.
//   Get this extended user data and output the content with the NN_LOG function.
//
//   The only action on the actual device's screen is a single text box being displayed.
//
//------------------------------------------------------------------


#include <nn/applet.h>
#include <nw/lyt.h>
#include <nw/demo.h>

namespace
{

typedef nw::ut::MoveArray<u8> File;

/*
  

  
  
 */
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  

  
  
 */
void
SetupCamera(nw::lyt::DrawInfo& drawInfo, const nw::lyt::Layout& layout)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44Ortho(
        &projMtx,
        layoutRect.bottom,  // Left
        layoutRect.top,     // Right
        -layoutRect.right,  // Bottom
        -layoutRect.left,   // Top
        znear,
        zfar);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
    nw::math::VEC3 pos(0, 0, 1);
    nw::math::VEC3 up(-1, 0, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

/*
  

  
 */
void
PrintExtUserData(nw::lyt::Pane* pPane)
{
    // Get a pointer to the start of the extended user data array
    const nw::lyt::ExtUserData *const extUserDataAry =  pPane->GetExtUserDataArray();

    NN_LOG("\n ext user data\n");

    // Output each element of the extended user data array
    for (u16 udIdx = 0; udIdx < pPane->GetExtUserDataNum(); ++udIdx)
    {
        // Display the extended user data name
        NN_LOG("    %d \"%s\" [", udIdx, extUserDataAry[udIdx].GetName());

        // Divide processing according to the type of extended user data.
        switch (extUserDataAry[udIdx].GetType())
        {
        case nw::lyt::EXTUSERDATATYPE_STRING:             // Character string data
            NN_LOG("%s", extUserDataAry[udIdx].GetString());
            break;

        case nw::lyt::EXTUSERDATATYPE_INT:                // Integer (s32) list
            {
                const s32 *const intList = extUserDataAry[udIdx].GetIntArray();
                for (int i = 0; i < extUserDataAry[udIdx].GetNum(); ++i)
                {
                    NN_LOG("%d ", intList[i]);
                }
            }
            break;

        case nw::lyt::EXTUSERDATATYPE_FLOAT:              // Floating point (f32) list
            {
                const f32 *const floatList = extUserDataAry[udIdx].GetFloatArray();
                for (int i = 0; i < extUserDataAry[udIdx].GetNum(); ++i)
                {
                    NN_LOG("%f ", floatList[i]);
                }
            }
            break;
        }

        NN_LOG("]\n");
    }

    NN_LOG("\n");

    // Search for extended user data by name
    const nw::lyt::ExtUserData *const pExtUserDataIntList = pPane->FindExtUserDataByName("int1");
    NW_NULL_ASSERT(pExtUserDataIntList);
    NN_LOG("find ext user data \"%s\"\n", pExtUserDataIntList->GetName());
}

}   // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Loads layout binary resources (archives).
    File fileLayout = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout.arc"), 128);

    if (fileLayout.empty())
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.begin(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "extUserData.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::Pane *const pExtUserDataPane = pLayout->GetRootPane()->FindPaneByName("TextBox");
    NW_NULL_ASSERT(pExtUserDataPane);

    // Display extended user data
    PrintExtUserData(pExtUserDataPane);

    f32 animFrame = 0;

    nw::lyt::GraphicsResource graphicsResource;
    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file = nw::demo::Utility::LoadFile(&demoApp.GetAllocator(), resourcePath);

            if (file.empty())
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }

            graphicsResource.SetResource(i, file.begin(), file.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    bool loop = true;
    while (loop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            loop = false;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            SetupCamera(drawInfo, *pLayout);

            pLayout->CalculateMtx(drawInfo);

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;

    delete pResAccessor;
}

