/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: orthoStereo
//
// Overview
//   This sample an orthogonal projection of the layout in stereoscopic view.
//
//------------------------------------------------------------------

#include <nn/applet.h>
#include <nn/os.h>
#include <nn/fs.h>

#include <nw/types.h>
#include <nw/sdk.h>
#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/lyt.h>
#include <nw/ut.h>

namespace
{

nw::demo::RenderSystem* s_pRenderSystem = NULL;
nw::gfx::IRenderTarget* s_pRenderTarget = NULL;
nw::gfx::IRenderTarget* s_pRenderTargetLower = NULL;

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
typedef nw::ut::MoveArray<u8> File;

nn::ulcd::CTR::StereoCamera s_StereoCamera;

f32     s_DepthLevel    = 10.0f;    // The distance from the camera to the location laid out on the LCD (base plane).
f32     s_StereoFactor  =  1.0f;    // Stereoscopic adjustment factor (0 = No parallax. 1 = Standard.）

/*
  
 */
void
InitializeGraphics()
{
    nw::demo::RenderSystem::Description renderDesc;

    nw::demo::DisplayBufferSwapper::Description& upperScreenDesc =
        renderDesc.upperScreenDescription;

    nw::demo::DisplayBufferSwapper::Description& lowerScreenDesc =
        renderDesc.lowerScreenDescription;

    renderDesc.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_pRenderSystem = nw::demo::RenderSystem::Create(
        &s_DeviceAllocator,
        renderDesc);

    NW_ASSERT(s_pRenderSystem != NULL);

    s_pRenderTarget = nw::gfx::IRenderTarget::Builder()
        .BufferSize(upperScreenDesc.height, upperScreenDesc.width)
        .ColorFormat(renderDesc.renderColorFormat)
        .Create(&s_DeviceAllocator);

    NW_ASSERT(s_pRenderTarget != NULL);

    s_pRenderTargetLower = nw::gfx::IRenderTarget::Builder()
        .BufferSize(lowerScreenDesc.height, lowerScreenDesc.width)
        .ColorFormat(renderDesc.renderColorFormat)
        .Create(&s_DeviceAllocator);

    NW_ASSERT(s_pRenderTargetLower != NULL);

    NW_GL_ASSERT();
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_pRenderTarget);
    nw::gfx::SafeDestroy(s_pRenderTargetLower);
    nw::gfx::SafeDestroy(s_pRenderSystem);
}

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
void
SetupCamera(
    nw::lyt::DrawInfo&      drawInfo,
    const nw::lyt::Layout&  layout
)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    f32 znear = 0.f;
    f32 zfar = 500.f;

    // Set the projection matrix for an orthogonal projection
    // (Since the layout data is horizontal, change its direction.)
    nw::math::MTX44 projMtx;
    nw::math::MTX44OrthoPivot(
        &projMtx,
        layoutRect.left,   // Left
        layoutRect.right,  // Right
        layoutRect.bottom, // Bottom
        layoutRect.top,    // Top
        znear,
        zfar,
        nw::math::PIVOT_UPSIDE_TO_TOP);
    drawInfo.SetProjectionMtx(projMtx);

    // Set the modelview matrix
    nw::math::VEC3 pos(0, 0, s_DepthLevel);
    nw::math::VEC3 up(0, 1, 0);
    nw::math::VEC3 target(0, 0, 0);

    nw::math::MTX34 viewMtx;
    nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
    drawInfo.SetViewMtx(viewMtx);
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
f32
CalcParallaxDot(nw::lyt::DrawInfo& drawInfo)
{
    // Calculates the view matrix and projection matrix for rendering the image that corresponds to the left and right parallax
    nn::math::MTX34 view = drawInfo.GetViewMtx();
    s_StereoCamera.SetBaseCamera(&view);

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    s_StereoCamera.CalculateMatrices(
        &projL,
        &viewL,
        &projR,
        &viewR,
        s_DepthLevel,
        s_StereoFactor,
        nn::math::PIVOT_UPSIDE_TO_TOP);

    // Determine half the parallax value, in units of LCD pixels.
    f32 wkParallax = s_StereoCamera.GetCoefficientForParallax();
    return wkParallax * nw::demo::SimpleApp::DISPLAY0_WIDTH;
}

//---------------------------------------------------------------------------
//
//
//
//
//
//
//---------------------------------------------------------------------------
GLuint
CreateCmdList(
    GLsizei     bufSize,
    GLsizei     requestCount
)
{
    // Saves the current.
    GLuint crntCmdListId;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_BINDING, reinterpret_cast<GLint*>(&crntCmdListId));

    // Generate
    GLuint lytCmdListId;
    nngxGenCmdlists(1, &lytCmdListId);
    nngxBindCmdlist(lytCmdListId);
    nngxCmdlistStorage(bufSize, requestCount);

    // Restores the current.
    nngxBindCmdlist(crntCmdListId);

    return lytCmdListId;
}

//---------------------------------------------------------------------------
//
//
//
//
//
//---------------------------------------------------------------------------
GLuint
StartCmdListSave(GLuint cmdListId)
{
    // Saves the current.
    GLuint crntCmdListId;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_BINDING, reinterpret_cast<GLint*>(&crntCmdListId));

    nngxBindCmdlist(cmdListId);     // Bind
    nngxClearCmdlist();             // Clear
    nngxStartCmdlistSave();         // Starts saving.

    return crntCmdListId;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
struct CmdListInfo
{
    GLuint  bufferOffset;
    GLsizei bufferSize;
    GLuint  requestId;
    GLsizei requestSize;
};

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void
StopCmdListSave(
    CmdListInfo*    pCmdListInfo,
    GLuint          cmdListId
)
{
    nngxStopCmdlistSave(
        &pCmdListInfo->bufferOffset,
        &pCmdListInfo->bufferSize,
        &pCmdListInfo->requestId,
        &pCmdListInfo->requestSize);
    NW_GL_ASSERT();

    // Restores the current.
    nngxBindCmdlist(cmdListId);
}

//---------------------------------------------------------------------------
//
//
//
//
//---------------------------------------------------------------------------
void
UseSavedCmdlist(
    GLuint              cmdListId,
    const CmdListInfo&  cmdListInfo
)
{
    nngxUseSavedCmdlist(
        cmdListId,
        cmdListInfo.bufferOffset,
        cmdListInfo.bufferSize,
        cmdListInfo.requestId,
        cmdListInfo.requestSize,
        0,
        GL_TRUE);
}

}   // namespace {anonymous}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    InitializeGraphics();

    // Initializes the layout library.
    // The demos allocate all memory from device memory, so the same allocator gets passed.
    // 
    // To handle this and make a distinction with heap memory, be sure to specify device memory for the second argument.
    nw::lyt::Initialize(&s_DeviceAllocator, &s_DeviceAllocator);

    // Loads layout binary resources (archives).
    File resLayout = nw::demo::Utility::LoadFile(
        &s_DeviceAllocator, NW_DEMO_FILE_PATH(L"layout.arc"), 128);

    if (resLayout.empty())
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(resLayout.begin(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "orthoStereo.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::GraphicsResource graphicsResource;

    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File resource = nw::demo::Utility::LoadFile(&s_DeviceAllocator, resourcePath);

            if (resource.empty())
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }

            graphicsResource.SetResource(i, resource.begin(), resource.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    // Initializes the camera for stereoscopic view.
    s_StereoCamera.Initialize();
    s_StereoCamera.SetBaseFrustum(
        -0.04f,
         0.04f,
        -0.04f * nw::demo::SimpleApp::DISPLAY0_HEIGHT / nw::demo::SimpleApp::DISPLAY0_WIDTH,
         0.04f * nw::demo::SimpleApp::DISPLAY0_HEIGHT / nw::demo::SimpleApp::DISPLAY0_WIDTH,
         0.2f,
         30.f);

    // Generates command list for layout rendering.
    const GLuint lytCmdListId = CreateCmdList(2048, 1);

    bool loop = true;
    while (loop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            loop = false;
        }

        SetupCamera(drawInfo, *pLayout);
        NW_GL_ASSERT();

        pLayout->Animate();
        pLayout->CalculateMtx(drawInfo);
        NW_GL_ASSERT();

        // Generates layout rendering commands.
        CmdListInfo lytCmdListInfo;
        const GLuint crntCmdListId = StartCmdListSave(lytCmdListId);
        drawer.DrawBegin(drawInfo, nw::lyt::Drawer::DONT_USE_SETUP_COMMAND);
        drawer.Draw(pLayout, drawInfo);
        drawer.DrawEnd(drawInfo, nw::lyt::Drawer::DONT_USE_SETUP_COMMAND);
        StopCmdListSave(&lytCmdListInfo, crntCmdListId);

        s_pRenderSystem->SetRenderTarget(s_pRenderTarget);
        NW_GL_ASSERT();

        // Finds half of the parallax value in liquid crystal dot units.
        const f32 halfParallaxDot = CalcParallaxDot(drawInfo);

        // Upper screen (for left eye)
        {
            s_pRenderSystem->ClearBuffer(
                nw::ut::FloatColor(0.2f, 0.2f, 0.2f, 0.0f));

            InitDraw(
                nw::demo::SimpleApp::DISPLAY0_WIDTH,
                nw::demo::SimpleApp::DISPLAY0_HEIGHT);

            // Uses Drawer to render the layout.
            drawer.SetParallax(-halfParallaxDot, s_StereoCamera.GetDistanceToLevel());
            drawer.DrawBegin(drawInfo);
            UseSavedCmdlist(lytCmdListId, lytCmdListInfo);
            drawer.DrawEnd(drawInfo);

            s_pRenderSystem->TransferBuffer(nw::demo::UPPER_SCREEN);
        }

        // Upper screen (for right eye)
        {
            s_pRenderSystem->ClearBuffer(
                nw::ut::FloatColor(0.2f, 0.2f, 0.2f, 0.0f));

            InitDraw(
                nw::demo::SimpleApp::DISPLAY0_WIDTH,
                nw::demo::SimpleApp::DISPLAY0_HEIGHT);

            // Uses Drawer to render the layout.
            drawer.SetParallax( halfParallaxDot, s_StereoCamera.GetDistanceToLevel());
            drawer.DrawBegin(drawInfo);
            UseSavedCmdlist(lytCmdListId, lytCmdListInfo);
            drawer.DrawEnd(drawInfo);

            s_pRenderSystem->TransferBuffer(nw::demo::EXTENSION_SCREEN);
        }

        s_pRenderSystem->SetRenderTarget(s_pRenderTargetLower);

        // Lower screen
        {
            s_pRenderSystem->ClearBuffer(
                nw::ut::FloatColor(0.1f, 0.1f, 0.1f, 0.0f));

            s_pRenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);
        }

        s_pRenderSystem->PresentBuffer(nw::demo::BOTH_SCREENS);
        NW_GL_ASSERT();
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;
    delete pResAccessor;

    TerminateGraphics();

    nw::demo::PadFactory::Finalize();
}

