/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//------------------------------------------------------------------
// Demo: perspective
//
// Overview
//   This is a demo for layouts displayed with a three-dimensional appearance.
//
// Controls
//   Use the A Button to switch the projection matrix between orthogonal and perspective projections.
//   With a perspective projection, the display appears three-dimensional.
//   With an orthogonal projection, the display is ordinary and two-dimensional.
//
//------------------------------------------------------------------

#include <nn/applet.h>
#include <nw/lyt.h>
#include <nw/demo.h>

namespace
{

typedef nw::ut::MoveArray<u8> File;

/*
  
  
  
 
*/
void
InitDraw(int width, int height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo =
    {
        height, width, PICA_DATA_DEPTH24_STENCIL8_EXT
    };

    const u32 commands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT(0, 0, colBufInfo.width, colBufInfo.height),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR(colBufInfo),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(commands, sizeof(commands), true);

    static const u32 constCommands[] =
    {
        // Clear the cache so frame-buffer access control can be changed
        // 0x111
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1),
        // 0x110
        1, PICA_CMD_HEADER_SINGLE(PICA_REG_COLOR_DEPTH_BUFFER_CLEAR0),

        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE(NW_FONT_CMD_CULL_FACE_DISABLE),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(constCommands, sizeof(constCommands), true);
}

/*
  
  
 
*/
void
EnableZCompare(bool bEnable)
{
    if (bEnable)
    {
        // Perform z-comparisons.
        glEnable(GL_DEPTH_TEST);
    }
    else
    {
        // Do not perform z-comparisons. Objects lower in the layout hierarchy are always displayed in front of objects higher in the hierarchy.
        glDisable(GL_DEPTH_TEST);
    }
}

/*
  
  
  
  
                       
                       
 
*/
void
SetupCamera(
    nw::lyt::DrawInfo& drawInfo,
    const nw::lyt::Layout& layout,
    bool bPerspective
)
{
    nw::ut::Rect layoutRect = layout.GetLayoutRect();

    if (bPerspective)
    {
        const f32 fovy = 45.f;
        const f32 znear = 0.05f;
        const f32 zfar = 10000.f;
        const f32 width = layoutRect.top - layoutRect.bottom;
        const f32 height = layoutRect.right - layoutRect.left;

        // Set the projection matrix to be a perspective projection.
        nw::math::MTX44 projMtx;
        nw::math::MTX44Perspective(
            &projMtx,
            fovy,
            width / height,
            znear,
            zfar);
        drawInfo.SetProjectionMtx(projMtx);

        //  View matrix
        //      Set the camera position in front, such that when the viewpoint is (0, 0, 0), a plane with a z-position of 0 will have the same appearance as it does for an orthogonal projection.
        // 
        const f32 dist = height * 0.5f / nw::math::TanDeg(fovy * 0.5f);

        nw::math::VEC3 pos(0.f, 0.f, dist);
        nw::math::VEC3 up(-1, 0, 0);
        nw::math::VEC3 target(0.f, 0.f, 0.f);

        nw::math::MTX34 viewMtx;  // View matrix
        nw::math::MTX34LookAt(
            &viewMtx,
            &pos,                 // Camera position
            &up,                  // Camera up direction
            &target);             // Viewpoint
        drawInfo.SetViewMtx(viewMtx);
    }
    else
    {
        const f32 znear   =  -1000.f;
        const f32 zfar    =   1000.f;

        // Set the projection matrix for an orthogonal projection
        // (Since the layout data is horizontal, change its direction.)
        nw::math::MTX44 projMtx;
        nw::math::MTX44Ortho(
            &projMtx,
            layoutRect.bottom,  // Left
            layoutRect.top,     // Right
            -layoutRect.right,  // Bottom
            -layoutRect.left,   // Top
            znear,
            zfar);
        drawInfo.SetProjectionMtx(projMtx);

        // Set the modelview matrix
        // (Since the layout data is horizontal, the upper part of the screen is the -X direction of the layout.)
        nw::math::VEC3 pos(0, 0, 1);
        nw::math::VEC3 up(-1, 0, 0);
        nw::math::VEC3 target(0, 0, 0);

        nw::math::MTX34 viewMtx;
        nw::math::MTX34LookAt(&viewMtx, &pos, &up, &target);
        drawInfo.SetViewMtx(viewMtx);
    }
}

}   // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::SimpleApp& demoApp = nw::demo::SimpleApp::GetInstance();
    demoApp.Initialize();

    nw::lyt::Initialize(&demoApp.GetAllocator(), &demoApp.GetDeviceAllocator());

    // Loads layout binary resources (archives).
    File fileLayout = nw::demo::Utility::LoadFile(
        &demoApp.GetDeviceAllocator(), NW_DEMO_FILE_PATH(L"layout.arc"), 128);

    if (fileLayout.empty())
    {
        NW_FATAL_ERROR("can not open layout archive.\n");
    }

    // Specifies the binary resource root directory and generates the resource assessor.
    nw::lyt::ArcResourceAccessor* pResAccessor = new nw::lyt::ArcResourceAccessor;
    if (!pResAccessor->Attach(fileLayout.begin(), "."))
    {
        NW_FATAL_ERROR("can not attach layout archive.\n");
    }

    nw::lyt::Layout* pLayout = new nw::lyt::Layout();

    // Load the layout resource
    {
        const void* lytRes = pResAccessor->GetResource(0, "perspective.bclyt");
        NW_NULL_ASSERT(lytRes);
        pLayout->Build(lytRes, pResAccessor);
    }

    nw::lyt::AnimTransform* pAnimTrans = 0;
    // Loads the animation resource.
    {
        const void* lpaRes = pResAccessor->GetResource(0, "perspective.bclan");
        NW_NULL_ASSERT(lpaRes);
        pAnimTrans = pLayout->CreateAnimTransform(lpaRes, pResAccessor);
        pLayout->BindAnimation(pAnimTrans);
    }

    f32 animFrame = 0;
    bool bPerspective = true;

    nw::lyt::GraphicsResource graphicsResource;
    // Loads the global resource file.
    {
        graphicsResource.StartSetup();
        const wchar_t* resourcePath = 0;
        for (int i = 0;
             (resourcePath = graphicsResource.GetResourcePath(i)) != NULL;
             ++i)
        {
            File file = nw::demo::Utility::LoadFile(&demoApp.GetAllocator(), resourcePath);

            if (file.empty())
            {
                NW_FATAL_ERROR("can not read lyt resource file.");
            }

            graphicsResource.SetResource(i, file.begin(), file.size(), false);
        }

        graphicsResource.FinishSetup();
    }

    nw::lyt::DrawInfo drawInfo;
    drawInfo.SetGraphicsResource(&graphicsResource);

    nw::lyt::Drawer drawer;
    drawer.Initialize(graphicsResource);

    nw::demo::Pad& pad = *nw::demo::PadFactory::GetPad();

    const nw::ut::FloatColor clearColor(0.3f, 0.3f, 0.3f, 1.0f);
    const f32 clearDepth = 0.0f;

    bool loop = true;
    while (loop)
    {
        pad.Update();

        if (pad.IsButtonDown(pad.BUTTON_START))
        {
            loop = false;
        }

        if (pad.IsButtonDown(pad.BUTTON_A))
        {
            // Switch between perspective and orthogonal projections.
            bPerspective = ! bPerspective;
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY0);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);

            InitDraw(demoApp.DISPLAY0_WIDTH, demoApp.DISPLAY0_HEIGHT);

            EnableZCompare(bPerspective);
            SetupCamera(drawInfo, *pLayout, bPerspective);

            pLayout->Animate();
            pLayout->CalculateMtx(drawInfo);

            drawer.DrawBegin(drawInfo);
            drawer.Draw(pLayout, drawInfo);
            drawer.DrawEnd(drawInfo);

            // Frame update.
            animFrame += 1.0f;
            while (animFrame >= pAnimTrans->GetFrameSize())
            {
                animFrame -= pAnimTrans->GetFrameSize();
            }
            pAnimTrans->SetFrame(animFrame);
        }

        demoApp.SetRenderingTarget(demoApp.DISPLAY1);
        {
            demoApp.GetFrameBufferObject().ClearBuffer(clearColor, clearDepth);
        }

        demoApp.SwapBuffer(demoApp.DISPLAY_BOTH);
    }

    drawer.Finalize();
    graphicsResource.Finalize();
    delete pLayout;
    delete pResAccessor;
}

