﻿/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"
#include <nw/lyt/lyt_Layout.h>
#include <nw/lyt/lyt_Pane.h>
#include <nw/lyt/lyt_Drawer.h>
#include <nw/lyt/lyt_DrawerCommand.h>
#include <nw/lyt/lyt_DrawInfo.h>
#include <nw/lyt/lyt_GraphicsResource.h>

#ifdef NW_LYT_DRAWER_ENABLE

namespace nw { namespace lyt {

//----------------------------------------------------------
// Drawer

void
Drawer::Draw( const Layout* pLayout, DrawInfo& drawInfo )
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->Draw(pCmdStream, pLayout, drawInfo);
    SetCurrentCommandBuffer(pCmdStream);
}

u32*
Drawer::Draw( u32* pCmdStream, const Layout* pLayout, DrawInfo& drawInfo )
{
    if ( pLayout == NULL )
    {
        return pCmdStream;
    }

    if ( pLayout->GetRootPane() == NULL ||
         pLayout->GetRootPane()->IsVisible() == false )
    {
        return pCmdStream;
    }

    pCmdStream = pLayout->GetRootPane()->MakeUniformData( pCmdStream, &drawInfo, this );

    return pCmdStream;
}

void
Drawer::Draw( const Pane* pPane, DrawInfo& drawInfo )
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->Draw(pCmdStream, pPane, drawInfo);
    SetCurrentCommandBuffer(pCmdStream);
}

u32*
Drawer::Draw( u32* pCmdStream, const Pane* pPane, DrawInfo& drawInfo )
{
    if ( pPane == NULL )
    {
        return pCmdStream;
    }

    if ( pPane->IsVisible() )
    {
        pCmdStream = pPane->MakeUniformData( pCmdStream, &drawInfo, this );
    }

    return pCmdStream;
}

void
Drawer::DrawBegin( const DrawInfo& drawInfo, u32 flag )
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->DrawBegin(pCmdStream, drawInfo, flag);
    SetCurrentCommandBuffer(pCmdStream);
}

u32*
Drawer::DrawBegin( u32* pCmdStream, const DrawInfo& drawInfo, u32 flag )
{
    pCmdStream = DrawBegin(pCmdStream, flag);

    if (!(flag & DONT_USE_SETUP_COMMAND))
    {
        pCmdStream = Base::SetProjectionMtx(pCmdStream, drawInfo.GetProjectionMtx());
    }

    return pCmdStream;
}

/* virtual */
u32*
Drawer::DrawBegin(u32* pCmdStream, u32 flag)
{
    m_ActiveTexureNum = 0;
    m_TexCoordNum = 0;
    m_PrevTexObj = 0;
    m_CurrentTexEnvType = TEX_ENV_TYPE_NUM;

    m_IsBlendDefault = true;

    pCmdStream = SetUpTextures( pCmdStream, NULL, false );

    pCmdStream = Base::DrawBegin(pCmdStream, flag);

    return pCmdStream;
}

/* virtual, deprecated */
void
Drawer::DrawBegin(u32 flag)
{
    Base::DrawBegin(flag);
}

void
Drawer::DrawEnd( const DrawInfo& drawInfo, u32 flag )
{
    u32* pCmdStream = GetCurrentCommandBuffer();
    pCmdStream = this->DrawEnd(pCmdStream, drawInfo, flag);
    SetCurrentCommandBuffer(pCmdStream);
}

u32*
Drawer::DrawEnd( u32* pCmdStream, const DrawInfo& drawInfo, u32 flag )
{
    NW_UNUSED_VARIABLE(drawInfo);
    pCmdStream = this->DrawEnd(pCmdStream, flag);
    return pCmdStream;
}

/* virtual */
u32*
Drawer::DrawEnd(u32* pCmdStream, u32 flag)
{
    pCmdStream = FlushBuffer(pCmdStream);
    pCmdStream = Base::DrawEnd(pCmdStream, flag);
    return pCmdStream;
}

/* virtual, deprecated */
void
Drawer::DrawEnd(u32 flag)
{
    Base::DrawEnd(flag);
}

Drawer::Drawer()
{
}

Drawer::Drawer( GraphicsResource& graphicsResource )
{
    this->Initialize( graphicsResource );
}

void
Drawer::Initialize( GraphicsResource& graphicsResource, void* vertexBuffer )
{
    const u32 cmdBufSize = GetCommandBufferSize(
        graphicsResource.GetRectShaderBinary(),
        graphicsResource.GetRectShaderBinarySize());

    const u32 allocSize = cmdBufSize + (vertexBuffer == NULL ? GetVertexBufferSize(): 0);

    void* commandBuffer = Layout::AllocDeviceMemory(allocSize, 4);
    NW_NULL_ASSERT(commandBuffer);

    if (vertexBuffer == NULL)
    {
        Base::Initialize(
            commandBuffer,
            graphicsResource.GetRectShaderBinary(),
            graphicsResource.GetRectShaderBinarySize());
    }
    else
    {
        Base::Initialize(
            vertexBuffer,
            commandBuffer,
            graphicsResource.GetRectShaderBinary(),
            graphicsResource.GetRectShaderBinarySize());
    }

    m_PrevTexObj = 0;
    m_ActiveTexureNum = 0;
    m_TexCoordNum = 0;
    m_IsBlendDefault = true;
    m_IsAlphaTestDefault = true;
    m_CurrentTexEnvType = TEX_ENV_TYPE_NUM;
#if ! defined(NW_RELEASE)
    m_PreviousTexEnvType = TEX_ENV_TYPE_NUM;
#endif
}

void
Drawer::Finalize()
{
    void *const comamndBuffer = m_CommandBuffer;

    Base::Finalize();
    Layout::FreeDeviceMemory(comamndBuffer);
}

}   // namespace lyt
}   // namespace nw
#endif
