/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"
#include <nw/lyt/lyt_Drawer.h>
#include <nw/lyt/lyt_Layout.h>
#include <nw/lyt/lyt_TextBox.h>
#include <nw/lyt/lyt_Material.h>
#include <nw/lyt/lyt_DrawerCommand.h>
#include <nw/lyt/lyt_DrawInfo.h>
#include <nw/lyt/lyt_GraphicsResource.h>
#include <nw/font/font_DispStringBuffer.h>

#ifdef NW_LYT_DRAWER_ENABLE

#define TEX_ENV_COLOR_OFFS            3

// Clears the color depth buffer  (required once for the 0x100 ~ 0x130 settings)
// 0x111
#define ADD_COMMAND_COLOR_DEPTH_BUFFER_CLEAR1(ptr, clear) \
    NW_FONT_ADD_SINGLE_COMMAND_PTR( \
        (ptr), \
        PICA_CMD_HEADER_SINGLE( PICA_REG_COLOR_DEPTH_BUFFER_CLEAR1 ), \
        PICA_CMD_DATA_COLOR_DEPTH_BUFFER_CLEAR( clear ) ) \

namespace nw { namespace lyt {

u32*
Drawer::SetUpTexEnv( u32* pCmdStream, const Material* __restrict pMaterial )
{
    NW_PROFILE("nw::lyt::Drawer::SetUpTexEnv");
#if ! defined(NW_RELEASE)
    m_PreviousTexEnvType = m_CurrentTexEnvType;
#endif

    if ( pMaterial->GetTextureOnly() )
    {
        switch ( m_CurrentTexEnvType )
        {
        case TEX_ENV_TYPE_0_TEX:
        case TEX_ENV_TYPE_1_TEX:
            break;

        default:
            pCmdStream = FlushBuffer(pCmdStream);
            break;
        }

        return pCmdStream;
    }

    bool bBufferClear = false;

#ifdef NW_LYT_DRAWER_ENABLE_ALPHA_TEST
    // Alpha test settings
    if ( pMaterial->IsAlphaCompareCap() )
    {
        pCmdStream = FlushBuffer(pCmdStream);
        m_IsAlphaTestDefault = false;

        // 0x111
        ADD_COMMAND_COLOR_DEPTH_BUFFER_CLEAR1( pCmdStream, true );
        bBufferClear = true;

        pCmdStream = SetUpAlphaTest( pCmdStream, pMaterial );
    }
    else if ( !m_IsAlphaTestDefault )
    {
        // Setting that returns the alpha test to the default (GREATER 0).
        pCmdStream = FlushBuffer(pCmdStream);
        m_IsAlphaTestDefault = true;

        // 0x111
        ADD_COMMAND_COLOR_DEPTH_BUFFER_CLEAR1( pCmdStream, true );
        bBufferClear = true;

        // 0x104
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_FRAGOP_ALPHA_TEST ),
            PICA_CMD_DATA_FRAGOP_ALPHA_TEST( true, PICA_DATA_ALPHA_TEST_GREATER, 0 ) );
    }
#endif

#ifdef NW_LYT_DRAWER_ENABLE_BLEND_FUNC
    // Blend mode settings
    if ( pMaterial->IsBlendModeCap() )
    {
        pCmdStream = FlushBuffer(pCmdStream);
        m_IsBlendDefault = false;

        if (!bBufferClear)
        {
            // 0x111
            ADD_COMMAND_COLOR_DEPTH_BUFFER_CLEAR1( pCmdStream, true );
            bBufferClear = true;
        }

        // Blend settings
        pCmdStream = SetUpBlendMode( pCmdStream, pMaterial );
    }
    else if ( m_IsBlendDefault == false )
    {
        // Setting that returns to the default (alpha blend).
        pCmdStream = FlushBuffer(pCmdStream);
        m_IsBlendDefault = true;

        if (!bBufferClear)
        {
            // 0x111
            ADD_COMMAND_COLOR_DEPTH_BUFFER_CLEAR1( pCmdStream, true );
            bBufferClear = true;
        }

        const u32 command[] =
        {
            NW_FONT_COMMAND_SET_BLEND_DEFAULT
        };

        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }
#endif

#ifdef NW_LYT_DRAWER_ENABLE_USER_TEX_ENV
    //----------------------------------------------------------
    // User defined texture combiner setting
    if ( pMaterial->GetTevStageNum() > 0 )
    {
        pCmdStream = FlushBuffer(pCmdStream);

        pCmdStream = SetUpGLTexEnvUser( pCmdStream, pMaterial );

        m_CurrentTexEnvType = TEX_ENV_TYPE_USER;

        return pCmdStream;
    }
#endif

    //----------------------------------------------------------
    // Preset texture combiner setting
    switch ( pMaterial->GetTexMapNum() )
    {
      case 0:
        {
            nw::ut::Color8 white = pMaterial->GetColor( INTERPOLATECOLOR_WHITE );

            static u32 command[] NW_LYT_DRAWER_ALIGN_32 =
            {
                PICA_CMD_DATA_TEX_ENV_SRC(
                    PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT, PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT ),
                PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV5, 5 ),
                PICA_CMD_DATA_TEX_ENV_OPERAND(
                    PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
                    PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
                PICA_CMD_DATA_TEX_ENV_COMBINE(
                    PICA_DATA_TEX_ENV_COMBINE_MODULATE,
                    PICA_DATA_TEX_ENV_COMBINE_MODULATE ),
                PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_WHITE,
                PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),
            };

            if ( m_CurrentTexEnvType != TEX_ENV_TYPE_0_TEX )
            {
                pCmdStream = FlushBuffer(pCmdStream);

                *reinterpret_cast< nw::ut::Color8* >( &command[ 4 ] ) = white;

                NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );

                m_CurrentTexEnvType = TEX_ENV_TYPE_0_TEX;
            }
            else if ( *reinterpret_cast< nw::ut::Color8* >( &command[ 4 ] ) != white )
            {
                pCmdStream = FlushBuffer(pCmdStream);

                *reinterpret_cast< nw::ut::Color8* >( &command[ 4 ] ) = white;

                NW_FONT_ADD_SINGLE_COMMAND_PTR(
                    pCmdStream,
                    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV5 + TEX_ENV_COLOR_OFFS),
                    command[ 4 ] );
            }

            break;
        }
      case 1 :
        {

            nw::ut::Color8 white = pMaterial->GetColor( INTERPOLATECOLOR_WHITE );
            nw::ut::Color8 black = pMaterial->GetColor( INTERPOLATECOLOR_BLACK );

            static u32 command[] NW_LYT_DRAWER_ALIGN_32 =
            {
                PICA_CMD_DATA_TEX_ENV_SRC(
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT ),
                PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, 5 ),
                PICA_CMD_DATA_TEX_ENV_OPERAND(
                    PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_ONE_MINUS_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
                    PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
                PICA_CMD_DATA_TEX_ENV_COMBINE(
                    PICA_DATA_TEX_ENV_COMBINE_MODULATE,
                    PICA_DATA_TEX_ENV_COMBINE_MODULATE ),
                PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
                PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

                PICA_CMD_DATA_TEX_ENV_SRC(
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
                    PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                    PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
                PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, 5 ),
                PICA_CMD_DATA_TEX_ENV_OPERAND(
                    PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
                    PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
                PICA_CMD_DATA_TEX_ENV_COMBINE(
                    PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP,
                    PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP ),
                PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_WHITE,
                PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

                PICA_CMD_DATA_TEX_ENV_SRC(
                    PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
                    PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
                PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV5, 5 ),
                PICA_CMD_DATA_TEX_ENV_OPERAND(
                    PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
                    PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
                PICA_CMD_DATA_TEX_ENV_COMBINE(
                    PICA_DATA_TEX_ENV_COMBINE_MODULATE,
                    PICA_DATA_TEX_ENV_COMBINE_MODULATE ),
                PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
                PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),
            };

            if ( m_CurrentTexEnvType != TEX_ENV_TYPE_1_TEX )
            {
                pCmdStream = FlushBuffer(pCmdStream);

                *reinterpret_cast< nw::ut::Color8* >( &command[ 4 ] ) = black;
                *reinterpret_cast< nw::ut::Color8* >( &command[ 10 ] ) = white;

                NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );

                m_CurrentTexEnvType = TEX_ENV_TYPE_1_TEX;
            }
            else if ( *reinterpret_cast< nw::ut::Color8* >( &command[ 4 ] ) != black ||
                      *reinterpret_cast< nw::ut::Color8* >( &command[ 10 ] ) != white )
            {
                pCmdStream = FlushBuffer(pCmdStream);

                *reinterpret_cast< nw::ut::Color8* >( &command[ 4 ] ) = black;
                *reinterpret_cast< nw::ut::Color8* >( &command[ 10 ] ) = white;

                NW_FONT_ADD_SINGLE_COMMAND_PTR(
                    pCmdStream,
                    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3 + TEX_ENV_COLOR_OFFS ),
                    command[ 4 ] );

                NW_FONT_ADD_SINGLE_COMMAND_PTR(
                    pCmdStream,
                    PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4 + TEX_ENV_COLOR_OFFS ),
                    command[ 10 ] );
            }

            break;
        }
      case 2 :
        {
            pCmdStream = FlushBuffer(pCmdStream);

            pCmdStream = SetUpTexEnvType2( pCmdStream, pMaterial );

            m_CurrentTexEnvType = TEX_ENV_TYPE_2_TEX;

            break;
        }
      case 3 :
        {
            pCmdStream = FlushBuffer(pCmdStream);

            pCmdStream = SetUpTexEnvType3( pCmdStream, pMaterial );

            m_CurrentTexEnvType = TEX_ENV_TYPE_3_TEX;

            break;
        }
    }

    return pCmdStream;
}

u32* Drawer::SetUpTexEnvType2( u32* pCmdStream, const Material* __restrict pMaterial )
{
    static u32 command[] NW_LYT_DRAWER_ALIGN_32 =
    {
        // Blends Texture0 and Texture1.
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE1,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE1,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_ALPHA,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE,
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE ),
        // cmdPosBlend
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        // Black and white color interpolation using texture.
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE,
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE ),
        // cmdPosWhite
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_WHITE,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        // cmdPosBlack
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV_BUFFER_COLOR ), // 0x0fd

        // Vertex color and multiplication
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV5, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_MODULATE,
            PICA_DATA_TEX_ENV_COMBINE_MODULATE ),
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        NW_LYT_CMD_DATA_TEX_ENV_BUF_INPUT_THROUGH,
        PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEX_ENV_BUFFER_INPUT, 0x2 ) // 0x0e0
    };

    nw::ut::Color8 blend = pMaterial->GetColor( TEVKONSTSEL_K5 );
    nw::ut::Color8 white = pMaterial->GetColor( INTERPOLATECOLOR_WHITE );
    nw::ut::Color8 black = pMaterial->GetColor( INTERPOLATECOLOR_BLACK );

    const int cmdPosBlend = 4;
    const int cmdPosWhite = 10;
    const int cmdPosBlack = 12;

    if ( m_CurrentTexEnvType != TEX_ENV_TYPE_2_TEX )
    {
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosWhite ] ) = white;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlack ] ) = black;

        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }
    else if ( *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend ] ) != blend ||
              *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosWhite ] ) != white ||
              *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlack ] ) != black )
    {
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosWhite ] ) = white;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlack ] ) = black;

        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3 + TEX_ENV_COLOR_OFFS ),
            command[ cmdPosBlend ] );

        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4 + TEX_ENV_COLOR_OFFS ),
            command[ cmdPosWhite ] );

        // 0x0fd
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV_BUFFER_COLOR ),
            command[ cmdPosBlack ] );
    }

    return pCmdStream;
}

u32* Drawer::SetUpTexEnvType3( u32* pCmdStream, const Material* __restrict pMaterial )
{
    static u32 command[] NW_LYT_DRAWER_ALIGN_32 =
    {
        // Multiplies the blend ratio for Texture0.
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV1, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_ALPHA, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_MODULATE,
            PICA_DATA_TEX_ENV_COMBINE_MODULATE ),
        // cmdPosBlend0
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        // Multiplies the blend ratio for Texture1.
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE1,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE1,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV2, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_B_DMP, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_B_DMP, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP,
            PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP ),
        // cmdPosBlend1
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        // Multiplies the blend ratio for Texture2.
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE2,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE2,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV3, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_G_DMP, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_G_DMP, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP,
            PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP ),
        // cmdPosBlend2
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        // Black and white color interpolation using texture.
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV4, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE,
            PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE ),
        // cmdPosWhite
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_WHITE,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        // cmdPosBlack
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV_BUFFER_COLOR ), // 0x0fd

        // Vertex color and multiplication
        PICA_CMD_DATA_TEX_ENV_SRC(
            PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS ),
        PICA_CMD_HEADER_BURSTSEQ( PICA_REG_TEX_ENV5, 5 ),
        PICA_CMD_DATA_TEX_ENV_OPERAND(
            PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR, PICA_DATA_OPE_RGB_SRC_COLOR,
            PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA, PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
        PICA_CMD_DATA_TEX_ENV_COMBINE(
            PICA_DATA_TEX_ENV_COMBINE_MODULATE,
            PICA_DATA_TEX_ENV_COMBINE_MODULATE ),
        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_BLACK,
        PICA_CMD_DATA_TEX_ENV_SCALE( PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),

        NW_LYT_CMD_DATA_TEX_ENV_BUF_INPUT_THROUGH,
        PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEX_ENV_BUFFER_INPUT, 0x2 ) // 0x0e0
    };

    nw::ut::Color8 blend = pMaterial->GetColor( TEVKONSTSEL_K5 );
    nw::ut::Color8 white = pMaterial->GetColor( INTERPOLATECOLOR_WHITE );
    nw::ut::Color8 black = pMaterial->GetColor( INTERPOLATECOLOR_BLACK );

    const int cmdPosBlend0 = 4;
    const int cmdPosBlend1 = 10;
    const int cmdPosBlend2 = 16;
    const int cmdPosWhite = 22;
    const int cmdPosBlack = 24;

    if ( m_CurrentTexEnvType != TEX_ENV_TYPE_3_TEX )
    {
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend0 ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend1 ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend2 ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosWhite ] ) = white;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlack ] ) = black;

        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }
    else if ( *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend0 ] ) != blend ||
              *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosWhite ] ) != white ||
              *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlack ] ) != black )
    {
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend0 ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend1 ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlend2 ] ) = blend;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosWhite ] ) = white;
        *reinterpret_cast< nw::ut::Color8* >( &command[ cmdPosBlack ] ) = black;

        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV1 + TEX_ENV_COLOR_OFFS ),
            command[ cmdPosBlend0 ] );

        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV2 + TEX_ENV_COLOR_OFFS ),
            command[ cmdPosBlend1 ] );

        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV3 + TEX_ENV_COLOR_OFFS ),
            command[ cmdPosBlend2 ] );

        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV4 + TEX_ENV_COLOR_OFFS ),
            command[ cmdPosWhite ] );

        // 0x0fd
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV_BUFFER_COLOR ),
            command[ cmdPosBlack ] );
    }

    return pCmdStream;
}

u32* Drawer::SetUpTextures( u32* pCmdStream, const Material* __restrict pMaterial, bool addDisableCommand )
{
    NW_PROFILE("nw::lyt::Drawer::SetUpTextures");
    static u32 command[] NW_LYT_DRAWER_ALIGN_32 =
    {
        // 0x082
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE0_SIZE ),
        // 0x085
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE0_ADDR1 ),
        // 0x08e
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE0_FORMAT ),
        // 0x083
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE0_WRAP_FILTER ),

        // 0x092
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE1_SIZE ),
        // 0x095
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE1_ADDR ),
        // 0x096
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE1_FORMAT ),
        // 0x093
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE1_WRAP_FILTER ),

        // 0x09a
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE2_SIZE ),
        // 0x09d
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE2_ADDR ),
        // 0x09e
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE2_FORMAT ),
        // 0x09b
        PICA_CMD_DATA_ZERO(), PICA_CMD_HEADER_SINGLE( PICA_REG_TEXTURE2_WRAP_FILTER ),
    };

    enum
    {
        idxWidthHeight = 0,
        idxTextureAddr = 2,
        idxFormat      = 4,
        idxWrapFilter  = 6,
        settingSize    = 8,
        commandSize    = sizeof( u32 ) * settingSize
    };

    if ( pMaterial == NULL )
    {
        for (int i = 0; i < TexMapMax; ++i)
        {
            command[ i * settingSize + idxTextureAddr ] = PICA_CMD_DATA_ZERO();
        }

        #if defined(NW_FONT_RECTDRAWER_USE_DRAW_BUFFER)
            NW_ASSERT(m_UniformAddrIndex == 0);
        #endif

        if (addDisableCommand)
        {
            pCmdStream = DisableTextures(pCmdStream);
        }

        return pCmdStream;
    }

    bool isTexCoordModified = true;
    #if defined(NW_FONT_RECTDRAWER_USE_DRAW_BUFFER)
        if (m_UniformAddrIndex > 0) // Not flushed.
        {
            NW_ASSERT(
                    m_PreviousTexEnvType == m_CurrentTexEnvType
                &&  (m_CurrentTexEnvType == TEX_ENV_TYPE_0_TEX || m_CurrentTexEnvType == TEX_ENV_TYPE_1_TEX));
            isTexCoordModified = false;
        }
    #endif

    const u32 num = ut::Min( (u32)pMaterial->GetTexMapNum(), (u32)TexMapMax );

    bool isTexturesModified = false;

    int texbit = 0;
    for ( int i = 0; i < num; ++i )
    {
        texbit |= 1 << i;

        const TexMap& __restrict texMap = pMaterial->GetTexMap( i );
        u32* __restrict c = &command[ i * settingSize ];

        if ( c[ idxTextureAddr ] != texMap.GetPhysicalAddress() / 8
          || c[ idxWrapFilter  ] != texMap.GetU32WrapFilter()
        )
        {
            c[ idxWidthHeight ] = texMap.GetU32WidthHeight();
            c[ idxTextureAddr ] = texMap.GetPhysicalAddress() / 8;
            c[ idxFormat      ] = texMap.GetU32Format();
            c[ idxWrapFilter  ] = texMap.GetU32WrapFilter();
            isTexturesModified = true;
        }
    }

    if (isTexCoordModified || isTexturesModified)
    {
        // Writes data from the buffer.
        pCmdStream = FlushBuffer(pCmdStream);
    }

    if (isTexCoordModified || isTexturesModified)
    {
        // 0x080 x 3 (dummy command. Needed before 0x80 to 0xbf.)
        static const u32 dummyCommands[] =
        {
            0, PICA_CMD_HEADER_BURST_BE( PICA_REG_TEXTURE_FUNC, 3, 0x0), 0, 0,
        };

        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, dummyCommands, sizeof(dummyCommands) );
    }

    // 0x080
    const u32 texFuncData =
        PICA_CMD_DATA_TEXTURE_FUNC(
            PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_FALSE /* texture0SamplerType: texbit */,
            0 /* texture1SamplerType: texbit */,
            0 /* texture2SamplerType: texbit */,
            0 /* texture3Texcoord */,
            0 /* texture3SamplerType */,
            0 /* texture2Texcoord */,
            1 /* clearTextureCache */) | texbit;

    if (isTexCoordModified)
    {
        // Enabling the texture coordinates.
        // 0x06f
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_VS_OUT_ATTR_CLK, 0x2 ),
            PICA_CMD_DATA_VS_GS_OUT_ATTR_CLK(
                0 /* verZ */,
                0 /* col */,
                0 /* tex0 : texbit */,
                0 /* tex1 : texbit */,
                0 /* tex2 : texbit */,
                0 /* tex0_w */,
                0 /* view_quat */ ) | texbit << 8 );

        // Select sampler type, coordinates
        // 0x080
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEXTURE_FUNC, 0xB ),
            texFuncData );
    }

    if (isTexCoordModified || isTexturesModified)
    {
        // Clears the texture cache.
        // 0x080
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEXTURE_FUNC, 0x4 ),
            texFuncData );
    }

    if (isTexturesModified)
    {
        // Adds to command list.
        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, commandSize * num );
    }

    return pCmdStream;
}

u32* Drawer::UniformAndDraw(u32* pCmdStream)
{
    NW_LYT_DRAWER_ASSERT(m_UniformAddrIndex > 0);

    { // Uniform that goes to uUniformAddr.
    	const u32 size = SetUniformCommand(m_UniformAddrBuffer, m_UniformAddrIndex);
        NW_FONT_ADD_COMMANDS_PTR(pCmdStream, m_UniformAddrBuffer, size);
    }

    if ( m_UniformMtxIndex > 0 )
    { // Uniform to uUniformMtx

        pCmdStream = Base::AddUniformMtx(pCmdStream);
    }

    if ( m_UniformDataIndex > 0 )
    { // Uniform that goes to uUniformData
        const u32 size = SetUniformCommand(m_UniformDataBuffer, m_UniformDataIndex);
        NW_FONT_ADD_COMMANDS_PTR(pCmdStream, m_UniformDataBuffer, size);
        m_UniformDataIndex = 0;
    }

    const int vtxNum = DRAW_VTX_NUM * m_UniformAddrIndex;       // Number of vertices to draw
#if defined(NW_FONT_RECTDRAWER_USE_DRAW_ARRAYS)
    const u32 vtxIdxAddrOffset = GetVertexIndexAddressOffset(vtxNum);
#else
    const u32 vtxIdxAddrOffset = 0;     // Dummy
#endif

    // Sets the number of texture coordinates in the vertex shader.
    // 0x2b1
    NW_FONT_ADD_SINGLE_COMMAND_PTR(
        pCmdStream,
        PICA_CMD_HEADER_SINGLE( PICA_REG_VS_INT0 ),
        PICA_CMD_DATA_VS_INT( m_TexCoordNum /* x */ , 0 /* y */, 0 /* z */ ));

    pCmdStream = font::internal::SetVertexNumCmd(pCmdStream, vtxIdxAddrOffset, vtxNum);

    // Adds to command list.
    NW_FONT_ADD_COMMANDS_PTR( pCmdStream, GetDrawCommands(), GetDrawCommandSize() );

    // Initializes the index.
    m_UniformAddrIndex = 0;

    return pCmdStream;
}


#if !defined(NW_FONT_ENABLE_EDGE_FONT)
u32*
Drawer::SetUpTextBox(
    u32 *pCmdStream,
    const TextBox*  __restrict pTextBox,
    const Material* __restrict pMaterial,
    const DrawInfo& drawInfo )
#else
u32*
Drawer::SetUpTextBox(
    u32 *pCmdStream,
    const TextBox* pTextBox,
    const Material* pMaterial,
    const DrawInfo& drawInfo,
    font::FontCombinerType combiner )
#endif
{
    font::DispStringBuffer* __restrict pStringBuffer = pTextBox->GetDispStringBuffer();

    // Ends if there are no characters.
    if (pStringBuffer->IsCommandEmpty())
    {
        return pCmdStream;
    }

    // Configure text rendering.
    m_CurrentTexEnvType = TEX_ENV_TYPE_TEXT;

    // Blend is the default.
    m_IsBlendDefault = true;

    // Alpha test is the default.
    m_IsAlphaTestDefault = true;

    // Configure array information.
    pTextBox->GetTextGlobalMtx( (nw::math::MTX34*)m_UniformMtx );
    m_UniformMtxIndex = 3;
    pCmdStream = Base::AddUniformMtx(pCmdStream);  // Uniform to uUniformMtx

    font::WideTextWriter& writer = drawInfo.GetGraphicsResource()->GetTextWriter();
    writer.SetDispStringBuffer(pStringBuffer);

    // Text color
    ut::Color8 topCol = pTextBox->GetTextColor(TEXTCOLOR_TOP);
    ut::Color8 btmCol = pTextBox->GetTextColor(TEXTCOLOR_BOTTOM);
    writer.SetGradationMode(
        topCol != btmCol ?
            font::CharWriter::GRADMODE_V:
            font::CharWriter::GRADMODE_NONE);
    writer.SetTextColor(topCol, btmCol);

    // 2-color interpolation + alpha
    writer.SetColorMapping(
        pMaterial->GetColor( INTERPOLATECOLOR_BLACK ),
        pMaterial->GetColor( INTERPOLATECOLOR_WHITE ));
    writer.SetAlpha(pTextBox->GetGlobalAlpha());

    // Adds commands from the cache.
#if !defined(NW_FONT_ENABLE_EDGE_FONT)
    pCmdStream = writer.UseCommandBuffer( pCmdStream, this );
#else
    pCmdStream = writer.UseCommandBuffer( pCmdStream, combiner, this );
#endif

    writer.SetDispStringBuffer(NULL);

    // Texture initialization
    pCmdStream = SetUpTextures( pCmdStream, NULL );

    return pCmdStream;
}

#ifdef NW_LYT_DRAWER_ENABLE_USER_TEX_ENV
u32*
Drawer::SetUpGLTexEnvUser(
    u32* pCmdStream,
    const Material* __restrict pMaterial )
{
    // TODO: Will make it so that retention occurs in the command format when a member is of the material class.
    // TODO: Will determine in advance if it is necessary to use stage 0.

    const int tevSrc[] =
    {
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0,
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE1,
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE2,
        PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE3,
        PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
        PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
        PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS_BUFFER_DMP,
    };

    const int tevMode[] =
    {
        PICA_DATA_TEX_ENV_COMBINE_REPLACE,
        PICA_DATA_TEX_ENV_COMBINE_MODULATE,
        PICA_DATA_TEX_ENV_COMBINE_ADD,
        PICA_DATA_TEX_ENV_COMBINE_ADD_SIGNED,
        PICA_DATA_TEX_ENV_COMBINE_INTERPOLATE,
        PICA_DATA_TEX_ENV_COMBINE_SUBTRACT,
        PICA_DATA_TEX_ENV_COMBINE_ADD_MULT_DMP,
        PICA_DATA_TEX_ENV_COMBINE_MULT_ADD_DMP,
    };

    const int tevOpRgb[] =
    {
        PICA_DATA_OPE_RGB_SRC_COLOR,
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_COLOR,
        PICA_DATA_OPE_RGB_SRC_ALPHA,
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_ALPHA,
        PICA_DATA_OPE_RGB_SRC_R_DMP,
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_R_DMP,
        PICA_DATA_OPE_RGB_SRC_G_DMP,
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_G_DMP,
        PICA_DATA_OPE_RGB_SRC_B_DMP,
        PICA_DATA_OPE_RGB_ONE_MINUS_SRC_B_DMP,
    };

    const int tevOpAlp[] =
    {
        PICA_DATA_OPE_ALPHA_SRC_ALPHA,
        PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_ALPHA,
        PICA_DATA_OPE_ALPHA_SRC_R_DMP,
        PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_R_DMP,
        PICA_DATA_OPE_ALPHA_SRC_G_DMP,
        PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_G_DMP,
        PICA_DATA_OPE_ALPHA_SRC_B_DMP,
        PICA_DATA_OPE_ALPHA_ONE_MINUS_SRC_B_DMP,
    };

    const u32 scale[] =
    {
        PICA_DATA_TEX_ENV_SCALE_1,
        PICA_DATA_TEX_ENV_SCALE_2,
        PICA_DATA_TEX_ENV_SCALE_4,
    };

    const u32 reg[] =
    {
        PICA_REG_TEX_ENV0,
        PICA_REG_TEX_ENV1,
        PICA_REG_TEX_ENV2,
        PICA_REG_TEX_ENV3,
        PICA_REG_TEX_ENV4,
        PICA_REG_TEX_ENV5,
    };

    u32 bufferInput = NW_LYT_CMD_DATA_TEX_ENV_BUF_INPUT_THROUGH;

    u32 stageNum = nw::ut::Min( (u32)TEX_ENV_STAGE_NUM, (u32)pMaterial->GetTevStageNum() );
    for ( int i = 0; i < stageNum; ++i )
    {
        const TevStage& tevStage = pMaterial->GetTevStage( i );

        nw::ut::Color8 rgb = pMaterial->GetColor( tevStage.GetKonstSelRgb() );
        nw::ut::Color8 alpha = pMaterial->GetColor( tevStage.GetKonstSelAlpha() );

        int stageNo = i + ( TEX_ENV_STAGE_NUM - stageNum );

        if (i == 0)
        {
            // Determines if stage 0 must be used.
            if ( ( tevStage.GetSrcRgb0() != TEVSRC_CONSTANT &&
                   tevStage.GetSrcRgb1() != TEVSRC_CONSTANT &&
                   tevStage.GetSrcRgb2() != TEVSRC_CONSTANT ) ||
                 ( tevStage.GetSrcAlpha0() != TEVSRC_CONSTANT &&
                   tevStage.GetSrcAlpha1() != TEVSRC_CONSTANT &&
                   tevStage.GetSrcAlpha2() != TEVSRC_CONSTANT ) )
            {
                // Through setting
                for ( int j = 1; j <= stageNo; ++j)
                {
                    // Header
                    NW_FONT_ADD_SINGLE_COMMAND_PTR(
                        pCmdStream,
                        PICA_CMD_HEADER_BURSTSEQ( reg[ j ], 5 ),
                        PICA_CMD_DATA_TEX_ENV_SRC(
                            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
                            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                            PICA_DATA_TEX_ENV_SRC_RGBA_PREVIOUS,
                            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT,
                            PICA_DATA_TEX_ENV_SRC_RGBA_CONSTANT ) );

                    // Continuation
                    static const u32 command[] =
                    {
                        PICA_CMD_DATA_TEX_ENV_OPERAND(
                            PICA_DATA_OPE_RGB_SRC_COLOR,
                            PICA_DATA_OPE_RGB_SRC_COLOR,
                            PICA_DATA_OPE_RGB_SRC_COLOR,
                            PICA_DATA_OPE_ALPHA_SRC_ALPHA,
                            PICA_DATA_OPE_ALPHA_SRC_ALPHA,
                            PICA_DATA_OPE_ALPHA_SRC_ALPHA ),
                        PICA_CMD_DATA_TEX_ENV_COMBINE(
                            PICA_DATA_TEX_ENV_COMBINE_REPLACE,
                            PICA_DATA_TEX_ENV_COMBINE_REPLACE ),
                        PICA_CMD_DATA_TEX_ENV_BUFFER_COLOR_WHITE,
                        PICA_CMD_DATA_TEX_ENV_SCALE(
                            PICA_DATA_TEX_ENV_SCALE_1, PICA_DATA_TEX_ENV_SCALE_1 ),
                    };

                    NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
                }

                // Uses stage 0.
                stageNo = 0;
            }
        }

        if ( 1 <= i && i <= 4 )
        {
            bufferInput |= PICA_CMD_DATA_TEX_ENV_BUFFER_INPUT(
                stageNo,
                ( tevStage.GetSavePrevRgb() ) ?
                    PICA_DATA_TEX_ENV_BUFFER_INPUT_PREVIOUS :
                    PICA_DATA_TEX_ENV_BUFFER_INPUT_PREVIOUS_BUFFER_DMP,
                ( tevStage.GetSavePrevAlpha() ) ?
                    PICA_DATA_TEX_ENV_BUFFER_INPUT_PREVIOUS :
                    PICA_DATA_TEX_ENV_BUFFER_INPUT_PREVIOUS_BUFFER_DMP );
        }

        const u32 command[] =
        {
            PICA_CMD_DATA_TEX_ENV_SRC(
                tevSrc[ tevStage.GetSrcRgb0() ],   tevSrc[ tevStage.GetSrcRgb1() ],  tevSrc[ tevStage.GetSrcRgb2() ],
                tevSrc[ tevStage.GetSrcAlpha0() ], tevSrc[ tevStage.GetSrcAlpha1()], tevSrc[ tevStage.GetSrcAlpha2() ] ),
            PICA_CMD_HEADER_BURSTSEQ( reg[ stageNo ], 5 ),
            PICA_CMD_DATA_TEX_ENV_OPERAND(
                tevOpRgb[ tevStage.GetOperandRgb0() ],   tevOpRgb[ tevStage.GetOperandRgb1() ],   tevOpRgb[ tevStage.GetOperandRgb2() ],
                tevOpAlp[ tevStage.GetOperandAlpha0() ], tevOpAlp[ tevStage.GetOperandAlpha1() ], tevOpAlp[ tevStage.GetOperandAlpha2() ] ),
            PICA_CMD_DATA_TEX_ENV_COMBINE( tevMode[ tevStage.GetCombineRgb() ], tevMode[ tevStage.GetCombineAlpha() ] ),
            NW_LYT_CMD_DATA_COLOR_ALPHA( rgb, alpha ),
            PICA_CMD_DATA_TEX_ENV_SCALE( scale[ tevStage.GetScaleRgb() ], scale[ tevStage.GetScaleAlpha() ] ),
        };

        // Adds to command list.
        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }

    // Buffer color and buffer input settings
    {
        nw::ut::Color8 black = pMaterial->GetColor( TEVKONSTSEL_BUFFER );

        // 0x0fd
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE( PICA_REG_TEX_ENV_BUFFER_COLOR ),
            NW_LYT_CMD_DATA_COLOR(black) );

        // 0x0e0
        NW_FONT_ADD_SINGLE_COMMAND_PTR(
            pCmdStream,
            PICA_CMD_HEADER_SINGLE_BE( PICA_REG_TEX_ENV_BUFFER_INPUT, 0x2 ),
            bufferInput );
    }

    return pCmdStream;
}
#endif

#ifdef NW_LYT_DRAWER_ENABLE_ALPHA_TEST
u32*
Drawer::SetUpAlphaTest( u32* pCmdStream, const Material* __restrict pMaterial )
{
    const AlphaCompare& __restrict alphaCompare = pMaterial->GetAlphaCompare();
    register f32 refValue = alphaCompare.GetRef() * 255.f + 0.5f;

    const u32 alphaTest[] =
    {
        PICA_DATA_ALPHA_TEST_NEVER,
        PICA_DATA_ALPHA_TEST_LESS,
        PICA_DATA_ALPHA_TEST_LEQUAL,
        PICA_DATA_ALPHA_TEST_EQUAL,
        PICA_DATA_ALPHA_TEST_NOTEQUAL,
        PICA_DATA_ALPHA_TEST_GEQUAL,
        PICA_DATA_ALPHA_TEST_GREATER,
        PICA_DATA_ALPHA_TEST_ALWAYS,
    };

    // 0x104
    NW_FONT_ADD_SINGLE_COMMAND_PTR(
        pCmdStream,
        PICA_CMD_HEADER_SINGLE( PICA_REG_FRAGOP_ALPHA_TEST ),
        PICA_CMD_DATA_FRAGOP_ALPHA_TEST( true, alphaTest[ alphaCompare.GetFunc() ], (u8)refValue ) );

    return pCmdStream;
}
#endif

#ifdef NW_LYT_DRAWER_ENABLE_BLEND_FUNC
u32*
Drawer::SetUpBlendMode( u32* pCmdStream, const Material* __restrict pMaterial )
{
    const BlendMode& blendMode = pMaterial->GetBlendMode();

    if ( blendMode.GetBlendOp() == BLENDOP_DISABLE &&
         blendMode.GetLogicOp() == LOGICOP_DISABLE )
    {
        // Command with no blend
        u32 command[] =	{
            NW_FONT_COMMAND_SET_BLEND_FUNC(
                PICA_DATA_BLEND_EQUATION_ADD,
                PICA_DATA_BLEND_FUNC_ONE,
                PICA_DATA_BLEND_FUNC_ZERO )
        };

        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }
    else if ( blendMode.GetBlendOp() != BLENDOP_DISABLE )
    {
        // Blend settings

        const u32 blendOp[] =
        {
            0, // DISABLE
            PICA_DATA_BLEND_EQUATION_ADD,
            PICA_DATA_BLEND_EQUATION_SUBTRACT,
            PICA_DATA_BLEND_EQUATION_REVERSE_SUBTRACT,
        };

        const u32 srcFactor[] =
        {
            PICA_DATA_BLEND_FUNC_ZERO,
            PICA_DATA_BLEND_FUNC_ONE,
            PICA_DATA_BLEND_FUNC_DST_COLOR,
            PICA_DATA_BLEND_FUNC_ONE_MINUS_DST_COLOR,
            PICA_DATA_BLEND_FUNC_SRC_ALPHA,
            PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA,
            PICA_DATA_BLEND_FUNC_DST_ALPHA,
            PICA_DATA_BLEND_FUNC_ONE_MINUS_DST_ALPHA,
        };

        const u32 dstFactor[] =
        {
            PICA_DATA_BLEND_FUNC_ZERO,
            PICA_DATA_BLEND_FUNC_ONE,
            PICA_DATA_BLEND_FUNC_SRC_COLOR,
            PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_COLOR,
            PICA_DATA_BLEND_FUNC_SRC_ALPHA,
            PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA,
            PICA_DATA_BLEND_FUNC_DST_ALPHA,
            PICA_DATA_BLEND_FUNC_ONE_MINUS_DST_ALPHA,
        };

        const u32 command[] =
        {
            NW_FONT_COMMAND_SET_BLEND_FUNC(
                blendOp[   blendMode.GetBlendOp() ],
                srcFactor[ blendMode.GetSrcFactor() ],
                dstFactor[ blendMode.GetDstFactor() ] ),
        };
        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }
    else if ( blendMode.GetLogicOp() != LOGICOP_DISABLE )
    {
        // Logical operation setting

        const u32 logicOp[] =
        {
            0 , // DISABLE
            PICA_DATA_LOGIC_NOOP,
            PICA_DATA_LOGIC_CLEAR,
            PICA_DATA_LOGIC_SET,
            PICA_DATA_LOGIC_COPY,
            PICA_DATA_LOGIC_COPY_INVERTED,
            PICA_DATA_LOGIC_INVERT,
            PICA_DATA_LOGIC_AND,
            PICA_DATA_LOGIC_NAND,
            PICA_DATA_LOGIC_OR,
            PICA_DATA_LOGIC_NOR,
            PICA_DATA_LOGIC_XOR,
            PICA_DATA_LOGIC_EQUIV,
            PICA_DATA_LOGIC_AND_REVERSE,
            PICA_DATA_LOGIC_AND_INVERTED,
            PICA_DATA_LOGIC_OR_REVERSE,
            PICA_DATA_LOGIC_OR_INVERTED,
        };

        const u32 command[] =
        {
            NW_LYT_COMMAND_SET_LOGIC_OP( logicOp[ blendMode.GetLogicOp() ] ),
        };
        NW_FONT_ADD_COMMANDS_PTR( pCmdStream, command, sizeof( command ) );
    }

    return pCmdStream;
}
#endif

}   // namespace lyt
}   // namespace nw

#endif // NW_LYT_DRAWER_ENABLE
