/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"
#include <nw/lyt/lyt_Layout.h>
#include <nw/lyt/lyt_Pane.h>
#include <nw/lyt/lyt_Drawer.h>

#ifdef NW_LYT_DRAWER_ENABLE

// For implementing MakeUniformDataSelf.
#include <nw/lyt/lyt_Picture.h>
#include <nw/lyt/lyt_Window.h>
#include <nw/lyt/lyt_TextBox.h>

namespace nw { namespace lyt {

namespace local { namespace {

const nw::math::VEC2 Transform(
    const nw::math::MTX22& mtx,
    const nw::math::VEC2& vec)
{
    return nw::math::VEC2(
        mtx.f._00 * vec.x + mtx.f._01 * vec.y,
        mtx.f._10 * vec.x + mtx.f._11 * vec.y);
}

}}

u32*
Drawer::SetUniformDataEnd(u32* pCmdStream)
{
    NW_PROFILE("nw::lyt::Drawer::SetUniformDataEnd");
    m_UniformAddr[ m_UniformAddrIndex ] = m_UniformAddress;

    ++m_UniformAddrIndex;

#ifdef NW_FONT_RECTDRAWER_USE_DRAW_BUFFER

    if ( m_UniformAddrIndex + 1                    > UNIFORM_ADDR_NUM
      || m_UniformMtxIndex  + UNIFORM_MTX_USE_MAX  > UNIFORM_MTX_NUM
      || m_UniformDataIndex + UNIFORM_DATA_USE_MAX > UNIFORM_DATA_NUM
    )
    {
        pCmdStream = UniformAndDraw(pCmdStream);
    }

#else

    pCmdStream = UniformAndDraw(pCmdStream);

#endif

    return pCmdStream;
}

u8
Drawer::CalcTextureCoords( const Material* pMaterial,
                                const TexCoordQuad* pTexCoordQuad,
                                nw::math::VEC4 texCoords[ TexMapMax * 2 ] ) const
{
    NW_PROFILE("nw::lyt::Drawer::CalcTextureCoords");
    // Texture coordinate settings
    const u32 num = ut::Min( (u32)pMaterial->GetTexCoordGenNum(), (u32)TexMapMax );

    if ( 0 == num ) return 0;

    // TODO: Fix to center... Would like to allow editing with the editor and be able to "get" from Material.
    static const nw::math::VEC2 center( 0.5f, 0.5f );

    for ( int i = 0; i < num; ++i )
    {
        const int src = pMaterial->GetTexCoordGen( i ).GetTexGenSrc();

        const TexMap& texMap = pMaterial->GetTexMap( i );
        const f32 adjust_x = static_cast<f32>(texMap.GetWidth()) / texMap.GetRealWidth();
        const f32 adjust_y = static_cast<f32>(texMap.GetHeight()) / texMap.GetRealHeight();

        const TexSRT& texSRT = pMaterial->GetTexSRT( i );
        register const f32 rotate = texSRT.rotate;
        register const f32 scalex = texSRT.scale.x;
        register const f32 scaley = texSRT.scale.y;

        if ( rotate != 0.f || scalex != 1.f || scaley != 1.f )
        {
            f32 cos, sin;
            nw::math::SinCosDeg( &sin, &cos, rotate );

            register nw::math::MTX22 srMtx(
                cos * scalex * adjust_x, -sin * scaley * adjust_x,
                sin * scalex * adjust_y,  cos * scaley * adjust_y );

            register nw::math::VEC2 ctrans( ( center.x + texSRT.translate.x ) * adjust_x,
                                            ( center.y + texSRT.translate.y ) * adjust_y );

            { // Upper left, lower right
                nw::math::VEC4* pTexCoord = &texCoords[ i ];

                register nw::math::VEC2 lt( pTexCoordQuad[ src ][ VERTEX_LT ].x,
                                            pTexCoordQuad[ src ][ VERTEX_LT ].y);
                register nw::math::VEC2 rb( pTexCoordQuad[ src ][ VERTEX_RB ].x,
                                            pTexCoordQuad[ src ][ VERTEX_RB ].y );

                lt = local::Transform(srMtx, lt - center) + ctrans;
                rb = local::Transform(srMtx, rb - center) + ctrans;

                pTexCoord->x = lt.x ;
                pTexCoord->y = 1.f - lt.y;
                pTexCoord->z = rb.x ;
                pTexCoord->w = 1.f - rb.y;
            }

            { // Lower left, upper right and shift downward.
                nw::math::VEC4* pTexCoord = &texCoords[ i + TexMapMax ];

                register nw::math::VEC2 lb( pTexCoordQuad[ src ][ VERTEX_LB ].x,
                                            pTexCoordQuad[ src ][ VERTEX_LB ].y );
                register nw::math::VEC2 rt( pTexCoordQuad[ src ][ VERTEX_RT ].x,
                                            pTexCoordQuad[ src ][ VERTEX_RT ].y );

                lb = local::Transform(srMtx, lb - center) + ctrans;
                rt = local::Transform(srMtx, rt - center) + ctrans;

                pTexCoord->x = lb.x;
                pTexCoord->y = 1.f - lb.y;
                pTexCoord->z = rt.x;
                pTexCoord->w = 1.f - rt.y;
            }
        }
        else
        {
            register const nw::math::VEC2 trans( texSRT.translate.x, texSRT.translate.y );

            { // Upper left, lower right
                nw::math::VEC4* pTexCoord = &texCoords[ i ];

                pTexCoord->x = ( pTexCoordQuad[ src ][ VERTEX_LT ].x + trans.x ) * adjust_x;
                pTexCoord->y = 1.f - ( pTexCoordQuad[ src ][ VERTEX_LT ].y + trans.y ) * adjust_y;
                pTexCoord->z = ( pTexCoordQuad[ src ][ VERTEX_RB ].x + trans.x ) * adjust_x;
                pTexCoord->w = 1.f - ( pTexCoordQuad[ src ][ VERTEX_RB ].y + trans.y ) * adjust_y;
            }

            { // Lower left, upper right and shift downward.
                nw::math::VEC4* pTexCoord = &texCoords[ i + TexMapMax ];

                pTexCoord->x = ( pTexCoordQuad[ src ][ VERTEX_LB ].x + trans.x) * adjust_x;
                pTexCoord->y = 1.f - ( pTexCoordQuad[ src ][ VERTEX_LB ].y + trans.y ) * adjust_y;
                pTexCoord->z = ( pTexCoordQuad[ src ][ VERTEX_RT ].x + trans.x) * adjust_x;
                pTexCoord->w = 1.f - ( pTexCoordQuad[ src ][ VERTEX_RT ].y + trans.y ) * adjust_y;
            }
        }
    }

    // Uses register 6 if texture coordinates cannot be omitted.
    for ( int i = 0; i < num; ++i )
    {
        if ( texCoords[ i ].x != texCoords[ i + TexMapMax ].x || // Determines if the left side is aligned.
             texCoords[ i ].z != texCoords[ i + TexMapMax ].z || // Determines if the right side is aligned.
             texCoords[ i ].y != texCoords[ i + TexMapMax ].w || // Determines if the top is aligned.
             texCoords[ i ].w != texCoords[ i + TexMapMax ].y )  // Determines if the bottom is aligned.
        {
            return TexMapMax * 2;
        }
    }

    return num;
}


u32*
Drawer::SetUpTextureCoords(
    u32* pCmdStream,
    const nw::math::VEC4 texCoords[ TexMapMax * 2 ],
    const int texCoordNum )
{
    NW_PROFILE("nw::lyt::Drawer::SetUpTextureCoords");
    if ( texCoordNum != m_TexCoordNum )
    {
        pCmdStream = FlushBuffer(pCmdStream);
        m_TexCoordNum = texCoordNum;
    }

    if ( texCoordNum != 0 )
    {
        if ( 2 <= texCoordNum )
        {
            pCmdStream = FlushBuffer(pCmdStream);
        }

        m_UniformAddress.addrTexCoord = m_UniformDataIndex;
        nw::math::VEC4* pTexCoord = &m_UniformData[ m_UniformDataIndex ];

        std::memcpy( &m_UniformData[ m_UniformDataIndex ],
                     texCoords,
                     sizeof( nw::math::VEC4 ) * texCoordNum );

        m_UniformDataIndex += texCoordNum;
    }

    return pCmdStream;
}


void
Drawer::SetCurrentUniformAlpha( const f32 alpha )
{
    m_UniformAddress.addrColor = alpha;
}

void
Drawer::SetUpMtx( const nw::math::MTX34& __restrict mtx )
{
    // std::memcpy( &m_UniformMtx[ m_UniformMtxIndex ], &mtx, sizeof(nw::math::MTX34) );
    MTX34Copy( ( nw::math::MTX34* )&m_UniformMtx[ m_UniformMtxIndex ], &mtx );

    m_UniformAddress.addrMtx = m_UniformMtxIndex;
    m_UniformMtxIndex += 3;
}

void
Drawer::SetUpQuad( const nw::lyt::Size&  __restrict size,
                        const nw::math::VEC2& __restrict basePt )
{
    nw::math::VEC4* __restrict pQuadInfo = &m_UniformData[ m_UniformDataIndex ];
    pQuadInfo->x = size.width;
    pQuadInfo->y = size.height;
    pQuadInfo->z = basePt.x;
    pQuadInfo->w = basePt.y;

    m_UniformAddress.addrSizeAndVtx = m_UniformDataIndex;
    ++m_UniformDataIndex;
}

void
Drawer::SetUpVtxColors( const nw::ut::Color8 vtxColors[ VERTEX_MAX ], const u8 globalAlpha )
{
    NW_PROFILE("nw::lyt::Drawer::SetUpVtxColors");
    register f32 d = 1.f / 255.f;
    register f32 colors[ VERTEX_MAX ][ 4 ];
    bool isWhite = true;
    register f32 alpha = globalAlpha * d;

    for ( int i = 0; i < VERTEX_MAX; ++i )
    {
        if ( nw::ut::Color8::WHITE != vtxColors[i] )
        {
            colors[ i ][ 0 ] = vtxColors[ i ].r * d;
            colors[ i ][ 1 ] = vtxColors[ i ].g * d;
            colors[ i ][ 2 ] = vtxColors[ i ].b * d;
            colors[ i ][ 3 ] = vtxColors[ i ].a * d;

            for ( int j = 0; isWhite && j < i; ++j )
            {
                colors[ j ][ 0 ] = 1.f;
                colors[ j ][ 1 ] = 1.f;
                colors[ j ][ 2 ] = 1.f;
                colors[ j ][ 3 ] = alpha;
            }

            isWhite = false;

            // Multiply the alpha.
            colors[ i ][ 3 ] *= alpha;
        }
        else if ( !isWhite )
        {
            colors[ i ][ 0 ] = 1.f;
            colors[ i ][ 1 ] = 1.f;
            colors[ i ][ 2 ] = 1.f;
            colors[ i ][ 3 ] = alpha;
        }
    }

    if ( isWhite )
    {
        m_UniformAddress.addrColor = alpha;
    }
    else
    {
        m_UniformAddress.addrColor = m_UniformMtxIndex;

        for ( int i = 0; i < VERTEX_MAX; ++i )
        {
            m_UniformMtx[ m_UniformMtxIndex + i ].x = colors[ i ][ 0 ];
            m_UniformMtx[ m_UniformMtxIndex + i ].y = colors[ i ][ 1 ];
            m_UniformMtx[ m_UniformMtxIndex + i ].z = colors[ i ][ 2 ];
            m_UniformMtx[ m_UniformMtxIndex + i ].w = colors[ i ][ 3 ];
        }
        m_UniformMtxIndex += VERTEX_MAX;
    }
}


//----------------------------------------------------------
// TextBox
u32*
TextBox::MakeUniformDataSelf(
    u32* pCmdStream,
    DrawInfo* pDrawInfo,
    Drawer* __restrict pDrawer ) const
{
    NW_PROFILE("nw::lyt::TexBox::MakeUniformDataSelf");
#ifdef NW_LYT_DRAWER_DISABLE_DRAW_TEXTBOX
    return;
#endif

    if ( m_TextLen <= 0 || !m_pFont || !m_pMaterial )
    {
        return pCmdStream;
    }

    pCmdStream = pDrawer->FlushBuffer(pCmdStream);

    if (!(this->GetDispStringBuffer()->GetCommandBufferMode() & font::DispStringBuffer::CBF_STATIC))
    {
        const_cast<TextBox*>(this)->UpdateDrawCharDataImpl(pDrawer);
    }

#if !defined(NW_FONT_ENABLE_EDGE_FONT)
    pCmdStream = pDrawer->SetUpTextBox( pCmdStream, this, m_pMaterial, *pDrawInfo );
#else
    pCmdStream = pDrawer->SetUpTextBox( pCmdStream, this, m_pMaterial, *pDrawInfo, m_CombinerType );
#endif

    return pCmdStream;
}

//----------------------------------------------------------
// Picture
u32*
Picture::MakeUniformDataSelf(
    u32* pCmdStream,
    DrawInfo* pDrawInfo,
    Drawer* __restrict pDrawer ) const
{
    NW_PROFILE("nw::lyt::Picture::MakeUniformDataSelf");
    NW_UNUSED_VARIABLE( pDrawInfo );

#ifdef NW_LYT_DRAWER_DISABLE_DRAW_PICTURE
    return;
#endif

    { // TexEnv settings
         pCmdStream = pDrawer->SetUpTexEnv( pCmdStream, m_pMaterial );
    }

    { // Texture settings
        pCmdStream = pDrawer->SetUpTextures( pCmdStream, m_pMaterial );
    }

    { // Texture coordinate settings
        if ( !m_IsTexCoordInited || m_pMaterial->IsTextureDirty() )
        {
            m_UniformTexCoordNum = pDrawer->CalcTextureCoords( m_pMaterial,
                                                                    m_TexCoordAry.GetArray(),
                                                                    m_UniformTexCoords );
            m_IsTexCoordInited = true;
            m_pMaterial->SetTextureDirty( false );
        }

        pCmdStream = pDrawer->SetUpTextureCoords( pCmdStream, m_UniformTexCoords, m_UniformTexCoordNum );
    }

    { // Settings for information required to calculate vertex coordinates.
        pDrawer->SetUpMtx( GetGlobalMtx() );
        pDrawer->SetUpQuad( GetSize(), GetVtxPos() );
    }

    { // Set vertex color.
        pDrawer->SetUpVtxColors( m_VtxColors, GetGlobalAlpha() );
    }

    { // Ending UniformData settings.
        pCmdStream = pDrawer->SetUniformDataEnd(pCmdStream);
    }

    return pCmdStream;
}


//----------------------------------------------------------
// Pane
u32*
Pane::MakeUniformData(
    u32* pCmdStream,
    DrawInfo* pDrawInfo,
    Drawer* __restrict pDrawer ) const
{
    NW_UNUSED_VARIABLE( pDrawInfo );

    // Renders itself.
    if ( !this->GetInvisibleByAlphaEnable() || this->GetGlobalAlpha() != 0 )
    {
        pCmdStream = MakeUniformDataSelf( pCmdStream, pDrawInfo, pDrawer );
    }

    // TODO: Would like to stop recursion.

    // Child rendering
    for ( PaneList::ConstIterator child = m_ChildList.begin(); child != m_ChildList.end(); ++child )
    {
        if ( child->IsVisible() )
        {
            pCmdStream = child->MakeUniformData( pCmdStream, pDrawInfo, pDrawer );
        }
    }

    return pCmdStream;
}

u32*
Pane::MakeUniformDataSelf(
    u32* pCmdStream,
    DrawInfo* /*pDrawInfo*/,
    Drawer* /*pDrawer*/ ) const
{
    return pCmdStream;
}

}}
#endif
