/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/lyt/lyt_Common.h>
#include <nw/lyt/lyt_DrawInfo.h>
#include <nw/lyt/lyt_Layout.h>
#include <nw/lyt/lyt_Picture.h>
#include <nw/lyt/lyt_TexMap.h>
#include <nw/lyt/lyt_Material.h>
#include <nw/lyt/lyt_Animation.h>
#include <nw/lyt/lyt_ResourceAccessor.h>
#include <nw/lyt/lyt_Stopwatch.h>

using namespace nw::math;

namespace nw
{
namespace lyt
{

NW_UT_RUNTIME_TYPEINFO_DEFINITION(Picture, Picture::Base);    // Defines an instance of runtime type information

Picture::Picture(u8 texNum)
{
    NW_ASSERT(texNum <= TexMapMax);

    Init(texNum);

    // Material creation
    m_pMaterial = Layout::NewObj<Material>();
    if (m_pMaterial)
    {
        m_pMaterial->ReserveMem(texNum, texNum, texNum);
    }
}

Picture::Picture(const TexMap& texMap)
{
    const int texNum = 1;
    Init(texNum);

    // Material creation
    m_pMaterial = Layout::NewObj<Material>();
    if (m_pMaterial)
    {
        m_pMaterial->ReserveMem(texNum, texNum, texNum);
        Append(texMap);
    }
}

Picture::Picture(const TextureInfo& texInfo)
{
    const int texNum = 1;
    Init(texNum);

    // Material creation
    m_pMaterial = Layout::NewObj<Material>();
    if (m_pMaterial)
    {
        m_pMaterial->ReserveMem(texNum, texNum, texNum);
        Append(texInfo);
    }
}

Picture::Picture(
    const res::Picture* pBlock,
    const ResBlockSet& resBlockSet
)
:   Base(pBlock)
{
    const u8 texCoordNum = ut::Min((int)pBlock->texCoordNum, TexMapMax);

    Init(texCoordNum);

    // Vertex color
    for (int i = 0; i < VERTEXCOLOR_MAX; ++i)
    {
        m_VtxColors[i] = pBlock->vtxCols[i];
    }

    // Texture coordinates
    if (texCoordNum > 0)
    {
        if (! m_TexCoordAry.IsEmpty())
        {
            m_TexCoordAry.Copy(reinterpret_cast<const char*>(pBlock) + sizeof(*pBlock), texCoordNum);
        }
    }

    // Material creation
    {
        NW_NULL_ASSERT(resBlockSet.pMaterialList);
        const u32 *const matOffsTbl = internal::ConvertOffsToPtr<u32>(resBlockSet.pMaterialList, sizeof(*resBlockSet.pMaterialList));
        const res::Material *const pResMaterial = internal::ConvertOffsToPtr<res::Material>(resBlockSet.pMaterialList, matOffsTbl[pBlock->materialIdx]);
        m_pMaterial = Layout::NewObj<Material>(pResMaterial, resBlockSet);
    }
}

void
Picture::Init(u8 texNum)
{
    if (texNum > 0)
    {
        ReserveTexCoord(texNum);
    }

#ifdef NW_LYT_DRAWER_ENABLE
    m_IsTexCoordInited = false;
#endif

}

Picture::~Picture()
{
    // OSReport("Picture::~Picture()\n");

    if (m_pMaterial && ! m_pMaterial->IsUserAllocated())
    {
        Layout::DeleteObj(m_pMaterial);
        m_pMaterial = 0;
    }

    m_TexCoordAry.Free();
}

u8
Picture::GetMaterialNum() const
{
    return m_pMaterial? 1 : 0;
}

Material*
Picture::GetMaterial(u32 idx) const
{
    NW_WARNING(idx < GetMaterialNum(), "idx >= GetMaterialNum() : %d >= %d", idx, GetMaterialNum());

    return idx == 0 ? m_pMaterial : 0;
}

void Picture::SetMaterial(Material* pMaterial)
{
    if (m_pMaterial == pMaterial)
    {
        return;
    }

    if (m_pMaterial != NULL && !m_pMaterial->IsUserAllocated())
    {
        Layout::DeleteObj(m_pMaterial);
    }

    m_pMaterial = pMaterial;
    if (pMaterial != NULL)
    {
        pMaterial->SetTextureDirty();
    }
}

void
Picture::Append(const TexMap& texMap)
{
    if (m_pMaterial->GetTexMapNum() >= m_pMaterial->GetTexMapCap() || m_pMaterial->GetTexCoordGenNum() >= m_pMaterial->GetTexCoordGenCap())
    {
        NW_WARNING(false, "m_pMaterial->GetTexMapNum(%d) is large. m_pMaterial->GetTexMapCap(%d), m_pMaterial->GetTexCoordGenCap(%d)\n", m_pMaterial->GetTexMapNum(), m_pMaterial->GetTexMapCap(), m_pMaterial->GetTexCoordGenCap());
        return;
    }

    const u8 texIdx = m_pMaterial->GetTexMapNum();
    m_pMaterial->SetTexMapNum(u8(texIdx + 1));
    m_pMaterial->SetTexMap(texIdx, texMap);

    m_pMaterial->SetTexCoordGenNum(m_pMaterial->GetTexMapNum());
    m_pMaterial->SetTexCoordGen(texIdx, TexCoordGen());

    SetTexCoordNum(m_pMaterial->GetTexMapNum());

    // Set the size if it has not been set.
    if (GetSize() == Size(0.f, 0.f) && m_pMaterial->GetTexMapNum() == 1)
    {
        const TexSize& texSize = m_pMaterial->GetTexMap(0).GetSize();
        SetSize(Size(texSize.width, texSize.height));
    }
}

void
Picture::ReserveTexCoord(u8 num)
{
    // TODO:
    // * There is no method for obtaining the Cap value.
    // * It can be expanded, but not shrunk.
    m_TexCoordAry.Reserve(num);
}

u8
Picture::GetTexCoordNum() const
{
    return m_TexCoordAry.GetSize();
}

void
Picture::SetTexCoordNum(u8 num)
{
    m_TexCoordAry.SetSize(num);
}

void
Picture::GetTexCoord(
    u32 idx,
    TexCoordQuad coords
) const
{
    return m_TexCoordAry.GetCoord(idx, coords);
}

void
Picture::SetTexCoord(
    u32 idx,
    const TexCoordQuad coords
)
{
    m_TexCoordAry.SetCoord(idx, coords);

    if (m_pMaterial != NULL)
    {
        m_pMaterial->SetTextureDirty();
    }
}

const ut::Color8
Picture::GetVtxColor(u32 idx) const
{
    NW_ASSERT(idx < VERTEXCOLOR_MAX);

    return m_VtxColors[idx];
}

void
Picture::SetVtxColor(
    u32 idx,
    ut::Color8 value
)
{
    NW_ASSERT(idx < VERTEXCOLOR_MAX);

    m_VtxColors[idx] = value;
}

u8
Picture::GetVtxColorElement(u32 idx) const
{
    return internal::GetVtxColorElement(m_VtxColors, idx);
}

void
Picture::SetVtxColorElement(u32 idx, u8 value)
{
    internal::SetVtxColorElement(m_VtxColors, idx, value);
}

#ifdef NW_LYT_DMPGL_ENABLED
void
Picture::DrawSelf(const DrawInfo& drawInfo)
{
    NW_LYT_STOPWATCH_MEASURE(-300, "nw::lyt::Picture::DrawSelf");

    if (! m_pMaterial)
    {
        return;
    }

    LoadMtx(drawInfo);  // Matrix

    NW_GL_ASSERT();

    m_pMaterial->SetupGraphics(drawInfo, GetGlobalAlpha());

    NW_GL_ASSERT();

    internal::DrawQuad(
        drawInfo,
        GetVtxPos(),
        GetSize(),
        m_TexCoordAry.GetSize(),
        m_TexCoordAry.GetArray(),
        m_VtxColors);

    NW_GL_ASSERT();
}
#endif

} // namespace nw::lyt
} // namespace nw
