/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "precompiled.h"

#include <nw/lyt/lyt_DrawInfo.h>
#include <nw/lyt/lyt_GraphicsResource.h>
#include <nw/lyt/lyt_Layout.h>
#include <nw/lyt/lyt_TextBox.h>
#include <nw/lyt/lyt_Material.h>
#include <nw/lyt/lyt_Animation.h>
#include <nw/lyt/lyt_Common.h>
#include <nw/lyt/lyt_ResourceAccessor.h>
#include <nw/lyt/lyt_Stopwatch.h>
#include <nw/font/font_DispStringBuffer.h>

namespace nw
{
namespace lyt
{
namespace internal
{
namespace
{

inline u8
ClampColor(s16 colVal)
{
    return u8(colVal < 0 ? 0: (colVal > 255 ? 255: colVal));
}

/*
  

  
  
                        

  
 */
inline
f32
AdjustCenterValue(
    f32 value,
    bool isCeil
)
{
    f32 ret = value / 2;
    return isCeil ? math::FCeil(ret): ret;
}

} // namespace nw::lyt::internal::{anonymous}

} // namespace nw::lyt::internal

using namespace math;

NW_UT_RUNTIME_TYPEINFO_DEFINITION(TextBox, TextBox::Base);    // Defines an instance of runtime type information

TextBox::TextBox(
    u16 allocStrLen,
    CommandBufferMode cmdBuffMode)
{
    Init(allocStrLen, cmdBuffMode);
    InitMaterial();
}

TextBox::TextBox(
    u16 allocStrLen,
    const wchar_t* str,
    const font::Font* pFont,
    CommandBufferMode cmdBuffMode
)
{
    Init(allocStrLen, cmdBuffMode);
    SetFont(pFont);
    SetString(str);
    InitMaterial();
}

TextBox::TextBox(
    u16 allocStrLen,
    const wchar_t* str,
    u16 strLen,
    const font::Font* pFont,
    CommandBufferMode cmdBuffMode
)
{
    Init(allocStrLen, cmdBuffMode);
    SetFont(pFont);
    SetString(str, 0, strLen);
    InitMaterial();
}

TextBox::TextBox(
    const res::TextBox* pBlock,
    const ResBlockSet& resBlockSet,
    CommandBufferMode cmdBuffMode)
:   Pane(pBlock)
{
    u16 allocStrBufLen = static_cast<u16>(pBlock->textBufBytes / sizeof(wchar_t));
    if (allocStrBufLen > 0)
    {
        allocStrBufLen -= 1;
    }

    Init(allocStrBufLen, cmdBuffMode);
#if defined(NW_FONT_ENABLE_EDGE_FONT)
    m_CombinerType = resBlockSet.pResAccessor->GetDefaultFontCombiner();
#endif

    for (int i = 0; i < TEXTCOLOR_MAX; ++i)
    {
        m_TextColors[i] = pBlock->textCols[i];
    }

    m_FontSize = pBlock->fontSize;
    m_TextPosition = pBlock->textPosition;
    m_TextAlignment = pBlock->textAlignment;
    m_CharSpace = pBlock->charSpace;
    m_LineSpace = pBlock->lineSpace;

    // Build font
    NW_NULL_ASSERT(resBlockSet.pFontList);
    NW_ASSERT(pBlock->fontIdx < resBlockSet.pFontList->fontNum);
    const res::Font *const fonts = internal::ConvertOffsToPtr<res::Font>(resBlockSet.pFontList, sizeof(*resBlockSet.pFontList));
    const char *const fontName = internal::ConvertOffsToPtr<char>(fonts, fonts[pBlock->fontIdx].nameStrOffset);

    m_pFont = resBlockSet.pResAccessor->GetFont(fontName);

    // copy the initial string
    if (pBlock->textStrBytes >= sizeof(wchar_t) && m_TextBuf)
    {
        const wchar_t *const pBlockText = internal::ConvertOffsToPtr<wchar_t>(pBlock, pBlock->textStrOffset);
        const u16 resStrLen = static_cast<u16>(pBlock->textStrBytes / sizeof(wchar_t) - 1);
        SetString(pBlockText, 0, resStrLen);
    }

    // Material creation
    {
        NW_NULL_ASSERT(resBlockSet.pMaterialList);
        const u32 *const matOffsTbl = internal::ConvertOffsToPtr<u32>(resBlockSet.pMaterialList, sizeof(*resBlockSet.pMaterialList));
        const res::Material *const pResMaterial = internal::ConvertOffsToPtr<res::Material>(resBlockSet.pMaterialList, matOffsTbl[pBlock->materialIdx]);
        m_pMaterial = Layout::NewObj<Material>(pResMaterial, resBlockSet);
    }
}

void
TextBox::Init(u16 allocStrLen, CommandBufferMode cmdBuffMode)
{
#if defined(NW_FONT_ENABLE_EDGE_FONT)
    m_CombinerType = font::FONT_COMBINER_TYPE_DEFAULT;
#endif

    m_TextBuf = 0;
    m_TextBufBytes = 0;
    m_TextLen = 0;
    m_pFont = 0;
    m_FontSize = Size(0, 0);
    m_TextPosition = internal::SetHorizontalPosition(m_TextPosition, u8(HORIZONTALPOSITION_CENTER));
    m_TextPosition = internal::SetVerticalPosition(m_TextPosition, u8(VERTICALPOSITION_CENTER));
    m_LineSpace = 0;
    m_CharSpace = 0;
    m_pTagProcessor = 0;
    m_pDispStringBuf = 0;
    m_TextAlignment = 0;
    m_IsPTDirty = 0;

    AllocStringBuffer(allocStrLen, cmdBuffMode);
}

void
TextBox::InitMaterial()
{
    // Material creation
    m_pMaterial = Layout::NewObj<Material>();
    if (m_pMaterial)
    {
        m_pMaterial->ReserveMem(0, 0, 0);
    }
}

TextBox::~TextBox()
{
    // OSReport("TextBox::~TextBox()\n");

    if (m_pMaterial && !m_pMaterial->IsUserAllocated())
    {
        Layout::DeleteObj(m_pMaterial);
        m_pMaterial = 0;
    }

    FreeStringBuffer();
}

u8
TextBox::GetMaterialNum() const
{
    return m_pMaterial? 1 : 0;
}

Material*
TextBox::GetMaterial(u32 idx) const
{
    NW_WARNING(idx < GetMaterialNum(), "idx >= GetMaterialNum() : %d >= %d", idx, GetMaterialNum());

    return idx == 0 ? m_pMaterial : 0;
}

void TextBox::SetMaterial(Material* pMaterial)
{
    if (m_pMaterial && !m_pMaterial->IsUserAllocated())
    {
        Layout::DeleteObj(m_pMaterial);
    }
    m_pMaterial = pMaterial;
}

const ut::Color8
TextBox::GetVtxColor(u32 idx) const
{
    NW_ASSERT(idx < VERTEXCOLOR_MAX);

    return GetTextColor(idx / 2);
}

void
TextBox::SetVtxColor(
    u32         idx,
    ut::Color8 value
)
{
    NW_ASSERT(idx < VERTEXCOLOR_MAX);

    SetTextColor(idx / 2, value);
}

u8
TextBox::GetVtxColorElement(u32 idx) const
{
    NW_ASSERT(idx < ANIMTARGET_VERTEXCOLOR_MAX);

    return reinterpret_cast<const u8*>(&m_TextColors[idx / (2 * sizeof(ut::Color8))])[idx % sizeof(ut::Color8)];
}

void
TextBox::SetVtxColorElement(u32 idx, u8 value)
{
    NW_ASSERT(idx < ANIMTARGET_VERTEXCOLOR_MAX);

    u8& elm =
        reinterpret_cast<u8*>(&m_TextColors[idx / (2 * sizeof(ut::Color8))])[idx % sizeof(ut::Color8)];

    elm = value;
}

const ut::Rect
TextBox::GetTextDrawRect() const
{
    if (m_pFont == NULL)
    {
        NW_WARNING(false, "m_pFont is NULL");
        return ut::Rect();
    }

    font::WideTextWriter writer;
    writer.SetCursor(0, 0);
    SetFontInfo(&writer);

    ut::Rect textRect;
    writer.CalcStringRect(&textRect, m_TextBuf, m_TextLen);

    const Size textSize(textRect.GetWidth(), textRect.GetHeight());

    VEC2 ltPos = GetVtxPos();

    const VEC2 curPos  = AdjustTextPos(GetSize(), false);
    const VEC2 textPos = AdjustTextPos(textSize, true);

    ltPos.x += curPos.x - textPos.x;
    ltPos.y -= curPos.y - textPos.y;

    textRect.left   = ltPos.x;
    textRect.top    = ltPos.y;
    textRect.right  = ltPos.x + textSize.width;
    textRect.bottom = ltPos.y - textSize.height;

    return textRect;
}

#ifdef NW_LYT_DMPGL_ENABLED
void
TextBox::DrawSelf(const DrawInfo& drawInfo)
{
    NW_LYT_STOPWATCH_MEASURE(-400, "nw::lyt::TextBox::DrawSelf");

    if (m_TextLen <= 0 || !m_pFont || !m_pMaterial)
    {
        return;
    }

    // Do not allow the lyt render settings to have an affect on the font.
    internal::FinalizeGraphics();

    GraphicsResource& graphicsResource = *drawInfo.GetGraphicsResource();
    graphicsResource.ResetGlState();

    font::TextWriterResource& writerResource = graphicsResource.GetTextWriterResource();
    font::WideTextWriter writer;

    writer.SetTextWriterResource(&writerResource);
    SetupTextWriter(&writer);

    writerResource.ActiveGlProgram();
    NW_GL_ASSERT();

    // Matrix
    LoadMtx(drawInfo);

    ut::Color8 minCol = m_pMaterial->GetColor(INTERPOLATECOLOR_BLACK);
    ut::Color8 maxCol = m_pMaterial->GetColor(INTERPOLATECOLOR_WHITE);

    writer.SetColorMapping(minCol, maxCol);
    writer.SetAlpha(GetGlobalAlpha());

    writer.SetupGX();

    (void)writer.Print(m_TextBuf, m_TextLen);

    // Do not allow the font render settings to have an affect on lyt.
    writer.FinalizeGX();
}
#endif // NW_LYT_DMPGL_ENABLED

u16
TextBox::GetStringBufferLength() const
{
    if (m_TextBufBytes == 0)
    {
        return 0;
    }

    NW_ASSERT(m_TextBufBytes >= sizeof(wchar_t));

    return static_cast<u16>(m_TextBufBytes / sizeof(wchar_t) - 1);
}

void
TextBox::AllocStringBuffer(u16 minLen, CommandBufferMode cmdBuffMode)
{
    // Always recreates when the command buffer mode changes.
    if (m_pDispStringBuf &&
        m_pDispStringBuf->GetCommandBufferMode() != cmdBuffMode)
    {
        FreeStringBuffer();
    }

    if (minLen == 0)
    {
        return;
    }

    u32 allocLen = minLen;
    ++allocLen; // increment by the number of terminating characters

    const u32 textBufBytes = allocLen * sizeof(wchar_t);
    if (textBufBytes >= 0x10000)
    {
        NW_WARNING(false, "minLen is too large. - %d\n", minLen);
    }

    // does nothing if the buffer of the requested length has already been allocated
    if (textBufBytes <= m_TextBufBytes)
    {
        return;
    }

    FreeStringBuffer();

    // Allocate enough command buffer for the specified number of characters.
    const u32 COMMAND_BUFFER_ALIGNMENT = 16;
    u32 drawBufSize = font::CharWriter::GetDispStringBufferSize(minLen);
    u32 cmdBufSize = 0;
    if (cmdBuffMode & font::DispStringBuffer::CBF_DOUBLE_BUFFER)
    {
        drawBufSize = ut::RoundUp(drawBufSize, COMMAND_BUFFER_ALIGNMENT);
        cmdBufSize = font::DispStringBuffer::CalcCommandBufferCapacity(minLen) * sizeof(u32);
    }

    wchar_t* textBuf     = Layout::NewArray<wchar_t>(allocLen);

    void* pDispStringBuf = NULL;
    if (cmdBuffMode & font::DispStringBuffer::CBF_DOUBLE_BUFFER)
    {
        pDispStringBuf = Layout::AllocDeviceMemory(
            drawBufSize + cmdBufSize,
            COMMAND_BUFFER_ALIGNMENT);
    }
    else
    {
        pDispStringBuf = Layout::AllocMemory(drawBufSize);
    }

    if (NULL == textBuf || NULL == pDispStringBuf)
    {
        if (NULL != textBuf)
        {
            Layout::DeletePrimArray(textBuf);
        }
        if (NULL != pDispStringBuf)
        {
            if (cmdBuffMode & font::DispStringBuffer::CBF_DOUBLE_BUFFER)
            {
                Layout::FreeDeviceMemory(pDispStringBuf);
            }
            else
            {
                Layout::FreeMemory(pDispStringBuf);
            }
        }
        return;
    }

    m_TextBuf = textBuf;
    m_TextBufBytes = static_cast<u16>(textBufBytes);
    // Initialize the display string buffer
    m_pDispStringBuf = font::CharWriter::InitDispStringBuffer(
        pDispStringBuf,
        minLen);

    if (cmdBuffMode & font::DispStringBuffer::CBF_DOUBLE_BUFFER)
    {
        void* pSecondBuffer = ut::AddOffsetToPtr(pDispStringBuf, drawBufSize);
        m_pDispStringBuf->SetSecondCommandBuffer(pSecondBuffer);
    }
}

void
TextBox::FreeStringBuffer()
{
    if (m_TextBuf)
    {
        Layout::DeletePrimArray(m_TextBuf);
        m_TextBuf = 0;
        m_TextBufBytes = 0;
        m_TextLen = 0;
    }

    if (m_pDispStringBuf)
    {
        if (m_pDispStringBuf->GetCommandBufferMode() &
            font::DispStringBuffer::CBF_DOUBLE_BUFFER)
        {
            Layout::FreeDeviceMemory(m_pDispStringBuf);
        }
        else
        {
            Layout::FreeMemory(m_pDispStringBuf);
        }

        m_pDispStringBuf = 0;
    }
}

u16
TextBox::SetString(
    const wchar_t* str,
    u16 dstIdx
)
{
    return SetStringImpl(str, dstIdx, std::wcslen(str));
}

u16
TextBox::SetString(
    const wchar_t* str,
    u16 dstIdx,
    u16 strLen
)
{
    return SetStringImpl(str, dstIdx, strLen);
}

u16
TextBox::SetStringImpl(
    const wchar_t*  str,
    u16             dstIdx,
    u32             strLen
)
{
    if (m_pFont == 0)
    {
        NW_WARNING(false, "m_pFont is NULL.\n");
        return 0;
    }

    if (m_TextBuf == 0)
    {
        NW_WARNING(false, "m_TextBuf is NULL.\n");
        return 0;
    }

    const u16 bufLen = GetStringBufferLength();

    if (dstIdx >= bufLen)   // not copied because buffer length exceeded
    {
        NW_WARNING(false, "dstIdx is out of range.\n");
        return 0;
    }

    // Generates gap from the current string
    if (dstIdx > m_TextLen)
    {
        NW_WARNING(false, "dstIdx is out of range.\n");
        return 0;
    }

    // Do a special test here because invalid empty strings can be critical
    if (strLen > 0 && (str == 0 || str[0] == '\0'))
    {
        NW_WARNING(false, "empty string with invalid strLen.\n");
        return 0;
    }

    u32 cpLen = bufLen;
    cpLen -= dstIdx;        // Number of characters that can be copied

    cpLen = ut::Min(strLen, cpLen);
    NW_WARNING(cpLen >= strLen, "%d character(s) droped.\n", strLen - cpLen);

    std::memcpy(m_TextBuf + dstIdx, str, cpLen * sizeof(wchar_t));

    // Sets a 16-bit value no greater than bufLen in m_TextLen.
    m_TextLen = static_cast<u16>(dstIdx + cpLen);
    m_TextBuf[m_TextLen] = 0;

    UpdatePTDirty(true);

    return static_cast<u16>(cpLen);
}

const font::Font*
TextBox::GetFont() const
{
    return m_pFont;
}

void
TextBox::SetFont(const font::Font* pFont)
{
    if (pFont && pFont->GetCharacterCode() != font::CHARACTER_CODE_UNICODE)
    {
        NW_WARNING(false, "pFont is not uincode encoding.");
        return;
    }

    if (UpdatePTDirty(m_pFont != pFont))
    {
        m_pFont = pFont;

        if (m_pFont)
        {
            SetFontSize(Size(static_cast<f32>(m_pFont->GetWidth()), static_cast<f32>(m_pFont->GetHeight())));
        }
        else
        {
            SetFontSize(Size(0.f, 0.f));
        }
    }
}

#ifdef NW_LYT_DMPGL_ENABLED
void
TextBox::LoadMtx(const DrawInfo& drawInfo)
{
    MTX34 mtx;

    GetTextGlobalMtx(&mtx);

    drawInfo.GetGraphicsResource()->GetTextWriterResource().SetViewMtx(mtx);
    NW_GL_ASSERT();
}
#endif

void
TextBox::SetFontInfo(font::WideTextWriter* pWriter) const
{
    pWriter->SetFont(m_pFont);
    if (m_pFont != NULL)
    {
        pWriter->SetFontSize(m_FontSize.width, m_FontSize.height);
        pWriter->SetLineSpace(m_LineSpace);
        pWriter->SetCharSpace(m_CharSpace);
        pWriter->SetWidthLimit(GetSize().width);
    }

    if (m_pTagProcessor)
    {
        pWriter->SetTagProcessor(m_pTagProcessor);
    }
}

void
TextBox::SetTextPos(font::WideTextWriter* pWriter) const
{
    u32 value = 0;

    switch (GetTextAlignment())
    {
    case TEXTALIGNMENT_SYNCHRONOUS:
    default:
        switch (GetTextPositionH())
        {
        case HORIZONTALPOSITION_LEFT:
        default:                        value = font::WideTextWriter::HORIZONTAL_ALIGN_LEFT;   break;
        case HORIZONTALPOSITION_CENTER: value = font::WideTextWriter::HORIZONTAL_ALIGN_CENTER; break;
        case HORIZONTALPOSITION_RIGHT:  value = font::WideTextWriter::HORIZONTAL_ALIGN_RIGHT;  break;
        }
        break;

    case TEXTALIGNMENT_LEFT:        value = font::WideTextWriter::HORIZONTAL_ALIGN_LEFT;    break;
    case TEXTALIGNMENT_CENTER:      value = font::WideTextWriter::HORIZONTAL_ALIGN_CENTER;  break;
    case TEXTALIGNMENT_RIGHT:       value = font::WideTextWriter::HORIZONTAL_ALIGN_RIGHT;   break;
    }

    switch (GetTextPositionH())
    {
    case HORIZONTALPOSITION_LEFT:
    default:
        value |= font::WideTextWriter::HORIZONTAL_ORIGIN_LEFT;
        break;
    case HORIZONTALPOSITION_CENTER:
        value |= font::WideTextWriter::HORIZONTAL_ORIGIN_CENTER;
        break;
    case HORIZONTALPOSITION_RIGHT:
        value |= font::WideTextWriter::HORIZONTAL_ORIGIN_RIGHT;
        break;
    }

    switch (GetTextPositionV())
    {
    case VERTICALPOSITION_TOP:
    default:
        value |= font::WideTextWriter::VERTICAL_ORIGIN_TOP;
        break;
    case VERTICALPOSITION_CENTER:
        value |= font::WideTextWriter::VERTICAL_ORIGIN_MIDDLE;
        break;
    case VERTICALPOSITION_BOTTOM:
        value |= font::WideTextWriter::VERTICAL_ORIGIN_BOTTOM;
        break;
    }

    pWriter->SetDrawFlag(value);
}

VEC2
TextBox::AdjustTextPos(
    const Size&     size,
    bool            isCeil
) const
{
    VEC2 pos;

    switch (GetTextPositionH())
    {
    case HORIZONTALPOSITION_LEFT:
    default:
        pos.x = 0.f;
        break;
    case HORIZONTALPOSITION_CENTER:
        pos.x = internal::AdjustCenterValue(size.width, isCeil);
        break;
    case HORIZONTALPOSITION_RIGHT:
        pos.x = size.width;
        break;
    }

    switch (GetTextPositionV())
    {
    case VERTICALPOSITION_TOP:
    default:
        pos.y = 0.f;
        break;
    case VERTICALPOSITION_CENTER:
        pos.y = internal::AdjustCenterValue(size.height, isCeil);
        break;
    case VERTICALPOSITION_BOTTOM:
        pos.y = size.height;
        break;
    }

    return pos;
}

void
TextBox::GetTextGlobalMtx(nw::math::MTX34* pMtx) const
{
    MTX34Copy(pMtx, &GetGlobalMtx());

    // Uses ViewModel to adjust the text render start position.
    VEC2 pos = GetVtxPos();
    const VEC2 txtPos = AdjustTextPos(GetSize(), false);

    pos.x += txtPos.x;
    pos.y -= txtPos.y;

    pMtx->m[0][3] += pMtx->m[0][0] * pos.x + pMtx->m[0][1] * pos.y;
    pMtx->m[1][3] += pMtx->m[1][0] * pos.x + pMtx->m[1][1] * pos.y;
    pMtx->m[2][3] += pMtx->m[2][0] * pos.x + pMtx->m[2][1] * pos.y;

    // Inverts Y and Z because TextWriter places the positive Y axis below.
    //     Inverse of Y
    pMtx->m[0][1] = - pMtx->m[0][1];
    pMtx->m[1][1] = - pMtx->m[1][1];
    pMtx->m[2][1] = - pMtx->m[2][1];

    //     Inverse of Z
    //     Calculation of Z has been omitted because it has no affect.
    /*
    pMtx->m[0][2] = - pMtx->m[0][2];
    pMtx->m[1][2] = - pMtx->m[1][2];
    pMtx->m[2][2] = - pMtx->m[2][2];
    */
}

#ifdef NW_LYT_DRAWER_ENABLE
void
TextBox::SetupDrawCharData(Drawer* pDrawer)
{
    font::WideTextWriter writer;

    writer.SetDispStringBuffer(m_pDispStringBuf);
    SetupTextWriter(&writer);

    if (this->IsPTDirty())
    {
        writer.StartPrint();
        (void)writer.Print(m_TextBuf, m_TextLen);
        writer.EndPrint();

        m_IsPTDirty = false;
    }

    // After CharWriter::StartPrint, DispStringBuf::IsGeneratedCommand becomes false.
    if (! m_pDispStringBuf->IsGeneratedCommand() && pDrawer)
    {
        if (m_pDispStringBuf->GetCommandBufferMode() &
            font::DispStringBuffer::CBF_DOUBLE_BUFFER)
        {
            m_pDispStringBuf->SwapCurrentCommandBuffer();
        }

        // Creates character string command cache.
    #if !defined(NW_FONT_ENABLE_EDGE_FONT)
        pDrawer->BuildTextCommand(&writer);
    #else
        pDrawer->BuildTextCommand(&writer, m_CombinerType);
    #endif
    }
}
#endif

void
TextBox::SetupTextWriter(font::WideTextWriter* pWriter)
{
    SetFontInfo(pWriter);
    SetTextPos(pWriter);

    ut::Color8 topCol = m_TextColors[TEXTCOLOR_TOP];
    ut::Color8 btmCol = m_TextColors[TEXTCOLOR_BOTTOM];
    pWriter->SetGradationMode(topCol != btmCol ? font::CharWriter::GRADMODE_V: font::CharWriter::GRADMODE_NONE);
    pWriter->SetTextColor(topCol, btmCol);
}

} // namespace nw::lyt
} // namespace nw
