/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#define NW_DEBUG_CHECK_MEMORY_LEAK

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/applet.h>

#include <nw/types.h>

#include <nw/demo.h>
#include <nw/dev.h>
#include <nw/gfx.h>
#include <nw/ut.h>
#include <nw/anim.h>

// !Viewer
// Viewer library header include
#include <nw/viewer.h>

namespace
{

//----------------------------------------
// Memory-related

// Allocator to allocate device memory.
nw::demo::DemoAllocator s_DeviceAllocator;

//----------------------------------------
// Filename definitions.
const wchar_t* SKY_SPHERE_FILE_NAME  = NW_DEMO_FILE_PATH(L"SkySphere.bcmdl");

const wchar_t* MODEL_RESOURCE_FILES[] =
{
    NW_DEMO_FILE_PATH(L"Male.bcmdl"),
    NW_DEMO_FILE_PATH(L"SceneEnvironmentSetting.bcenv"),
    NW_DEMO_FILE_PATH(L"FragmentLight.bcenv"),
};

const wchar_t* SKELETAL_ANIM_RESOURCE_FILE = NW_DEMO_FILE_PATH(L"Walk.bcskla");

//----------------------------------------
// Rendering-related
const s32 RENDER_TARGET_COUNT = 1;
typedef nw::ut::FixedSizeArray<nw::gfx::IRenderTarget*, RENDER_TARGET_COUNT> RenderTargetArray;

RenderTargetArray s_RenderTargets;
nw::demo::SceneSystem*  s_SceneSystem = NULL;
nw::demo::RenderSystem* s_RenderSystem = NULL;

nw::demo::GraphicsDrawing  s_GraphicsDrawing;

//----------------------------------------
// Resource-related
nw::demo::ResourceArray s_Resources;

//----------------------------------------
// Scene-related
const int SCENE_NODE_COUNT = 4;
nw::gfx::SceneNode* s_SceneRoot = NULL;
nw::gfx::SceneNode* s_ModelRoot = NULL;
s32 s_FrameCount = 0;
nw::gfx::Camera* s_BaseCamera = NULL;
nw::gfx::Camera* s_LeftCamera = NULL;
nw::gfx::Camera* s_RightCamera = NULL;
const f32 s_fNearPlane = 0.1f;

//----------------------------------------
// Scene environment-related
const s32 ENVIRONMENT_SETTINGS_COUNT = 1;

typedef nw::ut::FixedSizeArray<nw::gfx::SceneEnvironmentSetting*, ENVIRONMENT_SETTINGS_COUNT> SceneEnvironmentSettingArray;
SceneEnvironmentSettingArray s_SceneEnvironmentSettings;

const s32 s_BaseCameraIndex = 0;

//----------------------------------------
// Animation-related
nw::gfx::SkeletalModel* s_AnimModel = NULL;
nw::gfx::TransformAnimEvaluator* s_AnimEvaluator = NULL;

float s_StepFrame = 1.0f;

//----------------------------------------
// !Viewer
// Viewer-related
nw::gfx::SceneTraverser* s_SceneTraverser = NULL;

// !Viewer
// Viewer input mode flag
static bool sViewerMode = false;

// !Viewer
// Flag for whether to perform pad operations with Viewer input mode
#define MOVE_CAMERA_BY_VIEWER_PAD

/*
  
 */
void
InitializeGraphics()
{
    // Perform stereo settings to renderDescription.
    nw::demo::RenderSystem::Description renderDescription;

    renderDescription.reusableCommandBufferSize = 0x100000;
    renderDescription.reusableCommandRequestCount = 512;
    renderDescription.upperScreenMode = nw::demo::UPPER_SCREEN_MODE_STEREO;

    s_RenderSystem = nw::demo::RenderSystem::Create(&s_DeviceAllocator, renderDescription);

    s_GraphicsDrawing.SetScreenSize(
        renderDescription.lowerScreenDescription.width,
        renderDescription.lowerScreenDescription.height
    );

    nw::demo::Utility::InitializeGraphicsDrawing(&s_DeviceAllocator, s_GraphicsDrawing);

    s_RenderTargets.push_back(
        nw::demo::Utility::CreateUpperScreenBuffer(&s_DeviceAllocator, renderDescription)
    );
    NW_ASSERT(!s_RenderTargets.empty());
    s_RenderSystem->GetRenderContext()->SetRenderTarget(s_RenderTargets.front());

    // Standard settings to sceneDescription are performed with the constructor.
    nw::demo::SceneSystem::Description sceneDescription;
    sceneDescription.isFixedSizeMemory = true;
    s_SceneSystem = nw::demo::SceneSystem::Create(&s_DeviceAllocator, sceneDescription);

    // Sets the furthest background model for the demo to the rendering system.
    // With the gfx demo, glClear, etc. is not used, and a fill is applied with the background.
    s_RenderSystem->LoadSkyModel(SKY_SPHERE_FILE_NAME);

    // !Viewer
    // Initialization of the Traverser for Viewer resource rendering
    s_SceneTraverser = nw::gfx::SceneTraverser::Builder().Create(&s_DeviceAllocator);

    NW_GL_ASSERT();
}

/*
  
 */
void
TerminateGraphics()
{
    nw::gfx::SafeDestroy(s_LeftCamera);

    nw::gfx::SafeDestroy(s_RightCamera);

    nw::gfx::SafeDestroy(s_SceneSystem);

    nw::gfx::SafeDestroyAll(s_RenderTargets);

    s_GraphicsDrawing.Finalize();

    nw::gfx::SafeDestroy(s_RenderSystem);

    // !Viewer
    // Release of Traverser for Viewer resource rendering
    nw::gfx::SafeDestroy(s_SceneTraverser);

    NW_GL_ASSERT();
}

/*
  
  
  
  
  
  
 


*/
nw::gfx::TransformAnimEvaluator*
CreateTransformAnimEvaluator(
    const int maxMembers,
    const bool translateAnimEnabled,
    const wchar_t* filePath
)
{
    // Generate an instance of animation data from a file.
    nw::demo::ResourceSet* resourceSet = nw::demo::Utility::LoadResources(s_Resources, filePath, &s_DeviceAllocator);
    if (resourceSet->resource.GetSkeletalAnimsCount() == 0)
    {
        // There is no animation data for skeletal animations.
        return NULL;
    }
    nw::anim::ResAnim resAnim = resourceSet->resource.GetSkeletalAnims(0);

    if (!resAnim.IsValid())
    {
        return NULL;
    }

    // Generate AnimEvaluator using AnimEvaluator::Builder from resAnim.
    // For skeletal animations, use TransformAnimEvaluator::Builder.
    // 
    // When an animation is application to only one model, and for frame-format data, setting AllocCache to FALSE will lower the processing load.
    // 
    nw::gfx::TransformAnimEvaluator* evaluator = nw::gfx::TransformAnimEvaluator::Builder()
        .AnimData(resAnim)
        .MaxMembers(maxMembers)
        .MaxAnimMembers(resAnim.GetMemberAnimSetCount())
        .AllocCache(false)
        .Create(&s_DeviceAllocator);

    // Set the invalidation flag for movement animation.
    evaluator->SetIsTranslateDisabled(!translateAnimEnabled);

    return evaluator;
}

/*
  
                
                
  
  
 


*/
nw::gfx::TransformAnimEvaluator*
InitializeSkeletalAnim(nw::gfx::SkeletalModel* model)
{
    // Get the animation group from the model.
    //
    // The animation group maintains a pointer to the animation target member.
    // The target member refers to elements that can animated (for example, Diffuse and Texture, etc., for a material animation) within a model.
    //
    // Here we get the animation group that includes the bone's transform from the skeletal model.
    // 
    //
    // Used functions differ according to the animation targets and types.
    // - SkeletalModel::GetSkeletalAnimGroup
    // - Model::GetVisibilityAnimGroup
    // - Model::GetMaterialAnimGroup
    // - Light::GetAnimGroup
    // - Camera::GetAnimGroup
    nw::gfx::AnimGroup* animGroup = model->GetSkeletalAnimGroup();
    if (animGroup == NULL) // There is no animation group for skeletal animations.
    {
        return NULL;
    }

    nw::gfx::ResSkeletalModel resModel = model->GetResSkeletalModel();
    nw::gfx::ResSkeleton resSkeleton = resModel.GetSkeleton();
    const int maxBones = resSkeleton.GetBonesCount();

    // When models with different body types share animations, you must disable the animation's translation component.
    // 
    // In this demo, according to the flag obtained from ResSkeleton, TransformAnimEvaluator's movement animation is invalidated, but in the current version there is no means available to set a flag to ResSkeleton in advance.
    // 
    // 
    // In addition, there is no feature that automatically disables movement animation when the flag is enabled.
    // We plan to add support for this feature in future.
    const bool translateAnimEnabled =
        nw::ut::CheckFlag(resSkeleton.GetFlags(), nw::gfx::ResSkeletonData::FLAG_TRANSLATE_ANIMATION_ENABLED);

    nw::gfx::TransformAnimEvaluator* evaluator = CreateTransformAnimEvaluator(
        maxBones, translateAnimEnabled, SKELETAL_ANIM_RESOURCE_FILE);
    if (evaluator == NULL)
    {
        return NULL;
    }

    // The animation group is bound to AnimEvaluator/
    // By doing so, animation data is associated with the animation target member.
    bool bindResult = evaluator->Bind(animGroup);

    // Set AnimEvaluator to the model.
    // AnimEvaluator can be set multiple times for one model. When doing so, specify an objectIndex for each AnimEvaluator.
    // 
    // For details, see PartialAnimationDemo.
    // 
    // Used functions differ according to the animation targets and types.
    // SkeletalModel::SetSkeletalAnimObject
    // Model::SetVisibilityAnimObject
    // Model::SetMaterialAnimObject
    // Light::SetAnimObject
    // Camera::SetAnimObject
    model->SetSkeletalAnimObject(evaluator);

    return evaluator;
}

/*
  
                
  
 

*/
void
ChangeSpeed(nw::gfx::SkeletalModel* model)
{
    NW_NULL_ASSERT(model);

    nw::gfx::AnimObject* animObject = model->GetSkeletalAnimObject();
    nw::gfx::TransformAnimEvaluator* evaluator =
        nw::ut::DynamicCast<nw::gfx::TransformAnimEvaluator*>(animObject);
    if (evaluator != NULL)
    {
        nw::demo::Pad* pad = nw::demo::PadFactory::GetPad();
        if (pad->IsButtonPress(nw::demo::Pad::BUTTON_LEFT))
        {
            s_StepFrame -= 0.02f;
        }
        if (pad->IsButtonPress(nw::demo::Pad::BUTTON_RIGHT))
        {
            s_StepFrame += 0.02f;
        }
        s_StepFrame = nw::ut::Clamp(s_StepFrame, 0.0f, 50.0f);

        // Change the animation's update frame.
        evaluator->SetStepFrame(s_StepFrame);
    }
}

/*
  
 */
void
BuildRootNodes()
{
    NW_ASSERT(s_SceneRoot == NULL);
    s_SceneRoot = nw::gfx::TransformNode::DynamicBuilder()
        .Create(&s_DeviceAllocator);
    NW_NULL_ASSERT(s_SceneRoot);

    NW_ASSERT(s_ModelRoot == NULL);
    s_ModelRoot = nw::gfx::TransformNode::DynamicBuilder()
        .Create(&s_DeviceAllocator);
    s_SceneRoot->AttachChild(s_ModelRoot);
    NW_NULL_ASSERT(s_ModelRoot);
}

/*
  
 */
void
BuildCameras()
{
    nw::demo::Utility::CreateStereoCameras(
        &s_BaseCamera,
        &s_LeftCamera,
        &s_RightCamera,
        &s_DeviceAllocator,
        nw::math::VEC3(20.0f, 15.0f, 20.0f),
        nw::math::VEC3(0.0f, 10.0f, 0.0f),
        s_fNearPlane
    );

    s_SceneRoot->AttachChild(s_BaseCamera);
    s_SceneSystem->GetCameraController()->Register(s_BaseCamera);
}


/*
  
 */
void
BuildResources(nw::demo::ResourceSet* resourceSet)
{
    nw::gfx::ResGraphicsFile resFile = resourceSet->resource;
    void* imageData = resFile.GetImageBlockData();

    // The texture and vertex buffers are manually transferred to VRAM.
    if (imageData)
    {
        s32 size = resFile.GetImageBlockDataSize();

        void* vramAddr = nw::demo::AllocateGraphicsMemory(NN_GX_MEM_VRAMA, NN_GX_MEM_TEXTURE, 0, size);

        // The data cache is flushed and transferred to VRAM.
        // Flushing of the cache occurs via nngxAddVramDmaCommand.
        nngxAddVramDmaCommand( imageData, vramAddr, size );

        // Set transferred image data address in resources prior to setup.
        nw::gfx::TransferedVramAddressSetter locationSetter( imageData, vramAddr );
        resFile.ForeachTexture(locationSetter);
        resFile.ForeachIndexStream(locationSetter);
        resFile.ForeachVertexStream(locationSetter);
    }

    nw::gfx::Result result = resFile.Setup(&s_DeviceAllocator);
    if (result.IsFailure())
    {
        NW_FATAL_ERROR("Fail to set up model. A result code is 0x%x", result.GetCode());
    }

    nw::ut::MoveArray<nw::gfx::SceneNode*> sceneNodeArray(SCENE_NODE_COUNT, &s_DeviceAllocator);

    // Generate an instance of the skeletal model.
    nw::gfx::ResModelArray models = resFile.GetModels();
    nw::gfx::ResModelArray::iterator modelsEnd = models.end();
    for (nw::gfx::ResModelArray::iterator modelResource = models.begin();
         modelResource != modelsEnd; ++modelResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*modelResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);

        s_AnimModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);
    }

    nw::gfx::ResLightArray lights = resFile.GetLights();
    nw::gfx::ResLightArray::iterator lightsEnd = lights.end();
    for (nw::gfx::ResLightArray::iterator lightResource = lights.begin();
         lightResource != lightsEnd; ++lightResource)
    {
        nw::gfx::SceneNode* node = nw::demo::Utility::CreateSceneNode(
            &s_DeviceAllocator,
            (*lightResource)
        );
        NW_NULL_ASSERT(node);
        sceneNodeArray.push_back(node);
    }

    // Resolve parent-child reference relationships
    nw::gfx::SceneHelper::ResolveReference(sceneNodeArray);

    // Add a model to the scene
    nw::gfx::SceneHelper::ForeachRootNodes(
        sceneNodeArray.Begin(),
        sceneNodeArray.End(),
        nw::gfx::AttachNode(s_ModelRoot)
    );

    nw::gfx::ResSceneEnvironmentSettingArray settings = resFile.GetSceneEnvironmentSettings();
    nw::gfx::ResSceneEnvironmentSettingArray::iterator settingsEnd = settings.end();
    for (nw::gfx::ResSceneEnvironmentSettingArray::iterator settingResource = settings.begin();
        settingResource != settingsEnd; ++settingResource)
    {
        nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
            .Resource(*settingResource)
            .CreateObject(&s_DeviceAllocator, &s_DeviceAllocator);

        nw::gfx::SceneEnvironmentSetting* sceneEnvironmentSetting =
            nw::ut::DynamicCast<nw::gfx::SceneEnvironmentSetting*>(sceneObject);

        NW_NULL_ASSERT(sceneEnvironmentSetting);
        s_SceneEnvironmentSettings.push_back(sceneEnvironmentSetting);
    }
}

/*
  
                
 
*/
void
InitializeScenes()
{
    s_StepFrame = 1.0f;

    BuildRootNodes();

    BuildCameras();

    NW_FOREACH(const wchar_t* name, MODEL_RESOURCE_FILES)
    {
        BuildResources(nw::demo::Utility::LoadResources(s_Resources, name, &s_DeviceAllocator));
    }

    // Initialize a skeletal animation.
    s_AnimEvaluator = InitializeSkeletalAnim(s_AnimModel);
    NW_NULL_ASSERT(s_AnimEvaluator);

    s_SceneSystem->InitializeScene(s_SceneRoot);
    s_SceneSystem->UpdateScene();

    s_RenderSystem->SetSceneEnvironmentSettings(s_SceneSystem, &s_SceneEnvironmentSettings);

    nw::gfx::SceneEnvironment& sceneEnvironment = s_RenderSystem->GetSceneEnvironment();
    sceneEnvironment.SetCamera(s_BaseCameraIndex, s_BaseCamera);
    nw::demo::Utility::SetCameraAspectRatio(s_BaseCamera, s_RenderTargets[0]);

    NW_GL_ASSERT();

    s_FrameCount = 0;
}

/*
  
 */
void
TerminateScenes()
{
    nw::gfx::SafeDestroyBranch(s_SceneRoot);
    nw::demo::SafeCleanupResources(s_Resources);
    nw::ut::SafeDestroyAll(s_SceneEnvironmentSettings);

    nw::ut::SafeDestroy(s_AnimEvaluator);

    NW_GL_ASSERT();

    s_Resources.clear();
    s_SceneEnvironmentSettings.clear();

    s_ModelRoot = NULL;
}

/*
  
 */
void
UpdateScene()
{
    // !Viewer
    // In the built-in Viewer operation mode, application operation is not performed.
    if( !sViewerMode )
    {
        ChangeSpeed(s_AnimModel);

    #ifndef MOVE_CAMERA_BY_VIEWER_PAD
        s_SceneSystem->GetCameraController()->Update();
    #endif
    }

    // SceneUpdater::UpdateAll is called in SceneSystem::UpdateScene.
    // By performing UpdateAll, the AnimEvaluator bound to the model is evaluated and the animation is updated at each frame.
    // 
    s_SceneSystem->UpdateScene();

    s_BaseCamera->UpdateCameraMatrix();
    s_RenderSystem->CalcStereoCamera(s_LeftCamera, s_RightCamera, s_BaseCamera, s_fNearPlane + 5.0f);

    ++s_FrameCount;
}

/*
  
 */
void
ReportDemo()
{
    NW_PROFILE("ReportDemo");

    // This load is exempted from the load display.
    s_RenderSystem->SuspendLoadMeter();

    nw::demo::DebugUtility::CalcLoadMeter(s_RenderSystem);

    s_GraphicsDrawing.BeginDrawingShape();

    nw::demo::DebugUtility::DrawLoadMeter(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    s_GraphicsDrawing.EndDrawingShape();

    s_GraphicsDrawing.BeginDrawingString();

    nw::demo::DebugUtility::DrawLoadMeterText(
        s_RenderSystem,
        &s_GraphicsDrawing
    );

    const int dumpPositionX = 10;
    const int dumpPositionY = 10;
    s_GraphicsDrawing.DrawString(
        dumpPositionX,
        dumpPositionY,
        "step frame: %0.2f",
        s_StepFrame
    );

    s_GraphicsDrawing.EndDrawingString();

    s_RenderSystem->ResumeLoadMeter();
}

/*
  
 */
// !Viewer
// Function to build SceneContext that includes Viewer resources
void
TraverseScene()
{
    // Gets the interface to perform resource access
    nw::viewer::g3d::Module3D& module3D =
        nw::viewer::sys::App::GetInstance().GetModule3D();
    nw::viewer::ef::ModuleEF& moduleEF =
        nw::viewer::sys::App::GetInstance().GetModuleEF();

    // Rebuilds scene context
    s_SceneTraverser->Begin(s_SceneSystem->GetSceneContext());
    {
        // Adds application-side resources to the scene context
        s_SceneRoot->Accept(s_SceneTraverser);

        // Scene update that includes application-side animation
        // Executes scene update at the state when only the application-side resources are included in the scene context
        // 
        UpdateScene();

#if 1 // Resource access via index

        // Adds Viewer-side model resources to the scene context
        u32 modelNum = module3D.GetModelNum();
        for ( int i = 0; i < modelNum; ++i ) {
            nw::viewer::g3d::ModelIter iter = module3D.GetModel( i );
            if ( iter.IsValid() ) { iter.GetModel()->Accept(s_SceneTraverser); }
        }

        // Adds Viewer-side particle model resources to the scene context
        u32 pModelNum = moduleEF.GetParticleModelNum();
        for ( int i = 0; i < pModelNum; ++i ) {
            nw::viewer::ef::ParticleModelIter iter = moduleEF.GetParticleModel( i );
            if ( iter.IsValid() ) { iter.GetParticleModel()->Accept(s_SceneTraverser); }
        }

        // Adds Viewer-side particle emitter resources to the scene context
        u32 pEmitterNum = moduleEF.GetParticleEmitterNum();
        for ( int i = 0; i < pEmitterNum; ++i ) {
            nw::viewer::ef::ParticleEmitterIter iter = moduleEF.GetParticleEmitter( i );
            if ( iter.IsValid() ) {
                nw::gfx::ParticleEmitter* emitter = iter.GetParticleEmitter();
                // Determines if the parent node is a particle model
                nw::gfx::ParticleModel* parent =
                    nw::ut::DynamicCast<nw::gfx::ParticleModel*>( emitter->GetParent() );
                // Adds emitter only when the parent node is not a particle emitter
                // When a particle model is the parent, it is already added by recursive calls of the parent node Accept
                if ( parent == NULL ) {
                    emitter->Accept(s_SceneTraverser);
                }
            }
        }

#else // Resource access via iterator increment

        // Adds Viewer-side model resources to the scene context
        for (nw::viewer::g3d::ModelIter iter = module3D.GetModel(); iter.IsValid(); ++iter ) {
            iter.GetModel()->Accept(s_SceneTraverser);
        }

        // Adds Viewer-side particle model resources to the scene context
        for (nw::viewer::ef::ParticleModelIter iter = moduleEF.GetParticleModel(); iter.IsValid(); ++iter) {
            iter.GetParticleModel()->Accept(s_SceneTraverser);
        }

        // Adds Viewer-side particle emitter resources to the scene context
        for ( nw::viewer::ef::ParticleEmitterIter iter = moduleEF.GetParticleEmitter(); iter.IsValid(); ++iter ) {
            nw::gfx::ParticleEmitter* emitter = iter.GetParticleEmitter();
            // Determines if the parent node is a particle model
            nw::gfx::ParticleModel* parent =
                nw::ut::DynamicCast<nw::gfx::ParticleModel*>( emitter->GetParent() );
            // Adds emitter only when the parent node is not a particle emitter
            // When a particle model is the parent, it is already added by recursive calls of the parent node Accept
            if ( parent == NULL ) {
                emitter->Accept(s_SceneTraverser);
            }
        }
#endif
    }
    s_SceneTraverser->End();

    // Sorts in order of dependency on the particle set.
    std::sort(
        s_SceneSystem->GetSceneContext()->GetParticleSetBegin(),
        s_SceneSystem->GetSceneContext()->GetParticleSetEnd(),
        nw::gfx::ParticleSetCompare());
}


//----------------------------------------
struct RenderSceneInternalFunctor
    : public std::unary_function<nw::gfx::RenderElement&, void>
{
    nw::gfx::RenderContext* m_RenderContext;
    nw::gfx::MeshRenderer* m_MeshRenderer;

    RenderSceneInternalFunctor(nw::gfx::RenderContext* renderContext, nw::gfx::MeshRenderer* meshRenderer)
    : m_RenderContext(renderContext), m_MeshRenderer(meshRenderer)
    {
        NW_NULL_ASSERT(renderContext);
        NW_NULL_ASSERT(meshRenderer);
    }

    void operator()(nw::gfx::RenderElement& element)
    {
        if (element.IsCommand())
        {
            nw::gfx::RenderCommand* command = element.GetCommand();
            NW_NULL_ASSERT(command);
            command->Invoke(this->m_RenderContext);
        }
        else
        {
            nw::gfx::ResMesh mesh = element.GetMesh();
            nw::gfx::Model* model = element.GetModel();
            model->PreRenderSignal()(model, mesh, this->m_RenderContext);
            this->m_MeshRenderer->RenderMesh(mesh, model);
            model->PostRenderSignal()(model, mesh, this->m_RenderContext);
        }
        NW_GL_ASSERT();
    }
};


/*
  
 */
// !Viewer
// Function to render stereo scenes
void
RenderStereoScene(bool viewerModeFlag)
{
    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    //----------------------------------------
    // Create a command for re-use and save it.
    //
    nw::demo::CommandListSwapper* commandListSwapper = s_RenderSystem->GetCommandListSwapper();

    // Render the scene based on the base camera.
    commandListSwapper->StartCommandSave();

    s_RenderSystem->ClearBySkyModel(renderContext->GetActiveCamera());

    nw::gfx::RenderQueue* renderQueue = s_RenderSystem->GetRenderQueue();
    nw::gfx::MeshRenderer* meshRenderer = s_RenderSystem->GetMeshRenderer();

    std::for_each(
        renderQueue->Begin(),
        renderQueue->End(),
        RenderSceneInternalFunctor(renderContext, meshRenderer));

    commandListSwapper->EndCommandSave();

    //----------------------------------------
    // Re-use the commands saved for the left eye and the right eye.

    // Start GPU process time measurement
    int profilerLeft = commandListSwapper->AddGpuProfilingStartPoint(true);
    NW_ASSERT(profilerLeft >= 0);

    // Left eye use
    renderContext->SetCameraMatrix(s_LeftCamera);
    commandListSwapper->ReuseCommand(false);

    // !Viewer
    // Built-in Viewer 3D information display
    if( viewerModeFlag )
    {
        nw::viewer::sys::App::GetInstance().GetModule3D().Draw3DInfo( s_LeftCamera, s_BaseCamera, nw::viewer::g3d::RenderAll);
    }

    // End GPU process time measurement
    commandListSwapper->SetGpuProfilingEndPoint(profilerLeft);

    s_RenderSystem->TransferBuffer(nw::demo::UPPER_SCREEN);
    NW_GL_ASSERT();

    // !Viewer
    // Resets the state.
    renderContext->ResetState();
    renderContext->SetActiveCamera(s_BaseCameraIndex);

    // Start GPU process time measurement
    int profilerRight = commandListSwapper->AddGpuProfilingStartPoint(true);
    NW_ASSERT(profilerRight >= 0);

    // Right eye use
    renderContext->SetCameraMatrix(s_RightCamera);
    commandListSwapper->ReuseCommand(false);

    // !Viewer
    // Built-in Viewer 3D information display
    if( viewerModeFlag )
    {
        nw::viewer::sys::App::GetInstance().GetModule3D().Draw3DInfo( s_RightCamera, s_BaseCamera, nw::viewer::g3d::RenderAll);
    }

    // End GPU process time measurement
    commandListSwapper->SetGpuProfilingEndPoint(profilerRight);

    s_RenderSystem->TransferBuffer(nw::demo::EXTENSION_SCREEN);
    NW_GL_ASSERT();

    // !Viewer
    // Resets the state.
    renderContext->ResetState();
    renderContext->SetActiveCamera(s_BaseCameraIndex);
    renderContext->SetCameraMatrix(s_RightCamera);
}
/*
  
 */
void
DemoScene()
{
    NW_ASSERT(!s_RenderTargets.empty());

    nw::gfx::RenderContext* renderContext = s_RenderSystem->GetRenderContext();

    InitializeScenes();

    nw::demo::DebugUtility::PostInitializeScenes();

    bool isContinuing = true;

    // !Viewer
    // Built-in Viewer application interface
    nw::viewer::sys::App& app = nw::viewer::sys::App::GetInstance();

#ifdef MOVE_CAMERA_BY_VIEWER_PAD
    // !Viewer
    // Gets the interface to perform resource access
    nw::viewer::g3d::Module3D& module3D =
        nw::viewer::sys::App::GetInstance().GetModule3D();

    // !Viewer
    // Add camera
    module3D.AddCamera(s_BaseCamera);
#endif

    while ( isContinuing )
    {
        nw::demo::DebugUtility::AdvanceAutoTestFrame();

        nw::demo::PadFactory::GetPad()->Update();

        // !Viewer
        // Switches Viewer and application operations with Pad L input
        nw::demo::Pad* pad = nw::demo::PadFactory::GetPad();
        if ( pad->IsButtonDown(nw::demo::Pad::BUTTON_L) )
        {
            sViewerMode = !sViewerMode;
        }

        // !Viewer
        // Communications processing of the embedded viewer.
        app.PollStream();

        // !Viewer
        // Built-in Viewer scene update
        app.GetModule3D().Calc();

        if( sViewerMode )
        {
            // !Viewer
            // Built-in Viewer operation processing
            app.Main();

            // !Viewer
            // Reset the Pad so that application operations are not performed while the Viewer is operating.
            nw::demo::PadFactory::GetPad()->Reset();
            nw::demo::PadFactory::GetPad()->Update();
        }

        // !Viewer
        // Rebuild SceneContext
        TraverseScene();

        renderContext->SetActiveCamera(s_BaseCameraIndex);
        s_RenderSystem->SubmitView(s_SceneSystem);

        s_RenderSystem->SetRenderTarget(s_RenderTargets[0]);

        // !Viewer
        // Viewer rendering start processing
        app.BeginDraw();

        // !Viewer
        // Stereo scene rendering
        RenderStereoScene( sViewerMode );

        // !Viewer
        // Switch the lower screen display according to the mode.
        if( sViewerMode )
        {
            // !Viewer
            // Display Viewer lower screen.
            s_RenderSystem->ClearBuffer(nw::ut::FloatColor(0.5f, 0.5f, 0.5f, 1.0f));
            app.GetVideoSetting().SetScreenSize(320, 240);
            app.DrawMenu();
            app.GetVideoSetting().SetScreenSize(400, 240);
        }
        else
        {
            s_RenderSystem->ClearBySkyModel(s_BaseCamera);
            ReportDemo();
        }

        // !Viewer
        // Viewer rendering finalization
        app.EndDraw();

        s_RenderSystem->TransferBuffer(nw::demo::LOWER_SCREEN);

        s_RenderSystem->PresentBuffer(nw::demo::UPPER_SCREEN | nw::demo::LOWER_SCREEN | nw::demo::EXTENSION_SCREEN);

        renderContext->ResetState();

        if (nw::demo::Utility::IsTerminating())
        {
            isContinuing = false;
        }
    }

    nw::demo::DebugUtility::PreTerminateScenes();

    TerminateScenes();
}

} // Namespace

/*
  
 */
void
nnMain()
{
    nn::applet::Enable();

    nw::demo::InitializeGraphicsSystem(&s_DeviceAllocator);

    nw::demo::PadFactory::Initialize(&s_DeviceAllocator);

    void* HioWorkMemory = s_DeviceAllocator.Alloc(nn::hio::CTR::WORKMEMORY_SIZE);
    nn::hio::Initialize(HioWorkMemory);

    NW_DEMO_TEST_LOOP(&s_DeviceAllocator, NULL, &s_RenderSystem)
    {
        InitializeGraphics();

        // !Viewer
        // Viewer library initialization
        nw::viewer::sys::App::GetInstance().Init( &s_DeviceAllocator, &s_DeviceAllocator );
        nw::viewer::sys::App::GetInstance().ViewerMenuInit();
        nw::viewer::sys::App::GetInstance().ViewerGraphicsInit();

        DemoScene();

        // !Viewer
        // Viewer library release
        nw::viewer::sys::App::GetInstance().Finalize();

        TerminateGraphics();
    }

    nn::hio::Finalize();
    s_DeviceAllocator.Free(HioWorkMemory);

    nw::demo::PadFactory::Finalize();

    nw::demo::FinalizeGraphicsSystem();
}
