/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_VIEWER_COMMON_COMMON_H_
#define NW4R_VIEWER_COMMON_COMMON_H_

#include <nn/types.h>

#define     GET_U32(x)                              (reinterpret_cast<u32>(x))
#define     GET_PTR(x)                              (reinterpret_cast<void*>(x))
#define     ADD_U32_TO_PTR(addr,offset)             GET_PTR(GET_U32(addr)+(offset))
#define     GET_OFFSET_FROM_PTR(begin, end)         (GET_U32(end)-GET_U32(begin))


namespace vw     {
namespace common {

u32 strlcpy( char* dst, const char* src, u32 bufSize );
u32 strlcat( char* dst, const char* src, u32 bufSize );
s32 strcmpi( const char* str1, const char* str2 );
s32 strncmpi( const char* str1, const char* str2, u32 num );


u16     GetAngle( f32 sin, f32 cos );


namespace LEndian
{

/*---------------------------------------------------------------------------*
  Name:         ReadU16
                ReadU32
                WriteU16
                WriteU32

  Description:  Reverses endianness and accesses memory.

  Arguments:    addr

  Returns:      None.
 *---------------------------------------------------------------------------*/
inline u16
ReadU16( const void* addr )
{
    const u8* p = reinterpret_cast<const u8*>( addr );
    u16 val = static_cast<u16>( (*(p + 1) << 8) + *p );
    return val;
}

inline u32
ReadU24( const void* addr )
{
    const u8* p = reinterpret_cast< const u8*>( addr );
    u32 val = static_cast<u32>( (*(p + 2) << 16) + (*(p + 1) << 8) + *p );
    return val;
}

inline u32
ReadU32( const void* addr )
{
    const u8* p   = reinterpret_cast<const u8*>( addr );
    u32 val = static_cast<u32>( (*(p + 3) << 24) + (*(p + 2) << 16) + (*(p + 1) << 8) + *p );
    return val;
}

inline u32
ReadUBytes( const void* addr, u8 bytes )
{
    switch ( bytes )
    {
    case 4:  return ReadU32( addr );
    case 3:  return ReadU24( addr );
    case 2:  return ReadU16( addr );
    default: return * reinterpret_cast<const u8*>( addr );
    }
}

inline void
WriteU16( void* addr, u16 data )
{
    u8* p   = reinterpret_cast<u8*>( addr );

    *(p + 1) = static_cast<u8>( data >> 8 );
    *p       = static_cast<u8>( data & 0xFF );
}

inline void
WriteU24( void* addr, u32 data )
{
    u8* p   = reinterpret_cast<u8*>( addr );

    *(p + 2) = static_cast<u8>( (data >> 16) & 0xFF );
    *(p + 1) = static_cast<u8>( (data >> 8 ) & 0xFF );
    *p       = static_cast<u8>( data & 0xFF );
}

inline void
WriteU32( void* addr, u32 data )
{
    u8* p   = reinterpret_cast<u8*>( addr );

    *(p + 3) = static_cast<u8>( (data >> 24) & 0xFF );
    *(p + 2) = static_cast<u8>( (data >> 16) & 0xFF );
    *(p + 1) = static_cast<u8>( (data >> 8 ) & 0xFF );
    *p       = static_cast<u8>( data & 0xFF );
}


inline s16
ReadS16( const void* addr )
{
    return static_cast<s16>( ReadU16( addr ) );
}

inline s32
ReadS32( const void* addr )
{
    return static_cast<s32>( ReadU32( addr ) );
}

inline void
WriteS16( void* addr, s16 data )
{
    WriteU16( addr, u16( data ) );
}

inline void
WriteS32( void* addr, s32 data )
{
    WriteU32( addr, u32( data ) );
}

}   /* namespace LEndian */


namespace BEndian
{

/*---------------------------------------------------------------------------*
  Name:         ReadU16
                ReadU32
                WriteU16
                WriteU32

  Description:  Reverses endianness and accesses memory.

  Arguments:    addr

  Returns:      None.
 *---------------------------------------------------------------------------*/
inline u16
ReadU16( const void* addr )
{
    const u8* p = reinterpret_cast<const u8*>( addr );
    u16 val = static_cast<u16>( (*p << 8) + *(p + 1) );
    return val;
}

inline u32
ReadU24( const void* addr )
{
    const u8* p = reinterpret_cast< const u8*>( addr );
    u32 val = static_cast<u32>( (*p << 16) + (*(p + 1) << 8) + *(p + 2) );
    return val;
}

inline u32
ReadU32( const void* addr )
{
    const u8* p   = reinterpret_cast<const u8*>( addr );
    u32 val = static_cast<u32>( (*p << 24) + (*(p + 1) << 16) + (*(p + 2) << 8) + *(p + 3) );
    return val;
}

inline u32
ReadUBytes( const void* addr, u8 bytes )
{
    switch ( bytes )
    {
    case 4:  return ReadU32( addr );
    case 3:  return ReadU24( addr );
    case 2:  return ReadU16( addr );
    default: return * reinterpret_cast<const u8*>( addr );
    }
}

inline void
WriteU16( void* addr, u16 data )
{
    u8* p   = reinterpret_cast<u8*>( addr );

    *p       = static_cast<u8>( data >> 8 );
    *(p + 1) = static_cast<u8>( data & 0xFF );
}

inline void
WriteU24( void* addr, u32 data )
{
    u8* p   = reinterpret_cast<u8*>( addr );

    *p       = static_cast<u8>( (data >> 16) & 0xFF );
    *(p + 1) = static_cast<u8>( (data >> 8 ) & 0xFF );
    *(p + 2) = static_cast<u8>( data & 0xFF );
}

inline void
WriteU32( void* addr, u32 data )
{
    u8* p   = reinterpret_cast<u8*>( addr );

    *p       = static_cast<u8>( (data >> 24) & 0xFF );
    *(p + 1) = static_cast<u8>( (data >> 16) & 0xFF );
    *(p + 2) = static_cast<u8>( (data >> 8 ) & 0xFF );
    *(p + 3) = static_cast<u8>( data & 0xFF );
}

inline s16
ReadS16( const void* addr )
{
    return static_cast<s16>( ReadU16( addr ) );
}

inline s32
ReadS32( const void* addr )
{
    return static_cast<s32>( ReadU32( addr ) );
}

inline void
WriteS16( void* addr, s16 data )
{
    WriteU16( addr, u16( data ) );
}

inline void
WriteS32( void* addr, s32 data )
{
    WriteU32( addr, u32( data ) );
}

}   /* namespace BEndian */


//=========================================
//
// Utility to extract reference
//  common::Ref, common::CRef
//
//=========================================
template <class T>
class ReferenceWrapper
{
public:
    typedef T type;
    /* ctor */ explicit ReferenceWrapper( T& t ) : mT( &t ) {}

    operator T& () const { return *mT; }

private:
    T* mT;
};

template <class T>
inline ReferenceWrapper<T> Ref( T& t )
{
    return ReferenceWrapper<T>( t );
}

template <class T>
inline ReferenceWrapper<const T> CRef( const T& t )
{
    return ReferenceWrapper<const T>( t );
}


}   /* namespace common */
}   /* namespace vw     */



/* NW4R_VIEWER_COMMON_COMMON_H_ */
#endif
