/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef PROTOCOL_EM4C_DEF
#define PROTOCOL_EM4C_DEF

#include <cstring>
#include <cassert>

namespace ProtocolEM4C
{

    struct Constants
    {
        static const unsigned int VERSION_EM4C = 0x02000000;          // Protocol version.
        static const unsigned int VERSION_MAJOR_EM4C_MASK = 0xFF000000;          // Major protocol version mask.
        static const unsigned int VERSION_MINOR_EM4C_MASK = 0x00FF0000;          // Minor protocol version mask.
        static const unsigned int VERSION_REVISION_EM4C_MASK = 0x0000FF00;          // Protocol revision mask.
        static const unsigned int MESSAGE_CATEGORY_EM4C_MASK = 0x0000FF00;          // Message category mask.
        static const unsigned int CHANNEL_ID_EFFECTMAKER_ACTIVE = 0x04;          // Channel ID for EffectMaker. Sent.
        static const unsigned int CHANNEL_ID_EFFECTMAKER_PASSIVE = 0x05;          // Channel ID for EffectMaker. Received.
        static const unsigned int CHANNEL_ID_EFFECTMAKER_SURVIVAL = 0x06;          // Channel ID for EffectMaker. For existence confirmation
    };

    enum EditTargetType // Edit target types.
    {
        EditTargetType_Node = 0x01,           // Node Target
        EditTargetType_Material = 0x02,           // Material target
        EditTargetType_Emitter = 0x03,           // Emitter.
        EditTargetType_ParticleSet = 0x04,           // Particle set target
        EditTargetType_ParticleModel = 0x05,           // Particle model target
        EditTargetType_Mask = 0xFF,           // Mask
        //
        EditTargetType_FORCE32 = 0xFFFFFFFF
    };

    enum BinaryDataType // Binary data type
    {
        BinaryDataType_Bcmdl = 0x00,           // model
        BinaryDataType_Bctex = 0x01,           // Texture
        BinaryDataType_Bcskla = 0x02,           // Skeletal animations
        BinaryDataType_Bclut = 0x03,           // Lookup table
        BinaryDataType_Bcmata = 0x04,           // Material Animations
        BinaryDataType_Bcptl = 0x10,           // Effect
        //
        BinaryDataType_FORCE32 = 0xFFFFFFFF
    };

    enum FlagBitType // Bit flag definitions (used as in =IsEnabled|IsResourceCopyEnabled)
    {
        FlagBitType_None = 0x00,           // None.
        FlagBitType_IsEnabled = 0x01,           // Initializer/Updater flags: Allowed/not allowed
        FlagBitType_IsResourceCopyEnabled = 0x02,           // Initializer/Updater Flag: Resource copy flag
        //
        FlagBitType_FORCE32 = 0xFFFFFFFF
    };

    enum FollowBitType // Bit definition for Flag (use similar to =IsFollowScale|IsFollowRotate)
    {
        FollowBitType_None = 0x00,           // None.
        FollowBitType_IsFollowScale = 0x01,           // Track: Scale
        FollowBitType_IsFollowRotate = 0x02,           // Track: Rotate
        FollowBitType_IsFollowTranslate = 0x04,           // Track: Move
        //
        FollowBitType_FORCE32 = 0xFFFFFFFF
    };

    enum MagnetAxisType // Bit definition for Flag (use similar to =MagnetXEnable|MagnetYEnable)
    {
        MagnetAxisType_None = 0x00,           // None.
        MagnetAxisType_MagnetXEnable = 0x01,           // Magnet: X-axis enabled
        MagnetAxisType_MagnetYEnable = 0x02,           // Magnet: Y-axis enabled
        MagnetAxisType_MagnetZEnable = 0x04,           // Magnet: Z-axis enabled
        //
        MagnetAxisType_FORCE32 = 0xFFFFFFFF
    };

    enum CollisionReactionType // Collision behavior type
    {
        CollisionReactionType_Reflection = 0x00,           // Collision: Reflection
        CollisionReactionType_Cesser = 0x01,           // Collision: Elimination
        //
        CollisionReactionType_FORCE32 = 0xFFFFFFFF
    };

    enum TextureCombineMode // Definition of the combiner function for the texture combiner
    {
        TextureCombineMode_Replace = 0x0,           // Use only operation 0
        TextureCombineMode_Modulate = 0x1,           // Multiply operation 0 and operation 1
        TextureCombineMode_Add = 0x2,           // Add operation 0 and operation 1
        TextureCombineMode_AddSigned = 0x3,           // Subtract 0.5 from the sum of operation 0 and operation 1
        TextureCombineMode_Interpolate = 0x4,           // Interpolate operation 0 and operation 1 with operation 2
        TextureCombineMode_Subtract = 0x5,           // Subtract operation 1 from operation 0
        TextureCombineMode_DotProduct3Rgb = 0x6,           // Replace each RGB with the inner product of operation 0 and operation 1
        TextureCombineMode_DotProduct3Rgba = 0x7,           // Replace each RGBA with the inner product of operation 0 and operation 1
        TextureCombineMode_AddMult = 0x8,           // Multiply the sum of operation 0 and operation 1 with operation 2
        TextureCombineMode_MultAdd = 0x9,           // Add the product of operation 0 and operation 1 with operation 2
        //
        TextureCombineMode_FORCE32 = 0xFFFFFFFF
    };

    enum TextureCombinerSourceKind // Definition of texture combiner source
    {
        TextureCombinerSourceKind_Texture0 = 0x0,           // Texture ０
        TextureCombinerSourceKind_Texture1 = 0x1,           // Texture １
        TextureCombinerSourceKind_Texture2 = 0x2,           // Texture ２
        TextureCombinerSourceKind_Texture3 = 0x3,           // Texture 3
        TextureCombinerSourceKind_Constant = 0x4,           // Constant color
        TextureCombinerSourceKind_PrimaryColor = 0x5,           // Fragment color input in texture combiner 0 before texture processing
        TextureCombinerSourceKind_Previous = 0x6,           // Fragment color input from the previous texture combiner
        TextureCombinerSourceKind_PreviousBuffer = 0x7,           // Color input from the previous texture combiner buffer
        TextureCombinerSourceKind_FragmentPrimaryColor = 0x8,           // Fragment light primary color (represents diffuse, ambient,and emissive)
        TextureCombinerSourceKind_FragmentSecondaryColor = 0x9,           // Fragment light secondary color (represents specular)
        //
        TextureCombinerSourceKind_FORCE32 = 0xFFFFFFFF
    };

    enum TextureCombinerOperandKind // Definition of texture combiner operand
    {
        TextureCombinerOperandKind_Color = 0x0,           // Source color
        TextureCombinerOperandKind_OneMinusColor = 0x1,           // 1-Color
        TextureCombinerOperandKind_Alpha = 0x2,           // Source alpha value
        TextureCombinerOperandKind_OneMinusAlpha = 0x3,           // 1-Alpha
        TextureCombinerOperandKind_Red = 0x4,           // Source Red value
        TextureCombinerOperandKind_Green = 0x5,           // Source Green value
        TextureCombinerOperandKind_Blue = 0x6,           // Source Blue value
        TextureCombinerOperandKind_OneMinusRed = 0x7,           // 1-Red
        TextureCombinerOperandKind_OneMinusGreen = 0x8,           // 1-Green
        TextureCombinerOperandKind_OneMinusBlue = 0x9,           // 1-Blue
        //
        TextureCombinerOperandKind_FORCE32 = 0xFFFFFFFF
    };

    enum TextureSamplerWrap // Definition for display methods after the texture edge
    {
        TextureSamplerWrap_Repeat = 0x0,           // Repeats at the texture edge.
        TextureSamplerWrap_MirroredRepeat = 0x1,           // Mirrors and repeats at the texture edge.
        TextureSamplerWrap_ClampToEdge = 0x2,           // Extends the pixels at the edge of a texture.
        TextureSamplerWrap_ClampToBorder = 0x3,           // Extends the border color at the edge of a texture.
        //
        TextureSamplerWrap_FORCE32 = 0xFFFFFFFF
    };

    enum TextureMatrixModeCtr // Texture coordinate calculation method.
    {
        TextureMatrixModeCtr_DccMaya = 0x0,           // Calculates texture coordinates in the same way as Maya.
        TextureMatrixModeCtr_DccSoftimage = 0x1,           // Calculates texture coordinates in the same way as Softimage.
        TextureMatrixModeCtr_Dcc3dsMax = 0x2,           // Calculates texture coordinates in the same way as 3dsmax.
        //
        TextureMatrixModeCtr_FORCE32 = 0xFFFFFFFF
    };

    enum MessageCategory  // Message categories.
    {
        MessageCategory_File = 0x0000,      // File
        MessageCategory_Object = 0x0100,      // Object
        MessageCategory_ParticleModel = 0x0200,      // model
        MessageCategory_Emitter = 0x0300,      // Emitter
        MessageCategory_ParticleForm = 0x0400,      // ParticleForm
        MessageCategory_ParticleInitializer = 0x0500,      // Initializer
        MessageCategory_ParticleUpdater = 0x0600,      // Updater
        MessageCategory_ParticleSet = 0x0700,      // ParticleSet
        MessageCategory_Material = 0x0800,      // Material
        MessageCategory_BoneLinkage = 0x0900,      // BoneLinkage
        //
        MessageCategory_FORCE32  = 0xFFFFFFFF
    };

    enum MessageType   // Message types.
    {
        MessageType_File_CloseAll = 0x00 | MessageCategory_File,    // Closes all files.
        MessageType_File_Close = 0x01 | MessageCategory_File,    // Close selected models.
        MessageType_File_LoadBinaryData = 0x02 | MessageCategory_File,    // Load binary data.
        MessageType_Object_SelectEM = 0x00 | MessageCategory_Object,    // Select an edit target.
        MessageType_Object_MoveFirstFrame = 0x01 | MessageCategory_Object,    // Rewind to first frame.
        MessageType_Object_StopUpdateAll = 0x02 | MessageCategory_Object,    // Stop all animation updates.
        MessageType_Object_StartUpdateAll = 0x03 | MessageCategory_Object,    // Start all animation updates.
        MessageType_Object_StepUpdateAll = 0x04 | MessageCategory_Object,    // Update all animations by 1 frame (step forward 1 frame).
        MessageType_Object_SetFrameRange = 0x05 | MessageCategory_Object,    // Specify range of frames to play back.
        MessageType_ParticleModel_SetTransform = 0x00 | MessageCategory_ParticleModel,    // Sets "placement."
        MessageType_Emitter_SetTransform = 0x00 | MessageCategory_Emitter,    // Sets "placement."
        MessageType_Emitter_SetEmission = 0x01 | MessageCategory_Emitter,    // Sets "emission."
        MessageType_ParticleForm_CylinderForm = 0x00 | MessageCategory_ParticleForm,    // Sets "cylinder."
        MessageType_ParticleForm_SphereForm = 0x01 | MessageCategory_ParticleForm,    // Sets "sphere."
        MessageType_ParticleForm_DiscForm = 0x02 | MessageCategory_ParticleForm,    // Sets "disk."
        MessageType_ParticleForm_RectangleForm = 0x03 | MessageCategory_ParticleForm,    // Sets "rectangle."
        MessageType_ParticleForm_CubeForm = 0x04 | MessageCategory_ParticleForm,    // Sets "cube."
        MessageType_ParticleInitializer_YAxisVelocityInitializer = 0x00 | MessageCategory_ParticleInitializer,    // ParticleYAxisVelocityInitializer|Sets "Diffusion from Y-axis" for "initial speed."
        MessageType_ParticleInitializer_OriginVelocityInitializer = 0x01 | MessageCategory_ParticleInitializer,    // ParticleOriginVelocityInitializer|Sets "Diffusion from origin" for "initial speed."
        MessageType_ParticleInitializer_RandomVelocityInitializer = 0x02 | MessageCategory_ParticleInitializer,    // ParticleRandomVelocityInitializer|Sets "random direction" for "initial speed."
        MessageType_ParticleInitializer_DirectionalVelocityInitializer = 0x03 | MessageCategory_ParticleInitializer,    // ParticleDirectionalVelocityInitializer|Sets "specified direction" for "initial speed."
        MessageType_ParticleInitializer_RandomDirectionalVelocityInitializer = 0x04 | MessageCategory_ParticleInitializer,    // ParticleRandomDirectionalVelocityInitializer|Sets "random specified direction" for "initial speed."
        MessageType_ParticleInitializer_Vector3MultRandomInitializer = 0x05 | MessageCategory_ParticleInitializer,    // ParticleRandomDirectionalVelocityInitializer|Sets "randomize results" for "initial speed."
        MessageType_ParticleInitializer_FloatRangeRandomInitializer = 0x06 | MessageCategory_ParticleInitializer,    // ParticleFloatRangeRandomInitializer|Sets "lifetime."
        MessageType_ParticleInitializer_FloatImmediateInitializer = 0x07 | MessageCategory_ParticleInitializer,    // ParticleFloatImmediateInitializer|Sets "alpha."
        MessageType_ParticleInitializer_FloatRandomInitializer = 0x08 | MessageCategory_ParticleInitializer,    // Sets ParticleFloatRandomInitializer "alpha."
        MessageType_ParticleInitializer_Vector3Random1Initializer = 0x09 | MessageCategory_ParticleInitializer,    // ParticleVector3Random1Initializer|Sets "lifetime."
        MessageType_ParticleInitializer_Vector3Random3Initializer = 0x0A | MessageCategory_ParticleInitializer,    // ParticleVector3Random3Initializer|Sets "rotate."
        MessageType_ParticleInitializer_Vector3ImmediateInitializer = 0x0B | MessageCategory_ParticleInitializer,    // ParticleVector3ImmediateInitializer|Sets "color."
        MessageType_ParticleInitializer_Vector2ImmediateInitializer = 0x0C | MessageCategory_ParticleInitializer,    // ParticleVector2ImmediateInitializer|Sets "texture UV."
        MessageType_ParticleUpdater_GravityUpdater = 0x00 | MessageCategory_ParticleUpdater,    // ParticleGravityUpdater|"Add to specified direction" "gravity"
        MessageType_ParticleUpdater_SpinUpdater = 0x01 | MessageCategory_ParticleUpdater,    // ParticleSpinUpdater|"Spin"
        MessageType_ParticleUpdater_RandomUpdater = 0x02 | MessageCategory_ParticleUpdater,    // ParticleRandomUpdater|"Random"
        MessageType_ParticleUpdater_AccelarationUpdater = 0x03 | MessageCategory_ParticleUpdater,    // ParticleAccelarationUpdater|Sets "acceleration" for "field."
        MessageType_ParticleUpdater_RotateUpVectorUpdater = 0x04 | MessageCategory_ParticleUpdater,    // Sets ParticleRotateUpVectorUpdater "rotate."
        MessageType_ParticleUpdater_FloatImmediateUpdater = 0x05 | MessageCategory_ParticleUpdater,    // ParticleFloatImmediateUpdater|Sets "alpha."
        MessageType_ParticleUpdater_Vector3AdditiveUpdater = 0x06 | MessageCategory_ParticleUpdater,    // ParticleVector3AdditiveUpdater|Sets "scale, scale expansion."
        MessageType_ParticleUpdater_Vector3ImmediateUpdater = 0x07 | MessageCategory_ParticleUpdater,    // ParticleVector3ImmediateUpdater|Sets "scale, scale expansion."
        MessageType_ParticleUpdater_Vector2ImmediateUpdater = 0x08 | MessageCategory_ParticleUpdater,    // ParticleVector2ImmediateUpdater|Sets "texture UV."
        MessageType_ParticleUpdater_TexturePatternUpdater = 0x09 | MessageCategory_ParticleUpdater,    // ParticleTexturePatternUpdater|Sets "texture UV."
        MessageType_ParticleUpdater_ChildUpdater = 0x0A | MessageCategory_ParticleUpdater,    // ParticleChildUpdater|Sets form for child.
        MessageType_ParticleUpdater_FloatImmediate4KeyUpdater = 0x0B | MessageCategory_ParticleUpdater,    // ParticleFloatImmediate4KeyUpdater|Sets 4 key.
        MessageType_ParticleUpdater_Vector3Immediate4KeyUpdater = 0x0C | MessageCategory_ParticleUpdater,    // ParticleVector3Immediate4KeyUpdater|Sets 4 key.
        MessageType_ParticleUpdater_MagnetUpdater = 0x0D | MessageCategory_ParticleUpdater,    // ParticleMagnetUpdater|Sets "magnet."
        MessageType_ParticleUpdater_CollisionUpdater = 0x0E | MessageCategory_ParticleUpdater,    // CollisionUpdater|「Sets "Collision"
        MessageType_ParticleSet_ShapeBuilder = 0x00 | MessageCategory_ParticleSet,    // ParticleShapeBuilder|Sets "ShapeBuilder."
        MessageType_Material_MaterialBlender = 0x00 | MessageCategory_Material,    // MaterialBlender|Sets "Material Blender."
        MessageType_Material_MaterialColor = 0x01 | MessageCategory_Material,    // MaterialColor|Sets "Material Color."
        MessageType_Material_MaterialMapper = 0x02 | MessageCategory_Material,    // MaterialMapper|Sets "Material Mapper."
        MessageType_Material_MaterialCombiner = 0x03 | MessageCategory_Material,    // MaterialCombiner|Sets "Material Combiner."
        MessageType_BoneLinkage_ModelNameList = 0x00 | MessageCategory_BoneLinkage,    // Have the model name list send from the viewer side
        MessageType_BoneLinkage_BoneNameList = 0x01 | MessageCategory_BoneLinkage,    // To start having the bone name list sent from the viewer side, when a packet is sent to the viewer, the bone list can be received.
        MessageType_BoneLinkage_SetBoneLink = 0x02 | MessageCategory_BoneLinkage,    // Send BoneLink information to the viewer
        //
        MessageType_FORCE32  = 0xFFFFFFFF
    };

    struct Header
    {
        static const int StructSize = 12;

        MessageType type;
        int size;
        int arraySize;
    };

    typedef int Int;
    typedef unsigned int Uint;
    typedef float Float;

    struct String
    {
        wchar_t Value[260];

        const wchar_t *ToString()
        {
            return Value;
        }
    };

    struct Rgb
    {
        float R;
        float G;
        float B;
    };

    struct Rgba
    {
        float R;
        float G;
        float B;
        float A;
    };

    struct Vector2
    {
        float X;
        float Y;
    };

    struct Vector3
    {
        float X;
        float Y;
        float Z;
    };

    struct Vector4
    {
        float X;
        float Y;
        float Z;
        float W;
    };

    // Closes all files.
    struct MESSAGEPACKET_File_CloseAll
    {
        static const MessageType MessagePacketType = MessageType_File_CloseAll;    // Message types.
        static const int StructSize = 0 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

    };

    // Close selected models.
    struct MESSAGEPACKET_File_Close
    {
        static const MessageType MessagePacketType = MessageType_File_Close;    // Message types.
        static const int StructSize = 8 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Type = 1<<0,
            PropertyFlag_Id = 1<<1,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        BinaryDataType    Type;       // Type.
        Uint    Id;       // Returns the ID.
    };

    // Load binary data.
    struct MESSAGEPACKET_File_LoadBinaryData
    {
        static const MessageType MessagePacketType = MessageType_File_LoadBinaryData;    // Message types.
        static const int StructSize = 532 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Type = 1<<0,
            PropertyFlag_Id = 1<<1,
            PropertyFlag_FileSize = 1<<2,
            PropertyFlag_FilePath = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        BinaryDataType    Type;       // Type.
        Uint    Id;       // Returns the ID.
        Uint    FileSize;       // The size of the file.
        String    FilePath;       // The file path.
    };

    // Select an edit target.
    struct MESSAGEPACKET_Object_SelectEM
    {
        static const MessageType MessagePacketType = MessageType_Object_SelectEM;    // Message types.
        static const int StructSize = 532 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_ObjectType = 1<<0,
            PropertyFlag_TargetType = 1<<1,
            PropertyFlag_Id = 1<<2,
            PropertyFlag_TargetName = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        BinaryDataType    ObjectType;       // Object type.
        EditTargetType    TargetType;       // Edit target type.
        Uint    Id;       // Binary hash.
        String    TargetName;       // Edit target name.
    };

    // Rewind to first frame.
    struct MESSAGEPACKET_Object_MoveFirstFrame
    {
        static const MessageType MessagePacketType = MessageType_Object_MoveFirstFrame;    // Message types.
        static const int StructSize = 0 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

    };

    // Stop all animation updates.
    struct MESSAGEPACKET_Object_StopUpdateAll
    {
        static const MessageType MessagePacketType = MessageType_Object_StopUpdateAll;    // Message types.
        static const int StructSize = 0 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

    };

    // Start all animation updates.
    struct MESSAGEPACKET_Object_StartUpdateAll
    {
        static const MessageType MessagePacketType = MessageType_Object_StartUpdateAll;    // Message types.
        static const int StructSize = 0 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

    };

    // Update all animations by 1 frame (step forward 1 frame).
    struct MESSAGEPACKET_Object_StepUpdateAll
    {
        static const MessageType MessagePacketType = MessageType_Object_StepUpdateAll;    // Message types.
        static const int StructSize = 0 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

    };

    // Specify range of frames to play back.
    struct MESSAGEPACKET_Object_SetFrameRange
    {
        static const MessageType MessagePacketType = MessageType_Object_SetFrameRange;    // Message types.
        static const int StructSize = 8 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_StartFrame = 1<<0,
            PropertyFlag_EndFrame = 1<<1,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    StartFrame;       // The starting frame.
        Int    EndFrame;       // The ending frame
    };

    // Sets "placement."
    struct MESSAGEPACKET_ParticleModel_SetTransform
    {
        static const MessageType MessagePacketType = MessageType_ParticleModel_SetTransform;    // Message types.
        static const int StructSize = 36 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Scale = 1<<0,
            PropertyFlag_Rotate = 1<<1,
            PropertyFlag_Translate = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Vector3    Scale;       // Scaling.
        Vector3    Rotate;       // Rotation.
        Vector3    Translate;       // Position.
    };

    // Sets "placement."
    struct MESSAGEPACKET_Emitter_SetTransform
    {
        static const MessageType MessagePacketType = MessageType_Emitter_SetTransform;    // Message types.
        static const int StructSize = 36 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Scale = 1<<0,
            PropertyFlag_Rotate = 1<<1,
            PropertyFlag_Translate = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Vector3    Scale;       // Scaling.
        Vector3    Rotate;       // Rotation.
        Vector3    Translate;       // Position.
    };

    // Sets "emission."
    struct MESSAGEPACKET_Emitter_SetEmission
    {
        static const MessageType MessagePacketType = MessageType_Emitter_SetEmission;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_EmissionRatio = 1<<0,
            PropertyFlag_EmissionRatioRandom = 1<<1,
            PropertyFlag_EmissionInterval = 1<<2,
            PropertyFlag_EmissionIntervalRandom = 1<<3,
            PropertyFlag_EmissionSpanInfinity = 1<<4,
            PropertyFlag_EmissionSpan = 1<<5,
            PropertyFlag_EmissionStart = 1<<6,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Float    EmissionRatio;       // Emission amount.
        Float    EmissionRatioRandom;       // Random % of the emission amount.
        Int    EmissionInterval;       // Emission interval.
        Float    EmissionIntervalRandom;       // Random % of the emission interval.
        Int    EmissionSpanInfinity;       // Emission time: infinity flag. (=0..false/=1..true)
        Int    EmissionSpan;       // Emission time: emission time.
        Int    EmissionStart;       // Emission time: emission start time.
    };

    // Sets "cylinder."
    struct MESSAGEPACKET_ParticleForm_CylinderForm
    {
        static const MessageType MessagePacketType = MessageType_ParticleForm_CylinderForm;    // Message types.
        static const int StructSize = 36 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Scale = 1<<0,
            PropertyFlag_Inner = 1<<1,
            PropertyFlag_AngleOffset = 1<<2,
            PropertyFlag_AngleWidth = 1<<3,
            PropertyFlag_AngleSwing = 1<<4,
            PropertyFlag_FixedOffset = 1<<5,
            PropertyFlag_EvenInterval = 1<<6,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Vector3    Scale;       // Size.
        Float    Inner;       // Cave.
        Float    AngleOffset;       // Orphan: Center.
        Float    AngleWidth;       // Width.
        Float    AngleSwing;       // Angle correction.
        Int    FixedOffset;       // Center fixed.
        Int    EvenInterval;       // Uniform placement.
    };

    // Sets "sphere."
    struct MESSAGEPACKET_ParticleForm_SphereForm
    {
        static const MessageType MessagePacketType = MessageType_ParticleForm_SphereForm;    // Message types.
        static const int StructSize = 36 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Scale = 1<<0,
            PropertyFlag_Inner = 1<<1,
            PropertyFlag_AngleOffset = 1<<2,
            PropertyFlag_AngleWidth = 1<<3,
            PropertyFlag_AngleSwing = 1<<4,
            PropertyFlag_FixedOffset = 1<<5,
            PropertyFlag_EvenInterval = 1<<6,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Vector3    Scale;       // Size.
        Float    Inner;       // Cave.
        Float    AngleOffset;       // Orphan: Center.
        Float    AngleWidth;       // Width.
        Float    AngleSwing;       // Angle correction.
        Int    FixedOffset;       // Center fixed.
        Int    EvenInterval;       // Uniform placement.
    };

    // Sets "disk."
    struct MESSAGEPACKET_ParticleForm_DiscForm
    {
        static const MessageType MessagePacketType = MessageType_ParticleForm_DiscForm;    // Message types.
        static const int StructSize = 32 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Scale = 1<<0,
            PropertyFlag_Inner = 1<<1,
            PropertyFlag_AngleOffset = 1<<2,
            PropertyFlag_AngleWidth = 1<<3,
            PropertyFlag_AngleSwing = 1<<4,
            PropertyFlag_FixedOffset = 1<<5,
            PropertyFlag_EvenInterval = 1<<6,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Vector2    Scale;       // Size.
        Float    Inner;       // Cave.
        Float    AngleOffset;       // Orphan: Center.
        Float    AngleWidth;       // Width.
        Float    AngleSwing;       // Angle correction.
        Int    FixedOffset;       // Center fixed.
        Int    EvenInterval;       // Uniform placement.
    };

    // Sets "rectangle."
    struct MESSAGEPACKET_ParticleForm_RectangleForm
    {
        static const MessageType MessagePacketType = MessageType_ParticleForm_RectangleForm;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Scale = 1<<0,
            PropertyFlag_Inner = 1<<1,
            PropertyFlag_EvenInterval = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Vector2    Scale;       // Size.
        Float    Inner;       // Cave.
        Int    EvenInterval;       // Uniform placement.
    };

    // Sets "cube."
    struct MESSAGEPACKET_ParticleForm_CubeForm
    {
        static const MessageType MessagePacketType = MessageType_ParticleForm_CubeForm;    // Message types.
        static const int StructSize = 20 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Inner = 1<<0,
            PropertyFlag_Scale = 1<<1,
            PropertyFlag_EvenInterval = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Float    Inner;       // Get or set the inner diameter ratio.
        Vector3    Scale;       // Get or set the shape scale.
        Int    EvenInterval;       // Get or set the flag to place evenly.
    };

    // ParticleYAxisVelocityInitializer|Sets "Diffusion from Y-axis" for "initial speed."
    struct MESSAGEPACKET_ParticleInitializer_YAxisVelocityInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_YAxisVelocityInitializer;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Velocity.
    };

    // ParticleOriginVelocityInitializer|Sets "Diffusion from origin" for "initial speed."
    struct MESSAGEPACKET_ParticleInitializer_OriginVelocityInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_OriginVelocityInitializer;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Velocity.
    };

    // ParticleRandomVelocityInitializer|Sets "random direction" for "initial speed."
    struct MESSAGEPACKET_ParticleInitializer_RandomVelocityInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_RandomVelocityInitializer;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Velocity.
    };

    // ParticleDirectionalVelocityInitializer|Sets "specified direction" for "initial speed."
    struct MESSAGEPACKET_ParticleInitializer_DirectionalVelocityInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_DirectionalVelocityInitializer;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_Direction = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Velocity.
        Vector3    Direction;       // Direction.
    };

    // ParticleRandomDirectionalVelocityInitializer|Sets "random specified direction" for "initial speed."
    struct MESSAGEPACKET_ParticleInitializer_RandomDirectionalVelocityInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_RandomDirectionalVelocityInitializer;    // Message types.
        static const int StructSize = 32 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_Angle = 1<<4,
            PropertyFlag_Direction = 1<<5,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Velocity.
        Float    Angle;       // Random angle.
        Vector3    Direction;       // Direction.
    };

    // ParticleRandomDirectionalVelocityInitializer|Sets "randomize results" for "initial speed."
    struct MESSAGEPACKET_ParticleInitializer_Vector3MultRandomInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_Vector3MultRandomInitializer;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Random = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Random;       // Random.
    };

    // ParticleFloatRangeRandomInitializer|Sets "lifetime."
    struct MESSAGEPACKET_ParticleInitializer_FloatRangeRandomInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_FloatRangeRandomInitializer;    // Message types.
        static const int StructSize = 20 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_MaxValue = 1<<3,
            PropertyFlag_MinValue = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    MaxValue;       // Maximum value
        Float    MinValue;       // Minimum value
    };

    // ParticleFloatImmediateInitializer|Sets "alpha."
    struct MESSAGEPACKET_ParticleInitializer_FloatImmediateInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_FloatImmediateInitializer;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_ImmediateValue = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    ImmediateValue;       // Initial value
    };

    // Sets ParticleFloatRandomInitializer "alpha."
    struct MESSAGEPACKET_ParticleInitializer_FloatRandomInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_FloatRandomInitializer;    // Message types.
        static const int StructSize = 20 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_BaseValue = 1<<3,
            PropertyFlag_Random = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    BaseValue;       // Standard value
        Float    Random;       // Random
    };

    // ParticleVector3Random1Initializer|Sets "lifetime."
    struct MESSAGEPACKET_ParticleInitializer_Vector3Random1Initializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_Vector3Random1Initializer;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_BaseValue = 1<<3,
            PropertyFlag_Random = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector3    BaseValue;       // Standard value
        Float    Random;       // Amplitude
    };

    // ParticleVector3Random3Initializer|Sets "rotate."
    struct MESSAGEPACKET_ParticleInitializer_Vector3Random3Initializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_Vector3Random3Initializer;    // Message types.
        static const int StructSize = 36 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_BaseValue = 1<<3,
            PropertyFlag_Random = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector3    BaseValue;       // Standard value
        Vector3    Random;       // Amplitude
    };

    // ParticleVector3ImmediateInitializer|Sets "color."
    struct MESSAGEPACKET_ParticleInitializer_Vector3ImmediateInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_Vector3ImmediateInitializer;    // Message types.
        static const int StructSize = 24 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_ImmediateValue = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector3    ImmediateValue;       // Color
    };

    // ParticleVector2ImmediateInitializer|Sets "texture UV."
    struct MESSAGEPACKET_ParticleInitializer_Vector2ImmediateInitializer
    {
        static const MessageType MessagePacketType = MessageType_ParticleInitializer_Vector2ImmediateInitializer;    // Message types.
        static const int StructSize = 20 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_ImmediateValue = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector2    ImmediateValue;       // UV
    };

    // ParticleGravityUpdater|"Add to specified direction" "gravity"
    struct MESSAGEPACKET_ParticleUpdater_GravityUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_GravityUpdater;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_Direction = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Strength
        Vector3    Direction;       // Position.
    };

    // ParticleSpinUpdater|"Spin"
    struct MESSAGEPACKET_ParticleUpdater_SpinUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_SpinUpdater;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_Axis = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Strength
        Vector3    Axis;       // Position.
    };

    // ParticleRandomUpdater|"Random"
    struct MESSAGEPACKET_ParticleUpdater_RandomUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_RandomUpdater;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Interval = 1<<3,
            PropertyFlag_Power = 1<<4,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Int    Interval;       // Axis strength
        Vector3    Power;       // Strength.
    };

    // ParticleAccelarationUpdater|Sets "acceleration" for "field."
    struct MESSAGEPACKET_ParticleUpdater_AccelarationUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_AccelarationUpdater;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Factor = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Factor;       // Acceleration coefficient.
    };

    // Sets ParticleRotateUpVectorUpdater "rotate."
    struct MESSAGEPACKET_ParticleUpdater_RotateUpVectorUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_RotateUpVectorUpdater;    // Message types.
        static const int StructSize = 16 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_IsUpdaterEnabled = 1<<2,
            PropertyFlag_Source = 1<<3,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    IsUpdaterEnabled;       // "Enable" check box
        Int    Source;       // Source Types
    };

    // ParticleFloatImmediateUpdater|Sets "alpha."
    struct MESSAGEPACKET_ParticleUpdater_FloatImmediateUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_FloatImmediateUpdater;    // Message types.
        static const int StructSize = 24 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_DefaultValue = 1<<3,
            PropertyFlag_LoopDuration = 1<<4,
            PropertyFlag_RandomOffset = 1<<5,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    DefaultValue;       // Standard value
        Int    LoopDuration;       // ParticleFrameLoopAnimationOption｜Length of one loop
        Int    RandomOffset;       // ParticleFrameLoopAnimationOption｜Flag to add to random offset
    };

    // ParticleVector3AdditiveUpdater|Sets "scale, scale expansion."
    struct MESSAGEPACKET_ParticleUpdater_Vector3AdditiveUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_Vector3AdditiveUpdater;    // Message types.
        static const int StructSize = 32 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_DefaultValue = 1<<3,
            PropertyFlag_LoopDuration = 1<<4,
            PropertyFlag_RandomOffset = 1<<5,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector3    DefaultValue;       // Amplitude
        Int    LoopDuration;       // ParticleFrameLoopAnimationOption｜Length of one loop
        Int    RandomOffset;       // ParticleFrameLoopAnimationOption｜Flag to add to random offset
    };

    // ParticleVector3ImmediateUpdater|Sets "scale, scale expansion."
    struct MESSAGEPACKET_ParticleUpdater_Vector3ImmediateUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_Vector3ImmediateUpdater;    // Message types.
        static const int StructSize = 32 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_DefaultValue = 1<<3,
            PropertyFlag_LoopDuration = 1<<4,
            PropertyFlag_RandomOffset = 1<<5,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector3    DefaultValue;       // Standard value
        Int    LoopDuration;       // ParticleFrameLoopAnimationOption｜Length of one loop
        Int    RandomOffset;       // ParticleFrameLoopAnimationOption｜Flag to add to random offset
    };

    // ParticleVector2ImmediateUpdater|Sets "texture UV."
    struct MESSAGEPACKET_ParticleUpdater_Vector2ImmediateUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_Vector2ImmediateUpdater;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_DefaultValue = 1<<3,
            PropertyFlag_LoopDuration = 1<<4,
            PropertyFlag_RandomOffset = 1<<5,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Vector2    DefaultValue;       // Standard value
        Int    LoopDuration;       // ParticleFrameLoopAnimationOption｜Length of one loop
        Int    RandomOffset;       // ParticleFrameLoopAnimationOption｜Flag to add to random offset
    };

    // ParticleTexturePatternUpdater|Sets "texture UV."
    struct MESSAGEPACKET_ParticleUpdater_TexturePatternUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_TexturePatternUpdater;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_DivisionX = 1<<3,
            PropertyFlag_DivisionY = 1<<4,
            PropertyFlag_LoopDuration = 1<<5,
            PropertyFlag_RandomOffset = 1<<6,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Int    DivisionX;       //
        Int    DivisionY;       //
        Int    LoopDuration;       // ParticleFrameLoopAnimationOption｜Length of one loop
        Int    RandomOffset;       // ParticleFrameLoopAnimationOption｜Flag to add to random offset
    };

    // ParticleChildUpdater|Sets form for child.
    struct MESSAGEPACKET_ParticleUpdater_ChildUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_ChildUpdater;    // Message types.
        static const int StructSize = 40 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Scale = 1<<1,
            PropertyFlag_Inner = 1<<2,
            PropertyFlag_AngleOffset = 1<<3,
            PropertyFlag_AngleWidth = 1<<4,
            PropertyFlag_AngleSwing = 1<<5,
            PropertyFlag_FixedOffset = 1<<6,
            PropertyFlag_EvenInterval = 1<<7,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Vector3    Scale;       // Size.
        Float    Inner;       // Cave.
        Float    AngleOffset;       // Orphan: Center.
        Float    AngleWidth;       // Width.
        Float    AngleSwing;       // Angle correction.
        Int    FixedOffset;       // Center fixed.
        Int    EvenInterval;       // Uniform placement.
    };

    // ParticleFloatImmediate4KeyUpdater|Sets 4 key.
    struct MESSAGEPACKET_ParticleUpdater_FloatImmediate4KeyUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_FloatImmediate4KeyUpdater;    // Message types.
        static const int StructSize = 40 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_InTime = 1<<3,
            PropertyFlag_OutTime = 1<<4,
            PropertyFlag_Value0 = 1<<5,
            PropertyFlag_Value1 = 1<<6,
            PropertyFlag_Value2 = 1<<7,
            PropertyFlag_Value3 = 1<<8,
            PropertyFlag_Value4 = 1<<9,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Uint    InTime;       // End time for fade in
        Uint    OutTime;       // Start time for fade out
        Int    Value0;       // Value 0 for fade in
        Int    Value1;       // Value 1 for fade in
        Int    Value2;       // Value 2 for fade in
        Int    Value3;       // Value 3 for fade in
        Int    Value4;       // Value 4 for fade in
    };

    // ParticleVector3Immediate4KeyUpdater|Sets 4 key.
    struct MESSAGEPACKET_ParticleUpdater_Vector3Immediate4KeyUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_Vector3Immediate4KeyUpdater;    // Message types.
        static const int StructSize = 80 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_InTime = 1<<3,
            PropertyFlag_OutTime = 1<<4,
            PropertyFlag_Value0 = 1<<5,
            PropertyFlag_Value1 = 1<<6,
            PropertyFlag_Value2 = 1<<7,
            PropertyFlag_Value3 = 1<<8,
            PropertyFlag_Value4 = 1<<9,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Uint    InTime;       // End time for fade in
        Uint    OutTime;       // Start time for fade out
        Vector3    Value0;       // Value 0 for fade in
        Vector3    Value1;       // Value 1 for fade in
        Vector3    Value2;       // Value 2 for fade in
        Vector3    Value3;       // Value 3 for fade in
        Vector3    Value4;       // Value 4 for fade in
    };

    // ParticleMagnetUpdater|Sets "magnet."
    struct MESSAGEPACKET_ParticleUpdater_MagnetUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_MagnetUpdater;    // Message types.
        static const int StructSize = 32 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_Position = 1<<4,
            PropertyFlag_AxisType = 1<<5,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Strength.
        Vector3    Position;       // Position.
        Int    AxisType;       // Type of axis that issues the magnet.
    };

    // CollisionUpdater|「Sets "Collision"
    struct MESSAGEPACKET_ParticleUpdater_CollisionUpdater
    {
        static const MessageType MessagePacketType = MessageType_ParticleUpdater_CollisionUpdater;    // Message types.
        static const int StructSize = 28 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_Usage = 1<<1,
            PropertyFlag_Flag = 1<<2,
            PropertyFlag_Power = 1<<3,
            PropertyFlag_Position = 1<<4,
            PropertyFlag_IsWorldCollision = 1<<5,
            PropertyFlag_ReactionType = 1<<6,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        Int    Usage;       // ParticleUsage value.
        FlagBitType    Flag;       // IsEnabled|Use similar to IsResourceCopyEnabled
        Float    Power;       // Strength.
        Float    Position;       // Position.
        Int    IsWorldCollision;       // Flag indicating whether collision occurs in world coordinate system.
        CollisionReactionType    ReactionType;       // Type of behaivor when collision occurs.
    };

    // ParticleShapeBuilder|Sets "ShapeBuilder."
    struct MESSAGEPACKET_ParticleSet_ShapeBuilder
    {
        static const MessageType MessagePacketType = MessageType_ParticleSet_ShapeBuilder;    // Message types.
        static const int StructSize = 20 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Usage = 1<<0,
            PropertyFlag_AscendingOrder = 1<<1,
            PropertyFlag_DrawOffset = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Usage;       // ParticleUsage value.
        Int    AscendingOrder;       // Render in playback order
        Vector3    DrawOffset;       // Offset of the center of the particle SRT
    };

    // MaterialBlender|Sets "Material Blender."
    struct MESSAGEPACKET_Material_MaterialBlender
    {
        static const MessageType MessagePacketType = MessageType_Material_MaterialBlender;    // Message types.
        static const int StructSize = 8 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_TranslucencyKind = 1<<0,
            PropertyFlag_IsFogEnabled = 1<<1,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    TranslucencyKind;       //
        Int    IsFogEnabled;       //
    };

    // MaterialColor|Sets "Material Color."
    struct MESSAGEPACKET_Material_MaterialColor
    {
        static const MessageType MessagePacketType = MessageType_Material_MaterialColor;    // Message types.
        static const int StructSize = 40 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Emission = 1<<0,
            PropertyFlag_Ambient = 1<<1,
            PropertyFlag_Diffuse = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Rgb    Emission;       // Emission.
        Rgb    Ambient;       // Ambient.
        Rgba    Diffuse;       // Diffuse.
    };

    // MaterialMapper|Sets "Material Mapper."
    struct MESSAGEPACKET_Material_MaterialMapper
    {
        static const MessageType MessagePacketType = MessageType_Material_MaterialMapper;    // Message types.
        static const int StructSize = 48 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_Index = 1<<0,
            PropertyFlag_WrapS = 1<<1,
            PropertyFlag_WrapT = 1<<2,
            PropertyFlag_MinLod = 1<<3,
            PropertyFlag_LodBias = 1<<4,
            PropertyFlag_IsMipMapUsed = 1<<5,
            PropertyFlag_IsMipMapInterpolated = 1<<6,
            PropertyFlag_TexCoordScale = 1<<7,
            PropertyFlag_TexCoordTranslate = 1<<8,
            PropertyFlag_MatrixMode = 1<<9,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Int    Index;       // Index in the array.
        TextureSamplerWrap    WrapS;       // Repeat setting U direction
        TextureSamplerWrap    WrapT;       // Repeat setting V direction
        Int    MinLod;       // LOD minimum level
        Float    LodBias;       // LOD Bias
        Int    IsMipMapUsed;       // MipMap.
        Int    IsMipMapInterpolated;       // Interpolation
        Vector2    TexCoordScale;       // UV scale.
        Vector2    TexCoordTranslate;       // UV movement.
        TextureMatrixModeCtr    MatrixMode;       // Matrix mode.
    };

    // MaterialCombiner|Sets "Material Combiner."
    struct MESSAGEPACKET_Material_MaterialCombiner
    {
        static const MessageType MessagePacketType = MessageType_Material_MaterialCombiner;    // Message types.
        static const int StructSize = 64 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_CombineRgb = 1<<0,
            PropertyFlag_SourceRgb0 = 1<<1,
            PropertyFlag_SourceRgb1 = 1<<2,
            PropertyFlag_SourceRgb2 = 1<<3,
            PropertyFlag_OperandRgb0 = 1<<4,
            PropertyFlag_OperandRgb1 = 1<<5,
            PropertyFlag_OperandRgb2 = 1<<6,
            PropertyFlag_Constant0 = 1<<7,
            PropertyFlag_CombineAlpha = 1<<8,
            PropertyFlag_SourceAlpha0 = 1<<9,
            PropertyFlag_SourceAlpha1 = 1<<10,
            PropertyFlag_OperandAlpha0 = 1<<11,
            PropertyFlag_OperandAlpha1 = 1<<12,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        TextureCombineMode    CombineRgb;       // CombineRgb.
        TextureCombinerSourceKind    SourceRgb0;       // SourceRgbSources0.
        TextureCombinerSourceKind    SourceRgb1;       // SourceRgbSources1.
        TextureCombinerSourceKind    SourceRgb2;       // SourceRgbSources2.
        TextureCombinerOperandKind    OperandRgb0;       // OperandRgbOperands0.
        TextureCombinerOperandKind    OperandRgb1;       // OperandRgbOperands1.
        TextureCombinerOperandKind    OperandRgb2;       // OperandRgbOperands2.
        Rgba    Constant0;       // Constant0 color R.
        TextureCombineMode    CombineAlpha;       // CombineAlpha.
        TextureCombinerSourceKind    SourceAlpha0;       // SourceAlphaSources0.
        TextureCombinerSourceKind    SourceAlpha1;       // SourceAlphaSources1.
        TextureCombinerOperandKind    OperandAlpha0;       // OperandAlphaOperands0.
        TextureCombinerOperandKind    OperandAlpha1;       // OperandAlphaOperands1.
    };

    // Have the model name list send from the viewer side
    struct MESSAGEPACKET_BoneLinkage_ModelNameList
    {
        static const MessageType MessagePacketType = MessageType_BoneLinkage_ModelNameList;    // Message types.
        static const int StructSize = 8 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_PanelID = 1<<0,
            PropertyFlag_Count = 1<<1,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Uint    PanelID;       // ID to determine the sent panel.
        Uint    Count;       // Number of elements
    };

    // To start having the bone name list sent from the viewer side, when a packet is sent to the viewer, the bone list can be received.
    struct MESSAGEPACKET_BoneLinkage_BoneNameList
    {
        static const MessageType MessagePacketType = MessageType_BoneLinkage_BoneNameList;    // Message types.
        static const int StructSize = 12 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_PanelID = 1<<0,
            PropertyFlag_ModelNameHashID = 1<<1,
            PropertyFlag_Count = 1<<2,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        Uint    PanelID;       // ID to determine the sent panel.
        Uint    ModelNameHashID;       // Hash ID for the specified model name
        Uint    Count;       // Number of elements
    };

    // Send BoneLink information to the viewer
    struct MESSAGEPACKET_BoneLinkage_SetBoneLink
    {
        static const MessageType MessagePacketType = MessageType_BoneLinkage_SetBoneLink;    // Message types.
        static const int StructSize = 1104 + 4 + 4;    // Size as a structure.  The +4 is for ModifiedProperties and ModifiedPropertyElements.

        enum PropertyFlag
        {
            PropertyFlag_ObjectType = 1<<0,
            PropertyFlag_TargetType = 1<<1,
            PropertyFlag_BcptlNameHashID = 1<<2,
            PropertyFlag_ModelNameHashID = 1<<3,
            PropertyFlag_StartFrame = 1<<4,
            PropertyFlag_EmissionLength = 1<<5,
            PropertyFlag_FollowFlag = 1<<6,
            PropertyFlag_Scale = 1<<7,
            PropertyFlag_Rotate = 1<<8,
            PropertyFlag_Translate = 1<<9,
            PropertyFlag_EmitterName = 1<<10,
            PropertyFlag_BoneName = 1<<11,
            PropertyFlag_FORCE32 = 0xFFFFFFFF
        };

        Uint    ModifiedProperties;              // Modified properties.
        Uint    ModifiedPropertyElements;        // Modified property elements.

        BinaryDataType    ObjectType;       // SELECT: Bcptl Object type.
        EditTargetType    TargetType;       // SELECT: Emitter edit target type.
        Uint    BcptlNameHashID;       // SELECT: Hash of the binary name acquired from the Bcptl name.
        Uint    ModelNameHashID;       // BoneLink : Hash ID for the specified model name
        Int    StartFrame;       // Start frame
        Uint    EmissionLength;       // Inherited number of frames
        Uint    FollowFlag;       // FollowBitType ; Continue|Follow flag
        Vector3    Scale;       //
        Vector3    Rotate;       //
        Vector3    Translate;       //
        String    EmitterName;       // SELECT: Emitter Name  Edit target name.
        String    BoneName;       // Name of bone to constrain to.
    };

    // Specifies the object to edit.
    struct EditSelectedObject
    {
        static const int StructSize = 8;    // Size as a structure.

        Uint    Id;    // Returns the ID.
        Uint    Index;    // The index.
    };

    // Support for object that is the material
    struct MaterialBindObject
    {
        static const int StructSize = 8;    // Size as a structure.

        Uint    Id;    // Returns the ID.
        Uint    Index;    // The index.
    };

}
#endif
