/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_VIEWER_SYSTEM_ANM_CTRL_H_
#define NW4R_VIEWER_SYSTEM_ANM_CTRL_H_

#include <viewer/system/App.h>
#include <viewer/system/AppId.h>

#define USE_MESSAGE_QUEUE

namespace vw     {
namespace system {

enum AnmCmd
{
    ANM_CMD_RESET,       // Reset
    ANM_CMD_PLAY,        // Played back
    ANM_CMD_REV_PLAY,    // Reverse playback
    ANM_CMD_STOP,        // Stop
    ANM_CMD_STEP,        // Frame advance.
    ANM_CMD_REV_STEP,    // Reverse frame advance
    ANM_CMD_LOOP,        // Loop at 0th frame
    ANM_CMD_SET_FRAME    // Frame settings
};

enum AnmTgt
{
    ANM_TGT_ALL,         // All objects are targeted
    ANM_TGT_3D_MDL,      // Object related to specified 3D model
    ANM_TGT_3D_SCN,      // Object related to specified 3D scene
    ANM_TGT_2D_LYT       // Object related to specified 2D layout
};

enum AnmPlayArgFlags
{
    ANM_ARG_NONE            = 0,   // No particular specification
    ANM_ARG_FORCE_LOOP      = 0x1, // Force loop playback without any relation to the model animation information
    ANM_ARG_FORCE_ONETIME   = 0x2  // Force one-time playback without any relation to the model animation information
};

typedef struct
{
    f32 speed;      // Playback speed
    u32 loopArg;    // Forced loop parameter
} AnmPlayArg;

#if defined( USE_MESSAGE_QUEUE )
void  SysAnmCtrl_Send( u32 command,  u32 target, u32 targetID, void* arg, u32 trigModule, u32 arglen );
void  SysAnmCtrl_Exec( void );
#else
inline void
SysAnmCtrl_Send( u32 command,  u32 target, u32 targetID, void* arg, u32 trigModule, u32 arglen )
{
#pragma unused( arglen )
    if ( ( trigModule == vw::system::SYSTEM_ID )                      ||
         (vw::system::App::GetInstance().GetFocusID() == trigModule ) )
    // Send only when the focus module and trigger module are equivalent
    {
        App::GetInstance().OnAnmCmd( command, target, targetID, arg, trigModule );
    }
}

inline void
SysAnmCtrl_Exec( void ) {}

#endif /* defined( USE_MESSAGE_QUEUE ) */

/*---------------------------------------------------------------------------*
  Name:         F32ToPtr
  Description:  Converts f32 to void* type.
  Arguments:    val     f32 value to convert
  Returns:      None.
 *---------------------------------------------------------------------------*/
inline void*
F32ToPtr( f32 val )
{
    union
    {
        f32   fval;
        void* ptr;
    } data;
    data.fval = val;
    return data.ptr;
}

/*---------------------------------------------------------------------------*
  Name:         PtrToF32
  Description:  Converts void* to f32 type.
  Arguments:    ptr     void* value to convert
  Returns:      None.
 *---------------------------------------------------------------------------*/
inline f32
PtrToF32( void* ptr )
{
    union
    {
        f32   fval;
        void* ptr;
    } data;
    data.ptr = ptr;
    return data.fval;
}


}   /* namespace system */


#define SYS_ANM_CTRL_FUNC_PLAY( func, cmd, id ) \
    inline void SysAnmCtrl_##func##All( f32 speed = 1.0f, u32 arg = ::vw::system::ANM_ARG_NONE )                      \
    {                                                                                                                 \
        ::vw::system::AnmPlayArg param = { speed, arg };                                                              \
        ::vw::system::SysAnmCtrl_Send( cmd, ::vw::system::ANM_TGT_ALL, 0, (void*)&param, id, sizeof(param) );         \
    }                                                                                                                 \
    inline void SysAnmCtrl_##func( u32 target, u32 targetID, f32 speed = 1.0f, u32 arg = ::vw::system::ANM_ARG_NONE ) \
    {                                                                                                                 \
        ::vw::system::AnmPlayArg param = { speed, arg };                                                              \
        ::vw::system::SysAnmCtrl_Send( cmd, target, targetID, (void*)&param, id, sizeof(param) );                     \
    }


#define SYS_ANM_CTRL_FUNC( func, cmd, id )  \
    inline void SysAnmCtrl_##func##All() { ::vw::system::SysAnmCtrl_Send( cmd, ::vw::system::ANM_TGT_ALL, 0, NULL, id, 0 );  } \
    inline void SysAnmCtrl_##func( u32 target, u32 targetID ) { ::vw::system::SysAnmCtrl_Send( cmd, target, targetID, NULL, id, 0 ); }

#define SYS_ANM_CTRL_FUNC_F32( func, cmd, id )  \
    inline void SysAnmCtrl_##func##All( f32 arg = 0.0f ) { ::vw::system::SysAnmCtrl_Send( cmd, ::vw::system::ANM_TGT_ALL, 0, ::vw::system::F32ToPtr(arg), id, 0 );  } \
    inline void SysAnmCtrl_##func( u32 target, u32 targetID, f32 arg = 0.0f ) { ::vw::system::SysAnmCtrl_Send( cmd, target, targetID, ::vw::system::F32ToPtr(arg), id, 0 ); }

#define SYS_ANM_CTRL_FUNC_F32_1( func, cmd, id )  \
    inline void SysAnmCtrl_##func##All( f32 arg = 1.0f ) { ::vw::system::SysAnmCtrl_Send( cmd, ::vw::system::ANM_TGT_ALL, 0, ::vw::system::F32ToPtr(arg), id, 0 );  } \
    inline void SysAnmCtrl_##func( u32 target, u32 targetID, f32 arg = 1.0f ) { ::vw::system::SysAnmCtrl_Send( cmd, target, targetID, ::vw::system::F32ToPtr(arg), id, 0 ); }

namespace system {
namespace detail {
SYS_ANM_CTRL_FUNC      ( Reset,    ::vw::system::ANM_CMD_RESET,     SYSTEM_ID )    // SysAnmCtrl_Reset,    SysAnmCtrl_ResetAll
SYS_ANM_CTRL_FUNC_PLAY ( Play,     ::vw::system::ANM_CMD_PLAY,      SYSTEM_ID )    // SysAnmCtrl_Play,     SysAnmCtrl_PlayAll
SYS_ANM_CTRL_FUNC_PLAY ( RevPlay,  ::vw::system::ANM_CMD_REV_PLAY,  SYSTEM_ID )    // SysAnmCtrl_RevPlay,  SysAnmCtrl_RevPlayAll
SYS_ANM_CTRL_FUNC      ( Stop,     ::vw::system::ANM_CMD_STOP,      SYSTEM_ID )    // SysAnmCtrl_Stop,     SysAnmCtrl_StopAll
SYS_ANM_CTRL_FUNC_F32_1( Step,     ::vw::system::ANM_CMD_STEP,      SYSTEM_ID )    // SysAnmCtrl_Step,     SysAnmCtrl_StepAll
SYS_ANM_CTRL_FUNC_F32_1( RevStep,  ::vw::system::ANM_CMD_REV_STEP,  SYSTEM_ID )    // SysAnmCtrl_RevStep,  SysAnmCtrl_RevStepAll
SYS_ANM_CTRL_FUNC      ( Loop,     ::vw::system::ANM_CMD_LOOP,      SYSTEM_ID )    // SysAnmCtrl_Loop,     SysAnmCtrl_LoopAll
SYS_ANM_CTRL_FUNC_F32  ( SetFrame, ::vw::system::ANM_CMD_SET_FRAME, SYSTEM_ID )    // SysAnmCtrl_SetFrame, SysAnmCtrl_SetFrameAll
}   /* namespace detail */
}   /* namespace system */

namespace g3d {

SYS_ANM_CTRL_FUNC      ( Reset,    ::vw::system::ANM_CMD_RESET,     G3D_ID )    // SysAnmCtrl_Reset,    SysAnmCtrl_ResetAll
SYS_ANM_CTRL_FUNC_PLAY ( Play,     ::vw::system::ANM_CMD_PLAY,      G3D_ID )    // SysAnmCtrl_Play,     SysAnmCtrl_PlayAll
SYS_ANM_CTRL_FUNC_PLAY ( RevPlay,  ::vw::system::ANM_CMD_REV_PLAY,  G3D_ID )    // SysAnmCtrl_RevPlay,  SysAnmCtrl_RevPlayAll
SYS_ANM_CTRL_FUNC      ( Stop,     ::vw::system::ANM_CMD_STOP,      G3D_ID )    // SysAnmCtrl_Stop,     SysAnmCtrl_StopAll
SYS_ANM_CTRL_FUNC_F32_1( Step,     ::vw::system::ANM_CMD_STEP,      G3D_ID )    // SysAnmCtrl_Step,     SysAnmCtrl_StepAll
SYS_ANM_CTRL_FUNC_F32_1( RevStep,  ::vw::system::ANM_CMD_REV_STEP,  G3D_ID )    // SysAnmCtrl_RevStep,  SysAnmCtrl_RevStepAll
SYS_ANM_CTRL_FUNC      ( Loop,     ::vw::system::ANM_CMD_LOOP,      G3D_ID )    // SysAnmCtrl_Loop,     SysAnmCtrl_LoopAll
SYS_ANM_CTRL_FUNC_F32  ( SetFrame, ::vw::system::ANM_CMD_SET_FRAME, G3D_ID )    // SysAnmCtrl_SetFrame, SysAnmCtrl_SetFrameAll

}   /* namespace g3d */

namespace effect {

SYS_ANM_CTRL_FUNC      ( Reset,    ::vw::system::ANM_CMD_RESET,     EFFECT_ID ) // SysAnmCtrl_Reset,    SysAnmCtrl_ResetAll
SYS_ANM_CTRL_FUNC_PLAY ( Play,     ::vw::system::ANM_CMD_PLAY,      EFFECT_ID ) // SysAnmCtrl_Play,     SysAnmCtrl_PlayAll
SYS_ANM_CTRL_FUNC_PLAY ( RevPlay,  ::vw::system::ANM_CMD_REV_PLAY,  EFFECT_ID ) // SysAnmCtrl_RevPlay,  SysAnmCtrl_RevPlayAll
SYS_ANM_CTRL_FUNC      ( Stop,     ::vw::system::ANM_CMD_STOP,      EFFECT_ID ) // SysAnmCtrl_Stop,     SysAnmCtrl_StopAll
SYS_ANM_CTRL_FUNC_F32_1( Step,     ::vw::system::ANM_CMD_STEP,      EFFECT_ID ) // SysAnmCtrl_Step,     SysAnmCtrl_StepAll
SYS_ANM_CTRL_FUNC_F32_1( RevStep,  ::vw::system::ANM_CMD_REV_STEP,  EFFECT_ID ) // SysAnmCtrl_RevStep,  SysAnmCtrl_RevStepAll
SYS_ANM_CTRL_FUNC      ( Loop,     ::vw::system::ANM_CMD_LOOP,      EFFECT_ID ) // SysAnmCtrl_Loop,     SysAnmCtrl_LoopAll
SYS_ANM_CTRL_FUNC_F32  ( SetFrame, ::vw::system::ANM_CMD_SET_FRAME, EFFECT_ID ) // SysAnmCtrl_SetFrame, SysAnmCtrl_SetFrameAll

}   /* namespace effect */

namespace g2d {

SYS_ANM_CTRL_FUNC      ( Reset,    ::vw::system::ANM_CMD_RESET,     G2D_ID )   // SysAnmCtrl_Reset,    SysAnmCtrl_ResetAll
SYS_ANM_CTRL_FUNC_PLAY ( Play,     ::vw::system::ANM_CMD_PLAY,      G2D_ID )   // SysAnmCtrl_Play,     SysAnmCtrl_PlayAll
SYS_ANM_CTRL_FUNC_PLAY ( RevPlay,  ::vw::system::ANM_CMD_REV_PLAY,  G2D_ID )   // SysAnmCtrl_RevPlay,  SysAnmCtrl_RevPlayAll
SYS_ANM_CTRL_FUNC      ( Stop,     ::vw::system::ANM_CMD_STOP,      G2D_ID )   // SysAnmCtrl_Stop,     SysAnmCtrl_StopAll
SYS_ANM_CTRL_FUNC_F32_1( Step,     ::vw::system::ANM_CMD_STEP,      G2D_ID )   // SysAnmCtrl_Step,     SysAnmCtrl_StepAll
SYS_ANM_CTRL_FUNC_F32_1( RevStep,  ::vw::system::ANM_CMD_REV_STEP,  G2D_ID )   // SysAnmCtrl_RevStep,  SysAnmCtrl_RevStepAll
SYS_ANM_CTRL_FUNC      ( Loop,     ::vw::system::ANM_CMD_LOOP,      G2D_ID )   // SysAnmCtrl_Loop,     SysAnmCtrl_LoopAll
SYS_ANM_CTRL_FUNC_F32  ( SetFrame, ::vw::system::ANM_CMD_SET_FRAME, G2D_ID )   // SysAnmCtrl_SetFrame, SysAnmCtrl_SetFrameAll

}   /* namespace g3d */

namespace sound {

SYS_ANM_CTRL_FUNC      ( Reset,    ::vw::system::ANM_CMD_RESET,     SOUND_ID ) // SysAnmCtrl_Reset,    SysAnmCtrl_ResetAll
SYS_ANM_CTRL_FUNC_PLAY ( Play,     ::vw::system::ANM_CMD_PLAY,      SOUND_ID ) // SysAnmCtrl_Play,     SysAnmCtrl_PlayAll
SYS_ANM_CTRL_FUNC_PLAY ( RevPlay,  ::vw::system::ANM_CMD_REV_PLAY,  SOUND_ID ) // SysAnmCtrl_RevPlay,  SysAnmCtrl_RevPlayAll
SYS_ANM_CTRL_FUNC      ( Stop,     ::vw::system::ANM_CMD_STOP,      SOUND_ID ) // SysAnmCtrl_Stop,     SysAnmCtrl_StopAll
SYS_ANM_CTRL_FUNC_F32_1( Step,     ::vw::system::ANM_CMD_STEP,      SOUND_ID ) // SysAnmCtrl_Step,     SysAnmCtrl_StepAll
SYS_ANM_CTRL_FUNC_F32_1( RevStep,  ::vw::system::ANM_CMD_REV_STEP,  SOUND_ID ) // SysAnmCtrl_RevStep,  SysAnmCtrl_RevStepAll
SYS_ANM_CTRL_FUNC      ( Loop,     ::vw::system::ANM_CMD_LOOP,      SOUND_ID ) // SysAnmCtrl_Loop,     SysAnmCtrl_LoopAll
SYS_ANM_CTRL_FUNC_F32  ( SetFrame, ::vw::system::ANM_CMD_SET_FRAME, SOUND_ID ) // SysAnmCtrl_SetFrame, SysAnmCtrl_SetFrameAll

}   /* namespace sound */


}   /* namespace vw     */



/* NW4R_VIEWER_SYSTEM_ANM_CTRL_H_ */
#endif
