/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4R_VIEWER_SYSTEM_CONFIG_FILE_H_
#define NW4R_VIEWER_SYSTEM_CONFIG_FILE_H_

#include <viewer/common.h>
#include <nw/ut/ut_LinkList.h>
//#include <viewer/system/FileBrowseWindow.h>

using namespace vw::common;

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw     {
namespace system {

/*---------------------------------------------------------------------------*
  Name:         <CODE>ConfigNode</CODE>
  Description:  The Config node class.
 *---------------------------------------------------------------------------*/
class ConfigNode
{
public:
    /* ctor */          ConfigNode( const char* name, const char* value = NULL );       // Constructor
    /* dtor */         ~ConfigNode(){mName=CharArray(); mValue=CharArray();}            // Destructor
    const char*         GetValue( void )                { return mValue.Get(); }        // Get value
    void                SetValue( const char* value )   { mValue = value; }             // Set value

    const char*         GetName() const                 { return mName.Get();  }        // Get name
    bool                HasName() const                 { return (this->GetName() != NULL); }

    s32                 CompareName( const char* name );                                // Check if matches the name string

    s32                 Dump( char* buf = NULL );                                       // Dump output.

    nw::ut::LinkListNode mLink;

private:
    CharArray           mName;                                                          // Name string
    CharArray           mValue;                                                         // Value string
};

#if 0
class ConfigNode;
class MessageDialog;
/*---------------------------------------------------------------------------*
  Name:         ConfigFileExporter
  Description:  Configuration file exporter
 *---------------------------------------------------------------------------*/
class ConfigFileExporter : public Newable
{
public:
    /* ctor */  ConfigFileExporter() : mState( STATE_IDLE ), mFilter("ini"), mDialogWindow(NULL) {}
    bool        IsWorking() const { return (mState != STATE_IDLE); }
    void        StartImport();
    void        StartExport();
    void        Cancel();
    void        Update();
    void        StartSave( const char* filename );
    void        StartLoad( const char* filename );
private:
    static const u32 MAX_PATH = 260;

    enum
    {
        STATE_IDLE,
        STATE_SELECTING,
        STATE_IMPORTING,
        STATE_EXPORTING
    };
    u32             mState;
    FileFilter      mFilter;
    MessageDialog*  mDialogWindow;  // Message window
    char            mFilePath[ MAX_PATH ];

    static void     ImportFileSelectClbk_( u32 result, const char* filename, void* arg );
    static void     ExportFileSelectClbk_( u32 result, const char* filename, void* arg );

    void        CreateMsgDialog_( const char* title, const char* message );
};
#endif

/*---------------------------------------------------------------------------*
  Name:         ConfigFile
  Description:  The Settings file interface.
 *---------------------------------------------------------------------------*/
class ConfigFile
{
public:
    static ConfigFile*      Create();                                           // Create configuration file object
    static ConfigFile*      GetInstance() { return sInstance; }
    static void             Destroy( void );                                    // Destroy configuration file object

    static bool             IsReloaded()    { return sReloadedTrig;  }          // Confirm configuration file update
    static void             UpdateReloaded() { sReloadedTrig = sReloadedNext; sReloadedNext = false; } // Update reload flag

    bool                    ReadFromMEM( void* data, u32 size );                // Load from memory
    bool                    ReadFromRom( const char* filePath );                // Read from DVD
    bool                    ReadFromHost( const char* filePath );               // Read from MCS
    void                    Write( const char* filePath );                      // Write
    void                    Reload( const char* filePath );                     // Reload configuration file
    void                    Reload( void* data, u32 size );                     // Reload configuration file

    const char*             GetString( const char* path );                      // Get the value as a string
    void                    SetString( const char* path, const char* value );   // Set the string value
    void                    SetComment( const char* comment );

    bool                    GetInt   ( const char* path, s32* pValue );         // Get the value as int type
    bool                    GetUInt  ( const char* path, u32 *pValue );         // Get the value as uint type
    void                    SetInt   ( const char* path, s32 value );           // Set a decimal integer value
    void                    SetUInt  ( const char* path, u32 value );           // Set an unsigned decimal integer value
    void                    SetInt16 ( const char* path, s32 value );           // Set a hexadecimal integer value
    void                    SetUInt16( const char* path, u32 value );           // Set an unsigned hexadecimal integer value

    bool                    GetFloat( const char* path, f32* pValue );          // Get the value as a floating point number
    void                    SetFloat( const char* path, f32 value );            // Set a floating point number value

    void                    Remove( const char* path ) { SetString( path, NULL ); }    // Delete
    void                    Clear();                                            // Clear
    const char*             GetDefaultFilePath() { return mDefaultFilePath.Get(); }

    //ConfigFileExporter&       GetExporter()       { return mExporter; }
    //const ConfigFileExporter& GetExporter() const { return mExporter; }

    static void             SetConfigFilePath( const char* filePath );
    static const char*      GetConfigFilePath();

private:

    static bool             sReloadedNext;                                      // Flag for determining to reload next frame
    static bool             sReloadedTrig;                                      // Flag for determining reload
    static ConfigFile*      sInstance;
    /* ctor */              ConfigFile();                                       // Constructor
    /* dtor */              ~ConfigFile();                                      // Destructor

    nw::ut::LinkList<ConfigNode, offsetof( ConfigNode, mLink )> mChildList;     // Child node
    // ConfigFileExporter      mExporter;                                          // Configuration file exporter
    bool                    mExistsFile;                                        // Flag for whether the configuration file exists

    void                    Parse_( char* fileBuf, s32 size );                  // Parse the configuration file
    s32                     DumpNode_( const char* path, char* buf );           // Output a node

    ConfigNode*             SeekNode ( const char* path );                      // Search for a child node
    ConfigNode*             CreateNode( const char* path );                     // Create a node
    void                    AppendNode( ConfigNode* child );                    // Add a child node

    vw::common::CharArray   mDefaultFilePath;
    static vw::common::CharArray   sHostFilePath;
};

inline ConfigFile* GetConfigFile()
{
    return ConfigFile::GetInstance();
}



}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif



/* NW4R_VIEWER_SYSTEM_CONFIG_FILE_H_ */
#endif
