/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_SYSTEM_MODULE_H_
#define NW4C_VIEWER_SYSTEM_MODULE_H_

#include <viewer/system/App.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw     {
namespace system {

class App;

/*===========================================================================*
  Name:         Module
  Description:  Module class
 *===========================================================================
*/
class Module
{
//--------public members ------------------------
public:
    //NW4R_UT_RUNTIME_TYPEINFO;

    static Module*      GetTopModule( void ) { return sTopModule; }
    static Module*      GetModuleById( u16 id );
    static Module*      GetInstance() { return NULL; }

    /* ctor */          Module( u16 id, u32 priority, u32 initOrder = DEFAULT_SETUP_ORDER );
    /* dtor */          ~Module();

    // Initialize.
    virtual void        Init( void ) = 0;

    // Perform communication processing. Measurement of the process meter begins after these processes complete
    virtual void        PollStream( void ) {}

    // Main routine pre-processing (Perform issuing GX commands, etc.)
    virtual void        PreMain( HidStatus &hidStatus ) {}

    // Main routine
    virtual void        Main( HidStatus &hidStatus ) = 0;

    // Update the animation frame
    virtual void        UpdateFrame( void ) {}

    // Calculation routine
    virtual void        Calc( void ) {}

    // Render routine
    virtual void        Draw( void ) {}

    // Finalization (called during reset)
    virtual void        Destroy( void ) {}

    // Process called when focus moves
    virtual void        OnModuleFocused( void ) {}

    // Process called when focus is removed
    virtual void        OnModuleUnfocused( void ) {}

    // Command for playback all is executed
    virtual void        OnAnmCmd( u32   /* command    */,     // Command (RESET/PLAY/REV_PLAY/STOP/STEP/REV_STEP/LOOP/SET_FRAME)
                                  u32   /* target     */,     // Target (ALL/3D_MDL/2D_LYT)
                                  u32   /* targetID   */,     // Target ID (do not use when target is ANM_TGT_ALL)
                                  void* /* arg        */,     // Option parameters
                                  u32   /* trigModule */ ) {} // Module ID that is the trigger

    // Information render routine
    virtual void        DrawInfo( void ) = 0;

    // Write back handler for configuration file
    virtual void        WriteBackConfigFile() {}

    void                ChangePriority( u32 priority );
    Module*             GetNext()       { return mNext; }
    virtual u16         GetId()   const { return mId; }

    //Heap&               GetHeap() { return mHeap;   }
    //VWPad&              GetPad()  { return mVWPad;    }
    bool                HasFocus() const {
        App *app;//::GetInstance();
        return false;
    }

    virtual bool        Available( void ) = 0;
    virtual bool        IsHidden( void ) { return false; }  // Focus does not return if this return value is TRUE
    virtual bool        IsPerformVisible() const { return true; } // Performance meter does not display if this return value is FALSE
    virtual bool        IsIgnoringChangeMode()    const { return false; } // Skip during mode changes if this return value is TRUE
    virtual bool        IsIgnoringChangeVisible() const { return false; } // Information hide switching does not work if this return value is TRUE

    // Callback settings
    typedef void (*CallbackFunc)( u32 id, void* arg );
    void                SetPostCalcCallback( u32 id, CallbackFunc callback, void* arg = NULL ) { mPostCalcClbk.Set( id, callback, arg ); }
    void                SetPostDrawCallback( u32 id, CallbackFunc callback, void* arg = NULL ) { mPostDrawClbk.Set( id, callback, arg ); }
    void                ClearPostCalcCallback( u32 id ) { mPostCalcClbk.Clear( id ); }
    void                ClearPostDrawCallback( u32 id ) { mPostDrawClbk.Clear( id ); }
    void                RisePostCalcCallback() { mPostCalcClbk.Rise(); }
    void                RisePostDrawCallback() { mPostDrawClbk.Rise(); }
    void                RisePostCalcCallback( u32 targetID ) { mPostCalcClbk.Rise( targetID ); }
    void                RisePostDrawCallback( u32 targetID ) { mPostDrawClbk.Rise( targetID ); }

    static Module*      GetFirstSetupModule() { return sFirstSetupModule; }
    Module*             GetNextSetup()        { return mNextSetup;        }

//--------Private members ------------------------
private:
    static  Module*     sTopModule;     // Pointer to the start of the module list structure

    u16                 mId;            // ID
    u32                 mPriority;      // Priority
    u32                 mSetupOrder;    // Initialization order
    Module*             mNext;          // Pointer to the next module

    static Module*      sFirstSetupModule; // Pointer to the first module of the initialization order
    Module*             mNextSetup;        // Pointer to the next module in the initialization order

    //Heap                mHeap;          // Heap library wrapper class
    //VWPad               mVWPad;           // Pad library wrapper class

    void                Insert( void );
    void                Remove( void );

    void                InsertToList_();
    void                InsertToSetupList_();

    // Callback management
    class ClbkMgr
    {
    public:
        typedef void (*Func)( u32 id, void* arg );
        void        Set( u32 id, Func func, void* arg = NULL );
        void        Clear( u32 id ) { this->Set( id, NULL, NULL ); }
        void        Rise();
        void        Rise( u32 targetID );
        /* ctor */  ClbkMgr( u32 id ) : mNum( 0 ), mId( id )
        {
            for ( u32 i = 0; i < MODULE_NUM; i++ ) { mInfo[ i ].mFunc = NULL; mInfo[ i ].mArg = NULL; }
        }
    private:
        u32         mNum;
        u32         mId;
        struct
        {
            Func    mFunc;
            void*   mArg;
        } mInfo[ MODULE_NUM ];
    };

    ClbkMgr       mPostCalcClbk;
    ClbkMgr       mPostDrawClbk;
};


}   /* namespace system */
}   /* namespace vw     */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif



/* NW4C_VIEWER_SYSTEM_MODULE_H_ */
#endif
