/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <cstring>
#include <nw/os/os_Memory.h>
#include <viewer/common.h>
//#include <viewer/system/AppId.h>
//#include <viewer/system/New.h>
#include <viewer/system.h>
#include <nw/ut.h>

//using namespace vw::system;
//using namespace nw4r;

namespace vw     {
namespace common {


/*---------------------------------------------------------------------------*
  Name:         CharArray::CharArray
  Description:  Constructor
  Arguments:    allocator Specify allocator
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
CharArray::CharArray( nw::os::IAllocator* allocator /* = NULL */ )
 : mAllocator( allocator ),
   mStr( NULL ),
   mLen( 0 )
{
}



/*---------------------------------------------------------------------------*
  Name:         CharArray::CharArray
  Description:  Constructor
  Arguments:    str       String to set
                allocator Specify allocator
  Returns:      None.
 *---------------------------------------------------------------------------*/
/* ctor */
CharArray::CharArray( const char* str, nw::os::IAllocator* allocator /* = NULL */ )
 : mAllocator( allocator ),
   mStr( NULL ),
   mLen( 0 )
{
    Set( str );
}


/*---------------------------------------------------------------------------*
  Name:         CharArray::CharArray
  Description:  Constructor (with size specification)
  Arguments:    str       String to set
                len       Character size
                allocator Specify allocator
  Returns:      None.
 *---------------------------------------------------------------------------*/
/* ctor */
CharArray::CharArray( const char* str, u32 len, nw::os::IAllocator* allocator /* = NULL */ )
 : mAllocator( allocator ),
   mStr( NULL ),
   mLen( 0 )
{
    Set( str, len );
}

/*---------------------------------------------------------------------------*
  Name:         CharArray::CharArray
  Description:  Constructor
  Arguments:    str       String to set
                allocator Specify allocator
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
CharArray::CharArray( const CharArray& str, nw::os::IAllocator* allocator /* = NULL */ )
 : mAllocator( allocator ),
   mStr( NULL ),
   mLen( 0 )
{
    Set( str.mStr );
}

/*---------------------------------------------------------------------------*
  Name:         CharArray::~CharArray
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
/* dtor */
CharArray::~CharArray()
{
    if ( mStr != NULL )
    {
        this->Free_( mStr );
    }
}

/*---------------------------------------------------------------------------*
  Name:         CharArray::Free_
  Description:
  Arguments:    str
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
CharArray::Free_( char* str )
{
    if ( mAllocator )
    {
        mAllocator->Free( str );
    }
    else
    {
        vw::system::Memory_Free( str );
    }
}


/*---------------------------------------------------------------------------*
  Name:         CharArray::Alloc_
  Description:
  Arguments:    size
  Returns:
 *---------------------------------------------------------------------------


*/
char*
CharArray::Alloc_( u32 size )
{
    if ( mAllocator )
    {
        void* buf = mAllocator->Alloc( size );
        return reinterpret_cast<char*>( buf );
    }
    else
    {
        void* buf = vw::system::Memory_AllocSys( size );
        return reinterpret_cast<char*>( buf );
    }
}

/*---------------------------------------------------------------------------*
  Name:         CharArray::Set
  Description:  Set string
  Arguments:    str:    Pointer to the character string
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
CharArray::Set( const char* str )
{
    u32 len = 0;

    if ( str )
    {
        len = std::strlen( str );
    }
    Set( str, len );
}


/*---------------------------------------------------------------------------*
  Name:         CharArray::Set
  Description:  Sets string (with size specification)
  Arguments:    str:    Pointer to the character string
                len     Size of character to set
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
CharArray::Set( const char* str, u32 len )
{
    // Deallocate character buffer when str is NULL
    if ( str == NULL )
    {
        NW_ASSERT( len == 0 );
        if ( mStr != NULL )
        {
            this->Free_( mStr );
            mStr = NULL;
        }
        mLen = 0;
        return;
    }

    u32 strLen = nw::ut::RoundUp( len + 1, 32 );

    if ( mStr != NULL && strLen > mLen )
    {
        this->Free_( mStr );
        mStr = NULL;
        mLen = 0;
    }
    if ( mStr == NULL )
    {
        mStr = this->Alloc_( strLen );
        mLen = strLen;
    }
    (void)std::strncpy( mStr, str, len );
    mStr[ len ] = '\0';
}


/*---------------------------------------------------------------------------*
  Name:         CharArray::Cat
  Description:  Adds string
  Arguments:    str:    Pointer to the character string
  Returns:      None.
 *---------------------------------------------------------------------------*/
const char*
CharArray::Cat( const char* str )
{
    if ( ! str )
    {
        return Get();
    }

    u32 len = std::strlen( str );

    return Cat( str, len );
}


/*---------------------------------------------------------------------------*
  Name:         CharArray::Cat
  Description:  Adds string (with size specification)
  Arguments:    str:    Pointer to the character string
                len     Size of character to set
  Returns:      None.
 *---------------------------------------------------------------------------*/
const char*
CharArray::Cat( const char* str, u32 len )
{
    if ( ! str )
    {
        return Get();
    }

    u32 strLen = len + 1;
    if ( mStr )
    {
        strLen += std::strlen( mStr ) + len + 1;
    }

    if ( strLen <= mLen )
    {
        // Use the buffer as is when the buffer size is insufficient
        return std::strncat( mStr, str, len );
    }

    // When the buffer size is insufficient, after allocating a new buffer and copying data, deallocate old buffer.
    // 
    strLen = nw::ut::RoundUp( strLen, 32 );

    char* newBuf = this->Alloc_( strLen );

    if ( mStr )
    {
        vw::common::strlcpy( newBuf, mStr, strLen );
        this->Free_( mStr );
    }
    else
    {
        newBuf[ 0 ] = '\0';
    }
    mStr = newBuf;
    mLen = strLen;

    return std::strncat( mStr, str, len );
}



}   /* namespace common */
}   /* namespace vw     */

