/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

//#include <dolphin.h>
#include <cctype>
#include <nw/math.h>
#include <nw/os/os_Memory.h>
#include <viewer/common.h>

namespace vw     {
namespace common {

//using namespace nw4r;

/*---------------------------------------------------------------------------*
  Name:         strlcpy

  Description:  strcpy without buffer overrun

  Arguments:    dst
                src
                bufSize

  Returns:      None.
 *---------------------------------------------------------------------------*/
u32
strlcpy( char* dst, const char* src, u32 bufSize )
{
    char* p = dst;
    char* terminal = dst + bufSize - 1;

    while ( *src != '\0' && p != terminal )
    {
        *p++ = *src++;
    }
    *p = '\0';

    return (u32)p - (u32)dst;
}

/*---------------------------------------------------------------------------*
  Name:         strlcat

  Description:  strcat without buffer overrun

  Arguments:    dst
                src
                bufSize

  Returns:      None.
 *---------------------------------------------------------------------------*/
u32
strlcat( char* dst, const char* src, u32 bufSize )
{
    char* p        = dst;
    char* terminal = dst + bufSize - 1;

    while ( *p != '\0' && p != terminal )
    {
        ++p;
    }

    while ( *src != '\0' && p != terminal )
    {
        *p++ = *src++;
    }
    *p = '\0';

    return (u32)p - (u32)dst;
}


/*---------------------------------------------------------------------------*
  Name:         chrcmpi

  Description:  Function to compare characters without distinguishing upper/lower case

  Arguments:    chr     Character 1
                ch2     Character 2

  Returns:
 *---------------------------------------------------------------------------*/
static inline u32
chr2val( const char ch )
{
    u32 val = u32( std::tolower( ch ) );

    // Symbol => Number=> Alphabet order
    if ( '0' <= val && val <= '9' )
    {
        val += 0x100;
    }
    if ( 'a' <= val && val <= 'z' )
    {
        val += 0x200;
    }
    return val;
}

/*---------------------------------------------------------------------------*
  Name:         strcmpi

  Description:  Function to compare strings without distinguishing upper/lower case

  Arguments:    str1     String 1
                str2    String 2

  Returns:      Convert to lower case characters and compare, and if the character values are:
                str1 > str2, TRUE
                str1 = str2, 0
                str1 < str2, FALSE
 *---------------------------------------------------------------------------*/
s32
strcmpi( const char* str1, const char* str2 )
{
    while ( std::tolower(*str1) == std::tolower(*str2) )
    {
        if ( *str1 == '\0' ) { return 0; }
        ++str1; ++str2;
    }
    return ( chr2val( *str1 ) > chr2val( *str2 ) ) ? 1 : -1;
}


/*---------------------------------------------------------------------------*
  Name:         strncmpi

  Description:  Function to compare strings without distinguishing upper/lower case (with specified maximum number of characters to search)

  Arguments:    str1     String 1
                str2    String 2
                num     number of characters to compare

  Returns:      Convert to lower case characters and compare, and if the character values are:
                str1 > str2, TRUE
                str1 = str2, 0
                str1 < str2, FALSE
 *---------------------------------------------------------------------------*/
s32
strncmpi( const char* str1, const char* str2, u32 num )
{
    if ( num == 0 ) { return 0; }

    while ( *str1 == *str2 )
    {
        if ( *str1 == '\0' ) { return 0; }
        if ( --num == 0 ) { return 0; }
        ++str1; ++str2;
    }
    return ( (u8)std::tolower( u8(*str1) ) > (u8)std::tolower( u8(*str2) ) ) ? 1 : -1;
}


/*
 * NOTE:
 *   Plan to replace with math library later
 *
 */

/*-------------------------------------------------------
  Name:         ArcSin

  Description:  Function to calculate ArcSin

  Arguments:    sin: sin value

  Returns:      Angle with 2 pi as 65536 (-0x4000 to 0x4000)

  -------------------------------------------------------*/
static s16
ArcSin( f32 sin )
{
    u16  neg;
    f32  sinAbs, sinTmp;
    s16  top, bottom, angle;
    enum { MAX_ANGLE = 0x4000 };

    if ( sin >= 0 )
    {
        sinAbs = sin;
        neg = 0;
    }
    else
    {
        sinAbs = -sin;
        neg = 1;
    }

    if ( sinAbs == 0 )
    {
        return 0;
    }
    else if ( sinAbs == 1.0f )
    {
        if ( neg )
        {
            return -MAX_ANGLE;
        }
        else
        {
            return MAX_ANGLE;
        }
    }

    // Get angle with binary search
    bottom = 0;
    top = MAX_ANGLE;
    while ( top - bottom > 1 )
    {
        angle = (s16)( (top + bottom) / 2 );
        sinTmp = nw::math::SinIdx( u16(angle) );
        if ( sinTmp > sinAbs )
        {
            top = angle;
        }
        else if ( sinTmp < sinAbs )
        {
            bottom = angle;
        }
        else
        {
            if ( neg )
            {
                angle = (s16)( -angle );
            }
            return angle;
        }
    }

    if ( neg )
    {
        angle = (s16)(-angle);
    }
    return angle;
}



/*-------------------------------------------------------
  Name:         GetAngle

  Description:  Function to get the angle from the sin and cos values

  Arguments:    sin: sin value
                cos: cos value

  Returns:      Angle using 65536 with 2 pi

  -------------------------------------------------------*/
u16
GetAngle( f32 sin, f32 cos )
{
    enum { PI = 0x8000 };
    u16 angle;

//    angle = F32ToU16( NW4R_fAsin( sin ) );
    angle = static_cast<u16>( ArcSin( sin ) );

    if ( cos < 0 )
    {
        angle = (u16)(PI - angle);
    }

    return angle;
}

}   /* namespace system */
}   /* namespace vw     */

