/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/AnimManager.h"
#include "../../g3d/include/SceneElemSelector.h"
#include "../../g3d/include/AnimManager.h"
#include "../../system/include/HidManager.h"
#include "../../g3d/include/SceneManager.h"

namespace vw  {
namespace effect {

#define FOREACH_PARTICLES( OPERATIONS )\
{\
    vw::g3d::SceneManager::ParticleModelList::iterator it = vw::g3d::SceneManager::GetInstance()->GetParticleModelsBegin();\
    vw::g3d::SceneManager::ParticleModelList::iterator end_it = vw::g3d::SceneManager::GetInstance()->GetParticleModelsEnd();\
    for (; it!=end_it; ++it)\
    {\
        {\
            vw::g3d::IAnimController *controller = it->GetParticleAnimController();\
            OPERATIONS;\
        }\
        {\
            vw::g3d::IAnimController *controller = it->GetAnimController();\
            OPERATIONS;\
        }\
    }\
}

AnimManager* AnimManager::m_pInstance = NULL;

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
AnimManager*
AnimManager::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
AnimManager::AnimManager()
    : m_Frame(0.f)
    , m_PrevFrame(0.f)
    , m_Step(1.f)
    , m_StartFrame(0)
    , m_IsActive(true)
    , m_MaxFrameCount( 0.f )
    , m_LoopFlag( false )
    , m_DefaultRate( 0.0f )
{
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
AnimManager::~AnimManager()
{
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(AnimManager));
    if (buf == NULL) return;

    m_pInstance = new(buf) AnimManager();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~AnimManager();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::UpdateFrame()
{
    if (system::App::GetInstance().GetFocusID() == system::EFFECT_ID)
    {
        m_PrevFrame = m_Frame;
        m_Frame += m_Step;

        // Perform loop process when the largest frame and loop flag are set.
        if (m_LoopFlag && m_MaxFrameCount > 0.0f)
        {
            if ( m_Frame >= m_MaxFrameCount )
            {
                m_Frame = 0;

                this->Reset();
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::UpdateAll()
{
    if (m_IsActive)
    {
        UpdateFrame();

        FOREACH_PARTICLES( controller->Update() );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::Synchronize()
{
    m_IsActive = false;

    m_Frame = 0;

    FOREACH_PARTICLES(
        controller->SetStepFrame( m_Step );
        controller->Stop();
    );

    vw::g3d::AnimManager::GetInstance()->StopAll();
    vw::g3d::AnimManager::GetInstance()->SetStepFrameAll( m_Step );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::Reset()
{
    m_Frame = m_StartFrame;
    FOREACH_PARTICLES( controller->Reset() );
    vw::g3d::AnimManager::GetInstance()->SetFrame(m_StartFrame);
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::StopAll()
{
    m_IsActive = false;
    m_Frame = 0;
    m_PrevFrame = 0;
    FOREACH_PARTICLES( controller->Stop() );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::PlayAll()
{
    m_IsActive = true;
    if (m_Frame < m_StartFrame)
    {
        Reset();
    }
    FOREACH_PARTICLES( controller->Play() );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::PauseAll()
{
    m_IsActive = false;
    FOREACH_PARTICLES( controller->Pause() );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::SpeedUpAll(f32 diff)
{
    m_Step += diff;
    if (m_Step<0.f) m_Step = 0.f;
    FOREACH_PARTICLES( controller->SetStepFrame(m_Step) );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::SetStepFrame(f32 step)
{
    m_Step = step > 0.f ? step : 0.f;
    FOREACH_PARTICLES( controller->SetStepFrame(m_Step) );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::StepForwardAll()
{
    m_IsActive = false;
    UpdateFrame();
    FOREACH_PARTICLES( controller->StepForward() );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::StepBackwardAll()
{
    m_IsActive = false;
    FOREACH_PARTICLES( controller->StepBackward() );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::OnTargetModuleChanged( u16 id )
{
    if (system::EFFECT_ID == id)
    {
        m_IsActive = false;
        m_Frame = 0;
        m_PrevFrame = 0;

        FOREACH_PARTICLES(
            controller->OnTargetModuleChanged(id);
            controller->Stop();
        );

        vw::g3d::AnimManager::GetInstance()->StopAll();
    }
    else
    {
        FOREACH_PARTICLES(
            controller->OnTargetModuleChanged(id);
        );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::SetStartFrame(s32 frame)
{
    m_StartFrame = frame;
    if (m_Frame < m_StartFrame)
    {
        Reset();
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
AnimManager::HidControl(system::HidStatus &hidStatus)
{
    static s32 status = vw::system::G3D_NONE;

    // Get current status
    s32 currentStatus = system::GetPadEvent( hidStatus, vw::system::GetEffectPadBindings() );
    if (currentStatus != system::G3D_NONE) status = currentStatus;

    // Process during scene update frame
    // For the purpose of frame rate synchronization when advancing the frame when the frame rate is not 60 Hz
    if (system::App::GetInstance().IsUpdateScene())
    {
        switch ( status )
        {
        case system::EF_SLOW_DOWN_ALL_ANIMATIONS:
            // Slow the animation playback.
            SpeedUpAll( -0.1f );
            vw::g3d::AnimManager::GetInstance()->SetStepFrameAll( m_Step );
            break;
        case system::EF_SPEED_UP_ALL_ANIMATIONS:
            // Speed up the animation playback.
            SpeedUpAll( 0.1f );
            vw::g3d::AnimManager::GetInstance()->SetStepFrameAll( m_Step );
            break;
        case system::EF_INCREMENT_ALL_ANIMATION_FRAMES:
            // Stop the animation, and advance one frame.
            StepForwardAll();
            vw::g3d::AnimManager::GetInstance()->StepForwardAll();
            break;
        case system::EF_STOP_ALL_ANIMATIONS:
            // Stop all model and effect animations.
            StopAll();
            vw::g3d::AnimManager::GetInstance()->StopAll();
            break;
        case system::EF_START_ALL_ANIMATIONS:
            // Play all model and effect animations.
            if (m_IsActive)
            {
                PauseAll();
                vw::g3d::AnimManager::GetInstance()->PauseAll();
            }
            else
            {
                PlayAll();
                vw::g3d::AnimManager::GetInstance()->PlayAll();
            }
            break;
        case system::EF_INCREMENT_MAX_FRAME:
            m_MaxFrameCount += 10;
            break;
        case system::EF_DECREMENT_MAX_FRAME:
            m_MaxFrameCount -= 10;
            if (m_MaxFrameCount < 0)
            {
                m_MaxFrameCount = 0;
            }
            break;
        case system::EF_CHANGE_ANIMATION_LOOP_STATUS:
            m_LoopFlag = !m_LoopFlag;
            break;
        default:
            break;
        }

        // Reset status
        status = vw::system::G3D_NONE;
    }
}

void
AnimManager::OnAnmCmd(u32 command, void* arg)
{
    switch (command)
    {
    case system::ANM_CMD_RESET:
        break;
    case system::ANM_CMD_PLAY:
        PlayAll();
        break;
    case system::ANM_CMD_REV_PLAY:
        break;
    case system::ANM_CMD_STOP:
        StopAll();
        break;
    case system::ANM_CMD_STEP:
        StepForwardAll();
        break;
    case system::ANM_CMD_REV_STEP:
        StepBackwardAll();
        break;
    case system::ANM_CMD_LOOP:
        break;
    case system::ANM_CMD_SET_FRAME:
        break;
    }
}

}   /* namespace g3d */
}   /* namespace vw  */

