/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G2D_ANIM_MANAGER_H_
#define NW4C_VIEWER_G2D_ANIM_MANAGER_H_

#include <nw/lyt/lyt_Layout.h>
#include <nw/lyt/lyt_Resources.h>
#include <nw/lyt/lyt_DrawInfo.h>
#include <nw/ut/ut_MoveArray.h>
#include <viewer/system.h>
//#include <viewer/system/STL.h>

namespace vw  {
namespace g2d {

//======================================
// Frame Management
//======================================
class FrameCtrl
{
public:
    FrameCtrl();

    //
    //
    //
    void StepFrame( f32 step, bool isLoop );

    //
    //
    //
    void SetFrame( f32 frame );

    //
    //
    //
    f32 GetMaxFrame() const
    { return mMaxFrame; }

    //
    //
    //
    void SetMaxFrame( f32 max );

    //
    //
    //
    void SetStepRate( f32 rate )
    { mStepRate = rate; }

    //
    //
    //
    f32 GetStepRate() const
    { return mStepRate; }

    //
    //
    //
    void SetLoop( bool loop )
    { mIsLoop = loop; }

    //
    //
    //
    bool IsLoop() const
    { return mIsLoop; }

    //
    //
    //
    bool IsStopAtEnd() const;

    //
    //
    //
    f32 GetCurrentFrame() const
    { return mFrame; }

    //
    //
    //
    f32 GetAbsoluteFrame() const
    { return mAbsoluteFrame; }

private:
    f32 mFrame;
    f32 mMaxFrame;
    f32 mAbsoluteFrame;
    f32 mStepRate;
    bool mIsLoop;
#if defined(NW_PLATFORM_CTR)
    u8 padding[3];
#endif
};

//======================================
// Animation management for layout
//======================================
class AnimCtrl : public FrameCtrl
{
public:
    AnimCtrl();

    //
    //
    //
    void Set(
        const char* name,
        nw::lyt::Layout* pLayout,
        const nw::lyt::AnimResource& animRes);

    //
    //
    //
    void Clear();

    //
    //
    //
    bool IsValid() const
    {
        return mName != NULL;
    }

    //
    //
    //
    const char* GetName() const
    {
        return mName;
    }

    //
    //
    //
    const nw::lyt::AnimResource& GetAnimResource() const
    { return mAnimRes; }

    //
    //
    //
    void StepFrame(f32 step, bool isLoop)
    {
        FrameCtrl::StepFrame(step, isLoop);
        this->FlushToAnim_();
    }

    //
    //
    //
    void SetFrame(f32 frame)
    {
        FrameCtrl::SetFrame(frame);
        this->FlushToAnim_();
    }

    //
    //
    //
    //
    //
    void SetEndFrame(bool bPrev);

//----- Private members --------
private:
    void SetMaxFrame(f32 max);
    void SetLoop(bool loop);
    void FlushToAnim_();

    f32 mFrame;
    const char* mName;
    nw::lyt::Layout* mpLayout;
    nw::lyt::AnimResource mAnimRes;
};

typedef nw::ut::MoveArray<AnimCtrl> AnimCtrlVec;

//======================================
// Animation management class
//======================================
class AnimManager
{
//----- Public members --------
public:
    AnimManager();

    //
    //
    //
    void Init();

    //
    //
    //
    void Clear();

    //
    //
    //
    void SetLayout(nw::lyt::Layout* layout);

    //
    //
    //
    void SetAnimNum(s32 num);

    //
    //
    //
    void RegistAnim(
        s32 idx,
        const char* name,
        const nw::lyt::AnimResource& animRes);

    //
    //
    //
    bool HasAnim() const
    { return ! mAnimCtrlVec.empty(); }

    //
    //
    //
    void Play()
    {
        if (this->HasAnim())
        {
            mPlayState = STATE_PLAY;
        }
    }

    //
    //
    //
    void Stop()
    { mPlayState = STATE_STOP; }

    //
    //
    //
    bool IsPlaying() const
    { return mPlayState == STATE_PLAY; }

    //
    //
    //
    bool IsStopAtEnd() const;

    //
    //
    //
    void SetFrame(f32 frame);

    //
    //
    //
    void Step( f32 step );

    //
    //
    //
    void Update();

    //
    //
    //
    void ResetAnim()
    {
        this->Stop();
        this->SetFrame(0.0f);
    }

    //
    //
    //
    f32 GetCurrentFrame() const;

    //
    //
    //
    // @details
    // An absolute frame is a frame that does not reset even if the animation is looped after the animation is reset.
    // 
    //
    f32 GetAbsoluteFrame() const;

    //
    //
    //
    f32 GetMaxFrame() const;

    //
    //
    //
    u32 GetAnimNum() const
    { return mAnimCtrlVec.size(); }

    //
    //
    //
    s32 GetAnimIdx() const
    { return mCrAnim; }

    //
    //
    //
    void SetAnim(s32 idx);

    //
    //
    //
    void ChangeAnimToNext();

    //
    //
    //
    void ChangeAnimToPrev();

    void SetWaitFrame() { mIsWaitOneFrame = true; }

    //
    //
    //
    const char* GetAnimName() const
    {
        return this->GetAnimCtrl()->GetName();
    }

    //
    //
    //
    enum TargetType
    {
        TargetType_Pane,
        TargetType_Group,
        TargetType_Max
    };

    //
    //
    //
    bool IsTargetSelectable() const;

    //
    //
    //
    //
    //
    bool SetTargetType(TargetType type);

    //
    //
    //
    void ChangeTargetTypeToPrev();

    //
    //
    //
    void ChangeTargetTypeToNext();

    //
    //
    //
    //
    TargetType GetTargetType() const
    { return mTargetType; }

    //
    //
    //
    const char* GetTargetTypeName() const;

    //
    //
    //
    const nw::lyt::Pane* GetTargetPane() const
    {
        s32 idx = mCrTarget[TargetType_Pane];

        if (idx <= 0)
        {
            return mPaneVec[0];
        }
        else
        {
            return mPaneVec[idx];
        }
    }

    //
    //
    //
    nw::lyt::Pane* GetTargetPane()
    {
        s32 idx = mCrTarget[TargetType_Pane];

        if (idx <= 0)
        {
            return mPaneVec[0];
        }
        else
        {
            return mPaneVec[idx];
        }
    }

    //
    //
    //
    const nw::lyt::Group* GetTargetGroup() const
    {
        s32 idx = mCrTarget[TargetType_Group];

        if (idx <= 0)
        {
            return mGroupVec[0];
        }
        else
        {
            return mGroupVec[idx];
        }
    }

    //
    //
    //
    nw::lyt::Group* GetTargetGroup()
    {
        s32 idx = mCrTarget[TargetType_Group];

        if (idx <= 0)
        {
            return mGroupVec[0];
        }
        else
        {
            return mGroupVec[idx];
        }
    }

    //
    //
    //
    const char* GetTargetName() const;

    //
    //
    //
    // @details
    // 0 displays the whole.
    //
    s32 GetTargetIdx() const
    {
        return mCrTarget[this->GetTargetType()];
    }

    //
    //
    //
    u32 GetTargetNum() const;

    //
    //
    //
    void SetTargetIdx(s32 idx);

    //
    //
    //
    void ChangeTargetToPrev();

    //
    //
    //
    void ChangeTargetToNext();

    //
    //
    //
    bool IsTargetAnimate()
    {
        return mTargetAnimate;
    }

    //
    //
    //
    bool IsTargetGroupInTag()
    {
        return mTargetType == TargetType_Group && mTargetGroupInTag;
    }

    //
    //
    //
    AnimCtrl* GetAnimCtrl(s32 idx)
    {
        if (idx < 0 || idx >= mAnimCtrlVec.size()) { return 0; }
        return &mAnimCtrlVec[std::size_t(idx)];
    }

    //
    //
    //
    const AnimCtrl* GetAnimCtrl( s32 idx ) const
    {
        if (idx < 0 || idx >= mAnimCtrlVec.size()) { return 0; }
        return &mAnimCtrlVec[std::size_t(idx)];
    }

    //
    //
    //
    AnimCtrl*GetAnimCtrl()
    {
        return GetAnimCtrl(mCrAnim);
    }

    //
    //
    //
    const AnimCtrl* GetAnimCtrl() const
    {
        return GetAnimCtrl(mCrAnim);
    }

//----- Private members --------
private:
    AnimCtrl& GetAnimCtrlRef(s32 idx)
    {
        NW_ASSERT(0 <= idx && idx < mAnimCtrlVec.size());
        return mAnimCtrlVec[std::size_t(idx)];
    }

    const AnimCtrl& GetAnimCtrlRef(s32 idx) const
    {
        NW_ASSERT(0 <= idx && idx < mAnimCtrlVec.size());
        return mAnimCtrlVec[std::size_t(idx)];
    }

    void SetAnimationEnable(bool enable);

    bool IsTargetAnimate(
        nw::lyt::Pane* pPane,
        const nw::lyt::AnimResource& animRes,
        bool bRecursive);

    nw::lyt::Layout* mpLayout;

    AnimCtrlVec mAnimCtrlVec;

    typedef nw::ut::MoveArray<nw::lyt::Pane*> PaneVec;
    PaneVec mPaneVec;

    typedef nw::ut::MoveArray<nw::lyt::Group*> GroupVec;
    GroupVec mGroupVec;

    enum
    {
        STATE_STOP,
        STATE_PLAY
    };

    u32 mPlayState;

    s32 mCrAnim;
    s32 mCrTarget[TargetType_Max];
    TargetType mTargetType;

    bool mIsWaitOneFrame;
    // Is the current target affected by the current animation?
    bool mTargetAnimate;
    // Is the target group registered in the animation interval tag?
    bool mTargetGroupInTag;
};


}   /* namespace g2d */
}   /* namespace vw  */


/* NW4C_VIEWER_G2D_ANIM_MANAGER_H_ */
#endif


