/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G2D_CAMERAMGR_H_
#define NW4C_VIEWER_G2D_CAMERAMGR_H_

#include <viewer/system.h>
#include <nn/math.h>
#include <nn/ulcd.h>
#include <nw/ut/ut_Rect.h>
#include <nw/lyt/lyt_Layout.h>

namespace vw  {
namespace g2d {

// Camera and projection management
class CameraMgr
{
//----- Public members -------------
public:
    /* ctor */  CameraMgr();

    void Init();

    f32         GetNear() const { return mNear; }
    void        SetNear( f32 near ) { mNear = near; }

    f32         GetMinNear() const { return mMinNear; }
    void        SetMinNear( f32 near ) { mMinNear = near; }

    f32         GetFar() const { return mFar; }
    void        SetFar( f32 far ) { mFar = far; }

    f32         GetFovy() const { return mFovy; }
    void        SetFovy( f32 fovy ) { mFovy = fovy; }

    bool        IsWide() const { return mbWideLocation; }
    void        SetWide( bool wide ) { mbWideLocation = wide; }

    void        SetLayoutRect( nw::ut::Rect& rect ) { mRect = rect; }
    void        SetLayoutRect( f32 left, f32 top, f32 right, f32 bottom )
                { mRect.left = left; mRect.top = top; mRect.right = right; mRect.bottom = bottom; }
    f32         GetLayoutWidth()  const { return mRect.right - mRect.left; }
    f32         GetLayoutHeight() const { return mRect.top - mRect.bottom; }

    bool             IsOrtho()           const { return !mbPerspective; }
    bool             IsPerspective()     const { return mbPerspective;  }
    void             SetPerspective( bool persp ) { mbPerspective = persp; }

    void        CalcProjMtx() { this->CalcPerspMtx(); this->CalcOrthoMtx(); }

    void        GetPerspMtx( nn::math::MTX44* pMtx ) const { nn::math::MTX44Copy( pMtx, &mPerspMtx ); }
    void        GetOrthoMtx( nn::math::MTX44* pMtx ) const { nn::math::MTX44Copy( pMtx, &mOrthoMtx ); }
    void        GetProjMtx( nn::math::MTX44* pMtx ) const
    {
        if ( mbPerspective ) { this->GetPerspMtx( pMtx ); }
        else                 { this->GetOrthoMtx( pMtx ); }
    }

    void        GetViewMtx( nn::math::MTX34* pMtx ) const;

    void        SetupProjection( vw::system::ScreenKind screen, bool persp );
    void        SetupProjection( vw::system::ScreenKind screen ) { this->SetupProjection( screen, mbPerspective ); }

    f32         GetWideScale() const { return mWideScale; }
    void        SetWideScale( f32 scale ) { mWideScale = scale; }

    f32         GetTestScale() const { return mTestScale; }
    void        SetTestScale( f32 scale ) { mTestScale = scale; }

    f32         GetStereoBasePlane() const { return mStereoBasePlane; }
    void        SetStereoBasePlane(f32 value) { mStereoBasePlane = value; }

    f32         GetStereoFactor() const { return mStereoFactor; }
    void        SetStereoFactor(f32 value) { mStereoFactor = value; }

    bool        GetRealSwitch() const { return mRealSwitch; }
    void        SetRealSwitch(bool value) { mRealSwitch = value; }

    f32         GetOrthoParallax() const { return mOrthoParallax; }

    f32         GetDistance() const;

//----- Private members -------------
private:
    void CalcPerspMtx();
    void CalcOrthoMtx();

    nn::ulcd::StereoCamera mStereoCamera;

    nw::ut::Rect                mRect;
    nn::math::MTX44             mPerspMtx;
    nn::math::MTX44             mPerspMtxL;
    nn::math::MTX44             mPerspMtxR;
    nn::math::MTX44             mOrthoMtx;
    f32 mOrthoParallax;

    f32                         mFovy;
    f32                         mNear;
    f32                         mFar;
    f32                         mMinNear;

    f32                         mWideScale;
    f32                         mTestScale;

    // Position of base plane in layout coordinate system (Z coordinate).
    f32 mStereoBasePlane;

    // Adjustment coefficient for the degree of stereoscopy (0 parallax at 0, standard at 1)
    f32 mStereoFactor;

    bool mbWideLocation;
    bool mbPerspective;

    // Allow changing the view angle to have a sense of reality?
    bool mRealSwitch;

#if defined(NW_PLATFORM_CTR)
    u8 padding[1];
#endif
};


}   /* namespace g2d */
}   /* namespace vw  */


/* NW4C_VIEWER_G2D_CAMERAMGR_H_ */
#endif


