/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <algorithm>
#include "../include/AnimManager.h"

namespace vw
{
namespace g2d
{

FrameCtrl::FrameCtrl()
 : mFrame(0.0f)
 , mMaxFrame(1.0f)
 , mAbsoluteFrame( 0.0f)
 , mStepRate(1.0f)
 , mIsLoop(false)
{
}

void
FrameCtrl::SetMaxFrame( f32 max )
{
    NW_ASSERT( max > 0.0f );
    mMaxFrame = max;

    if ( mFrame >= mMaxFrame - 1 )
    {
        mFrame         = 0.0f;
        mAbsoluteFrame = 0.0f;
    }
}

void
FrameCtrl::StepFrame( f32 step, bool isLoop )
{
    mFrame         += step;
    mAbsoluteFrame += step;

    if ( isLoop )
    {
        while ( mFrame >= mMaxFrame )
        {
            mFrame -= mMaxFrame;
        }
        while ( mFrame < 0.0f )
        {
            mFrame += mMaxFrame;
        }
    }
    else
    {
        if ( mFrame > mMaxFrame - 1 )
        {
            mFrame = mMaxFrame - 1.0f;
        }
        if ( mFrame < 0.0f )
        {
            mFrame = 0.0f;
        }
    }
}

void
FrameCtrl::SetFrame( f32 frame )
{
    mFrame         = frame;
    mAbsoluteFrame = frame;
    if ( mIsLoop )
    {
        while ( mFrame >= mMaxFrame )
        {
            mFrame -= mMaxFrame;
        }
    }
    else
    {
        if ( mFrame > mMaxFrame - 1.0f )
        {
            mFrame = mMaxFrame - 1.0f;
        }
    }
    while ( mFrame < 0.0f )
    {
        mFrame += mMaxFrame;
    }
}

bool
FrameCtrl::IsStopAtEnd() const
{
    if ( (! mIsLoop) && mFrame >= mMaxFrame - 1 )
    {
        // NOTE: No support for reverse playback
        return true;
    }
    return false;
}

AnimCtrl::AnimCtrl()
{
    Clear();
}

void
AnimCtrl::Clear()
{
    mFrame = 0.0f;
    mName = NULL;
    mpLayout = NULL;
}

void
AnimCtrl::Set(
    const char* name,
    nw::lyt::Layout* pLayout,
    const nw::lyt::AnimResource& animRes
)
{
    mName = name;
    mpLayout = pLayout;
    mAnimRes = animRes;
    FrameCtrl::SetMaxFrame(animRes.GetResourceBlock()->frameSize);
    FrameCtrl::SetLoop(animRes.GetResourceBlock()->loop != 0);
}

void
AnimCtrl::FlushToAnim_()
{
    NW_FOREACH(
        nw::lyt::AnimTransform& animTrans,
        mpLayout->GetAnimTransformList())
    {
        if (animTrans.GetAnimResource() == mAnimRes.GetResourceBlock())
        {
            animTrans.SetFrame(this->GetCurrentFrame());
        }
    }
}

void
AnimCtrl::SetEndFrame(bool bPrev)
{
    const f32 endFrame = IsLoop() || bPrev ?
        0.f:
        GetMaxFrame() - 1;

    SetFrame(endFrame);
}

AnimManager::AnimManager()
{
}

void AnimManager::Init()
{
     mAnimCtrlVec = AnimCtrlVec(Memory_GetAllocator());
     mPaneVec = PaneVec(Memory_GetAllocator());
     mGroupVec = GroupVec(Memory_GetAllocator());
     this->Clear();
}

void AnimManager::Clear()
{
    mpLayout = NULL;
    mPlayState = STATE_STOP;
    mIsWaitOneFrame = false;
    mCrAnim = -1;
    mTargetType = TargetType_Pane;
    mTargetAnimate = false;
    mTargetGroupInTag = false;

    NW_FOREACH(s32& idx, mCrTarget)
    {
        idx = -1;
    }

    mAnimCtrlVec.resize(0);
    mPaneVec.resize(0);
    mGroupVec.resize(0);
}

void AnimManager::SetLayout(nw::lyt::Layout* pLayout)
{
    mpLayout = pLayout;
    mPaneVec.resize(0);
    mGroupVec.resize(0);

    if (pLayout == 0)
    {
        return;
    }

    // Collect panes.
    {
        // To show all animations, be sure to register the root pane.
        nw::lyt::Pane* pPane = pLayout->GetRootPane();
        mPaneVec.push_back(pPane);

        // For other panes, register only when they have animation.
        while ((pPane = nw::lyt::GetNextPane(pPane)) != NULL)
        {
            if (! pPane->GetAnimationList().empty())
            {
                mPaneVec.push_back(pPane);
            }
        }
    }

    // Collect groups.
    {
        // To indicate the entire animation, register a dummy NULL.
        mGroupVec.push_back(NULL);

        nw::lyt::GroupContainer* pGroupContainer = pLayout->GetGroupContainer();
        if (pGroupContainer != 0)
        {
            NW_FOREACH(nw::lyt::Group& group, pGroupContainer->GetGroupList())
            {
                mGroupVec.push_back(&group);
            }
        }
    }
}

void
AnimManager::RegistAnim(
    s32 idx,
    const char* name,
    const nw::lyt::AnimResource& animRes
)
{
    GetAnimCtrlRef(idx).Set(name, mpLayout, animRes);
}

f32
AnimManager::GetCurrentFrame() const
{
    if ( ! this->HasAnim() )
    {
        return 0.0f;
    }
    else
    {
        return this->GetAnimCtrl()->GetCurrentFrame();
    }
}

f32
AnimManager::GetAbsoluteFrame() const
{
    if ( ! this->HasAnim() )
    {
        return 0.0f;
    }
    else
    {
        return this->GetAnimCtrl()->GetAbsoluteFrame();
    }
}

f32
AnimManager::GetMaxFrame() const
{
    if ( ! this->HasAnim() )
    {
        return 0.0f;
    }
    else
    {
        return this->GetAnimCtrl()->GetMaxFrame();
    }
}

bool
AnimManager::IsStopAtEnd() const
{
    if ( ! this->HasAnim() )
    {
        return false;
    }
    return this->GetAnimCtrl()->IsStopAtEnd();
}

void
AnimManager::Update()
{
    if ( ! this->HasAnim() )
    {
        return;
    }

    if ( mIsWaitOneFrame )
    {
        mIsWaitOneFrame = false;
        return;
    }

    if ( mPlayState == STATE_PLAY )
    {
        f32  rate = this->GetAnimCtrl()->GetStepRate();
        bool loop = this->GetAnimCtrl()->IsLoop();
        this->GetAnimCtrl()->StepFrame( rate, loop );
    }
}

void
AnimManager::SetAnimNum( s32 num )
{
    NW_ASSERT(num >= 0);

    if (mCrAnim >= num)
    {
        SetAnim(num - 1);   // When num is 0, mCrAnim is -1.
    }

    mAnimCtrlVec.resize(std::size_t(num));
}

void
AnimManager::SetFrame( f32 frame )
{
    if ( ! this->HasAnim() )
    {
        return;
    }

    this->GetAnimCtrl()->SetFrame( frame );
}

void
AnimManager::Step( f32 step )
{
    if ( ! this->HasAnim() )
    {
        return;
    }

    this->Stop();
    this->GetAnimCtrl()->StepFrame( step, true );
}

void
AnimManager::SetAnim( s32 idx )
{
    if (idx < -1)
    {
        return;
    }
    if (idx != -1 && size_t(idx) >= mAnimCtrlVec.size())
    {
        return;
    }

    if ( idx != mCrAnim )
    {
        if ( mCrAnim != -1 )
        {
            if (idx != -1)
            {
                // When switching animation, advance the current animation to a convenient position.
                // 
                // When moving to the next animation with the animation being a one-time animation, go to end frame.
                // 
                // When moving to the previous animation with the animation being a loop-type animation, go to frame 0.
                // 
                const bool bNext = (mCrAnim + 1) % mAnimCtrlVec.size() == idx;
                this->GetAnimCtrl()->SetEndFrame(! bNext);
                mpLayout->Animate();
            }

            this->SetAnimationEnable(false);
        }

        mCrAnim = idx;

        if (mCrAnim != -1)
        {
            this->SetAnimationEnable(true);
        }
    }
}

void
AnimManager::SetAnimationEnable(bool enable)
{
    const nw::lyt::AnimResource& animRes = this->GetAnimCtrl()->GetAnimResource();
    const nw::lyt::res::AnimationBlock* pBlock = animRes.GetResourceBlock();

    switch (this->GetTargetType())
    {
    case TargetType_Pane:
        {
            nw::lyt::Pane* pPane = this->GetTargetPane();
            NW_NULL_ASSERT(pPane);

            bool bRecursive = false;
            if (pPane->GetParent() == NULL)
            {
                bRecursive = true;
            }

            NW_FOREACH(
                nw::lyt::AnimTransform& animTrans,
                mpLayout->GetAnimTransformList())
            {
                if (animTrans.GetAnimResource() == pBlock)
                {
                    pPane->SetAnimationEnable(&animTrans, enable, bRecursive);
                }
            }

            // Check if the current target is affected by the current animation.
            if (enable)
            {
                mTargetAnimate = this->IsTargetAnimate(pPane, animRes, bRecursive);
            }
        }
        break;

    case TargetType_Group:
        {
            nw::lyt::Group* pGroup = this->GetTargetGroup();

            if (pGroup == NULL)
            {
                // Entire animation
                nw::lyt::Pane* pPane = mPaneVec[0];

                NW_FOREACH(
                    nw::lyt::AnimTransform& animTrans,
                    mpLayout->GetAnimTransformList())
                {
                    if (animTrans.GetAnimResource() == pBlock)
                    {
                        pPane->SetAnimationEnable(&animTrans, enable, true);
                    }
                }
            }
            else
            {
                NW_FOREACH(
                    nw::lyt::AnimTransform& animTrans,
                    mpLayout->GetAnimTransformList())
                {
                    if (animTrans.GetAnimResource() == pBlock)
                    {
                        nw::lyt::SetAnimationEnable(pGroup, &animTrans, enable, true);
                    }
                }
            }

            // Check if the current target is affected by the current animation.
            if (enable)
            {
                if (pGroup == NULL)
                {
                    // Entire animation
                    mTargetAnimate = this->IsTargetAnimate(mPaneVec[0], animRes, true);
                }
                else
                {
                    mTargetAnimate = false;
                    NW_FOREACH(nw::lyt::PaneLink& link, pGroup->GetPaneList())
                    {
                        if (link.target == 0)
                        {
                            continue;
                        }

                        mTargetAnimate = this->IsTargetAnimate(link.target, animRes, true);
                        if (mTargetAnimate)
                        {
                            break;
                        }
                    }
                }
            }

            // Check whether the current target is registered in the animation interval tag.
            if (enable)
            {
                mTargetGroupInTag = false;

                const nw::lyt::AnimationGroupRef* groupArray =
                    animRes.GetGroupArray();

                if (pGroup != NULL && groupArray != NULL)
                {
                    u32 groupNum = animRes.GetGroupNum();
                    for (u32 i = 0; i < groupNum; ++i)
                    {
                        if (0 == std::strcmp(pGroup->GetName(), groupArray[i].name))
                        {
                            mTargetGroupInTag = true;
                            break;
                        }
                    }
                }
            }
        }
        break;

    default:
        break;
    }
}

void
AnimManager::ChangeAnimToNext()
{
    if ( this->HasAnim() )
    {
        s32 idx = this->GetAnimIdx() + 1;
        if (idx >= mAnimCtrlVec.size())
        {
            idx = 0;
        }
        this->SetAnim(idx);
    }
}

void
AnimManager::ChangeAnimToPrev()
{
    if (this->HasAnim())
    {
        s32 idx = this->GetAnimIdx() - 1;
        if (idx < 0)
        {
            idx = s32(mAnimCtrlVec.size() - 1);
        }
        this->SetAnim(idx);
    }
}

bool
AnimManager::IsTargetSelectable() const
{
    if (mPaneVec.size() > 1)
    {
        return true;
    }

    if (mGroupVec.size() > 1)
    {
        return true;
    }

    return false;
}

bool
AnimManager::SetTargetType(TargetType type)
{
    switch (type)
    {
    case TargetType_Pane:
        if (mPaneVec.size() <= 1)
        {
            return false;
        }
        break;

    case TargetType_Group:
        if (mGroupVec.size() <= 1)
        {
            return false;
        }
        break;

    default:
        return false;
    }

    if (type != mTargetType)
    {
        this->SetAnimationEnable(false);

        mTargetType = type;

        if (mCrTarget[mTargetType] == -1)
        {
            mCrTarget[mTargetType] = 0;
        }

        this->SetAnimationEnable(true);
    }

    return true;
}

void
AnimManager::ChangeTargetTypeToPrev()
{
    int crType = GetTargetType();

    int type = (crType + TargetType_Max - 1) % TargetType_Max;

    while (type != crType && !this->SetTargetType(TargetType(type)))
    {
         type = (type + TargetType_Max - 1) % TargetType_Max;
    }
}

void
AnimManager::ChangeTargetTypeToNext()
{
    int crType = GetTargetType();

    int type = (crType + 1) % TargetType_Max;

    while (type != crType && !this->SetTargetType(TargetType(type)))
    {
         type = (type + 1) % TargetType_Max;
    }
}

void
AnimManager::SetTargetIdx(s32 idx)
{
    if (idx < 0 || this->GetTargetNum() <= idx)
    {
        return;
    }

    if (idx != mCrTarget[this->GetTargetType()])
    {
        // Stop current pane animation
        this->SetAnimationEnable(false);

        mCrTarget[this->GetTargetType()] = idx;

        // Enable next animation
        this->SetAnimationEnable(true);

        if (this->IsStopAtEnd())
        {
            this->SetFrame(0.0f);
        }

        if (! this->IsPlaying())
        {
            this->Play();
        }
    }
}

const char*
AnimManager::GetTargetTypeName() const
{
    switch (this->GetTargetType())
    {
    case TargetType_Pane:
        return "PANE";

    case TargetType_Group:
        return "GROUP";

    default:
        return "N/A";
    }
}

const char*
AnimManager::GetTargetName() const
{
    if (this->GetTargetIdx() <= 0)
    {
        return "ALL";
    }
    else
    {
        switch (this->GetTargetType())
        {
        case TargetType_Pane:
            return this->GetTargetPane()->GetName();

        case TargetType_Group:
            return this->GetTargetGroup()->GetName();

        default:
            return "N/A";
        }
    }
}

u32
AnimManager::GetTargetNum() const
{
    switch (this->GetTargetType())
    {
    case TargetType_Pane:
        return mPaneVec.size();

    case TargetType_Group:
        return mGroupVec.size();

    default:
        return 0;
    }
}

void
AnimManager::ChangeTargetToPrev()
{
    if (this->HasAnim())
    {
        s32 idx = (this->GetTargetIdx() + this->GetTargetNum() - 1) % this->GetTargetNum();
        this->SetTargetIdx(idx);
    }
}

void
AnimManager::ChangeTargetToNext()
{
    if (this->HasAnim())
    {
        s32 idx = (this->GetTargetIdx() + 1) % this->GetTargetNum();
        this->SetTargetIdx(idx);
    }
}

bool
AnimManager::IsTargetAnimate(
    nw::lyt::Pane* pPane,
    const nw::lyt::AnimResource& animRes,
    bool bRecursive)
{
    NW_NULL_ASSERT(pPane);

    nw::lyt::Pane* pEnd = pPane->GetFirstChild();

    if (pEnd == NULL || bRecursive)
    {
        nw::lyt::Pane* pTmp = pPane;

        // After the recursion, find the next obtained pane.
        while (pTmp)
        {
            nw::lyt::Pane* pParent = pTmp->GetParent();
            if (pParent == NULL)
            {
                pEnd = NULL;
                break;
            }

            if ((pEnd = pParent->GetNextChild(pTmp)) != NULL)
            {
                break;
            }

            pTmp = pParent;
        }
    }

    while (pPane && pPane != pEnd)
    {
        if (pPane->FindAnimationLinkSelf(animRes))
        {
            return true;
        }

        u32 num = pPane->GetMaterialNum();
        for (u32 i = 0; i < num; ++i)
        {
            nw::lyt::Material* pMaterial = pPane->GetMaterial(i);
            if (pMaterial)
            {
                if (pMaterial->FindAnimationLink(animRes))
                {
                    return true;
                }
            }
        }

        pPane = nw::lyt::GetNextPane(pPane);
    }

    return false;
}

}   /* namespace g2d */
}   /* namespace vw  */
