/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/ArcFileResourceAccessor.h"
#include "../include/Module.h"

namespace vw {
namespace g2d {

ArcFileResourceAccessor::ArcFileResourceAccessor()
    : m_pArcResourceAccessor(0)
    , m_ArchiveFontInfoArray(Memory_GetAllocator())
{
}

ArcFileResourceAccessor::~ArcFileResourceAccessor()
{
    this->Finalize();
}

void ArcFileResourceAccessor::Finalize()
{
    if (m_pArcResourceAccessor)
    {
        NW_FOREACH(FontInfo& info, m_ArchiveFontInfoArray)
        {
            m_pArcResourceAccessor->UnregistFont(info.key);
            info.pFont->Destroy();
            nw::lyt::Layout::DeleteObj(info.pFont);
        }

        nw::lyt::Layout::DeleteObj(m_pArcResourceAccessor);
        m_pArcResourceAccessor = NULL;
    }

    m_Archive.clear();
    m_ArchiveFontInfoArray.clear();
}

void* ArcFileResourceAccessor::GetResource(
    nw::lyt::ResType resType,
    const char* name,
    u32* pSize)
{
    if (!m_pArcResourceAccessor)
    {
        return NULL;
    }

    return m_pArcResourceAccessor->GetResource(resType, name, pSize);
}

nw::font::Font* ArcFileResourceAccessor::GetFont(const char* name)
{
    if (!m_pArcResourceAccessor)
    {
        return NULL;
    }

    nw::lyt::FontContainer& fontContainer = m_pArcResourceAccessor->GetFontContainer();
    nw::font::Font* pFont = fontContainer.FindFontByName(name);
    if (pFont)
    {
        return pFont;
    }

    u32 size = 0;
    void* pRes = m_pArcResourceAccessor->GetResource(nw::lyt::res::RESOURCETYPE_ARCHIVEFONT, name, &size);
    if (pRes != NULL)
    {
        FontInfo fontInfo;

        // Allocate buffer to store glyphs
        {
            u32 buffSize = nw::font::ArchiveFont::GetRequireBufferSize(pRes);
            if (buffSize == 0)
            {
                return NULL;
            }

            const u8 alignment = 128;
            void* buffer = Memory_GetAllocatorDevice()->Alloc(buffSize, alignment);
            if (buffer == NULL)
            {
                NW_WARNING(false, "Memory allocation failed.");
                return NULL;
            }

            fontInfo.buffer = nw::ut::MoveArray<u8>(buffer, buffSize, Memory_GetAllocatorDevice());
        }

        // Build font
        {
            fontInfo.pFont = nw::lyt::Layout::NewObj<nw::font::ArchiveFont>();
            fontInfo.pFont->Construct(fontInfo.buffer.begin(), fontInfo.buffer.size(), pRes);
        }

        fontInfo.key = m_pArcResourceAccessor->RegistFont(name, fontInfo.pFont);
        m_ArchiveFontInfoArray.push_back(fontInfo);

        return fontInfo.pFont;
    }
    else
    {
        return m_pArcResourceAccessor->GetFont(name);
    }
}

const nw::lyt::TextureInfo ArcFileResourceAccessor::GetTexture(const char* name)
{
    if (!m_pArcResourceAccessor)
    {
        return nw::lyt::TextureInfo();
    }

    return m_pArcResourceAccessor->GetTexture(name);
}

bool ArcFileResourceAccessor::Attach(const char* /*path*/)
{
    this->Finalize();

    char path[1024];
    s32 len = 0;
    nn::hio::GetEnvironmentVariable(&len, path, sizeof(path), "NW4C_ROOT");
    if (len == 0)
    {
        NW_WARNING(false, "NW4C_ROOT is not set.");
        return false;
    }

    const char *arcPath = "\\lyt1\\tool\\LayoutEditor\\Viewer\\layout.arc";
    nw::ut::strncat(path, sizeof(path), arcPath, sizeof(path));

    // We delete the old archive first to prevent memory from temporarily being allocated for two archives at once.
    // 
    m_Archive.Reset(NULL, 0);

    m_Archive = vw::system::Utility::LoadFileFromHost(
        Memory_GetAllocatorDevice(), path);

    if (m_Archive.size() == 0)
    {
        NW_WARNING(false, "load failed.");
        return false;
    }

#ifdef NW_TEST_LYT_VRAM
    nw::lyt::MemArea texMemArea = ModuleG2dImpl::GetInstance().GetTexMemArea();
    if (texMemArea != nw::lyt::MEMAREA_FCRAM)
    {
        nw::lyt::MemArea texMemArea4bit =
          (texMemArea == nw::lyt::MEMAREA_VRAMA)?
            nw::lyt::MEMAREA_VRAMB : nw::lyt::MEMAREA_VRAMA;

        nw::lyt::ARCHandle arcHandle;
        nw::lyt::ARCInitHandle(m_Archive.begin(), &arcHandle);

        nw::lyt::ARCDir arcDir;
        if (nw::lyt::ArcUtil::OpenTextureDir(&arcHandle, L"/", &arcDir))
        {
            nw::lyt::ARCDirEntry arcDirEnt;
            while (nw::lyt::ArcUtil::ReadTextureDir(&arcDir, &arcDirEnt))
            {
                nw::lyt::ARCFileInfo arcFileInfo;
                if (!nw::lyt::ARCOpen(&arcHandle, &arcDirEnt, &arcFileInfo))
                {
                    continue;
                }

                nw::lyt::TexResource texResource;
                if (!texResource.Set(arcFileInfo))
                {
                    continue;
                }

                if (texResource.Is4bitFormat())
                {
                    texResource.SetImageArea(texMemArea4bit);
                }
                else
                {
                    texResource.SetImageArea(texMemArea);
                }

                nw::lyt::ARCClose(&arcFileInfo);
            }
        }

        nw::lyt::ARCCloseDir(&arcDir);
    }
#endif

    m_pArcResourceAccessor = nw::lyt::Layout::NewObj<nw::lyt::ArcResourceAccessor>();
    m_pArcResourceAccessor->Attach(m_Archive.begin(), ".");

    nw::lyt::ARCInitHandle(m_Archive.begin(), &m_ArcHandle);

    return true;
}

bool ArcFileResourceAccessor::OpenResourceDir(
    nw::lyt::ResType resType,
    nw::lyt::ARCDir* pArcDir)
{
    s32 entryNum = -1;

    static const wchar_t* resRootDir = L"/./";

    if (-1 == ARCConvertPathToEntrynum(&m_ArcHandle, resRootDir))
    {
        return false;
    }

    if (!ARCChangeDir(&m_ArcHandle, resRootDir))
    {
        return false;
    }

    wchar_t resTypeStr[5];
    resTypeStr[0] = u8(resType >> 24);
    resTypeStr[1] = u8(resType >> 16);
    resTypeStr[2] = u8(resType >> 8);
    resTypeStr[3] = u8(resType >> 0);
    resTypeStr[4] = L'\0';

    if (-1 == ARCConvertPathToEntrynum(&m_ArcHandle, resTypeStr))
    {
        return false;
    }

    if (!ARCOpenDir(&m_ArcHandle, resTypeStr, pArcDir))
    {
        return false;
    }

    return true;
}

FileResourceAccessor::FindInfoPtr
ArcFileResourceAccessor::FindResource(
    nw::lyt::ResType resType)
{
    nw::os::IAllocator* pAllocator = Memory_GetAllocator();
    FindInfoImpl* pInfo = pAllocator->AllocAndConstruct<FindInfoImpl>(1);

    nw::lyt::ARCDir arcDir;
    if (this->OpenResourceDir(resType, &arcDir))
    {
        pInfo->FindFirst(arcDir);
    }

    return FindInfoPtr(pInfo, pAllocator);
}

ArcFileResourceAccessor::FindInfoImpl::FindInfoImpl()
    : m_Name(Memory_GetAllocator())
{
}

ArcFileResourceAccessor::FindInfoImpl::~FindInfoImpl()
{
    nw::lyt::ARCCloseDir(&m_ArcDir);
}

bool ArcFileResourceAccessor::FindInfoImpl::IsValid()
{
    return !m_Name.empty();
}

void ArcFileResourceAccessor::FindInfoImpl::FindFirst(
    const nw::lyt::ARCDir& arcDir)
{
    m_ArcDir = arcDir;
    this->FindNext();
}

bool ArcFileResourceAccessor::FindInfoImpl::FindNext()
{
    while (nw::lyt::ARCReadDir(&m_ArcDir, &m_ArcDirEntry))
    {
        if (!m_ArcDirEntry.isDir)
        {
            size_t len = std::wcstombs(NULL, m_ArcDirEntry.name, 0);
            if (len == (size_t)(-1))
            {
                continue;
            }

            m_Name.resize(len + 1);
            std::wcstombs(m_Name.begin(), m_ArcDirEntry.name, len);
            m_Name[len] = '\0';

            return true;
        }
    }

    m_Name.resize(0);
    return false;
}

const char* ArcFileResourceAccessor::FindInfoImpl::GetName()
{
    return (m_Name.empty())? NULL : m_Name.begin();
}

} // namespace g2d
} // namespace vw

