/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nw/ut.h>
#include <viewer/system.h>
#include "../include/FlashAppeal.h"

namespace vw  {
namespace g2d {


/*---------------------------------------------------------------------------*
  Name:         FlashAppeal::FlashAppeal
  Description:  Constructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
FlashAppeal::FlashAppeal()
 : mInterval(8)
 , mFlashNum(4)
 , mFrameCnt(0)
 , mFlashCnt(0)
 , mpPane(0)
 , mpGroup(0)
 , mbPaneVisible(false)
{
}

void
FlashAppeal::Init()
{
    mGroupVisibleVec = GroupVisibleVec(Memory_GetAllocator());
}

/*---------------------------------------------------------------------------*
  Name:         FlashAppeal::Update
  Description:  Updates the frame counter for blinking display
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
FlashAppeal::Update()
{
    if ( mFrameCnt == 0 )
    {
        return;
    }

    if ( --mFrameCnt == 0 )
    {
        --mFlashCnt;
        if ( mFlashCnt > 0 )
        {
            mFrameCnt = u16( mInterval * 2 );
        }
    }

    this->UpdateVisibility_();
}


/*---------------------------------------------------------------------------*
  Name:         FlashAppeal::UpdateVisibility_
  Description:  Updates the resource render setting flag according to the frame counter value
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
FlashAppeal::UpdateVisibility_()
{
    if (mpPane)
    {
        mpPane->SetVisible(! IsTurnOff());
    }

    if (mpGroup)
    {
        NW_FOREACH(nw::lyt::PaneLink& link, mpGroup->GetPaneList())
        {
            if (link.target)
            {
                link.target->SetVisible(! IsTurnOff());
            }
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         FlashAppeal::IsTurnOff
  Description:  Gets whether the selected object is in a period of being turned off.

  Arguments:    None.
  Returns:      If in a period of being turned off, TRUE; if in a state that should be displayed, FALSE

 *---------------------------------------------------------------------------


*/
bool
FlashAppeal::IsTurnOff()
{
    if ( mFrameCnt > mInterval  )
    {
        return true;
    }
    return false;
}


/*---------------------------------------------------------------------------*
  Name:         FlashAppeal::Start
  Description:  Starts illumination
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
FlashAppeal::Start( void )
{
    mFrameCnt = u16(mInterval * 2 + 1);
    mFlashCnt = mFlashNum;
}

/*---------------------------------------------------------------------------*
  Name:         FlashAppeal::Stop
  Description:  Ends illumination.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
FlashAppeal::Stop( void )
{
    mFrameCnt = 0;
    mFlashCnt = 0;

    // Turns visibility on
    SetPane(0);
    SetGroup(0);
}

void
FlashAppeal::SetPane(nw::lyt::Pane* pPane)
{
    if (mpGroup)
    {
        this->SetGroup(0);
    }

    if (mpPane != pPane)
    {
        if (mpPane)
        {
            mpPane->SetVisible(mbPaneVisible);
        }

        mpPane = pPane;

        if (mpPane)
        {
            mbPaneVisible = mpPane->IsVisible();
        }

        Start();
    }
}

void
FlashAppeal::SetGroup(nw::lyt::Group* pGroup)
{
    if (mpPane)
    {
        this->SetPane(0);
    }

    if (mpGroup != pGroup)
    {
        if (mpGroup)
        {
            int index = 0;
            NW_FOREACH(nw::lyt::PaneLink& link, mpGroup->GetPaneList())
            {
                if (link.target)
                {
                    link.target->SetVisible(mGroupVisibleVec[index++]);
                }
            }
        }

        mGroupVisibleVec.clear();

        mpGroup = pGroup;

        if (mpGroup)
        {
            NW_FOREACH(nw::lyt::PaneLink& link, mpGroup->GetPaneList())
            {
                if (link.target)
                {
                    mGroupVisibleVec.push_back(link.target->IsVisible());
                }
            }
        }

        Start();
    }
}

}   /* namespace g2d */
}   /* namespace vw  */

