/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nw/viewer/sys/Application.h>
#include <nw/lyt/lyt_Resources.h>
#include <nw/lyt/lyt_Pane.h>
#include <viewer/system.h>
//#include <nw/lyt/McsResourceAccessor.h>
//#include <nw/lyt/DvdResourceAccessor.h>
#include "../include/FileResourceAccessor.h"
#include "../include/ResLayout.h"

namespace vw  {
namespace g2d {

namespace
{

struct AnimResInfo
{
    nw::lyt::AnimResource animRes;
    bool                    bWholeAnim;
#if defined(NW_PLATFORM_CTR)
    u8 padding[3];
#endif
};

typedef nw::ut::MoveArray<AnimResInfo> AnimResInfoVec;

bool
IsStartLayoutName(
    const char* testName,
    const char* layoutName,
    std::size_t layoutNameLen
)
{
    return
         0 == std::strcmp(testName, layoutName)   // Perfect match?
      || (0 == std::memcmp(testName, layoutName, layoutNameLen) && testName[layoutNameLen] == '_');   // Starts with layout name + '_'?
}

f32
GetReverseFrame(
    const AnimCtrl* pOldCtrl,
    const AnimCtrl* pNewCtrl
)
{
    if (pOldCtrl->GetMaxFrame() == 1)
    {
        return 0.f;
    }

    f32 framePt = pOldCtrl->GetCurrentFrame()
                    / (pOldCtrl->GetMaxFrame() - 1 - 0);
    framePt = 1.f - framePt;

    return framePt * (pNewCtrl->GetMaxFrame() - 1 - 0);
}

/*---------------------------------------------------------------------------*
  Name:         AnimResInfoLess
  Description:  Comparison function for animation sort
  Arguments:    elem1
                elem2
  Returns:
 *---------------------------------------------------------------------------


*/
bool
AnimResInfoLess(const AnimResInfo& elem1, const AnimResInfo& elem2)
{
    if (elem1.bWholeAnim || elem2.bWholeAnim)
    {
        if (elem1.bWholeAnim && elem2.bWholeAnim)
        {
            return elem1.animRes.GetResourceBlock() < elem2.animRes.GetResourceBlock();
        }
        // When one or the either is the entire animation
        return elem1.bWholeAnim;
    }

    const nw::lyt::AnimResource& animRes1 = elem1.animRes;
    const nw::lyt::AnimResource& animRes2 = elem2.animRes;
    if (animRes1.GetTagBlock() || animRes2.GetTagBlock())
    {
        if (animRes1.GetTagBlock() && animRes2.GetTagBlock())
        {
            if (animRes1.GetTagOrder() == animRes2.GetTagOrder())
            {
                return animRes1.GetResourceBlock() < animRes2.GetResourceBlock();
            }
            return animRes1.GetTagOrder() < animRes2.GetTagOrder();
        }

        // When one or the either has a tag block
        return ! animRes1.GetTagBlock();   // The ones without tag blocks are first
    }

    return animRes1.GetResourceBlock() < animRes2.GetResourceBlock();
}

nw::lyt::AnimTransform*
FindAnimTransform(
    nw::lyt::Pane*                pPane,
    const nw::lyt::AnimResource&  animRes
)
{
    if (nw::lyt::AnimationLink *const pAnimLink = pPane->FindAnimationLinkSelf(animRes))
    {
        return pAnimLink->GetAnimTransform();
    }

    const u8 materialNum = pPane->GetMaterialNum();
    for (u8 i = 0; i < materialNum; ++i)
    {
        if (nw::lyt::AnimationLink *const pAnimLink = pPane->GetMaterial(i)->FindAnimationLink(animRes))
        {
            return pAnimLink->GetAnimTransform();
        }
    }

    return 0;
}

bool
HasAnimPane(
    nw::lyt::Pane*    pPane,
    AnimResInfoVec&   /*animResInfoVec*/
)
{
    NW_NULL_ASSERT(pPane);
    return !pPane->GetAnimationList().empty();
}

}   // Namespace

nw::os::IAllocator*
ResLayout::GetAllocator()
{
    return Memory_GetAllocator();
}

/*---------------------------------------------------------------------------*
  Name:         ResLayout
  Description:  Constructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
ResLayout::ResLayout()
 :  mpLayout(NULL)
{
}

void
ResLayout::Init()
{
    mAnimManager.Init();

    std::memset( mName, 0, MAX_PATH );
}

/*---------------------------------------------------------------------------*
  Name:         ResLayout::Destroy
  Description:  Layout destruction
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ResLayout::Destroy()
{
    if ( mpLayout )
    {
        nw::lyt::Layout::DeleteObj(mpLayout);
        mpLayout = NULL;
    }

    mAnimManager.Clear();
}

/*---------------------------------------------------------------------------*
  Name:         ResLayout::Load
  Description:  Loads the animation.
  Arguments:    pathName    File path
                layoutName  Layout name
                type        type of file to load (DVD, MCS)
  Returns:      Returns true if loading the file succeeded, false if it fails
                
 *---------------------------------------------------------------------------


*/
bool
ResLayout::Load(
    const char* pathName,
    const char* layoutName,
    u32         type
)
{
    // Layout destruction
    this->Destroy();
    nw::ut::strcpy(mName, sizeof(mName), "");

    FileResourceAccessor* pResAccessor = 0;

#if 0 // FIXME!
    if (type == system::FileInfo::TYPE_DVD)
    {
        pResAccessor = mpDvdResAccessor;
    }
    else
    {
        pResAccessor = mpMcsResAccessor;
    }
#endif

    if (!mArcFileResourceAccessor.Attach(pathName))
    {
        NW_WARNING(false, "resource attach failed.\n");
        return false;
    }

    pResAccessor = &mArcFileResourceAccessor;

    nw::viewer::lyt::IResourceListener* pResListener = nw::viewer::sys::App::GetInstance().GetModuleLYT().GetResourceListener();
    if (pResListener && pResListener->OnLoadResource(pResAccessor))
    {
        // How resources are handled is left up to the user.
        return true;
    }

    if (layoutName)
    {
        nw::ut::strcpy(mName, sizeof(mName), layoutName);
    }
    else
    {
        // Displays the first .bclyt file found.
        FileResourceAccessor::FindInfoPtr findInfo = pResAccessor->FindResource(nw::lyt::res::RESOURCETYPE_LAYOUT);
        if (!findInfo->IsValid())
        {
            NW_WARNING(false, "lyt not found.");
            return false;
        }

        const char* pName = findInfo->GetName();
        NW_NULL_ASSERT(pName);
        const char* pExt = std::strstr(pName, ".bclyt");
        NW_NULL_ASSERT(pExt);

        size_t len = pExt - pName;
        nw::ut::strncpy(mName, sizeof(mName), pName, pExt - pName);
    }

    nw::ut::MoveArray<char> fileName(this->GetAllocator());

    size_t layoutNameLen = std::strlen(mName);

    fileName.resize(layoutNameLen + 6 + 1);

    // Generate the layout filename.
    nw::ut::strcpy(fileName.begin(), fileName.size(), mName);
    nw::ut::strncat(fileName.begin(), fileName.size(), ".bclyt", fileName.size());

    const void* lytRes = pResAccessor->GetResource(nw::lyt::res::RESOURCETYPE_LAYOUT, fileName.begin());
    if (!lytRes)
    {
        NW_WARNING(false, "not found. - [%s]\n", fileName.begin());
        return false;
    }

    if (!pResListener || NULL == (mpLayout = pResListener->OnBuildLayout(lytRes, mName, pResAccessor)))
    {
        mpLayout = nw::lyt::Layout::NewObj<nw::lyt::Layout>();
        if (!mpLayout)
        {
            NW_WARNING(false, "layout construction failed.\n");
            return false;
        }

        if (!mpLayout->Build(lytRes, pResAccessor))
        {
            NW_WARNING(false, "binary version error. - [%s]\n", fileName.begin());
            nw::lyt::Layout::DeleteObj(mpLayout);
            mpLayout = NULL;
            return false;
        }
    }

    // Generate the filename for the entire animation.
    nw::ut::strcpy(fileName.begin(), fileName.size(), mName);
    nw::ut::strncat(fileName.begin(), fileName.size(), ".bclan", fileName.size());

    // Check the total number of animation files
    u32 animFileNum = 0;
    FileResourceAccessor::FindInfoPtr findInfo;
    findInfo = pResAccessor->FindResource(nw::lyt::res::RESOURCETYPE_ANIMATION);
    for (; findInfo->IsValid(); findInfo->FindNext())
    {
        if (IsStartLayoutName(findInfo->GetName(), fileName.begin(), layoutNameLen))
        {
            ++animFileNum;
        }
    }

    AnimResInfoVec animResVec(this->GetAllocator());
    animResVec.resize(animFileNum);

    // Load all animation resources at once
    u32 animIdx = 0;
    findInfo = pResAccessor->FindResource(nw::lyt::res::RESOURCETYPE_ANIMATION);
    for (; findInfo->IsValid(); findInfo->FindNext())
    {
        const char* name = findInfo->GetName();

        if (!IsStartLayoutName(name, fileName.begin(), layoutNameLen))
        {
            continue;
        }

        // Load animation resources separated by tags
        const void* lpaRes = pResAccessor->GetResource(nw::lyt::res::RESOURCETYPE_ANIMATION, name);
        if (! lpaRes)
        {
            continue;
        }

        animResVec[animIdx].animRes.Set(lpaRes);

        // The animation file with the same name as the layout is the entire animation
        animResVec[animIdx].bWholeAnim = 0 == std::strcmp(name, fileName.begin());

        ++animIdx;
    }

    animResVec.resize(animIdx);

    // Sort animation resources
    std::sort(animResVec.begin(), animResVec.end(), AnimResInfoLess);

    // Bind animation
    animIdx = 0;
    for (AnimResInfoVec::iterator it = animResVec.begin(); it != animResVec.end(); ++it, ++animIdx)
    {
        if (!pResListener || !pResListener->OnBindAnimation(mpLayout, it->animRes, pResAccessor))
        {
            if (! mpLayout->BindAnimationAuto(it->animRes, pResAccessor))
            {
                NW_WARNING(false, "animation bind failed.\n");
                break;
            }
        }
    }

    animResVec.resize(animIdx);

    u32 animatePaneNum = 0;
    if (! animResVec.empty())
    {
        for (nw::lyt::Pane* pPane = mpLayout->GetRootPane(); 0 != pPane; pPane = nw::lyt::GetNextPane(pPane))
        {
            if (HasAnimPane(pPane, animResVec))
            {
                ++animatePaneNum;
            }
        }
    }

    if (! animResVec.empty())
    {
        // Extract animation target
        mAnimManager.SetLayout(mpLayout);

        // Set number of animations
        mAnimManager.SetAnimNum(s32(animResVec.size()));

        // Load animation resources applied to entire layout
        animIdx = 0;
        NW_FOREACH(AnimResInfo& info, animResVec)
        {
            mAnimManager.RegistAnim(s32(animIdx), info.bWholeAnim ? "ALL": info.animRes.GetTagName(), info.animRes);
            ++animIdx;
        }

        mAnimManager.SetAnim(0);
        mAnimManager.SetTargetIdx(0);
    }

    return true;
}

void
ResLayout::Update()
{
    mAnimManager.Update();
}

#if 0 // FIXME!
/*---------------------------------------------------------------------------*
  Name:         ChangeAnimeByDpd
  Description:  Sets animation with DPD points.
  Arguments:    pos:              Cursor position
                drawInfo:         Reference to nw::lyt::DrawInfo object
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ResLayout::ChangeAnimeByDpd(
    const nw::math::VEC2&     pos,
    s32                       animSel
)
{
    // Process from each nw::lyt::AnimShare child pane
    for (AnimManagerVec::iterator it = mAnimManVec.begin() + 1; it != mAnimManVec.end(); ++it)
    {
        const s16 FocusInAnimIdx = 0;
        const s16 OtherAnimStartIdx = 1;
        const s16 FocusOutAnimIdx = s16(it->GetAnimNum() >= 2 ? it->GetAnimNum() - 1: 1);
        const s32 AnimSelMax = s32(it->GetAnimNum() >= 2 ? it->GetAnimNum() - 2: 0);

        const bool bPointed = 0 != nw::lyt::FindHitPane(it->GetTargetPane(), pos);
        if (bPointed)
        {
            if (! it->IsPlaying())   // When animation is not enabled
            {
                it->Play();
            }

            s32 nextAnimIdx = FocusInAnimIdx;
            if (0 <= animSel && animSel < AnimSelMax)
            {
                nextAnimIdx = OtherAnimStartIdx + animSel;
            }

            const s32 crAnimIdx = it->GetAnimIdx();
            if (nextAnimIdx < it->GetAnimNum() && nextAnimIdx != crAnimIdx)  // When animation is switched
            {
                it->SetAnim(nextAnimIdx);

                if (nextAnimIdx == FocusInAnimIdx)
                {
                    // When focus out -> in
                    if (crAnimIdx == FocusOutAnimIdx)
                    {
                        it->Set(GetReverseFrame(it->GetAnimCtrl(FocusOutAnimIdx), it->GetAnimCtrl(FocusInAnimIdx)));
                    }
                    else if (crAnimIdx == -1)
                    {
                        it->Set(0.f);
                    }
                }
                else
                {
                    it->Set(0.f);
                }
            }
        }
        else
        {
            // Playback focus out only when animation is enabled
            // (If moving from a stopped state to focus out suddenly, focus out starts when the B Button is pressed for the first time)
            //  
            if (it->IsPlaying())
            {
                const s32 crAnimIdx = it->GetAnimIdx();
                // When animation is switched
                if (FocusOutAnimIdx < it->GetAnimNum() && FocusOutAnimIdx != crAnimIdx)
                {
                    f32 frame = 0.f;
                    // When focus in -> out
                    if (crAnimIdx == FocusInAnimIdx)
                    {
                        frame = GetReverseFrame(it->GetAnimCtrl(FocusInAnimIdx), it->GetAnimCtrl(FocusOutAnimIdx));
                    }
                    it->SetAnim(FocusOutAnimIdx);
                    it->Set(frame);
                }
            }
        }
    }
}
#endif


}   /* namespace g2d */
}   /* namespace vw  */
