/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_ANIM_ELEM_H_
#define NW4C_VIEWER_G3D_ANIM_ELEM_H_

#include <nw/gfx.h>
#include "SceneElem.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw {
namespace g3d {

class AnimSet;
class AnimController;

class AnimElem : public ResElem<nw::anim::ResAnim>
{
friend class SceneManager;
friend class MoveArray;
public:
    enum AnimElemType
    {
        Anim_Skeletal = 0,
        Anim_Material,
        Anim_Visibility,
        Anim_Camera,
        Anim_Light,
        Anim_Fog,
        //
        Anim_Num
    };

    AnimElemType GetAnimType( void ) { return m_AnimType; }
    nw::anim::ResAnim GetResAnim( void ) { return m_ResObject; }

    class Builder
    {
    public:
        Builder();
        ~Builder();

        AnimElem* Create(nw::os::IAllocator *pAllocator, nw::os::IAllocator *pDeviceAllocator);
        Builder& ResSource(ResourceSet* resSource);
        Builder& ResObject(nw::anim::ResAnim resAnim);
        Builder& AnimType(AnimElemType type);

    private:
        ResourceSet* m_ResSource;
        nw::anim::ResAnim m_ResAnim;
        AnimElemType m_AnimType;
    };

private:
    AnimElem(
        AnimElemType animType,
        ResourceSet* resSource,
        nw::anim::ResAnim resObject,
        nw::os::IAllocator* pAllocator,
        nw::os::IAllocator* pDeviceAllocator
    );
    virtual ~AnimElem();
    virtual void Destroy();

    // Animation type
    AnimElemType m_AnimType;
};

class AnimSetLinker
{
friend class SceneManager;
public:
    typedef nw::ut::MoveArray<nw::gfx::BaseAnimEvaluator*> BaseEvaluatorArray;

    class Builder
    {
    public:
        Builder();
        ~Builder();

        AnimSetLinker* Create( nw::os::IAllocator *pAllocator );
        Builder& Node( nw::gfx::SceneNode *node );
        Builder& Controller( IAnimController *pController);

    private:
        nw::gfx::SceneNode* m_pNode;
        IAnimController* m_pController;
    };

    // Resolve the name, and generate multiple groups
    void GroupByName( nw::ut::MoveArray<AnimElem*> &anims );
    // Generate one group that gathers the AnimElems
    void Group( const char *name, const nw::ut::MoveArray<AnimElem*> &anims );
    // Generate a group with only one AnimElem
    void Group( AnimElem* anim );
    // From multiple animations, generate the same number of groups
    void Group( const nw::ut::MoveArray<AnimElem*> &anims );
    // Change to the previous animation set
    void ChangeToPrevAnimation();
    // Change to the next animation set
    void ChangeToNextAnimation();
    // Unbind the current AnimSet, and bind the specified AnimSet
    void ChangeTo(const char *animSetName);
    // Unbind the current AnimSet, and bind the specified index AnimSet
    void ChangeTo(s32 index);
    // Unbind the current AnimSet, and bind the StrayAnimation start AnimSet
    void ChangeToFirstStrayAnimation();
    // Unbind the animation
    void UnbindAnimation();
    // Get the name of the current animation set
    const char* GetCurrentAnimSetName();
    // Get the index of the current animation set
    s32 GetCurrentAnimSetIndex();
    // Get the scene node
    nw::gfx::SceneNode *GetSceneNode(){ return m_pNode; }
    // Release all maintained animation sets
    void Clear();
    // Rename animation set
    void RenameAnimSet(const char* oldName, const char* newName);
    // Get the number of currently bound animations
    s32 GetNumCurrentAnim(AnimElem::AnimElemType type);
    //
    BaseEvaluatorArray::iterator GetEvaluatorsBegin(){ return m_CurrentEvaluators.begin(); }
    //
    BaseEvaluatorArray::iterator GetEvaluatorsEnd(){ return m_CurrentEvaluators.end(); }
    // Destroy the resource
    void Destroy();
    // Get the current number of frames
    f32 GetFrame();
    // Get the number of frames at the end
    f32 GetEndFrame();
    // Get the animation with the longest frame
    const nw::gfx::BaseAnimEvaluator *GetLongestAnimation(){ return m_pLongestAnimEvaluator; }
    // Sort in alphabetical order
    void SortAnimSets();
    // Perform binding resolution for animation with names and order
    static void SolveAnimBindingsByName(
        nw::ut::MoveArray<AnimSetLinker*> &animSetLinkers,
        nw::ut::MoveArray<AnimElem*> &animElems,
        nw::os::IAllocator *pAllocator);
    // Erase animation element
    void RemoveAnimElem(AnimElem *anim);

private:
    // Ctor
    AnimSetLinker(nw::gfx::SceneNode *node, IAnimController *pController, nw::os::IAllocator *pAllocator);
    // Dtor
    ~AnimSetLinker();
    // Unbind the current AnimSet, and bind the specified AnimSet.
    void ChangeTo(AnimSet *animSet);
    // Attach animation set
    void AttachAnimSet(AnimSet *anim);
    // Detach animation set
    void DetachAnimSet(AnimSet *anim);
    // Search for the animation with the maximum frame count.
    void FindLongestAnimation();
    // Eras animation created by name resolution
    void DeleteStrayAnimations();
    // Pointer to the animation set currently applied
    AnimSet* m_pCurrentAnimSet;
    // Array of animations sets that can be applied
    nw::ut::MoveArray<AnimSet*> m_AnimSets;
    // Nodes for which animation is set
    nw::gfx::SceneNode* m_pNode;
    // Animation controller
    IAnimController *m_pController;
    // Currently applied Evaluator
    BaseEvaluatorArray m_CurrentEvaluators;
    // Animation with the maximum frame count
    nw::gfx::BaseAnimEvaluator* m_pLongestAnimEvaluator;
    // Allocator
    nw::os::IAllocator *m_pAllocator;
    // Animation set for unbinding
    AnimSet *m_pVoidAnimSet;
};


}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#endif

#endif
