/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_ANIM_SETUP_UTILS_H_
#define NW4C_VIEWER_G3D_ANIM_SETUP_UTILS_H_

#include "../../system/include/LogSystem.h"

namespace vw {
namespace g3d {

struct AnimBindingSet
{
    nw::gfx::BaseAnimEvaluator *eval;
    nw::gfx::Model *model;
};

/*
  

  
  
  

  
 */
inline nw::gfx::TransformAnimEvaluator*
CreateTransformAnimEvaluator(
    const int maxMembers,
    const bool translateAnimEnabled,
    nw::anim::ResAnim resAnim
)
{
    //----------------------------------------
    // Generate the animation resource.
    if (!resAnim.IsValid())
    {
        return NULL;
    }

    //----------------------------------------
    // Creates a transform animation evaluator.
    nw::gfx::TransformAnimEvaluator* evaluator = nw::gfx::TransformAnimEvaluator::Builder()
        .AnimData(resAnim)
        .MaxMembers(maxMembers)
        .MaxAnimMembers(resAnim.GetMemberAnimSetCount())
        .AllocCache(true)
        .Create(vw::g3d::Memory_GetAllocator());

    // Set the invalidation flag for movement animation.
    evaluator->SetIsTranslateDisabled(!translateAnimEnabled);

    return evaluator;
}

/*
  

  
  

  
 */
inline nw::gfx::AnimEvaluator*
CreateAnimEvaluator(
    const int maxMembers,
    nw::anim::ResAnim resAnim)
{
    //----------------------------------------
    // Generate the animation resource.
    if (!resAnim.IsValid())
    {
        return NULL;
    }

    if ( maxMembers <= 0 )
    {
        return NULL;
    }
    if ( resAnim.GetMemberAnimSetCount() <= 0 )
    {
        return NULL;
    }

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator = nw::gfx::AnimEvaluator::Builder()
        .AnimData(resAnim)
        .MaxMembers(maxMembers)
        .MaxAnimMembers(resAnim.GetMemberAnimSetCount())
        .AllocCache(true)
        .Create(vw::g3d::Memory_GetAllocator());

    return evaluator;
}


/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
InitializeSkeletalAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim,
    u32 idx = 0)
{
    nw::gfx::SkeletalModel* model = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);
    if (!model)
    {
        return NULL;
    }
    nw::gfx::AnimGroup* animGroup = model->GetSkeletalAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    nw::gfx::ResSkeletalModel resModel = model->GetResSkeletalModel();
    nw::gfx::ResSkeleton resSkeleton = resModel.GetSkeleton();
    const int maxBones = resSkeleton.GetBonesCount();
    const bool translateAnimEnabled =
        nw::ut::CheckFlag(resSkeleton.GetFlags(), nw::gfx::ResSkeletonData::FLAG_TRANSLATE_ANIMATION_ENABLED);

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::TransformAnimEvaluator* evaluator0 = CreateTransformAnimEvaluator(
        maxBones, translateAnimEnabled, resAnim);
    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Set whether it is a full-baked animation.
    model->SetFullBakedAnimEnabled(resAnim.IsFullBakedAnim());

    // Bind the animation.
    evaluator0->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    model->SetSkeletalAnimObject(evaluator0, idx);

    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
InitializeMaterialAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim,
    u32 idx = 0)
{
    nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>(node);
    if (!model)
    {
        return NULL;
    }
    nw::gfx::AnimGroup* animGroup = model->GetMaterialAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    model->SetMaterialAnimObject(evaluator0, idx);
    return evaluator0;
}


/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
InitializeVisibilityAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim,
    u32 idx = 0)
{
    nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>(node);
    if (!model)
    {
        return NULL;
    }
    nw::gfx::AnimGroup* animGroup = model->GetVisibilityAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    model->SetVisibilityAnimObject(evaluator0, idx);
    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
InitializeCameraAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Camera* camera = nw::ut::DynamicCast<nw::gfx::Camera*>(node);
    if (!camera)
    {
        return NULL;
    }
    nw::gfx::AnimGroup* animGroup = camera->GetAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    // Is the camera animation of the same type?
    nw::anim::ResCameraAnim anim(resAnim.ptr());
    if( (anim.GetProjectionUpdaterKind() != camera->GetProjectionUpdater()->Kind())
     || (anim.GetViewUpdaterKind() != camera->GetViewUpdater()->Kind()) )
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    camera->SetAnimObject(evaluator0);
    return evaluator0;
}

bool
CheckLigthType(nw::anim::res::ResLightAnim anim, nw::gfx::Light *light)
{
    nw::gfx::res::ResSceneObject resObject = light->GetResSceneObject();
    switch (resObject.GetTypeInfo())
    {
    case nw::gfx::res::ResAmbientLight::TYPE_INFO:
         return (anim.GetLightType() == nw::anim::ResLightAnimData::LIGHT_TYPE_AMBIENT);
    case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
        return (anim.GetLightType() == nw::anim::ResLightAnimData::LIGHT_TYPE_HEMISPHERE);
    case nw::gfx::res::ResVertexLight::TYPE_INFO:
        return (anim.GetLightType() == nw::anim::ResLightAnimData::LIGHT_TYPE_VERTEX)
            && (anim.GetLightKind() == nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject).GetLightKind());
    case nw::gfx::res::ResFragmentLight::TYPE_INFO:
        return (anim.GetLightType() == nw::anim::ResLightAnimData::LIGHT_TYPE_FRAGMENT)
            && (anim.GetLightKind() == nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject).GetLightKind());
    default:
        NW_LOG("Unknown light type\n");
        return false;
    }
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
InitializeLightAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Light* light = nw::ut::DynamicCast<nw::gfx::Light*>(node);
    if (!light)
    {
        return NULL;
    }
    nw::gfx::AnimGroup* animGroup = light->GetAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    // Is the light animation of the same type?
    nw::anim::ResLightAnim anim(resAnim.ptr());
    if( !CheckLigthType( anim, light ) )
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    light->SetAnimObject(evaluator0);

    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
InitializeFogAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Fog* fog = nw::ut::DynamicCast<nw::gfx::Fog*>(node);
    if (!fog)
    {
        return NULL;
    }
    nw::gfx::AnimGroup* animGroup = fog->GetAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    //----------------------------------------
    // Set animation for model.
    fog->SetAnimObject(evaluator0);

    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
CreateSkeletalAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::SkeletalModel* model = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);
    nw::gfx::AnimGroup* animGroup = model->GetSkeletalAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    nw::gfx::ResSkeletalModel resModel = model->GetResSkeletalModel();
    nw::gfx::ResSkeleton resSkeleton = resModel.GetSkeleton();
    const int maxBones = resSkeleton.GetBonesCount();
    const bool translateAnimEnabled =
        nw::ut::CheckFlag(resSkeleton.GetFlags(), nw::gfx::ResSkeletonData::FLAG_TRANSLATE_ANIMATION_ENABLED);

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::TransformAnimEvaluator* evaluator0 = CreateTransformAnimEvaluator(
        maxBones, translateAnimEnabled, resAnim);
    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Set whether it is a full-baked animation.
    model->SetFullBakedAnimEnabled(resAnim.IsFullBakedAnim());

    // Bind the animation.
    evaluator0->Bind(animGroup);

    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
CreateMaterialAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>(node);
    nw::gfx::AnimGroup* animGroup = model->GetMaterialAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    return evaluator0;
}


/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
CreateVisibilityAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Model* model = nw::ut::DynamicCast<nw::gfx::Model*>(node);
    nw::gfx::AnimGroup* animGroup = model->GetVisibilityAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
CreateCameraAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Camera* camera = nw::ut::DynamicCast<nw::gfx::Camera*>(node);
    nw::gfx::AnimGroup* animGroup = camera->GetAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    return evaluator0;
}

/*
  

  
  
 */
inline nw::gfx::BaseAnimEvaluator*
CreateLightAnim(
    nw::gfx::SceneNode* node,
    nw::anim::ResAnim resAnim)
{
    nw::gfx::Light* light = nw::ut::DynamicCast<nw::gfx::Light*>(node);
    nw::gfx::AnimGroup* animGroup = light->GetAnimGroup();

    if (animGroup == NULL) // There is no animation group for animations.
    {
        return NULL;
    }

    int maxMembers = animGroup->GetMemberCount();

    //----------------------------------------
    // Generate animation evaluation.
    nw::gfx::AnimEvaluator* evaluator0 =
        CreateAnimEvaluator(maxMembers, resAnim);

    if (evaluator0 == NULL)
    {
        return NULL;
    }

    // Bind the animation.
    evaluator0->Bind(animGroup);

    return evaluator0;
}

inline void
UnbindSkeletalAnim(
    nw::gfx::SceneNode* node,
    int idx = 0)
{
    nw::gfx::SkeletalModel* model
        = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>(node);

    if (model)
    {
        nw::gfx::BaseAnimEvaluator* eval
            = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(model->GetSkeletalAnimObject(idx));
        if (eval)
        {
            // Deletes it from the log.
            vw::system::LogSystem::GetInstance()->Clear(eval);

            model->SetSkeletalAnimObject(NULL, idx);
            eval->Reset();
            nw::gfx::SafeDestroy(eval);
        }
    }
}

inline void
UnbindMaterialAnim(
    nw::gfx::SceneNode* node,
    int idx = 0)
{
    nw::gfx::Model* model
        = nw::ut::DynamicCast<nw::gfx::Model*>(node);

    if (model)
    {
        nw::gfx::BaseAnimEvaluator* eval
            = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(model->GetMaterialAnimObject(idx));
        if (eval)
        {
            // Deletes it from the log.
            vw::system::LogSystem::GetInstance()->Clear(eval);

            model->SetMaterialAnimObject(NULL, idx);
            eval->Reset();
            nw::gfx::SafeDestroy(eval);
        }
    }
}

inline void
UnbindVisibilityAnim(
    nw::gfx::SceneNode* node,
    int idx = 0)
{
    nw::gfx::Model* model
        = nw::ut::DynamicCast<nw::gfx::Model*>(node);

    if (model)
    {
        nw::gfx::BaseAnimEvaluator* eval
            = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(model->GetVisibilityAnimObject(idx));
        if (eval)
        {
            // Deletes it from the log.
            vw::system::LogSystem::GetInstance()->Clear(eval);

            model->SetVisibilityAnimObject(NULL, idx);
            eval->Reset();
            nw::gfx::SafeDestroy(eval);
        }
    }
}

inline void
UnbindCameraAnim( nw::gfx::SceneNode* node )
{
    nw::gfx::Camera* camera
        = nw::ut::DynamicCast<nw::gfx::Camera*>(node);

    if (camera)
    {
        nw::gfx::BaseAnimEvaluator* eval
            = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(camera->GetAnimObject());
        if (eval)
        {
            // Deletes it from the log.
            vw::system::LogSystem::GetInstance()->Clear(eval);

            camera->SetAnimObject(NULL);
            eval->Reset();
            nw::gfx::SafeDestroy(eval);
        }
    }
}

inline void
UnbindLightAnim( nw::gfx::SceneNode* node )
{
    nw::gfx::Light* light
        = nw::ut::DynamicCast<nw::gfx::Light*>(node);

    if (light)
    {
        nw::gfx::BaseAnimEvaluator* eval
            = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(light->GetAnimObject());
        if (eval)
        {
            // Deletes it from the log.
            vw::system::LogSystem::GetInstance()->Clear(eval);

            light->SetAnimObject(NULL);
            eval->Reset();
            nw::gfx::SafeDestroy(eval);
        }
    }
}

inline void
UnbindFogAnim( nw::gfx::SceneNode* node )
{
    nw::gfx::Fog* fog
        = nw::ut::DynamicCast<nw::gfx::Fog*>(node);

    if (fog)
    {
        nw::gfx::BaseAnimEvaluator* eval
            = nw::ut::DynamicCast<nw::gfx::BaseAnimEvaluator*>(fog->GetAnimObject());
        if (eval)
        {
            // Deletes it from the log.
            vw::system::LogSystem::GetInstance()->Clear(eval);

            fog->SetAnimObject(NULL);
            eval->Reset();
            nw::gfx::SafeDestroy(eval);
        }
    }
}

}
}
#endif
