/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_INFO_DRAWER_H_
#define NW4C_VIEWER_G3D_INFO_DRAWER_H_

#include "SceneElemSelector.h"
#include "SceneRenderer.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw  {
namespace g3d {


/*===========================================================================*
  Name:         InfoDrawer

  Description:  information display
 *===========================================================================*/
class InfoDrawer
{
public:
    enum SelectType {
        SELECT_VIEW_CAMERA = 0,
        //SELECT_CONTROL_TARGET_TYPE,
        SELECT_CONTROL_TARGET,
        //SELECT_RESOURCE_TYPE,
        SELECT_RESOURCE,
        SELECT_ANIMATION,
        SELECT_SCENE,
        SELECT_ANIMATION_FRAME,
        SELECT_ANIMATION_RATE,
        SELECT_NUM
    };

    enum ControlTargetType {
        CONTROL_CAMERA = 0,
        CONTROL_LIGHT,
        CONTROL_NUM
    };

    enum InfoResType {
        INFO_MODEL = SceneElemSelector::ELEM_MODEL,
        INFO_CAMERA = SceneElemSelector::ELEM_CAMERA,
        INFO_LIGHT = SceneElemSelector::ELEM_LIGHT,
        INFO_FOG = SceneElemSelector::ELEM_FOG,
        INFO_NUM
    };

    static InfoDrawer* GetInstance();

    // Instance initialization, finalization
    static void Initialize();
    static void Finalize();

    // Initialization when using menus
    void InitializeMenu();

    // Key input
    void Main(vw::system::HidStatus &hidStatus);

    // Display the measurement result
    void DrawInfo();

    // Set the current number of frames
    void SetFrameCount(f32 frame);

    // Set the maximum number of frames
    void SetMaxFrame(f32 frame);

    // Set frame rate
    void SetFrameRate(f32 frameRate);

    // Sets sub-mode name.
    void SetSubModeName(const char* name);

    // Set view camera name
    void SetViewCameraName(const char* name);

    // Set operation target type.
    void SetControlTargetType(const ControlTargetType type);

    // Set operation target name
    void SetControlTargetName(const char* name);

    // Set display resource type
    void SetResourceType(const InfoResType type);

    // Set resource name
    void SetResouceName(const char* name);

    // Configuring an Animation
    void SetAnimationGroup(const char* name, bool skeletal, bool material, bool visibility);

    // Sets the scene.
    void SetSceneName(const char* name);

    // Render seek bar
    void DrawSeekBar(f32 frameCount, f32 maxFrame);

    // Seek bar information
    bool GetSeekBarInfo(float& value);

    // Gets the execution command of the selected target.
    s32 GetSelectionCommand( s32 padEvent ) const;

    // Selected type
    SelectType GetSelectType() const;

    // Return the selected target to what it was.
    void ChangeToPrevSelection();

    // Advance the selected target to the next step.
    void ChangeToNextSelection();

    // Control target type
    ControlTargetType GetControlTargetType() const;

    // Returns the selected target control type to what it was.
    void ChangeToPrevControlTargetType();

    // Advances the selected target to the next step.
    void ChangeToNextControlTargetType();

    // The control target checks the first one.
    bool IsFirstControlTarget();

    // The control target checks the end one.
    bool IsLastControlTarget();

    // Returns the operation target matrix to what it was before.
    void ChangeToFirstControlTarget();

    // Advances to the end of the control target.
    void ChangeToLastControlTarget();

    // Returns the operation target matrix to what it was before.
    void ChangeToPrevControlTarget();

    // Advances to the end of the control target.
    void ChangeToNextControlTarget();

    // Display resource type
    InfoResType GetInfoResourceType() const;

    // Return the display target to what it was.
    void ChangeToPrevResourceType();

    // Advances the display resource target to the next step.
    void ChangeToNextResourceType();

    // Existence of 2D information display
    bool IsRender2DInfo() const;

    // Existence of 3D information display
    bool IsRender3DInfo() const;

    // Display mode setting [0: Display all; 1: Display only 2D information; 2: Hide all]
    void SetRenderInfoFlag(u32 flag);

    // Switches display modes.
    void ChangeRenderInfoFlag();

    // Gets the display flags for various 3D information.
    bool* GetRenderFlag();

    // Switches resources using the touch panel.
    void GetTouchPanelEvent(const vw::system::HidStatus& hidStatus, s32& padEvent);

private:
    void SetupFontBuffer();

private:
    // Constructor
    InfoDrawer();
    // Destructor
    ~InfoDrawer();

    const char*             mSubModeName;
    SelectType              mSelectType;
    const char*             mViewCameraName;
    ControlTargetType       mControlTargetType;
    const char*             mControlTargetName;
    InfoResType             mResType;
    const char*             mResName;
    const char*             mAnimationName;
    bool                    mAnimationType[3];
    const char*             mSceneName;
    f32                     mFrameCount;
    f32                     mMaxFrame;
    f32                     mFrameRate;
    s32                     mRenderInfoFlag;
    bool                    mRenderFlag[SceneRenderer::RenderNum];
    nw::ut::Color8          mSelectColor;
    vw::system::FontBuffer  mFontBuffer;
    bool                    mFontCacheSetup;
    bool                    mTouch;

    static InfoDrawer*      mInstance;
};



}   /* namespace g3d */
}   /* namespace vw  */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif


/* NW4C_VIEWER_G3D_SUB_MODE_EM4C_H_ */
#endif
