/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_G3D_SCENE_ELEM_H_
#define NW4C_VIEWER_G3D_SCENE_ELEM_H_

#include <stdlib.h>
#include "ResourceLinker.h"
#include "AnimController.h"
#include "NodeConstraint.h"

#include <nw/ut/ut_LinkList.h>

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
namespace vw {
namespace g3d {
    class AnimSetLinker;
    class IAnimController;
}
}

namespace nw     {
namespace viewer {
namespace g3d    {

///
/// Class to allow access from the interface class when making into a library
///
class NodeEntity
{
public:
    // Get AnimLinker
    virtual vw::g3d::AnimSetLinker* GetAnimLinker() = 0;
    // Get AnimSetLinker
    virtual vw::g3d::IAnimController* GetAnimController() = 0;
    // Get SceneNode
    virtual nw::gfx::SceneObject* GetSceneObject() = 0;
    // Dtor
    virtual ~NodeEntity(){};
};

} // g3d
} // viewer
} // nw

namespace vw {
namespace g3d {

class ResourceSet;
class AnimSetLinker;

#define SceneElemBuilder( classname )\
class Builder\
{\
public:\
    Builder() : m_ResSource( NULL ), m_pSceneNode( NULL ), m_Destroy( true ) {}\
    ~Builder() {}\
\
    classname<NODE, RESOURCE>* \
    Create(nw::os::IAllocator *pAllocator, nw::os::IAllocator *pDeviceAllocator)\
    {\
        void *buf = pAllocator->Alloc(sizeof(classname<NODE, RESOURCE>));\
        if (buf == NULL) return NULL;\
\
        classname<NODE, RESOURCE>* elem = new(buf) classname<NODE, RESOURCE>(\
            m_ResSource, m_ResObject, m_pSceneNode, m_Destroy,\
            pAllocator, pDeviceAllocator\
        );\
\
        return elem;\
    }\
\
    Builder&\
    ResSource(ResourceSet* resSource)\
    { \
        m_ResSource = resSource;\
        return *this;\
    }\
\
    Builder&\
    SceneNode( NODE *node, bool destroyNode = true )\
    {\
        m_pSceneNode = node;\
        m_Destroy = destroyNode;\
        return *this;\
    }\
\
    Builder&\
    ResObject(RESOURCE resObject)\
    { \
        m_ResObject = resObject;\
        return *this;\
    }\
\
    Builder&\
    DestroyNode(bool destroyNode)\
    { \
        m_Destroy = destroyNode;\
        return *this;\
    }\
\
private:\
    ResourceSet* m_ResSource;\
    RESOURCE m_ResObject;\
    NODE *m_pSceneNode;\
    bool m_Destroy;\
};

template <class RESOURCE>
class ResElem
{
public:
    nw::ut::LinkListNode m_Link;

    class Builder
    {
    public:
        Builder() : m_ResSource( NULL ) {}
        ~Builder() {}

        ResElem<RESOURCE>*
        Create(nw::os::IAllocator *pAllocator, nw::os::IAllocator *pDeviceAllocator)
        {
            void *buf = pAllocator->Alloc(sizeof(ResElem<RESOURCE>));
            if (buf == NULL) return NULL;

            ResElem<RESOURCE>* elem = new(buf) ResElem<RESOURCE>(
                m_ResSource, m_ResObject,
                pAllocator, pDeviceAllocator
            );

            return elem;
        }

        Builder&
        ResSource(ResourceSet* resSource)
        {
            m_ResSource = resSource;
            return *this;
        }

        Builder&
        ResObject(RESOURCE resObject)
        {
            m_ResObject = resObject;
            return *this;
        }

    private:
        ResourceSet* m_ResSource;
        RESOURCE m_ResObject;
    };

    virtual void Destroy()
    {
        this->~ResElem();
        m_pAllocator->Free( this );
    }

    ResourceSet*
    GetResSource()
    {
        NW_NULL_ASSERT( m_ResSource );
        return m_ResSource;
    }

    RESOURCE
    GetResObject()
    {
        return m_ResObject;
    }

protected:
    ResourceSet* m_ResSource;
    RESOURCE m_ResObject;
    nw::os::IAllocator* m_pAllocator;

    ResElem(
        ResourceSet* resSource,
        RESOURCE resObject,
        nw::os::IAllocator* pAllocator,
        nw::os::IAllocator* pDeviceAllocator
    )
    {
        m_ResSource = resSource;
        //NW_NULL_ASSERT( m_ResSource );
        m_ResObject = resObject;
        m_pAllocator = pAllocator;
        NW_NULL_ASSERT( m_pAllocator );
    }

    virtual ~ResElem()
    {
    }
};

template <class NODE, class RESOURCE>
class SceneElem : public ResElem<RESOURCE>, public nw::viewer::g3d::NodeEntity
{
public:
    SceneElemBuilder( SceneElem )

    virtual void
    Destroy()
    {
        this->~SceneElem();
        this->m_pAllocator->Free( this );
    }

    virtual AnimSetLinker*
    GetAnimLinker()
    {
        return NULL;
    }

    virtual IAnimController*
    GetAnimController()
    {
        return NULL;
    }

    virtual nw::gfx::SceneObject *
    GetSceneObject()
    {
        return m_pNode;
    }

    NODE*
    GetSceneNode()
    {
        return m_pNode;
    }

    ResourceLinker*
    GetResourceLinker()
    {
        return m_pResourceLinker;
    }

    const char*
    GetName()
    {
        return m_pName;
    }

protected:
    NODE *m_pNode;
    ResourceLinker *m_pResourceLinker;
    const char *m_pName;
    bool m_NameBufferAllocated;
    bool m_DestroyNode;

    SceneElem(
        ResourceSet* resSource,
        RESOURCE resObject,
        NODE* sceneNode,
        bool destroyNode,
        nw::os::IAllocator* pAllocator,
        nw::os::IAllocator* pDeviceAllocator
    ) : ResElem<RESOURCE>( resSource, resObject, pAllocator, pDeviceAllocator ),
        m_DestroyNode( destroyNode )
    {
        if (sceneNode) {
            m_pNode = sceneNode;
        } else {
            nw::gfx::SceneObject* sceneObject = nw::gfx::SceneBuilder()
                .Resource( resObject )
                .BufferOption( nw::gfx::Model::MULTI_FLAG_BUFFER_MATERIAL )
                .MaxAnimObjectsPerGroup( 128 )
                .IsFixedSizeMemory( false )
                .CreateObject(pAllocator, pDeviceAllocator);
            m_pNode = nw::ut::DynamicCast<NODE*>(sceneObject);
        }

        m_pResourceLinker = ResourceLinker::Builder().Create( pAllocator );
        NW_NULL_ASSERT( m_pResourceLinker );
        m_pResourceLinker->ParseLink( (nw::gfx::SceneNode*)m_pNode );

        // Convert name if necessary.
        {
            const char *identifier     = "3DEditor";                    // Identifier for determination
            const s32 identifierLength = strlen(identifier);            // Number of characters for identifier
            s32 digit                  = 3;                             // Number of digits for number at end
            size_t strLength           = 0;
            if ( resObject.GetName() ) strLength = strlen( resObject.GetName() ); // Number of characters in original name

            if ( resObject.GetName() == NULL ) {
                const char* defaultName = "Nameless Node";
                char* nameBuffer = (char*)pAllocator->Alloc(strlen(defaultName) + 1);
                strcpy(nameBuffer, defaultName);
                m_pName = nameBuffer;
                m_NameBufferAllocated = true;
            } else if ( strLength >= (identifierLength + digit)
              && !strncmp( resObject.GetName() + strLength - digit - identifierLength, identifier, identifierLength))
            {
                // When a special filename ("3DEditor***" is at the end, with *** being a 3-digit number) is attached from the 3DEditor
                s32 instanceNameLength = atoi( resObject.GetName() + strLength - digit );
                char* nameBuffer = (char*)pAllocator->Alloc( strLength - identifierLength - digit + 1 );
                {
                    // Copy the string, and convert the underbars to parentheses
                    strncpy( nameBuffer, resObject.GetName(), strLength - identifierLength - digit );
                    nameBuffer[ instanceNameLength ] = '(';
                    nameBuffer[ strLength - identifierLength - digit - 1 ] = ')';
                    nameBuffer[ strLength - identifierLength - digit ] = '\0';
                }
                m_pName = nameBuffer;
                m_NameBufferAllocated = true;
            }
            else
            {
                m_pName = resObject.GetName();
                m_NameBufferAllocated = false;
            }
        }

        NW_NULL_ASSERT( m_pNode );
    }

    virtual ~SceneElem()
    {
        m_pResourceLinker->Destroy();
        if (m_DestroyNode)
        {
            nw::gfx::SafeDestroy( m_pNode );
        }
        if (m_NameBufferAllocated)
        {
            this->m_pAllocator->Free( (void*)(m_pName) );
        }
    }
};

template <class NODE, class RESOURCE>
class AnimSceneElem : public SceneElem<NODE,RESOURCE>
{
public:
    SceneElemBuilder( AnimSceneElem )

    virtual void Destroy()
    {
        this->~AnimSceneElem();
        this->m_pAllocator->Free( this );
    }

    virtual AnimSetLinker*
    GetAnimLinker()
    {
        return m_pAnimController->GetAnimSetLinker();
    }

    virtual IAnimController*
    GetAnimController()
    {
        return m_pAnimController;
    }

protected:
    AnimController *m_pAnimController;

    AnimSceneElem(
        ResourceSet* resSource,
        RESOURCE resObject,
        NODE* sceneNode,
        bool destroyNode,
        nw::os::IAllocator* pAllocator,
        nw::os::IAllocator* pDeviceAllocator
    ) : SceneElem<NODE,RESOURCE>( resSource, resObject, sceneNode, destroyNode, pAllocator, pDeviceAllocator )
    {
        m_pAnimController = AnimController::Builder().Node(this->m_pNode).Create(pAllocator);
        NW_NULL_ASSERT( m_pAnimController );
    }

    virtual ~AnimSceneElem()
    {
        m_pAnimController->Destroy();
    }
};

template <class NODE, class RESOURCE>
class ConstraintSceneElem : public AnimSceneElem<NODE,RESOURCE>
{
public:
    SceneElemBuilder( ConstraintSceneElem )

    virtual void Destroy()
    {
        this->~ConstraintSceneElem();
        this->m_pAllocator->Free( this );
    }

    NodeConstraint&
    GetNodeConstraint()
    {
        return m_Constraints;
    }

protected:
    NodeConstraint m_Constraints;

    ConstraintSceneElem(
        ResourceSet* resSource,
        RESOURCE resObject,
        NODE* sceneNode,
        bool destroyNode,
        nw::os::IAllocator* pAllocator,
        nw::os::IAllocator* pDeviceAllocator
    ) : AnimSceneElem<NODE,RESOURCE>( resSource, resObject, sceneNode, destroyNode, pAllocator, pDeviceAllocator ),
        m_Constraints( this->m_pNode, pAllocator )
    {
    }

    virtual ~ConstraintSceneElem()
    {
    }
};

typedef ConstraintSceneElem<nw::gfx::Model, nw::gfx::res::ResModel> ModelElem;
typedef ResElem<nw::gfx::res::ResTexture> TextureElem;
typedef ResElem<nw::gfx::res::ResLookupTableSet> LookupTableSetElem;
typedef AnimSceneElem<nw::gfx::Camera, nw::gfx::res::ResCamera> CameraElem;
typedef AnimSceneElem<nw::gfx::Light, nw::gfx::res::ResLight> LightElem;
typedef AnimSceneElem<nw::gfx::Fog, nw::gfx::res::ResFog> FogElem;
typedef SceneElem<nw::gfx::SceneEnvironmentSetting, nw::gfx::res::ResSceneEnvironmentSetting> SceneEnvironmentSettingElem;

}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif


#endif
