/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/SceneManager.h"
#include "../include/AnimElem.h"
#include "../include/AnimController.h"
#include "../../effect/include/AnimManager.h"
#include "../include/AnimManager.h"
#include <viewer/system.h>

namespace vw  {
namespace g3d {

AnimController::AnimController( nw::gfx::SceneNode *node, nw::os::IAllocator* pAllocator )
    : m_IsActive( true )
{
    m_pLinker = AnimSetLinker::Builder().Node(node).Controller(this).Create(pAllocator);
    m_pAllocator = pAllocator;

    if ( NULL != nw::ut::DynamicCast<nw::gfx::ParticleModel*>(node) )
    {
        m_StepFrame = effect::AnimManager::GetInstance()->GetStepFrame();
    }
    else
    {
        m_StepFrame = AnimManager::GetInstance()->GetDefaultRate();
    }
}

AnimController::~AnimController()
{
    m_pLinker->Destroy();
}

void
AnimController::Update()
{
    if (m_IsActive)
    {
        if (vw::system::App::GetInstance().GetFocusID() == vw::system::EFFECT_ID)
        {
            vw::effect::AnimManager *effectAnimManager = vw::effect::AnimManager::GetInstance();
            // Before the start frame is not updated.
            if (effectAnimManager->GetFrame() < m_StartFrame)
            {
                SetFrame( 0.f );
                return;
            }

            // When the global start frame is before the local start frame, it is not updated.
            // 
            if (m_StartFrame < effectAnimManager->GetStartFrame())
            {
                SetFrame( 0.f );
                return;
            }

            // For the instant that the start frame straddles, the difference is the set frame and update is not performed.
            if (effectAnimManager->GetFrame() >= m_StartFrame
             && effectAnimManager->GetPreFrame() <= m_StartFrame )
            {
                this->SetFrame( effectAnimManager->GetFrame() - m_StartFrame);
                return;
            }
        }

        // Auto repeat process for a one-time animation
        if ( AnimManager::GetInstance()->GetLoopFlag() )
        {
            AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
            AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
            for (; it!=end_it; ++it)
            {
                if (m_StepFrame < 0.f && (*it)->GetFrame() == 0
                && (*it)->GetAnimData().GetLoopMode() == nw::anim::res::ResAnimData::LOOP_MODE_ONETIME)
                {
                    (*it)->SetFrame( (*it)->GetEndFrame() );
                }
                else if (m_StepFrame > 0.f && (*it)->GetFrame() >= (*it)->GetEndFrame() - FLT_EPSILON
                    &&  (*it)->GetAnimData().GetLoopMode() == nw::anim::res::ResAnimData::LOOP_MODE_ONETIME)
                {
                    (*it)->SetFrame( 0.f );
                }
            }
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::SetFrame(f32 frame)
{
    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->SetFrame( frame );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
f32
AnimController::GetFrame()
{
    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        return (*it)->GetFrame();
    }

    return 0.f;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::SetStepFrame(f32 step)
{
    m_StepFrame = step;

    if (m_IsActive)
    {
        AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
        AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
        for (; it!=end_it; ++it)
        {
            (*it)->SetStepFrame( step );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::StepForward()
{
    m_IsActive = true;
    this->Update();
    m_IsActive = false;

    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->SetFrame( (*it)->GetFrame() + m_StepFrame );
        (*it)->SetStepFrame( 0.f );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::StepBackward()
{
    m_IsActive = false;

    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->SetFrame( (*it)->GetFrame() - m_StepFrame );
        (*it)->SetStepFrame( 0.f );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::PauseAndPlay()
{
    m_IsActive = !m_IsActive;
    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        if ( m_IsActive ) (*it)->SetStepFrame( m_StepFrame );
        else                 (*it)->SetStepFrame( 0.f );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::Play()
{
    m_IsActive = true;
    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        if (m_StepFrame < 0.f && (*it)->GetFrame() == 0
        && (*it)->GetAnimData().GetLoopMode() == nw::anim::res::ResAnimData::LOOP_MODE_ONETIME)
        {
            (*it)->SetFrame( (*it)->GetEndFrame() );
        }
        else if (m_StepFrame > 0.f && (*it)->GetFrame() >= (*it)->GetEndFrame() - FLT_EPSILON
            &&  (*it)->GetAnimData().GetLoopMode() == nw::anim::res::ResAnimData::LOOP_MODE_ONETIME)
        {
            (*it)->SetFrame( 0 );
        }
        (*it)->SetStepFrame( m_StepFrame );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::Pause()
{
    m_IsActive = false;
    AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
    AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->SetStepFrame( 0.f );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::SpeedUp(f32 step)
{
    SetStepFrame( m_StepFrame + step );
}


/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::Stop()
{
    SetFrame( 0.f );
    Pause();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimController::Reset()
{
    SetFrame( 0.f );
}

bool
AnimController::IsActive()
{
    if (!m_IsActive)
    {
        return false;
    }
    else
    {
        // Animations that are not looped may be stopped even if m_IsActive=true, so if even one evaluator is running for a check, TRUE is returned.
        // 
        // 
        bool isActive = false;
        AnimSetLinker::BaseEvaluatorArray::iterator it = m_pLinker->GetEvaluatorsBegin();
        AnimSetLinker::BaseEvaluatorArray::iterator end_it = m_pLinker->GetEvaluatorsEnd();
        for (; it!=end_it; ++it)
        {
            if (m_StepFrame < 0.f && (*it)->GetFrame() == 0
            && (*it)->GetAnimData().GetLoopMode() == nw::anim::res::ResAnimData::LOOP_MODE_ONETIME)
            {
                continue;
            }
            else if (m_StepFrame > 0.f && (*it)->GetFrame() >= (*it)->GetEndFrame() - FLT_EPSILON
                &&  (*it)->GetAnimData().GetLoopMode() == nw::anim::res::ResAnimData::LOOP_MODE_ONETIME)
            {
                continue;
            }
            else
            {
                isActive = true;
                break;
            }
        }

        return isActive;
    }
}

}   /* namespace g3d */
}   /* namespace vw  */

