/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/AnimElem.h"
#include "../include/AnimSetupUtils.h"
#include "../include/AnimController.h"
#include "../include/SceneManager.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
#define FOREACH_ELEM( type, iteratorName, name ) \
    for (SceneManager::type::iterator iteratorName = SceneManager::GetInstance()->Get##name##sBegin(), end_it_ = SceneManager::GetInstance()->Get##name##sEnd(); \
        iteratorName != end_it_; ++iteratorName )

namespace vw {
namespace g3d {

class AnimSet
{
public:
    // Get start iterator
    nw::ut::MoveArray<AnimElem*>::iterator
    GetAnimElemBegin( void )
    {
        return m_AnimElems.begin();
    }

    // Get end iterator
    nw::ut::MoveArray<AnimElem*>::iterator
    GetAnimElemEnd( void )
    {
        return m_AnimElems.end();
    }

    // Get the animation count
    s32
    GetAnimNum( void )
    {
        return m_AnimElems.size();
    }

    // Get the animation count
    s32
    GetAnimNum( const AnimElem::AnimElemType type )
    {
        return m_AnimNum[ type ];
    }

    void
    Clear()
    {
        memset(m_AnimNum, 0, sizeof(u32) * AnimElem::Anim_Num);
        m_AnimElems.clear();
        m_CurrentFrame = 0.0f;
    }

    // Get name
    const char *
    GetName() const
    {
        return m_Name;
    }

    void
    SetName( const char *name )
    {
        strcpy( m_Name, name );
    }

    // Resource addition
    void
    AddAnim( AnimElem* anim )
    {
        m_AnimElems.push_back( anim );
        m_AnimNum[ anim->GetAnimType() ]++;
    }

    // Remove resources.
    void
    RemoveAnimElem( AnimElem *anim )
    {
        nw::ut::MoveArray<AnimElem*>::iterator it = m_AnimElems.begin();
        nw::ut::MoveArray<AnimElem*>::iterator end_it = m_AnimElems.end();
        while (it!=end_it)
        {
            if ((*it) == anim)
            {
                nw::ut::MoveArray<AnimElem*>::iterator next = it;
                ++next;
                it = m_AnimElems.erase(it, next);
                end_it = m_AnimElems.end();
            }
            else
            {
                ++it;
            }
        }
    }

    bool
    IsStray() const
    {
        return m_IsStray;
    }

    f32
    GetCurrentFrame() const
    {
        return m_CurrentFrame;
    }

    void
    SetCurrentFrame( const f32 frame )
    {
        m_CurrentFrame = frame;
    }

    class Builder
    {
    public:
        Builder() : m_IsStray(false)
        {
        }

        Builder&
        Name(const char* name)
        {
            strcpy(m_Name, name);
            return *this;
        }

        AnimSet*
        Create(nw::os::IAllocator *pAllocator)
        {
            void *buf = pAllocator->Alloc(sizeof(AnimSet));
            if (buf == NULL) return NULL;

            AnimSet* pAnimSet = new(buf) AnimSet( pAllocator, m_IsStray );
            memcpy( pAnimSet->m_Name, m_Name, 256 );

            return pAnimSet;
        }

        Builder&
        IsStray(bool isStray)
        {
            m_IsStray = isStray;
            return *this;
        }
    private:
        char m_Name[256];
        bool m_IsStray;
    };

    void
    Destroy()
    {
        this->~AnimSet();
        m_pAllocator->Free(this);
    }

private:

    AnimSet(nw::os::IAllocator *pAllocator, bool isStray )
    {
        m_pAllocator = pAllocator;
        m_AnimElems = nw::ut::MoveArray<AnimElem*>(pAllocator);
        memset(m_AnimNum, 0, sizeof(u32) * AnimElem::Anim_Num);
        m_IsStray = isStray;
        m_CurrentFrame = 0.0f;
    }

    ~AnimSet() {}

    char  m_Name[256];
    u32 m_AnimNum[ AnimElem::Anim_Num ];
    nw::ut::MoveArray<AnimElem*> m_AnimElems;
    bool m_IsStray;
    f32 m_CurrentFrame;
    nw::os::IAllocator *m_pAllocator;
};

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::Builder::Builder() : m_ResSource( NULL )
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::Builder::~Builder()
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem*
AnimElem::Builder::Create(nw::os::IAllocator *pAllocator, nw::os::IAllocator *pDeviceAllocator)
{
    void *buf = pAllocator->Alloc(sizeof(AnimElem));
    if (buf == NULL) return NULL;

    AnimElem* pAnimElem = new(buf) AnimElem(
        m_AnimType,
        m_ResSource,
        m_ResAnim,
        pAllocator,
        pDeviceAllocator);

    return pAnimElem;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::Builder&
AnimElem::Builder::ResSource(ResourceSet* resSource)
{
    m_ResSource = resSource;
    return *this;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::Builder&
AnimElem::Builder::ResObject(nw::anim::ResAnim resAnim)
{
    m_ResAnim = resAnim;
    return *this;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::Builder&
AnimElem::Builder::AnimType(AnimElemType type)
{
    m_AnimType = type;
    return *this;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::AnimElem(
    AnimElemType animType,
    ResourceSet* resSource,
    nw::anim::ResAnim resObject,
    nw::os::IAllocator* pAllocator,
    nw::os::IAllocator* pDeviceAllocator
) : ResElem<nw::anim::ResAnim>( resSource, resObject, pAllocator, pDeviceAllocator )
{
    m_AnimType = animType;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimElem::~AnimElem()
{
    switch (m_AnimType)
    {
    case Anim_Skeletal:
    case Anim_Material:
    case Anim_Visibility:
        {
            FOREACH_ELEM( ModelList, it, Model ) { it->GetAnimLinker()->RemoveAnimElem(this); }
        }
        break;
    case Anim_Light:
        {
            FOREACH_ELEM( LightList, it, Light ) { it->GetAnimLinker()->RemoveAnimElem(this); }
        }
        break;
    case Anim_Camera:
        {
            FOREACH_ELEM( CameraList, it, Camera ) { it->GetAnimLinker()->RemoveAnimElem(this); }
        }
        break;
    case Anim_Fog:
        {
            FOREACH_ELEM( FogList, it, Fog ) { it->GetAnimLinker()->RemoveAnimElem(this); }
        }
        break;
    default:
        break;
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimElem::Destroy()
{
    this->~AnimElem();
    m_pAllocator->Free(this);
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker::Builder::Builder() : m_pNode(NULL)
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker::Builder::~Builder()
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker*
AnimSetLinker::Builder::Create( nw::os::IAllocator *pAllocator )
{
    NW_NULL_ASSERT( m_pNode );
    NW_NULL_ASSERT( m_pController );
    void *buf = pAllocator->Alloc(sizeof(AnimSetLinker));
    if (buf == NULL) return NULL;

    AnimSetLinker* pLinker = new(buf) AnimSetLinker( m_pNode, m_pController, pAllocator );
    return pLinker;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker::Builder&
AnimSetLinker::Builder::Node( nw::gfx::SceneNode *node )
{
    m_pNode = node;
    return *this;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker::Builder&
AnimSetLinker::Builder::Controller( IAnimController *pController )
{
    m_pController = pController;
    return *this;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSet*
FindAnimSetByName(nw::ut::MoveArray<AnimSet*> &animSets, const char* name)
{
    nw::ut::MoveArray<AnimSet*>::iterator it = animSets.begin();
    nw::ut::MoveArray<AnimSet*>::const_iterator end_it = animSets.end();
    for (; it!=end_it; ++it)
    {
        if ( (*it)->IsStray() && !strcmp((*it)->GetName(), name) )
        {
            return *it;
        }
    }
    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::SolveAnimBindingsByName(
    nw::ut::MoveArray<AnimSetLinker*> &animSetLinkers,
    nw::ut::MoveArray<AnimElem*> &animElems,
    nw::os::IAllocator *pAllocator)
{
    // Separate AnimElem into groups by name and generate AnimSet
    nw::ut::MoveArray<AnimSet*> animSets(pAllocator);
    {
        nw::ut::MoveArray<AnimElem*>::iterator it = animElems.begin();
        nw::ut::MoveArray<AnimElem*>::const_iterator end_it = animElems.end();
        for (; it!=end_it; ++it)
        {
            AnimSet* animSet = FindAnimSetByName(animSets, (*it)->GetResAnim().GetName());
            if (animSet)
            {
                animSet->AddAnim( *it );
            }
            else
            {
                animSet = AnimSet::Builder().Name( (*it)->GetResAnim().GetName() )
                                            .IsStray( true )
                                            .Create(pAllocator);
                animSet->AddAnim( *it );
                animSets.push_back( animSet );
            }
        }
    }

    // At the same time as setting AnimSet into each Linker, generate linker array
    nw::ut::MoveArray<AnimSetLinker*> linkers(pAllocator);
    for (int i=0; i<animSetLinkers.size(); ++i)
    {
        for (int j=0; j<animSets.size(); ++j)
        {
            // Generate copy and set (because AnimSet is maintained and released inside each linker)
            AnimSet* animSet = AnimSet::Builder().Name( animSets[j]->GetName() ).IsStray(true).Create(pAllocator);
            nw::ut::MoveArray<AnimElem*>::iterator it = animSets[j]->GetAnimElemBegin();
            nw::ut::MoveArray<AnimElem*>::iterator end_it = animSets[j]->GetAnimElemEnd();
            for (; it!=end_it; ++it) animSet->AddAnim( *it );
            animSetLinkers[i]->m_AnimSets.push_back( animSet );
        }
        // Sort in name order
        animSetLinkers[i]->SortAnimSets();
        linkers.push_back( animSetLinkers[i] );
    }

    // Activate the animation set that matches the name
    nw::ut::MoveArray<AnimSetLinker*>::iterator linker_it = linkers.begin();
    while (linker_it != linkers.end())
    {
        bool animNotFound = true;
        nw::ut::MoveArray<AnimSet*>::iterator animset_it = animSets.begin();
        while (animset_it != animSets.end())
        {
            if(!strcmp((*linker_it)->GetSceneNode()->GetName(), (*animset_it)->GetName()))
            {
                (*linker_it)->ChangeTo( (*animset_it)->GetName() );
                linker_it = linkers.erase(linker_it, linker_it+1);
                (*animset_it)->Destroy();
                animset_it = animSets.erase(animset_it, animset_it+1);
                animNotFound = false;
                break;
            }
            else
            {
                ++animset_it;
            }
        }

        if (animNotFound) ++linker_it;
    }

    // Bind the linker and animation for which the name could not be resolved in order
    linker_it = linkers.begin();
    nw::ut::MoveArray<AnimSet*>::iterator animset_it = animSets.begin();
    for (;linker_it!=linkers.end() && animset_it!=animSets.end(); ++linker_it, ++animset_it)
    {
        (*linker_it)->ChangeTo( (*animset_it)->GetName() );
    }

    // Release the generated animation set
    for (int i=0; i<animSets.size(); ++i)
    {
        animSets[i]->Destroy();
    }
}
/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::GroupByName( nw::ut::MoveArray<AnimElem*> &anims )
{
    DeleteStrayAnimations();

    nw::ut::MoveArray<AnimSet*> animSets(m_pAllocator);
    nw::ut::MoveArray<AnimElem*>::iterator it = anims.begin();
    nw::ut::MoveArray<AnimElem*>::const_iterator end_it = anims.end();
    for (; it!=end_it; ++it)
    {
        AnimSet* animSet = FindAnimSetByName(animSets, (*it)->GetResAnim().GetName());
        if (animSet)
        {
            animSet->AddAnim( *it );
        }
        else
        {
            animSet = AnimSet::Builder().Name( (*it)->GetResAnim().GetName() ).IsStray(true).Create(m_pAllocator);
            animSet->AddAnim( *it );
            animSets.push_back( animSet );
        }
    }

    for (int i=0; i<animSets.size(); ++i)
    {
        m_AnimSets.push_back( animSets[i] );
    }

    SortAnimSets();

    if (m_pCurrentAnimSet == m_pVoidAnimSet)  ChangeToFirstStrayAnimation();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::DeleteStrayAnimations()
{
    AnimSet *unbindAnimSet = NULL;
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    while (it != m_AnimSets.end())
    {
        if ( (*it)->IsStray() )
        {
            nw::ut::MoveArray<AnimSet*>::iterator next = it;
            ++next;
            if ( m_pCurrentAnimSet == *it )
            {
                DetachAnimSet( m_pCurrentAnimSet );
                m_pCurrentAnimSet = NULL;
            }
            (*it)->Destroy();
            it = m_AnimSets.erase( it, next );
        }
        else
        {
            ++it;
        }
    }

    if (!m_pCurrentAnimSet)
    {
        m_pCurrentAnimSet = m_pVoidAnimSet;
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::Group( const char *name, const nw::ut::MoveArray<AnimElem*> &anims )
{
    nw::os::IAllocator *pAllocator = Memory_GetAllocator();
    AnimSet *animSet = NULL;
    {
        // Overwrite if one with the same name already exists
        nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
        nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
        for (; it!=end_it; ++it)
        {
            if (!(*it)->IsStray() && !strcmp((*it)->GetName(), name))
            {
                animSet = *it;
                if (animSet == m_pCurrentAnimSet)
                {
                    DetachAnimSet( animSet );
                    m_pCurrentAnimSet = NULL;
                }
                animSet->Clear();
                break;
            }
        }
        if (!animSet)
        {
            animSet = AnimSet::Builder().Name(name).IsStray(false).Create(pAllocator);
            m_AnimSets.push_back( animSet );
        }
    }

    nw::ut::MoveArray<AnimElem*>::const_iterator it = anims.begin();
    nw::ut::MoveArray<AnimElem*>::const_iterator end_it = anims.end();
    for (; it!=end_it; ++it)
    {
        animSet->AddAnim( *it );
    }

    if (!m_pCurrentAnimSet)
    {
        AttachAnimSet( animSet );

        m_pCurrentAnimSet = animSet;

        FindLongestAnimation();
    }

    SortAnimSets();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::Group( const nw::ut::MoveArray<AnimElem*> &anims )
{
    nw::os::IAllocator *pAllocator = Memory_GetAllocator();
    nw::ut::MoveArray<AnimElem*>::const_iterator it = anims.begin();
    nw::ut::MoveArray<AnimElem*>::const_iterator end_it = anims.end();
    for (; it!=end_it; ++it)
    {
        AnimSet *animSet = AnimSet::Builder().Name((*it)->GetResAnim().GetName()).Create(pAllocator);
        animSet->AddAnim( *it );
        m_AnimSets.push_back( animSet );
    }

    SortAnimSets();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::Group( AnimElem* anim )
{
    nw::os::IAllocator *pAllocator = Memory_GetAllocator();
    AnimSet *animSet = AnimSet::Builder().Name(anim->GetResObject().GetName()).Create(pAllocator);

    animSet->AddAnim( anim );

    m_AnimSets.push_back( animSet );

    SortAnimSets();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::ChangeToPrevAnimation()
{
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
    nw::ut::MoveArray<AnimSet*>::iterator prev;
    for (; it!=end_it; ++it)
    {
        if ( (*it) == m_pCurrentAnimSet )
        {
            prev = it;
            if (prev==m_AnimSets.begin()) prev = end_it;
            --prev;
            break;
        }
    }

    ChangeTo(*prev);
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::ChangeToNextAnimation()
{
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
    nw::ut::MoveArray<AnimSet*>::iterator next;
    for (; it!=end_it; ++it)
    {
        if ( (*it) == m_pCurrentAnimSet )
        {
            next = it+1;
            if (next==end_it) next = m_AnimSets.begin();
            break;
        }
    }

    ChangeTo(*next);
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
s32
AnimSetLinker::GetNumCurrentAnim(AnimElem::AnimElemType type)
{
    return m_pCurrentAnimSet->GetAnimNum( type );
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::FindLongestAnimation()
{
    f32 endFrame = 0;
    BaseEvaluatorArray::iterator it = m_CurrentEvaluators.begin();
    BaseEvaluatorArray::iterator end_it = m_CurrentEvaluators.end();
    for (; it!=end_it; ++it)
    {
        if ((*it)->GetEndFrame()>endFrame)
        {
            endFrame = (*it)->GetEndFrame();
            m_pLongestAnimEvaluator = *it;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
const char*
AnimSetLinker::GetCurrentAnimSetName()
{
    return m_pCurrentAnimSet->GetName();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
s32
AnimSetLinker::GetCurrentAnimSetIndex()
{
    for ( u32 idx = 0; idx < m_AnimSets.size(); ++idx)
    {
        if ( m_AnimSets[ idx ] == m_pCurrentAnimSet )
        {
            return idx;
        }
    }

    return -1;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::Destroy()
{
    this->~AnimSetLinker();
    m_pAllocator->Free(this);
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker::AnimSetLinker(
    nw::gfx::SceneNode *node,
    IAnimController *pController,
    nw::os::IAllocator *pAllocator )
{
    m_pNode = node;
    m_pController = pController;
    m_pAllocator = pAllocator;
    m_AnimSets = nw::ut::MoveArray<AnimSet*>(pAllocator);
    m_CurrentEvaluators = nw::ut::MoveArray<nw::gfx::BaseAnimEvaluator*>(pAllocator);
    m_pLongestAnimEvaluator = NULL;

    // Add empty animation set (for unbinding)
    m_pCurrentAnimSet = AnimSet::Builder().Name("N/A").Create(pAllocator);
    m_pVoidAnimSet = m_pCurrentAnimSet;
    m_AnimSets.push_back( m_pCurrentAnimSet );
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
AnimSetLinker::~AnimSetLinker()
{
    if (m_pCurrentAnimSet)
    {
        DetachAnimSet( m_pCurrentAnimSet );
    }

    nw::os::IAllocator *pAllocator = Memory_GetAllocator();
    nw::ut::MoveArray<AnimSet*>::const_iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::const_iterator end_it = m_AnimSets.end();
    for (; it!=end_it; ++it)
    {
        (*it)->Destroy();
    }
}

inline bool
CompareByAlphabet(const AnimSet* rLeft, const AnimSet* rRight)
{
    if (rLeft->IsStray() &&  !rRight->IsStray())
        return false;
    return strcmp(rLeft->GetName(),rRight->GetName()) < 0;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::SortAnimSets()
{
    std::sort( m_AnimSets.begin(), m_AnimSets.end(), CompareByAlphabet );
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::ChangeTo(const char *animSetName)
{
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
    for (; it!=end_it; ++it)
    {
        if (!strcmp(animSetName, (*it)->GetName()))
        {
            ChangeTo( *it );
            break;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::ChangeTo(s32 index)
{
    if (index >= 0 && index < m_AnimSets.size())
    {
        ChangeTo( m_AnimSets[ index ] );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::ChangeTo(AnimSet *animSet)
{
    f32 step = m_pLongestAnimEvaluator ? m_pLongestAnimEvaluator->GetStepFrame() : 1.f;

    // Save the current frame count only when in a stop state
    m_pCurrentAnimSet->SetCurrentFrame( m_pController->GetFrame() );

    DetachAnimSet( m_pCurrentAnimSet );

    AttachAnimSet( animSet );

    BaseEvaluatorArray::iterator it = GetEvaluatorsBegin();
    BaseEvaluatorArray::iterator end_it = GetEvaluatorsEnd();
    for (; it!=end_it; ++it)
    {
        if (m_pController->IsActive())
            (*it)->SetStepFrame( m_pController->GetStepFrame() );
        else
            (*it)->SetStepFrame( 0.f );
    }

    m_pCurrentAnimSet = animSet;

    // Restore the current frame count only when in a stop state
    if (!m_pController->IsActive()) {
        m_pController->SetFrame( m_pCurrentAnimSet->GetCurrentFrame() );
    }

    FindLongestAnimation();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::RenameAnimSet(const char* oldName, const char* newName)
{
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
    for (; it!=end_it; ++it)
    {
        if (!strcmp(oldName, (*it)->GetName()))
        {
            (*it)->SetName(newName);
            break;
        }
    }

    SortAnimSets();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::ChangeToFirstStrayAnimation()
{
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
    for (; it!=end_it; ++it)
    {
        if ((*it)->IsStray())
        {
            ChangeTo( *it );
            return;
        }
    }

    UnbindAnimation();
}


/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::UnbindAnimation()
{
    ChangeTo( m_pVoidAnimSet );
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::DetachAnimSet(AnimSet *anim)
{
    m_CurrentEvaluators.clear();

    for (int i=0; i<anim->GetAnimNum( AnimElem::Anim_Skeletal ); ++i)
    {
        UnbindSkeletalAnim( m_pNode, i );
    }
    for (int i=0; i<anim->GetAnimNum( AnimElem::Anim_Material ); ++i)
    {
        UnbindMaterialAnim( m_pNode, i );
    }
    for (int i=0; i<anim->GetAnimNum( AnimElem::Anim_Visibility ); ++i)
    {
        UnbindVisibilityAnim( m_pNode, i );
    }
    for (int i=0; i<anim->GetAnimNum( AnimElem::Anim_Camera ); ++i)
    {
        UnbindCameraAnim( m_pNode );
    }
    for (int i=0; i<anim->GetAnimNum( AnimElem::Anim_Light ); ++i)
    {
        UnbindLightAnim( m_pNode );
    }
    for (int i=0; i<anim->GetAnimNum( AnimElem::Anim_Fog ); ++i)
    {
        UnbindFogAnim( m_pNode );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::AttachAnimSet(AnimSet *anim)
{
    u32 numSkeletalAnimSet = 0;
    u32 numMaterialAnimSet = 0;
    u32 numVisibilityAnimSet = 0;
    u32 numCameraAnimSet = 0;
    u32 numLightAnimSet = 0;
    nw::gfx::BaseAnimEvaluator *eval;
    nw::ut::MoveArray<AnimElem*>::iterator it = anim->GetAnimElemBegin();
    nw::ut::MoveArray<AnimElem*>::iterator end_it = anim->GetAnimElemEnd();
    for (; it!=end_it; ++it)
    {
        switch ((*it)->GetAnimType())
        {
        case AnimElem::Anim_Skeletal:
            eval = InitializeSkeletalAnim( m_pNode, (*it)->GetResAnim(), numSkeletalAnimSet++ );
            if (eval) m_CurrentEvaluators.push_back(eval);
            break;
        case AnimElem::Anim_Material:
            eval = InitializeMaterialAnim( m_pNode, (*it)->GetResAnim(), numMaterialAnimSet++ );
            if (eval) m_CurrentEvaluators.push_back(eval);
            break;
        case AnimElem::Anim_Visibility:
            eval = InitializeVisibilityAnim( m_pNode, (*it)->GetResAnim(), numVisibilityAnimSet++ );
            if (eval) m_CurrentEvaluators.push_back(eval);
            break;
        case AnimElem::Anim_Camera:
            eval = InitializeCameraAnim( m_pNode, (*it)->GetResAnim() );
            if (eval) m_CurrentEvaluators.push_back(eval);
            break;
        case AnimElem::Anim_Light:
            eval = InitializeLightAnim( m_pNode, (*it)->GetResAnim() );
            if (eval) m_CurrentEvaluators.push_back(eval);
            break;
        case AnimElem::Anim_Fog:
            eval = InitializeFogAnim( m_pNode, (*it)->GetResAnim() );
            if (eval) m_CurrentEvaluators.push_back(eval);
            break;
        default:
            break;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::Clear()
{
    AnimSet *unbindAnimSet = NULL;
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    while (it != m_AnimSets.end())
    {
        if ( !(*it)->IsStray() && (*it) != m_pVoidAnimSet)
        {
            nw::ut::MoveArray<AnimSet*>::iterator next = it;
            ++next;
            if ( m_pCurrentAnimSet == *it )
            {
                DetachAnimSet( m_pCurrentAnimSet );
                m_pCurrentAnimSet = NULL;
            }
            (*it)->Destroy();
            it = m_AnimSets.erase( it, next );
        }
        else
        {
            ++it;
        }
    }

    if (!m_pCurrentAnimSet) m_pCurrentAnimSet = m_pVoidAnimSet;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:  Patrols all animSets held by the Linker and deletes the link with the specified AnimElem
                
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
AnimSetLinker::RemoveAnimElem(AnimElem *anim)
{
    nw::ut::MoveArray<AnimSet*>::iterator it = m_AnimSets.begin();
    nw::ut::MoveArray<AnimSet*>::iterator end_it = m_AnimSets.end();
    for (; it!=end_it; ++it)
    {
        if( *it == m_pCurrentAnimSet )
        {
            ChangeTo( m_pVoidAnimSet );
            (*it)->RemoveAnimElem( anim );
            ChangeTo( *it );
        }else
        {
            (*it)->RemoveAnimElem( anim );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
f32
AnimSetLinker::GetFrame()
{
    if (m_pCurrentAnimSet == m_pVoidAnimSet) return 0.f;

    if (m_pLongestAnimEvaluator)
        return m_pLongestAnimEvaluator->GetFrame();

    return 0;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
f32
AnimSetLinker::GetEndFrame()
{
    if (m_pCurrentAnimSet == m_pVoidAnimSet) return 0.f;

    if (m_pLongestAnimEvaluator)
        return m_pLongestAnimEvaluator->GetEndFrame();
    return 0.f;
}

}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif
