/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <nw/gfx.h>
#include "../include/SceneManager.h"
#include "../include/SceneElemSelector.h"
#include "../../system/include/HidManager.h"
#include "../include/ModuleG3dImpl.h"
#include "../include/AnimManager.h"
#include "../include/InfoDrawer.h"
#include "../include/SceneManager.h"
#include "../../effect/include/AnimManager.h"

namespace vw  {
namespace g3d {

#define FOREACH_SCENEELEM( OPERATIONS )\
{\
    SceneManager* pScnMgr = SceneManager::GetInstance();\
    {\
        SceneManager::ModelList::iterator it = pScnMgr->GetModelsBegin();\
        SceneManager::ModelList::iterator end_it = pScnMgr->GetModelsEnd();\
        for (; it!=end_it; ++it)\
        {\
            IAnimController *controller = it->GetAnimController();\
            OPERATIONS;\
        }\
    }\
    {\
        SceneManager::LightList::iterator it = pScnMgr->GetLightsBegin();\
        SceneManager::LightList::iterator end_it = pScnMgr->GetLightsEnd();\
        for (; it!=end_it; ++it)\
        {\
            IAnimController *controller = it->GetAnimController();\
            OPERATIONS;\
        }\
    }\
    {\
        SceneManager::CameraList::iterator it = pScnMgr->GetCamerasBegin();\
        SceneManager::CameraList::iterator end_it = pScnMgr->GetCamerasEnd();\
        for (; it!=end_it; ++it)\
        {\
            IAnimController *controller = it->GetAnimController();\
            OPERATIONS;\
        }\
    }\
    {\
        SceneManager::FogList::iterator it = pScnMgr->GetFogsBegin();\
        SceneManager::FogList::iterator end_it = pScnMgr->GetFogsEnd();\
        for (; it!=end_it; ++it)\
        {\
            IAnimController *controller = it->GetAnimController();\
            OPERATIONS;\
        }\
    }\
}

AnimManager* AnimManager::m_pInstance = NULL;

AnimManager*
AnimManager::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

AnimManager::AnimManager()
    : m_DefaultRate( 0.0f )
    , m_LoopFlag( false )
{
    for ( int i = 0; i < ANIM_RECORD_NUM; ++i ) {
        m_AnimFrameRecord[ i ].fileType = (ResourceSet::BinaryFileType)( ResourceSet::Type_Packed + 1 );
    }
}

AnimManager::~AnimManager()
{
}

void
AnimManager::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(AnimManager));
    if (buf == NULL) return;

    m_pInstance = new(buf) AnimManager();
}

void
AnimManager::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~AnimManager();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

void
AnimManager::UpdateAll()
{
    FOREACH_SCENEELEM( controller->Update() );
}

void
AnimManager::Reset()
{
    FOREACH_SCENEELEM( controller->Reset() );
}

void
AnimManager::StopAll()
{
    FOREACH_SCENEELEM( controller->Stop() );
}

void
AnimManager::PauseAll()
{
    FOREACH_SCENEELEM( controller->Pause() );
}

void
AnimManager::PlayAll()
{
    FOREACH_SCENEELEM( controller->Play() );
}

void
AnimManager::StepForwardAll()
{
    FOREACH_SCENEELEM( controller->StepForward() );
}

void
AnimManager::StepBackwardAll()
{
    FOREACH_SCENEELEM( controller->StepBackward() );
}

void
AnimManager::SpeedUpAll(f32 diff)
{
    FOREACH_SCENEELEM( controller->SpeedUp(diff) );
}

void
AnimManager::SetFrame(f32 frame)
{
    FOREACH_SCENEELEM( controller->SetFrame( frame ) );
}

void
AnimManager::SetStepFrameAll(f32 step)
{
    m_DefaultRate = step;
    FOREACH_SCENEELEM( controller->SetStepFrame( step ) );
}

bool
AnimManager::ActiveAnimationExists()
{
    FOREACH_SCENEELEM( if (controller->IsActive()) return true; )
    return false;
}

bool
AnimManager::StoreCurrentAnimFrame( ResourceSet::BinaryFileType fileType, unsigned int hashID, unsigned int index )
{
    int emptyIndex = -1;
    for ( int i = 0; i < ANIM_RECORD_NUM; ++i ) {
        if ( m_AnimFrameRecord[ i ].fileType ==  ResourceSet::Type_Packed + 1 ) {
            emptyIndex = i;
            break;
        }
    }
    if ( emptyIndex == -1 || emptyIndex == ANIM_RECORD_NUM ) return false;

    int animSetIndex = -1;
    float currentFrame = 0.0f;
    bool  isActive = true;
    switch( fileType )
    {
    case ResourceSet::Type_Model:
        {
            SceneManager::ModelList::iterator it = SceneManager::GetInstance()->FindModel( hashID, index );
            if ( it == SceneManager::GetInstance()->GetModelsEnd() ) return false;
            animSetIndex = it->GetAnimLinker()->GetCurrentAnimSetIndex();
            currentFrame = it->GetAnimController()->GetFrame();
            isActive     = it->GetAnimController()->IsActive();
        }
        break;
    case ResourceSet::Type_Camera:
        {
            SceneManager::CameraList::iterator it = SceneManager::GetInstance()->FindCamera( hashID, index );
            if ( it == SceneManager::GetInstance()->GetCamerasEnd() ) return false;
            animSetIndex = it->GetAnimLinker()->GetCurrentAnimSetIndex();
            currentFrame = it->GetAnimController()->GetFrame();
            isActive     = it->GetAnimController()->IsActive();
        }
        break;
    case ResourceSet::Type_Light:
        {
            SceneManager::LightList::iterator it = SceneManager::GetInstance()->FindLight( hashID, index );
            if ( it == SceneManager::GetInstance()->GetLightsEnd() ) return false;
            animSetIndex = it->GetAnimLinker()->GetCurrentAnimSetIndex();
            currentFrame = it->GetAnimController()->GetFrame();
            isActive     = it->GetAnimController()->IsActive();
        }
        break;
    case ResourceSet::Type_Fog:
        {
            SceneManager::FogList::iterator it = SceneManager::GetInstance()->FindFog( hashID, index );
            if ( it == SceneManager::GetInstance()->GetFogsEnd() ) return false;
            animSetIndex = it->GetAnimLinker()->GetCurrentAnimSetIndex();
            currentFrame = it->GetAnimController()->GetFrame();
            isActive     = it->GetAnimController()->IsActive();
        }
        break;
    default:
        return false;
    }

    m_AnimFrameRecord[ emptyIndex ].fileType = fileType;
    m_AnimFrameRecord[ emptyIndex ].hashID = hashID;
    m_AnimFrameRecord[ emptyIndex ].index = index;
    m_AnimFrameRecord[ emptyIndex ].frame = currentFrame;
    m_AnimFrameRecord[ emptyIndex ].animSetIndex = animSetIndex;
    m_AnimFrameRecord[ emptyIndex ].isActive = isActive;

    return true;
}

bool
AnimManager::RestoreCurrentAnimFrame( ResourceSet::BinaryFileType fileType, unsigned int hashID, unsigned int index )
{
    int findIndex = -1;
    for ( int i = 0; i < ANIM_RECORD_NUM; ++i ) {
        if (
            m_AnimFrameRecord[ i ].fileType == fileType &&
            m_AnimFrameRecord[ i ].hashID == hashID &&
            m_AnimFrameRecord[ i ].index == index
        ) {
            findIndex = i;
            break;
        }
    }
    if ( findIndex == -1 || findIndex == ANIM_RECORD_NUM ) return false;

    switch( fileType )
    {
    case ResourceSet::Type_Model:
        {
            SceneManager::ModelList::iterator it = SceneManager::GetInstance()->FindModel( hashID, index );
            if ( it == SceneManager::GetInstance()->GetModelsEnd() ) return false;
            it->GetAnimLinker()->ChangeTo( m_AnimFrameRecord[ findIndex ].animSetIndex );
            it->GetAnimController()->SetFrame( m_AnimFrameRecord[ findIndex ].frame );
            ( m_AnimFrameRecord[ findIndex ].isActive ) ? it->GetAnimController()->Play() : it->GetAnimController()->Pause();
        }
        break;
    case ResourceSet::Type_Camera:
        {
            SceneManager::CameraList::iterator it = SceneManager::GetInstance()->FindCamera( hashID, index );
            if ( it == SceneManager::GetInstance()->GetCamerasEnd() ) return false;
            it->GetAnimLinker()->ChangeTo( m_AnimFrameRecord[ findIndex ].animSetIndex );
            it->GetAnimController()->SetFrame( m_AnimFrameRecord[ findIndex ].frame );
            ( m_AnimFrameRecord[ findIndex ].isActive ) ? it->GetAnimController()->Play() : it->GetAnimController()->Pause();
        }
        break;
    case ResourceSet::Type_Light:
        {
            SceneManager::LightList::iterator it = SceneManager::GetInstance()->FindLight( hashID, index );
            if ( it == SceneManager::GetInstance()->GetLightsEnd() ) return false;
            it->GetAnimLinker()->ChangeTo( m_AnimFrameRecord[ findIndex ].animSetIndex );
            it->GetAnimController()->SetFrame( m_AnimFrameRecord[ findIndex ].frame );
            ( m_AnimFrameRecord[ findIndex ].isActive ) ? it->GetAnimController()->Play() : it->GetAnimController()->Pause();
        }
        break;
    case ResourceSet::Type_Fog:
        {
            SceneManager::FogList::iterator it = SceneManager::GetInstance()->FindFog( hashID, index );
            if ( it == SceneManager::GetInstance()->GetFogsEnd() ) return false;
            it->GetAnimLinker()->ChangeTo( m_AnimFrameRecord[ findIndex ].animSetIndex );
            it->GetAnimController()->SetFrame( m_AnimFrameRecord[ findIndex ].frame );
            ( m_AnimFrameRecord[ findIndex ].isActive ) ? it->GetAnimController()->Play() : it->GetAnimController()->Pause();
        }
        break;
    default:
        return false;
    }

    m_AnimFrameRecord[ findIndex ].fileType = (ResourceSet::BinaryFileType)( ResourceSet::Type_Packed + 1 );

    return true;
}

void
AnimManager::HidControl_(s32 padEvent, IAnimController *controller)
{
    static s32 status = system::G3D_NONE;

    // Get current status
    if (padEvent != system::G3D_NONE) status = padEvent;

    // Process during scene update frame
    // For the purpose of frame rate synchronization when advancing the frame when the frame rate is not 60 Hz
    if (vw::system::App::GetInstance().IsUpdateScene())
    {
        switch ( status )
        {
        case system::G3D_SLOW_DOWN_ANIMATION:
            // Slow the animation playback.
            if (controller) controller->SpeedUp( -0.1f );
            break;

        case system::G3D_SLOW_DOWN_ALL_ANIMATIONS:
            // Slow all animation playback.
            SetStepFrameAll( controller ? controller->GetStepFrame() - 0.1f : m_DefaultRate - 0.1f );
            vw::effect::AnimManager::GetInstance()->SetStepFrame(m_DefaultRate);
            break;

        case system::G3D_SPEED_UP_ANIMATION:
            // Speed up the animation playback.
            if (controller) controller->SpeedUp( 0.1f );
            break;

        case system::G3D_SPEED_UP_ALL_ANIMATIONS:
            // Speed up all animation playback.
            SetStepFrameAll( controller ? controller->GetStepFrame() + 0.1f : m_DefaultRate + 0.1f );
            vw::effect::AnimManager::GetInstance()->SetStepFrame(m_DefaultRate);
            break;

        case system::G3D_INCREMENT_ANIMATION_FRAME:
            // Stop the animation, and advance one frame.
            if (controller) controller->StepForward();
            break;

        case system::G3D_DECREMENT_ANIMATION_FRAME:
            // Stop the animation, and rewind one frame.
            if (controller) controller->StepBackward();
            break;

        case system::G3D_INCREMENT_ALL_ANIMATION_FRAMES:
            // Stop all animation, and advance one frame.
            StepForwardAll();
            vw::effect::AnimManager::GetInstance()->StepForwardAll();
            break;

        case system::G3D_DECREMENT_ALL_ANIMATION_FRAMES:
            // Stop all animation, and advance one frame.
            StepBackwardAll();
            vw::effect::AnimManager::GetInstance()->StepBackwardAll();
            break;

        case system::G3D_STOP_ALL_ANIMATIONS:
            // Stop all model and effect animations.
            StopAll();
            vw::effect::AnimManager::GetInstance()->StopAll();
            break;

        case system::G3D_START_ALL_ANIMATIONS:
            // Play all model and effect animations.
            if (ActiveAnimationExists() && vw::effect::AnimManager::GetInstance()->ActiveAnimationExists())
            {
                PauseAll();
                vw::effect::AnimManager::GetInstance()->PauseAll();
            }
            else
            {
                PlayAll();
                vw::effect::AnimManager::GetInstance()->PlayAll();
                //SysAnmCtrl_PlayAll();
            }
            break;

        case system::G3D_STOP_ANIMATION:
            // Stop the animation.
            if (controller) controller->Stop();
            break;

        case system::G3D_START_ANIMATION:
            // Plays the animation.
            if (controller) controller->PauseAndPlay();
            break;
        case system::G3D_TOGGLE_ONETIME_LOOP:
            m_LoopFlag = !m_LoopFlag;
        default:
            break;
        }

        // Reset status
        status = system::G3D_NONE;
    }
}

void
AnimManager::HidControl(s32 padEvent)
{
    SceneElemSelector *selector = SceneElemSelector::GetInstance();

    switch (InfoDrawer::GetInstance()->GetInfoResourceType())
    {
    case InfoDrawer::INFO_MODEL:
        if (selector->GetTargetModel())
        {
            HidControl_( padEvent, selector->GetTargetModel()->GetAnimController());
        }
        break;
    case InfoDrawer::INFO_LIGHT:
        if (selector->GetTargetLight())
        {
            HidControl_( padEvent, selector->GetTargetLight()->GetAnimController() );
        }
        break;
    case InfoDrawer::INFO_CAMERA:
        if (selector->GetTargetCamera())
        {
            HidControl_( padEvent, selector->GetTargetCamera()->GetAnimController() );
        }
        break;
    case InfoDrawer::INFO_FOG:
        if (selector->GetTargetFog())
        {
            HidControl_( padEvent, selector->GetTargetFog()->GetAnimController() );
        }
        break;
    default:
        break;
    }
}

void
AnimManager::OnTargetModuleChanged(u16 id)
{
}

void
AnimManager::OnAnmCmd(u32 command, void* arg)
{
    switch (command)
    {
    case system::ANM_CMD_RESET:
        break;
    case system::ANM_CMD_PLAY:
        PlayAll();
        break;
    case system::ANM_CMD_REV_PLAY:
        break;
    case system::ANM_CMD_STOP:
        StopAll();
        break;
    case system::ANM_CMD_STEP:
        StepForwardAll();
        break;
    case system::ANM_CMD_REV_STEP:
        StepBackwardAll();
        break;
    case system::ANM_CMD_LOOP:
        break;
    case system::ANM_CMD_SET_FRAME:
        break;
    }
}

}   /* namespace g3d */
}   /* namespace vw  */

