/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include <viewer/g3d/CameraMgr.h>
#include <viewer/g3d/LightMgr.h>
#include <viewer/system/Perform.h>
#include "../include/InfoDrawer.h"
#include "../include/AnimManager.h"
#include "../../system/include/PerformDrawer.h"
#include "../../system/include/HidManager.h"
#include "../../system/include/LogSystem.h"
#include "../../effect/include/AnimManager.h"

#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#endif
namespace vw  {
namespace g3d {

InfoDrawer* InfoDrawer::mInstance = NULL;

/*===========================================================================*
  Name:         InfoDrawer
  Description:  G3D information display management
 *===========================================================================
*/
InfoDrawer* InfoDrawer::GetInstance()
{
    NW_ASSERT( mInstance != NULL );
    return mInstance;
}

InfoDrawer::InfoDrawer()
{
    mSubModeName = "N/A";
    mSelectType = SELECT_VIEW_CAMERA;
    mViewCameraName = NULL;
    mControlTargetType = CONTROL_CAMERA;
    mControlTargetName = NULL;
    mResType = INFO_MODEL;
    mResName = NULL;
    mAnimationName = NULL;
    mAnimationType[0]=mAnimationType[1]=mAnimationType[2]=false;
    mSceneName = NULL;
    mFrameCount = 50.0f;
    mMaxFrame  = 0.0f;
    mFrameRate = 1.0f;
    mRenderInfoFlag = 0;    // Displays all defaults.
    memset(mRenderFlag, true, sizeof(bool) * SceneRenderer::RenderNum);
    mSelectColor = nw::ut::Color8(255, 110, 110, 255);
    mFontCacheSetup = true;
    mTouch = false;

    if( vw::system::App::GetUseMenuFlg() )
    {
        vw::system::FontDrawer::GetInstance().CreateCache(mFontBuffer,256);
    }

    // Does not display bounding box.
    mRenderFlag[SceneRenderer::RenderBoundingBox] = false;
    // Does not display stereoscopic base plane.
    mRenderFlag[SceneRenderer::RenderDepthLevelSurface] = false;
}

InfoDrawer::~InfoDrawer()
{
}

// Initialize
void
InfoDrawer::Initialize()
{
    if (mInstance != NULL) return;

    void *buf = Memory_Alloc(sizeof(InfoDrawer));
    if (buf == NULL) return;

    mInstance = new(buf) InfoDrawer();
}

// InitializeMenu
void
InfoDrawer::InitializeMenu()
{
    vw::system::FontDrawer::GetInstance().CreateCache(mFontBuffer,256);
}

// Finalize
void
InfoDrawer::Finalize()
{
    if (mInstance == NULL) return;

    mInstance->~InfoDrawer();
    Memory_Free(mInstance);

    mInstance = NULL;
}

/*===========================================================================*
  Name:         SeekBar
  Description:  Seek bar
 *===========================================================================
*/
class SeekBar
{
    struct Rect
    {
        f32 x, y, w, h;
    };
public:
    SeekBar() : m_Grab(false), m_Cursor(0.0f)
    {
        m_Rect.x = 6;
        m_Rect.y = 217;
        m_Rect.w = 306;
        m_Rect.h = 40;
    }
    void Main(bool touch, u16 x, u16 y)
    {
        x = ((x-11) / (296.0f)) * 320;

        m_Grab = false;
        if( touch == 1 )
        {
            if( isInRect(x, y) )
            {
                m_Grab = true;
                m_Cursor = f32(x - 4 - m_Rect.x) / 300.0f;
                if( m_Cursor < 0.0f ) m_Cursor = 0.0f;
                if( m_Cursor > 1.0f ) m_Cursor = 1.0f;
            }
        }
    }
    void SetCursor(f32 cursor)
    {
        m_Cursor = cursor;
    }
    bool IsGrab() const
    {
        return m_Grab;
    }
    f32 GetSeekBar() const
    {
        return m_Cursor;
    }
    void ResetGrab()
    {
        m_Grab = false;
    }

private:
    bool isInRect(f32 /*x*/, f32 y)
    {
        return m_Rect.y < y && y < m_Rect.y+m_Rect.h;
    }

private:
    bool    m_Grab;
    f32     m_Cursor;
    Rect    m_Rect;    // Rectangle for touch determination (not used for rendering)
};
static SeekBar sSeekBar;



// HID input
void
InfoDrawer::Main(vw::system::HidStatus &hidStatus)
{
    // Turns off touch operation during lower screen preview.
    {
        vw::system::SystemSetting setting;
        vw::system::System_GetSetting( &setting );
        if ( setting.lowerPreview ) { return; }
    }
    if( vw::system::App::GetInstance().GetFocusID() == vw::system::G3D_ID )
    {
        sSeekBar.Main(hidStatus.touchPanel.touch,
                      hidStatus.touchPanel.x,
                      hidStatus.touchPanel.y);
    }
    else
    {
        sSeekBar.ResetGrab();
    }
}

void
InfoDrawer::SetFrameCount(f32 frame)
{
    mFrameCount = frame;
}

void
InfoDrawer::SetMaxFrame(f32 frame)
{
    if(mMaxFrame == frame) return;
    mMaxFrame = frame;
    mFontCacheSetup |= true;
}

void
InfoDrawer::SetFrameRate(f32 frameRate)
{
    if(mFrameRate == frameRate) return;
    if(nn::math::FAbs(frameRate) < 0.001f)
        frameRate = 0.0f; // -0.0f revision
    mFrameRate = frameRate;
    mFontCacheSetup |= true;
}

// Sets sub-mode name.
void
InfoDrawer::SetSubModeName(const char* name)
{
    if(mSubModeName == name) return;

    if(mSubModeName && name)
    {
        if(0 != strcmp(mSubModeName,name))
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    else
    {
        if(mSubModeName || name)
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    mSubModeName = name;
}

// Sets the camera name.
void
InfoDrawer::SetViewCameraName(const char* name)
{
    if(mViewCameraName == name) return;

    if(mViewCameraName && name)
    {
        if(0 != strcmp(mViewCameraName,name))
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    else
    {
        if(mViewCameraName || name)
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    mViewCameraName = name;
}

void
InfoDrawer::SetControlTargetType(const ControlTargetType type)
{
    if(mControlTargetType == type) return;
    mControlTargetType = type;
    mFontCacheSetup |= true;
}

void
InfoDrawer::SetControlTargetName(const char* name)
{
    if(mControlTargetName == name) return;

    if(mControlTargetName && name)
    {
        if(0 != strcmp(mControlTargetName,name))
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    else
    {
        if(mControlTargetName || name)
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    mControlTargetName = name;
}

void
InfoDrawer::SetResourceType(const InfoResType type)
{
    if(mResType == type) return;
    mResType = type;
    mFontCacheSetup |= true;
}

void
InfoDrawer::SetResouceName(const char* name)
{
    if(mResName == name) return;

    if(mResName && name)
    {
        if(0 != strcmp(mResName,name))
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    else
    {
        if(mResName || name)
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    mResName = name;
}

void
InfoDrawer::SetAnimationGroup(const char* name, bool skeletal, bool material, bool visibility)
{
    if(mAnimationName == name &&
       mAnimationType[0] == skeletal &&
       mAnimationType[1] == material &&
       mAnimationType[2] == visibility )
    {
        return;
    }

    if(mAnimationName && name)
    {
        if(0 != strcmp(mAnimationName,name))
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    else
    {
        if(mAnimationName || name)
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }

    if( (mAnimationType[0] != skeletal) ||
        (mAnimationType[1] != material) ||
        (mAnimationType[2] != visibility) )
    {
        // Rebuilds the cache.
        mFontCacheSetup |= true;
    }
    mAnimationName = name;
    mAnimationType[0] = skeletal;
    mAnimationType[1] = material;
    mAnimationType[2] = visibility;
}

// Sets the scene.
void
InfoDrawer::SetSceneName(const char* name)
{
    if(mSceneName == name) return;

    if(mSceneName && name)
    {
        if(0 != strcmp(mSceneName,name))
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    else
    {
        if(mSceneName || name)
        {
            // Rebuilds the cache.
            mFontCacheSetup |= true;
        }
    }
    mSceneName = name;
}

// Display the measurement result
void InfoDrawer::DrawInfo()
{
    // Rebuilds the cache.
    if (mFontCacheSetup) {
        SetupFontBuffer();
        mFontCacheSetup = false;
    }

#if !defined( NN_PCGX )
    // Displays performance information.
    vw::system::PerformDrawer::GetInstance()->DrawMeter();

    // Displays performance bar.
    vw::system::PerformDrawer::GetInstance()->DrawGraph();
#endif

    // Display seek bar.
    DrawSeekBar(mFrameCount, mMaxFrame);

    // Renders frame bar.
    vw::system::FontDrawer& font = vw::system::FontDrawer::GetInstance();
    font.BeginDraw();
    if ( mSelectType == SELECT_ANIMATION_FRAME ) {
        font.SetColor( mSelectColor );
        font.DrawString(125,200,"%6.1f", mFrameCount);
        font.SetColor( nw::ut::Color8(~0) );
    } else {
        font.DrawString(125,200,"%6.1f", mFrameCount);
    }
    font.EndDraw();

    // Displays touch panel determination range.
    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    draw2D.BeginDraw();
    draw2D.FillRectangle( 5,119,310,1,nw::ut::Color8(95,95,95,255)); // Determines view camera selection.
    draw2D.FillRectangle( 5,139,310,1,nw::ut::Color8(95,95,95,255)); // Determines target selection.
    draw2D.FillRectangle( 5,159,310,1,nw::ut::Color8(95,95,95,255)); // Determines resource selection.
    draw2D.FillRectangle( 5,179,310,1,nw::ut::Color8(95,95,95,255)); // Determines animation selection.
    draw2D.FillRectangle( 5,199,310,1,nw::ut::Color8(95,95,95,255)); // Determines scene selection.
    draw2D.FillRectangle( 7,219,8.5,1,nw::ut::Color8(95,95,95,255)); // Determines animation controller selection.
    draw2D.FillRectangle(22,219,8.5,1,nw::ut::Color8(95,95,95,255)); // Determines animation controller selection.
    draw2D.FillRectangle(37,219, 34,1,nw::ut::Color8(95,95,95,255)); // Determines automatic looping of the one-time animation.
    draw2D.FillRectangle( 2, 39, 68,1,nw::ut::Color8(95,95,95,255)); // Determines the sub-mode selection.
    draw2D.EndDraw();

    // Displays the animation controller.
    font.BeginDraw();
    if( AnimManager::GetInstance()->ActiveAnimationExists()
        && vw::effect::AnimManager::GetInstance()->ActiveAnimationExists())
    {
        font.DrawString(7,200,"|");
    }
    else
    {
        font.DrawString(7,200,">");
    }
    font.DrawString(22,200,"#");
    font.EndDraw();

    // Displays the one-time animation automatic loop flag.
    bool loopFlag = AnimManager::GetInstance()->GetLoopFlag();
    font.BeginDraw();
    if( loopFlag )
    {
        font.DrawString(37,200,"LOOP");
    }
    else
    {
        font.DrawString(37,200,"ONCE");
    }
    font.EndDraw();

    // Displays the frame skip flag.
    bool frameSkip = vw::system::System_GetSetting()->frameSkip;
    if ( frameSkip )
    {
        font.BeginDraw();
        font.DrawString(75,200,"S");
        font.EndDraw();
    }

    // Renders the font cache.
    font.BeginDraw();
    font.DrawBuffer(mFontBuffer);
    font.EndDraw();

    // FPS display
    font.BeginDraw();
    font.SetColor( ~0 );

    font.DrawString(2, 40,
#if defined( NN_PCGX )
        "Frequency : %3.1f",
#else
        "FPS : %3.1f",
#endif
        vw::system::Video_GetFPS());
    font.EndDraw();
}

void
InfoDrawer::DrawSeekBar(f32 frameCount, f32 maxFrame)
{
    f32 seek = 0.0f;
    if( maxFrame > 0.0f )
    {
        seek = frameCount / maxFrame;
    }
    if( seek > 1.0f)
    {
        seek = 1.0f;
    }

    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    draw2D.BeginDraw();

    const nw::ut::Color8 white = nw::ut::Color8(~0);
    draw2D.SetLineWidth(2);
    draw2D.DrawLine(10,230,310,230,white);
    s32 x = 300 * seek;
    draw2D.FillRectangle(6+x,223,8,14,white);

    draw2D.EndDraw();
}

bool
InfoDrawer::GetSeekBarInfo(float& value)
{
    value = sSeekBar.GetSeekBar();
    return sSeekBar.IsGrab();
}

#define DRAW_TYPE_STRING( flag, pos_x, pos_y, arg1 ) \
    if ( flag ) { \
        font.SetColor( mSelectColor ); \
        font.DrawCache( pos_x, pos_y, arg1 ); \
        font.SetColor( nw::ut::Color8(~0) ); \
    } else { \
        font.DrawCache( pos_x, pos_y, arg1 ); \
    }

#define DRAW_NAME_STRING( flag, arg1, arg2 ) \
    if ( flag ) { \
        font.SetColor( mSelectColor ); \
        font.DrawCache( arg1, arg2 ); \
        font.SetColor( nw::ut::Color8(~0) ); \
    } else { \
        font.DrawCache( arg1, arg2 ); \
    }

void
InfoDrawer::SetupFontBuffer()
{
    vw::system::FontDrawer& font = vw::system::FontDrawer::GetInstance();
    font.BeginCache(mFontBuffer);

    // Animation name
    vw::common::CharArray AnimationName("N/A", vw::system::Memory_GetSysAllocator());
    if( mAnimationName )
    {
        AnimationName = mAnimationName;
        if (mAnimationType[0] || mAnimationType[1] || mAnimationType[2])
        {
            AnimationName += " [";
            if( mAnimationType[0] ) { AnimationName += "S"; }
            if( mAnimationType[1] ) { AnimationName += "M"; }
            if( mAnimationType[2] ) { AnimationName += "V"; }
            AnimationName += "]";
        }
    }

    // Data for the operation target display.
    const char* control_names[CONTROL_NUM] = { "Target Cam: ", "Target Lgt: " };
    const int control_offset[CONTROL_NUM] = { 6, 6 };

    // Resource display data
    const char* res_names[INFO_NUM] = { "Model: ", "Camera: ", "Light: ", "Fog: " };
    const int res_offset[INFO_NUM] = { 48, 40, 48, 65 };

    // Object display
    DRAW_TYPE_STRING( false, 23, 100, "View Cam: " )
    DRAW_NAME_STRING( mSelectType == SELECT_VIEW_CAMERA, "%s", mViewCameraName ? mViewCameraName : "N/A" )

    //DRAW_TYPE_STRING( mSelectType == SELECT_CONTROL_TARGET_TYPE, control_offset[mControlTargetType], 120, control_names[mControlTargetType] )
    DRAW_TYPE_STRING( false, control_offset[mControlTargetType], 120, control_names[mControlTargetType] )
    DRAW_NAME_STRING( mSelectType == SELECT_CONTROL_TARGET, "%s", mControlTargetName ? mControlTargetName : "N/A" )

    //DRAW_TYPE_STRING( mSelectType == SELECT_RESOURCE_TYPE, res_offset[mResType], 140, res_names[mResType] )
    DRAW_TYPE_STRING( false, res_offset[mResType], 140, res_names[mResType] )
    DRAW_NAME_STRING( mSelectType == SELECT_RESOURCE, "%s", mResName ? mResName : "N/A" )

    DRAW_TYPE_STRING( false, 14, 160, "Animation: " )
    DRAW_NAME_STRING( mSelectType == SELECT_ANIMATION, "%s", AnimationName.Get() )

    DRAW_TYPE_STRING( false, 48, 180, "Scene: " )
    DRAW_NAME_STRING( mSelectType == SELECT_SCENE, "%s", mSceneName ? mSceneName : "N/A" )

    // Display frame rate
    DRAW_TYPE_STRING( false, 175, 200, "/" )
    DRAW_NAME_STRING( false, "%6.1f[RATE", mMaxFrame )
    DRAW_NAME_STRING( mSelectType == SELECT_ANIMATION_RATE, "%4.1f", mFrameRate )
    DRAW_NAME_STRING( false, "%s", "]" )

    // Displays the main mode and sub-mode.
    vw::common::CharArray ModeText("Graphics/", vw::system::Memory_GetSysAllocator());
    ModeText += mSubModeName;
    s32 NumLog = vw::system::LogSystem::GetInstance()->GetNumLog();
    if( NumLog > 0 )
    {
        char str[20];
        sprintf(str, "(%d)", NumLog);
        ModeText += str;
    }
    font.DrawCache(2,20, ModeText.Get());

    font.EndCache(mFontBuffer);
}

// Gets the execution command of the selected target.
s32
InfoDrawer::GetSelectionCommand( s32 padEvent ) const
{
    #define KeyDefine( name ) \
        { SELECT_##name, system::G3D_SWITCH_PREV_SELECTION_ITEM_MAIN, system::G3D_SWITCH_PREV_##name }, \
        { SELECT_##name, system::G3D_SWITCH_NEXT_SELECTION_ITEM_MAIN, system::G3D_SWITCH_NEXT_##name },

    #define AnimKeyDefine( name ) \
        { SELECT_##name, system::G3D_SWITCH_PREV_SELECTION_ITEM_SUB, system::G3D_DECREMENT_ALL_ANIMATION_FRAMES }, \
        { SELECT_##name, system::G3D_SWITCH_NEXT_SELECTION_ITEM_SUB, system::G3D_INCREMENT_ALL_ANIMATION_FRAMES },

    struct keyType { s32 select; s32 padEvent; s32 command; };
    const keyType keys[] = {
        KeyDefine( VIEW_CAMERA )
        AnimKeyDefine( VIEW_CAMERA )
        //KeyDefine( CONTROL_TARGET_TYPE )
        //KeyDefine( CONTROL_TARGET )
        { SELECT_CONTROL_TARGET, system::G3D_SWITCH_PREV_SELECTION_ITEM_MAIN, system::G3D_SWITCH_PREV_CONTROL_TARGET_SEQUENTIALLY },
        { SELECT_CONTROL_TARGET, system::G3D_SWITCH_NEXT_SELECTION_ITEM_MAIN, system::G3D_SWITCH_NEXT_CONTROL_TARGET_SEQUENTIALLY },
        AnimKeyDefine( CONTROL_TARGET )
        //KeyDefine( RESOURCE_TYPE )
        //KeyDefine( RESOURCE )
        { SELECT_RESOURCE, system::G3D_SWITCH_PREV_SELECTION_ITEM_MAIN, system::G3D_SWITCH_PREV_RESOURCE_SEQUENTIALLY },
        { SELECT_RESOURCE, system::G3D_SWITCH_NEXT_SELECTION_ITEM_MAIN, system::G3D_SWITCH_NEXT_RESOURCE_SEQUENTIALLY },
        AnimKeyDefine( RESOURCE )
        KeyDefine( ANIMATION )
        AnimKeyDefine( ANIMATION )
        KeyDefine( SCENE )
        AnimKeyDefine( SCENE )
        { SELECT_ANIMATION_FRAME, system::G3D_SWITCH_PREV_SELECTION_ITEM_MAIN, system::G3D_DECREMENT_ALL_ANIMATION_FRAMES },
        { SELECT_ANIMATION_FRAME, system::G3D_SWITCH_NEXT_SELECTION_ITEM_MAIN, system::G3D_INCREMENT_ALL_ANIMATION_FRAMES },
        { SELECT_ANIMATION_RATE, system::G3D_SWITCH_PREV_SELECTION_ITEM_MAIN, system::G3D_SLOW_DOWN_ALL_ANIMATIONS },
        { SELECT_ANIMATION_RATE, system::G3D_SWITCH_NEXT_SELECTION_ITEM_MAIN, system::G3D_SPEED_UP_ALL_ANIMATIONS },
        { SELECT_ANIMATION_FRAME, system::G3D_SWITCH_PREV_SELECTION_ITEM_SUB, system::G3D_DECREMENT_ANIMATION_FRAME },
        { SELECT_ANIMATION_FRAME, system::G3D_SWITCH_NEXT_SELECTION_ITEM_SUB, system::G3D_INCREMENT_ANIMATION_FRAME },
        { SELECT_ANIMATION_RATE, system::G3D_SWITCH_PREV_SELECTION_ITEM_SUB, system::G3D_SLOW_DOWN_ANIMATION },
        { SELECT_ANIMATION_RATE, system::G3D_SWITCH_NEXT_SELECTION_ITEM_SUB, system::G3D_SPEED_UP_ANIMATION },
    };

    const int keyNum = sizeof( keys ) / sizeof( keyType );
    for ( int i = 0; i < keyNum; ++ i )
    {
        if ( keys[ i ].select == mSelectType && keys[ i ].padEvent == padEvent )
        {
            padEvent = keys[ i ].command;
            break;
        }
    }

    return padEvent;
}

// Selected type
InfoDrawer::SelectType
InfoDrawer::GetSelectType() const
{
    return mSelectType;
}

// Return the selected target to what it was.
void
InfoDrawer::ChangeToPrevSelection()
{
    int type = mSelectType;
    --type;
    if ( type == -1 ) {
        mSelectType = (SelectType)( SELECT_NUM - 1 );
    } else {
        mSelectType = (SelectType)type;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Advance the selected target to the next step.
void
InfoDrawer::ChangeToNextSelection()
{
    int type = mSelectType;
    ++type;
    if ( type == SELECT_NUM ) {
        mSelectType = (SelectType)0;
    } else {
        mSelectType = (SelectType)type;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Control target type
InfoDrawer::ControlTargetType
InfoDrawer::GetControlTargetType() const
{
    return mControlTargetType;
}

// Returns the selected target control type to what it was.
void
InfoDrawer::ChangeToPrevControlTargetType()
{
    int type = mControlTargetType;
    --type;
    if ( type == -1 ) {
        mControlTargetType = (ControlTargetType)( CONTROL_NUM - 1 );
    } else {
        mControlTargetType = (ControlTargetType)type;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Advances the selected target to the next step.
void
InfoDrawer::ChangeToNextControlTargetType()
{
    int type = mControlTargetType;
    ++type;
    if ( type == CONTROL_NUM ) {
        mControlTargetType = (ControlTargetType)0;
    } else {
        mControlTargetType = (ControlTargetType)type;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// The control target checks the first one.
bool
InfoDrawer::IsFirstControlTarget()
{
    switch (mControlTargetType)
    {
    case InfoDrawer::CONTROL_CAMERA:
        return CameraMgr::GetInstance()->IsFirstTargetCamera();
    case InfoDrawer::CONTROL_LIGHT:
        return LightMgr::GetInstance()->IsFirstTargetLight();
    default:
        break;
    }

    return false;
}

// The control target checks the end one.
bool
InfoDrawer::IsLastControlTarget()
{
    switch (mControlTargetType)
    {
    case InfoDrawer::CONTROL_CAMERA:
        return CameraMgr::GetInstance()->IsLastTargetCamera();
    case InfoDrawer::CONTROL_LIGHT:
        return LightMgr::GetInstance()->IsLastTargetLight();
    default:
        break;
    }

    return false;
}

// Returns the operation target matrix to what it was before.
void
InfoDrawer::ChangeToFirstControlTarget()
{
    switch (mControlTargetType)
    {
    case CONTROL_CAMERA:
        CameraMgr::GetInstance()->SwitchToFirstTargetCamera();
        break;
    case CONTROL_LIGHT:
        LightMgr::GetInstance()->SwitchToFirstTargetLight();
        break;
    default:
        break;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Advances to the end of the control target.
void
InfoDrawer::ChangeToLastControlTarget()
{
    switch (mControlTargetType)
    {
    case CONTROL_CAMERA:
        CameraMgr::GetInstance()->SwitchToLastTargetCamera();
        break;
    case CONTROL_LIGHT:
        LightMgr::GetInstance()->SwitchToLastTargetLight();
        break;
    default:
        break;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Returns the operation target matrix to what it was before.
void
InfoDrawer::ChangeToPrevControlTarget()
{
    switch (mControlTargetType)
    {
    case InfoDrawer::CONTROL_CAMERA:
        CameraMgr::GetInstance()->SwitchToPrevTargetCamera();
        break;
    case InfoDrawer::CONTROL_LIGHT:
        LightMgr::GetInstance()->SwitchToPrevTargetLight();
        break;
    default:
        break;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Advances to the end of the control target.
void
InfoDrawer::ChangeToNextControlTarget()
{
    switch (mControlTargetType)
    {
    case InfoDrawer::CONTROL_CAMERA:
        CameraMgr::GetInstance()->SwitchToNextTargetCamera();
        break;
    case InfoDrawer::CONTROL_LIGHT:
        LightMgr::GetInstance()->SwitchToNextTargetLight();
        break;
    default:
        break;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Display resource type
InfoDrawer::InfoResType
InfoDrawer::GetInfoResourceType() const
{
    return mResType;
}

// Return the display target to what it was.
void
InfoDrawer::ChangeToPrevResourceType()
{
    int type = mResType;
    --type;
    if ( type == -1 ) {
        mResType = (InfoResType)( INFO_NUM - 1 );
    } else {
        mResType = (InfoResType)type;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Advances the display resource target to the next step.
void
InfoDrawer::ChangeToNextResourceType()
{
    int type = mResType;
    ++type;
    if ( type == INFO_NUM ) {
        mResType = (InfoResType)0;
    } else {
        mResType = (InfoResType)type;
    }
    // Rebuilds the cache.
    mFontCacheSetup |= true;
}

// Whether information is displayed.
bool
InfoDrawer::IsRender2DInfo() const
{
    return (mRenderInfoFlag==0 || mRenderInfoFlag==1);
}

// Whether a grid is displayed.
bool
InfoDrawer::IsRender3DInfo() const
{
    return (mRenderInfoFlag==0);
}


// Display mode settings: [0: display all; 1: display lower screen only; 2: display nothing]
void
InfoDrawer::SetRenderInfoFlag(u32 flag)
{
    if (flag > 2)
    {
        flag = 2;
    }
    mRenderInfoFlag = flag;
}

// Switches display modes.
void
InfoDrawer::ChangeRenderInfoFlag()
{
    // Switches in sequence, as in switching from display all to display lower screen only.
    // 
    mRenderInfoFlag = (mRenderInfoFlag+1) % 2;
}

// Gets the display flags for various 3D information.
bool*
InfoDrawer::GetRenderFlag()
{
    return mRenderFlag;
}

static bool
inRect(const nw::ut::Rect& rect, u16 x, u16 y)
{
    return (rect.left<= x && x <= rect.right &&
            rect.top <= y && y <= rect.bottom);
}

// Switches resources using the touch panel.
void
InfoDrawer::GetTouchPanelEvent(const vw::system::HidStatus& hidStatus, s32& padEvent)
{
    // Turns off touch operation during lower screen preview.
    {
        vw::system::SystemSetting setting;
        vw::system::System_GetSetting( &setting );
        if ( setting.lowerPreview )
        {
            //padEvent = vw::system::G3D_NONE;
            return;
        }
    }

    if ( hidStatus.isClicked )
    {
        nw::ut::Rect viewCamRect, ctrlTargetRect, ctrlResRect, animRect, sceneRect, playRect, stopRect, animLoopRect;
        const u32 offset=5, width=320-offset*2, height=14;

        viewCamRect.SetOriginAndSize(   offset, 104, width, height+2 );    // Sets the height to slightly larger than "height" because there is space above the first line.

        ctrlTargetRect.SetOriginAndSize(offset, 124, width, height );
        ctrlResRect.SetOriginAndSize(   offset, 142, width, height );
        animRect.SetOriginAndSize(      offset, 160, width, height );
        sceneRect.SetOriginAndSize(     offset, 178, width, height );

        playRect.SetOriginAndSize(           7, 200,     8, height );
        stopRect.SetOriginAndSize(          22, 200,     8, height );
        animLoopRect.SetOriginAndSize(      37, 200,    36, height );

        u16 x = hidStatus.clickPosX;
        u16 y = hidStatus.clickPosY;

        // Determines the view camera selection.
        if (inRect(viewCamRect, x, y) )
        {
            padEvent = vw::system::G3D_SWITCH_NEXT_VIEW_CAMERA;
            return;
        }
        // Determines target selection.
        if( inRect(ctrlTargetRect, x, y) )
        {
            padEvent = vw::system::G3D_SWITCH_NEXT_CONTROL_TARGET_SEQUENTIALLY;
            return;
        }
        // Determines resource selection.
        if( inRect(ctrlResRect, x, y) )
        {
            padEvent = vw::system::G3D_SWITCH_NEXT_RESOURCE_SEQUENTIALLY;
            return;
        }
        // Determines animation selection.
        if( inRect(animRect, x, y) )
        {
            padEvent = vw::system::G3D_SWITCH_NEXT_ANIMATION;
            return;
        }
        // Determines scene selection.
        if( inRect(sceneRect, x, y) )
        {
            padEvent = vw::system::G3D_SWITCH_NEXT_SCENE;
            return;
        }
        // Determines animation controller selection.
        if( inRect(playRect, x, y) )
        {
            if( AnimManager::GetInstance()->ActiveAnimationExists()
                && vw::effect::AnimManager::GetInstance()->ActiveAnimationExists())
            {
                AnimManager::GetInstance()->PauseAll();
                vw::effect::AnimManager::GetInstance()->PauseAll();
                return;
            }
            else
            {
                AnimManager::GetInstance()->PlayAll();
                vw::effect::AnimManager::GetInstance()->PlayAll();
                return;
            }
        }
        if( inRect(stopRect, x, y) )
        {
            AnimManager::GetInstance()->StopAll();
            vw::g3d::AnimManager::GetInstance()->StopAll();
            return;
        }
        // Determines selection of the one-time animation loop.
        if( inRect(animLoopRect, x, y) )
        {
            AnimManager::GetInstance()->SetLoopFlag( !AnimManager::GetInstance()->GetLoopFlag() );
            return;
        }
    }

    if( hidStatus.touchPanel.touch && !mTouch )
    {
        mTouch = true;
    }
    else
    {
        mTouch = hidStatus.touchPanel.touch;
        return;
    }
}



}   /* namespace g3d */
}   /* namespace vw  */
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 1301 // padding inserted in struct.
#endif
