/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/SceneManager.h"
#include "../include/AnimElem.h"
#include "../include/ParticleAnimController.h"
#include "../../effect/include/AnimManager.h"

namespace vw  {
namespace g3d {

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::SetFrame(f32 frame)
{
    m_pParticleModel->ParticleAnimFrameController().SetFrame(frame);

    ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
    ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->ParticleAnimFrameController().SetFrame(frame);
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
ParticleAnimController::ParticleAnimController( ParticleModelElem *pParticleModelElem, nw::os::IAllocator *pAllocator )
    : m_pParticleModelElem( pParticleModelElem )
    , m_IsActive( true )
    , m_EmissionControlActive( false )
    , m_pAllocator( pAllocator )
    , m_EmissionStartFrame( 0 )
    , m_EmissionLength( 0 )
{
    m_pParticleContext = SceneManager::GetInstance()->GetParticleContext();
    m_pParticleModel = pParticleModelElem->GetSceneNode();

    m_pParticleModel->ParticleAnimFrameController().SetStepFrame( vw::effect::AnimManager::GetInstance()->GetStepFrame() );
    ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
    ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->ParticleAnimFrameController().SetStepFrame( vw::effect::AnimManager::GetInstance()->GetStepFrame() );
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
ParticleAnimController::~ParticleAnimController()
{
}


/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
f32
ParticleAnimController::GetFrame()
{
    return 0.f;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::SetStepFrame(f32 step)
{
    m_pParticleModel->ParticleAnimFrameController().SetStepFrame(step);

    ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
    ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->ParticleAnimFrameController().SetStepFrame(step);
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::StepForward()
{
    m_IsActive = true;
    Update();
    m_IsActive = false;
}


/*---------------------------------------------------------------------------*
  Name:
  Description:  Particle frame rewind is not performed.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::StepBackward()
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::Update()
{
    if (m_IsActive)
    {
        if (m_EmissionControlActive)
        {
            vw::effect::AnimManager *effectAnimManager = vw::effect::AnimManager::GetInstance();
            // Before the start frame is not updated.
            if (effectAnimManager->GetFrame() < m_EmissionStartFrame)
            {
                return;
            }

            // When the global start frame is before the local start frame, it is not updated.
            // 
            if (m_EmissionStartFrame < effectAnimManager->GetStartFrame())
            {
                return;
            }

            // For the instant that the start frame straddles, the difference is the set frame and update is not performed.
            if (effectAnimManager->GetFrame() > m_EmissionStartFrame
             && effectAnimManager->GetPreFrame() < m_EmissionStartFrame )
            {
                this->SetFrame( effectAnimManager->GetFrame() - m_EmissionStartFrame);
                return;
            }
        }

        m_pParticleModel->UpdateParticleFrame();

        {
            ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
            ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
            for (; it!=end_it; ++it)
            {
                (*it)->UpdateParticleFrame();
                (*it)->Emission( m_pParticleContext );
            }
        }

        nw::gfx::ParticleModel *pParticleModel = m_pParticleModelElem->GetSceneNode();
        u32 numParticleSets = pParticleModel->GetParticleSetsCount();
        for (int i=0; i<numParticleSets; ++i)
        {
            pParticleModel->GetParticleSets(i)->UpdateParticles( m_pParticleContext );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::PauseAndPlay()
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::Stop()
{
    m_IsActive = false;
    m_pParticleModel->ParticleAnimFrameController().SetFrame( 0.f );
    m_pParticleModel->ForeachParticleSet( nw::gfx::ParticleSetsClear() );

    ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
    ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->ParticleAnimFrameController().SetFrame( 0.f );
        (*it)->Reset();
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::Play()
{
    m_IsActive = true;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::Pause()
{
    if (m_IsActive)
    {
        m_IsActive = false;
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::SpeedUp(f32 step)
{
    {
        f32 currentStep = m_pParticleModel->ParticleAnimFrameController().GetStepFrame();
        if (currentStep + step >= 0)
            m_pParticleModel->ParticleAnimFrameController().SetStepFrame( currentStep + step );
        else
            m_pParticleModel->ParticleAnimFrameController().SetStepFrame( 0.f );
    }
    {
        ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
        ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
        for (; it!=end_it; ++it)
        {
            f32 currentStep = (*it)->ParticleAnimFrameController().GetStepFrame();
            if (currentStep + step >= 0)
                (*it)->ParticleAnimFrameController().SetStepFrame( currentStep + step );
            else
                (*it)->ParticleAnimFrameController().SetStepFrame( 0.f );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::Reset()
{
    m_pParticleModel->ForeachParticleSet(nw::gfx::ParticleSetsClear());

    ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
    ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
    for (; it!=end_it; ++it)
    {
        (*it)->Reset();
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::OnTargetModuleChanged(u16 id)
{
    if (id == vw::system::G3D_ID)
    {
        m_EmissionControlActive = false;
        ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
        ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
        for (; it!=end_it; ++it)
        {
            (*it)->ParticleAnimFrameController().SetEndFrame(16777215.f);
        }
    }
    else if (id == vw::system::EFFECT_ID)
    {
        m_EmissionControlActive = true;
        ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
        ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
        for (; it!=end_it; ++it)
        {
            (*it)->ParticleAnimFrameController().SetEndFrame(
                m_EmissionLength == 0 ? 16777215.f : m_EmissionLength);
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ParticleAnimController::SetEmissionLength(s32 length)
{
    m_EmissionLength = length;
    if (m_EmissionControlActive)
    {
        ParticleModelElem::ParticleEmitterArray::iterator  it = m_pParticleModelElem->GetParticleEmittersBegin();
        ParticleModelElem::ParticleEmitterArray::iterator  end_it = m_pParticleModelElem->GetParticleEmittersEnd();
        for (; it!=end_it; ++it)
        {
            (*it)->ParticleAnimFrameController().SetEndFrame(
                m_EmissionLength == 0 ? 16777215.f : m_EmissionLength);
        }
    }
}

}   /* namespace g3d */
}   /* namespace vw  */

