/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <math.h>
#include <viewer/system.h>
#include <viewer/g3d/CameraMgr.h>
#include <viewer/g3d/LightMgr.h>
#include "../include/Grid.h"
#include "../include/ResourceManager.h"
#include "../include/SceneManager.h"
#include "../include/AnimManager.h"
#include "../../effect/include/AnimManager.h"
#include "../include/InfoDrawer.h"
#include "../include/SceneRenderer.h"
#include <viewer/system/capture.h>
#include "../../system/include/RenderSystem.h"


#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_suppress 1301 // padding inserted in struct.
#pragma diag_suppress 2530 // padding added to end of struct.
#endif
extern "C" {
#if defined( NN_PCGX )
    extern u8* ICON_BEGIN;
    extern u8* ICON_END;
#else
    extern u8* ICON_BEGIN[];
    extern u8* ICON_END[];
#endif
}

namespace vw {
namespace g3d {

SceneRenderer* SceneRenderer::m_pInstance = NULL;

// The function object that enters preview display settings in the model's user parameters and divides up the display in that state.
class IsPreviewVisibleModelFunctor : public nw::gfx::ISceneUpdater::IsVisibleModelFunctor
{
public:
    virtual bool IsVisible(const nw::gfx::Model* model)
    {
        bool bVisible = model->IsVisible() && model->IsEnabledResults(nw::gfx::SceneNode::FLAG_IS_VISIBLE);
        if ( model->GetUserParameter<u32>() != 0 ) bVisible = false;
        return bVisible;
    }
} gPreviewVisibleModel;


// Render command generation function class
struct RenderSceneInternalFunctor : public std::unary_function<nw::gfx::RenderElement&, void>
{
    nw::gfx::RenderContext* m_RenderContext;
    nw::gfx::MeshRenderer* m_MeshRenderer;

    RenderSceneInternalFunctor(nw::gfx::RenderContext* renderContext, nw::gfx::MeshRenderer* meshRenderer)
    : m_RenderContext(renderContext), m_MeshRenderer(meshRenderer)
    {
        NW_NULL_ASSERT(renderContext);
        NW_NULL_ASSERT(meshRenderer);
    }

    void operator()(nw::gfx::RenderElement& element)
    {
        if (element.IsCommand())
        {
            nw::gfx::RenderCommand* command = element.GetCommand();
            NW_NULL_ASSERT(command);
            command->Invoke(this->m_RenderContext);
        }
        else
        {
            nw::gfx::ResMesh mesh = element.GetMesh();
            nw::gfx::Model* model = element.GetModel();
            model->PreRenderSignal()(model, mesh, this->m_RenderContext);
            this->m_MeshRenderer->RenderMesh(mesh, model);
            model->PostRenderSignal()(model, mesh, this->m_RenderContext);
        }
        NW_GL_ASSERT();
    }
};

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
SceneRenderer*
SceneRenderer::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::UpdateScene()
{
    // Skip checks the scene update process.
    if (vw::system::App::GetInstance().IsUpdateScene())
    {
        // Gets the scene context.
        nw::gfx::SceneContext* sceneContext = SceneManager::GetInstance()->GetSceneContext();
        nw::gfx::ParticleContext* particleContext = SceneManager::GetInstance()->GetParticleContext();

        // Starts scene update CPU load measurements.
        vw::system::PerformMgr::GetInstance()->StartCalcMeasure();
        {
            // Updates scene.
            {
            #if 0
                // Sorts in order of dependency on the particle set.
                std::sort(
                    this->m_pSceneContext->GetParticleSetBegin(),
                    this->m_pSceneContext->GetParticleSetEnd(),
                    nw::gfx::ParticleSetCompare());
            #endif

                m_pSceneUpdater->UpdateAll(sceneContext);
            }

            // Updates particles.
            vw::effect::AnimManager::GetInstance()->UpdateAll();
            vw::g3d::AnimManager::GetInstance()->UpdateAll();
        }
        vw::system::PerformMgr::GetInstance()->StopCalcMeasure();

        // Updates camera.
        CameraMgr::GetInstance()->UpdateCameraMatrix();

        // Starts render CPU load measurements.
        vw::system::PerformMgr::GetInstance()->StartDrawMeasure();
        {
            // Resets the render queue.
            m_pRenderQueue->Reset();

            f32 elapsedTime = 0;
            if (!m_BlinkTargets.empty())
            {
                int currentTick = nn::os::Tick::GetSystemCurrent();
                const f32 NANO_TO_MILLI = 1.0f / (1000.0f * 1000.0f);
                s64 span = static_cast<nn::os::Tick>(currentTick - m_BlinkStartTick).ToTimeSpan().GetNanoSeconds();
                elapsedTime = static_cast<f32>(span) * NANO_TO_MILLI;
            }

            // Inverts flashing of selected object.
            FlipSelectedObjectsVisibility(elapsedTime);
            {
                // Creates a render queue.
                SubmitView(
                    sceneContext,
                    *CameraMgr::GetInstance()->GetViewCamera(),
                    &gPreviewVisibleModel
                );
            }
            FlipSelectedObjectsVisibility(elapsedTime);

            // Sort the render queue.
            std::sort(
                m_pRenderQueue->Begin(),
                m_pRenderQueue->End(),
                nw::gfx::RenderElementCompare()
            );
        }
        vw::system::PerformMgr::GetInstance()->StopDrawMeasure();

        // Updates the directional light and spherical light.
        UpdateLightDir();
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static int GetRenderFlag()
{
    // Gets the 3D information render flag.
    bool IsRender3DInfo = InfoDrawer::GetInstance()->IsRender3DInfo();

    // Gets the 3D information individual render flag.
    bool* RenderFlag = InfoDrawer::GetInstance()->GetRenderFlag();

    // Flag setting auxiliary macro
    #define SET_RENDER_FLAG( name ) \
        if ( IsRender3DInfo && RenderFlag[SceneRenderer::name] ) { \
            renderFlag |= 0x01 << SceneRenderer::name; \
        }

    // Creating flags.
    int renderFlag = 0x01 << SceneRenderer::RenderQueue;
    SET_RENDER_FLAG( RenderGrid )
    SET_RENDER_FLAG( RenderLightDir )
    SET_RENDER_FLAG( RenderParticleEmitterForm )
    SET_RENDER_FLAG( RenderTargetCameraAxis )
    SET_RENDER_FLAG( RenderModelAxis )
    SET_RENDER_FLAG( RenderBoneAxis )
    SET_RENDER_FLAG( RenderParticleEmitterAxis )
    SET_RENDER_FLAG( RenderConstraintTargetBoneAxis )
    SET_RENDER_FLAG( RenderCameraIcon )
    SET_RENDER_FLAG( RenderLightIcon )
    SET_RENDER_FLAG( RenderTargetLine )
    SET_RENDER_FLAG( RenderViewFrustum )
    SET_RENDER_FLAG( RenderBoundingBox )
    SET_RENDER_FLAG( RenderDepthLevelSurface )

    return renderFlag;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::GetCorrectedRenderFlag( bool* RenderFlag )
{
    // Gets scene context.
    nw::gfx::SceneContext* sceneContext = SceneManager::GetInstance()->GetSceneContext();

    int renderFlag = GetRenderFlag();
    for ( int i = 0; i < RenderNum; ++i )
    {
        RenderFlag[ i ] = (bool)( renderFlag & ( 0x01 << i ) );
    }

    // Adds the flag conditions.
    {
        // Render flag of the target queue.
        RenderFlag[ RenderQueue ] =
            RenderFlag[ RenderQueue ] &&
            (m_pRenderQueue->Begin() != m_pRenderQueue->End());

        // Auxiliary information render flag
        RenderFlag[ RenderLightDir ] =
            RenderFlag[ RenderLightDir ] &&
            (InfoDrawer::CONTROL_LIGHT == InfoDrawer::GetInstance()->GetControlTargetType());

        // Emitter information render flag.
        RenderFlag[ RenderParticleEmitterForm ] =
            RenderFlag[ RenderParticleEmitterForm ] &&
            (sceneContext->GetParticleEmitterBegin() != sceneContext->GetParticleEmitterEnd());

        // Axis render flag
        RenderFlag[ RenderConstraintTargetBoneAxis ] =
            RenderFlag[ RenderConstraintTargetBoneAxis ] &&
            (SceneElemSelector::GetInstance()->GetModelsCount() != 0);

        // Icon render flag
        RenderFlag[ RenderCameraIcon ] =
            RenderFlag[ RenderCameraIcon ] &&
            (SceneElemSelector::GetInstance()->GetCamerasCount() != 0);
        RenderFlag[ RenderLightIcon ] =
            RenderFlag[ RenderLightIcon ] &&
            (SceneElemSelector::GetInstance()->GetLightsCount() != 0);
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::SetupDraw(bool* RenderFlag, nw::gfx::Camera* activeCamera)
{
    // Gets render context.
    nw::gfx::RenderContext* renderContext = vw::system::Video_GetRenderContext();
    m_pMeshRenderer->SetRenderContext( renderContext );

    // Sets active camera index.
    int cameraIndex = SceneElemSelector::GetInstance()->GetCameraIndex( activeCamera );
    renderContext->SetActiveCamera(
        cameraIndex
    );

    // Removes flag.
    GetCorrectedRenderFlag( RenderFlag );

    // Regenerates the command buffer only if the render queue is empty.
    if(RenderFlag[ RenderQueue ])
    {
        GLint command_id;
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_BINDING, &command_id);

        // Generates and saves the render command buffer.
        vw::system::CommandListSwapper* CommandListSwapper = vw::system::CommandListSwapper::GetInstance();
        CommandListSwapper->StartCommandSave();
        {
            // Starts render CPU load measurements.
            vw::system::PerformMgr::GetInstance()->StartDrawMeasure();
            {
                // Generates the commands of each model.
                std::for_each(
                    m_pRenderQueue->Begin(),
                    m_pRenderQueue->End(),
                    RenderSceneInternalFunctor(renderContext, m_pMeshRenderer)
                );
            }
            vw::system::PerformMgr::GetInstance()->StopDrawMeasure();
        }
        CommandListSwapper->EndCommandSave();

        nngxBindCmdlist( command_id );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::RenderStereoScene()
{
    // Does not skip when capture is executed.
    if( !vw::system::CaptureInterface::IsExecuting() )
    {
        // Returns immediately when there is a scene update skip.
        if(!vw::system::App::GetInstance().IsUpdateScene()) return;

        // Returns immediately if there is a render skip when a frame drops.
        if(!vw::system::App::GetInstance().IsRenderScene()) return;
    }

    // Rendering setting
    bool RenderFlag[ RenderNum ];
    SetupDraw(RenderFlag, CameraMgr::GetInstance()->GetViewCamera());

    // Executes capture.
    if( vw::system::CaptureInterface::IsExecuting() )
    {
        // Scene rendering
        DrawScene(
            CameraMgr::GetInstance()->GetLeftViewCamera(),
            RenderFlag,
            false
        );
        return;
    }

    // Changes render target.
    bool lowerPreview = vw::system::System_GetSetting()->lowerPreview;
    if( lowerPreview )
    {
        // Render on the lower screen
        vw::system::Video_SetRenderTarget(vw::system::LOWER_SCREEN);
        {
            // Scene rendering
            DrawScene(
                CameraMgr::GetInstance()->GetLeftViewCamera(),
                RenderFlag,
                false
            );
        }
    }
    else
    {
        // Renders screen for left eye.
        vw::system::Video_SetRenderTarget(vw::system::UPPER_SCREEN);
        {
            // Scene rendering
            DrawScene(
                CameraMgr::GetInstance()->GetLeftViewCamera(),
                RenderFlag,
                false
            );
        }

#if !defined( NN_PCGX )
        // Renders scene for right eye.
        vw::system::Video_SetRenderTarget(vw::system::EXTENSION_SCREEN);
        {
            // Scene rendering
            DrawScene(
                CameraMgr::GetInstance()->GetRightViewCamera(),
                RenderFlag,
                false
            );
        }
#endif
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::RenderScene(nw::gfx::IRenderTarget* renderTarget, nw::gfx::Camera* camera)
{
    vw::system::RenderSystem::GetInstance()->SetRenderTarget( renderTarget );

    // Rendering setting
    bool RenderFlag[ RenderNum ];
    SetupDraw(RenderFlag, CameraMgr::GetInstance()->GetViewCamera());

    // Scene rendering
    DrawScene(
        camera,
        RenderFlag,
        true
    );

    vw::system::CommandListSwapper::GetInstance()->ResetGpuProfiling();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::SetBlinkSpan(u32 span)
{
    m_BlinkSpan = span;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::SetBlinkNum(u32 num)
{
    m_BlinkNum = num;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::ut::MoveArray<SceneRenderer::BlinkAnimInfo>::iterator
SceneRenderer::GetBlinkAnimationsBegin()
{
    return m_BlinkTargets.begin();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
nw::ut::MoveArray<SceneRenderer::BlinkAnimInfo>::iterator
SceneRenderer::GetBlinkAnimationsEnd()
{
    return m_BlinkTargets.end();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::SetBlinkAnimation(nw::ut::MoveArray<BlinkAnimTarget>& targets)
{
    ResetBlinkAnimation();

    nw::ut::MoveArray<BlinkAnimTarget>::iterator it = targets.begin();
    nw::ut::MoveArray<BlinkAnimTarget>::iterator end = targets.end();
    for ( ; it != end; ++it )
    {
        switch( it->type )
        {
        case BlinkAnimTarget::BLINK_MODEL:
            AddBlinkModel(it->hashID);
            break;
        case BlinkAnimTarget::BLINK_MESH:
            AddBlinkMesh(it->hashID, it->index);
            break;
        case BlinkAnimTarget::BLINK_TEXTURE:
            AddBlinkTexture(it->hashID);
            break;
        case BlinkAnimTarget::BLINK_MATERIAL:
            AddBlinkMaterial(it->hashID, it->index);
            break;
        default:
            break;
        }
    }

    if (!m_BlinkTargets.empty())
    {
        m_BlinkStartTick = nn::os::Tick::GetSystemCurrent();
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::ResetBlinkAnimation()
{
    m_BlinkStartTick = 0;
    m_BlinkTargets.clear();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::AddDrawAxis(DrawAxisInfo& axis)
{
    DeleteDrawAxis( axis.key1, axis.key2 );
    m_DrawAxis.push_back( axis );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DeleteDrawAxis(AxisType type)
{
    nw::ut::MoveArray<DrawAxisInfo>::iterator it = m_DrawAxis.begin();
    for ( ; it != m_DrawAxis.end(); )
    {
        if ( it->axisType == type ) {
            nw::ut::MoveArray<DrawAxisInfo>::iterator it_next = it;
            ++it_next;
            it = m_DrawAxis.erase( it, it_next );
            continue;
        }
        it++;
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DeleteDrawAxis(void* key1, void* key2)
{
    nw::ut::MoveArray<DrawAxisInfo>::iterator it = m_DrawAxis.begin();
    for ( ; it != m_DrawAxis.end(); )
    {
        if ( it->key1 == key1 && it->key2 == key2 ) {
            nw::ut::MoveArray<DrawAxisInfo>::iterator it_next = it;
            ++it_next;
            it = m_DrawAxis.erase( it, it_next );
            continue;
        }
        it++;
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::ClearDrawAxis()
{
    m_DrawAxis.clear();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::SetRenderSortMode(nw::gfx::ISceneUpdater::RenderSortMode mode)
{
    // Releases key factory.
    nw::gfx::SafeDestroy(m_pPriorDepthRenderKeyFactory);

    // Regenerates key factory.
    if (mode == nw::gfx::ISceneUpdater::OPAQUE_MESH_BASE_AND_TRANSLUCENT_MODEL_BASE_SORT) {
        m_pPriorDepthRenderKeyFactory =
            nw::gfx::CreateTopPriorDepthReverseDepthRenderKeyFactory(m_pAllocator);
    } else {
        m_pPriorDepthRenderKeyFactory =
            nw::gfx::CreatePriorDepthReverseDepthRenderKeyFactory(m_pAllocator);
    }
    NW_NULL_ASSERT(m_pPriorDepthRenderKeyFactory);

    // Initializes render queue.
    m_pRenderQueue->Reset(
        m_pPriorMaterialRenderKeyFactory,
        m_pPriorDepthRenderKeyFactory,
        m_pPriorMaterialRenderKeyFactory,
        m_pPriorMaterialRenderKeyFactory
    );
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
SceneRenderer::SceneRenderer()
:   m_Grid(NULL), m_Axis(NULL), m_Cursor(NULL), m_HemiSphereCursor(NULL),
    m_LineModel(NULL), m_ViewFrustum(NULL), m_BoundingBoxLine(NULL),
    m_LightTexture(NULL), m_CameraTexture(NULL),
    m_BlinkTargets(nw::ut::MoveArray<BlinkAnimInfo>(Memory_GetAllocator())),
    m_BlinkStartTick(0), m_BlinkSpan(250), m_BlinkNum(3),
    m_DrawAxis(nw::ut::MoveArray<DrawAxisInfo>(Memory_GetAllocator())),
    m_ManipulatorScale(1.0f), m_CameraManipulatorScale(1.0f),
    m_DepthLevelSurfaceAlpha(0.25f)
{
    // Sets the allocator.
    m_pAllocator       = Memory_GetAllocator();
    m_pDeviceAllocator = Memory_GetAllocatorDevice();

    // Generates the key factory.
    m_pPriorMaterialRenderKeyFactory =
        nw::gfx::CreatePriorMaterialRenderKeyFactory(m_pAllocator);
    m_pPriorDepthRenderKeyFactory =
        nw::gfx::CreatePriorDepthReverseDepthRenderKeyFactory(m_pAllocator);

    // Generates the render queue.
    m_pRenderQueue = nw::gfx::RenderQueue::Builder()
        .IsFixedSizeMemory(false)
        .Create(m_pAllocator);

    // Initializes render queue.
    m_pRenderQueue->Reset(
        m_pPriorMaterialRenderKeyFactory,
        m_pPriorDepthRenderKeyFactory,
        m_pPriorMaterialRenderKeyFactory,
        m_pPriorMaterialRenderKeyFactory
    );

    // Generates the scene updater.
    m_pWorldMatrixUpdater = nw::gfx::WorldMatrixUpdater::Builder()
        .Create(m_pAllocator);
    m_pSkeletonUpdater = nw::gfx::SkeletonUpdater::Builder()
        .Create(m_pAllocator);

    m_pSceneUpdater = nw::gfx::SceneUpdater::Builder()
        .WorldMatrixUpdaterPtr(m_pWorldMatrixUpdater)
        .SkeletonUpdaterPtr(m_pSkeletonUpdater)
        .Create(m_pAllocator);

    // Generates the mesh renderer.
    m_pMeshRenderer = nw::gfx::MeshRenderer::Create(m_pAllocator);

    // Initializes resources when graphics are used.
    if( vw::system::App::GetUseGraphicsFlg() )
    {
        InitializeGraphics();
    }
}

template<typename TMemory>
inline void
SafeDelete( TMemory*& memory, nw::os::IAllocator* allocator )
{
    if (memory == NULL) return;
    memory->~TMemory();
    allocator->Free(static_cast<void*>(memory));
    memory = NULL;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
SceneRenderer::~SceneRenderer()
{
    nw::gfx::SafeDestroy(m_pPriorMaterialRenderKeyFactory);
    nw::gfx::SafeDestroy(m_pPriorDepthRenderKeyFactory);
    nw::gfx::SafeDestroy(m_pRenderQueue);
    nw::gfx::SafeDestroy(m_pSceneUpdater);
    nw::gfx::SafeDestroy(m_pMeshRenderer);
    SafeDelete(m_Grid, m_pAllocator);             // Releases grid.
    SafeDelete(m_Axis, m_pAllocator);             // Releases axis.
    SafeDelete(m_Cursor, m_pAllocator);           // Releases the cursor.
    SafeDelete(m_HemiSphereCursor, m_pAllocator); // Releases the hemisphere light cursor.
    SafeDelete(m_LineModel, m_pAllocator);        // Releases the light line model.
    SafeDelete(m_ViewFrustum, m_pAllocator);      // Releases the view frustum model.
    SafeDelete(m_BoundingBoxLine, m_pAllocator);  // Releases the bounding box model.
    SafeDelete(m_DepthLevelSurface, m_pAllocator);// Releases the stereoscopic view base plane model.
    SafeDelete(m_LightTexture, m_pAllocator);     // Releases the light icon texture.
    SafeDelete(m_CameraTexture, m_pAllocator);    // Releases the camera icon texture.
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(SceneRenderer));
    if (buf == NULL) return;

    m_pInstance = new(buf) SceneRenderer();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~SceneRenderer();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::InitializeGraphics()
{
    // Generates the grids.
    void *buf = m_pAllocator->Alloc(sizeof(Grid));
    m_Grid = new(buf) Grid();

    // Generates the axis.
    buf = m_pAllocator->Alloc(sizeof(vw::system::XYZTC));
    m_Axis = new(buf) vw::system::XYZTC();
    InitAxis();

    // Generates the cursor.
    buf = m_pAllocator->Alloc(sizeof(vw::system::XYZTC));
    m_Cursor = new(buf) vw::system::XYZTC();
    CreateCursor(m_Cursor);

    // Generates the hemisphere light cursor.
    buf = m_pAllocator->Alloc(sizeof(vw::system::XYZTC));
    m_HemiSphereCursor = new(buf) vw::system::XYZTC();
    CreateCursor(m_HemiSphereCursor);

    // Generates the light line model.
    buf = m_pAllocator->Alloc(sizeof(vw::system::XYZTC));
    m_LineModel = new(buf) vw::system::XYZTC();
    m_LineModel->CreateBuffer(1024, 1024);

    // Generates the view frustum model.
    buf = m_pAllocator->Alloc(sizeof(vw::system::XYZTC));
    m_ViewFrustum = new(buf) vw::system::XYZTC();
    m_ViewFrustum->CreateBuffer(1024, 1024);

    // Generates the bounding box model.
    buf = m_pAllocator->Alloc(sizeof(vw::system::XYZTC));
    m_BoundingBoxLine = new(buf) vw::system::XYZTC();
    m_BoundingBoxLine->CreateBuffer(4*4096, 6*4096);

    // Generates the stereoscopic base plane model.
    buf = m_pAllocator->Alloc( sizeof(vw::system::XYZTC) );
    m_DepthLevelSurface = new(buf) vw::system::XYZTC();
    m_DepthLevelSurface->CreateBuffer( 6, 6 );

    // Loads the light icon texture.
    buf = m_pAllocator->Alloc(sizeof(vw::system::Texture));
    m_LightTexture = new(buf) vw::system::Texture();
    m_LightTexture->SetData(ICON_BEGIN, "light.tga");

    // Loads the camera icon texture.
    buf = m_pAllocator->Alloc(sizeof(vw::system::Texture));
    m_CameraTexture = new(buf) vw::system::Texture();
    m_CameraTexture->SetData(ICON_BEGIN, "camera.tga");

    // Combiner settings
    nn::gr::Combiner combiner;
    combiner.stage[0].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    combiner.stage[0].rgb.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0;
    combiner.stage[0].rgb.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    combiner.stage[0].alpha.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    combiner.stage[0].alpha.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0;
    combiner.stage[0].alpha.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    combiner.stage[0].alpha.operand[0] = PICA_DATA_OPE_RGB_SRC_ALPHA;
    m_LightTexture->SetCombiner(combiner);
    m_CameraTexture->SetCombiner(combiner);
    m_LightTexture->SetFilter(PICA_DATA_TEXTURE_MAG_FILTER_LINEAR, PICA_DATA_TEXTURE_MIN_FILTER_LINEAR_MIPMAP_LINEAR);
    m_CameraTexture->SetFilter(PICA_DATA_TEXTURE_MAG_FILTER_LINEAR, PICA_DATA_TEXTURE_MIN_FILTER_LINEAR_MIPMAP_LINEAR);
    m_LightTexture->SetWrapMode(PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE, PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE);
    m_CameraTexture->SetWrapMode(PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE, PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE);
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawScene(nw::gfx::Camera* camera, bool isRender[RenderNum], bool copyCommand)
{
   // information display
    DrawInfo(camera,
             vw::g3d::CameraMgr::GetInstance()->GetTargetCamera(),
             vw::g3d::CameraMgr::GetInstance()->GetViewCamera(),
             isRender);

    // Camera and light icons
    DrawIcons(camera,
             vw::g3d::CameraMgr::GetInstance()->GetViewCamera(),
             isRender);

    if( isRender[RenderDepthLevelSurface] &&
        GetDepthLevelSurfaceAlpha() >= GetDepthLevelSurfaceMaxAlpha() )
    {
        // Renders the stereoscopic base plane. Note: When incorporated in a model with transparency, the plane is rendered in front of the model so that it does not disappear.
        DrawInfoDepthLevelSurface(camera,
            vw::g3d::CameraMgr::GetInstance()->GetViewCamera(),
            false);
    }

    // Gets the management object of the render command buffer.
    vw::system::CommandListSwapper* CommandListSwapper = vw::system::CommandListSwapper::GetInstance();

    // Gets render context.
    nw::gfx::RenderContext* renderContext = vw::system::Video_GetRenderContext();

    // Resets the state.
    renderContext->ResetState();

    // Sets camera information.
    renderContext->SetCameraMatrix(camera);

    // Starts GPU load measurement.
    int profiler = CommandListSwapper->AddGpuProfilingStartPoint(true);

    NW_ASSERT(profiler >= 0);
    {
        // Renders the queue.
        if(isRender[RenderQueue]) CommandListSwapper->ReuseCommand(copyCommand);
    }
    CommandListSwapper->SetGpuProfilingEndPoint(profiler);

    NW_GL_ASSERT();


    if( isRender[RenderDepthLevelSurface] &&
        GetDepthLevelSurfaceAlpha() < GetDepthLevelSurfaceMaxAlpha() )
    {
        // Stereoscopic base plane rendering * If a transmissive plane, this is render behind the model so that a non-transparent model in the back will not disappear.
        DrawInfoDepthLevelSurface(camera,
            vw::g3d::CameraMgr::GetInstance()->GetViewCamera(),
            true);
    }

#if defined( NN_PCGX )
    nngxSplitDrawCmdlist();
#endif
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawInfo(nw::gfx::Camera* camera, nw::gfx::Camera* baseCamera, bool isRender[RenderNum])
{
    DrawInfo(camera, baseCamera, baseCamera, isRender);
}

/*---------------------------------------------------------------------------*
  Name:

  Description: 3D information display

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawInfo(nw::gfx::Camera* camera,
                        nw::gfx::Camera* targetCamera,
                        nw::gfx::Camera* viewCamera,
                        bool isRender[RenderNum])
{
    // Render Grids
    if(isRender[RenderGrid]) m_Grid->Draw( camera, viewCamera );

    // Gets the axis render flag.
    bool isRenderAxis[AxisNum] = {
        isRender[RenderModelAxis],
        isRender[RenderBoneAxis],
        isRender[RenderParticleEmitterAxis],
        isRender[RenderConstraintTargetBoneAxis],
    };
    bool renderAxis = false;
    for ( int i = 0; i < AxisNum; ++i ) renderAxis |= isRenderAxis[ i ];

    // Renders auxiliary information.
    if (
        isRender[RenderParticleEmitterForm] ||
        renderAxis ||
        isRender[RenderTargetCameraAxis] ||
        isRender[RenderLightDir] ||
        isRender[RenderTargetLine] ||
        isRender[RenderViewFrustum] ||
        isRender[RenderBoundingBox]
    ) {
        // Prepares 3D rendering.
        vw::system::Draw3DHelper& draw3D = vw::system::Draw3DHelper::GetInstance();
        draw3D.SetScreenSize(vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight());
        draw3D.Begin();
        {
            // Sets the state.
            draw3D.Enable( vw::system::Draw3DHelper::ALPHABLEND );
            draw3D.Disable( vw::system::Draw3DHelper::DEPTHWRITE );

            // Configure the camera matrix
            draw3D.SetView( camera->ViewMatrix() );
            draw3D.SetProj( camera->ProjectionMatrix() );

            // Renders emitter form.
            if(isRender[RenderParticleEmitterForm]) DrawEmitterForm();

             // Axis rendering
            if(renderAxis) DrawAxis(isRenderAxis, viewCamera);

            // Renders the gaze point of the target camera.
            if(isRender[RenderTargetCameraAxis]) DrawTargetCameraAxis(
                targetCamera,
                viewCamera );

            // Light direction display
            if(isRender[RenderLightDir]) DrawLightDir( camera );

            // Renders target lines.
            if(isRender[RenderTargetLine]) DrawTargetLine( camera, viewCamera );

            // Renders view frustum.
            if(isRender[RenderViewFrustum]) DrawViewFrustum( camera, viewCamera );

            // Renders bounding box.
            if(isRender[RenderBoundingBox]) DrawBoundingBox( camera, viewCamera );
        }
        draw3D.End();
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description: Renders the stereoscopic base plane.

  Arguments:   Camera
               viewCamera:
               isTransparent semi-transparent

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawInfoDepthLevelSurface(nw::gfx::Camera* camera,
                        nw::gfx::Camera* viewCamera,
                        bool isTransparent)
{
    // Prepares 3D rendering.
    vw::system::Draw3DHelper& draw3D = vw::system::Draw3DHelper::GetInstance();
    draw3D.SetScreenSize(vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight());
    draw3D.Begin();
    {
        // Sets the state.
        if( isTransparent )
            draw3D.Enable( vw::system::Draw3DHelper::ALPHABLEND );
        draw3D.Enable( vw::system::Draw3DHelper::DEPTHWRITE );

         // Configure the camera matrix
         draw3D.SetView( camera->ViewMatrix() );
         draw3D.SetProj( camera->ProjectionMatrix() );

         // Renders the stereoscopic base plane.
         DrawDepthLevelSurface( camera, viewCamera );
    }
    draw3D.End();
}

/*---------------------------------------------------------------------------*
  Name:

  Description: Displays the camera and light icons.

  Arguments:    camera,
                viewCamera,
                isRender[RenderNum]    Same array as DrawInf third parameter.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawIcons(nw::gfx::Camera* camera,
                        nw::gfx::Camera* viewCamera,
                        bool isRender[RenderNum])
{
     // Icon rendering
    if (isRender[RenderCameraIcon] || isRender[RenderLightIcon])
    {
        // Prepares 2D rendering.
        vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
        draw2D.SetScreenSize(vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight());
        draw2D.BeginDraw();
        {
            // State setting - Conducts alpha testing because the icon will display properly even when overlapped with a semi-transparent model.
            draw2D.Enable( vw::system::Draw2DHelper::ALPHATEST | vw::system::Draw2DHelper::DEPTHTEST | vw::system::Draw2DHelper::DEPTHWRITE );
            draw2D.Disable(vw::system::Draw2DHelper::ALPHABLEND | vw::system::Draw2DHelper::CULLFACE );

            // Light icon rendering
            if (isRender[RenderLightIcon])
            {
                m_LightTexture->SetState();
                DrawLightIcon( camera );
                draw2D.Flush();
            }

            // Camera icon rendering
            if (isRender[RenderCameraIcon])
            {
                m_CameraTexture->SetState();
                DrawCameraIcon( camera, viewCamera );
            }
        }

        draw2D.EndDraw();
        draw2D.Disable( vw::system::Draw2DHelper::ALPHATEST );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawEmitterForm()
{
    nw::gfx::SceneContext *sceneContext = SceneManager::GetInstance()->GetSceneContext();
    nw::gfx::ParticleEmitterArray::iterator it = sceneContext->GetParticleEmitterBegin();
    nw::gfx::ParticleEmitterArray::iterator end = sceneContext->GetParticleEmitterEnd();
    for ( ; it != end; ++it )
    {
        nw::gfx::ParticleEmitter* emitter = *it;
        NW_NULL_ASSERT(emitter);

        const nw::ut::FloatColor color = nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 0.5f);
        nw::math::MTX34 world = emitter->WorldMatrix();

        const nw::gfx::ResParticleEmitter resEmitter = emitter->GetResParticleEmitter();
        const nw::gfx::ResParticleForm    resForm    = resEmitter.GetParticleForm();
        if (!resForm.IsValid()) continue;

        switch(resForm.GetTypeInfo())
        {
        case nw::gfx::ResParticleCubeForm::TYPE_INFO:
            {
                vw::system::Draw3DHelper::GetInstance().Disable( vw::system::Draw3DHelper::CULLFACE );
                const nw::gfx::ResParticleCubeForm form = nw::ut::ResStaticCast<nw::gfx::ResParticleCubeForm>( resForm );
                vw::system::Draw3DUtil::DrawCube( world, form.GetScale(), color );
            }
            break;
        case nw::gfx::ResParticleCylinderForm::TYPE_INFO:
            {
                vw::system::Draw3DHelper::GetInstance().Enable( vw::system::Draw3DHelper::CULLFACE );
                const nw::gfx::ResParticleCylinderForm form = nw::ut::ResStaticCast<nw::gfx::ResParticleCylinderForm>( resForm );
                vw::system::Draw3DUtil::DrawCylinder( world, form.GetScale(), color );
            }
            break;
        case nw::gfx::ResParticleDiscForm::TYPE_INFO:
            {
                vw::system::Draw3DHelper::GetInstance().Disable( vw::system::Draw3DHelper::CULLFACE );
                const nw::gfx::ResParticleDiscForm form = nw::ut::ResStaticCast<nw::gfx::ResParticleDiscForm>( resForm );
                nw::math::VEC3 scale( form.GetScale().x,1.0f,form.GetScale().y );
                vw::system::Draw3DUtil::DrawDisc( world, scale, color );
            }
            break;
        case nw::gfx::ResParticlePointForm::TYPE_INFO:
            {
                vw::system::Draw3DHelper::GetInstance().Enable( vw::system::Draw3DHelper::CULLFACE );
                const nw::gfx::ResParticlePointForm form = nw::ut::ResStaticCast<nw::gfx::ResParticlePointForm>( resForm );
                nw::math::VEC3 scale( 0.2f,0.2f,0.2f );
                vw::system::Draw3DUtil::DrawSphere( world, scale, color );
            }
            break;
        case nw::gfx::ResParticleRectangleForm::TYPE_INFO:
            {
                vw::system::Draw3DHelper::GetInstance().Disable( vw::system::Draw3DHelper::CULLFACE );
                const nw::gfx::ResParticleRectangleForm form = nw::ut::ResStaticCast<nw::gfx::ResParticleRectangleForm>( resForm );
                nw::math::VEC3 scale( form.GetScale().x,1.0f,form.GetScale().y );
                vw::system::Draw3DUtil::DrawRectangle( world, scale, color );
            }
            break;
        case nw::gfx::ResParticleSphereForm::TYPE_INFO:
            {
                vw::system::Draw3DHelper::GetInstance().Enable( vw::system::Draw3DHelper::CULLFACE );
                const nw::gfx::ResParticleSphereForm form = nw::ut::ResStaticCast<nw::gfx::ResParticleSphereForm>( resForm );
                vw::system::Draw3DUtil::DrawSphere( world, form.GetScale(), color );
            }
            break;
        default:
            NW_FATAL_ERROR("unknown form type");
            break;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawLightIcon( nw::gfx::Camera* camera )
{
    const nw::ut::FloatColor color = nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 0.5f);
    nw::math::MTX44 wvpMat;
    nw::math::MTX44Mult(&wvpMat,camera->ProjectionMatrix(),nw::math::MTX44(camera->ViewMatrix()));

    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    SceneElemSelector::LightList::iterator it = SceneElemSelector::GetInstance()->GetLightsBegin();
    SceneElemSelector::LightList::iterator end = SceneElemSelector::GetInstance()->GetLightsEnd();

    u32 width  = vw::system::Video_GetScreenWidth();
    u32 height = vw::system::Video_GetScreenHeight();

    for ( ; it != end; ++it )
    {
        nw::gfx::Light* light = (*it)->GetSceneNode();
        NW_NULL_ASSERT(light);

        nw::math::VEC3 p = light->Transform().GetTranslate();
        nw::math::VEC4 t;
        nw::math::VEC3Transform(&t, wvpMat, p);
        if (t.w > 0.0f)
        {
            t /= t.w;
#if defined( NN_PCGX )
            nw::math::VEC4 t2 = t;
            t.x = (-0.5f * t2.y + 0.5f) * height + 0.5f;
            t.y = (-0.5f * -t2.x + 0.5f) * width + 0.5f;
#else
            t.x = (-0.5f * t.x + 0.5f) * height + 0.5f;
            t.y = (-0.5f * t.y + 0.5f) * width + 0.5f;
#endif

            const f32 r = 8.0f;
            draw2D.SetDepth(t.z);
            draw2D.FillRectangle( t.y-r, t.x-r, r*2, r*2, color );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawCameraIcon( nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera )
{
    const nw::ut::FloatColor color = nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 0.5f);
    nw::math::Matrix44 wvpMat;
    nw::math::MTX44Mult(
        &wvpMat,camera->ProjectionMatrix(),
        nw::math::Matrix44(camera->ViewMatrix())
    );

    u32 width  = vw::system::Video_GetScreenWidth();
    u32 height = vw::system::Video_GetScreenHeight();

    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();

    SceneElemSelector::CameraList::iterator it = SceneElemSelector::GetInstance()->GetCamerasBegin();
    SceneElemSelector::CameraList::iterator end = SceneElemSelector::GetInstance()->GetCamerasEnd();
    for ( ; it != end; ++it )
    {
        nw::gfx::Camera* cam = (*it)->GetSceneNode();
        NW_NULL_ASSERT(cam);
        if( cam == viewCamera ) continue;

        nw::math::VEC3 cameraPos;
        cam->Transform().GetTranslate( &cameraPos );

        nw::math::VEC4 t;
        nw::math::VEC3Transform(&t, wvpMat, cameraPos);
        if (t.w > 0.0f)
        {
            t /= t.w;
#if defined( NN_PCGX )
            nw::math::VEC4 t2 = t;
            t.x = (-0.5f * t2.y + 0.5f) * height + 0.5f;
            t.y = (-0.5f * -t2.x + 0.5f) * width + 0.5f;
#else
            t.x = (-0.5f * t.x + 0.5f) * height + 0.5f;
            t.y = (-0.5f * t.y + 0.5f) * width + 0.5f;
#endif

            const f32 r = 8.0f;
            draw2D.SetDepth(t.z);
            draw2D.FillRectangle( t.y-r, t.x-r, r*2, r*2, color );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawAxis(bool isRenderAxis[AxisNum], nw::gfx::Camera* viewCamera)
{
    vw::system::Draw3DHelper::GetInstance().Disable(vw::system::Draw3DHelper::ALPHABLEND |
                                                    vw::system::Draw3DHelper::CULLFACE);
    vw::system::Draw3DHelper::GetInstance().Enable (vw::system::Draw3DHelper::DEPTHTEST |
                                                    vw::system::Draw3DHelper::DEPTHWRITE);

    nw::ut::MoveArray<DrawAxisInfo>::iterator it = m_DrawAxis.begin();
    nw::ut::MoveArray<DrawAxisInfo>::iterator end = m_DrawAxis.end();
    for ( ; it != end; ++it )
    {
        const nw::gfx::CalculatedTransform* transform = it->axisTransform;
        if( transform && isRenderAxis[it->axisType] )
        {
            nw::math::MTX34 world = transform->TransformMatrix();

            nw::math::VEC3 objPos, cameraPos;
            transform->GetTranslate(&objPos);
            viewCamera->Transform().GetTranslate(&cameraPos);

            f32 dist = nw::math::VEC3Len(cameraPos - objPos);
            if(dist > 0.f)
            {
                f32 scale = dist / 75.f;
                scale *= m_ManipulatorScale;

                nw::math::MTX34 m;
                nw::math::MTX34Scale(&m, nw::math::VEC3(scale, scale, scale));
                nw::math::MTX34Mult(&world, &world, &m);

                vw::system::Draw3DHelper::GetInstance().SetWorld( world );
                m_Axis->Draw();
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawTargetCameraAxis(nw::gfx::Camera* targetCamera, nw::gfx::Camera* viewCamera)
{
    vw::system::Draw3DHelper::GetInstance().Disable(vw::system::Draw3DHelper::ALPHABLEND |
                                                    vw::system::Draw3DHelper::CULLFACE);
    vw::system::Draw3DHelper::GetInstance().Enable (vw::system::Draw3DHelper::DEPTHTEST |
                                                    vw::system::Draw3DHelper::DEPTHWRITE);

    f32 dist = 0.0f;
    nw::math::VEC3 axisPos;

    if (targetCamera->GetResCamera().GetViewType() == nw::gfx::ResCamera::VIEWTYPE_LOOKAT)
    {
        nw::gfx::LookAtTargetViewUpdater* viewUpdater =
            nw::ut::DynamicCast< nw::gfx::LookAtTargetViewUpdater* >( targetCamera->GetViewUpdater() );
        nw::gfx::ResLookAtTargetViewUpdater resViewUpdater =
            nw::gfx::ResStaticCast< nw::gfx::ResLookAtTargetViewUpdater >( viewUpdater->GetResource() );

        axisPos = resViewUpdater.GetTargetPosition();

        nw::math::VEC3 cameraPos;
        viewCamera->Transform().GetTranslate( &cameraPos );
        dist = nw::math::VEC3Len( cameraPos - axisPos );
    }
    else if(targetCamera->GetResCamera().GetViewType() == nw::gfx::ResCamera::VIEWTYPE_AIM)
    {
        nw::gfx::AimTargetViewUpdater* viewUpdater =
            nw::ut::DynamicCast< nw::gfx::AimTargetViewUpdater* >( targetCamera->GetViewUpdater() );
        nw::gfx::ResAimTargetViewUpdater resViewUpdater =
            nw::gfx::ResStaticCast< nw::gfx::ResAimTargetViewUpdater >( viewUpdater->GetResource() );

        axisPos = resViewUpdater.GetTargetPosition();

        nw::math::VEC3 cameraPos;
        viewCamera->Transform().GetTranslate( &cameraPos );
        dist = nw::math::VEC3Len( cameraPos - axisPos );
    }

    if( dist > 0.0f )
    {
        nw::math::MTX34 world;
        nw::math::MTX34Identity( &world );
        world.SetupTranslate( axisPos );

        f32 scale = dist / 75.0f;
        scale *= m_CameraManipulatorScale;
        nw::math::MTX34 scaleMat;
        nw::math::MTX34Scale( &scaleMat, nw::math::VEC3(scale,scale,scale) );
        nw::math::MTX34Mult( &world, &world, &scaleMat );
        vw::system::Draw3DHelper::GetInstance().SetWorld(world);
        m_Axis->Draw();
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawLightDir( nw::gfx::Camera* camera )
{
    struct Inner
    {
        static nw::math::MTX34 CalcWorld( const nw::math::MTX34& view, nw::math::VEC3 dir )
        {
            if( nw::math::VEC3SquareLen( &dir ) == 0.0f )
            {
                dir = nw::math::VEC3(0.0f, 1.0f, 0.0f);
            }
            nn::math::VEC3TransformNormal(&dir, &view, &dir);
            nw::math::VEC3Normalize(&dir, &dir);

            nw::math::VEC3 camUp(0.0f, 1.0f, 0.0f);
            if( dir.x == 0.0f && dir.z == 0.0f )
            {
                camUp = nw::math::VEC3(1.0f, 0.0f, 0.0f);
            }
            nn::math::VEC3TransformNormal(&camUp, &view, &camUp);
            nn::math::VEC3 camPos(45.0f, 25.0f, -100.0f);
            nw::math::VEC3 target = camPos + dir;

            nw::math::MTX34 world;
            nw::math::MTX34LookAt (&world, &camPos, &camUp, &target);
            nw::math::MTX34Inverse(&world, &world);

            return world;
        }
        // VertexLight, FragmentLight
        static void RenderCursor(s32 kind, nw::math::MTX34 world, vw::system::XYZTC* cursor)
        {
            switch(kind){
            case nw::gfx::ResLight::KIND_DIRECTIONAL:
                {
                    nw::math::VEC3 mov(1.0f,0.0f,0.0f);
                    nw::math::MTX34MultTranslate(&world, &world, &mov);
                    vw::system::Draw3DHelper::GetInstance().SetWorld( world );
                    cursor->Draw();
                    mov = nw::math::VEC3(-2.0f,0.0f,0.0f);
                    nn::math::MTX34MultTranslate(&world, &world, &mov);
                    vw::system::Draw3DHelper::GetInstance().SetWorld( world );
                    cursor->Draw();
                }
                break;
            case nw::gfx::ResLight::KIND_SPOT:
                {
                    nw::math::MTX34 rot;
                    nw::math::MTX34RotAxisDeg(&rot, nw::math::VEC3(1.0f,0.0f,0.0f), 90.0f);
                    nw::math::MTX34Mult(&world, &world, &rot);
                    vw::system::Draw3DHelper::GetInstance().SetWorld( world );
                    const nw::ut::FloatColor color = nw::ut::FloatColor(0.0f, 1.0f, 0.0f, 1.0f);
                    vw::system::Draw3DUtil::DrawCone( world, nw::math::VEC3(1.5f,2.5f,1.5f), color );
                }
                break;
            }
        }
        // HemiSphereLight
        static void RenderHemiSphereCursor(nw::math::MTX34 world, vw::system::XYZTC* cursor, nw::gfx::res::ResHemiSphereLight HemiSphere)
        {
            const nw::ut::Color8 SkyColor = HemiSphere.GetSkyColor();
            const nw::ut::Color8 GroundColor = HemiSphere.GetGroundColor();

            vw::system::XYZTC::Vertex* vertex = cursor->GetVertexData();
            vertex[0].setColor(GroundColor); vertex[1].setColor(SkyColor); vertex[2].setColor(GroundColor);
            vertex[3].setColor(GroundColor); vertex[4].setColor(SkyColor); vertex[5].setColor(GroundColor);
            cursor->UpdateVertex(0,6);

            nw::math::VEC3 mov(1.0f,0.0f,0.0f);
            nw::math::MTX34MultTranslate(&world, &world, &mov);
            vw::system::Draw3DHelper::GetInstance().SetWorld( world );
            cursor->Draw();

            mov = nw::math::VEC3 (-2.0f,0.0f,0.0f);
            nn::math::MTX34MultTranslate(&world, &world, &mov);
            vw::system::Draw3DHelper::GetInstance().SetWorld( world );
            cursor->Draw();
        }
    };
    nw::gfx::Light* light = vw::g3d::LightMgr::GetInstance()->GetTargetLight();
    if( light )
    {
        vw::system::Draw3DHelper::GetInstance().Disable(vw::system::Draw3DHelper::ALPHABLEND |
                                                        vw::system::Draw3DHelper::CULLFACE |
                                                        vw::system::Draw3DHelper::DEPTHTEST );
        nw::math::MTX34 view;
        nw::math::MTX34Identity( &view );
        vw::system::Draw3DHelper::GetInstance().SetView( view );

        nw::math::MTX44 proj;
        nw::math::MTX44PerspectivePivotRad(&proj, 0.5973f, 1.6667f, 1.0f, 10000.0f, nn::math::PIVOT_UPSIDE_TO_TOP);
        vw::system::Draw3DHelper::GetInstance().SetProj( proj );

        nw::gfx::res::ResSceneObject resObject = light->GetResSceneObject();
        switch (resObject.GetTypeInfo())
        {
    #if defined(NW_GFX_VERTEX_LIGHT_ENABLED)
        case nw::gfx::res::ResVertexLight::TYPE_INFO:
            {
                s32 kind = nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject).GetLightKind();
                nw::math::VEC3 dir = nw::ut::DynamicCast<nw::gfx::VertexLight*>(light)->Direction();
                nw::math::MTX34 world = Inner::CalcWorld(camera->ViewMatrix(), dir);
                Inner::RenderCursor(kind, world, m_Cursor);
            }
            break;
    #endif
        case nw::gfx::res::ResFragmentLight::TYPE_INFO:
            {
                s32 kind = nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject).GetLightKind();
                nw::math::VEC3 dir = nw::ut::DynamicCast<nw::gfx::FragmentLight*>(light)->Direction();
                nw::math::MTX34 world = Inner::CalcWorld(camera->ViewMatrix(), dir);
                Inner::RenderCursor(kind, world, m_Cursor);
            }
            break;
        case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
            {
                nw::gfx::res::ResHemiSphereLight HemiSphere = nw::ut::ResDynamicCast<nw::gfx::res::ResHemiSphereLight>(resObject);
                nw::math::VEC3 dir = HemiSphere.GetDirection();
                nw::math::MTX34 world = Inner::CalcWorld(camera->ViewMatrix(), dir);
                Inner::RenderHemiSphereCursor(world, m_HemiSphereCursor, HemiSphere);
            }
            break;
        }
        vw::system::Draw3DHelper::GetInstance().Enable(vw::system::Draw3DHelper::DEPTHTEST);
        vw::system::Draw3DHelper::GetInstance().SetView( camera->ViewMatrix() );
        vw::system::Draw3DHelper::GetInstance().SetProj( camera->ProjectionMatrix() );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawTargetLine( nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera )
{
    struct LineRenderer
    {
    public:
        LineRenderer(vw::system::XYZTC* model, nw::math::VEC3 cameraPos) : m_LineModel( model ), m_DrawCount( 0 ), m_CameraPos( cameraPos ){
        }
        ~LineRenderer()
        {
            if( m_DrawCount > 0 )
            {
                m_LineModel->UpdateVertex(0, m_DrawCount * 4);
                m_LineModel->UpdateIndex (0, m_DrawCount * 6);
                m_LineModel->Draw( m_DrawCount * 6 );
            }
        }
        void DrawLine(const nw::math::VEC3& pt1, const nw::math::VEC3& pt2 )
        {
            if( pt1 == pt2 || m_DrawCount >= 127)
                return;

            nw::math::VEC3 center = (pt1 + pt2) * 0.5f;
            nw::math::VEC3 ray = center - m_CameraPos;
            nw::math::VEC3 dir = pt1 - pt2;
            nw::math::VEC3Cross( &dir, &dir, &ray );
            if( nw::math::VEC3Len( &dir ) == 0.0f )
                return;

            nw::math::VEC3Normalize( &dir, &dir );
            f32 dist1 = nw::math::VEC3Dist( &pt1, &m_CameraPos );
            f32 dist2 = nw::math::VEC3Dist( &pt2, &m_CameraPos );

            vw::system::XYZTC::Vertex* vertex = m_LineModel->GetVertexData() + m_DrawCount * 4;
            u16*                        index = m_LineModel->GetIndexData()  + m_DrawCount * 6;

            const f32 scale = 500.0f;
            f32 scale1 = dist1 / scale;
            f32 scale2 = dist2 / scale;
            nw::math::VEC3 p0 = pt1 - dir * scale1;
            nw::math::VEC3 p1 = pt1 + dir * scale1;
            nw::math::VEC3 p2 = pt2 - dir * scale2;
            nw::math::VEC3 p3 = pt2 + dir * scale2;

            const nw::ut::FloatColor color = nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 0.5f);
            vertex[0].setPos( p0.x, p0.y, p0.z ).setColor( color );
            vertex[1].setPos( p1.x, p1.y, p1.z ).setColor( color );
            vertex[2].setPos( p2.x, p2.y, p2.z ).setColor( color );
            vertex[3].setPos( p3.x, p3.y, p3.z ).setColor( color );

            const u32 offset = m_DrawCount * 4;
            index[0] = offset + 0; index[1] = offset + 1; index[2] = offset + 2;
            index[3] = offset + 2; index[4] = offset + 1; index[5] = offset + 3;

            m_DrawCount++;
        }

    private:
        vw::system::XYZTC*  m_LineModel;
        u32                 m_DrawCount;
        nw::math::VEC3      m_CameraPos;
    };

    vw::system::Draw3DHelper::GetInstance().Disable(vw::system::Draw3DHelper::ALPHABLEND |
                                                    vw::system::Draw3DHelper::CULLFACE);
    vw::system::Draw3DHelper::GetInstance().Enable (vw::system::Draw3DHelper::DEPTHTEST |
                                                    vw::system::Draw3DHelper::DEPTHWRITE);

    nw::math::MTX34 invViewMat;
    nw::math::MTX34Inverse( &invViewMat, &camera->ViewMatrix() );
    nw::math::VEC3 cameraPos( invViewMat.m[0][3],
                              invViewMat.m[1][3],
                              invViewMat.m[2][3] );

    nw::math::MTX34 world;
    nw::math::MTX34Identity( &world );
    vw::system::Draw3DHelper::GetInstance().SetWorld( world );

    LineRenderer lineRenderer( m_LineModel, cameraPos );

    // draw light line
    nw::ut::MoveArray<g3d::Light>* lights = LightMgr::GetInstance()->GetLightArray();
    for(nw::ut::MoveArray<g3d::Light>::iterator i=lights->begin(); i!=lights->end(); ++i)
    {
        nw::gfx::Light* light = i->GetLight();
        if( light->GetResSceneObject().GetTypeInfo() == nw::gfx::res::ResAmbientLight::TYPE_INFO )
        {
            continue;
        }

        nw::math::VEC3 pt1 = i->GetPosition();
        nw::math::VEC3 pt2 = i->GetTargetPos();
        lineRenderer.DrawLine( pt1, pt2 );
    }

    // draw camera line
    for (SceneElemSelector::CameraList::iterator it = SceneElemSelector::GetInstance()->GetCamerasBegin();
                                                 it!= SceneElemSelector::GetInstance()->GetCamerasEnd(); ++it )
    {
        nw::gfx::Camera* cam = (*it)->GetSceneNode();
        NW_NULL_ASSERT(cam);
        if( cam == viewCamera )
            continue;

        nw::gfx::ResCameraViewUpdater       resViewUpdater             = cam->GetViewUpdater()->GetResource();
        nw::gfx::ResAimTargetViewUpdater    resAimTargetViewUpdater    = nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>(resViewUpdater);
        nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>(resViewUpdater);
        if( resAimTargetViewUpdater.IsValid() )
        {
            nw::math::VEC3 targetPos = resAimTargetViewUpdater.GetTargetPosition();
            lineRenderer.DrawLine( cam->GetPosition(), targetPos );
        }
        else if( resLookAtTargetViewUpdater.IsValid() )
        {
            nw::math::VEC3 targetPos = resLookAtTargetViewUpdater.GetTargetPosition();
            lineRenderer.DrawLine( cam->GetPosition(), targetPos );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawViewFrustum( nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera )
{
    struct FrustumRenderer
    {
    public:
        FrustumRenderer(vw::system::XYZTC* model, nw::math::VEC3 cameraPos) : m_LineModel( model ), m_DrawCount( 0 ), m_CameraPos( cameraPos ){
        }
        ~FrustumRenderer()
        {
            if( m_DrawCount > 0 )
            {
                m_LineModel->UpdateVertex(0, m_DrawCount * 4);
                m_LineModel->UpdateIndex (0, m_DrawCount * 6);
                m_LineModel->Draw( m_DrawCount * 6 );
            }
        }
        void DrawFrustum( nw::gfx::Camera* camera )
        {
            nw::math::MTX34 invView = camera->InverseViewMatrix();
            nw::math::MTX44 invProj = camera->InverseProjectionMatrix();

            nw::math::VEC3 p[8] = {nw::math::VEC3( 1.0f, 1.0f, 0.0f), nw::math::VEC3(-1.0f, 1.0f, 0.0f),
                                   nw::math::VEC3(-1.0f,-1.0f, 0.0f), nw::math::VEC3( 1.0f,-1.0f, 0.0f),
                                   nw::math::VEC3( 1.0f, 1.0f,-1.0f), nw::math::VEC3(-1.0f, 1.0f,-1.0f),
                                   nw::math::VEC3( 1.0f,-1.0f,-1.0f), nw::math::VEC3(-1.0f,-1.0f,-1.0f)};
            nw::math::VEC4 p4;
            for( u32 i=0; i<8; ++i )
            {
                p4 = nw::math::VEC4(p[i].x, p[i].y, p[i].z, 1.0f);
                nw::math::VEC4Transform( &p4, &invProj, &p4 );
                p4 /= p4.w;
                p[i] = nw::math::VEC3(p4.x, p4.y, p4.z);
                nw::math::VEC3Transform( &p[i], &invView, &p[i] );
            }

            f32 dist = 25.0f;

            nw::math::VEC3 org = camera->GetPosition();
            nw::gfx::ResCameraViewUpdater       resViewUpdater             = camera->GetViewUpdater()->GetResource();
            nw::gfx::ResAimTargetViewUpdater    resAimTargetViewUpdater    = nw::gfx::ResDynamicCast<nw::gfx::ResAimTargetViewUpdater>(resViewUpdater);
            nw::gfx::ResLookAtTargetViewUpdater resLookAtTargetViewUpdater = nw::gfx::ResDynamicCast<nw::gfx::ResLookAtTargetViewUpdater>(resViewUpdater);
            if (resAimTargetViewUpdater.IsValid())
            {
                nw::math::VEC3 targetPos = resAimTargetViewUpdater.GetTargetPosition();
                dist = nw::math::VEC3Dist( &targetPos, &org );
            }
            else if (resLookAtTargetViewUpdater.IsValid())
            {
                nw::math::VEC3 targetPos = resLookAtTargetViewUpdater.GetTargetPosition();
                dist = nw::math::VEC3Dist( &targetPos, &org );
            }

            const nw::ut::FloatColor white = nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 0.5f);
            const nw::ut::FloatColor red   = nw::ut::FloatColor(1.0f, 0.5f, 0.5f, 0.5f);
            const nw::ut::FloatColor blue  = nw::ut::FloatColor(0.5f, 0.5f, 1.0f, 0.5f);

            nw::math::VEC3 zAxis = -invView.GetColumn(2);
            nw::math::VEC3Normalize( &zAxis, &zAxis );

            switch( camera->GetResCamera().GetProjectionType() )
            {
            case nw::gfx::ResCamera::PROJTYPE_PERSPECTIVE:

                // draw line
                for(u32 i = 0; i < 4; ++i)
                {
                    // draw near clip
                    DrawLine( p[i%4], p[(i+1) % 4], red );

                    // draw frustum
                    nw::math::VEC3 dir = p[4 + i] - org;
                    if(nw::math::VEC3Len( &dir ) > 0.0f )
                    {
                        nw::math::VEC3Normalize( &dir, &dir );
                        f32 d = nw::math::VEC3Dot( &dir, &zAxis );
                        if( d > 0.0f )
                        {
                            DrawLine( org, org + dir * ( dist / d ), white );
                        }
                    }
                }
                break;
            case nw::gfx::ResCamera::PROJTYPE_ORTHO:

                // draw line
                for(u32 i = 0; i < 4; ++i)
                {
                    // draw near clip
                    DrawLine( p[i], p[(i+1) % 4], blue );

                    // draw frustum
                    DrawLine( org, p[i], white );

                    nw::math::VEC3 dir = p[4 + i] - org;
                    if(nw::math::VEC3Len( &dir ) > 0.0f )
                    {
                        nw::math::VEC3Normalize( &dir, &dir );
                        DrawLine( p[i], p[i] + dir * dist, white );
                    }
                }
                break;
            }
        }
        void DrawLine(const nw::math::VEC3& pt1, const nw::math::VEC3& pt2, nw::ut::FloatColor color )
        {
            if( pt1 == pt2 )
                return;

            nw::math::VEC3 center = (pt1 + pt2) * 0.5f;
            nw::math::VEC3 ray = center - m_CameraPos;
            nw::math::VEC3 dir = pt1 - pt2;
            nw::math::VEC3Cross( &dir, &dir, &ray );
            if( nw::math::VEC3Len( &dir ) == 0.0f )
                return;

            nw::math::VEC3Normalize( &dir, &dir );
            f32 dist1 = nw::math::VEC3Dist( &pt1, &m_CameraPos );
            f32 dist2 = nw::math::VEC3Dist( &pt2, &m_CameraPos );

            vw::system::XYZTC::Vertex* vertex = m_LineModel->GetVertexData() + m_DrawCount * 4;
            u16*                        index = m_LineModel->GetIndexData()  + m_DrawCount * 6;

            const f32 scale = 500.0f;
            f32 scale1 = dist1 / scale;
            f32 scale2 = dist2 / scale;
            nw::math::VEC3 p0 = pt1 - dir * scale1;
            nw::math::VEC3 p1 = pt1 + dir * scale1;
            nw::math::VEC3 p2 = pt2 - dir * scale2;
            nw::math::VEC3 p3 = pt2 + dir * scale2;

            vertex[0].setPos( p0.x, p0.y, p0.z ).setColor( color );
            vertex[1].setPos( p1.x, p1.y, p1.z ).setColor( color );
            vertex[2].setPos( p2.x, p2.y, p2.z ).setColor( color );
            vertex[3].setPos( p3.x, p3.y, p3.z ).setColor( color );

            const u32 offset = m_DrawCount * 4;
            index[0] = offset + 0; index[1] = offset + 1; index[2] = offset + 2;
            index[3] = offset + 2; index[4] = offset + 1; index[5] = offset + 3;

            m_DrawCount++;
        }

    private:
        vw::system::XYZTC*  m_LineModel;
        u32                 m_DrawCount;
        nw::math::VEC3      m_CameraPos;
    };

    // set state
    vw::system::Draw3DHelper::GetInstance().Disable(vw::system::Draw3DHelper::ALPHABLEND |
                                                    vw::system::Draw3DHelper::CULLFACE );
    vw::system::Draw3DHelper::GetInstance().Enable( vw::system::Draw3DHelper::DEPTHTEST);

    nw::math::MTX34 world;
    nw::math::MTX34Identity( &world );
    vw::system::Draw3DHelper::GetInstance().SetWorld( world );

    nw::math::MTX34 invViewMat;
    nw::math::MTX34Inverse( &invViewMat, &camera->ViewMatrix() );
    nw::math::VEC3 cameraPos( invViewMat.m[0][3],
                              invViewMat.m[1][3],
                              invViewMat.m[2][3] );

    FrustumRenderer frustumRenderer( m_ViewFrustum, cameraPos );
    SceneElemSelector::CameraList::iterator it = SceneElemSelector::GetInstance()->GetCamerasBegin();
    SceneElemSelector::CameraList::iterator end = SceneElemSelector::GetInstance()->GetCamerasEnd();
    for ( ; it != end; ++it )
    {
        nw::gfx::Camera* cam = (*it)->GetSceneNode();
        NW_NULL_ASSERT(cam);
        if( cam == viewCamera ) continue;

        frustumRenderer.DrawFrustum( cam );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawBoundingBox( nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera )
{
    struct LineRenderer
    {
    public:
        LineRenderer(vw::system::XYZTC* model, nw::math::VEC3 cameraPos) : m_LineModel( model ), m_DrawCount( 0 ), m_CameraPos( cameraPos ){
        }
        ~LineRenderer()
        {
            if( m_DrawCount > 0 )
            {
                m_LineModel->UpdateVertex(0, m_DrawCount * 4);
                m_LineModel->UpdateIndex (0, m_DrawCount * 6);
                m_LineModel->Draw( m_DrawCount * 6 );
            }
        }
        void DrawLine(const nw::math::VEC3& pt1, const nw::math::VEC3& pt2 )
        {
            if( pt1 == pt2)
                return;

            nw::math::VEC3 center = (pt1 + pt2) * 0.5f;
            nw::math::VEC3 ray = center - m_CameraPos;
            nw::math::VEC3 dir = pt1 - pt2;
            nw::math::VEC3Cross( &dir, &dir, &ray );
            if( nw::math::VEC3Len( &dir ) == 0.0f )
                return;

            nw::math::VEC3Normalize( &dir, &dir );
            f32 dist1 = nw::math::VEC3Dist( &pt1, &m_CameraPos );
            f32 dist2 = nw::math::VEC3Dist( &pt2, &m_CameraPos );

            vw::system::XYZTC::Vertex* vertex = m_LineModel->GetVertexData() + m_DrawCount * 4;
            u16*                        index = m_LineModel->GetIndexData()  + m_DrawCount * 6;

            const f32 scale = 500.0f;
            f32 scale1 = dist1 / scale;
            f32 scale2 = dist2 / scale;
            nw::math::VEC3 p0 = pt1 - dir * scale1;
            nw::math::VEC3 p1 = pt1 + dir * scale1;
            nw::math::VEC3 p2 = pt2 - dir * scale2;
            nw::math::VEC3 p3 = pt2 + dir * scale2;

            const nw::ut::FloatColor color = nw::ut::FloatColor(1.0f, 1.0f, 1.0f, 0.5f);
            vertex[0].setPos( p0.x, p0.y, p0.z ).setColor( color );
            vertex[1].setPos( p1.x, p1.y, p1.z ).setColor( color );
            vertex[2].setPos( p2.x, p2.y, p2.z ).setColor( color );
            vertex[3].setPos( p3.x, p3.y, p3.z ).setColor( color );

            const u32 offset = m_DrawCount * 4;
            index[0] = offset + 0; index[1] = offset + 1; index[2] = offset + 2;
            index[3] = offset + 2; index[4] = offset + 1; index[5] = offset + 3;

            m_DrawCount++;
        }

    private:
        vw::system::XYZTC*  m_LineModel;
        u32                 m_DrawCount;
        nw::math::VEC3      m_CameraPos;
    };

    vw::system::Draw3DHelper::GetInstance().Disable(vw::system::Draw3DHelper::ALPHABLEND |
                                                    vw::system::Draw3DHelper::CULLFACE);
    vw::system::Draw3DHelper::GetInstance().Enable (vw::system::Draw3DHelper::DEPTHTEST );

    nw::math::VEC3 cameraPos = viewCamera->GetPosition();

    LineRenderer lineRenderer( m_BoundingBoxLine, cameraPos );

    nw::math::MTX34 world;
    for ( SceneElemSelector::ModelList::iterator n = SceneElemSelector::GetInstance()->GetModelsBegin();
                                                 n!= SceneElemSelector::GetInstance()->GetModelsEnd(); ++n )
    {
        nw::gfx::Model* model = (*n)->GetSceneNode();
        NW_NULL_ASSERT(model);

        nw::gfx::SkeletalModel* skeletalModel = nw::ut::DynamicCast<nw::gfx::SkeletalModel*>( model );

        for (nw::gfx::ResMeshArray::iterator m = model->GetResMeshes().begin() ;
                                             m!= model->GetResMeshes().end(); ++m )
        {
            nw::gfx::res::ResMesh mesh = *m;
            nw::gfx::res::ResShape shape = model->GetResModel().GetShapes( mesh.GetShapeIndex() );

            nw::ut::ResOrientedBoundingBox obb = shape.GetOrientedBoundingBox();

            for(s32 j = 0; j < shape.GetPrimitiveSetsCount(); ++j)
            {
                nw::gfx::res::ResPrimitiveSet primitiveSet = shape.GetPrimitiveSets(j);
                nw::math::VEC3 vertex[8] = {
                     nw::math::VEC3(-0.5f, 0.5f,-0.5f),
                     nw::math::VEC3(-0.5f, 0.5f, 0.5f),
                     nw::math::VEC3( 0.5f, 0.5f, 0.5f),
                     nw::math::VEC3( 0.5f, 0.5f,-0.5f),
                     nw::math::VEC3(-0.5f,-0.5f,-0.5f),
                     nw::math::VEC3(-0.5f,-0.5f, 0.5f),
                     nw::math::VEC3( 0.5f,-0.5f, 0.5f),
                     nw::math::VEC3( 0.5f,-0.5f,-0.5f)
                };
                for(u32 i = 0; i < 8; ++i){
                    vertex[i] *= obb.GetSize();
                    nw::math::VEC3Transform(&vertex[i], &obb.GetOrientationMatrix(), &vertex[i]);
                    vertex[i] += obb.GetCenterPosition();
                }

                // Supports only no skinning.
                bool noneSkinning = primitiveSet.GetSkinningMode() == nw::gfx::res::ResPrimitiveSet::SKINNING_MODE_NONE;
                if( !noneSkinning )
                    continue;

                s32 boneIndexCount = primitiveSet.GetBoneIndexTableCount();
                if(boneIndexCount != 1)
                    continue;

                if (skeletalModel == NULL || boneIndexCount == 0)
                {
                    for(size_t k = 0; k < 8; ++k)
                    {
                        nw::math::VEC3Transform(&vertex[k], &model->WorldMatrix(), &vertex[k]);
                    }
                }
                else
                {
                    s32 matrixPalletCount = primitiveSet.GetBoneIndexTableCount();

                    nw::gfx::Skeleton* skeleton = skeletalModel->GetSkeleton();
                    nw::math::MTX34 matrix;
                    for(size_t k = 0; k < 8; ++k)
                    {
                        nw::math::VEC3 temp_pos(0.0f,0.0f,0.0f);

                        for (s32 count = 0; count < matrixPalletCount; ++count)
                        {
                            s32 boneIndex = primitiveSet.GetBoneIndexTable( count );
                            nw::gfx::res::ResBone bone = skeleton->GetResSkeleton().GetBones( boneIndex );

                            if( noneSkinning )
                            {
                                matrix = *skeleton->WorldMatrixPose().GetMatrix( boneIndex );
                            }
                            /*else
                            {
                                // For rigid skinning and smooth skinning, OBB is output in the world coordinate system.
                                // The conversion matrix converts the world vertex to the local coordinates of the bones in the binding pose.
                                // Next, these are converted by the world coordinates of the animated bones.
                                nw::math::MTX34Mult(&matrix,
                                                    skeleton->WorldMatrixPose().GetMatrix( boneIndex ),
                                                    &bone.GetInverseBaseMatrix());
                               //matrix = model->WorldMatrix(); // If the pose is binding pose, then the identity matrix is OK.
                            }*/
                            nw::math::VEC3 p;
                            nw::math::VEC3Transform(&p, &matrix, &vertex[k]);
                            temp_pos += p;
                        }

                        temp_pos /= matrixPalletCount;
                        vertex[k] = temp_pos;
                    }
                }
                for(int i = 0; i < 4; ++i){
                    lineRenderer.DrawLine(vertex[i],  vertex[(i+1)%4]);
                    lineRenderer.DrawLine(vertex[i+4],vertex[4+(i+1)%4]);
                    lineRenderer.DrawLine(vertex[i],  vertex[4+i]);
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ComputeCameraAngle
(
    const nw::math::VEC3&   cameraPos,
    const nw::math::VEC3&   targetPos,
    const f32               twist,
    nw::math::VEC3&         result
)
{
    nw::math::VEC3 lookReverse = ( cameraPos - targetPos );
    f32 dist = nw::math::VEC3Len( &lookReverse );
    nw::math::VEC3SafeNormalize( &lookReverse, &lookReverse, lookReverse );

    result.x = nw::math::AsinFIdx( -lookReverse.y );
    result.y = nw::math::Atan2FIdx( -lookReverse.x, lookReverse.z );
    result.z = twist;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::DrawDepthLevelSurface(nw::gfx::Camera* camera, nw::gfx::Camera* viewCamera)
{
    const nw::ut::FloatColor color = nw::ut::FloatColor(0.85f, 0.85f, 0.85f, m_DepthLevelSurfaceAlpha);
    {
        nw::math::VEC3 surfacePos;
        const DirectionalInfo&    info   = vw::g3d::DirectionalObj::GetDirectionalInfo();
        vw::system::Draw3DHelper& draw3d = vw::system::Draw3DHelper::GetInstance();

        // Camera updater
        nw::gfx::ResCameraViewUpdater       viewUpdater   = viewCamera->GetViewUpdater()->GetResource();
        nw::gfx::ResAimTargetViewUpdater    aimUpdater    = nw::gfx::ResDynamicCast< nw::gfx::ResAimTargetViewUpdater >( viewUpdater );
        nw::gfx::ResLookAtTargetViewUpdater lookAtUpdater = nw::gfx::ResDynamicCast< nw::gfx::ResLookAtTargetViewUpdater >( viewUpdater );
        nw::gfx::ResRotateViewUpdater       rotateUpdater = nw::gfx::ResDynamicCast< nw::gfx::ResRotateViewUpdater >( viewUpdater );

        // Gets camera position.
        viewCamera->Transform().GetTranslate( &surfacePos );

        // Calculates the default virtual space distance.
        f32 depthLevel = viewCamera->GetProjectionUpdater()->GetResource().GetNear() + 5.0f;

        // Calculates the virtual space distance.
        if ( info.depthLevel != 0.0f )
        { depthLevel = info.depthLevel; }
        else
        { depthLevel += info.depthLevelBias; }

        nw::math::VEC3 rotate( 0.0f, 0.0f, 0.0f );
        nw::math::VEC3 dir( 0.0f, 0.0f, 1.0f );

        // Aim mode
        if ( aimUpdater.IsValid() )
        {
            // Calculates the rotation angle.
            ComputeCameraAngle( viewCamera->GetPosition(), aimUpdater.GetTargetPosition(), aimUpdater.GetTwist(), rotate );

            // Calculates the directional vectors.
            dir = aimUpdater.GetTargetPosition() - viewCamera->GetPosition();
            nw::math::VEC3SafeNormalize( &dir, &dir, dir );
        }
        // LooktAt mode
        else if ( lookAtUpdater.IsValid() )
        {
            // Calculates the rotation angle.
            ComputeCameraAngle( viewCamera->GetPosition(), lookAtUpdater.GetTargetPosition(), 0.0f, rotate );

            // Calculates the directional vectors.
            dir = lookAtUpdater.GetTargetPosition() - viewCamera->GetPosition();
            nw::math::VEC3SafeNormalize( &dir, &dir, dir );
        }
        // Rotate mode
        else if ( rotateUpdater.IsValid() )
        {
            // Gets the rotational angle.
            rotate.x = rotateUpdater.GetViewRotate().x;
            rotate.y = rotateUpdater.GetViewRotate().y;
            rotate.z = rotateUpdater.GetViewRotate().z;

            // Calculates the directional vectors.
            f32 xs = nw::math::SinFIdx( rotate.x );
            f32 xc = nw::math::CosFIdx( rotate.x );
            f32 ys = nw::math::SinFIdx( -rotate.y );
            f32 yc = nw::math::CosFIdx( -rotate.y );

            dir.x = -xc * ys;
            dir.y = -xs;
            dir.z = xc * yc;
            nw::math::VEC3SafeNormalize( &dir, &dir, dir );
        }

        // Calculates the render position.
        surfacePos += dir * ( depthLevel );

        // Sets the matrix.
        {
            nw::math::Matrix34 world;
            nw::math::Matrix34 translation;
            nw::math::Matrix34 rotation;
            nw::math::Matrix34 twistaion;

            // Parallel displacement
            nw::math::MTX34Translate( &translation, &surfacePos );
            // Billboard processing (Twists, etc., are not supported.)
            nw::math::MTX34RotXYZFIdx( &rotation, rotate.x, -rotate.y, 0.0f );
            // Calculates the world matrix.
            nw::math::MTX34Mult( &world, &translation, &rotation );
            // Sets the world matrix.
            draw3d.SetWorld( world );
        }

        // Temporary saving of the state.
        vw::system::Draw3DHelper::STATE enableState = draw3d.GetEnable();
        vw::system::Draw3DHelper::STATE disableState = draw3d.GetDisable();

        // Disables culling.
        draw3d.Disable( vw::system::Draw3DHelper::CULLFACE );
        // Enables alpha blending.
        draw3d.Enable( vw::system::Draw3DHelper::ALPHABLEND );

        // Sets the vertex data.
        {
            const f32 s_BoxVtxStreamPos[] = {
                -2.0f, -1.0f, 0.0f,    2.0f, -1.0f,  0.0f,   -2.0f, 1.0f, 0.0f,
                -2.0f,  1.0f, 0.0f,    2.0f, -1.0f,  0.0f,    2.0f, 1.0f, 0.0f
            };

            vw::system::XYZTC::Vertex* vertex = m_DepthLevelSurface->GetVertexData();
            u16* index  = m_DepthLevelSurface->GetIndexData();

            // Sets the vertex coordinates and vertex index.
            for( u32 i=0; i<6; ++i )
            {
                u32 j = i * 3;
                vertex[ i ].x = s_BoxVtxStreamPos[ j + 0 ];
                vertex[ i ].y = s_BoxVtxStreamPos[ j + 1 ];
                vertex[ i ].z = s_BoxVtxStreamPos[ j + 2 ];
                vertex[ i ].color = color;
                index[ i ] = i;
            }
        }

        // Updates the data.
        m_DepthLevelSurface->UpdateData();

        // Renders
        m_DepthLevelSurface->Draw();

        // Restores the state.
        draw3d.Disable( disableState );
        draw3d.Enable( enableState );
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::InitAxis()
{
    // Generates the axis model.
    typedef nw::math::VEC3 VEC3;
    struct Inner
    {
        static void axis( vw::system::XYZTC::Vertex* v, VEC3 w, VEC3 h, nw::ut::Color8 color )
        {
            VEC3 p1=w+h, p2=h, p3=w-h, p4=-h;
            v[0].setPos(p1.x,p1.y,p1.z).setColor( color );
            v[1].setPos(p2.x,p2.y,p2.z).setColor( color );
            v[2].setPos(p3.x,p3.y,p3.z).setColor( color );
            v[3].setPos(p3.x,p3.y,p3.z).setColor( color );
            v[4].setPos(p2.x,p2.y,p2.z).setColor( color );
            v[5].setPos(p4.x,p4.y,p4.z).setColor( color );
        }
    };

    u32 vertexNum = 6 * 3 * 2;
    m_Axis->CreateBuffer(vertexNum, vertexNum);
    vw::system::XYZTC::Vertex* vertex = m_Axis->GetVertexData();
    u16*                        index = m_Axis->GetIndexData();

    const float width=5.0f, height=0.2f;
    const nw::ut::Color8 colorXAxis = 0xFF0000FF;
    const nw::ut::Color8 colorYAxis = 0x00FF00FF;
    const nw::ut::Color8 colorZAxis = 0x0000FFFF;

    // set x axis
    Inner::axis( vertex+6*0, VEC3(width,0.0f,0.0f), VEC3(0.0f,height,0.0f), colorXAxis );
    Inner::axis( vertex+6*1, VEC3(width,0.0f,0.0f), VEC3(0.0f,0.0f,height), colorXAxis );
    // set y axis
    Inner::axis( vertex+6*2, VEC3(0.0f,width,0.0f), VEC3(0.0f,0.0f,height), colorYAxis );
    Inner::axis( vertex+6*3, VEC3(0.0f,width,0.0f), VEC3(height,0.0f,0.0f), colorYAxis );
    // set z axis
    Inner::axis( vertex+6*4, VEC3(0.0f,0.0f,width), VEC3(height,0.0f,0.0f), colorZAxis );
    Inner::axis( vertex+6*5, VEC3(0.0f,0.0f,width), VEC3(0.0f,height,0.0f), colorZAxis );

    for(s32 i=0; i<vertexNum; ++i)
    {
        index[i] = i;
    }

    m_Axis->UpdateData();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::CreateCursor(vw::system::XYZTC* cursor)
{
    // Generates the axis model.
    typedef nw::math::VEC3 VEC3;

    u32 vertexNum = 6;
    cursor->CreateBuffer(vertexNum, vertexNum);
    vw::system::XYZTC::Vertex* vertex = cursor->GetVertexData();
    u16*                        index = cursor->GetIndexData();

    const float height=1.0f, depth=5.0f;
    const nw::ut::Color8 color = 0x0000FFFF;

    vertex[0].setPos( 0.0f, height, depth/2).setColor( color );
    vertex[1].setPos( 0.0f,   0.0f,-depth/2).setColor( color );
    vertex[2].setPos( 0.0f,-height, depth/2).setColor( color );

    vertex[3].setPos(-height, 0.0f, depth/2).setColor( color );
    vertex[4].setPos(   0.0f, 0.0f,-depth/2).setColor( color );
    vertex[5].setPos( height, 0.0f, depth/2).setColor( color );

    for(s32 i=0; i<vertexNum; ++i)
    {
        index[i] = i;
    }

    cursor->UpdateData();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::UpdateLightDir()
{
    SceneElemSelector::LightList::iterator it = SceneElemSelector::GetInstance()->GetLightsBegin();
    SceneElemSelector::LightList::iterator end = SceneElemSelector::GetInstance()->GetLightsEnd();

    for ( ; it != end; ++it )
    {
        nw::gfx::Light* light = (*it)->GetSceneNode();
        NW_NULL_ASSERT(light);

        s32 kind = nw::gfx::ResLight::KIND_UNUSED;
        nw::math::VEC3 direction;
        nw::gfx::res::ResSceneObject resObject = light->GetResSceneObject();
        switch ( resObject.GetTypeInfo() )
        {
        case nw::gfx::res::ResAmbientLight::TYPE_INFO:
            continue;
    #if defined(NW_GFX_VERTEX_LIGHT_ENABLED)
        case nw::gfx::res::ResVertexLight::TYPE_INFO:
            {
                kind = nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject).GetLightKind();
                direction = nw::ut::DynamicCast<nw::gfx::VertexLight*>(light)->Direction();
            }
            break;
    #endif
        case nw::gfx::res::ResFragmentLight::TYPE_INFO:
            {
                kind = nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject).GetLightKind();
                direction = nw::ut::DynamicCast<nw::gfx::FragmentLight*>(light)->Direction();
            }
            break;
        case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
            {
                kind = nw::gfx::ResLight::KIND_DIRECTIONAL;
                direction = nw::ut::ResDynamicCast<nw::gfx::res::ResHemiSphereLight>(resObject).GetDirection();
            }
            break;
        }

        // Resets parallel lighting and hemisphere lighting.
        if(kind == nw::gfx::ResLight::KIND_DIRECTIONAL)
        {
            nw::math::VEC3 lookReverse = -direction;
            if( nw::math::VEC3Len( &lookReverse ) != 0.0f )
            {
                nw::math::VEC3Normalize(&lookReverse, &lookReverse);
                f32 pitch = nw::math::AsinFIdx(-lookReverse.y);
                f32 yaw   = nw::math::Atan2FIdx(-lookReverse.x, lookReverse.z);
                f32 xs=nw::math::SinFIdx(pitch), xc=nw::math::CosFIdx(pitch);
                f32 ys=nw::math::SinFIdx(yaw),   yc=nw::math::CosFIdx(yaw);
                const f32 dist = 10.0f;
                light->Transform().SetTranslate(nw::math::VEC3(-xc*ys*dist, -xs*dist, xc*yc*dist));
            }
        }
    }

    nw::ut::MoveArray<g3d::Light>* lights = LightMgr::GetInstance()->GetLightArray();
    for(nw::ut::MoveArray<g3d::Light>::iterator i=lights->begin(); i!=lights->end(); ++i)
    {
        nw::gfx::Light* light = i->GetLight();
        NW_NULL_ASSERT(light);

        s32 kind = nw::gfx::ResLight::KIND_UNUSED;
        nw::math::VEC3 direction;
        nw::gfx::res::ResSceneObject resObject = light->GetResSceneObject();
        switch ( resObject.GetTypeInfo() )
        {
        case nw::gfx::res::ResAmbientLight::TYPE_INFO:
            continue;
    #if defined(NW_GFX_VERTEX_LIGHT_ENABLED)
        case nw::gfx::res::ResVertexLight::TYPE_INFO:
            {
                kind = nw::ut::ResDynamicCast<nw::gfx::res::ResVertexLight>(resObject).GetLightKind();
                direction = nw::ut::DynamicCast<nw::gfx::VertexLight*>(light)->Direction();
            }
            break;
    #endif
        case nw::gfx::res::ResFragmentLight::TYPE_INFO:
            {
                kind = nw::ut::ResDynamicCast<nw::gfx::res::ResFragmentLight>(resObject).GetLightKind();
                direction = nw::ut::DynamicCast<nw::gfx::FragmentLight*>(light)->Direction();
            }
            break;
        case nw::gfx::res::ResHemiSphereLight::TYPE_INFO:
            {
                kind = nw::gfx::ResLight::KIND_DIRECTIONAL;
                direction = nw::ut::ResDynamicCast<nw::gfx::res::ResHemiSphereLight>(resObject).GetDirection();
            }
            break;
        }

        // Resets parallel lighting and hemisphere lighting.
        if(kind == nw::gfx::ResLight::KIND_DIRECTIONAL)
        {
            nw::math::VEC3 lookReverse = -direction;
            if( nw::math::VEC3Len( &lookReverse ) != 0.0f )
            {
                nw::math::VEC3Normalize(&lookReverse, &lookReverse);
                f32 pitch = nw::math::AsinFIdx(-lookReverse.y);
                f32 yaw   = nw::math::Atan2FIdx(-lookReverse.x, lookReverse.z);
                f32 xs=nw::math::SinFIdx(pitch), xc=nw::math::CosFIdx(pitch);
                f32 ys=nw::math::SinFIdx(yaw),   yc=nw::math::CosFIdx(yaw);
                const f32 dist = 10.0f;
                light->Transform().SetTranslate(nw::math::VEC3(-xc*ys*dist, -xs*dist, xc*yc*dist));
            }
        }

        // Resets the spotlight's virtual gaze point.
        const f32 dist = 10.0f;
        if(kind == nw::gfx::ResLight::KIND_SPOT)
        {
            nw::math::VEC3 lightPos = i->GetPosition();
            nw::math::VEC3 lightDir = i->GetDirection();
            if( nw::math::VEC3Len( &lightDir ) != 0.0f )
                nw::math::VEC3Normalize( &lightDir, &lightDir );

            nw::math::VEC3 targetPos = lightPos + lightDir * dist;
            i->SetTargetPos( targetPos );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::AddBlinkAnim(BlinkAnimInfo& info)
{
    nw::ut::MoveArray<BlinkAnimInfo>::iterator it = m_BlinkTargets.begin();
    nw::ut::MoveArray<BlinkAnimInfo>::iterator end = m_BlinkTargets.end();
    for ( ; it != end; ++it ) {
        if ( it->model == info.model && it->mesh.ptr() == info.mesh.ptr() ) break;
    }
    if ( it == end ) m_BlinkTargets.push_back(info);
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::AddBlinkModel(unsigned int hashID)
{
    nw::gfx::Model* model = SceneManager::GetInstance()->FindModelNode(hashID);
    if (model) {
        nw::gfx::ResMeshArray::iterator it = model->GetResMeshes().begin();
        nw::gfx::ResMeshArray::iterator end = model->GetResMeshes().end();
        for ( ; it != end; ++it ) {
            BlinkAnimInfo info;
            info.model = model; info.mesh = *it;
            AddBlinkAnim(info);
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::AddBlinkMesh(unsigned int hashID, unsigned int index)
{
    nw::gfx::Model* model = SceneManager::GetInstance()->FindModelNode(hashID);
    if (model) {
        if ( index < model->GetResMeshes().size() ) {
            BlinkAnimInfo info;
            info.model = model; info.mesh = model->GetResMeshes()[ index ];
            AddBlinkAnim(info);
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::AddBlinkTexture(unsigned int hashID)
{
    nw::gfx::res::ResTexture texture = SceneManager::GetInstance()->FindResTexture(hashID);

    SceneManager::ModelList::iterator it = SceneManager::GetInstance()->GetModelsBegin();
    SceneManager::ModelList::iterator end = SceneManager::GetInstance()->GetModelsEnd();
    for ( ; it != end; ++it )
    {
        nw::gfx::Model* model = it->GetSceneNode();
        if (!model) continue;

        for ( int i = 0; i < model->GetMaterialCount(); ++i )
        {
            nw::gfx::res::ResMaterial material = model->GetMaterial( i )->GetBuffer();
            for ( int j = 0; j < material.GetTextureMappersCount(); ++j )
            {
                nw::gfx::res::ResTextureMapper mapper = material.GetTextureMappers(j);
                if (!mapper.IsValid()) continue;

                if (mapper.GetTexture().Dereference().ptr() == texture.ptr())
                {
                    AddBlinkMaterial( it->GetResSource()->GetHashID(), i );
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::AddBlinkMaterial(unsigned int hashID, unsigned int index)
{
    nw::gfx::Model* model = SceneManager::GetInstance()->FindModelNode(hashID);
    if (model) {
        nw::gfx::ResMeshArray::iterator it = model->GetResMeshes().begin();
        nw::gfx::ResMeshArray::iterator end = model->GetResMeshes().end();
        for ( ; it != end; ++it  ) {
            if ((*it).GetMaterialIndex() == index) {
                BlinkAnimInfo info;
                info.model = model; info.mesh = *it;
                AddBlinkAnim(info);
            }
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::FlipSelectedObjectsVisibility(f32 time)
{
    if (m_BlinkTargets.empty()) return;

    if (time >= m_BlinkSpan * (m_BlinkNum - 0.25)) {
        ResetBlinkAnimation();
        return;
    }

    if ( (s32)(time / (m_BlinkSpan / 2.f)) % 2 == 0)
    {
        nw::ut::MoveArray<BlinkAnimInfo>::iterator it = m_BlinkTargets.begin();
        nw::ut::MoveArray<BlinkAnimInfo>::iterator end = m_BlinkTargets.end();
        for ( ; it != end; ++it ) {
             it->mesh.SetVisible( !it->mesh.IsVisible() );
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SceneRenderer::SubmitView(
    nw::gfx::SceneContext* sceneContext,
    const nw::gfx::Camera& camera,
    nw::gfx::ISceneUpdater::IsVisibleModelFunctor* isVisibleModel
) {
    // Updates Fog.
    nw::gfx::FogArray::iterator fog = sceneContext->GetFogBegin();
    nw::gfx::FogArray::iterator fogEnd = sceneContext->GetFogEnd();
    for (; fog != fogEnd; ++fog) {
        if (*fog) (*fog)->Update(&camera);
    }
    m_pSceneUpdater->SubmitView(m_pRenderQueue, sceneContext, camera, 0, 0, isVisibleModel);
}


}
}
#if NN_CURRENT_VERSION_NUMBER >= NN_VERSION_NUMBER(4,0,0,0)
#pragma diag_default 2530 // padding added to end of struct.
#pragma diag_default 1301 // padding inserted in struct.
#endif
