/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../../system/include/HidManager.h"
#include "../include/SubMode.h"
#include "../include/MenuSubMode.h"
#include "../include/LogSubMode.h"

namespace {
bool
inRect(const nw::ut::Rect& rect, u16 x, u16 y)
{
    return (rect.left<= x && x <= rect.right &&
            rect.top <= y && y <= rect.bottom);
}
}

namespace vw  {
namespace g3d {


#define ADD_SUBMODE( type ) \
{ \
    void* buf = pAllocator->Alloc(sizeof(type##SubMode)); \
    type##SubMode* mode = new(buf) type##SubMode(); \
    mMode.push_back( mode ); \
} \

#define DELETE_SUBMODE( type ) \
{ \
    type##SubMode* mode = (type##SubMode*)mMode[i]; \
    mode->~type##SubMode(); \
    pAllocator->Free( mode ); \
} \

SubModeManager* SubModeManager::m_pInstance = NULL;

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SubModeManagerBase::Main(vw::system::HidStatus& hidStatus)
{
    // Get current pad status
    s32 modeSelect = 0;
    switch( system::GetPadEvent( hidStatus, system::GetSubModePadBindings() ) )
    {
    case system::SUB_SWITCH_PREV_SUBMODE:
        modeSelect = -1;
        break;
    case system::SUB_SWITCH_NEXT_SUBMODE:
        modeSelect = +1;
        break;
    default:
        break;
    }

    if( modeSelect != 0 )
    {
        mCurrentModeIndex += modeSelect;
        if(mCurrentModeIndex < 0) mCurrentModeIndex = mMode.size()-1;
        if(mCurrentModeIndex >= mMode.size()) mCurrentModeIndex = 0;
    }
    else
    {
        if(mCurrentModeIndex < mMode.size()) {
            mMode[mCurrentModeIndex]->Main(hidStatus);
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SubModeManagerBase::Draw()
{
    if (mCurrentModeIndex < mMode.size()) {
        mMode[mCurrentModeIndex]->Draw();
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SubModeManagerBase::SetMode(SubModeType mode)
{
    for (int i = 0; i < mMode.size(); ++i )
    {
        if (mMode[i]->GetModeType() == mode)
        {
            mCurrentModeIndex = i;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SubModeType
SubModeManagerBase::GetCurrentModeType() const
{
    SubModeType mode = SubMode_Camera;
    if(mCurrentModeIndex < mMode.size())
    {
        mode = mMode[mCurrentModeIndex]->GetModeType();
    }
    return mode;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
const char*
SubModeManagerBase::GetCurrentModeName() const
{
    if (mCurrentModeIndex < mMode.size()) {
        return mMode[mCurrentModeIndex]->GetModeName();
    } else {
        return "N/A";
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SubModeManagerBase::SubModeManagerBase() : mCurrentModeIndex(0)
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SubModeManagerBase::~SubModeManagerBase()
{
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SubModeManager*
SubModeManager::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SubModeManager::SubModeManager() : SubModeManagerBase()
{
     nw::os::IAllocator* pAllocator = Memory_GetAllocator();

     const SubModeType modes[] = {
        SubMode_Camera,
        SubMode_Menu,
        SubMode_Log
    };
    const int modeNum = sizeof( modes ) / sizeof( SubModeType );

    mMode = nw::ut::MoveArray<ISubMode*>( modeNum, pAllocator );
    for ( int i = 0; i < modeNum; ++i )
    {
        switch ( modes[i] )
        {
        case SubMode_Camera:
            ADD_SUBMODE( Camera )
            break;
        case SubMode_Menu:
            ADD_SUBMODE( Menu )
            break;
        case SubMode_Log:
            ADD_SUBMODE( Log );
            break;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
SubModeManager::~SubModeManager()
{
    nw::os::IAllocator* pAllocator = Memory_GetAllocator();

    for ( int i = 0; i < mMode.size(); ++i )
    {
        switch ( mMode[i]->GetModeType() )
        {
        case SubMode_Camera:
            DELETE_SUBMODE( Camera )
            break;
        case SubMode_Menu:
            DELETE_SUBMODE( Menu )
            break;
        case SubMode_Log:
            DELETE_SUBMODE( Log );
            break;
        }
    }
    mMode.clear();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SubModeManager::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(SubModeManager));
    if (buf == NULL) return;

    m_pInstance = new(buf) SubModeManager();
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SubModeManager::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~SubModeManager();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

/*---------------------------------------------------------------------------*
  Name:
  Description:
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
SubModeManager::GetTouchPanelEvent(const vw::system::HidStatus &hidStatus)
{
    // Turns off touch operation during lower screen preview.
    {
        vw::system::SystemSetting setting;
        vw::system::System_GetSetting( &setting );
        if ( setting.lowerPreview ) { return; }
    }

    if (hidStatus.isClicked)
    {
        nw::ut::Rect modeRect;
        modeRect.SetOriginAndSize( 2, 20, 68, 20 );
        u16 x = hidStatus.clickPosX;
        u16 y = hidStatus.clickPosY;

        if( inRect(modeRect, x, y) )
        {
            mCurrentModeIndex += 1;
            if(mCurrentModeIndex >= mMode.size()) mCurrentModeIndex = 0;
        }
    }

    // Get current touch panel status
    if ( hidStatus.touchPanel.touch && !mTouch )
    {
        mTouch = true;
    }
    else
    {
        mTouch = hidStatus.touchPanel.touch;
        return;
    }
}


}   /* namespace g3d */
}   /* namespace vw  */
