/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/ResourceManager.h"
#include "../include/UserDataManager.h"

namespace vw {
namespace g3d {

UserDataManager* UserDataManager::m_pInstance = NULL;

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
UserDataManager::UserDataManager() : m_userDataArray( Memory_GetAllocator() )
{
    m_pAllocator = Memory_GetAllocator();
    m_pDeviceAllocator = Memory_GetAllocatorDevice();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
UserDataManager::~UserDataManager()
{
    Clear();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
UserDataManager*
UserDataManager::GetInstance()
{
    return m_pInstance;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
UserDataManager::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_GetAllocator()->Alloc(sizeof(UserDataManager));
    if (buf == NULL) return;

    m_pInstance = new(buf) UserDataManager();
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
UserDataManager::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~UserDataManager();
    Memory_GetAllocator()->Free(m_pInstance);

    m_pInstance = NULL;
}

/*---------------------------------------------------------------------------*
  Name:

  Description:  adds dataLink to m_userDataArray if it's not empty

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
UserDataManager::Link(
    nw::ut::MoveArray<LinkData>& targets,
    nw::gfx::res::ResSceneObject replace,
    unsigned int hashID
)
{
    // Data check
    if ( m_pInstance == NULL ) return;
    if ( targets.size() == 0 ) return;
    if ( !replace.IsValid() ) return;
    if( replace.ref().m_UserDataDicCount == 0 ) return;

    // Gets iterator.
    nw::ut::MoveArray<LinkData>::iterator it  = targets.begin();
    nw::ut::MoveArray<LinkData>::iterator end = targets.end();

    // Creates a data storage object.
    UserDataLink dataLink;
    dataLink.m_targets =
        nw::ut::MoveArray<UserDataTarget>( m_pAllocator );
    dataLink.m_replace = replace;
    dataLink.m_hashID = hashID;

    // Iterates.
    for ( ; it != end; ++it )
    {
        // Determines if data is valid.
        if ( it->m_target == NULL || it->m_resourceSet == NULL ) continue;

        // Cancels the user data that is currently configured.
        Unlink( it->m_target );

        // Saves the original user data.
        UserDataTarget target;
        target.m_target = it->m_target;
        target.m_targetType = it->m_targetType;
        target.m_resourceSet = it->m_resourceSet;

        // When the target is the object derived from the scene object.
        if ( it->m_targetType == TARGET_SCENE_OBJECT )
        {
            // Gets the scene object.
            nw::gfx::res::ResSceneObject obj( it->m_target );

            // Saves the original user data.
            target.m_originalDictCount = obj.GetUserDataCount();
            target.m_originalDict = obj.ref().toUserDataDic.to_ptr();

            // Replaces the reference location of the user data.
            obj.ref().m_UserDataDicCount = replace.ref().m_UserDataDicCount;
            obj.ref().toUserDataDic.set_ptr( replace.ref().toUserDataDic.to_ptr() );
        }
        // When the target is a bone object
        else if ( it->m_targetType == TARGET_BONE )
        {
            // Gets the animations.
            nw::gfx::res::ResBone bone( it->m_target );

            // Saves the original user data.
            target.m_originalDictCount = bone.GetUserDataCount();
            target.m_originalDict = bone.ref().toUserDataDic.to_ptr();

            // Replaces the reference location of the user data.
            bone.ref().m_UserDataDicCount = replace.ref().m_UserDataDicCount;
            bone.ref().toUserDataDic.set_ptr( replace.ref().toUserDataDic.to_ptr() );
        }
        // When the target is the object derived from the scene object.
        else if ( it->m_targetType == TARGET_ANIMATION )
        {
            // Gets the animations.
            nw::anim::res::ResAnim anim( it->m_target );

            // Saves the original user data.
            target.m_originalDictCount = anim.GetUserDataCount();
            target.m_originalDict = anim.ref().toUserDataDic.to_ptr();

            // Replaces the reference location of the user data.
            anim.ref().m_UserDataDicCount = replace.ref().m_UserDataDicCount;
            anim.ref().toUserDataDic.set_ptr( replace.ref().toUserDataDic.to_ptr() );
        }
        else
        {
            continue;
        }

        // Adds to the array.
        dataLink.m_targets.push_back( target );
    }

    // Registers data if the size is not zero.
    if ( dataLink.m_targets.size() != 0 )
    {
        m_userDataArray.push_back( dataLink );
    }
}



namespace
{
    // If defined as a local class, compile will not accepted as an argument of the template function.
    struct ResourceSetChecker : public std::unary_function<UserDataLink::UserDataTargetList::iterator, bool>
    {
        ResourceSetChecker(ResourceSet* resourceSet) : m_resourceSet(resourceSet){}
        result_type operator()(argument_type it) { return (it->m_resourceSet==m_resourceSet); }
        ResourceSet* m_resourceSet;
    };

    // If defined as a local class, compile will not accepted as an argument of the template function.
    struct TargetChecker : public std::unary_function<UserDataLink::UserDataTargetList::iterator, bool>
    {
        TargetChecker(void* target) : m_target(target){}
        result_type operator()(argument_type it) { return (it->m_target==m_target); }
        void* m_target;
    };
}


/*---------------------------------------------------------------------------*
  Name:         Unlink

  Description:  Unlink object

  Arguments:    object

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
UserDataManager::Unlink( void* object )
{
    Unlink(TargetChecker(object));
}


/*---------------------------------------------------------------------------*
  Name:         Unlink

  Description:  Unlink ResourceSet

  Arguments:    ResourceSet

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
UserDataManager::Unlink( ResourceSet* resourceSet )
{
    Unlink(ResourceSetChecker(resourceSet));
}

/*---------------------------------------------------------------------------*
  Name:         Unlink

  Description:  Factorises code from Unlink( void* object ) and Unlink( ResourceSet* resourceSet ), using template argument checkFunction to validate iterator

  Arguments:    CheckFunction checkFunction : must derive from std::unary_function<UserDataLink::UserDataTargetList::iterator, bool>

  Returns:      None.
 *---------------------------------------------------------------------------*/
template<class CheckFunction>
void
UserDataManager::Unlink(CheckFunction checkFunction)
{

    // Data check
    if ( m_pInstance == NULL ) return;

    // Gets iterator.
    UserDataLinkList::iterator it  = m_userDataArray.begin();

    // Iterates.
    for ( ; it != m_userDataArray.end(); )
    {
        // Gets iterator.
        nw::ut::MoveArray<UserDataTarget>::iterator it_  = it->m_targets.begin();

        // Iterates.
        for ( ; it_ != it->m_targets.end(); )
        {
            // Determines if the deletion target and rewrite source object are the same.
            if ( checkFunction(it_))
            {
                // When the target is the object derived from the scene object.
                if ( it_->m_targetType == TARGET_SCENE_OBJECT )
                {
                    // Gets the scene object.
                    nw::gfx::res::ResSceneObject obj( it_->m_target );

                    // Returns to the source data.
                    obj.ref().m_UserDataDicCount = it_->m_originalDictCount;
                    obj.ref().toUserDataDic.set_ptr( it_->m_originalDict );
                }
                // When the target is a bone object
                else if ( it_->m_targetType == TARGET_BONE )
                {
                    // Gets the animations.
                    nw::gfx::res::ResBone bone( it_->m_target );

                    // Replaces the reference location of the user data.
                    bone.ref().m_UserDataDicCount = it_->m_originalDictCount;
                    bone.ref().toUserDataDic.set_ptr( it_->m_originalDict );
                }
                // When the target is the object derived from the scene object.
                else if ( it_->m_targetType == TARGET_ANIMATION )
                {
                    // Gets the animations.
                    nw::anim::res::ResAnim anim( it_->m_target );

                    // Returns to the source data.
                    anim.ref().m_UserDataDicCount = it_->m_originalDictCount;
                    anim.ref().toUserDataDic.set_ptr( it_->m_originalDict );
                }

                // Deletes from the rewrite source object.
                nw::ut::MoveArray<UserDataTarget>::iterator it_next = it_;
                ++it_next;
                it_ = it->m_targets.erase( it_, it_next );
            } else {
                ++it_;
            }
        }

        // When the size of the rewrite source object list is zero.
        if ( it->m_targets.size() == 0 )
        {
            // Releases the rewrite file.
            g3d::ResourceManager::GetInstance()->DeleteResource(
                g3d::ResourceSet::Type_UserData,
                it->m_hashID
            );

            // Deletes the rewrite source object list.
            UserDataLinkList::iterator it_next = it;
            ++it_next;
            it = m_userDataArray.erase( it, it_next );
        } else {
            ++it;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:

  Description:

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
UserDataManager::Clear()
{
    // Data check
    if ( m_pInstance == NULL ) return;

    // Gets iterator.
    UserDataLinkList::iterator it  = m_userDataArray.begin();
    UserDataLinkList::iterator end = m_userDataArray.end();

    // Iterates.
    for ( ; it != end; ++it )
    {
        // Releases the rewrite file.
        g3d::ResourceManager::GetInstance()->DeleteResource(
            g3d::ResourceSet::Type_UserData,
            it->m_hashID
        );
    }

    m_userDataArray.clear();
}

#if 0
inline u32
GetBit(const char* str, size_t len, u32 refBit)
{
    u32 wd = (refBit >> 3);
    u32 pos = (refBit & 7);
    // If the reference bit is outside of the name, 0
    return wd < len ? (str[wd] >> pos) & 0x1 : 0;
}

inline u32
GetBit(const nw::ut::ResName* name, u32 refBit)
{
    return GetBit(name->GetName(), name->GetLength(), refBit);
}

inline const s32
max(const s32& a, const s32& b)
{
    return a < b ? b : a;
}

bool ResDicPatriciaBuild( nw::ut::ResDicPatricia dic )
{
    // Rebuilds the dictionary from the state in which the numData and each node's ofsName are set accurately.
    // The size and each node's ofsData will not change.

    if (dic.ptr() == NULL)
    {
        return true; // If empty, nothing happens.
    }

    nw::ut::ResDicPatriciaData::ResDicNodeData* node = dic.ref().data;
    nw::ut::ResDicPatriciaData::ResDicNodeData& root = node[0];
    root.idxLeft = root.idxRight = 0;
    root.ref = ~0u;

    // Temporarily sets ResName to length 0 so the length can be obtained.
    // Borrows the ResDicPatriciaData::size area, not the stack, so that 64-bit version can also be used.
    nw::ut::ResNameData* empty = reinterpret_cast<nw::ut::ResNameData*>(&dic.ref().length);
    nw::ut::ResU32 len = empty->len;
    empty->len = 0;
    root.ofsString.set_ptr(empty->str);

    for (int index = 1, end = dic.ref().numData; index <= end; ++index)
    {
        nw::ut::ResDicPatriciaData::ResDicNodeData& current = node[index];
        const nw::ut::ResName name = dic.GetResName( index );
        u32 refBit = 0;
        nw::ut::ResName childName;

        // Goes all the way back and gets the target bit comparison character string.
        {
            const nw::ut::ResDicPatriciaData::ResDicNodeData* parent = &root;
            const nw::ut::ResDicPatriciaData::ResDicNodeData* child = &node[parent->idxLeft];
            childName = dic.GetResName(parent->idxLeft);

            while (parent->ref > child->ref)
            {
                // If from the child to the location of refBit is 1, the loop follows the right link. If it is 0, it follows the left link.
                // Upon return to the above node, the loop ends.
                parent = child;
                u32 bit = GetBit(&name, child->ref);
                child = bit ? &node[child->idxRight] : &node[child->idxLeft];
                childName = bit ? dic.GetResName(parent->idxRight) : dic.GetResName(parent->idxLeft);
            }

            refBit = static_cast<u32>(max(name.GetLength(), childName.GetLength())) * CHAR_BIT;
            while (GetBit(&childName, refBit) == GetBit(&name, refBit))
            {
                // Searches for a bit different from the bit at the node entry at the location returned to.
                // Makes a determination based on the bit location of refBit.
                if (refBit == 0)
                {
                    return false; // If the names are duplicated, it is a failure.
                }
                --refBit;
            }
            current.ref = refBit;
        }

        // Insert
        {
            nw::ut::ResDicPatriciaData::ResDicNodeData* parent = &root;
            nw::ut::ResDicPatriciaData::ResDicNodeData* child = &node[parent->idxLeft];

            while (parent->ref > child->ref && child->ref > refBit)
            {
                parent = child;
                child = GetBit(&name, child->ref) ? &node[child->idxRight] : &node[child->idxLeft];
            }
            // At this point, it is confirmed that a node is inserted between parent and child.

            u16 idxCurrent = static_cast<u16>(index);
            u16 idxChild = static_cast<u16>(std::distance(node, child));

            if (GetBit(&name, current.ref))
            {
                // If the relevant bit is 1, it attaches t(itself) to the right (current -> (current, child)).
                current.idxLeft = idxChild;
                current.idxRight = idxCurrent;
            }
            else
            {
                // If the relevant bit is 0, it attaches t(itself) to the left (current -> (child, current)).
                current.idxLeft = idxCurrent;
                current.idxRight = idxChild;
            }

            // Determines which of the parent's children it will be.
            if (GetBit(&name, parent->ref))
            {
                parent->idxRight = idxCurrent;
            }
            else
            {
                parent->idxLeft  = idxCurrent;
            }
        }
    }

    root.ofsString.set_ptr(NULL);
    empty->len = len;

    return true;
}
#endif

}
}
