/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#ifndef NW4C_VIEWER_SYSTEM_TGA_DATA_H_
#define NW4C_VIEWER_SYSTEM_TGA_DATA_H_

#include "ImageFmt.h"

namespace vw     {
namespace system {


class TgaData : public ImageFmt
{
//--- Public members --------------
public:
    typedef enum
    {
        FMT_NONE      = 0x00, // No image
        FMT_CI8       = 0x01, // Color index 256 colors
        FMT_RGB       = 0x02, // Full color
        FMT_I8        = 0x03, // Gray
        FMT_CMP_CI8   = 0x09, // Compressed color index 256 colors
        FMT_CMP_RGB24 = 0x0A, // Compressed full color
        FMT_CMP_I8    = 0x0B  // Compressed gray
    } Format;

    typedef struct
    {
        u8  bfIdLength;      // ID field length
        u8  bfType;          // Existence of color map
        u8  bfFormat;        // Image format
        u8  bfPlttIdx[2];    // Color map index
        u8  bfPlttLength[2]; // Color map length
        u8  bfPlttSize;      // Color map size
        u8  bfOriginX[2];    // X-coordinate
        u8  bfOriginY[2];    // Y-coordinate
        u8  bfWidth[2];      // Image width
        u8  bfHeight[2];     // Image height
        u8  bfBpp;           // bit per pixel
        union
        {
            struct
            {
                u8 padding_   :2;   // 0x00
                u8 bfDirY     :1;   // Store direction (0 = bottom to top; 1 = top to bottom)
                u8 bfDirX     :1;   // Store direction (0 = left to right; 1 = right to left)
                u8 bfAttribute:4;   // Attribute.
            };
            u8 bfDescript;
        };
    } Header;

    typedef struct
    {
        u8      bfOffset[4];     // File position
        u8      bfDevOffset[4];  // Developer directory file position
        u8      bfSignature[17]; // "TRUEVISION-TARGA" character
        u8      bfPadding_;      // 0x00
    } Footer;

    /* ctor */          TgaData( u32 width, u32 height, u8 bpp = 3 );
    /* dtor */ virtual  ~TgaData( void );

    virtual u32         GetBodySize() const   { return mBodySize; }

    virtual u8*         GetHeader()           { return reinterpret_cast<u8*>( &mHeader ); }
    virtual u32         GetHeaderSize() const { return sizeof(Header); }

    virtual void        ConvertData( void );    // Convert RGBA data in buffer to match format.
    virtual void        ConvertData( void* src_image, u32 width, u32 height, u32 dst_ofs_x, u32 dst_ofs_y );
    virtual void        ConvertData( void* src_image,
                                     u32 src_width, u32 src_height,
                                     u32 dst_ofs_x, u32 dst_ofs_y,
                                     u32 src_ofs_x, u32 src_ofs_y,
                                     u32 width,     u32 height );
    virtual void        ConvertAlphaData( void );    // Convert RGBA data in buffer to match format.
    virtual void        ConvertAlphaData( void* src_image, u32 width, u32 height, u32 dst_ofs_x, u32 dst_ofs_y, Operation op );
    virtual void        ConvertAlphaData( void* src_image, // Converts RGBA data in buffer to match format, and applies alpha only.
                                     u32 src_width, u32 src_height,
                                     u32 dst_ofs_x, u32 dst_ofs_y,
                                     u32 src_ofs_x, u32 src_ofs_y,
                                     u32 width,     u32 height,
                                     Operation op = OPE_SETVALUE );
    virtual void        Finalize( void );

//--- Private members--------------
private:
    void        ConvertDataImpl_( void );    // Convert RGBA data in buffer to match format.
    void        ConvertDataImpl_( void* src_image,
                                  u32 src_width, u32 src_height,
                                  u32 dst_ofs_x, u32 dst_ofs_y,
                                  u32 src_ofs_x, u32 src_ofs_y,
                                  u32 width, u32 height );
    void        ConvertAlphaDataImpl_( void );    // Convert RGBA data in buffer to match format.
    void        ConvertAlphaDataImpl_( void* src_image,
                                  u32 src_width, u32 src_height,
                                  u32 dst_ofs_x, u32 dst_ofs_y,
                                  u32 src_ofs_x, u32 src_ofs_y,
                                  u32 width, u32 height,
                                  Operation op = OPE_SETVALUE );

    void        CompressDataImpl_( void );      // Data RLE compression
    void        CompressDataImplThin_( void );  // Data RLE compression (memory conservation version)
    u32         GetCompressedDataSize_( void ); // Calculate size when data is compressed.

    u32 mBodySize;  // Data size

#define  TGA_SIGNATURE      "TRUEVISION-TARGA"

    Header mHeader;
    Footer mFooter;
};


}   /* namespace system */
}   /* namespace vw     */



/* NW4C_VIEWER_SYSTEM_TGA_DATA_H_ */
#endif
