/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/BitmapData.h"
#include "../include/TexUtil.h"

namespace vw     {
namespace system {


/*---------------------------------------------------------------------------*
  Name:         BitmapData::BitmapData

  Description:  Constructor

  Arguments:    width
                height
                bpp

  Returns:      None.
 *---------------------------------------------------------------------------*/
/* ctor */
BitmapData::BitmapData( u32 width, u32 height, u8 bpp /* = 3 */ )
 : ImageFmt( width, height, bpp )
{
    u32 size = nn::math::RoundUp(mWidth, 4) * nn::math::RoundUp(mHeight, 4) * sizeof(u32);

    mBody = Memory_AllocSys( size, 128 );

    mHeader.bfType[0] = 'B';
    mHeader.bfType[1] = 'M';
    LEndian::WriteU32( &mHeader.bfSize,     size );
    mHeader.bfReserved1 = 0;
    mHeader.bfReserved2 = 0;
    LEndian::WriteU32( &mHeader.bfOffBits,  sizeof( Header ) );
    LEndian::WriteU32( &mHeader.bcSize,     40 );
    LEndian::WriteU32( &mHeader.bcWidth,    mWidth );
    LEndian::WriteU32( &mHeader.bcHeight,   mHeight );
    LEndian::WriteU16( &mHeader.bcPlanes,   1 );
    LEndian::WriteU16( &mHeader.bcBitCount, static_cast<u16>( mBpp * 8 ) );
    LEndian::WriteU32( &mHeader.compMode,   0 ); // No compression
    LEndian::WriteU32( &mHeader.dataSize,   mWidth * mHeight * mBpp );
    LEndian::WriteU32( &mHeader.hPpm,       0 );
    LEndian::WriteU32( &mHeader.vPpm,       0 );
    LEndian::WriteU32( &mHeader.pltNum,     0 );
    LEndian::WriteU32( &mHeader.impPltNum,  0 );
}

/*---------------------------------------------------------------------------*
  Name:         BitmapData::~BitmapData

  Description:  Destructor

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
/* dtor */
BitmapData::~BitmapData( void )
{
}


/*---------------------------------------------------------------------------*
  Name:         BitmapData::ConvertData

  Description:  Converts RGBA data to match the file format.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
BitmapData::ConvertData( void )
{
    const u32 SRC_BPP = 4;

    u8* pTmp   = static_cast<u8*>( Memory_AllocSys( system::SYSTEM_ID, nn::math::RoundUp(mWidth, 4) * nn::math::RoundUp(mHeight, 4) * SRC_BPP ) );
    std::memcpy( pTmp, mBody, mWidth * mHeight * SRC_BPP );

    this->ConvertData( pTmp, mWidth, mHeight, 0, 0 );

    Memory_Free( pTmp );
}



/*---------------------------------------------------------------------------*
  Name:         BitmapData::ConvertData

  Description:  Converts RGB8 texture to BMP data.

  Arguments:    src_image   The pointer to the texture data
                src_width   Texture horizontal size
                src_height  Texture vertical size
                dst_ofs_x   X position in the overall BMP image
                dst_ofs_y   Y position in the overall BMP image
                src_ofs_x   X-offset of the rectangle used in the texture
                src_ofs_y   Y-offset of the rectangle used in the texture
                width       Horizontal size of the rectangle used in the texture
                height      Vertical size of the rectangle used in the texture

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
BitmapData::ConvertData( void* src_image,
                         u32 src_width, u32 src_height,
                         u32 dst_ofs_x, u32 dst_ofs_y,
                         u32 src_ofs_x, u32 src_ofs_y,
                         u32 width,     u32 height )
{
    u8* pRead  = reinterpret_cast<u8*>( src_image );
    u8* pWrite = reinterpret_cast<u8*>( mBody );
    u32 i;

    u32 line = 0;

    for ( line = 0; line < height; line++ )
    {
        u32 y = src_height - line - 1 - (src_height - height - src_ofs_y);
        u32 dst_addr = ( (dst_ofs_y + line) * mWidth + dst_ofs_x ) * mBpp;
        pWrite = reinterpret_cast<u8*>( mBody ) + dst_addr;

        for ( i = 0; i < width; i ++ )
        {
            u32 x = i + src_ofs_x;
            nw::ut::Color8 col = UtilRGBA8::GetColor( pRead, x, y, src_width );

            switch ( mBpp )
            {
            //---------------
            // Output 24-bit color BMP
            case 3:
                {
                    *pWrite++   = col.b;
                    *pWrite++   = col.g;
                    *pWrite++   = col.r;
                }
                break;
            //---------------
            // Output 32-bit color BMP
            case 4:
                {
                    *pWrite++   = col.b;
                    *pWrite++   = col.g;
                    *pWrite++   = col.r;
                    *pWrite++   = col.a;
                }
            }
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         BitmapData::ConvertData

  Description:  Converts RGB8 texture to BMP data.

  Arguments:    src_image   The pointer to the texture data
                width       Texture horizontal size
                height      Texture vertical size
                dst_ofs_x   X position in the overall BMP image
                dst_ofs_y   Y position in the overall BMP image

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
BitmapData::ConvertData( void* src_image, u32 width, u32 height, u32 dst_ofs_x, u32 dst_ofs_y )
{
    this->ConvertData( src_image, width, height, dst_ofs_x, dst_ofs_y, 0, 0, width, height );
}


}   /* namespace system */
}   /* namespace vw     */


