/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/


#include "../include/ClearImage.h"
#include "../include/Imagereader.h"


namespace vw     {
namespace system {


/* static */ ClearImage* ClearImage::sInstance = NULL;
/* static */ u32         ClearImage::sPosition = ClearImage::POS_STRETCH;
/* static */ char        ClearImage::sLatestPath[ FILE_PATH_MAX ] = "";




ClearImage& ClearImage::GetInstance()
{
    NW_ASSERT( sInstance != NULL );
    return *sInstance;
}

void ClearImage::Initialize()
{
    if (sInstance != NULL) return;
    void *buf = Memory_AllocSys(sizeof(ClearImage));
    if (buf == NULL) return;

    sInstance = new(buf) ClearImage();
}

void ClearImage::Finalize()
{
    if (sInstance == NULL) return;

    sInstance->~ClearImage();
    Memory_Free(sInstance);

    sInstance = NULL;
}



/*---------------------------------------------------------------------------*
  Name:         ClearImage::ClearImage
  Description:  Constructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
ClearImage::ClearImage()
 : m_TexImg( NULL ),
   mImgWidth( 0 ),
   mImgHeight( 0 ),
   mTexWidth( 0 ),
   mTexHeight( 0 ),
   m_Texture( 0 )
{

}


/*---------------------------------------------------------------------------*
  Name:         ClearImage::~ClearImage
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
ClearImage::~ClearImage()
{
    this->Destroy();
}


/*---------------------------------------------------------------------------*
  Name:         ClearImage::Destroy
  Description:  Destroys loaded image data.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ClearImage::Destroy()
{
    if ( m_TexImg )
    {
        Memory_Free( m_TexImg );
        m_TexImg = NULL;
    }
    if ( m_Texture )
    {
        m_Texture->~Texture();
        Memory_Free( m_Texture );
        m_Texture = NULL;
    }
}


/*---------------------------------------------------------------------------*
  Name:         ClearImage::ReadImg
  Description:  Reads image
  Arguments:    reader
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ClearImage::ReadImg( void* data, u32 dataSize /* = 0 */ )
{
    // Destroys current image data
    this->Destroy();

    u8* fileBuf = static_cast<u8*>( data );

    // Determines data type
    ImageReader imgReader( data, dataSize );

    if ( ! imgReader.IsValid() )
    {
        return;
    }

    mImgWidth  = u16( imgReader.GetWidth()  );
    mImgHeight = u16( imgReader.GetHeight() );

    mTexWidth  = u16( 1 << nn::math::ILog2( mImgWidth )  );
    mTexHeight = u16( 1 << nn::math::ILog2( mImgHeight ) );
    if ( mTexWidth < mImgWidth )
    {
        mTexWidth <<= 1;
    }
    if ( mTexHeight < mImgHeight )
    {
        mTexHeight <<= 1;
    }

    m_TexImg = static_cast<u8*>( Memory_AllocSys( mTexWidth * mTexHeight * sizeof(u32) ) );

    imgReader.Expand( m_TexImg, mTexWidth, mTexHeight );
}


/*---------------------------------------------------------------------------*
  Name:         ClearImage::Draw
  Description:  Renders the background image.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ClearImage::Draw(u32 width, u32 height)
{
    if ( ! this->IsValid() )
    {
        return;
    }

    if( this->m_TexImg && this->m_Texture == 0 )
    {
        void* buf = Memory_AllocSys(sizeof(vw::system::Texture));
        m_Texture = new(buf) vw::system::Texture(m_TexImg, mTexWidth, mTexHeight);

        // Combiner settings
        nn::gr::Combiner combiner;
        combiner.stage[0].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        combiner.stage[0].rgb.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0;
        combiner.stage[0].rgb.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
        combiner.stage[0].alpha.combine = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
        combiner.stage[0].alpha.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_TEXTURE0;
        combiner.stage[0].alpha.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
        combiner.stage[0].alpha.operand[0] = PICA_DATA_OPE_RGB_SRC_ALPHA;
        m_Texture->SetCombiner(combiner);
    }

    this->DrawImage(width, height);
}

/*---------------------------------------------------------------------------*
  Name:         ClearImage::DrawImage
  Description:  Renders the background image.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
ClearImage::DrawImage(u32 CANVAS_WIDTH, u32 CANVAS_HEIGHT)
{
    vw::system::VideoSetting setting;
    vw::system::Video_GetSetting( &setting );
    const nw::ut::FloatColor color = setting.clearColor;

    vw::system::Draw2DHelper& draw2D = vw::system::Draw2DHelper::GetInstance();
    draw2D.SetScreenSize(CANVAS_WIDTH, CANVAS_HEIGHT);
    draw2D.BeginDraw();
    draw2D.Disable( vw::system::Draw2DHelper::ALPHABLEND |
                    vw::system::Draw2DHelper::CULLFACE   |
                    vw::system::Draw2DHelper::DEPTHWRITE |
                    vw::system::Draw2DHelper::DEPTHTEST );
    m_Texture->SetFilter(PICA_DATA_TEXTURE_MAG_FILTER_LINEAR, PICA_DATA_TEXTURE_MIN_FILTER_LINEAR);

    if ( sPosition == POS_STRETCH )
    {
        m_Texture->SetWrapMode(PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE, PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE);
        m_Texture->SetState();
        draw2D.SetUVScale(mImgWidth / f32(mTexWidth), mImgHeight / f32(mTexHeight));
        draw2D.FillRectangle( 0, CANVAS_HEIGHT, CANVAS_WIDTH, -CANVAS_HEIGHT, color );
    }
    else if ( sPosition == POS_CENTER )
    {
        s32 x_org = (s32(CANVAS_WIDTH)  - mImgWidth)  / 2;
        s32 y_org = (s32(CANVAS_HEIGHT) - mImgHeight) / 2;

        draw2D.FillRectangle( 0, 0, CANVAS_WIDTH, CANVAS_HEIGHT, color );
        draw2D.Flush();
        m_Texture->SetWrapMode(PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE, PICA_DATA_TEXTURE_WRAP_CLAMP_TO_EDGE);
        m_Texture->SetState();
        draw2D.SetUVScale(mImgWidth / f32(mTexWidth), mImgHeight / f32(mTexHeight));
        draw2D.FillRectangle( x_org, CANVAS_HEIGHT-y_org, mImgWidth, -mImgHeight, color );
    }
    else if ( sPosition == POS_TILE )
    {
        m_Texture->SetWrapMode(PICA_DATA_TEXTURE_WRAP_REPEAT, PICA_DATA_TEXTURE_WRAP_REPEAT);
        m_Texture->SetState();

        s32 x_num = (CANVAS_WIDTH / mImgWidth);
        s32 y_num = (CANVAS_HEIGHT / mImgHeight);
        s32 imgW  = mImgWidth;
        s32 imgH  = mImgHeight;
        if ( (CANVAS_WIDTH % mImgWidth) != 0 )
        {
            x_num += 1;
        }

        if ( (CANVAS_HEIGHT % mImgHeight) != 0 )
        {
            y_num += 1;
        }

        if ( mImgWidth == mTexWidth )
        {
            // Can repeat in X-direction
            imgW = imgW * x_num;
            x_num = 1;
        }

        if ( mImgHeight == mTexHeight )
        {
            // Can repeat in Y-direction
            imgH = imgH * y_num;
            y_num = 1;
        }

        draw2D.SetUVScale(imgW / f32(mTexWidth), imgH / f32(mTexHeight));
        for ( s32 x = 0; x < x_num; x++ )
        {
            for ( s32 y = 0; y < y_num; y++ )
            {
                draw2D.FillRectangle( imgW * x, imgH * (y + 1), imgW, -imgH, color );
            }
        }
    }
    draw2D.EndDraw();
}

/*---------------------------------------------------------------------------*
  Name:         ClearImage::SetLatestPath
  Description:  Records the path of the file last loaded.
  Arguments:    path        File path
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* static */ void
ClearImage::SetLatestPath( const char* path )
{
    if ( ! path )
    {
        sLatestPath[ 0 ] = '\0';
        return;
    }
    vw::common::strlcpy( sLatestPath, path, FILE_PATH_MAX );
}

/*---------------------------------------------------------------------------*
  Name:         ClearImage::GetLatestPath
  Description:  Gets the path of the background file last loaded.
  Arguments:    None.
  Returns:      Background file path
 *---------------------------------------------------------------------------


*/
/* static */ const char*
ClearImage::GetLatestPath()
{
    if ( sLatestPath[ 0 ] == '\0' )
    {
        return NULL;
    }
    return sLatestPath;
}



}   /* namespace system */
}   /* namespace vw     */


