/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <cstring>
#include <cstdlib>
#include <viewer/system.h>
#include <viewer/common.h>
#include "../include/ClearImage.h"



namespace vw     {
namespace system {

static void ConfigSys_ExecuteVideo_();
static void ConfigSys_ExecuteMisc_();
static void ConfigSys_WriteBackVideo_();
static void ConfigSys_WriteBackMisc_();


/*---------------------------------------------------------------------------*
  Name:         ConfigSys_Execute

  Description:  Executes the initial configuration file

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ConfigSys_Execute( void )
{
    if ( GetConfigFile() == NULL )
    {
        return;
    }

    // Video-related settings
    ConfigSys_ExecuteVideo_();

    // Other system settings
    ConfigSys_ExecuteMisc_();

}


/*---------------------------------------------------------------------------*
  Name:         ConfigSys_WriteBack

  Description:  Writes back system-related settings to the Config class instance.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
ConfigSys_WriteBack( void )
{
    // Video settings
    ConfigSys_WriteBackVideo_();
    // Other system settings
    ConfigSys_WriteBackMisc_();
}


/*---------------------------------------------------------------------------*
  Name:         ConfigSys_ExecuteVideo_

  Description:  Video-related initial settings

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ConfigSys_ExecuteVideo_()
{
    s32 value;
    u32 uValue;
    f32 fValue;

    VideoSetting setting;
    Video_GetSetting( &setting );

#if 0
    // Screen size settings
    if ( GetConfigFile()->GetInt("/system/video/fb_width", &value) )
    {
        if ( value > 0 )
        {
            setting.efbWidth = VIPadFrameBufferWidth ( static_cast<u16>( ut::RoundUp( value, 16 ) ) );
        }
    }
    if ( GetConfigFile()->GetInt("/system/video/fb_height", &value) )
    {
        if ( value > 0 )
        {
            setting.efbHeight = static_cast<u16>( ut::RoundUp( value, 2 ) );
        }
    }
    if ( GetConfigFile()->GetInt("/system/video/vi_width", &value) )
    {
        if ( value > 0 )
        {
            setting.viWidth = ut::Max( static_cast<u16>( value ), setting.efbWidth );
        }
    }
    if ( GetConfigFile()->GetInt("/system/video/vi_height", &value) )
    {
        if ( value > 0 )
        {
            setting.viHeight = ut::Max( static_cast<u16>( value ), setting.efbHeight );
            setting.viHeight = ut::RoundUp( setting.viHeight, 2 );
        }
    }

    if ( GetConfigFile()->GetInt("/g3d/projection/wide_screen", &value ) )
    {
        setting.wideScreen = (value != 0)? true : false;
    }
#endif

    // Clear color setting
    if ( GetConfigFile()->GetUInt("/system/video/clr_color", &uValue ) )
    {
        setting.clearColor = nw::ut::Color8( uValue );
    }

    // Clear Z value settings
    if ( GetConfigFile()->GetFloat("/system/video/clr_z", &fValue ) )
    {
        if (fValue < 0.0f)
        {
            fValue = 0.0f;
        }
        if (fValue > 1.0f)
        {
            fValue = 1.0f;
        }
        setting.clearZ = fValue;
    }

    // Buffer format
    {
        // Color format.
        const char* fmt = GetConfigFile()->GetString("/system/video/color_fmt");
        if ( fmt )
        {
            if ( std::strncmp( fmt, "RGBA8", 5 ) == 0 )
            {
                setting.colorFormat = nw::gfx::RENDER_COLOR_FORMAT_RGBA8;
            }
            else if ( std::strncmp( fmt, "RGBA4", 5 ) == 0 )
            {
                setting.colorFormat = nw::gfx::RENDER_COLOR_FORMAT_RGBA4;
            }
            else if ( std::strncmp( fmt, "RGB5_A1", 7 ) == 0 )
            {
                setting.colorFormat = nw::gfx::RENDER_COLOR_FORMAT_RGB5_A1;
            }
            else if ( std::strncmp( fmt, "RGB565", 6 ) == 0 )
            {
                setting.colorFormat = nw::gfx::RENDER_COLOR_FORMAT_RGB565;
            }
        }

        // Depth format.
        fmt = GetConfigFile()->GetString("/system/video/depth_fmt");
        if ( fmt )
        {
            if ( std::strncmp( fmt, "D16", 3 ) == 0 )
            {
                setting.depthFormat = nw::gfx::RENDER_DEPTH_FORMAT_16;
            }
            else if ( std::strncmp( fmt, "D24_S8", 6 ) == 0 )
            {
                setting.depthFormat = nw::gfx::RENDER_DEPTH_FORMAT_24_STENCIL8;
            }
            else if ( std::strncmp( fmt, "D24", 3 ) == 0 )
            {
                setting.depthFormat = nw::gfx::RENDER_DEPTH_FORMAT_24;
            }
        }
    }

    Video_ChangeSetting( &setting );
}

/*---------------------------------------------------------------------------*
  Name:         ConfigSys_ExecuteMisc_

  Description:  Other system settings

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ConfigSys_ExecuteMisc_()
{
    s32 value;
    u32 uValue;
    s32 iValue;

    SystemSetting setting;
    System_GetSetting( &setting );

    // Initial load model path setting
    const char* default_model = GetConfigFile()->GetString("/system/config/default_model");
    setting.defaultModelPath = default_model;

    // File load path settings
    const char* reload_target = GetConfigFile()->GetString("/system/config/reload_target");
    setting.reloadTargetPath = reload_target;

    // Screen capture save path settings
    const char* capture_image = GetConfigFile()->GetString("/system/config/capture_image");
    setting.captureImagePath = capture_image;

    // Animation frame rate
    if ( GetConfigFile()->GetInt("/system/config/frame_rate", &value) )
    {
        if ( value > 0 && value <= 60 )
        {
            setting.updateRate =  static_cast<u32>( 60 / value );
            setting.frameRate =  static_cast<u32>( 60 / setting.updateRate );

            // Set measurement count
            PerformMgr::GetInstance()->SetMeasureCount(setting.frameRate);
        }
    }

    // Frame skip process
    if ( GetConfigFile()->GetInt("/system/config/frame_skip", &iValue ) )
    {
        setting.frameSkip = ( iValue != 0 );
    }




#if 0
        if ( cmode )
        {
            switch ( BEndian::ReadU16( cmode ) )
            {
            case '3D':
                App::GetInstance().SetFocusID( G3D_ID );
                break;
            case 'EF':
                App::GetInstance().SetFocusID( EFFECT_ID );
                break;
            case 'LY':
                App::GetInstance().SetFocusID( G2D_ID );
                break;
            case 'SN':
                App::GetInstance().SetFocusID( SOUND_ID );
            }
        }
    }

    // Processing meter measure
    if ( GetConfigFile()->GetUInt("/system/config/perform_meter_length", &uValue) )
    {
        if ( uValue > 5  ) { uValue = 5; }
        PerformDrawer::SetMaxFrame( uValue );
    }

    // Controller port setting
    if ( GetConfigFile()->GetUInt("/system/config/pad_port", &uValue) )
    {
        if ( uValue < PAD_MAX_CONTROLLERS )
        {
            Pad_SetPort( uValue );
        }
    }

    // Controller diagonal input disable settings
    if ( GetConfigFile()->GetUInt("/system/config/pad_just_direction", &uValue) )
    {
        Pad_SetForceRightAngle( (uValue)? true : false );
    }

    // WPAD auto sleep settings
    {
        const char* auto_sleep = GetConfigFile()->GetString("/system/config/wpad_auto_sleep");
        if ( auto_sleep )
        {
            if ( std::strncmp( auto_sleep, "5min", 4) == 0)
            {
                WPad_SetAutoSleepTime( WPAD_SLEEP_TIME_5 );
            }
            else if ( std::strncmp( auto_sleep, "15min", 5 ) == 0 )
            {
                WPad_SetAutoSleepTime( WPAD_SLEEP_TIME_15 );
            }
            else if ( std::strncmp( auto_sleep, "30min", 5 ) == 0 )
            {
                WPad_SetAutoSleepTime( WPAD_SLEEP_TIME_30 );
            }
            else if ( std::strncmp( auto_sleep, "60min", 5 ) == 0 )
            {
                WPad_SetAutoSleepTime( WPAD_SLEEP_TIME_60 );
            }
            else if ( std::strncmp( auto_sleep, "off", 3 ) == 0 )
            {
                WPad_SetAutoSleepTime( WPAD_SLEEP_TIME_OFF );
            }
        }
    }
#endif

    // Load background image
    {
        const char* path = GetConfigFile()->GetString("/system/clr_image/file_path");
        GetClearImage().SetLatestPath( path );

        void *pData = NULL;
        s32  dataSize = 0;
        if ( path )
        {
            pData = LoadFile_FromHIO( system::SYSTEM_ID, path, &dataSize );
        }
        if ( pData )
        {
            GetClearImage().ReadImg( pData, u32(dataSize) );
            Memory_Free( pData );
        }
        else
        {
            GetClearImage().Destroy();
        }
    }

    // Placement settings for background image
    {
        const char* position = GetConfigFile()->GetString("/system/clr_image/image_position");

        if ( position )
        {
            switch ( BEndian::ReadU16( position ) )
            {
            case 'ST':
                GetClearImage().SetPosition( ClearImage::POS_STRETCH );
                break;
            case 'TI':
                GetClearImage().SetPosition( ClearImage::POS_TILE );
                break;
            case 'CE':
                GetClearImage().SetPosition( ClearImage::POS_CENTER );
                break;
            }
        }
    }

    System_ChangeSetting( &setting );
}


/*---------------------------------------------------------------------------*
  Name:         ConfigSys_WriteBackVideo_

  Description:  Writes back video-related settings

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ConfigSys_WriteBackVideo_()
{
#if 0
    VideoSetting setting;
    Video_GetSetting( &setting );
    // Screen size settings
    GetConfigFile()->SetInt("/system/video/fb_width",  setting.efbWidth );
    GetConfigFile()->SetInt("/system/video/fb_height", setting.efbHeight);
    GetConfigFile()->SetInt("/system/video/vi_width",  setting.viWidth  );
    GetConfigFile()->SetInt("/system/video/vi_height", setting.viHeight );

    GetConfigFile()->SetInt("/g3d/projection/wide_screen", setting.wideScreen ? 1:0 );
    // Clear color setting
    GetConfigFile()->SetUInt16("/system/video/clr_color", u32( nw4r::ut::Color(setting.clearColor) ) );
    GetConfigFile()->SetUInt16("/system/video/clr_z", setting.clearZ );

    // Gamma correction settings
    {
        const char* GAMMA_STR[] = { "GM_1_0", "GM_1_7", "GM_2_2" };

        GetConfigFile()->SetString("/system/video/gamma", GAMMA_STR[setting.gamma] );
    }

    // Dither Settings
    GetConfigFile()->SetInt("/system/video/dither", setting.dither);
    // Trap filter settings
    GetConfigFile()->SetInt("/system/video/trap_filter", setting.trapFilter);

    // Screen mode settings
    GetConfigFile()->SetInt("/system/video/scan_mode", static_cast<s32>( setting.scanning ) );
    GetConfigFile()->SetInt("/system/video/anti_alias", setting.antiAlias? 1:0 );
    GetConfigFile()->SetInt("/system/video/vfilter", setting.vfilter? 1:0 );
    GetConfigFile()->SetInt("/system/video/field_render", setting.fieldRender? 1:0 );

    // EFB pixel format
    if ( setting.efbPixFmt == GX_PF_RGBA6_Z24 )
    {
        GetConfigFile()->SetString("/system/video/efb_pix_fmt", "RGBA6");
    }
    else
    {
        GetConfigFile()->SetString("/system/video/efb_pix_fmt", "RGB8");
    }

    {
        const char* EFB_Z_FMT_STR[] = {"LINEAR", "NEAR", "MID", "FAR"};

        GetConfigFile()->SetString("/system/video/efb_z_fmt", EFB_Z_FMT_STR[ setting.efbZFmt ]);
    }

    // Frame rate
    GetConfigFile()->SetInt("/system/video/frame_rate", static_cast<s32>(Video_GetFrameRate()) );
#endif
}

/*---------------------------------------------------------------------------*
  Name:         ConfigSys_WriteBackMisc_

  Description:  Writes back other system settings

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
static void
ConfigSys_WriteBackMisc_()
{
#if 0
    // Default mode setting (no changes)

    // Processing meter measure
    GetConfigFile()->SetUInt("/system/config/perform_meter_length", PerformDrawer::GetMaxFrame());

    // Controller diagonal input disable settings
    GetConfigFile()->SetUInt("/system/config/pad_just_direction", Pad_GetForceRightAngle()? 1U:0U );

    // WPAD auto sleep settings
    {
        const char* SLEEP_TIME_STR[] = { "5min", "15min", "30min", "60min", "off" };
        u32 sleepTime = WPad_GetAutoSleepTime();
        GetConfigFile()->SetString("/system/config/wpad_auto_sleep", SLEEP_TIME_STR[sleepTime] );
    }

    // Background image path settings
    const char* latest_path = GetClearImage().GetLatestPath();

    if ( ! latest_path )
    {
        GetConfigFile()->SetString("/system/clr_image/dvd_path", "");
    }
    else
    {
        GetConfigFile()->SetString("/system/clr_image/dvd_path", latest_path);
    }

    // Placement settings for background image
    switch ( GetClearImage().GetPosition() )
    {
    case ClearImage::POS_STRETCH:
        GetConfigFile()->SetString("/system/clr_image/image_position", "STRETCH");
        break;
    case ClearImage::POS_TILE:
        GetConfigFile()->SetString("/system/clr_image/image_position", "TILE"   );
        break;
    case ClearImage::POS_CENTER:
        GetConfigFile()->SetString("/system/clr_image/image_position", "CENTER" );
        break;
    }
#endif
}

}   /* namespace system */
}   /* namespace viewer */




