/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>
#include <nn/pl.h>
#include "../include/TextureConverter.h"



extern "C" {
#if defined( NN_PCGX )
    extern u8* COMMON_2D_BEGIN;
    extern u8* COMMON_2D_END;
#else
    extern u8* COMMON_2D_BEGIN[];
    extern u8* COMMON_2D_END[];
#endif
}

extern "C" {
#if defined( NN_PCGX )
    extern u8* FONT_BEGIN;
    extern u8* FONT_END;
    extern u8* FONT_SHADER_BEGIN;
    extern u8* FONT_SHADER_END;
#else
    extern u8* FONT_BEGIN[];
    extern u8* FONT_END[];
    extern u8* FONT_SHADER_BEGIN[];
    extern u8* FONT_SHADER_END[];
#endif
}



namespace vw     {
namespace system {


static bool sFontLoadFlag = false;
FontDrawer* FontDrawer::m_pInstance = NULL;


FontDrawer& FontDrawer::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return *m_pInstance;
}

FontDrawer::FontDrawer() : mScreenSize(320,240), mVtxBufCmdBuf(NULL)
{
    mDrawStringBuf = AllocDispStringBuffer(256);
    mColor = nw::ut::Color8(~0);

    InitShaders();
    InitFont();
}

FontDrawer::~FontDrawer()
{
    // Free the render string buffer
    Memory_Free(mDrawStringBuf);

    mDrawer.Finalize();

    if( mVtxBufCmdBuf )
    {
        Memory_Free(mVtxBufCmdBuf);
    }

    // Destroy font
    CleanupFont(&mFont);
}

void
FontDrawer::Initialize()
{
    if (m_pInstance != NULL) return;

    void *buf = Memory_AllocSys(sizeof(FontDrawer));
    if (buf == NULL) return;

    m_pInstance = new(buf) FontDrawer();
}

void
FontDrawer::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~FontDrawer();
    Memory_Free(m_pInstance);

    m_pInstance = NULL;
}

void
FontDrawer::InitShaders()
{
    const void* shaderBinary = FONT_SHADER_BEGIN;
    const u32 shaderSize = static_cast<u32>( reinterpret_cast<int>(FONT_SHADER_END) - reinterpret_cast<int>(FONT_SHADER_BEGIN) );
    NN_NULL_ASSERT(shaderBinary);

    const u32 vtxBufCmdBufSize = nw::font::RectDrawer::GetVertexBufferCommandBufferSize(shaderBinary, shaderSize);

    void *const vtxBufCmdBuf = Memory_AllocSysDevice( vtxBufCmdBufSize );

    NN_NULL_ASSERT(vtxBufCmdBuf);
    mDrawer.Initialize(vtxBufCmdBuf, shaderBinary, shaderSize);

    mVtxBufCmdBuf = vtxBufCmdBuf;
}


void
FontDrawer::InitFont()
{
    void* buffer = NULL;

#ifdef NW_PLATFORM_CTRWIN

    buffer = FONT_SHADER_BEGIN;

#else

    if(nn::pl::SHARED_FONT_LOAD_STATE_LOADED != nn::pl::GetSharedFontLoadState())
    {
        nn::pl::Initialize();
        NN_UTIL_PANIC_IF_FAILED(nn::pl::InitializeSharedFont());
    }

    // Wait until shared font has completed loading
    while (nn::pl::GetSharedFontLoadState() != nn::pl::SHARED_FONT_LOAD_STATE_LOADED)
    {
    // Confirm that shared font loading has not failed
    if (nn::pl::GetSharedFontLoadState() == nn::pl::SHARED_FONT_LOAD_STATE_FAILED)
    {
    NN_TPANIC_("failed to load shared font!\n");
    }
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(10));
    NN_LOG("loading SharedFont ...\n");
    }
    buffer = nn::pl::GetSharedFontAddress();
    nn::pl::Finalize();
#endif

    // Sets the font resource.
    NN_NULL_ASSERT( buffer );
    bool bSuccess = mFont.SetResource( buffer );
    NN_ASSERT( bSuccess );

    // Sets the render buffer.
    const u32 drawBufferSize = nw::font::ResFont::GetDrawBufferSize( buffer );
    void* drawBuffer = Memory_AllocSys(drawBufferSize, 4);
    NN_NULL_ASSERT( drawBuffer );
    mFont.SetDrawBuffer( drawBuffer );
}

nw::font::DispStringBuffer*
FontDrawer::AllocDispStringBuffer(u32 charMax)
{
    const u32 DrawBufferSize = nw::font::CharWriter::GetDispStringBufferSize(charMax);
    void *const bufMem = Memory_AllocSys(DrawBufferSize);
    NN_NULL_ASSERT(bufMem);

    return nw::font::CharWriter::InitDispStringBuffer(bufMem, charMax);
}

void
FontDrawer::SetScreenSize(u32 width, u32 height)
{
    mScreenSize = nw::math::VEC2( static_cast<f32>(width), static_cast<f32>(height) );
}

void
FontDrawer::SetColor(nw::ut::Color8 color)
{
    mColor = color;
    mWriter.SetTextColor(color);
}

void
FontDrawer::InitDraw(u32 width, u32 height)
{
    // Color buffer information
    // Switches the width and height to match the LCD direction.
    const nw::font::ColorBufferInfo colBufInfo = { height, width, PICA_DATA_DEPTH24_STENCIL8_EXT };

    const u32 screenSettingCommands[] =
    {
        // Viewport settings
        NW_FONT_CMD_SET_VIEWPORT( 0, 0, colBufInfo.width, colBufInfo.height ),

        // Disable scissoring
        NW_FONT_CMD_SET_DISABLE_SCISSOR( colBufInfo ),

        // Disable W buffer
        // Set the depth range
        // Disable polygon offset
        NW_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(
            0.0f,           // wScale : W buffer is disabled at 0.0
            0.0f,           // depth range near
            1.0f,           // depth range far
            0,              // polygon offset units : polygon offset is disabled with 0.0
            colBufInfo),
    };

    nngxAdd3DCommand(screenSettingCommands, sizeof(screenSettingCommands), true);

    static const u32 s_InitCommands[] =
    {
        // Disable culling
        NW_FONT_CMD_SET_CULL_FACE( NW_FONT_CMD_CULL_FACE_DISABLE ),

        // Disable stencil test
        NW_FONT_CMD_SET_DISABLE_STENCIL_TEST(),

        // Disable depth test
        // Make all elements of the color buffer writable
        NW_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(
            false,  // isDepthTestEnabled
            0,      // depthFunc
            true,   // depthMask
            true,   // red
            true,   // Green
            true,   // Blue
            true),  // Alpha

        // Disable early depth test
        NW_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST( false ),

        // Framebuffer access control
        NW_FONT_CMD_SET_FBACCESS(
            true,   // colorRead
            true,   // colorWrite
            false,  // depthRead
            false,  // depthWrite
            false,  // stencilRead
            false), // stencilWrite
    };

    nngxAdd3DCommand(s_InitCommands, sizeof(s_InitCommands), true);
}

void
FontDrawer::SetupTextCamera(nw::font::RectDrawer* pDrawer, u32 width, u32 height)
{
    // Set the projection matrix for an orthogonal projection
    {
        // Sets the origin at the upper left, and the Y and Z axes in opposite directions.
        nn::math::MTX44 proj;
        f32 znear   =  0.0f;
        f32 zfar    = -1.0f;
        f32 t       =  0;
        f32 b       =  static_cast<f32>(height);
        f32 l       =  0;
        f32 r       =  static_cast<f32>(width);
        nn::math::MTX44OrthoPivot(&proj, l, r, b, t, znear, zfar, nn::math::PIVOT_UPSIDE_TO_TOP);
        pDrawer->SetProjectionMtx(proj);
    }

    // Set the model view matrix as an identity matrix
    {
        nn::math::MTX34 mv;
        nn::math::MTX34Identity(&mv);
        pDrawer->SetViewMtxForText(mv);
    }
}

void
FontDrawer::CleanupFont(nw::font::ResFont* pFont)
{
    // Disable render buffer
    // When the render buffer is set, the pointer to the buffer passed to SetDrawBuffer at build time is returned.
    // 
    void *const drawBuffer = pFont->SetDrawBuffer(NULL);
    if (drawBuffer != NULL)
    {
        Memory_Free(drawBuffer);
    }

    // Returns a pointer to the resource passed when SetResource was called if a font is set.
    // 
    void *const resource = pFont->RemoveResource();
    if (resource != NULL)
    {
      //  Memory_Free(resource);
    }
}

void
FontDrawer::SetFontSize( f32 width, f32 height)
{
    mWriter.SetFontSize( width, height);
}

void
FontDrawer::SetFixedWidth( f32 width)
{
    mWriter.SetFixedWidth(width);
}

void
FontDrawer::BeginDraw()
{
    InitDraw(mScreenSize.x, mScreenSize.y);

    mWriter.SetDispStringBuffer(mDrawStringBuf);
    mWriter.SetFont(&mFont);

    mWriter.SetCursor(0, 0);
    mWriter.SetFontSize(12.0f, 18.0f);
    mWriter.SetFixedWidth(8.5f);
    mWriter.EnableFixedWidth(true);
    mWriter.SetTextColor(mColor);

    mDrawer.DrawBegin();

    SetupTextCamera(&mDrawer, mScreenSize.x, mScreenSize.y);
}

nw::ut::Rect
FontDrawer::GetDrawRect( s32 posh, s32 posv, const char* format, ... )
{
    nw::ut::Rect rect;

    mWriter.SetCursor(static_cast<f32>(posh), static_cast<f32>(posv));
    std::va_list vargs;
    va_start(vargs, format);
    mWriter.CalcVStringRect(&rect, format, vargs);
    va_end(vargs);

    rect.Move(static_cast<f32>(posh), static_cast<f32>(posv));
    return rect;
}

/* ------------------------------------------------------------------------
        String Size Calculation
   ------------------------------------------------------------------------
template <typename CharType>
f32
TextWriterBase<CharType>::CalcFormatStringWidth(
    StreamType  format,
    ...
) const;
{
    NN_POINTER_ASSERT(format);
    ut::Rect rect;
    std::va_list vargs;
    va_start(vargs, format);
    CalcVStringRect(, format, vargs);
    va_end(vargs);
    return rect.GetWidth();
}


*/

/*
template typename CharType
void
TextWriterBaseCharType::CalcVStringRect(
    ut::Rect*       pRect,
    StreamType  format,
    std::va_list    args
) const;
{
    NN_POINTER_ASSERT(pRect);
    NN_POINTER_ASSERT(format);
    CharType* buffer = GET_FORMAT_BUFFER();
    int length = VSNPrintf(buffer, s_FormatBufferSize, format, args);
    length = math::Min(length, int(s_FormatBufferSize - 1));
    CalcStringRect(pRect, buffer, length);
}


*/

void
FontDrawer::EndDraw()
{
    mDrawer.DrawEnd();
}

f32
FontDrawer::DrawString( s32 posh, s32 posv, const char* format, ... )
{
    mWriter.StartPrint();
    mWriter.SetCursor(static_cast<f32>(posh), static_cast<f32>(posv));
    std::va_list vargs;
    va_start(vargs, format);
    f32 width = mWriter.VPrintf(format, vargs);
    va_end(vargs);
    mWriter.EndPrint();

    mDrawer.BuildTextCommand(&mWriter);
    mWriter.UseCommandBuffer();

    return width;
}

void
FontDrawer::DrawString( const char* format, ... )
{
    mWriter.StartPrint();
    std::va_list vargs;
    va_start(vargs, format);
    f32 width = mWriter.VPrintf(format, vargs);
    va_end(vargs);
    mWriter.EndPrint();

    mDrawer.BuildTextCommand(&mWriter);
    mWriter.UseCommandBuffer();
}

f32
FontDrawer::DrawString( s32 posh, s32 posv, const char* format, std::va_list vargs )
{
    mWriter.StartPrint();
    mWriter.SetCursor(static_cast<f32>(posh), static_cast<f32>(posv));
    f32 width = mWriter.VPrintf(format, vargs);
    mWriter.EndPrint();

    mDrawer.BuildTextCommand(&mWriter);
    mWriter.UseCommandBuffer();

    return width;
}

void
FontDrawer::DrawString( const char* format, std::va_list vargs )
{
    mWriter.StartPrint();
    f32 width = mWriter.VPrintf(format, vargs);
    mWriter.EndPrint();

    mDrawer.BuildTextCommand(&mWriter);
    mWriter.UseCommandBuffer();
}

void
FontDrawer::CreateCache(FontBuffer& buf, u32 charMax)
{
    buf.ReleaseBuffer();
    buf.mDrawStringBuf = AllocDispStringBuffer(charMax);
}

void
FontDrawer::BeginCache(FontBuffer& buf)
{
    mWriter.SetDispStringBuffer(buf.mDrawStringBuf);
    mWriter.SetFont(&mFont);
    mWriter.SetCursor(0, 0);
    mWriter.SetFontSize(12.0f, 18.0f);
    mWriter.SetFixedWidth(8.5f);
    mWriter.EnableFixedWidth(true);
    mWriter.StartPrint();
}

void
FontDrawer::DrawCache(s32 posh, s32 posv, const char* format, ... )
{
    mWriter.SetCursor(posh,posv);
    std::va_list vargs;
    va_start(vargs, format);
    f32 width = mWriter.VPrintf(format, vargs);
    va_end(vargs);
}

void
FontDrawer::DrawCache(const char* format, ... )
{
    std::va_list vargs;
    va_start(vargs, format);
    f32 width = mWriter.VPrintf(format, vargs);
    va_end(vargs);
}

void
FontDrawer::EndCache(FontBuffer& buf)
{
    mWriter.EndPrint();
    mDrawer.BuildTextCommand(&mWriter);
}

void
FontDrawer::DrawBuffer(const FontBuffer& buf, s32 posh, s32 posv)
{
    if(0 != buf.mDrawStringBuf)
    {
        mWriter.SetCursor(posh, posv);
        mWriter.SetDispStringBuffer(buf.mDrawStringBuf);
        mWriter.UseCommandBuffer();
    }
}

/*
  
 */
class XYTC
{
public:
    struct Vertex
    {
        f32 x,y,z;
        f32 u,v;
        u32 color;
        Vertex& setPos( float x_, float y_, float z_ )
        {
            x=x_, y=y_, z=z_;
            return *this;
        }
        Vertex& setUV( float u_, float v_ )
        {
            u=u_, v=v_;
            return *this;
        }
        Vertex& setColor( u32 col )
        {
            color = col;
            return *this;
        }
    };
    XYTC() : m_Size(0), m_Data(0), m_VertexData(0), m_IndexData(0)
    {
        m_VertexSize = 0;
        m_IndexSize  = 0;
    }
    ~XYTC()
    {
        ReleaseBuffer();
    }
    void CreateBuffer( u32 vertexNum, u32 indexNum )
    {
        ReleaseBuffer();

        // interleave info
        nn::gr::CTR::Vertex::InterleaveInfo interleave;
        interleave.dataNum = 3;
        interleave.dataType[0] = PICA_DATA_SIZE_3_FLOAT;
        interleave.dataType[1] = PICA_DATA_SIZE_2_FLOAT;
        interleave.dataType[2] = PICA_DATA_SIZE_4_UNSIGNED_BYTE;
        interleave.symbol[0]   = Draw2DHelper::GetInstance().GetSymbolPos();
        interleave.symbol[1]   = Draw2DHelper::GetInstance().GetSymbolTex();
        interleave.symbol[2]   = Draw2DHelper::GetInstance().GetSymbolClr();

        // create buffer
        m_VertexSize = sizeof(Vertex) * vertexNum;
        m_IndexSize  = sizeof(u16)    * indexNum;
        m_Size = m_VertexSize + m_IndexSize;
        m_Data = reinterpret_cast<u8*>( vw::system::Memory_AllocSysDevice(m_Size, 128) );
        m_VertexData = (Vertex*)(m_Data);
        m_IndexData  =    (u16*)(m_Data + m_VertexSize);

        // Sets up vertex information.
        uptr vtx_physical_addr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_VertexData ) );
        m_VertexStream.EnableInterleavedArray( interleave, vtx_physical_addr );

        // Sets up index.
        uptr idx_physical_addr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_IndexData ) );
        m_IndexStream.physicalAddr = idx_physical_addr;
        m_IndexStream.drawVtxNum = indexNum;
        m_IndexStream.isUnsignedByte = false;
    }
    void ReleaseBuffer()
    {
        if( m_Data )
            vw::system::Memory_Free( (void*)m_Data );
        m_Data = 0;
    }
    void UpdateVertex(u32 start, u32 size)
    {
        size = std::min( size, m_VertexSize );
        nngxUpdateBuffer( reinterpret_cast<void*>(m_VertexData+start), sizeof(Vertex)*size );
    }
    void UpdateIndex(u32 start, u32 size)
    {
        size = std::min( size, m_IndexSize );
        nngxUpdateBuffer( reinterpret_cast<void*>(m_IndexData+start), sizeof(u16)*size );
    }
    void UpdateData()
    {
        nngxUpdateBuffer( reinterpret_cast<void*>(m_Data), m_Size );
    }
    void Draw(u32 drawVtxNum=~0)
    {
        u32*& command = Draw2DHelper::GetInstance().GetCommand();

        if( drawVtxNum>= m_IndexSize)
            drawVtxNum = m_IndexSize;

        m_IndexStream.drawVtxNum = drawVtxNum;

        command = m_VertexStream.MakeDisableAttrCommand( command );
        command = m_VertexStream.MakeEnableAttrCommand( command );
        command = m_VertexStream.MakeDrawCommand( command, m_IndexStream );
        command = m_VertexStream.MakeDisableAttrCommand( command );
        command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand(command);
    }
    Vertex* GetVertexData() const
    {
        return m_VertexData;
    }
    u16* GetIndexData()  const
    {
        return m_IndexData;
    }
    void DrawOffset(u32 start, u32 size)
    {
        u32*& command = Draw2DHelper::GetInstance().GetCommand();

        nn::gr::Vertex::IndexStream stream;
        uptr idx_physical_addr = nngxGetPhysicalAddr( reinterpret_cast< uptr >( m_IndexData+start ) );
        stream.physicalAddr   = idx_physical_addr;
        stream.drawVtxNum     = size;
        stream.isUnsignedByte = false;

        command = m_VertexStream.MakeDisableAttrCommand( command );
        command = m_VertexStream.MakeEnableAttrCommand( command );
        command = m_VertexStream.MakeDrawCommand( command, stream );
        command = m_VertexStream.MakeDisableAttrCommand( command );
        command = nn::gr::CTR::FrameBuffer::MakeClearCacheCommand(command);
    }

private:
    nn::gr::Vertex              m_VertexStream;
    nn::gr::Vertex::IndexStream m_IndexStream;
    u32                         m_Size;
    u8*                         m_Data;
    Vertex*                     m_VertexData;
    u32                         m_VertexSize;
    u16*                        m_IndexData;
    u32                         m_IndexSize;
    CommandCache                m_Command;
};
static XYTC* sXYTC = NULL;

Draw2DHelper* Draw2DHelper::m_pInstance = NULL;

/*
  
*/
Draw2DHelper&
Draw2DHelper::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return *m_pInstance;
}

//--------------------------------------------------------------------------
Draw2DHelper::Draw2DHelper(u32 TriangleBufferSize)
{
    m_Viewport.Set( 0, 0, 240, 320 );
    m_Scissor.Set( 0, 0, 240, 320 );
    m_RenderState.blend.isEnable = false;
    m_RenderState.cullingTest.isEnable = false;
    m_RenderState.depthTest.isEnable = false;
    m_RenderState.alphaTest.isEnable = false;

    m_PrimitiveCount = 0;
    m_PrimitiveSize  = 0;
    m_Width = 320;
    m_Height = 240;
    m_LineWidth = 2.0f;
    m_Depth = 0.0f;
    m_PrimitiveStart = 0;
    m_PrimitiveCount = 0;
    m_PrimitiveSize  = TriangleBufferSize;

    m_UVScale = nn::math::VEC2(1.0f,1.0f);
    m_Command = 0;

    // Load shader binary
    u32 shader_file_size = static_cast<u32>( reinterpret_cast<int>(COMMON_2D_END) - reinterpret_cast<int>(COMMON_2D_BEGIN) );
    m_ShaderBinary = static_cast<u8*>( vw::system::Memory_AllocSysDevice(shader_file_size, 128) );
    memcpy( m_ShaderBinary, COMMON_2D_BEGIN, shader_file_size );

    // Specify shader binary and initialize
    // Enable the executable image at position 0 as the vertex shader
    m_Shader.SetupBinary( m_ShaderBinary, 0, -1 );

    // Search for "Location."
    m_Shader.SearchBindSymbol( &m_SymbolPos, "Position" );
    m_Shader.SearchBindSymbol( &m_SymbolTex, "Texcoord" );
    m_Shader.SearchBindSymbol( &m_SymbolClr, "Color" );

    // Uniform search
    m_Shader.SearchBindSymbol( &m_SymbolScale, "uScale" );
}

//--------------------------------------------------------------------------
Draw2DHelper::~Draw2DHelper()
{
    if( m_ShaderBinary != 0 )
    {
        vw::system::Memory_Free( m_ShaderBinary );
        m_ShaderBinary = 0;
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::Initialize( u32 TriangleBufferSize )
{
    if (m_pInstance != NULL) return;
    void *buf = Memory_AllocSys(sizeof(Draw2DHelper));
    if (buf == NULL) return;

    NN_GR_ASSERT( TriangleBufferSize > 0 );
    m_pInstance = new(buf) Draw2DHelper(TriangleBufferSize);

    // create primitive buffer
    buf = vw::system::Memory_AllocSys( sizeof(XYTC) );
    sXYTC = new(buf) XYTC();

    sXYTC->CreateBuffer( TriangleBufferSize*3, TriangleBufferSize*3 );

    u16* index = sXYTC->GetIndexData();
    for(u16 i=0; i<TriangleBufferSize*3; ++i)
        index[i] = i;
    sXYTC->UpdateIndex( 0,TriangleBufferSize * 3 );
}

//--------------------------------------------------------------------------
void
Draw2DHelper::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~Draw2DHelper();
    Memory_Free(m_pInstance);

    m_pInstance = NULL;

    if( sXYTC )
    {
        sXYTC->~XYTC();
        vw::system::Memory_Free( sXYTC );
    }
    sXYTC = 0;


    CommandCache::Finalize();
}

// Configures the screen size.
void
Draw2DHelper::SetScreenSize(u32 width, u32 height)
{
    m_Viewport.Set( 0, 0, height, width );
    m_Scissor.Set( 0, 0, height, width );
    m_Scissor.SetBufferSize(height, width);

    m_Width=width, m_Height=height;
}

//--------------------------------------------------------------------------
void
Draw2DHelper::Begin()
{
    m_PrimitiveCount = 0;
    m_PrimitiveStart = 0;
}

//--------------------------------------------------------------------------
void
Draw2DHelper::BeginDraw()
{
    // Get a pointer to the render command buffer
    nngxGetCmdlistParameteri( NN_GX_CMDLIST_CURRENT_BUFADDR, reinterpret_cast<GLint*>( &m_Command ) );

    // Get the start address (so that the pointer can be advanced at the end)
    m_StartAddr = m_Command;

    // Viewport settings
    m_Command = m_Viewport.MakeCommand( m_Command );

    // Shader binary settings ( glUseProgram() )
    m_Command = m_Shader.MakeFullCommand( m_Command );

    // Blend mode settings
    m_Command = m_RenderState.MakeCommand( m_Command );

    // Texture combiner settings
    nn::gr::Combiner combiner;
    combiner.stage[ 0 ].rgb.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    combiner.stage[ 0 ].rgb.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
    combiner.stage[ 0 ].rgb.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    combiner.stage[ 0 ].alpha.combine   = PICA_DATA_TEX_ENV_COMBINE_REPLACE;
    combiner.stage[ 0 ].alpha.source[0] = PICA_DATA_TEX_ENV_SRC_RGBA_PRIMARY_COLOR;
    combiner.stage[ 0 ].alpha.source[1] = PICA_DATA_TEX_ENV_SRC_RGBA_FRAGMENT_SECONDARY_COLOR_DMP;
    m_Command = combiner.MakeCommand( m_Command );

    // Texture settings
    nn::gr::Texture texture;
    texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_FALSE;
    texture.unit1.texType = PICA_DATA_TEXTURE1_SAMPLER_TYPE_FALSE;
    texture.unit2.texType = PICA_DATA_TEXTURE2_SAMPLER_TYPE_FALSE;
    texture.unit3.texType = PICA_DATA_TEXTURE3_SAMPLER_TYPE_FALSE;
    m_Command = texture.MakeCommand( m_Command );

    // Fog configuration
    nn::gr::CTR::Fog fog;
    fog.isEnable = false;
    m_Command = fog.MakeConfigCommand( m_Command );

    // Configures the screen scale.
    nn::math::VEC4 uScale(-2.0f/m_Width,-2.0f/m_Height,1.0f,0.0f);
    m_Command = m_SymbolScale.MakeUniformCommand( m_Command, uScale );

    m_PrimitiveStart = m_PrimitiveCount;

    m_Depth = 0.0f;
    m_UVScale = nn::math::VEC2(1.0f,1.0f);

#if defined( NN_PCGX )
    *m_Command++ = 0x1;
    *m_Command++ = PICA_CMD_HEADER_SINGLE( PCGX_EXTCMD_PIVOT );
#endif
}

//--------------------------------------------------------------------------
void
Draw2DHelper::EndDraw()
{
    // Flush the buffer
    if( m_PrimitiveCount - m_PrimitiveStart > 0 )
    {
        u32 size = m_PrimitiveCount - m_PrimitiveStart;
        sXYTC->DrawOffset(m_PrimitiveStart*3, size*3);
    }

#if defined( NN_PCGX )
    *m_Command++ = 0x0;
    *m_Command++ = PICA_CMD_HEADER_SINGLE( PCGX_EXTCMD_PIVOT );
#endif

    // Advance the pointer to the render command buffer
    if( m_Command-m_StartAddr > 0)
    {
        nngxMoveCommandbufferPointer( ( m_Command - m_StartAddr ) * sizeof(u32) );
    }
    m_Command = 0;
}

//--------------------------------------------------------------------------
void
Draw2DHelper::End()
{
    // Flushes last buffer.
    if( m_PrimitiveCount > 0 )
    {
        sXYTC->UpdateVertex( 0, m_PrimitiveCount * 3 );
        m_PrimitiveCount = 0;
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::Flush()
{
    // Flush the buffer
    if( m_PrimitiveCount - m_PrimitiveStart > 0 )
    {
        u32 size = m_PrimitiveCount - m_PrimitiveStart;
        sXYTC->DrawOffset(m_PrimitiveStart*3, size*3);

        m_PrimitiveStart = m_PrimitiveCount;
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::Enable(STATE state)
{
    if( state & ALPHATEST )
    {
        m_RenderState.alphaTest.isEnable = true;
        m_RenderState.alphaTest.func = PICA_DATA_ALPHA_TEST_GREATER;
        m_RenderState.alphaTest.refValue = 90;    // Since the current ALPHATEST flag has only been used from SceneRenderer::DrawIcons, refValue has been made a fixed value. If necessary, add a refValue accessor.

        if( m_Command )
        {
            m_Command = m_RenderState.alphaTest.MakeCommand( m_Command );
        }
    }
    if( state & ALPHABLEND )
    {
        m_RenderState.blend.isEnable = true;
        m_RenderState.blend.eqRgb  = PICA_DATA_BLEND_EQUATION_ADD;
        m_RenderState.blend.srcRgb = PICA_DATA_BLEND_FUNC_SRC_ALPHA;
        m_RenderState.blend.dstRgb = PICA_DATA_BLEND_FUNC_ONE_MINUS_SRC_ALPHA;

        if( m_Command )
        {
            m_Command = m_RenderState.blend.MakeCommand( m_Command );
        }
    }
    if( state & CULLFACE )
    {
        m_RenderState.cullingTest.isEnable = true;
        if( m_Command )
        {
            m_Command = m_RenderState.cullingTest.MakeCommand( m_Command );
        }
    }
    if( state & DEPTHWRITE )
    {
        m_RenderState.depthTest.isEnableWrite = true;
        if( m_Command )
        {
            m_Command = m_RenderState.depthTest.MakeCommand( m_Command );
        }
    }
    if ( state & DEPTHTEST )
    {
        m_RenderState.depthTest.isEnable = true;
        if( m_Command )
        {
            m_Command = m_RenderState.depthTest.MakeCommand ( m_Command );
        }
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::Disable(STATE state)
{
    if( state & ALPHATEST )
    {
        m_RenderState.alphaTest.isEnable = false;
        if( m_Command )
        {
            m_Command = m_RenderState.alphaTest.MakeCommand( m_Command );
        }
    }
    if ( state & ALPHABLEND )
    {
        m_RenderState.blend.isEnable = false;
        if( m_Command )
        {
            m_Command = m_RenderState.blend.MakeCommand( m_Command );
        }
    }
    if ( state & CULLFACE )
    {
        m_RenderState.cullingTest.isEnable = false;
        if( m_Command )
        {
            m_Command = m_RenderState.cullingTest.MakeCommand( m_Command );
        }
    }
    if ( state & DEPTHWRITE )
    {
        m_RenderState.depthTest.isEnableWrite = false;
        if( m_Command )
        {
            m_Command = m_RenderState.depthTest.MakeCommand( m_Command );
        }
    }
    if ( state & DEPTHTEST )
    {
        m_RenderState.depthTest.isEnable = false;
        if( m_Command )
        {
            m_Command = m_RenderState.depthTest.MakeCommand ( m_Command );
        }
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::DrawLine( const nw::math::VEC2& p1, const nw::math::VEC2& p2, nw::ut::Color8 color )
{
    u32* color32 = (u32*)&color;
    this->BuildLine(p1, p2, m_LineWidth, *color32);
}

//--------------------------------------------------------------------------
void
Draw2DHelper::DrawRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, nw::ut::Color8 color )
{
    u32* color32 = (u32*)&color;
    this->BuildLine(pos, nw::math::VEC2(pos.x + size.x, pos.y),        m_LineWidth, *color32);
    this->BuildLine(pos, nw::math::VEC2(pos.x, pos.y + size.y),        m_LineWidth, *color32);
    this->BuildLine(nw::math::VEC2(pos.x + size.x, pos.y), pos + size, m_LineWidth, *color32);
    this->BuildLine(nw::math::VEC2(pos.x, pos.y + size.y), pos + size, m_LineWidth, *color32);
}

//--------------------------------------------------------------------------
void
Draw2DHelper::FillRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, nw::ut::Color8 color )
{
    u32* color32 = (u32*)&color;
    this->BuildRectangle( pos, size, *color32 );
}

//--------------------------------------------------------------------------
void
Draw2DHelper::FillTriangle( const nw::math::VEC2& pt1, const nw::math::VEC2& pt2, const nw::math::VEC2& pt3, nw::ut::Color8 color )
{
    u32* color32 = (u32*)&color;
    this->BuildTriangle( pt1, pt2, pt3, *color32 );
}

//--------------------------------------------------------------------------
void
Draw2DHelper::FillTriangles( const nw::math::VEC2* pPointArray, s32 pointCount, nw::ut::Color8 color )
{
    NW_NULL_ASSERT( pPointArray );
    NW_ASSERT( pointCount > 0 );

    u32 count = pointCount / 3;
    for(u32 i=0; i<count; ++i){
        u32 j=i*3;
        this->FillTriangle(pPointArray[j+0],
                           pPointArray[j+1],
                           pPointArray[j+2], color);
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::BuildTriangle( const nw::math::VEC2& pt1, const nw::math::VEC2& pt2, const nw::math::VEC2& pt3, u32 color )
{
    XYTC::Vertex* vertex = sXYTC->GetVertexData() + m_PrimitiveCount * 3;
    m_PrimitiveCount++;
    vertex[0].setPos(pt1.x,pt1.y,m_Depth).setColor(color);
    vertex[1].setPos(pt2.x,pt2.y,m_Depth).setColor(color);
    vertex[2].setPos(pt3.x,pt3.y,m_Depth).setColor(color);
}

//--------------------------------------------------------------------------
void Draw2DHelper::BuildRectangle(  const nw::math::VEC2& pt1, const nw::math::VEC2& pt2,
                                    const nw::math::VEC2& pt3, const nw::math::VEC2& pt4, u32 color )
{
    const nw::math::VEC2 uv1(0.0f,m_UVScale.y), uv2(m_UVScale.x,m_UVScale.y), uv3(0.0f,0.0f), uv4(m_UVScale.x,0.0f);
    nw::math::VEC2 pos[] = {pt1,pt2,pt4, pt4,pt2,pt3};
    const nw::math::VEC2 uv[] = {uv1,uv2,uv3, uv3,uv2,uv4};

    BuildTriangle( pos,   uv,   color );
    BuildTriangle( pos+3, uv+3, color );
}

//--------------------------------------------------------------------------
void
Draw2DHelper::BuildTriangle( const nw::math::VEC2* pos, const nw::math::VEC2* uv, u32 color )
{
    XYTC::Vertex* vertex = sXYTC->GetVertexData() + m_PrimitiveCount * 3;
    m_PrimitiveCount++;
    for(u32 i=0; i<3; ++i)
    {
        vertex[i].setPos(pos[i].x,pos[i].y,m_Depth).
                  setUV(uv[i].x,uv[i].y).
                  setColor(color);
    }
}

//--------------------------------------------------------------------------
void
Draw2DHelper::BuildRectangle( const nw::math::VEC2& pos, const nw::math::VEC2& size, u32 color )
{
    BuildRectangle(
        pos,
        nw::math::VEC2(pos.x + size.x, pos.y),
        pos + size,
        nw::math::VEC2(pos.x, pos.y + size.y),
        color );
}

//--------------------------------------------------------------------------
void
Draw2DHelper::BuildLine( const nw::math::VEC2& pt1, const nw::math::VEC2& pt2, f32 lineWidth, u32 color )
{
    NW_ASSERT( lineWidth >= 0.f );

    if( pt1 == pt2 )
    {
        BuildRectangle( pt1, pt1, pt2, pt2, color );
        return;
    }

    nw::math::VEC2  dir  = pt1 - pt2;
    nw::math::VEC2 ortho = nw::math::VEC2(dir.y, -dir.x);

    VEC2Normalize( &ortho, &ortho );
    ortho *= lineWidth/2;

    VEC2Normalize( &dir, &dir );
    dir *= lineWidth/2;

    BuildRectangle( pt1-ortho+dir, pt1+ortho+dir, pt2+ortho-dir, pt2-ortho-dir, color );
}



static PicaDataTextureFormat GetPICAFormat(nn::tpl::CTR::CtrTexFormat ctr)
{
    switch(ctr)
    {
        case nn::tpl::TEXFMT_RGBA8888: return PICA_DATA_TEXTURE_FORMAT_RGBA_UNSIGNED_BYTE;
        case nn::tpl::TEXFMT_RGB888:   return PICA_DATA_TEXTURE_FORMAT_RGB_UNSIGNED_BYTE;
        case nn::tpl::TEXFMT_RGBA5551: return PICA_DATA_TEXTURE_FORMAT_RGBA_UNSIGNED_SHORT_5_5_5_1;
        case nn::tpl::TEXFMT_RGB565:   return PICA_DATA_TEXTURE_FORMAT_RGB_UNSIGNED_SHORT_5_6_5;
        case nn::tpl::TEXFMT_RGBA4444: return PICA_DATA_TEXTURE_FORMAT_RGBA_UNSIGNED_SHORT_4_4_4_4;
        case nn::tpl::TEXFMT_LA88:     return PICA_DATA_TEXTURE_FORMAT_LUMINANCE_ALPHA_UNSIGNED_BYTE;
        case nn::tpl::TEXFMT_HL8:      return PICA_DATA_TEXTURE_FORMAT_HILO8_DMP_UNSIGNED_BYTE;
        case nn::tpl::TEXFMT_L8:       return PICA_DATA_TEXTURE_FORMAT_LUMINANCE_UNSIGNED_BYTE;
        case nn::tpl::TEXFMT_A8:       return PICA_DATA_TEXTURE_FORMAT_ALPHA_UNSIGNED_BYTE;
        case nn::tpl::TEXFMT_LA44:     return PICA_DATA_TEXTURE_FORMAT_LUMINANCE_ALPHA_UNSIGNED_BYTE_4_4_DMP;
        case nn::tpl::TEXFMT_L4:       return PICA_DATA_TEXTURE_FORMAT_LUMINANCE_UNSIGNED_4BITS_DMP;
        case nn::tpl::TEXFMT_A4:       return PICA_DATA_TEXTURE_FORMAT_ALPHA_UNSIGNED_4BITS_DMP;
        case nn::tpl::TEXFMT_ETC1:     return PICA_DATA_TEXTURE_FORMAT_ETC1_RGB8_NATIVE_DMP;
        case nn::tpl::TEXFMT_ETC1_A4:  return PICA_DATA_TEXTURE_FORMAT_ETC1_ALPHA_RGB8_A4_NATIVE_DMP;
        default: return PICA_DATA_TEXTURE_FORMAT_RGBA_UNSIGNED_BYTE;
    }
}

Texture::Texture() : m_Data(0)
{
}
Texture::Texture(const char* filename, s16 index) : m_Data(0)
{
    LoadTexture(filename, index);
    InitState();
}
Texture::Texture(const char* filename, const char* tex_name) : m_Data(0)
{
    LoadTexture(filename, tex_name);
    InitState();
}
Texture::Texture(const void* data, u32 width, u32 height) : m_Data(0)
{
    u32 size = width * height * sizeof(u32);
    m_Data = static_cast<u8*>( vw::system::Memory_AllocSysDevice( size, 128) );

    ConvertGLTextureToNative( GL_RGBA_NATIVE_DMP, width, height, const_cast<void*>(data), m_Data );

    nngxUpdateBuffer(m_Data, size);

    // Texture settings
    m_Texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_2D;
    m_Texture.unit0.physicalAddr = nngxGetPhysicalAddr(reinterpret_cast<uptr>(m_Data));
    m_Texture.unit0.width  = width;
    m_Texture.unit0.height = height;
    m_Texture.unit0.format = PICA_DATA_TEXTURE_FORMAT_RGBA_UNSIGNED_BYTE;
}
Texture::~Texture()
{
    if( m_Data )
    {
        vw::system::Memory_Free(m_Data);
    }
    m_Data = 0;
}

bool
Texture::SetData(const void* data, const char* tex_name)
{
    if( data )
    {
        // Get the texture index
        s16 index = nn::tpl::CTR::GetTextureIndex(data,tex_name);
        if( index != -1 )
        {
            SetupData(data,index);
        }
    }
    InitState();
    return true;
}

void
Texture::SetState()
{
    if( 0 == m_Data )
    {
        return;
    }
    u32*& command = Draw2DHelper::GetInstance().GetCommand();
    command =  m_Texture.MakeCommand( command );
    command = m_Combiner.MakeCommand( command );
}

void
Texture::SetCombiner(nn::gr::Combiner combiner)
{
    m_Combiner = combiner;
}

void
Texture::SetFilter(PicaDataTextureMagFilter magFilter, PicaDataTextureMinFilter minFilter)
{
    m_Texture.unit0.magFilter = magFilter;
    m_Texture.unit0.minFilter = minFilter;
}

void
Texture::SetWrapMode(PicaDataTextureWrap T, PicaDataTextureWrap S)
{
    m_Texture.unit0.wrapT = T;
    m_Texture.unit0.wrapS = S;
}

void
Texture::ResetState()
{
    u32*& command = Draw2DHelper::GetInstance().GetCommand();
    nn::gr::Texture texture;
    nn::gr::Combiner combiner;
    command = texture.MakeCommand( command );
    command = combiner.MakeCommand( command );
}

void
Texture::LoadTexture(const char* filename, s16 index)
{
    void* buffer = LoadData(filename);
    if( buffer )
    {
        if(index>=0 && index<nn::tpl::GetTextureNum(buffer))
        {
            SetupData(buffer,index);
        }
        vw::system::Memory_Free(buffer);
    }
}

void
Texture::LoadTexture(const char* filename, const char* tex_name)
{
    void* buffer = LoadData(filename);
    if( buffer )
    {
        // Get the texture index
        s16 index = nn::tpl::CTR::GetTextureIndex(buffer,tex_name);
        if( index != -1 )
        {
            SetupData(buffer,index);
        }
        vw::system::Memory_Free(buffer);
    }
}

void
Texture::SetupData(const void* buffer, s16 index)
{
    s32 mipLevel = 0;
    u32 mipmapSize = 0;
    u8* tex_data = (u8*)nn::tpl::GetTexture(&mipLevel, &mipmapSize, buffer, index);
    if( !tex_data )
    {
        return;
    }
    bool result = nn::tpl::GetTextureInfo(&m_Info, buffer, index);
    if( !result )
    {
        return;
    }
    m_Data = static_cast<u8*>( vw::system::Memory_AllocSysDevice(m_Info.texDataSize, 128) );
    std::memcpy(m_Data, tex_data, m_Info.texDataSize);
    nngxUpdateBuffer(reinterpret_cast<void*>(m_Data), m_Info.texDataSize);
}

void*
Texture::LoadData(const char* filename)
{
    nn::fs::FileReader file(filename);
    size_t size = file.GetSize();
    if ( file.GetSize() > 0 )
    {
        u8* buffer = (u8*)vw::system::Memory_AllocSys( file.GetSize());
        file.Read(buffer, file.GetSize());
        return buffer;
    }
    return 0;
}

void
Texture::InitState()
{
    if( m_Data )
    {
        // Texture settings
        m_Texture.unit0.texType = PICA_DATA_TEXTURE0_SAMPLER_TYPE_TEXTURE_2D;
        m_Texture.unit0.physicalAddr = nngxGetPhysicalAddr(reinterpret_cast<uptr>(m_Data));
        m_Texture.unit0.width  = m_Info.width;
        m_Texture.unit0.height = m_Info.height;
        m_Texture.unit0.format = GetPICAFormat(m_Info.texFormat);
    }
}



} /* namespace system */
} /* namespace vw   */
