/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/ImageReader.h"
#include "../include/BmpReaderImpl.h"
#include "../include/TgaReaderImpl.h"
#include "../include/JpegReaderImpl.h"

namespace vw     {
namespace system {

/*---------------------------------------------------------------------------*
  Name:         ImageReader::ImageReader

  Description:  Constructor

  Arguments:    data

  Returns:      None.
 *---------------------------------------------------------------------------*/
/* ctor */
ImageReader::ImageReader( void* data, u32 dataSize /* = 0 */ )
 : mReaderImpl( NULL ),
   mBuf( data ),
   mType( ImageReaderImpl::TYPE_UNKNOWN ),
   mWidth( 0 ),
   mHeight( 0 )
{
    bool isBmp = false;
    bool isJpeg = false;
    u8* buf = static_cast<u8*>( data );

    if ( (buf[0] == 'B') && (buf[1] == 'M') )
    {
        isBmp = true;
    }
    else if ( buf[0] == 0xFF )
    {
        isJpeg = true;
    }

    if ( isBmp )
    {
        void* impl_buf = vw::system::Memory_AllocSys( sizeof(BmpReaderImpl) );
        mReaderImpl = new(impl_buf) BmpReaderImpl( data );
    }
    else if ( isJpeg )
    {
        void* impl_buf = vw::system::Memory_AllocSys( sizeof(JpegReaderImpl) );
        mReaderImpl = new(impl_buf) JpegReaderImpl( data, dataSize );
    }
    else
    {
        void* impl_buf = vw::system::Memory_AllocSys( sizeof(TgaReaderImpl) );
        mReaderImpl = new(impl_buf) TgaReaderImpl( data );
    }

    if ( ! mReaderImpl->IsValid() )
    {
        mType = ImageReaderImpl::TYPE_UNKNOWN;
        vw::system::Memory_Free( mReaderImpl );
        mReaderImpl = NULL;
        return;
    }
    else
    {
        mType = mReaderImpl->GetType();
    }

    mWidth  = mReaderImpl->GetWidth();
    mHeight = mReaderImpl->GetHeight();
}


/*---------------------------------------------------------------------------*
  Name:         ImageReader::~ImageReader

  Description:  Destructor

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
/* dtor */
ImageReader::~ImageReader( void )
{
    if ( mReaderImpl )
    {
        mReaderImpl->Dispose();
        mReaderImpl->~ImageReaderImpl();
        vw::system::Memory_Free( mReaderImpl );
    }
}



/*---------------------------------------------------------------------------*
  Name:         ImageReader::Expand

  Description:  Expands the image data as RGB8 texture

  Arguments:    expBuf      Buffer to expand to
                bufWidth    Buffer horizontal size
                bufHeight   Buffer vertical size

  Returns:      None.
 *---------------------------------------------------------------------------*/
bool
ImageReader::Expand( void* expBuf, u32 bufWidth, u32 bufHeight )
{
    if ( mReaderImpl )
    {
        return mReaderImpl->Expand( expBuf, bufWidth, bufHeight );
    }
    return false;
}



}   /* namespace system */
}   /* namespace vw     */


