/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/LogSystem.h"


namespace vw     {
namespace system {


LogSystem* LogSystem::m_pInstance = NULL;

/*===========================================================================*
  Name:         LogSystem

  Description:  Log management class
 *===========================================================================*/



LogSystem::Log::Log()
{

}
LogSystem::Log::~Log()
{
    m_Message = vw::common::CharArray();
}

LogSystem*
LogSystem::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

void
LogSystem::Initialize()
{
    if (m_pInstance != NULL) return;
    void *buf = Memory_AllocSys(sizeof(LogSystem));
    if (buf == NULL) return;

    m_pInstance = new(buf) LogSystem();
}

void
LogSystem::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~LogSystem();
    Memory_Free(m_pInstance);

    m_pInstance = NULL;
}

/* ctor */
LogSystem::LogSystem() : m_Log( vw::system::Memory_GetSysAllocator() ),
                         m_Error( vw::system::Memory_GetSysAllocator() ),
                         m_Warning( vw::system::Memory_GetSysAllocator() ),
                         m_SystemLog( vw::system::Memory_GetSysAllocator() )
{
}

/* ctor */
LogSystem::~LogSystem()
{
    Clear();

    m_Log =  vw::system::Memory_GetSysAllocator();
    m_Error = vw::system::Memory_GetSysAllocator();
    m_Warning = vw::system::Memory_GetSysAllocator();
    m_SystemLog = vw::system::Memory_GetSysAllocator();
}

void
LogSystem::AddError(const char* message, u32 hashID, void* mainKey, u32 subKey)
{
    if( message == 0 || IsExist(ERROR, mainKey, subKey))
    {
        return;
    }

    void* buf = vw::system::Memory_AllocSys(sizeof(Log));
    Log* log = new(buf) Log();
    log->m_Message.SetAllocator(vw::system::Memory_GetSysAllocator());
    log->m_Type = ERROR;
    log->m_Message = message;
    log->m_HashID = hashID;
    log->m_MainKey = mainKey;
    log->m_SubKey = subKey;
    log->m_Index = m_Log.Size();
    m_Error.push_back( log->m_Index );
    m_Log.push_back( log );
}

void
LogSystem::AddWarning(const char* message, u32 hashID, void* mainKey, u32 subKey)
{
    if( message == 0 || IsExist(WARNING, mainKey, subKey))
    {
        return;
    }

    void* buf = vw::system::Memory_AllocSys(sizeof(Log));
    Log* log = new(buf) Log();
    log->m_Type = WARNING;
    log->m_Message = message;
    log->m_HashID = hashID;
    log->m_MainKey = mainKey;
    log->m_SubKey = subKey;
    log->m_Index = m_Log.size();
    m_Warning.push_back( log->m_Index );
    m_Log.push_back( log );
}

void
LogSystem::AddSystemLog(const char* message)
{
    if( message == 0)
    {
        return;
    }

    void* buf = vw::system::Memory_AllocSys(sizeof(Log));
    Log* log = new(buf) Log();
    log->m_Type = SYSTEM_LOG;
    log->m_Message = message;
    log->m_HashID = ~0;
    log->m_MainKey = NULL;
    log->m_SubKey = 0;
    log->m_Index = m_Log.size();
    m_SystemLog.push_back( log->m_Index );
    m_Log.push_back( log );
}

u32
LogSystem::GetNumError() const
{
    return (u32)m_Error.size();
}

u32
LogSystem::GetNumWarning() const
{
    return (u32)m_Warning.size();
}

u32
LogSystem::GetNumSystemLog() const
{
    return (u32)m_SystemLog.size();
}

u32
LogSystem::GetNumLog() const
{
    return (u32)m_Log.size();
}

LogSystem::Log*
LogSystem::GetError(u32 index)
{
    if( index < m_Error.size() )
    {
        return m_Log[ m_Error[index] ];
    }
    return 0;
}

LogSystem::Log*
LogSystem::GetWarning(u32 index)
{
    if( index < m_Warning.size() )
    {
        return m_Log[ m_Warning[index] ];
    }
    return 0;
}

LogSystem::Log*
LogSystem::GetSystemLog(u32 index)
{
    if( index < m_SystemLog.size() )
    {
        return m_Log[ m_SystemLog[index] ];
    }
    return 0;
}

LogSystem::Log*
LogSystem::GetLog(u32 index)
{
    if( index < m_Log.size() )
    {
        return m_Log[index];
    }
    return 0;
}

bool
LogSystem::IsExist(TYPE type, void* mainKey, u32 subKey)
{
    for(u32 i=0, n=0; i<m_Log.size(); ++i)
    {
        if( type == m_Log[i]->m_Type && mainKey == m_Log[i]->m_MainKey && subKey == m_Log[i]->m_SubKey )
        {
            return true;
        }
    }

    return false;
}

void
LogSystem::Clear()
{
    for(u32 i=0; i<m_Log.size(); ++i)
    {
        m_Log[i]->~Log();
        vw::system::Memory_Free( m_Log[i] );
    }

    m_Log.clear();
    m_Error.clear();
    m_Warning.clear();
    m_SystemLog.clear();
}

void
LogSystem::Clear(TYPE type)
{
    switch(type)
    {
    case ERROR:
        for(u32 i=0; i<m_Error.size(); ++i)
        {
            m_Log[ m_Error[i] ]->m_Index = ~0;
        }
        m_Error.clear();
        break;
    case WARNING:
        for(u32 i=0; i<m_Warning.size(); ++i)
        {
            m_Log[ m_Warning[i] ]->m_Index = ~0;
        }
        m_Warning.clear();
        break;
    case SYSTEM_LOG:
        for(u32 i=0; i<m_SystemLog.size(); ++i)
        {
            m_Log[ m_SystemLog[i] ]->m_Index = ~0;
        }
        m_SystemLog.clear();
        break;
    case HISTORY:
        m_Error.clear();
        m_Warning.clear();
        m_SystemLog.clear();
        m_Log.clear();
        return;
    }

    for(u32 i=0, n=0; i<m_Log.size(); ++i)
    {
        if( m_Log[i]->m_Index != ~0 )
        {
            m_Log[i]->m_Index = n++;
        }
    }

    this->ResetIndex();
}

void
LogSystem::Clear(u32 hashID)
{
    for(u32 i=0, n=0; i<m_Log.size(); ++i)
    {
        if( hashID != m_Log[i]->m_HashID )
        {
            m_Log[i]->m_Index = n++;
        }
        else
        {
            m_Log[i]->m_Index = ~0;
        }
    }
    this->ResetIndex();
}

void
LogSystem::Clear(void* mainKey)
{
    for(u32 i=0, n=0; i<m_Log.size(); ++i)
    {
        if( mainKey != m_Log[i]->m_MainKey )
        {
            m_Log[i]->m_Index = n++;
        }
        else
        {
            m_Log[i]->m_Index = ~0;
        }
    }
    this->ResetIndex();
}

void
LogSystem::Clear(void* mainKey, u32 subKey)
{
    for(u32 i=0, n=0; i<m_Log.size(); ++i)
    {
        if( mainKey != m_Log[i]->m_MainKey || subKey != m_Log[i]->m_SubKey )
        {
            m_Log[i]->m_Index = n++;
        }
        else
        {
            m_Log[i]->m_Index = ~0;
        }
    }
    this->ResetIndex();
}

void
LogSystem::ResetIndex()
{
    u32 count = 0;
    for(u32 i=0; i<m_Error.size(); ++i)
    {
        if( m_Log[ m_Error[i] ]->m_Index != ~0)
            m_Error[count++] = m_Log[ m_Error[i] ]->m_Index;
    }
    m_Error.resize( count );

    count = 0;
    for(u32 i=0; i<m_Warning.size(); ++i)
    {
        if( m_Log[ m_Warning[i] ]->m_Index != ~0)
            m_Warning[ count++ ] = m_Log[ m_Warning[i] ]->m_Index;
    }
    m_Warning.resize( count );

    count = 0;
    for(u32 i=0, n=0; i<m_SystemLog.size(); ++i)
    {
        if( m_Log[ m_SystemLog[i] ]->m_Index != ~0)
            m_SystemLog[ count++ ] = m_Log[ m_SystemLog[i] ]->m_Index;
    }
    m_SystemLog.resize( count );

    count = 0;
    for(u32 i=0, n=0; i<m_Log.size(); ++i)
    {
        u32 index = m_Log[i]->m_Index;
        if( index == ~0 )
        {
            m_Log[i]->~Log();
            vw::system::Memory_Free( m_Log[i] );
        }
        else
        {
            m_Log[index] = m_Log[i];
            count++;
        }
    }
    m_Log.resize( count );
}

void
LogSystem::OutputLog(const char* filename, TYPE type)
{
    nn::hio::CTR::HostFile file;
    nn::Result res = file.Open(filename, nn::hio::CTR::HostFile::ACCESS_MODE_WRITE,
                                         nn::hio::CTR::HostFile::OPEN_DISP_CREATE_ALWAYS);
    if( res.IsSuccess() )
    {
        struct Inner
        {
            static void OutputReturnCode( nn::hio::CTR::HostFile& file )
            {
                const u16 Ret = u16(0x0A << 8 | 0x0D);
                file.Write(&Ret, 2);
            }
        };

        // output error
        size_t size = 0;
        if( m_Error.Size() > 0 && type == ERROR )
        {
            for(u32 i=0; i<m_Error.Size(); ++i)
            {
                const char* error = GetError(i)->m_Message.Get();
                file.Write(&size, error, strlen( error ));
                Inner::OutputReturnCode( file );
            }
            Inner::OutputReturnCode( file );
        }

        // output warning
        if( m_Warning.Size() > 0 && type == WARNING )
        {
            for(u32 i=0; i<m_Warning.Size(); ++i)
            {
                const char* warning = GetWarning(i)->m_Message.Get();
                file.Write(&size, warning, strlen( warning ));
                Inner::OutputReturnCode( file );
            }
            Inner::OutputReturnCode( file );
        }

        // output system log
        if( m_SystemLog.Size() > 0 && type == SYSTEM_LOG )
        {
            for(u32 i=0; i<m_SystemLog.Size(); ++i)
            {
                const char* system_log = GetSystemLog(i)->m_Message.Get();
                file.Write(&size, system_log, strlen( system_log ));
                Inner::OutputReturnCode( file );
            }
            Inner::OutputReturnCode( file );
        }

        // output history
        if( m_Log.Size() > 0 && type == HISTORY )
        {
            for(u32 i=0; i<m_Log.Size(); ++i)
            {
                const char* log = GetLog(i)->m_Message.Get();
                file.Write(&size, log, strlen( log ));
                Inner::OutputReturnCode( file );
            }
            Inner::OutputReturnCode( file );
        }

        file.Close();
    }
}

}   /* namespace system */
}   /* namespace vw     */
