/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system/Module.h>

namespace vw     {
namespace system {

//NW4R_UT_RUNTIME_TYPEINFO_ROOT_DEFINITION( Module );

Module* Module::sTopModule        = NULL;
Module* Module::sFirstSetupModule = NULL;

/*---------------------------------------------------------------------------*
  Name:         Module::GetModuleById

  Description:  Gets the instance to the module having the specified ID.

  Arguments:    id          Module ID

  Returns:      Returns the instance to the module having the specified ID.
 *---------------------------------------------------------------------------*/
Module*
Module::GetModuleById( u16 id )
{
    return vw::system::App::GetModuleInstance(id);
/*
    Module* module = GetTopModule();

    while ( module != NULL )
    {
        if ( module->GetId() == id )
        {
            return module;
        }
        module = module->GetNext();
    }
    return NULL;*/
}



/*---------------------------------------------------------------------------*
  Name:         Module::Module

  Description:  Constructor

  Arguments:    id
                priority

  Returns:      None.
 *---------------------------------------------------------------------------*/
Module::Module( u16 id, u32 priority, u32 setupOrder /* = DEFAULT_SETUP_ORDER */ )
 : /*mHeap ( id ),
   mVWPad( id ),*/
   mId   ( id ),
   mPriority( priority ),
   mSetupOrder( setupOrder ),
   mNext ( NULL ),
   mNextSetup ( NULL ),
   mPostCalcClbk( id ),
   mPostDrawClbk( id )
{
    Insert();
}


/*---------------------------------------------------------------------------*
  Name:         Module::~Module

  Description:  Destructor

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
Module::~Module()
{
    Remove();
}


/*---------------------------------------------------------------------------*
  Name:         Module::Insert

  Description:  Registers self in module list.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::Insert( void )
{
    this->InsertToList_();
    this->InsertToSetupList_();
}


/*---------------------------------------------------------------------------*
  Name:         Module::InsertToList_

  Description:  Registers module in list.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::InsertToList_()
{
    // Register to the module list
    Module* module = Module::GetTopModule();
    if ( module == NULL || module->mPriority > this->mPriority )
    {
        // When inserted at the start of the list
        sTopModule  = this;
        this->mNext = module;
        return;
    }

    while ( true )
    {
        Module* next = module->GetNext();
        if ( next == NULL || next->mPriority > this->mPriority )
        {
            this->mNext   = next;
            module->mNext = this;
            return;
        }
        module = next;
    }
}

/*---------------------------------------------------------------------------*
  Name:         Module::InsertToSetupList_

  Description:  Registers module in initialization order list.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::InsertToSetupList_()
{
    // Register to the initialization order list
    Module* module = Module::GetFirstSetupModule();
    if ( module == NULL || module->mSetupOrder > this->mSetupOrder )
    {
        sFirstSetupModule = this;
        this->mNextSetup  = module;
        return;
    }

    while ( true )
    {
        Module* next = module->GetNextSetup();
        if ( next == NULL || next->mSetupOrder > this->mSetupOrder )
        {
            this->mNextSetup   = next;
            module->mNextSetup = this;
            return;
        }
        module = next;
    }
}



/*---------------------------------------------------------------------------*
  Name:         Module::Remove

  Description:  Removes self from module list.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::Remove( void )
{
    Module* module = Module::GetTopModule();
    if ( module == this )
    {
        sTopModule = this->mNext;
        return;
    }

    while ( module != NULL )
    {
        if ( module->mNext == this )
        {
            module->mNext = this->mNext;
            return;
        }
        module = module->GetNext();
    }
}


/*---------------------------------------------------------------------------*
  Name:         Module::ChangePriority

  Description:  Changes module priority.

  Arguments:    priority

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::ChangePriority( u32 priority )
{
    Remove();
    this->mPriority = priority;
    this->mNext = NULL;
    Insert();
}


/*---------------------------------------------------------------------------*
  Name:         Module::ClbkMgr::Set

  Description:  Sets callback function for Calc and Draw.

  Arguments:    id
                callback
                arg

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::ClbkMgr::Set( u32 id, Func func, void* arg )
{
    NW_ASSERT( id < MODULE_NUM );
    if ( mInfo[ id ].mFunc )
    {
        --mNum;
    }

    mInfo[ id ].mFunc = func;
    mInfo[ id ].mArg  = arg;

    if ( func )
    {
        ++mNum;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Module::ClbkMgr::Rise

  Description:  Issues the set callback.

  Arguments:    None.

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::ClbkMgr::Rise()
{
    if ( mNum == 0 )
    {
        return;
    }

    for ( u32 i = 0; i < MODULE_NUM; i++ )
    {
        if ( mInfo[ i ].mFunc )
        {
            mInfo[ i ].mFunc( mId, mInfo[ i ].mArg );
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         Module::ClbkMgr::Rise

  Description:  Issues the set callback from the specific target.

  Arguments:    targetID Target ID

  Returns:      None.
 *---------------------------------------------------------------------------*/
void
Module::ClbkMgr::Rise( u32 targetID )
{
    if ( targetID >= MODULE_NUM )
    {
        return;
    }

    if ( mInfo[ targetID ].mFunc )
    {
        mInfo[ targetID ].mFunc( mId, mInfo[ targetID ].mArg );
    }
}


}   /* namespace system */
}   /* namespace viewer */




