/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
//#include "WindowManager.h"

namespace vw     {
namespace system {

template<typename T> static void RightAngleFilter_( T *val_x, T* val_y );

static const u32 PAD_BUTTON_PLUS_MASK = (PAD_BUTTON_LEFT | PAD_BUTTON_RIGHT | PAD_BUTTON_UP | PAD_BUTTON_DOWN);



static nn::hid::CTR::PadStatus      sPadCtr;
static nn::hid::CTR::DebugPadStatus sDebugPad;
static PADInfo                      sPadBuf; // Internal use only
static PADInfo                      sPad;
static bool                         sForceRightAngle = false;
static u32                          sPadMask = 0xFFFFFFFF;

static struct
{
    u32     press;
    u32     press_plus;
    u32     last_pressed_plus; // The last pressed +Control Pad key
    u32     prev;
    u32     trig;
    u32     release;
    u32     repeat;
    u32     last_repeated_plus; // The last repeated +Control Pad key
    u32     repeatAccel;        // accelerated repeat
    u32     last_repeated_accel_plus; // The last repeated +Control Pad key
    u32     click;
    u32     click_ignore;
    u32     hold;
    f32     stickX;
    f32     stickY;
}
sPadStatus;

const u16 BUTTON_NUM   = 32;

static struct {
    u8   hold[ BUTTON_NUM ];
    u8   hold_dbl[ BUTTON_NUM ];
} sClickState;


static void Pad_CopyStatus_( PADInfo* dstPad, const PADInfo* srcPad );

/*---------------------------------------------------------------------------*
  Name:         Pad_IsEnable
  Description:  Function to get whether the current pad is enabled for the specified appId
  Arguments:    appId   Module ID
  Returns:      if pad operations are enabled, true; if operation focus is on another module or the system, false.
                
 *---------------------------------------------------------------------------


*/
bool
Pad_IsEnable( u16 appId )
{
    if ( appId == SYSTEM_ID )
    {
        return true;
    }

    //if ( App::GetInstance().HasFocus( appId )  (! GetWindowManager()-GetNextWindow( NULL ) ) )
    if ( App::GetInstance().HasFocus( appId ) )
    {
        return true;
    }
    return false;
}


/*---------------------------------------------------------------------------*
  Name:         Pad_Init
  Description:  Initializes PAD state.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_Init( void )
{
    //(void)PADInit();
    //(void)WPad_Init();
    //
    Pad_Read();
    Pad_Read();
    sPadStatus.click_ignore             = 0;
    sPadStatus.last_pressed_plus        = 0;
    sPadStatus.last_repeated_plus       = 0;
    sPadStatus.last_repeated_accel_plus = 0;
}

/*---------------------------------------------------------------------------*
  Name:         Pad_ReadDegitalTrigger_
  Description:  Converts Control Stick input to ON/OFF two-value input.
  Arguments:    pad         Pad state
                prev_press  Previous input value
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Pad_ReadDegitalTrigger_( PADInfo* pad, u32 prev_press )
{
    const f32 STICK_THRESHOLD_UPPER = 32.0f / 128.0f;
    const f32 STICK_THRESHOLD_LOWER = 20.0f / 128.0f;

    // Check if stick is on or off
    f32 threshold;
    // Change the on/off threshold value with the previous stick state
    // LEFT
    threshold = ( prev_press & PAD_STICK_LEFT ) ? STICK_THRESHOLD_LOWER : STICK_THRESHOLD_UPPER;
    if ( pad->stickX < -threshold )
    {
        pad->button |= PAD_STICK_LEFT;
    }
    // RIGHT
    threshold = ( prev_press & PAD_STICK_RIGHT ) ? STICK_THRESHOLD_LOWER : STICK_THRESHOLD_UPPER;
    if ( pad->stickX > threshold )
    {
        pad->button |= PAD_STICK_RIGHT;
    }
    // DOWN
    threshold = ( prev_press & PAD_STICK_DOWN ) ? STICK_THRESHOLD_LOWER : STICK_THRESHOLD_UPPER;
    if ( pad->stickY < -threshold )
    {
        pad->button |= PAD_STICK_DOWN;
    }
    // UP
    threshold = ( prev_press & PAD_STICK_UP ) ? STICK_THRESHOLD_LOWER : STICK_THRESHOLD_UPPER;
    if ( pad->stickY > threshold )
    {
        pad->button |= PAD_STICK_UP;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_UpdateLastRepeatKey_
  Description:  Performs anti-chattering processing when +Control Pad keys are pressed diagonally.
  Arguments:    last_repeat_plus -   The +Control Pad key for which the Repeat flag was last set
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static u32
Pad_UpdateLastRepeatKey_( u32 last_repeated_plus )
{
    // Handling diagonal presses of the +Control Pad keys
    // No new REPEAT flag will be set for any +Control Pad key until the key for which a REPEAT flag was last set is released
    if ( last_repeated_plus )
    {
        // When the last pressed +Control Pad key is released, release its state.
        if ( ! (sPadStatus.press & last_repeated_plus) )
        {
            last_repeated_plus = 0;
        }
    }

    if ( ! last_repeated_plus )
    {
        // When a new button is pressed, register its state.
        if ( sPadStatus.press & PAD_BUTTON_LEFT )
        {
            last_repeated_plus = PAD_BUTTON_LEFT;
        }
        else if ( sPadStatus.press & PAD_BUTTON_RIGHT )
        {
            last_repeated_plus = PAD_BUTTON_RIGHT;
        }
        else if ( sPadStatus.press & PAD_BUTTON_UP )
        {
            last_repeated_plus = PAD_BUTTON_UP;
        }
        else if ( sPadStatus.press & PAD_BUTTON_DOWN )
        {
            last_repeated_plus = PAD_BUTTON_DOWN;
        }
    }
    return last_repeated_plus;
}


/*---------------------------------------------------------------------------*
  Name:         Pad_ReadRepeat_
  Description:  Performs a key repeat count.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Pad_ReadRepeat_( void )
{
    const u16 REPEAT_START = 25;    // Number of frames until key repeat starts
    const u16 REPEAT_SPAN  = 8;     // Number of frames between key repeats
    static u16      repeat_count[ BUTTON_NUM ];
    u16             i;

    sPadStatus.repeat = sPadStatus.trig;

    // Key repeat calculations
    for ( i = 0; i < BUTTON_NUM; i++ )
    {
        if ( sPadStatus.press & ( 1 << i ) )
        {
            if ( (--repeat_count[ i ]) == 0 )
            {
                sPadStatus.repeat |= ( 1 << i );
                repeat_count[ i ] = REPEAT_SPAN;
            }
        }
        else
        {
            repeat_count[ i ] = REPEAT_START;
        }
    }

    // Update the information used for handling diagonal presses on the +Control Pad
    sPadStatus.last_repeated_plus = Pad_UpdateLastRepeatKey_( sPadStatus.last_repeated_plus );

    if ( sPadStatus.last_repeated_plus )
    {
        // Mask other +Control Pad keys
        sPadStatus.repeat &= ~(PAD_BUTTON_PLUS_MASK & ~sPadStatus.last_repeated_plus);
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_ReadRepeatAccel_
  Description:  Performs an accelerated key repeat count.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Pad_ReadRepeatAccel_( void )
{
    const s32 REPEAT_START = 25;    // Number of frames until key repeat starts
    const s32 REPEAT_SPAN  = 8;     // Number of frames between key repeats
    static s32      repeat_count[ BUTTON_NUM ];
    static s32      repeat_trig_count[ BUTTON_NUM ];
    u16             i;

    sPadStatus.repeatAccel = sPadStatus.trig;

    // Key repeat calculations
    for ( i = 0; i < BUTTON_NUM; i++ )
    {
        if ( sPadStatus.press & ( 1 << i ) )
        {
            if ( (--repeat_count[ i ]) == 0 )
            {
                sPadStatus.repeatAccel |= ( 1 << i );
                repeat_count[ i ] = nw::ut::Max<s32>( REPEAT_SPAN - (repeat_trig_count[ i ] >> 1), 1L );
            }
        }
        else
        {
            repeat_count[ i ] = REPEAT_START;
        }

        // Key repeat count
        if ( sPadStatus.repeatAccel & ( 1 << i ) )
        {
            ++repeat_trig_count[ i ];
        }

        if ( ! (sPadStatus.press & ( 1 << i ))  )
        {
            repeat_trig_count[ i ] = 0;
        }
    }

    // Update the information used for handling diagonal presses on the +Control Pad
    sPadStatus.last_repeated_accel_plus = Pad_UpdateLastRepeatKey_( sPadStatus.last_repeated_accel_plus );

    if ( sPadStatus.last_repeated_accel_plus )
    {
        // Mask other +Control Pad keys
        sPadStatus.repeatAccel &= ~(PAD_BUTTON_PLUS_MASK & ~sPadStatus.last_repeated_accel_plus);
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_ReadClick_
  Description:  Performs a count of click and hold determinations.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Pad_ReadClick_( void )
{
    const  s32  HOLD_COUNT   = 90;    // Number of frames until key repeat starts
    const  s32  CLICK_COUNT  = 20;    // Number of frames between key repeats

    sPadStatus.click = 0;
    sPadStatus.hold  = 0;

    // All click counts are invalid when there is input from the stick
    if ( sPadStatus.stickX > 0 || sPadStatus.stickY > 0 )
    {
        sPadStatus.click_ignore = 0xFFFF;
    }

    for ( s32 i = 0; i < BUTTON_NUM; i++ )
    {
        u16 i_mask = u16( 1 << i );

        if ( sPadStatus.press & i_mask )
        {
            ++sClickState.hold[ i ];

            if ( sClickState.hold[ i ] == HOLD_COUNT )
            {
                sPadStatus.hold |= i_mask;
            }
        }
        else
        {
            s32 count = sClickState.hold[ i ];
            sClickState.hold[ i ] = 0;

            if ( count > 0           &&
                 count < CLICK_COUNT &&
                 (! (sPadStatus.click_ignore & i_mask) ) )
            {
                sPadStatus.click |= i_mask;
            }

            sPadStatus.click_ignore &= ~i_mask;
        }
    }
}



/*---------------------------------------------------------------------------*
  Name:         Pad_Read_
  Description:  Debug pad input
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_Read_(PADInfo* padInfo)
{
    static nn::hid::CTR::PadReader padReader;
    padReader.SetStickClampMode(nn::hid::CTR::PadReader::STICK_CLAMP_MODE_CROSS);
    padReader.ReadLatest(&sPadCtr);

    padInfo->button = sPadCtr.hold;
    padInfo->stickX = padReader.NormalizeStick(sPadCtr.stick.x);
    padInfo->stickY = padReader.NormalizeStick(sPadCtr.stick.y);
}

/*---------------------------------------------------------------------------*
  Name:         DPad_Read_
  Description:  Debug pad input
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
DPad_Read_(nn::hid::CTR::DebugPadStatus* pad)
{
    static nn::hid::CTR::DebugPadReader debugPadReader;
    debugPadReader.ReadLatest(pad);
}

/*---------------------------------------------------------------------------*
  Name:         Merge_Flag_
  Description:  Merges the debug pad flag and the pad flag
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Merge_Flag_(bit32& pad_flag, bit16 debug_flag)
{
    using namespace nn::hid::CTR;
    if ( debug_flag & DEBUG_PAD_BUTTON_LEFT )   { pad_flag |= BUTTON_LEFT; }
    if ( debug_flag & DEBUG_PAD_BUTTON_RIGHT )  { pad_flag |= BUTTON_RIGHT; }
    if ( debug_flag & DEBUG_PAD_BUTTON_UP )     { pad_flag |= BUTTON_UP; }
    if ( debug_flag & DEBUG_PAD_BUTTON_DOWN )   { pad_flag |= BUTTON_DOWN; }
    if ( debug_flag & DEBUG_PAD_BUTTON_A )      { pad_flag |= BUTTON_A; }
    if ( debug_flag & DEBUG_PAD_BUTTON_B )      { pad_flag |= BUTTON_B; }
    if ( debug_flag & DEBUG_PAD_BUTTON_X )      { pad_flag |= BUTTON_X; }
    if ( debug_flag & DEBUG_PAD_BUTTON_Y )      { pad_flag |= BUTTON_Y; }
    if ( debug_flag & DEBUG_PAD_TRIGGER_R )     { pad_flag |= BUTTON_R; }
    if ( debug_flag & DEBUG_PAD_TRIGGER_L )     { pad_flag |= BUTTON_L; }
    if ( debug_flag & DEBUG_PAD_BUTTON_PLUS )   { pad_flag |= BUTTON_START; }
    if ( debug_flag & DEBUG_PAD_BUTTON_MINUS )  { pad_flag |= BUTTON_START; }
}


/*---------------------------------------------------------------------------*
  Name:         DPad_MergeToPad_
  Description:  Merges the debug pad input to the pad
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
DPad_MergeToPad_( PADInfo* pad,
                  nn::hid::CTR::DebugPadStatus* dpad )
{
    Merge_Flag_(pad->button, dpad->hold);

    if( nw::ut::Abs(pad->stickX) < 0.000001f )
    {
        pad->stickX = dpad->leftStickX;
    }
    if( nw::ut::Abs(pad->stickY) < 0.000001f )
    {
        pad->stickY = dpad->leftStickY;
    }
    if( nw::ut::Abs(pad->stickX) < 0.000001f )
    {
        pad->stickX = dpad->rightStickX;
    }
    if( nw::ut::Abs(pad->stickY) < 0.000001f )
    {
        pad->stickY = dpad->rightStickY;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_Read
  Description:  Gets the key information.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_Read( void )
{
    Pad_Read_(&sPadBuf);

    Pad_CopyStatus_( &sPad, &sPadBuf );

    DPad_Read_(&sDebugPad);
    DPad_MergeToPad_( &sPad, &sDebugPad );

    // Digital input of the control stick
    Pad_ReadDegitalTrigger_( &sPad, sPadStatus.press );

    sPadStatus.prev    = sPadStatus.press;
    sPadStatus.trig    = static_cast<u32>( (sPadStatus.press ^ sPad.button) & sPad.button );
    sPadStatus.release = static_cast<u32>( (sPadStatus.press ^ sPad.button) & sPadStatus.press );
    sPadStatus.press   = sPad.button;
    sPadStatus.stickX    = sPad.stickX;
    sPadStatus.stickY    = sPad.stickY;

    // Update the information used for handling diagonal presses on the +Control Pad
    sPadStatus.last_pressed_plus = Pad_UpdateLastRepeatKey_( sPadStatus.last_pressed_plus );

    if ( sPadStatus.last_pressed_plus )
    {
        // Mask other +Control Pad keys
        sPadStatus.press_plus = sPadStatus.press & sPadStatus.last_pressed_plus & PAD_BUTTON_PLUS_MASK;
    }
    else
    {
        sPadStatus.press_plus = 0;
    }

    // Make the Repeat determination
    Pad_ReadRepeat_();

    // Accelerated repeat determination
    Pad_ReadRepeatAccel_();

    // Click, hold determination
    Pad_ReadClick_();
}


/*---------------------------------------------------------------------------*
  Name:         Pad_GetKeyInfo
  Description:  Gets key information as the PADInfo structure.
  Arguments:    appId   Module ID
                keyInfo     Region for getting the key information
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_GetKeyInfo( u16 appId, PADInfo* keyInfo )
{
    const PADInfo PAD_INFO_NULL =
    {
        0, /* button */
        0.0f, /* stickX */
        0.0f, /* stickY */
        0, /* press */
        0, /* trig */
        0, /* Release  */
        0, /* repeat */
        0, /* repeatAccel */
        1, /* disabled */
    };
    if ( Pad_IsEnable( appId ) )
    {
        keyInfo->button      = sPadStatus.press;
        keyInfo->stickX      = sPadStatus.stickX;
        keyInfo->stickY      = sPadStatus.stickY;
        keyInfo->press       = sPadStatus.press;
        keyInfo->trig        = sPadStatus.trig;
        keyInfo->release     = sPadStatus.release;
        keyInfo->repeat      = sPadStatus.repeat;
        keyInfo->repeatAccel = sPadStatus.repeatAccel;
        keyInfo->disabled    = 0;

        RightAngleFilter_( &keyInfo->stickX,    &keyInfo->stickY    );
    }
    else
    {
        *keyInfo = PAD_INFO_NULL;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_IsPressed
  Description:  Determines whether the pad is being pressed.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key currently being pressed and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsPressed( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId ) )
    {
        return static_cast<u32>( button & sPadStatus.press );
    }
    else
    {
        return 0;
    }
}

/*---------------------------------------------------------------------------*
  Name:         Pad_IsPressedHold
  Description:  Determines whether the pad is being pressed. (Diagonal is invalid)
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key currently being pressed and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsPressedHold( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId ) )
    {
        return static_cast<u32>( button & sPadStatus.press_plus );
    }
    else
    {
        return 0;
    }
}

/*---------------------------------------------------------------------------*
  Name:         Pad_IsTriggered
  Description:  Determines whether the pad trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current trigger key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsTriggered( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId )  )
    {
        return static_cast<u32>( button & sPadStatus.trig );
    }
    else
    {
        return 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_IsReleased
  Description:  Determines whether the pad up trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current up trigger key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsReleased( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId )  )
    {
        return static_cast<u32>( button & sPadStatus.release );
    }
    else
    {
        return 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_IsRepeated
  Description:  Determines whether the pad repeat trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current repeat trigger key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsRepeated( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId )  )
    {
        return static_cast<u32>( button & sPadStatus.repeat );
    }
    else
    {
        return 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_IsClicked
  Description:  Determines pad single clicks.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key for which the click has been determined and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsClicked( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId ) )
    {
        return u32( button & sPadStatus.click );
    }
    else
    {
        return 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_IsLongPressed
  Description:  Determines pad long presses.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key for which the long press has been determined and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsLongPressed( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId ) )
    {
        return u32( button & sPadStatus.hold );
    }
    else
    {
        return 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_IsRepeatedAccel
  Description:  Determines whether the pad accelerated repeat trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current repeat key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
u32
Pad_IsRepeatedAccel( u16 appId, u32 button )
{
    if ( Pad_IsEnable( appId )  )
    {
        return static_cast<u32>( button & sPadStatus.repeatAccel );
    }
    else
    {
        return 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_GetStickXY
  Description:  Gets the main stick analog value.
  Arguments:    appId   Module ID
                stickX  Value of the X-component for the main stick
                stickY  Value of the Y-component for the main stick
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_GetStickXY( u16 appId, f32* stickX, f32* stickY )
{
    if ( Pad_IsEnable( appId )  )
    {
        *stickX = sPadStatus.stickX;
        *stickY = sPadStatus.stickY;

        RightAngleFilter_( stickX, stickY );
    }
    else
    {
        *stickX = 0;
        *stickY = 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_SetForceRightAngle
  Description:  Sets an option to forcibly convert to 4-direction
  Arguments:    force
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_SetForceRightAngle( bool force )
{
    sForceRightAngle = force;
}


/*---------------------------------------------------------------------------*
  Name:         Pad_GetForceRightAngle
  Description:  Gets the option settings to forcibly convert to 4-direction
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
Pad_GetForceRightAngle()
{
    return sForceRightAngle;
}

/*---------------------------------------------------------------------------*
  Name:         RightAngleFilter_
  Description:  Filter for control stick value
  Arguments:    val_x
                val_y
  Returns:      None.
 *---------------------------------------------------------------------------


*/
template<typename T>
static void
RightAngleFilter_( T *val_x, T* val_y )
{
    if ( ! sForceRightAngle )
    {
        return;
    }

    if ( nw::ut::Abs( *val_x ) >= nw::ut::Abs( *val_y ) )
    {
        *val_y = 0;
    }
    else
    {
        *val_x = 0;
    }
}

/*---------------------------------------------------------------------------*
  Name:         Pad_ClearState
  Description:  Clears state
  Arguments:    appId
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_ClearState( u16 appId )
{
    if ( Pad_IsEnable( appId )  )
    {
        sPadStatus.trig = 0;
        sPadStatus.click = 0;

        for ( s32 i = 0; i < BUTTON_NUM; i++ )
        {
            sClickState.hold[ i ] = 0;
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         Pad_GetRawInfo_
  Description:
  Arguments:    pad
                port
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
Pad_GetRawPADInfo_( nn::hid::CTR::PadStatus* pad )
{
    if ( ! pad )
    {
        return false;
    }

    std::memcpy( pad, &sPadCtr, sizeof( nn::hid::CTR::PadStatus ) );

    return true;
}

/*---------------------------------------------------------------------------*
  Name:         Pad_GetRawInfo_
  Description:
  Arguments:    pad
                port
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
Pad_GetRawDEBUGPADInfo_( nn::hid::CTR::DebugPadStatus* dpad )
{
    if ( ! dpad )
    {
        return false;
    }

    std::memcpy( dpad, &sDebugPad, sizeof( nn::hid::CTR::DebugPadStatus ) );

    return true;
}


/*---------------------------------------------------------------------------*
  Name:         Pad_SetButtonMask
  Description:  Mask settings
  Arguments:    mask
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Pad_SetButtonMask( u32 mask )
{
    sPadMask = mask;
}


/*---------------------------------------------------------------------------*
  Name:         Pad_CopyStatus_
  Description:
  Arguments:    dstPad
                srcPad
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
Pad_CopyStatus_( PADInfo* dstPad, const PADInfo* srcPad )
{
    NW_NULL_ASSERT( dstPad );
    NW_NULL_ASSERT( srcPad );

    std::memcpy( dstPad, srcPad, sizeof(PADInfo) );

    // Apply mask when mask is set
    dstPad->button &= sPadMask;

    if ( ! (sPadMask & PAD_BUTTONMASK_STICK) )
    {
        dstPad->stickX = 0.0f;
        dstPad->stickY = 0.0f;
    }
}



}   /* namespace system */
}   /* namespace vw     */


