/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>
#include "../include/PerformDrawer.h"

namespace vw     {
namespace system {



/*===========================================================================*
  Name:         PerformMgr

  Description:  Performance management
 *===========================================================================*/
PerformMgr* PerformMgr::m_pInstance = NULL;
PerformMgr* PerformMgr::GetInstance()
{
    NW_ASSERT( m_pInstance != NULL );
    return m_pInstance;
}

void
PerformMgr::Initialize()
{
    if (m_pInstance != NULL) return;
    void *buf = Memory_AllocSys(sizeof(PerformMgr));
    if (buf == NULL) return;

    m_pInstance = new(buf) PerformMgr();
}

void
PerformMgr::Finalize()
{
    if (m_pInstance == NULL) return;

    m_pInstance->~PerformMgr();
    Memory_Free(m_pInstance);

    m_pInstance = NULL;
}

PerformMgr::PerformMgr()
{
    mCount = 0;
    mMeasureCount = 60;
    mSystemTime = 0.0f;
    mCalcTmp = 0.0f;
    mDrawTmp = 0.0f;
    mCpuTmp = 0.0f;
    mGpuTmp = 0.0f;
}

PerformMgr::~PerformMgr()
{
}

// Get GPU process time
f32 PerformMgr::GetGpuTime() const
{
    return mGpuCounter.GetFrameTime();
}
// Get CPU process time
f32 PerformMgr::GetCpuTime() const
{
    return mCpuCounter.GetFrameTime();
}
// Get Calc process time
f32 PerformMgr::GetCalcTime() const
{
    return mCalcCounter.GetFrameTime();
}
// Get Draw process time
f32 PerformMgr::GetDrawTime() const
{
    return mDrawCounter.GetFrameTime();
}

// End 1 frame measurement
void
PerformMgr::EndFrame()
{
    // Add measurement data to measurement counter
    mCalcTmp += mCalc.GetCumulativeTime();
    mDrawTmp += mDraw.GetCumulativeTime();
    mCpuTmp += mCalc.GetCumulativeTime() + mDraw.GetCumulativeTime();
    if( CommandListSwapper::GetInstance() )
        mGpuTmp += CommandListSwapper::GetInstance()->GetGpuProfilingTotalCostTime();

    bool IsUpdateScene = vw::system::App::GetInstance().IsUpdateScene();
    if(IsUpdateScene)
    {
        const u32 updateRate = vw::system::System_GetSetting()->updateRate;
        mCalcCounter.AddTime( mCalcTmp / updateRate );
        mDrawCounter.AddTime( mDrawTmp / updateRate );
        mCpuCounter.AddTime( mCpuTmp / updateRate );
        mGpuCounter.AddTime( mGpuTmp / updateRate );
        mCalcTmp = 0.0f;
        mDrawTmp = 0.0f;
        mCpuTmp = 0.0f;
        if( vw::system::App::GetInstance().IsRenderScene() ) mGpuTmp = 0.0f;
    }

    // Reset frame measurement object
    mCalc.Reset();
    mDraw.Reset();

    // Calculate measurement time if specified number of measurements have been made
    if(++mCount >= mMeasureCount)
    {
        CalcTime();
    }
}

// Set calculation interval for measurement time
void
PerformMgr::SetMeasureCount(u32 count)
{
    mMeasureCount = count;
}

// Start 1 frame process time measurement
void
PerformMgr::SetTimerStart()
{
    mSystem.Reset();
    mSystem.Begin();
}

// End 1 frame process time measurement
void
PerformMgr::SetTimerEnd()
{
    mSystem.End();
    mSystemTime = mSystem.GetCumulativeTime();
}

// Get 1 frame process time
f32
PerformMgr::GetSystemTime()
{
    return mSystemTime;
}

// Measurement time calculation
void
PerformMgr::CalcTime()
{
    // Update of measurement counter
    mCalcCounter.CalcTime();
    mDrawCounter.CalcTime();
    mCpuCounter.CalcTime();
    mGpuCounter.CalcTime();

    // Counter reset
    mCount = 0;
}

}   /* namespace system */
}   /* namespace vw     */
