/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include "../include/TgaReaderImpl.h"
#include "../include/TgaData.h"
#include "../include/TexUtil.h"

namespace vw     {
namespace system {

/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::TgaReaderImpl
  Description:  Constructor
  Arguments:    data   Specifies TGA file data.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* ctor */
TgaReaderImpl::TgaReaderImpl( void* data )
 : ImageReaderImpl(),
   mBuf( static_cast<u8*>( data ) ),
   mBody( NULL ),
   mPltt( NULL ),
   mWidth( 0 ),
   mHeight( 0 ),
   mIsValid( true ),
   mCompress( false ),
   mBpp( 0 ),
   mPlttBpp( 0 ),
   mOrigin( ORIGIN_BL ),
   mFormat( TgaData::FMT_NONE )
{
    mIsValid = CheckValid_();

    if ( ! mIsValid )
    {
        //NW_WARNING( false, RESTXT_ON_UNSUPPORTED_IMAGE );
        return;
    }

    TgaData::Header *hd = reinterpret_cast<TgaData::Header*>( mBuf );
    mFormat = static_cast<TgaData::Format>( hd->bfFormat );

    if ( mFormat >= TgaData::FMT_CMP_CI8 )
    {
        mCompress = true;
    }

    mWidth   = LEndian::ReadU16( &hd->bfWidth  );
    mHeight  = LEndian::ReadU16( &hd->bfHeight );
    mBpp     = hd->bfBpp;
    mOrigin  = u32( (hd->bfDescript >> 4) & 0x3 );
    mPlttBpp = u32( hd->bfPlttSize / 8 );

    if ( mWidth > 1024 || mHeight > 1024 )
    {
        mIsValid = false;
        //NW_WARNING( false, RESTXT_ON_TOO_LARGE_IMAGE_dd, mWidth, mHeight );
    }

    u32 offset = sizeof(TgaData::Header) + hd->bfIdLength;
    u32 plttSize = u32( LEndian::ReadU16( &hd->bfPlttLength ) * mPlttBpp );

    if ( hd->bfType == 1 )
    {
        mPltt = static_cast<u8*>( nw::ut::AddOffsetToPtr( mBuf, offset ) );
    }

    mBody   = static_cast<u8*>( nw::ut::AddOffsetToPtr( mBuf, offset + plttSize ) );
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::~TgaReaderImpl
  Description:  Destructor
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
/* dtor */
TgaReaderImpl::~TgaReaderImpl()
{
    this->Dispose();
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::IsValid
  Description:  Determines if data is valid
  Arguments:    None.
  Returns:      If data is  valid, TRUE
                Returns FALSE if the data is invalid.
 *---------------------------------------------------------------------------


*/
bool
TgaReaderImpl::IsValid()
{
    return mIsValid;
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::GetWidth
  Description:  Gets the image horizontal size.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
u32
TgaReaderImpl::GetWidth()
{
    return mWidth;
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::GetHeight
  Description:  Gets the image vertical size.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
u32
TgaReaderImpl::GetHeight()
{
    return mHeight;
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::Expand
  Description:  Expands TGA data to RGBA8 texture
  Arguments:    expBuf      Buffer to expand to
                bufWidth    Buffer horizontal size
                bufHeight   Buffer vertical size
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
TgaReaderImpl::Expand( void* expBuf, u32 bufWidth, u32 bufHeight )
{
    u8* body = mBody;
    u8* uncompBuf = NULL;
    u32 bpp  = mBpp / 8;
    u32 cidx;
    u32 shift = 0;
    u32 mask  = 0xFF;

    if ( mCompress )
    // Expands RLE compressed files
    {
        uncompBuf = static_cast<u8*>( Memory_AllocSys( mWidth * mHeight * mBpp / 8 ) );

        this->Uncomp_( uncompBuf );
        body = uncompBuf;
    }

    if ( mBpp < 8 )
    {
        mask = u32( (~(0x100 - (1 << mBpp))) & 0xFF );
    }

    u8* pOut = static_cast<u8*>( expBuf );

    for ( u16 y = 0; y < mHeight; y++ )
    {
        for ( u16 x = 0; x < mWidth; x++ )
        {
            u8 a, b, g, r;
            if ( mPltt )
            {
                cidx = ( *body >> (8 - mBpp - shift) ) & mask;
                u8* pCol = &mPltt[ cidx * mPlttBpp ];
                a = (mPlttBpp == 4)? pCol[ 3 ] : u8(0xFF);
                b = pCol[ 0 ];
                g = pCol[ 1 ];
                r = pCol[ 2 ];
                shift += mBpp;
                if ( shift >= 8 )
                {
                    shift -= 8;
                    ++body;
                }
            }
            else
            {
                a = 0xFF;
                b = *body++;
                g = *body++;
                r = *body++;
                if ( bpp == 4 )
                {
                    a = *body++;
                }
            }

            // u16 ar  = u16( (a << 8) + r );
            // u16 gb  = u16( (g << 8) + b );
            // u32 col = u32( (ar << 16) + gb );

            u16 tx = ( mOrigin & ORIGIN_R )? u16(mWidth  - 1 - x) : x; // Reflects the right/left inversion flag
            u16 ty = ( mOrigin & ORIGIN_T )? u16(mHeight - 1 - y) : y; // Reflects the up/down inversion flag
            u16 line = u16(mHeight - 1 - ty);

            nw::ut::Color8 color(r,g,b,a);
            UtilRGBA8::SetColor( color, pOut, tx, line, bufWidth, bufHeight );
        }
    }

    if ( uncompBuf )
    {
        Memory_Free( uncompBuf );
    }

    return true;
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::Uncomp_
  Description:  RL expansion processing
  Arguments:    pOut    Buffer to expand to
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaReaderImpl::Uncomp_( u8* pOut )
{
    u8* pIn = mBody;
    u32 size = mWidth * mHeight * mBpp / 8;
    u32 count = 0;
    u32 bpp = mBpp / 8;

    while ( count < size )
    {
        u8 len = *pIn++;
        if ( len & 0x80 )
        // compressed data
        {
            len &= ~0x80;
            len += 1;
            while ( len > 0 )
            {
                for ( u32 i = 0; i < bpp; i++ )
                {
                    *pOut++ = *(pIn + i);
                }
                len--;
                count += bpp;
            }
            pIn += bpp;
        }
        else
        // Uncompressed data
        {
            len += 1;
            while ( len > 0 )
            {
                for ( u32 i = 0; i < bpp; i++ )
                {
                    *pOut++ = *pIn++;
                }
                len--;
                count += bpp;
            }
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::GetType
  Description:  Gets the data type.
  Arguments:    None.
  Returns:      Data type.
 *---------------------------------------------------------------------------


*/
u32
TgaReaderImpl::GetType()
{
    return TYPE_TGA;
}


/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::CheckValid_
  Description:  Determines if valid as TGA data.
  Arguments:    None.
  Returns:      If valid data, TRUE
                Returns FALSE if the data is invalid.
 *---------------------------------------------------------------------------


*/
bool
TgaReaderImpl::CheckValid_()
{
    TgaData::Header *hd = reinterpret_cast<TgaData::Header*>( mBuf );
    u8 format = hd->bfFormat;

    if ( (format & 0xF0) != 0 )
    {
        return false;
    }
    switch ( format & 0x07 )
    {
    case TgaData::FMT_CI8:
    case TgaData::FMT_RGB:
    case TgaData::FMT_I8:
        break;
    default:
        return false;
    }

    if ( hd->bfType == 1 )
    {
        if ( format != TgaData::FMT_CI8 && format != TgaData::FMT_CMP_CI8 )
        {
            return false;
        }
        if ( LEndian::ReadU16( &hd->bfPlttLength ) == 0 )
        {
            return false;
        }
        if ( hd->bfPlttSize != 15 && hd->bfPlttSize != 16 &&
             hd->bfPlttSize != 24 && hd->bfPlttSize != 32 )
        {
            return false;
        }
    }
    else if ( hd->bfType == 0 )
    {
        if ( format == TgaData::FMT_CI8 || format == TgaData::FMT_CMP_CI8 )
        {
            return false;
        }
        if ( LEndian::ReadU16( &hd->bfPlttLength ) != 0 )
        {
            return false;
        }
    }
    else
    {
        return false;
    }

    u16 width  = LEndian::ReadU16( &hd->bfWidth  );
    u16 height = LEndian::ReadU16( &hd->bfHeight );
    if ( width == 0 || height == 0 )
    {
        return false;
    }
    u8 bpp = hd->bfBpp;
    if ( bpp != 8 && bpp != 16 && bpp != 24 && bpp != 32 )
    {
        return false;
    }

    return true;
}

/*---------------------------------------------------------------------------*
  Name:         TgaReaderImpl::Dispose()
  Description:  Free allocated memory for internal processing.
  Arguments:    data
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TgaReaderImpl::Dispose()
{
    /* DO_NOTHING */
}


}   /* namespace system */
}   /* namespace vw     */


