/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system.h>


namespace vw     {
namespace system {



static nn::hid::CTR::TouchPanelStatus sTouchPanel;
static TouchPanelInfo                 sBuf;   // Internal use only

static struct
{
    u16    x;
    u16    y;
    u16    startX;
    u16    startY;
    u8     press;
    u8     press_plus;
    u8     prev;
    u8     trig;
    u8     release;
    u8     repeat;
    u8     repeatAccel;        // accelerated repeat
    u8     click;
    u8     click_ignore;
    u8     hold;
}
sPanelStatus;

static struct {
    u8   hold;
    u8   hold_dbl;
} sClickState;

/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsEnable
  Description:  Function to get whether the current pad is enabled for the specified appId
  Arguments:    appId   Module ID
  Returns:      if pad operations are enabled, true; if operation focus is on another module or the system, false.
                
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsEnable( u16 appId )
{
    if ( appId == SYSTEM_ID )
    {
        return true;
    }

    //if ( App::GetInstance().HasFocus( appId )  (! GetWindowManager()-GetNextWindow( NULL ) ) )
    if ( App::GetInstance().HasFocus( appId ) )
    {
        return true;
    }
    return false;
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_Init
  Description:  Initializes PAD state.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_Init( void )
{
    TouchPanel_Read();
    TouchPanel_Read();
    sPanelStatus.click_ignore             = 0;
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_ReadRepeat_
  Description:  Performs a key repeat count.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
TouchPanel_ReadRepeat_( void )
{
    const u16 REPEAT_START = 25;    // Number of frames until key repeat starts
    const u16 REPEAT_SPAN  = 8;     // Number of frames between key repeats
    static u16      repeat_count;
    u16             i;

    sPanelStatus.repeat = sPanelStatus.trig;

    // Key repeat calculations
    for ( i = 0; i < 1; i++ )
    {
        if ( sPanelStatus.press & ( 1 << i ) )
        {
            if ( (--repeat_count) == 0 )
            {
                sPanelStatus.repeat |= ( 1 << i );
                repeat_count = REPEAT_SPAN;
            }
        }
        else
        {
            repeat_count = REPEAT_START;
        }
    }

}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_ReadRepeatAccel_
  Description:  Performs an accelerated key repeat count.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
TouchPanel_ReadRepeatAccel_( void )
{
    const  s32 BUTTON_NUM   = 1;
    const  s32 REPEAT_START = 25;    // Number of frames until key repeat starts
    const  s32 REPEAT_SPAN  = 8;     // Number of frames between key repeats
    static s32 repeat_count[ BUTTON_NUM ];
    static s32 repeat_trig_count[ BUTTON_NUM ];
    u16             i;

    sPanelStatus.repeatAccel = sPanelStatus.trig;

    // Key repeat calculations
    for ( i = 0; i < BUTTON_NUM; i++ )
    {
        if ( sPanelStatus.press & ( 1 << i ) )
        {
            if ( (--repeat_count[ i ]) == 0 )
            {
                sPanelStatus.repeatAccel |= ( 1 << i );
                repeat_count[ i ] = nw::ut::Max<u32>( REPEAT_SPAN - (repeat_trig_count[ i ] >> 1), 1L );
            }
        }
        else
        {
            repeat_count[ i ] = REPEAT_START;
        }

        // Key repeat count
        if ( sPanelStatus.repeatAccel & ( 1 << i ) )
        {
            ++repeat_trig_count[ i ];
        }

        if ( ! (sPanelStatus.press & ( 1 << i ))  )
        {
            repeat_trig_count[ i ] = 0;
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_ReadClick_
  Description:  Performs a count of click and hold determinations.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
TouchPanel_ReadClick_( void )
{
    const  s32  BUTTON_NUM   = 1;
    const  s32  HOLD_COUNT   = 90;    // Number of frames until key repeat starts
    const  s32  CLICK_COUNT  = 20;    // Number of frames between key repeats

    sPanelStatus.click = 0;
    sPanelStatus.hold  = 0;

    for ( s32 i = 0; i < BUTTON_NUM; i++ )
    {
        u16 i_mask = u16( 1 << i );

        if ( sPanelStatus.press & i_mask )
        {
            ++sClickState.hold;

            if ( sClickState.hold == HOLD_COUNT )
            {
                sPanelStatus.hold |= i_mask;
            }
        }
        else
        {
            s32 count = sClickState.hold;
            sClickState.hold = 0;

            if ( count > 0           &&
                 count < CLICK_COUNT &&
                 (! (sPanelStatus.click_ignore & i_mask) ) )
            {
                sPanelStatus.click |= i_mask;
            }

            sPanelStatus.click_ignore &= ~i_mask;
        }
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_Dragged
  Description:  Touch panel drag processing
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_Dragged_()
{
    if ( sPanelStatus.trig != 0 )
    {
        sPanelStatus.startX = sPanelStatus.x;
        sPanelStatus.startY = sPanelStatus.y;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_Read_
  Description:  Touch panel pad input
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_Read_(nn::hid::CTR::TouchPanelStatus* status)
{
    static nn::hid::CTR::TouchPanelReader touchPanelReader;
    touchPanelReader.ReadLatest(status);

    sBuf.x = status->x;
    sBuf.y = status->y;
    sBuf.touch = status->touch;
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_Read
  Description:  Gets the key information.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_Read( void )
{
    TouchPanel_Read_(&sTouchPanel);

    sPanelStatus.prev    = sPanelStatus.press;
    sPanelStatus.trig    = static_cast<u8>( (sPanelStatus.press ^ sBuf.touch) & sBuf.touch );
    sPanelStatus.release = static_cast<u8>( (sPanelStatus.press ^ sBuf.touch) & sPanelStatus.press );
    sPanelStatus.press   = sTouchPanel.touch;
    sPanelStatus.x       = sTouchPanel.x;
    sPanelStatus.y       = sTouchPanel.y;

    // Make the Repeat determination
    TouchPanel_ReadRepeat_();

    // Accelerated repeat determination
    TouchPanel_ReadRepeatAccel_();

    // Click, hold determination
    TouchPanel_ReadClick_();

    // Drag processing
    TouchPanel_Dragged_();
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_GetKeyInfo
  Description:  Gets key information as the TouchPanelInfo structure.
  Arguments:    appId   Module ID
                keyInfo     Region for getting the key information
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_GetInfo( u16 appId, TouchPanelInfo* panelInfo )
{
    const TouchPanelInfo PANEL_INFO_NULL =
    {
        0, /* x */
        0, /* y */
        0, /* startX */
        0, /* startY */
        0, /* touch */
        0, /* press */
        0, /* trig */
        0, /* Release  */
        0, /* repeat */
        0, /* repeatAccel */
        1, /* disabled */
    };
    if ( Pad_IsEnable( appId ) )
    {
        panelInfo->x           = sPanelStatus.x;
        panelInfo->y           = sPanelStatus.y;
        panelInfo->startX      = sPanelStatus.startX;
        panelInfo->startY      = sPanelStatus.startY;
        panelInfo->touch       = sPanelStatus.press;
        panelInfo->press       = sPanelStatus.press;
        panelInfo->trig        = sPanelStatus.trig;
        panelInfo->release     = sPanelStatus.release;
        panelInfo->repeat      = sPanelStatus.repeat;
        panelInfo->repeatAccel = sPanelStatus.repeatAccel;
        panelInfo->disabled    = 0;
    }
    else
    {
        *panelInfo = PANEL_INFO_NULL;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsPressed
  Description:  Determines whether the pad is being pressed.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key currently being pressed and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsPressed( u16 appId )
{
    if ( TouchPanel_IsEnable( appId ) )
    {
        return ( sPanelStatus.press != 0 );
    }
    else
    {
        return false;
    }
}

/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsPressedHold
  Description:  Determines whether the pad is being pressed. (Diagonal is invalid)
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key currently being pressed and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsPressedHold( u16 appId )
{
    if ( TouchPanel_IsEnable( appId ) )
    {
        return ( sPanelStatus.press_plus != 0 );
    }
    else
    {
        return false;
    }
}

/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsTriggered
  Description:  Determines whether the pad trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current trigger key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsTriggered( u16 appId )
{
    if ( TouchPanel_IsEnable( appId )  )
    {
        return ( sPanelStatus.trig != 0);
    }
    else
    {
        return false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsReleased
  Description:  Determines whether the pad up trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current up trigger key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsReleased( u16 appId )
{
    if ( TouchPanel_IsEnable( appId )  )
    {
        return ( sPanelStatus.release != 0 );
    }
    else
    {
        return false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsRepeated
  Description:  Determines whether the pad repeat trigger has occurred.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the current repeat trigger key and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsRepeated( u16 appId )
{
    if ( TouchPanel_IsEnable( appId )  )
    {
        return ( sPanelStatus.repeat != 0 );
    }
    else
    {
        return false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsClicked
  Description:  Determines pad single clicks.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key for which the click has been determined and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsClicked( u16 appId )
{
    if ( TouchPanel_IsEnable( appId ) )
    {
        return ( sPanelStatus.click != 0 );
    }
    else
    {
        return false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_IsLongPressed
  Description:  Determines pad long presses.
  Arguments:    appId   Module ID
                button  Key for determination
  Returns:      Returns the AND value of the key for which the long press has been determined and the determination passed with an argument.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_IsDragged( u16 appId )
{
    if ( TouchPanel_IsEnable( appId ) )
    {
        return ( sPanelStatus.press != 0 );
    }
    else
    {
        return false;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_GetXY
  Description:  Gets the main stick analog value.
  Arguments:    appId   Module ID
                stickX  Value of the X-component for the main stick
                stickY  Value of the Y-component for the main stick
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_GetXY( u16 appId, u16* posX, u16* posY )
{
    if ( TouchPanel_IsEnable( appId )  )
    {
        *posX = sPanelStatus.x;
        *posY = sPanelStatus.y;
    }
    else
    {
        *posX = 0;
        *posY = 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_GetStartXY
  Description:  Gets the main stick analog value.
  Arguments:    appId   Module ID
                stickX  Value of the X-component for the main stick
                stickY  Value of the Y-component for the main stick
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
TouchPanel_GetStartXY( u16 appId, u16* posX, u16* posY )
{
    if ( TouchPanel_IsEnable( appId )  )
    {
        *posX = sPanelStatus.startX;
        *posY = sPanelStatus.startY;
    }
    else
    {
        *posX = 0;
        *posY = 0;
    }
}


/*---------------------------------------------------------------------------*
  Name:         TouchPanel_GetRawInfo_
  Description:
  Arguments:    pad
                port
  Returns:      None.
 *---------------------------------------------------------------------------


*/
bool
TouchPanel_GetRawInfo_( nn::hid::CTR::TouchPanelStatus* status )
{
    if ( ! status )
    {
        return false;
    }

    *status = sTouchPanel;

    return true;
}



}   /* namespace system */
}   /* namespace vw     */


