/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/system/Video.h>
#include <viewer/system/VTask.h>

namespace vw     {
namespace system {


static void VTask_Main_();

static nn::os::Thread               sVTaskThread;               // Task thread
static nn::os::StackBuffer<4096>    sVTaskThreadStack;          // Thread stack

static nn::os::CriticalSection      s_CriticalSection;

typedef nw::ut::LinkList<VTaskObj,  offsetof(VTaskObj,m_Link)> VTaskObjList;
static VTaskObjList                 sVTaskList;                 // Task list
static bool                         sActive = false;


/*---------------------------------------------------------------------------*
  Name:         VTask_Init
  Description:  Initializes VTask and creates a thread.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
VTask_Init( void )
{
    // Critical session initialization
    s_CriticalSection.Initialize();

    sActive = true;
}


/*---------------------------------------------------------------------------*
  Name:         VTask_Start
  Description:  Starts VTask.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
VTask_Start( void )
{
    sVTaskThread.Start(VTask_Main_, sVTaskThreadStack, nn::os::DEFAULT_THREAD_PRIORITY );
}


/*---------------------------------------------------------------------------*
  Name:         VTask_Finalize
  Description:  Finalizes VTask and deletes the thread.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
VTask_Finalize( void )
{
    // Wait for thread to end.
    sActive = false;
    sVTaskThread.Join();

    // Destroy critical session
    s_CriticalSection.Finalize();

    // Destroys a thread object.
    sVTaskThread.Finalize();

    // Delete task
    sVTaskList.Clear();
}


/*---------------------------------------------------------------------------*
  Name:         VTask_Main_
  Description:  Thread main function.
                After calling the registered VTask functions in order, sleeps until the next VIPostRetrace.
                
  Arguments:    param   Parameter arguments (unused)
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
VTask_Main_()
{
    while ( sActive )
    {
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(1000 / 60));

        s_CriticalSection.Enter();
        for(VTaskObjList::iterator i=sVTaskList.begin(); i!=sVTaskList.end(); i++)
        {
            i->func( i->arg );
        }

        s_CriticalSection.Leave();
    }
}

#if defined( NN_PCGX )
void
pcgx_VTask_Main_()
{
    if (sActive)
    {
        for (VTaskObjList::iterator i = sVTaskList.begin(); i != sVTaskList.end(); i++)
        {
            i->func( i->arg );
        }
    }
}
#endif


/*---------------------------------------------------------------------------*
  Name:         VTask_AddObj
  Description:  Adds VTask objects
  Arguments:    task    VTask object to add
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
VTask_AddObj( VTaskObj* task )
{
    s_CriticalSection.Enter();

    sVTaskList.push_back( task );

    s_CriticalSection.Leave();
}


/*---------------------------------------------------------------------------*
  Name:         VTask_RemoveObj
  Description:  Deletes VTask objects
  Arguments:    task    VTask object to delete
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
VTask_RemoveObj( VTaskObj* task )
{
    s_CriticalSection.Enter();

    sVTaskList.erase( task );

    s_CriticalSection.Leave();
}


}   /* namespace system */
}   /* namespace vw     */


