/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/
#include <viewer/system.h>
#include "../include/LogUtil.h"


namespace vw     {
namespace system {
namespace Memory {



static nn::fnd::ThreadSafeExpHeap    s_MainMemoryHeap;
static nn::fnd::ThreadSafeExpHeap    s_DeviceHeap;
static bool                s_IsInitialized = false;
static u32                 sAllocateCount = 0;
static u32                 sDeviceAllocateCount = 0;




//--------------------------------------------------------------------------
void InitializeMemory(size_t mainMemorySize = 0 )
{
    if ( s_IsInitialized )
    {
        return;
    }
    s_IsInitialized = true;

    const size_t DEFAULT_DEMO_MEMORY_SIZE = 0x2000000;  // 32MB
    nn::os::SetDeviceMemorySize(DEFAULT_DEMO_MEMORY_SIZE);

    uptr address = nn::os::GetHeapAddress();
    size_t size  = nn::os::GetHeapSize();

    s_MainMemoryHeap.Initialize(address, size);
    nwosPrintf("################# Main Memory : 0x%x - 0x%x #################\n", (uint)address, ((uint)address + size));

    s_DeviceHeap.Initialize(
        nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize());
    nwosPrintf(
        "################# Device Memory : 0x%x - 0x%x #################\n",
        nn::os::GetDeviceMemoryAddress(),
        nn::os::GetDeviceMemoryAddress() + nn::os::GetDeviceMemorySize()
    );
}

bool IsDeviceMemory(const void* memory)
{
    if (nn::os::GetDeviceMemoryAddress() <= (uint)memory &&
        (uint)memory <= nn::os::GetDeviceMemoryAddress() + nn::os::GetDeviceMemorySize())
    {
        return true;
    }
    else
    {
        return false;
    }
}

//--------------------------------------------------------------------------
void Dump()
{
    s_MainMemoryHeap.Dump();
}

//--------------------------------------------------------------------------
void DumpDeviceMemory()
{
    s_DeviceHeap.Dump();
}


namespace internal
{

//--------------------------------------------------------------------------
void* AllocInternal_(size_t size)
{
    InitializeMemory(); // When uninitialized, initialize with default settings.

    void* memory = s_MainMemoryHeap.Allocate(size);
    return memory;
}

//--------------------------------------------------------------------------
void* AllocDeviceMemoryInternal_(size_t size)
{
    InitializeMemory(); // When uninitialized, initialize with default settings.

    void* memory = s_DeviceHeap.Allocate(size);
    return memory;
}

//--------------------------------------------------------------------------
void FreeInternal_(void* memory)
{
    if (IsDeviceMemory(memory))
    {
        s_DeviceHeap.Free(memory);
    }
    else
    {
        s_MainMemoryHeap.Free(memory);
    }
}

} // namespace internal

/*
  
  
  
  
 

*/
static void*
AlignMemoryInternal_( void* memory, u8 align )
{
    void* top = nw::ut::AddOffsetToPtr( memory, 1 );
    top = nw::ut::RoundUp( top, align );

    u8* head = reinterpret_cast<u8*>( nw::ut::AddOffsetToPtr( top, - 1 ) );

    size_t offset = nw::ut::GetOffsetFromPtr( memory, top );
    NW_U8_RANGE_ASSERT( offset );

    *head = static_cast<u8>( offset );
    return top;
}


/*
  
  
  
 

*/
static void*
UnAlignMemoryInternal_( void* memory )
{
    u8* head = reinterpret_cast<u8*>( nw::ut::AddOffsetToPtr( memory, -1 ) );

    u8 offset = *head;

    return nw::ut::AddOffsetToPtr( memory, -offset );
}

/*
  
  
  
  
 

*/
void* Allocate(size_t size, u8 alignment = 4)
{
    // Check for power of 2
    NW_ASSERT( (alignment & (alignment - 1)) == 0 );

    if ( alignment == 0 ) { alignment = 1; }

    void* memory = internal::AllocInternal_( size + alignment );
    if(0 == memory)
    {
        return 0;
    }
    return AlignMemoryInternal_( memory, alignment );
}

/*
  
  
  
  
 

*/
void* AllocateDeviceMemory(size_t size, u8 alignment = 4)
{
    // Check for power of 2
    NW_ASSERT( (alignment & (alignment - 1)) == 0 );

    if ( alignment == 0 ) { alignment = 1; }

    void* memory = internal::AllocDeviceMemoryInternal_( size + alignment );
    if(0 == memory)
    {
        NW_VIEWER_DEV_LOG("AllocateDeviceMemory [Out of memory].\n");
        return 0;
    }
    return AlignMemoryInternal_( memory, alignment );
}

/*
  
  
                           
 
*/
void Free(void* ptr)
{
    ptr = UnAlignMemoryInternal_( ptr );
    internal::FreeInternal_( ptr );
}




    void* Allocator::Alloc(size_t size, u8 alignment)
    {
        return Memory::Allocate(size, alignment);
    }
    void Allocator::Free(void* memory)
    {
        Memory::Free(memory);
    }
    void Allocator::Dump()
    {
        Memory::Dump();
    }

class DeviceMemoryAllocator : public nw::os::IAllocator
{
    using nw::os::IAllocator::Alloc;
public:
    virtual void* Alloc(size_t size, u8 alignment)
    {
        return Memory::AllocateDeviceMemory(size, alignment);
    }
    virtual void Free(void* memory)
    {
        Memory::Free(memory);
    }
    void Dump()
    {
        Memory::DumpDeviceMemory();
    }
};

static Memory::Allocator             sAllocator;
static Memory::DeviceMemoryAllocator sDeviceMemoryAllocator;

}    /* namespace Memory */







static nw::os::IAllocator* sAllocator = &Memory::sAllocator;
static nw::os::IAllocator* sDeviceMemoryAllocator = &Memory::sDeviceMemoryAllocator;


// Initialize heap.
void Memory_Init( size_t mainMemorySize )
{
    Memory::InitializeMemory( mainMemorySize );
}

// Initialize heap.
void Memory_Init(nw::os::IAllocator* pAllocator, nw::os::IAllocator* pDeviceAllocator)
{
    Memory::s_IsInitialized = true;
    sAllocator = pAllocator;
    sDeviceMemoryAllocator = pDeviceAllocator;
}

// Perform finalization for heap.
void  Memory_Finalize()
{

}

// Allocates memory of the specified size.
void* Memory_Alloc( u16 id, u32 size, u32 align )
{
    return sAllocator->Alloc( size, align );
}

// Allocates device memory of the specified size.
void* Memory_AllocDevice( u16 id, u32 size, u32 align )
{
    return sDeviceMemoryAllocator->Alloc( size, align );
}

// Returns memory obtained from the heap.
void Memory_Free( void* addr )
{
    if ( Memory::IsDeviceMemory( addr ) )
    {
        sDeviceMemoryAllocator->Free( addr );
    }
    else
    {
        sAllocator->Free( addr );
    }
}

// Returns memory obtained from the heap.
void Memory_Free( const void* addr )
{
    Memory_Free(const_cast<void*>(addr));
}

// Checks the amount of heap memory currently available.
u32 Memory_FreeSize( void )
{
    return 0;
}

// Dump and output the heap state to the console.
void Memory_Dump()
{
    nwosPrintf("################# Main Memory #################\n");
    Memory::Dump();
    nwosPrintf("################# Device Memory #################\n");
    Memory::DumpDeviceMemory();
}

// Gets a pointer to the allocator.
nw::os::IAllocator* Memory_GetAllocator( u16 id )
{
    return sAllocator;
}

// Gets a pointer to the allocator for device memory.
nw::os::IAllocator* Memory_GetAllocatorDevice( u16 id )
{
    return sDeviceMemoryAllocator;
}

}   /* namespace system */
}   /* namespace viewer */

// malloc() override especially for the Viewer application.
void* vwmalloc( size_t size )
{
    vw::system::Memory::InitializeMemory();
    void* memory = vw::system::Memory::s_MainMemoryHeap.Allocate(size);
    return memory;
}

// free() override especially for the Viewer application.
void vwfree( void* p )
{
    vw::system::Memory::InitializeMemory();
    vw::system::Memory::s_MainMemoryHeap.Free( p );
}

// calloc() override especially for the Viewer application.
void* vwcalloc( size_t num, size_t size )
{
    size_t sum = num * size;
    void* ptr = vwmalloc( sum );
    if ( ptr != NULL ) { std::memset( ptr, 0, sum ); }
    return ptr;
}

// realloc() override especially for the Viewer application.
void* vwrealloc( void* p, size_t size )
{
    if ( p != NULL )
    {
        if ( size == 0 ) { std::free( p ); return NULL; }

        size_t newSize = vw::system::Memory::s_MainMemoryHeap.ResizeBlock( p, size );
        if ( newSize != 0 ) { return p; }

        void* pNewMem = vwmalloc( size );
        if ( pNewMem != NULL )
        {
            size_t oldSize = vw::system::Memory::s_MainMemoryHeap.GetSizeOf( p );
            std:memcpy( pNewMem, p, std::min( size, oldSize ) );
            vwfree( p );
        }

        return pNewMem;
    }
    else
    { return vwmalloc( size ); }
}
