/*---------------------------------------------------------------------------*
  Project:  NintendoWare

  Copyright (C)Nintendo/HAL Laboratory, Inc.  All rights reserved.

  These coded instructions, statements, and computer programs contain proprietary
  information of Nintendo and/or its licensed developers and are protected by
  national and international copyright laws. They may not be disclosed to third
  parties or copied or duplicated in any form, in whole or in part, without the
  prior written consent of Nintendo.

  The content herein is highly confidential and should be handled accordingly.
 *---------------------------------------------------------------------------*/

#include <viewer/common.h>
#include <viewer/system.h>
#include "../include/system_private.h"
#include "../../g3d/include/AnimManager.h"

extern const char SYS_DATE[];

namespace vw     {
namespace system {


static void     UpdateConfigFile_();
static void     UpdateVisibleInfo_();
static void     UpdateFocusModule_();
static void     DrawMode_();


#define DEFAULT_FIFO_SIZE (256 * 1024)

// System settings
static SystemSetting sSystemSetting;
static bool          sChangeModeEnable;

void SystemSetting::Finalize()
{
    defaultModelPath = vw::common::CharArray();
    reloadTargetPath = vw::common::CharArray();
    captureImagePath = vw::common::CharArray();
}

/*---------------------------------------------------------------------------*
    Name:           Video_ChangeSetting
    Description:    Changes video settings
    Arguments:      setting     Video settings
    Returns:        None.
 *---------------------------------------------------------------------------


*/
void
System_ChangeSetting( const SystemSetting* setting )
{
    sSystemSetting.Finalize();
    sSystemSetting.defaultModelPath.Clear();
    sSystemSetting.defaultModelPath.SetAllocator( Memory_GetAllocator( SYSTEM_ID ) );
    if ( setting->defaultModelPath.Get() ) {
        sSystemSetting.defaultModelPath.Set( setting->defaultModelPath );
    } else {
        sSystemSetting.defaultModelPath.Set( "" );
    }

    sSystemSetting.reloadTargetPath.Clear();
    sSystemSetting.reloadTargetPath.SetAllocator( Memory_GetAllocator( SYSTEM_ID ) );
    if ( setting->reloadTargetPath.Get() ) {
        sSystemSetting.reloadTargetPath.Set( setting->reloadTargetPath );
    } else {
        sSystemSetting.reloadTargetPath.Set( "" );
    }

    sSystemSetting.captureImagePath.Clear();
    sSystemSetting.captureImagePath.SetAllocator( Memory_GetAllocator( SYSTEM_ID ) );
    if ( setting->captureImagePath.Get() ) {
        sSystemSetting.captureImagePath.Set( setting->captureImagePath );
    } else {
        sSystemSetting.captureImagePath.Set( "" );
    }

    sSystemSetting.frameRate = setting->frameRate;
    sSystemSetting.updateRate = setting->updateRate;
    sSystemSetting.frameSkip = setting->frameSkip;
    sSystemSetting.maxFrameSkipCount = setting->maxFrameSkipCount;
    sSystemSetting.lowerPreview = setting->lowerPreview;
}

void
System_GetSetting( SystemSetting* setting )
{
    setting->Finalize();
    setting->defaultModelPath.Clear();
    setting->defaultModelPath.SetAllocator( Memory_GetAllocator( SYSTEM_ID ) );
    if ( sSystemSetting.defaultModelPath.Get() ) {
        setting->defaultModelPath.Set( sSystemSetting.defaultModelPath );
    } else {
        setting->defaultModelPath.Set( "" );
    }

    setting->reloadTargetPath.Clear();
    setting->reloadTargetPath.SetAllocator( Memory_GetAllocator( SYSTEM_ID ) );
    if ( sSystemSetting.reloadTargetPath.Get() ) {
        setting->reloadTargetPath.Set( sSystemSetting.reloadTargetPath );
    } else {
        setting->reloadTargetPath.Set( "" );
    }

    setting->captureImagePath.Clear();
    setting->captureImagePath.SetAllocator( Memory_GetAllocator( SYSTEM_ID ) );
    if ( sSystemSetting.captureImagePath.Get() ) {
        setting->captureImagePath.Set( sSystemSetting.captureImagePath );
    } else {
        setting->captureImagePath.Set( "" );
    }

    setting->frameRate = sSystemSetting.frameRate;
    setting->updateRate = sSystemSetting.updateRate;
    setting->frameSkip = sSystemSetting.frameSkip;
    setting->maxFrameSkipCount = sSystemSetting.maxFrameSkipCount;
    setting->lowerPreview = sSystemSetting.lowerPreview;
}
const SystemSetting* System_GetSetting()
{
    return &sSystemSetting;
}

/*---------------------------------------------------------------------------*
  Name:         Init
  Description:  Initialize system
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Init( void )
{/*
    if( !vw::system::App::GetInstance().IsEmbeddedViewer() )
    {
        // Initialize the heap
        size_t mainMemorySize = 0;
        Memory_Init(mainMemorySize);
        // Mount ROM.
        const int MAX_FILE = 256;
        const int MAX_DIRECTORY = 16;
        s32 workingMemorySize = nn::fs::GetRomRequiredMemorySize(MAX_FILE, MAX_DIRECTORY);
        void* workingMemory = Memory_AllocSys(workingMemorySize);
        nn::fs::MountRom(MAX_FILE, MAX_DIRECTORY, workingMemory, workingMemorySize);
    }
    // host io
    sHioWorkMemory = Memory_AllocSysDevice(nn::hio::CTR::WORKMEMORY_SIZE);
    nn::hio::Initialize( sHioWorkMemory );


*/

    // Pad initialization
    Pad_Init();

    // Initialize log system
    LogSystem::Initialize();

    // MSC initialization
    Stream_Init();

    // V task initialization
    VTask_Init();

    // Load the configuration file
    (void)ConfigFile::Create();

    sChangeModeEnable = true;

    PerformMgr::Initialize();
    ClearImage::Initialize();
}

/*---------------------------------------------------------------------------*
  Name:         Finalize
  Description:  System finalization
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Finalize( void )
{
    // V task initialization
    VTask_Finalize();

    // Destroy configuration file information.
    ConfigFile::Destroy();
    PerformMgr::Finalize();
    CaptureInterface::Finalize();
    LogSystem::Finalize();
    ClearImage::Finalize();
    sSystemSetting.Finalize();

    // Stream finalization
    Stream_UnInit();

    // Video finalization
    Video_Finalize();
}

/*---------------------------------------------------------------------------*
  Name:         Start
  Description:  Starts system operation
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Start( void )
{
    system::VTask_Start();
    //system::BGTask_Start();
}



/*---------------------------------------------------------------------------*
  Name:         PollStream
  Description:  MCS communication processing
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
PollStream( void )
{
//    Stream_Polling();
//
//    ExecCommandSystem();
}

/*---------------------------------------------------------------------------*
  Name:         PreMain
  Description:  System main
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
PreMain( void )
{
}

/*---------------------------------------------------------------------------*
  Name:         Main
  Description:  System main
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Main( void )
{
    UpdateConfigFile_();

    TouchPanel_Read();
    Pad_Read();

    //IconControl::GetInstance().ResetState();

    UpdateVisibleInfo_();

    UpdateFocusModule_();

    //GetWindowManager()->ExecPad();
    //
    //Debug_Update();
}


/*---------------------------------------------------------------------------*
  Name:         Reset
  Description:  System reset
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
Reset( void )
{
//    (void)VISetPreRetraceCallback( NULL );
//    (void)VISetPostRetraceCallback( NULL );
//    VISetBlack( GX_true );
//    VIFlush();
//    VIWaitForRetrace();
//    OSResetSystem( OS_RESET_RESTART, 0, false );
}


/*---------------------------------------------------------------------------*
  Name:         DrawInfo
  Description:  System rendering
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
DrawInfo( void )
{
}


/*---------------------------------------------------------------------------*
  Name:         DrawBG
  Description:  Background rendering
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
DrawBG( void )
{
    if( GetClearImage().IsValid() )
    {
        bool lowerPreview = vw::system::System_GetSetting()->lowerPreview;
        if( lowerPreview )
        {
            vw::system::Video_SetRenderTarget( LOWER_SCREEN );
            vw::system::Video_ClearDepth();
            GetClearImage().Draw( vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight() );

            vw::system::Video_SetRenderTarget( UPPER_SCREEN );
            vw::system::Video_ClearBuffer();

#if !defined( NN_PCGX )
            vw::system::Video_SetRenderTarget( EXTENSION_SCREEN );
            vw::system::Video_ClearBuffer();
#endif
        }
        else
        {
            vw::system::Video_SetRenderTarget( LOWER_SCREEN );
            vw::system::Video_ClearBuffer();

            vw::system::Video_SetRenderTarget( UPPER_SCREEN );
            vw::system::Video_ClearDepth();
            GetClearImage().Draw( vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight() );

#if !defined( NN_PCGX )
            vw::system::Video_SetRenderTarget( EXTENSION_SCREEN );
            vw::system::Video_ClearDepth();
            GetClearImage().Draw( vw::system::Video_GetScreenWidth(), vw::system::Video_GetScreenHeight() );
#endif
        }
    }
    else
    {
        vw::system::Video_ClearBuffers();
    }

#if defined( NN_PCGX )
    nngxSplitDrawCmdlist();
#endif

    Video_CalcFPS();
}

/*---------------------------------------------------------------------------*
  Name:         RestoreChangeMode
  Description:  Sets the mode change to enabled/disabled
  Arguments:    None.
  Returns:      Returns the value before setting.
 *---------------------------------------------------------------------------


*/
bool
RestoreChangeMode( bool enabled )
{
//    bool before = sChangeModeEnable;
//    sChangeModeEnable = enabled;
//    return before;
    return true;
}


/*---------------------------------------------------------------------------*
  Name:         WriteBackConfigFile
  Description:  Writes back current settings to the instance of the Config class.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
void
WriteBackConfigFile()
{
//    // Write back current settings to the Config structure
//    ConfigSys_WriteBack();
}


/*---------------------------------------------------------------------------*
  Name:         SetPointerType
  Description:  Switches the pointer display
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------*/
void
SetPointerType( u32 icon_type )
{
//    App::GetUIManager().GetCursor().SetIconType( icon_type );
}


/*---------------------------------------------------------------------------*
  Name:         DrawMode_
  Description:  Displays mode information.
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
DrawMode_()
{
//    static const char*
//    MODULE_NAME[ MODULE_NUM ] =
//    {
//        "System Mode",
//        "3D Mode",
//        "Effect Mode",
//        "Layout Mode",
//        "Sound Mode",
//#if defined(USE_SAMPLE)
//        "SMP",
//#endif
//    };
//
//    const s16 MODE_POS_Y  = 32;
//    const s16 MODE_POS_X  = 192;
//    const s16 MODE_SIZE_X = 256;
//    const s16 MODE_SIZE_Y = 32;
//
//    Canvas canvas;
//    canvas.SetupGX();
//    canvas.SetupCamera();
//
//    GXColor FG_COLOR = { 0x00, 0x00, 0x00, 0xFF };
//    GXColor BG_COLOR = { 0xAF, 0xAF, 0xAF, 0xFF };
//
//    f32 alpha = 2.0f * ( App::GetInstance().GetDrawIconCount() ) / 60.0f;
//    alpha = alpha > 1.0f ? 1.0f : alpha;
//
//    BG_COLOR.a = u8( ( 0xFF * alpha ) );
//    FG_COLOR.a = u8( ( 0xFF * alpha ) );
//
//    canvas.FillRect( MODE_POS_X, MODE_POS_Y, MODE_SIZE_X, MODE_SIZE_Y, BG_COLOR );
//
//    ut::Color col( FG_COLOR );
//
//    TextWriter_GetInstance()->SetDrawFlag(
//            ut::TextWriter::HORIZONTAL_ALIGN_CENTER  |
//            ut::TextWriter::HORIZONTAL_ORIGIN_CENTER |
//            ut::TextWriter::VERTICAL_ORIGIN_MIDDLE  );
//    TextWriter_GetInstance()->SetTextColor( col );
//    TextWriter_GetInstance()->SetScale( 0.75f );
//
//    (void)TextWriter_Printf( MODE_POS_X + MODE_SIZE_X / 2, MODE_POS_Y + MODE_SIZE_Y / 2, MODULE_NAME[ App::GetInstance().GetFocusID() ] );
}


/*---------------------------------------------------------------------------*
  Name:         UpdateConfigFile
  Description:  Update check of configuration file
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
UpdateConfigFile_()
{
    ConfigFile::UpdateReloaded();

    if ( ConfigFile::IsReloaded() )
    {
        ConfigSys_Execute();
    }

    //// NOTE: Reload configuration file with the reset button
    ////       Temporary process until a proper reset is implemented
    //static bool sResetButton = false;
    //bool before = sResetButton;
    //sResetButton = OSGetResetButtonState();
    //if ( before  ( ! sResetButton ) )
    //{
    //    GetConfigFile()-Reload( NULL );
    //}
}


/*---------------------------------------------------------------------------*
  Name:         UpdateVisibleInfo_
  Description:  Updates the show/hide state for the information
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
UpdateVisibleInfo_()
{
    //if ( ! sChangeModeEnable )
    //{
    //    return;
    //}
    //if ( App::GetInstance().GetFocusModule().IsIgnoringChangeVisible() )
    //{
    //    return;
    //}
    //
    //if ( Pad_IsTriggered( SYSTEM_ID, PAD_TRIGGER_R ) || WPad_IsTriggered( SYSTEM_ID, KPAD_BUTTON_2 ) )
    //{
    //    (void)App::GetInstance().ChangeVisibleInfo( ! App::GetInstance().IsVisibleInfo() );
    //}
}


/*---------------------------------------------------------------------------*
  Name:         UpdateFocusModule_
  Description:  Updates module with focus
  Arguments:    None.
  Returns:      None.
 *---------------------------------------------------------------------------


*/
static void
UpdateFocusModule_()
{
    if ( !sChangeModeEnable )
    {
        return;
    }

#if defined( NN_PCGX )
    if (App::GetInstance().GetFocusID() != SYSTEM_ID)
    {
        if ( App::GetInstance().GetFocusModule().IsIgnoringChangeMode() )
        {
            return;
        }
    }
#else
    if ( App::GetInstance().GetFocusModule().IsIgnoringChangeMode() )
    {
        return;
    }
#endif

    // Get current pad status
    s32 modeSelect = 0;
    HidStatus status = system::App::GetInstance().GetHidStatus();
    switch( system::GetPadEvent( status, system::GetSystemPadBindings() ) )
    {
    case system::SYS_SWITCH_MODE_PREV:
        modeSelect = -1;
        break;
    case system::SYS_SWITCH_MODE_NEXT:
        modeSelect = +1;
        break;
    default:
        break;
    }

    const u16 MODULES_NEXT[] =
    {
        G3D_ID,
        EFFECT_ID,
        G2D_ID,
        SOUND_ID,
        TEMPLATE_ID
    };
    if(0 != modeSelect)
    {
        App& app  = App::GetInstance();
        u16  focus = app.GetFocusID();

        while ( 1 )
        {
            switch(focus)
            {
            case G3D_ID:      focus=0; break;
            case EFFECT_ID:   focus=1; break;
            case G2D_ID:      focus=2; break;
            case SOUND_ID:    focus=3; break;
            case TEMPLATE_ID: focus=4; break;
            }
            focus = MODULES_NEXT[(focus+modeSelect+5) % 5];
            Module* module = Module::GetModuleById( focus );

            if ( module == NULL                 ) { continue; }
            if (!module->Available()            ) { continue; }
            if ( module->IsHidden()             ) { continue; }
            if ( module->IsIgnoringChangeMode() ) { continue; }

            break;
        }

        app.SetFocusID( focus );
    }
}

}   /* namespace system */
}   /* namespace vw     */
